
/**
 * @module internal_plugins
 */

/**
 * Common 'markup' RJS plugin.
 *
 * The markup plugin is the way to use a templating system (specifically, _.template) in
 *  UP.  This document will cover [example usage](#examples) and
 *  [why this plugin was created](#why).
 *
 * <a id="examples"></a>
 * Examples
 * --------
 *
 * If you prefer seeing actual changes made in our app to use the markup plugin, checkout
 *  [this](http://smartbear.us.oracle.com:8080/ui#file:review=29192/version=519629)
 *  and [this](http://smartbear.us.oracle.com:8080/ui#file:review=29170/version=519199).
 *  The following section represents guided examples to the functionality that markup!
 *  provides.
 *
 * So, let's go through some examples.  We'll start with what one would do by default (and,
 *  at time of writing, is currently being done in our codebase):
 *
 * **Deprecated Example**
 *
 *     // path/to/my/markup.htm:
 *     <div>Hello <%= name %>!</div>
 *     // myCode.js:
 *     define(['lodash', 'text!path/to/my/markup.htm'], function (_, markupText) {
 *         // This is now deprecated!  Don't do this, use 'markup!' instead!
 *         var markupFn = _.template(markupText);
 *         console.log(markupFn({
 *             name: 'Bob'
 *         }));                 // prints '<div>Hello Bob!</div>'
 *     });
 *
 * **Basic Usage**
 *
 * To convert the above example to use the markup plugin, we would use the following:
 *
 *     // path/to/my/markup.htm:
 *     <div>Hello <%= name %>!</div>
 *     // myCode.js:
 *     define(['markup!path/to/my/markup.htm'], function (markupFn) {
 *         console.log(markupFn({
 *             name: 'Bob'
 *         }));                 // prints '<div>Hello Bob!</div>'
 *     });
 *
 * In the above snippet, we're asking the markup plugin to load a file,
 *  'path/to/my/markup.htm'.  What markup is going to return is the function made by
 *  _.template.
 *
 * **Load & Pluck**
 *
 * Above shows the basic usage.  Optionally, if you include a : in the file import, it will
 *  pluck the specified portion of your template (similar to our old 'pluckMarkup' function).
 *  Let's take a look at an example:
 *
 *     // my/markup-combo.htm:
 *     <script class="hello-world">
 *         <div>Hello <%= name %>!</div>
 *     </script>
 *     <script class="hello-josh">
 *         <div>Hi Josh...</div>
 *     </script>
 *     // myCode.js:
 *     define(['markup!my/markup.htm:.hello-world', 'markup!my/markup.htm:.hello-josh'],
 *         function (helloWorld, helloJosh) {
 *
 *             console.log(helloWorld({
 *                 name: 'Bob'
 *             }));                   // prints '<div>Hello Bob!</div>'
 *
 *             console.log(helloJosh());
 *                                    // prints '<div>Hi Josh...</div>'
 *     });
 *
 * In my markup file, I defined two templates, not one.  Then, in my JS, I used the markup
 *  plugin and one of two filters: '.hello-world' '.hello-josh' to determine which template
 *  I wanted to use.  (Please note the '.' character! the markup plugin just uses jQuery to
 *  figure out which script to use)
 *
 * <a id="why"></a>
 * Why?
 * ----
 *
 * In brief, Content Security Policy (CSP); see
 *  [this Redmine issue](http://10.140.183.183/redmine/issues/2648) for more information.
 *
 * To elaborate just a little: CSP bans use of inline scripts and eval (or similar JS
 *  functionality, like new Function('string')).  \_.template makes use of
 *  new Function('string').  Therefore, to use \_.template in our codebase, it is necessary
 *  to precompile templates (as part of 'grunt migrate --optimize=true') as part of the
 *  build process.  markup! does just that.
 *
 * @class markup!
 * @namespace plugins
 */

define('markup', ['jquery', 'lodash'],

        function($, _) {

    'use strict';

    var buildMap = {};
    var isBuild = false;

    function getText(fileName, url, callback, errback) {
        if (!isBuild) {
            require(['text!' + fileName], callback);
        }
        else {
            try {
                var fs = require.nodeRequire('fs');
                var file = fs.readFileSync(url, 'utf8');
                //Remove BOM (Byte Mark Order) from utf8 files if it is there.
                if (file.indexOf('\uFEFF') === 0) {
                    file = file.substring(1);
                }
                callback(file);
            }
            catch (e) {
                errback(e);
            }
        }
    }

    function template(text) {
        var underscore = _;
        if (isBuild) {
            underscore = require.nodeRequire('lodash');
        }
        return underscore.template(text.trim());
    }

    function filter(text, filterOn) {
        if (isBuild) {
            var cheerio = require.nodeRequire('cheerio');
            var jq = cheerio.load(text);
            return jq('script' + filterOn).html().trim();
        }
        else {
            return $(text).filter(filterOn).html().trim();
        }
    }


    return {

        load : function(name, req, onLoad, config) {
            var fileName = name.split(':')[0];
            var optionalFilter = name.split(':')[1];

            if (config.isBuild && req.toUrl(fileName).match(/empty:/)) {
                /*global console:true*/
                console.log('Warning: Ignoring markup! import: ' + name);
                return onLoad();
            }
            else if (config.isBuild) {
                isBuild = true;
            }

            getText(fileName, req.toUrl(fileName), function (text) {
                var usableText = text;

                if (optionalFilter) {
                    usableText = filter(usableText, optionalFilter);
                }

                buildMap[name] = usableText;
                onLoad(template(usableText));
            }, function (e) {
                onLoad.error(e);
            });
        },

        write : function(pluginName, moduleName, write) {
            if(buildMap[moduleName]) {
                var source = template(buildMap[moduleName]).source;
                write('define(\'' + pluginName + '!' + moduleName  +
                                    '\', function () { return ' + source + ';});\n');
            }
        }

    };
});

