//!
//! Built on: Fri Nov 14 2014 08:35:45
//! Library version: 2.10.0
//! Branch: master  |  Commit: 32f8709
//!
/*global console:true*/
/*jshint quotmark:false*/
define('primutils.markup',['primutils.jquery', 'primutils.lodash', 'primutils.handlebars', 'primutils.logger'
], function($, _, Handlebars, logger) {
    

    if (require.nodeRequire) {
        Handlebars = Handlebars || require.nodeRequire('handlebars');
        _ = _ || require.nodeRequire('lodash');
        logger = {
            warn : console.log.bind(console)
        };
    }

    var VALID_HANDLEBARS_TYPES = ['helper', 'partial'];

    var buildMap = {},
        fullError = false; // Have we shown the full parse error?

    var markupPlugin = {};

    function doesMask(type, name, dep) {
        if (type === 'helper') {
            return _.has(Handlebars.helpers, name) && Handlebars.helpers[name] !== dep;
        } else if (type === 'partial') {
            return _.has(Handlebars.partials, name) && Handlebars.partials[name] !== dep;
        }
        return false;
    }

    /**
     * Provides several utility functions to the markup! plugin.
     * This helps reduce the size of compiled templates.
     *
     * The methods in this object should be considered private and internal to primutils and the markup! plugin.
     *
     * @class templates.utils.markup
     * @private
     */

    /**
     * Takes an array of loaded dependencies and converts it into an object that can be passed to Handlebars to inject local helpers and partials into a template.
     *
     * Only dependencies that fit the profile of a primutils Handlebars helper or template will be injected into the template.
     * This can potentially lead to situations where a dependency is listed but does not work as expected with the markup! plugin.
     * For best compatibility all locally injected helpers and partials should follow the template outlined in <a href="templates.helpers.theory.html">Helper Theory</a>.
     *
     * @method buildImported
     * @param dependencies {Array} Array of dependencies already loaded by require
     * @param [fileName] {String} The filename of the template which loaded the dependencies (for reporting errors)
     * @returns {Object} Collection of helpers and partials which can be injected into a Handlebars template
     */
    markupPlugin.buildImported = function (dependencies, fileName) {
        fileName = fileName || 'file';
        var imports = _.chain(dependencies)
            .map(function (dep) { // Allow for exporting multiple helpers per file (as named object)
                if (_.isObject(dep) && !_.isArray(dep) && !_.isFunction(dep)) {
                    return _.values(dep);
                }
                return dep;
            })
            .flatten() // Allow for exporting multiple helpers per file (in array)
            .filter(function (dep) {
                var isValid = (_.isFunction(dep) && _.has(dep, 'handlebarsType') && _.contains(VALID_HANDLEBARS_TYPES, dep.handlebarsType) && _.has(dep, 'handlebarsName'));
                if (!isValid) {
                    logger.warn('Markup Plugin: In ' + fileName + ', dependency was not a known primutils handlebars dependency. This may be recoverable.');
                }
                return isValid;
            })
            .groupBy(function (dep) {
                return dep.handlebarsType;
            })
            .reduce(function (res, deps, type) {
                res[type] = _.reduce(deps, function (depRes, dep) {
                    if (doesMask(type, dep.handlebarsName, dep)) {
                        logger.warn('Markup Plugin: In ' + fileName + ', "' + dep.handlebarsName + '" masks globally registered ' + type + '.');
                    }
                    depRes[dep.handlebarsName] = dep;
                    return depRes;
                }, {});
                return res;
            }, {})
            .value();

        return {
            helpers: imports.helper,
            partials: imports.partial
        };
    };

    /**
     * Converts a collection of helpers into a list that can be passed to the Handlers `knownHelpers` option.
     *
     * @method buildKnownHelpers
     * @param helpers {Object} Collection of key:helper name to value:helper function
     * @returns {Object} Collection of key:helper name to value:true
     */
    markupPlugin.buildKnownHelpers = function (helpers) {
        return _.reduce(helpers, function (res, helper, helperName) {
            res[helperName] = true;
            return res;
        }, {});
    };

    /**
     * Convenience method for injecting known helpers and partials from the markup! plugin into the options provided by the page author using the template.
     * A clone of the original options is created to prevent leaking imported helpers and partials from the markup! plugin back to the page.
     *
     * If the page author imports helpers or partials with the same name as those from `buildImported`, the author provided ones will mask the `buildImported` ones.
     *
     * @method extendOptions
     * @param options {Object} Page author provided options to the template
     * @param imports {Object} Object created with `buildImported`
     * @returns {Object} Clone of page author provided objects with helpers and partials injected from `buildImported`
     */
    markupPlugin.extendOptions =  function (options, imports) {
        var result = _.clone(options);
        if (imports) {
            result.helpers = _.defaults({}, options.helpers, imports.helpers);
            result.partials = _.defaults({}, options.partials, imports.partials);
        }
        return result;
    };

    function getText(fileName, url, isBuild, callback, errback) {
        if (!isBuild) {
            require(['primutils.text!' + fileName], callback, errback);
        }
        else {
            try {
                var fs = require.nodeRequire('fs');
                var file = fs.readFileSync(url, 'utf8');
                //Remove BOM (Byte Mark Order) from utf8 files if it is there.
                if (file.indexOf('\uFEFF') === 0) {
                    file = file.substring(1);
                }
                callback(file);
            }
            catch (e) {
                errback(e);
            }
        }
    }

    function inlineTemplate(text, importedHelpers, knownHelpersOnly, noEscape) {
        var knownHelpers = importedHelpers ? markupPlugin.buildKnownHelpers(importedHelpers.helpers) : {},
            func = Handlebars.compile(text.trim(), {
                knownHelpers : knownHelpers,
                knownHelpersOnly : !!knownHelpersOnly,
                noEscape : noEscape
            });

        return function (context, options) {
            options = markupPlugin.extendOptions(options || {}, importedHelpers);
            return func(context, options).trim(); // Trim is for jQuery -- must start with "<"
        };
    }

    function compileTemplate(text, knownHelpers, knownHelpersOnly, noEscape) {
        return Handlebars.precompile(text.trim(), {
            knownHelpers : knownHelpers,
            knownHelpersOnly : !!knownHelpersOnly,
            noEscape : noEscape
        });
    }

    function filter(text, filterOn, isBuild, origFilename) {
        if (!isBuild) {
            var lFilteredText = $(text).filter(filterOn).html();
            if (!lFilteredText) {
                throw new Error('File ' + origFilename + ' does not define a template for ' + filterOn + ', or that template is empty.');
            }
            return lFilteredText.trim();
        }
        else {
            var cheerio = require.nodeRequire('cheerio');
            var jq = cheerio.load(text);
            var filteredText = jq('script' + filterOn).html();
            if (!filteredText) {
                console.log('Error: File ' + origFilename + ' does not define a template for ' + filterOn + ', or that template is empty.');
                throw new Error('File ' + origFilename + ' does not define a template for ' + filterOn + ', or that template is empty.');
            }
            return filteredText.trim();
        }
    }

    function parseMeta (text) {
        var nodes = Handlebars.parse(text),
            result = {};
        if (nodes && nodes.statements) {
            var meta = nodes.statements[0];
            if (meta && meta.type === 'comment') {
                try {
                    result = JSON.parse(meta.comment.trim());
                } catch (e) {
                    result = { comment : meta.comment.trim() };
                }
            }
        }
        return result;
    }

    function makeDepString (deps) {
        var depsQuote = [],
            result;
        for (var i = 0; i < deps.length; i++) {
            depsQuote.push("'" + deps[i] + "'");
        }

        result = depsQuote.join(',');

        if (deps.length) {
            result = ',' + result;
        }

        return result;
    }

    function displayCompileImportsError (name) {
        console.log('Could not read dependency function: ' + name + '. Will not be able to minimize template with known helpers.');
        if (!fullError) {
            console.log('Make sure your markup dependencies are require plugins of the form:');
            console.log('define([\'...\'], function (...) { var helper = function () { ... }; helper.handlebarsType = \'helper\'; helper.handlebarsName = \'name\'');
            console.log('');
            fullError = true;
        }
    }

    function getImportInfoByEval (name, text) {
        /*jshint evil:true, strict:false, unused:false*/
        var importFn;
        function define (deps, fn) {
            importFn = fn;
        }
        try {
            eval(text);
            return importFn.call(undefined);
        } catch (e) {
            displayCompileImportsError(name);
            return {};
        }
    }

    function knownDependenciesAtBuild(req, dependencies, fileName, callback, errback) {
        var async = require.nodeRequire('async');
        async.map(dependencies, function (dep, callback) {
            // WARNING: All dependencies listed in .hbs files MUST be JavaScript
            var url = req.toUrl(dep + '.js');
            getText(dep, url, true, function (text) {
                callback(null, getImportInfoByEval(dep, text));
            }, function () {
                displayCompileImportsError(dep);
                callback(null, {});
            });
        }, function (err, results) {
            if (err) {
                errback(err);
                return;
            }

            var imported = markupPlugin.buildImported(results, fileName),
                knownHelpers = imported ? markupPlugin.buildKnownHelpers(imported.helpers) : {};
            callback(knownHelpers);
        });
    }

    markupPlugin.load = function(name, req, onLoad, config) {
        var fileName = name.split(':')[0],
            optionalFilter = name.split(':')[1],
            isBuild = config.isBuild,
            noEscape = (config.markup && config.markup.noEscape) || false,
            debug = (config.markup && config.markup.debug) || false;

        if (isBuild && req.toUrl(fileName).match(/empty:/)) {
            console.log('Warning: Ignoring markup! import:', name, 'assumed to be included from library file.');
            return onLoad();
        }

        getText(fileName, req.toUrl(fileName), isBuild, function (text) {
            var usableText = text;

            if (optionalFilter) {
                usableText = filter(usableText, optionalFilter, isBuild, fileName);
            }

            var meta = parseMeta(usableText),
                knownHelpersOnly = !!meta.knownHelpersOnly;

            if (!isBuild) {
                if (meta && meta.deps) {
                    req(meta.deps, function () {
                        var deps = _.toArray(arguments),
                            imported = markupPlugin.buildImported(deps, name);
                        var inlineFn = inlineTemplate(usableText, imported, knownHelpersOnly, noEscape);
                        if (debug) {
                            inlineFn.templateText = usableText;
                        }
                        onLoad(inlineFn);
                    }, onLoad.error);
                } else {
                    var inlineFn = inlineTemplate(usableText, undefined, knownHelpersOnly, noEscape);
                    if (debug) {
                        inlineFn.templateText = usableText;
                    }
                    onLoad(inlineFn);
                }
            }
            else {
                buildMap[name] = { text: usableText, noEscape : noEscape };
                if (meta && meta.deps) {
                    // Ensure dependencies are captured during build process
                    req(meta.deps, function () {
                        knownDependenciesAtBuild(req, meta.deps, name, function (deps) {
                            buildMap[name].knownDeps = deps;
                            onLoad();
                        }, onLoad.error);
                    }, onLoad.error);
                } else {
                    onLoad();
                }
            }
        }, onLoad.error);
    };

    /**
     * The `hbr` plugin is the preferred templating system in Prime.
     * It combines the ability to load Handlebars template files with a network of helpers such as `button`, `lz`, and many more.
     * These combine to make accessible DOM generation easier than writing raw HTML all the time.
     *
     * See also: https://confluence.oraclecorp.com/confluence/display/~brian.wyant@oracle.com/2014/05/21/Primutils+-+the+vision%2C+the+changes
     *
     * See also: https://confluence.oraclecorp.com/confluence/display/~dan.joseph@oracle.com/Using+the+new+markup!+%28hbr!%29+plugin
     *
     * See also: https://confluence.oraclecorp.com/confluence/display/~dan.joseph@oracle.com/Converting+Prime+.htm+Templates+to+Handlebars+Templates
     *
     * Examples
     * --------
     *
     * **Basic Usage**
     *
     * To use the `hbr` plugin, we would do the following:
     *
     *     // path/to/my/markup.hbs:
     *     <div>Hello \{{this.name}}!</div>
     *     // myCode.js:
     *     define(['hbr!path/to/my/markup.hbs'], function (markupFn) {
     *         console.log(markupFn({
     *             name: 'Bob'
     *         });                   // prints '<div>Hello Bob!</div>'
     *     });
     *
     * In the above snippet, we're using the plugin to load a file, `path/to/my/markup.hbs`.
     * Ultimately, [Handlebars](http://handlebarsjs.com/) here is used to render the template, in this case used to print out the DOM to the console so we can see it.
     *
     * **Load & Pluck**
     *
     * If you prefer defining more than one Handlebars template per file, you can use a `:` in the file import to indicate to the plugin which template to render.
     * Let's take a look:
     *
     *     // path/to/my/markup-combo.hbs:
     *     <script id="hello-world">
     *         <div>Hello \{{this.name}}!</div>
     *     </script>
     *     <script id="hello-josh">
     *         <div>Hi Josh...</div>
     *     </script>
     *     // myCode.js:
     *     define(['hbr!path/to/my/markup-combo.hbs:#hello-world',
     *             'hbr!path/to/my/markup-combo.hbs:#hello-josh'], function (customMarkupFn, joshMarkupFn) {
     *         console.log(customMarkupFn({
     *             name: 'Bob'
     *         });                   // prints '<div>Hello Bob!</div>'
     *
     *         console.log(joshMarkupFn({});
     *                               // prints '<div>Hi Josh...</div>'
     *     });
     *
     * The above `markup-combo.hbs` file defines two templates, and we distinguish between them in our JS file by using either `:#hello-world` or `:#hello-josh`.
     *
     * **Helpers and Subexpressions**
     *
     * Let's say we want to make a button with some localized text content and and icon.
     * To do this, we would want to use the [`button`](templates.helpers.button.html), [`lz`](templates.helpers.lz.html), and [`icon`](templates.helpers.icon.html) helpers.
     *
     * Something like the following could do:
     *
     *     // path/to/my/markup.hbs:
     *     \{{#button "toolbar" (lz "label.global_select")}}
     *         \{{icon "checkmark-checked" (lz "label.global_select")}}
     *         \{{lz "label.global_select"}}
     *     \{{/button}}
     *
     * This example is a bit nonsensical, but would produce a button, with an icon and text inside of it, and a tooltip.
     * Note the use of parentheses - (lz ...) - in the first and second line of the template.
     * These are [Subexpressions](http://handlebarsjs.com/expressions.html).
     *
     * @class plugins.hbr!
     * @module plugins
     */
    markupPlugin.write = function(pluginName, moduleName, write) {
        if(buildMap[moduleName]) {
            var moduleInfo = buildMap[moduleName],
                text = moduleInfo.text,
                knownDeps = moduleInfo.knownDeps,
                noEscape = moduleInfo.noEscape,
                meta = parseMeta(text),
                compiled = compileTemplate(text, knownDeps, !!meta.knownHelpersOnly, noEscape),
                depString = (meta && meta.deps) ? makeDepString(meta.deps) : '';

            write(
                "define('" + pluginName + "!" + moduleName  + "', ['primutils.handlebars','"+ pluginName +"'" + depString + "], " +
                "function (H, u) {\n" +
                "var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), '" + moduleName + "'),\n" +
                "    func = H.template(" + compiled.toString() + ");\n" +
                "return function (context, options) {\n" +
                "    options = u.extendOptions(options || {}, imports);\n" +
                "    return func(context, options).trim();\n" + // Trim is for jQuery -- must start with "<"
                "};\n" +
                "});\n"
            );
        }
    };

    return markupPlugin;
});


