
/**
 * Common 'module' plugin.
 *
 * Several of our common libraries make use of internally-defined JS modules that are meant to be exposed
 * to developers.
 *
 * An example: pgbucoreapi_nodeps.js is meant to have the 'pgbucoreapi' path point to it, and defines
 * the 'pgbucoreapi' module.
 * In that file, it also defines the module 'pgbucore/formatters/numberFormatter'.  This number formatting module
 * is intended to be used by a developer.  Naively, one would do something like the following:
 *
 *     // Start example
 *     define(['pgbucoreapi'], function () {
 *         // Do some stuff
 *         require(['pgbucore/formatters/numberFormatter'], function (formatter) {
 *             // Do something with the formatter
 *         });
 *         // Do other stuff
 *     });
 *
 * And, indeed, this will work without problems.  But, there is a simpler way, since we know that
 * the 'pgbucore/formatters/numberFormatter' module is contained in the same file as the 'pgbucoreapi' module:
 * we can use the 'module' plugin.  Compare the above to this simplified version:
 *
 *     // Start example
 *     define(['module!pgbucoreapi:pgbucore/formatters/numberFormatter', function (formatter) {
 *         // Do some stuff
 *         // Do something with the formatter
 *         // Do other stuff
 *     });
 *
 * With this, we can be sure that 'pgbucoreapi' is loaded first, but actually return the
 * 'pgbucore/formatters/numberFormatter' module instead of the empty 'pgbucoreapi' module.
 *
 * Having multiple modules from the same library is also fine:
 *
 *     define([
 *              'module!pgbucoreapi:pgbucore/formatters/numberFormatter',
 *              'module!pgbucoreapi:pgbucoreapi.dateFormatter'
 *         ],
 *         function (numFormatter, dateFormatter) {
 *             // Code that uses formatters here
 *         }
 *     );
 *
 * PS: If you want to use the module plugin in another project, you can copy it from the
 * pages/javascripts/shared/internalPlugins directory.
 * Please make sure your requireConfig is setup correctly to point to that file.
 * Also, if you are glomming files together that make use of the module plugin, make sure to read up on
 * [RequireJS optimization in general](http://requirejs.org/docs/optimization.html), and the
 * ['stubModules' option](https://github.com/jrburke/r.js/blob/66f1877b2fa93fe144eb8db85963e2d1edddbc7d/build/example.build.js#L286)
 * in particular.
 *
 * @class module!
 * @namespace plugins
 */

define({
    load: function (name, req, load, config) {
        'use strict';

        if (config.isBuild) {
            return load();
        }

        var parts = name.split(':');
        var libName = parts[0];
        var moduleName = parts[1];

        req([libName], function (lib) {
            req([moduleName], function (module) {
                load(module);
            });
        });
    }
});
