//!
//! OUI
//! Built on: Thu Oct 30 2014 17:38:00
//! Library version: 2.22.3
//! Branch: blue-sky-3  |  Commit: a2579c9
//!
define('oui.utils/subscribable',['oui.underscore'], function (_) {
    

    /**
     * @typedef {Object} Subscribable~EventOptions
     * @property {Boolean} [silent=false]
     */

    /**
     * @method Subscribable~normalize
     */
    function normalize (eventName) {
        var split = eventName.split('.');

        if(eventName === split[0]) {
            return { name : eventName };
        } else {
            return { name : split[1], namespace : split[0] };
        }
    }

    /**
     * The interface for all objects that can be subscribed to.
     *
     * @class Subscribable
     * @private
     */
    function Subscribable() {

        /**
         * The collection of events for the object.
         *
         * An event object has keys: name, namespace (optional), fn, context
         *
         * @memberof! Subscribable.prototype
         * @type {Array<Object>}
         * @default []
         * @private
         */
        this._events = [];
    }

    Subscribable.prototype = {
        constructor : Subscribable,
        /**
         * Calls each function that matches the argument.
         * Argument can either be namespace.event, or just event
         *
         * If argument is just event name than all functions will be called otherwise
         * only the namespaced event will be called.
         *
         * @example
         * <caption>Triggering a silent event with event args '1'</caption>
         *
         * this.trigger({ silent : true }, 'select', '1');
         *
         * @example
         * <caption>Triggering a default event with event args '1'</caption>
         *
         * this.trigger('select', '1');
         *
         * @example
         * <caption>Triggering a default namespaced event with event args '1'</caption>
         *
         * this.trigger('view.select', '1');
         *
         * @method Subscribable#trigger
         * @param {Subscribable~EventOptions} [options] The event options. Currently, only silent is an accepted property of the object
         * @param {String} eventName Either a string in the form of nameSpace.event or event
         */
        trigger : function (options, eventName) {
            var args = Array.prototype.slice.call(arguments);

            //No options supplied or using combined options + eventName
            if(_.isString(options) || (_.isObject(options) && options.name)) {
                args = _.rest(args, 1);

                eventName = options.name || options;
            //Options given to trigger
            } else {
                args = _.rest(args, 2);
            }

            options = options || { silent : false };

            _.chain(this._events)
                .where(normalize(eventName))
                .each(function(event) {
                    event.fn.apply(event.context || options, args);
                });

            args = _.union([eventName], args);

            _.chain(this._events)
                .where({ name : '*'})
                .each(function (event) {
                    event.fn.apply(event.context || options, args);
                });
        },
        /**
         * Sets the callback for the event name under the given namespace.
         *
         * '*' as the event name is a special property that will be
         * fired for any event called.
         *
         * @example
         * <caption>Listen to select event name and function</caption>
         *
         * this.on('select', function () {});
         *
         * @example
         * <caption>Listen to select event with specified context</caption>
         *
         * this.on('select', function () {}, that);
         *
         * @example
         * <caption>Listen to select through Object</caption>
         *
         * this.on({
         *    'select' : function () {}
         * })
         *
         * @example
         * <caption>Listen to select event through Object (variation)</caption>
         *
         * this.on({
         *     'select' : {
         *          fn : function () {},
         *          context : that
         *     }
         * })
         *
         * @method Subscribable#on
         * @chainable
         * @param {String|Object} eventName Either a string in the form of nameSpace.event or event
         * @param {Function} [callBack] The function to call when triggered.
         * @param {Object} [context] The context to use when calling the callBack.
         */
        on : function (eventName, callBack, context) {
            var that = this;

            function addEvent(ev, cb, ctx) {
                var events = ev.split(' ');

                if(that.name) {
                    //If this is a widget then always use 'that' for the context
                    ctx = ctx || that;
                } else {
                    //Allow undefined to be set as the context as that will allow the eventParams
                    //to be applied as the context in 'trigger'
                    ctx = ctx;
                }

                _.each(events, function (event) {
                    that._events.push(_.extend({ fn : cb, context : ctx }, normalize(event)));
                });
            }

            if (_.isObject(eventName)) {
                //Given a map
                //Ignore callback and context
                _.each(eventName, function (value, key) {
                    if (_.isFunction(value)) {
                        addEvent(key, value, undefined);
                    } else {
                        addEvent(key, value.fn, value.context);
                    }
                });
            } else if (_.isString(eventName)) {
                addEvent(eventName, callBack, context);
            }

            return this;
        },
        /**
         * Removes the function(s) for the given namespace. If no name space is given then all
         * functions are removed.
         *
         * @example
         * <caption>Remove all events</caption>
         *
         * this.off();
         *
         * @example
         * <caption>Remove all events of name 'select'</caption>
         *
         * this.off('select');
         *
         * @example
         * <caption>Remove only namespaced 'select' event</caption>
         *
         * this.off('view.select');
         *
         * @method Subscribable#off
         * @chainable
         * @param {String} [eventName] Either a string in the form of nameSpace.event or event
         */
        off : function (eventName) {
            if(eventName == null || eventName === '') {
                this._events = [];
            } else {
                var properties = normalize(eventName);
                this._events = _.reject(this._events, function (event) {
                    return _.isEqual(properties, _.pick(event, _.keys(properties)));
                });
            }

            return this;
        }
    };

    return Subscribable;

});

/**
 * @module uniqueId
 * @private
 */
define('oui.utils/uniqueId',[],function() {
    

    var ids = {};

    return function(clz) {
        if(!ids[clz]) {
            ids[clz] = 0;
        }

        return ++ids[clz];
    };
});

/**
 * Helper module for dealing with optional arguments to a function.
 *
 * @module argumentsHelper
 * @private
 */
define('oui.utils/argumentsHelper',['oui.underscore'], function(_) {
    

    //TODO: not really able to document this with yuidoc. Needs additional tag support
    //https://github.com/yui/yuidoc/issues/25
    /**
     * Helper function to deal with optional arguments to a function.
     *
     * Uses the arguments and a map of key expecting to a function to determine what case this is the argument
     *
     * @method
     * @param {Arguments|Array} args The given arguments to a function
     * @param {Object} mapper The map of key to argument value function
     *
     * @return {Object} fixedArgs The remapped arguments the function is expecting.
     */
    return function(args, mapper) {
        var fixedArgs = {};
        args = _.toArray(args);

        _.each(mapper, function(argFn, key) {
            var found = _.find(args, argFn);

            if(_.isUndefined(found)) {
                found = null;
            }

            fixedArgs[key] = found;
        });
        return fixedArgs;
    };
});


define("oui.json/strings.json", function(){ return {
    "loading" : "Loading",
    "help" : "Help",
    "today": "Today",
    "ok": "Ok",
    "cancel": "Cancel",
    "remove": "Remove",
    "close" : "Close",
    "expanded" : "Expanded",
    "collapsed": "Collapsed",
    "locked": "Locked",
    "calendar": {
        "nextYear": "Next Year",
        "prevYear": "Previous Year",
        "nextMonth": "Next Month",
        "prevMonth": "Previous Month",
        "selectDate": "Select Date"
    },
    "spinner": {
        "increment": "Increment",
        "decrement": "Decrement"
    },
    "months": [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December"
    ],
    "day": {
        "abbrev": {
            "sunday": "Sun",
            "monday": "Mon",
            "tuesday": "Tues",
            "wednesday": "Wed",
            "thursday": "Thu",
            "friday": "Fri",
            "saturday": "Sat"
        },
        "full": {
            "sunday": "Sunday",
            "monday": "Monday",
            "tuesday": "Tuesday",
            "wednesday": "Wednesday",
            "thursday": "Thursday",
            "friday": "Friday",
            "saturday": "Saturday"
        }
    },
    "menu" : "Dropdown menu",
    "test": "Test",
    "clear": "Clear",
    "resize": "Resize",
    "collapse": {
        "up": "Collapse Up",
        "down": "Collapse Down",
        "left": "Collapse Left",
        "right": "Collapse Right"
    },
    "search" : "Search",
    "placeholder" : "Search",
    "notification" : "{0} selected. Click to remove filter.",
    "no_results_found": "No Results Found",
    "picker" : {
        "select" : "Select"
    },
    "browse" : "Browse",
    "editor" : {
        "title" : "Rich Text Editor",
        "fontColor" : "Font Color",
        "backgroundColor" : "Background Color",
        "bold" : "Bold",
        "italic" : "Italic",
        "underline" : "Underline",
        "strikeThrough" : "Strikethrough",
        "subscript" : "Subscript",
        "superscript" : "Superscript",
        "justifyLeft" : "Left Justify",
        "justifyCenter" : "Center",
        "justifyRight" : "Right Justify",
        "orderedList" : "Ordered List",
        "unorderedList" : "Unordered List",
        "increaseIndent" : "Increase Indent",
        "decreaseIndent" : "Decrease Indent",
        "insertLink" : "Insert Link",
        "removeLink" : "Remove Link",
        "fontSize" : "Font Size",
        "fontName" : "Font Name",
        "fontFormat" : "Font Format",
        "linkText" : "Link Text",
        "openIn" : "Open in",
        "openInCurrent" : "Current Tab",
        "openInNew" : "New Tab or Window",
        "url" : "URL"
    },
    "icon" : {
        "select" : "Select Icon"
    },
    "color" : {
        "select" : "Select Color",
        "empty" : "Empty",
        "000000" : "Black",
        "800000" : "Dark Red",
        "FF0000" : "Red",
        "FF00FF" : "Magenta",
        "FF99CC" : "Pink",
        "993300" : "Darkest Orange",
        "FF6600" : "Dark Orange",
        "FF9900" : "Orange",
        "FFCC00" : "Gold",
        "FFCC99" : "Light Orange",
        "333300" : "Brown",
        "808000" : "Medium Brown",
        "99CC00" : "Yellow Green",
        "FFFF00" : "Yellow",
        "FFFF99" : "Light Yellow",
        "003300" : "Dark Green",
        "008000" : "Medium Green",
        "339966" : "Green",
        "00FF00" : "Neon Green",
        "CCFFCC" : "Light Green",
        "003366" : "Darkest Cyan",
        "008080" : "Dark Cyan",
        "33CCCC" : "Medium Cyan",
        "00FFFF" : "Cyan",
        "CCFFFF" : "Light Cyan",
        "000080" : "Darkest Blue",
        "0000FF" : "Dark Blue",
        "3366FF" : "Blue",
        "00CCFF" : "Neon Blue",
        "99CCFF" : "Light Blue",
        "333399" : "Darkest Violet",
        "666699" : "Blue Grey",
        "800080" : "Violet",
        "993366" : "Violet Red",
        "CC99FF" : "Light Violet",
        "333333" : "Darkest Grey",
        "808080" : "Dark Grey",
        "969696" : "Grey",
        "C0C0C0" : "Light Grey",
        "FFFFFF" : "White"
    },
    "instructions" : {
        "date" : "Date Field: Enter a date as text or press either F2 or Shift-F10 to open calendar popup.",
        "datetime" : "Date/Time Field: Enter a date with time as text or press either F2 or Shift-F10 to open the calendar popup.",
        "picker" : "Press F2 or Shift-F10 to open the dialog.",
        "typeahead" : "Enter a query to search.",
        "time" : "Time Field: Enter a time as text.",
        "number" : "Number Field: Enter a number as text.",
        "combo-input" : "Select Field: Press Down to show all options or type to find the value.",
        "combo-button" : "Select Field: Press Down to show all options or type a letter to filter values.",
        "spinner" : "Press up to increment the value and press down to decrement the value."
    }
}
;});

/**
 * @module language
 * @private
 */
define('oui.utils/language',['oui.jquery', 'oui.json/strings.json'], function ($, strings) {
    

    var _lang = strings;

    return {
        /**
         * Extends/returns the language used by OUI.
         *
         * If no arguments are given it will return the language object otherwise it will extend the defaults with
         * the provided object.
         *
         * @function lang
         * @param {Object} [language] key value pairs of localization key to localized string value
         * @return {Object} The current language object used by OUI to construct templates
         */
        lang : function (language) {
            if (!arguments.length) {
                return _lang;
            }

            _lang = $.extend(true, strings, language);
        }
    };
});

/**
 * @module defaults
 * @private
 */
define('oui.utils/defaults',[
    'oui.jquery'
],
    function ($) {
    

    var _defaults = {},
        _defaultsSet = false;

    return function (defs) {
        if (!arguments.length) {
            return _defaults;
        }

        if(!_defaultsSet) {
            _defaults = $.extend(true, {}, _defaults, defs);
            _defaultsSet = true;
        }
    };
});

/**
 * @module dom
 * @private
 */
define('oui.utils/dom',['oui.jquery', 'oui.underscore'], function($, _) {
    

    return {
        /**
         * Stores a widget with element to get them when the user has an element later.
         *
         * @method put
         * @param {jQuery} element The element to store the widget with
         * @param {Widget} widget The widget to store with the element.
         */
        put : function (element, widget) {
            var data = this.get(element, true) || {};
            data[widget.name()] = widget;

            if($(element).is('input, button')) {

                $(element).data('oui-control', data);
            } else {
                $(element).find('input, button').first().data('oui-control', data);
            }
        },
        /**
         * Gets the widget(s) associated with the element
         *
         * Will only turn the matching widget if name is specified otherwise returns an Object of name -> widget.
         *
         * **Note**: Boolean parameter for name is reserved for internal use only.
         *
         * @method get
         * @param {jQuery} element
         * @param {String|Boolean} [name] What specific widget to find.
         * @returns {Object|Widget} Either a map of widgets keyed by name or the specific widget.
         */
        get : function (element, name) {
            var data;

            if($(element).is('input, button')) {
                data = $(element).data('oui-control');
            } else {
                data = $(element).find('input, button').first().data('oui-control');
            }

            if ( name === true) {
                return data;
            }

            if ( name ) {
                return data[name];
            }

            //_.keys has protection for calling on non-object
            //Place _.isObject here as extra? protection
            if (_.isObject(data) && _.keys(data).length === 1) {
                return _.flatten(data)[0];
            }

            return data;
        }
    };
});

define('oui.base/widget',[
    'oui.underscore',
    'oui.jquery',
    'oui.moment',
    'oui.utils/subscribable',
    'oui.utils/uniqueId',
    'oui.utils/argumentsHelper',
    'oui.utils/language',
    'oui.utils/defaults',
    'oui.utils/dom',
    'oui.logger'
],
       function (_, $, moment, Subscribable, uniqueId, argsHelper, language, globalDefaults, dom, logger) {
    

    var wp;

    //Monky-patching jQuery because they do not trigger an event
    //when an element is removed from the DOM.
    //DOMNodeRemoved is deprecated - and when people want to use $.detach it would
    //    assume destroyed
    //MutationObservers are the future - but like DOMNodeRemoved $.detach would not work.
    var _clean = $.cleanData;
    $.cleanData = function (elems) {
        for (var i = 0, elem; (elem = elems[i]) != null; i++) {
            try {
                $(elem).triggerHandler('destroy.oui');
            } catch (e) {
            }
        }
        return _clean(elems);
    };

    /**
     * Widget base class. Sets up the optional arguments (options, initData and element) and maintains a unique ID across all
     * widgets.
     *
     * Widgets trigger each event with Subscribable as well as DOM Events. For DOM events each event is prefixed with 'oui-'.
     *
     * @class Widget
     * @private
     * @extends Subscribable
     * @since 1.3.0
     * @param {Object} [options] The widget's options
     * @param {jQuery} [element] JQuery object that the UI will be built
     *   on top of.
     * @param {Number|Array|moment|String} [initData] The data to be initialized
     *  with.
     */
    function Widget() {
        var defaults = this.defaults(), val, args, options, element, initData, that = this;

        //Super call
        Subscribable.call(this);

        args = argsHelper(arguments, {
            options : function(arg) {
                return _.isObject(arg) && !arg.jquery && !moment.isMoment(arg);
            },
            element : function(arg) {
                return _.isObject(arg) && arg.jquery;
            },
            initData : function(arg) {
                return !_.isObject(arg) || _.isArray(arg) || moment.isMoment(arg);
            }
        });

        options = args.options;
        element = args.element;
        initData = args.initData;

        if (element != null && element.jquery && element.length === 0) {
            //Log about an invalid selector being passed in - setting element to null
            logger.warn('An invalid selector was supplied as the element - constructing new element for creation');
            element = null;
        }

        if (!_.isNull(element) && _.isNull(initData) && element.is('input') && element.val() !== '') {
            val = element.val();
            if (!_.isUndefined(options.formatter) && options.formatter.validate(val)) {
                initData = options.formatter.unformat(val);
            } else {
                initData = val;
            }
        }


        /**
         * The options for the widget.
         *
         * @property _options
         * @private
         * @type {Object}
         */
        this._options = $.extend(true, {}, defaults, { locale : language.lang() }, globalDefaults()[this.name()] || {}, options || {});
        /**
         * The element that was passed into when creating the widget.
         *
         * @property _element
         * @private
         * @type {Object}
         */
        this._element = element;
        /**
         * The data to use as the objects initial data.
         *
         * @property _initData
         * @private
         * @type {Object}
         */
        this._initData = initData;
        /**
         * Each widget gets an id that will be unique across widgets on a page.
         *
         * @property _id
         * @type {Number}
         * @private
         */
        this._id = uniqueId('widget');

        Object.defineProperty(this, '_data', {
            enumerable : false,
            configurable : true,
            get : $.noop,
            set : $.noop
        });

        //TODO: is this really useful or can we trust it - users can update the options themselves
        if(this._options.storeFns) {
            _.each(this._options.storeFns, function(fn, key) {
                that._options.storeFns[key] = _.bind(fn, that);
            });
        }

        if(this._options.viewFns) {
            _.each(this._options.viewFns, function(fn, key) {
                that._options.viewFns[key] = _.bind(fn, that);
            });
        }

        this._settings = {
            id : this._id,
            name : this.name()
        };
    }

    Widget.prototype = Object.create(Subscribable.prototype);
    wp = Widget.prototype;

    wp.constructor = Widget;

    /**
     * Abstract method for getting the widgets name.
     *
     * @deprecated
     * @see Widget#name
     * @method Widget#getName
     * @return {String} name - The name of the widget.
     */
    /**
     * Abstract method for getting the widgets name.
     *
     * @method Widget#name
     * @return {String} name The name of the widget (defaults to 'widget').
     */
    wp.name = wp.getName = function () {
        return 'widget';
    };

    /**
     * Initializes the widget after it is on the DOM. This is necessary for all widgets to let them know that they can start
     * using computed CSS properties if needed. Not all widgets use CSS properties but dynamic positioning of additional DOM
     * structure will need to use some of these properties.
     *
     * Sets up the remove listener to clean itself up when it is removed from the DOM.
     *
     * @method Widget#init
     * @chainable
     */
    wp.init = function () {
        var that = this;

        $(this.element()).on('destroy.oui', function () {
            that.destroy({ self : false });
        });

        if(!_.isUndefined(this._view) && _.isFunction(this._view.init)) {
            this._view.init();
        }

        dom.put(this.element(), this);

        this._init();

        return this;
    };

    /**
     * Empty function that each widget can override to provide
     * custom initialization logic.
     *
     * @method Widget#_init
     * @private
     */
    wp._init = $.noop;

    /**
     * Function to explicitly remove the remove event listener.
     *
     * @method Widget#removeOff
     * @private
     */
    wp.removeOff = function () {
        $(this.element()).off('destroy.oui');
    };

    /**
     * Function to be called when the widget is either removed from the DOM
     * or is called directly.
     *
     * This method should not be overridden by subclasses.
     *
     * @method Widget#destroy
     * @param {Object} cleanUp Options for cleaning up widget.
     * @param {Boolean} [cleanUp.self=true] Whether to clean up the DOM element or false if jQuery is cleaning up the DOM
     * element.
     */
    wp.destroy = function (cleanUp) {
        if (_.isUndefined(cleanUp)) {
            cleanUp = { self : true };
        }
        this.removeOff();

        if(!_.isUndefined(this._view)) {
            this._view.destroy(cleanUp);
        }

        this._destroy(cleanUp);
    };

    /**
     * Empty function that each widget can implement to provide their default
     * destroy function.
     *
     * @method Widget#_destroy
     * @private
     */
    wp._destroy = $.noop;

    /**
     * Getter/setter for the current value of the widget.
     *
     * @method Widget#data
     * @param {*} [data] The new data/value to set on the widget
     * @returns {*} The current data/value of the widget.
     */
    wp.data = function (data) {
        if (!arguments.length) {
            return this._data;
        } else {
            this._data = data;
        }
    };

    /**
     * Returns the element that represents this widget.
     *
     * @method Widget#element
     * @return {HTMLElement} The element that makes up the component for placement on the DOM
     */
    /**
     * Alias for element.
     *
     * @see Widget#element
     * @method Widget#getElement
     */
    wp.element = wp.getElement = function () {
        if(!_.isUndefined(this._view)) {
            return this._view.element();
        }
    };

    /**
     * Returns the default options for a widget.
     *
     * Used by widgets that contain other other widgets such as the
     * DatePicker.
     *
     * @method Widget#defaults
     */
    wp.defaults = function () {
        logger.error('"defaults" method must be implemented in subclass');
    };

    /**
     * Update a single option at the specific key.
     *
     * Currently, does not work with nested keys.
     *
     * @method Widget#updateOption
     * @param {String} key The attribute to update.
     * @param {String|Date|Number|null} value The value of the attribute.
     */
    wp.updateOption = function (key, value) {
        if (_.isFunction(this[key])) {
            this[key](value);
        } else if (_.contains(_.keys(this._options), key)) {
            this._options[key] = value;
        } else {
            logger.warn('Attribute: ' + key + ' is not valid.');
        }
    };

    /**
     * Update multiple options at once.
     *
     * @method Widget#updateOptions
     * @param {Object} map The map of attributes to update.
     */
    /**
     * Update multiple options at once.
     *
     * @see Widget#updateOptions
     * @param {Object} map The map of attributes to update.
     */
    wp.options = wp.updateOptions = function (map) {
        var that = this, convertMap = {};

        if (_.isString(map) || arguments.length === 2) {
            logger.warn('Please use updateOption when only setting a single option.');
            logger.info('Converting to object.');
            convertMap[map] = arguments[1];
            map = convertMap;
        }

        _.each(map, function (value, key) {
            that.updateOption(key, value);
        });
    };

    /**
     * Disables/Enables the widget
     *
     * @method Widget#disable
     * @param {Boolean} disabled Whether or not the widget should be disabled.
     */
    wp.disable = function (disabled) {
        if(this._view && _.isBoolean(disabled)) {
            this._view.disable(disabled);
        }
    };

    /**
     * Locks or unlocks the widget.
     *
     * Used when the widget should be displayed but user does not have security rights to update the value.
     *
     * If the parameter is { locked.view : true } or locked is true then the widget will empty the data contained.
     * If the parameter is { locked.edit : true } then the widget applies the locked style (disables widget and displays
     * lock) without removing the data.
     *
     * @method Widget#lock
     * @param {Object|Boolean} locked Whether the widget is locked or not.
     * @param {Boolean} [locked.edit=false] Whether edit is allowed on the widget.
     * @param {Boolean} [locked.view=false] Whether the view is locked.
     */
    wp.lock = function (locked) {
        var emptyData = (locked === true) || locked.view || false;
        var lockView = emptyData || locked.edit || false;
        if(this._view) {

            this._view.lock(lockView);
            this._view.disable(lockView);

            if(emptyData) {
                this.data(null);
            }
        }
    };

    /**
     * Retrieves or sets the id attribute of the related DOM element that represents the widget.
     *
     * @method Widget#id
     * @since 2.7.3
     * @param {Object|String} [id] The id or settings to use when setting the attribute on the DOM element.
     * @return {String} The id of the widget as used in the DOM
     */
    wp.id = function (id) {
        if(!arguments.length) {
            return this._view.id();
        }

        this._view.id(id);
    };

    /**
     * Used when triggering events and arguments need to be manipulated/translated.
     *
     * @method Widget#_translate
     * @since 2.7.4
     * @private
     * @return {Array} arguments The translated arguments for the event being triggered.
     */
    wp._translate = function () {
        return arguments;
    };

    /**
     * Takes in the subscribable to subscribe to and sets up forwarding method.
     *
     * This function will not trigger any event if the event is described as 'silent'.
     *
     * Triggers DOM events with oui- prefixed to all events as well as triggering subscribable events.
     *
     * @method Widget#_setupEvents
     * @since 2.7.3
     * @private
     * @param {Array<Subscribable>|Subscribable} [subscribable] The subscribables to add event listeners for.
     */
    wp._setupEvents = function (subscribable) {
        var viewEl = this._view.element(),
            that = this;

        function addEvents(sub) {
            sub.on('*', function(name) {
                if(!this.silent) {
                    var args = _.chain(that._translate.apply(that, arguments))
                                .toArray()
                                .value();
                    if(that._element && that._element.length > 0) {
                        that._element.trigger('oui-' + name, _.union([that], args));
                    } else {
                        viewEl.trigger('oui-' + name, _.union([that], args));
                    }

                    if(that._options.events && that._options.events[name]) {
                        that._options.events[name].apply(that, args);
                    }

                    that.trigger.apply(that, args);
                }
            });
        }

        if(_.isArray(subscribable)) {
            //Context will be undefined in addEvents after this call
            _.each(subscribable, addEvents);
        } else {
            addEvents(subscribable);
        }

    };

    return Widget;

});

define('oui.base/input',['oui.base/widget'],
    function (Widget) {
        

        /**
         * Used as a point of extension to specific input widgets.
         *
         * @class Input
         * @private
         * @extends Widget
         * @since 1.3.0
         */
        function Input(options, element, initData) {
            Widget.call(this, options, element, initData);
        }

        Input.prototype = Object.create(Widget.prototype);

        Input.prototype.constructor = Input;

        return Input;
    }
);


define('oui.template/instructions.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<span class="spoken-text instructions" data-name="'+
((__t=( data.name ))==null?'':__t)+
'">'+
((__t=( data.instructions ))==null?'':__t)+
'</span>';
return __p;
};});


define("oui.json/keyCode.json", function(){ return {
    "UP": 38,
    "DOWN": 40,
    "LEFT": 37,
    "RIGHT": 39,

    "SPACE": 32,
    "TAB": 9,
    "BACKSPACE": 8,
    "ENTER": 13,
    "ESC": 27,

    "SHIFT": 16,
    "CTRL": 17,
    "ALT": 18,

    "PAUSE": 19,
    "CAPS": 20,
    "PAGEUP": 33,
    "PAGEDOWN": 34,
    "END": 35,
    "HOME": 36,
    "INSERT": 45,
    "DELETE": 46,

    "F1": 112,
    "F2": 113,
    "F3": 114,
    "F4": 115,
    "F5": 116,
    "F6": 117,
    "F7": 118,
    "F8": 119,
    "F9": 120,
    "F10": 121,
    "F11": 122,
    "F12": 123,

    "NUMLOCK": 144,
    "SCROLLLOCK": 145,

    "LEFTSUPER": 91,
    "RIGHTSUPER": 92
}
;});

/**
 * @module keyCode
 * @private
 */
define('oui.utils/keyCode',['oui.json/keyCode.json'], function(keyCode) {
    

    return keyCode;
});

define('oui.mixins/shortcut',['oui.underscore', 'oui.utils/keyCode'], function (_, keyCode) {
    

    /**
     * Mixin for views to quickly add shortcut handlers.
     *
     * @mixin shortcut
     * @private
     */
    return {
        _initShortcutProps : function() {
            var that = this;
            this._keyDownShortcuts = {};
            this._keyUpShortcuts = {};

            this._handler = function (shortcuts, event) {
                var shortcut = _.find(shortcuts, function(shortcut) {
                    return (shortcut.hasCtrl === event.ctrlKey) &&
                    (shortcut.hasShift === event.shiftKey &&
                    (event.altKey === shortcut.hasAlt) &&
                    (shortcut.hasMeta === event.metaKey)) &&
                    (event.keyCode === shortcut.key ||
                        (_.isString(shortcut.key) && String.fromCharCode(event.keyCode) === shortcut.key.toUpperCase()));
                });

                if(!_.isUndefined(shortcut)) {
                    if(_.isUndefined(shortcut.preventDefault) ||
                       (shortcut.preventDefault === true || shortcut.preventDefault.call(that))){
                        event.preventDefault();
                    }
                    if(!_.isUndefined(shortcut.stopPropagation) && (shortcut.stopPropagation === true ||
                       (shortcut.stopPropagation.call(that)))){
                        event.stopPropagation();
                    }
                    shortcut.cb();
                }
            };

            this._$el
                .on('keydown', _.partial(this._handler, this._keyDownShortcuts))
                .on('keyup', _.partial(this._handler, this._keyUpShortcuts));
        },
        /**
         * Adds a shortcut using a parseable format to a single common handler.
         *
         * Splits the name on the '-' and grabs the modifier(s) to use for the character.
         *
         * **Limitation** - can only have one callback per shortcut key
         *
         * @function shortcut#addShortcut
         * @param {String} name The event to split and set up the event handler for.
         * @param {Function} callback The function to call when the shortcut keys are held down. (keydown)
         */
        addShortcut : function (name, callback) {
            var keys = name.toUpperCase().split('-'),
                up = false || name === 'ESC',
                shortcut,
                preventDefault,
                stopPropagation;

            if(_.isUndefined(this._shortcutInit)) {
                this._initShortcutProps();
                this._shortcutInit = true;
            }

            if(_.isObject(callback) && !_.isFunction(callback)) {
                preventDefault = callback.prevention;
                stopPropagation = callback.stopPropagation;

                if(callback.up != null) {
                    up = callback.up;
                }
                callback = callback.fn;
            }

            shortcut = {
                name : name,
                preventDefault : preventDefault,
                stopPropagation : stopPropagation,
                hasCtrl : _.contains(keys, 'CTRL'),
                hasShift : _.contains(keys, 'SHIFT'),
                hasAlt : _.contains(keys, 'ALT'),
                hasMeta : _.contains(keys, 'META'),
                key : keyCode[_.last(keys)] || _.last(keys),
                cb : _.bind(callback, this)
            };

            if(!up) {
                this._keyDownShortcuts[name] = shortcut;
            } else {

                if(_.last(keys) === 'SHIFT') {
                    shortcut.hasShift = false;
                }
                this._keyUpShortcuts[name] = shortcut;
            }

            return this;
        },
        /**
         * Uses addShortcut to add a collection of shortcuts at one time.
         *
         * @function shortcut#addShortcuts
         * @param {Object} shortcuts An object of key value pairs where the key is the shortcut and the value
         *     is the callback
         */
        addShortcuts : function(shortcuts) {
            var addShortcutFn = _.bind(this.addShortcut, this);

            _.each(shortcuts, function(shortcut, name) {
                var multiple = name.split(' ');
                if(multiple.length > 1) {
                    _.each(multiple, function(singleName) {
                        addShortcutFn(singleName, shortcut);
                    });
                } else {
                    addShortcutFn(name, shortcut);
                }
            });
        },
        /**
         * Removes the shortcut from the collection of shortcuts.
         *
         * @function shortcut#removeShortcut
         * @param {String} name The name of the shortcut to remove.
         */
        removeShortcut : function (name) {
            delete this._shortcuts[name];
        }
    };
});

/**
 * @module inputTypes
 * @private
 */
define('oui.utils/inputTypes',[],function () {
    

    var available = {}, elem;

    [
        'date',
        'datetime',
        'color',
        'datetime-local',
        'email',
        'month',
        'number',
        'range',
        'search',
        'tel',
        'time',
        'url',
        'week',
        'file'
    ].forEach(function(key) {
        elem = document.createElement('input');
        elem.setAttribute('type', key);

        available[key] = elem.type === key;
    });

    return available;

});

/**
 * Utility module to determine if user is on a mobile browser.
 *
 * @module detectMobile
 * @private
 */
define('oui.utils/detectMobile',[],function () {
    

    return (/Android|webOS|iPhone|iPad|iPod|BlackBerry|Mobile/i).test(navigator.userAgent);
});

define('oui.base/view',['oui.underscore', 'oui.utils/subscribable', 'oui.template/instructions.htm', 'oui.mixins/shortcut', 'oui.utils/inputTypes', 'oui.utils/detectMobile'],
    function (_, Subscribable, instructions, shortcut, supportedInputTypes, detectMobile) {
    

    //Used to shorten the View.prototype call below.
    var vp;

    //TODO: look at these parameters. Reorganize to match all children constructors (model, controller, element, options)
    /**
     * Super class view that will subscribe to a subscribable object.
     * Not made to be instantiated by itself.
     *
     * Sets up the element to use as the widget. Might need to construct an entirely new DOM element or wrap an existing one.
     *
     * If the library is used on mobile devices it will use the native controls to allow the keyboards to change
     * appropriately.
     *
     * @class View
     * @extends Subscribable
     * @private
     * @mixes shortcut
     * @param {Object} options The widget's options. (They are not view specific)
     * @param {Object} [element] The element that will be used as the widget's UI.
     * @param {Object} [controller] The element the view will call into when the
     *   component is interacted with.
     */
    function View(options, element, controller, settings) {
        //Super call
        Subscribable.call(this);

        /**
         * The HTML element that gets created and returned to be appended to the page.
         *
         * It is a JQuery object.
         *
         * @property _$el
         * @type {Object}
         * @private
         */
        if(detectMobile && supportedInputTypes[this.nativeType(options)]) {
            this._$el = this.buildNativeElement(options, element, controller);

            this._$el.addClass('native');
        } else {
            this._$el = this.buildElement(options, element, controller);
            this._$el.append(instructions({
                name : settings.name,
                instructions : options.locale.instructions[options.type || settings.name]
            }));
        }

        this.id(settings);

        this._$el.toggleClass('fluid', options.fluid);

        /**
         * The options for the widget.
         *
         * @property _options
         * @private
         * @type {Object}
         */
        this._options = options;
        /**
         * The controller to use for the view.
         *
         * @property _controller
         * @private
         * @type {Controller}
         */
        this._controller = controller;
        /**
         * Settings built from the widget that are passed into the view.
         *
         * Contains the id and the name of the widget.
         *
         * @property _settings
         * @private
         * @type {Object}
         */
        this._settings = settings;

        return this;
    }

    View.prototype = Object.create(Subscribable.prototype);
    vp = View.prototype;

    _.extend(vp, shortcut);

    vp.constructor = View;

    /**
     * Returns the native type of the widget if it has one otherwise returns null.
     *
     * @method View#nativeType
     * @return {String} type The native type of the widget.
     */
    vp.nativeType = function() {
        return null;
    };

    vp.init = function () {};

    /**
     * Constructs the view's specific element that events are bound to.
     *
     * Overridden, in children classes when they need an element.
     *
     * @method View#buildElement
     * @return {null} The default element
     */
    vp.buildElement = function () {
        return null;
    };

    /**
     * Constructs the view's native element.
     *
     * @method View#buildNativeElement
     * @return {null} The native element of the widget.
     */
    vp.buildNativeElement = function () {
        return null;
    };

    /**
     * Alias to this._$el.find(selector)
     *
     * Utility to not issuing global DOM searches.
     *
     * @method View#$
     * @param {String} selector The selector to search for in the view's element.
     * @return {jQuery} The jQuery object represented by the selector.
     */
    vp.$ = function(selector) {
        return this._$el.find(selector);
    };

    /**
     * Injects other elements into this views element.
     *
     * @method View#inject
     * @param {Object} $element The jQuery object to be injected.
     * @param {String} selector The place to inject the object.
     */
    vp.inject = function ($element, selector) {
        this._$el.find(selector).append($element);
    };

    /**
     * Returns the view's element.
     *
     * @method View#element
     * @returns {Object} The jQuery object used by the view.
     */
    vp.element = vp.getElement = function () {
        return this._$el;
    };

    /**
     * Renders the view as locked or unlocked.
     *
     * @method View#lock
     * @param {Boolean} locked Whether the view should render as locked or unlocked.
     */
    vp.lock = function (locked) {
        this._$el.toggleClass('locked', locked);
    };

    /**
     * Cleans up the views created elements.
     *
     * @method View#destroy
     */
    vp.destroy = function () {
    };

    /**
     * Default id function for every view that can be overwritten if a view is not made up of an input element.
     *
     * Gets/Sets the id of the DOM element.
     *
     * @method View#id
     * @param {String|Object} [settings] The id or settings to use to set the id attribute of the DOM element.
     * @return id The id of the DOM element.
     */
    vp.id = function (settings) {
        var input = this._$el.find('input').first();

        if(!arguments.length) {
            return input.attr('id');
        }

        if(input.length > 0 && _.isUndefined(input.attr('id'))) {
            if(_.isObject(settings)) {
                input.attr('id', 'oui-' + settings.name + '-' + settings.id);
            } else if (_.isString(settings)) {
                input.attr('id', settings);
            }
        }
    };

    /**
     * Returns whether the widget is currently in a readonly state and additional popup elements should not be created.
     *
     * @method View#isReadOnly
     * @returns {Boolean} whether the element is in a readonly state.
     */
    vp.isReadOnly = function () {
        return this._$el.find('input').first().attr('readonly') != null;
    };

    return View;

});


define('oui.textField/textFieldTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="orcl-text-field">';
 if(data.multiline) { 
__p+=' <textarea></textarea>';
 } else { 
__p+=' <input type="text">';
 } 
__p+='</div>';
return __p;
};});


define('oui.textField/wrapTemplate.htm',[],function () { return '<div class="orcl-text-field">\n</div>\n';});

define('oui.textField/textView',[
    'oui.jquery',
    'oui.underscore',
    'oui.base/view',
    'oui.textField/textFieldTemplate.htm',
    'oui.textField/wrapTemplate.htm'
],
function($, _, View, textFieldTemplate, wrapTemplate) {
    

    var tvp;

    function TextView(element, options, settings) {
        var that = this;

        View.call(this, options, element, null, settings);

        this._$input = this._$el.find('input, textarea').first();
        this._inputEl = this._$input[0];

        this._$el.on('change', function() {
            that.trigger('text-changed', that._inputEl.value);
        });
    }

    TextView.prototype = Object.create(View.prototype);
    tvp = TextView.prototype;

    tvp.constructor = TextView;

    tvp.buildElement = function(opts, element) {
        var $element;

        if(_.isNull(element)) {
            $element = $(textFieldTemplate({ multiline : opts.multiline }));
        } else if(element.jquery) {
            $element = element
                .wrap(wrapTemplate)
                .parent();
        }

        return $element;
    };

    tvp.text = function(newText) {
        if(!arguments.length) {
            return this._inputEl.value;
        }

        this._inputEl.value = newText;
    };

    return TextView;

});

define('oui.textField/textField',['oui.base/input', 'oui.textField/textView'], function(Input, View) {
    

    var __WIDGET_NAME__ = 'text-field';
    var tfp;

    /**
     * @typedef {Object} TextField#Config
     *
     * @property {Boolean} [fluid=false] Whether or not the text field should take up all available widget.
     * @property {Boolean} [fill=false] Whether or not the widget should take up all available height. **Note**: Only works
     * with text areas.
     * @property {Boolean} [multiline=false] Whether or not the text field supports editing text that is more than one line.
     * Uses a text area instead of an input.
     * @property {Boolean} [required=false] Whether or not the textfield is required.
     */

    var __DEFAULTS__ = {
        fluid : false,
        fill : false, //Only works with textareas
        multiline : false,
        required : false
    };

    /**
     * Basic text field widget to provide a familiar API to an
     * input field.
     *
     * @class TextField
     * @extends Input
     * @param {TextField#Config} [options]
     * @param {jQuery} [element]
     * @param {String} [initData]
     */
    function TextField(options, element, initData) {
        Input.call(this, options, element, initData);

        this._view = new View(this._element, this._options, this._settings);

        Object.defineProperty(this, '_data', {
            get : this.text,
            set : this.text
        });

        this._setupEvents(this._view);
    }

    TextField.prototype = Object.create(Input.prototype);
    tfp = TextField.prototype;

    tfp.constructor = TextField;

    /**
     * Get the name of the widget.
     *
     * @method TextField#name
     * @return {String} 'text-field'
     */
    tfp.name = function() {
        return __WIDGET_NAME__;
    };

    /**
     * Combined getter/setter for the text fields value.
     *
     * @method TextField#text
     * @param {String} [newText] The new value of the text field.
     * @return {String} The current value of the text field.
     */
    tfp.text = function(newText) {
        if(!arguments.length) {
            return this._view.text();
        }

        this._view.text(newText);
    };

    /**
     * Get the defaults for the widget.
     *
     * @method TextField#defaults
     * @return {Object}
     */
    tfp.defaults = function() {
        return __DEFAULTS__;
    };

    return TextField;
});


define('oui.date/datePickerTemplate.htm',[],function () { return '<div class="orcl-date-picker">\n    <input type="date" spellcheck="false"/>\n</div>\n';});


define('oui.date/buttonTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button type="button" title="'+
((__t=( data.locale.calendar.selectDate  ))==null?'':__t)+
'" class="btn btn-picker" tabindex="-1"> <i class="pgbu-icon-small pgbu-icon-calendar"></i></button>';
return __p;
};});


define('oui.date/wrapTemplate.htm',[],function () { return '<div class="orcl-date-picker">\n</div>\n';});


define('oui.date/calendarTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="calendar" aria-hidden="true"> <span class="button-controls"> <div class="prev-buttons"> <button type="button" data-field="year" data-value="prev" title="'+
((__t=(data.locale.calendar.prevYear))==null?'':__t)+
'" tabindex="-1" class="btn btn-blue-light">&laquo;</button> <button type="button" data-field="month" data-value="prev" title="'+
((__t=(data.locale.calendar.prevMonth))==null?'':__t)+
'" tabindex="-1" class="btn btn-blue-light">&lsaquo;</button> </div> <label id="'+
((__t=( data.id  ))==null?'':__t)+
'" role="heading" aria-live="assertive" aria-atomic="true" class="title"></label> <div class="next-buttons"> <button type="button" data-field="month" data-value="next" title="'+
((__t=(data.locale.calendar.nextMonth))==null?'':__t)+
'" tabindex="-1" class="btn btn-blue-light">&rsaquo;</button> <button type="button" data-field="year" data-value="next" title="'+
((__t=(data.locale.calendar.nextYear))==null?'':__t)+
'" tabindex="-1" class="btn btn-blue-light">&raquo;</button> </div> </span> <div class="month"> <table aria-labelledby="'+
((__t=( data.id  ))==null?'':__t)+
'"> <thead> <tr> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.sunday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.sunday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.sunday ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.monday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.monday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.monday ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.tuesday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.tuesday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.tuesday ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.wednesday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.wednesday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.wednesday ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.thursday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.thursday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.thursday  ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.friday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.friday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.friday  ))==null?'':__t)+
'</th> <th scope="col" aria-label="'+
((__t=( data.locale.day.full.saturday ))==null?'':__t)+
'" title="'+
((__t=( data.locale.day.full.saturday ))==null?'':__t)+
'" >'+
((__t=( data.locale.day.abbrev.saturday ))==null?'':__t)+
'</th> </tr> </thead> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> </table> </div> <span class="button-controls today-cancel"> ';
 if (data.type === 'datetime' ) { 
__p+=' <input type="number"> ';
 } 
__p+=' <div class="group"> <button type="button" tabindex="-1" title="'+
((__t=( data.locale.today ))==null?'':__t)+
'" class="btn btn-small btn-blue-light today">'+
((__t=( data.locale.today ))==null?'':__t)+
'</button> <button type="button" tabindex="-1" title="'+
((__t=( data.locale.cancel ))==null?'':__t)+
'" class="btn btn-small btn-blue-light cancel                pull-right">'+
((__t=( data.locale.cancel ))==null?'':__t)+
'</button> </div> </span></div>';
return __p;
};});


define('oui.template/nativeElement.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<input type="'+
((__t=( data.type  ))==null?'':__t)+
'">';
return __p;
};});

define('oui.utils/dateHelper',['oui.underscore', 'oui.moment'], function (_, moment) {
    

    var numberComparator = function (num1, num2) {
        if (num1 > num2) {
            return 1;
        } else if (num1 < num2) {
            return -1;
        }

        return 0;
    };

    var flexComparator = function (fields, item1, item2) {
        var value;

        for (var i = 0; i < fields.length; i++) {
            value = numberComparator(item1[fields[i]](), item2[fields[i]]());

            if (value !== 0) {
                return value;
            }
        }


        return 0;
    };

    /**
     * Collection of functions to help with dealing with dates.
     *
     * @module dateHelper
     * @private
     */
    return {
        /**
         * Determine whether dates are before or after another date.
         *
         * Returns -1 when the first date is before the second
         * Returns 1 when the first date is after the second
         * Returns 0 when they are the same date.
         *
         * @method dateComparator
         * @return num Whether the first date is an earlier date than the second.
         */
        dateComparator : _.partial(flexComparator, ['year', 'month', 'date']),
        /**
         * Determine whether dates are before or after another date.
         * Only looks at time fields (hours, minutes, seconds)
         *
         * Returns -1 when the first date is before the second
         * Returns 1 when the first date is after the second
         * Returns 0 when they are the same date.
         *
         * @method timeComparator
         * @return num Whether the first date is an earlier date than the second.
         */
        timeComparator : _.partial(flexComparator, ['hours', 'minutes', 'seconds']),
        /**
         * Determines if a specific day of week or date is contained in an array.
         *
         * @method contains
         * @param {Array} dateArray The array of dates or integers (day of week) to check whether a date is contained in.
         * @param {moment} day The day to check
         * @return contained Whether or not the day is contained in the array.
         */
        contains : function (dateArray, day) {
            //dataArray may contain days of week [0-6]
            var contained = _.contains(dateArray, day.day());

            //dateArray may contain specific days (such as holidays)
            var specificDay = _.find(dateArray, function (val) {
                return moment.isMoment(val) && val.clone().startOf('day').isSame(day.clone().startOf('day'));
            });

            contained = contained || !_.isUndefined(specificDay);

            return contained;
        }
    };

});

define('oui.utils/error',[],function () {
    

    /**
     * @class OUIError
     * @param {String} type The type of error this represents
     * @param {Object} data The data for the error (past value, attempted value...)
     */
    function OUIError(type, data) {

        /**
         * @property name
         * @type {String}
         * @default 'OUIError'
         */
        this.name = 'OUIError';

        /**
         * @property type
         * @type {String}
         */
        this.type = type;

        /**
         * @property data
         * @type {Object}
         */
        this.data = data;
    }

    OUIError.prototype = {
        constructor : OUIError
    };

    return OUIError;
});

define('oui.spinner/numberModel',['oui.underscore', 'oui.utils/subscribable', 'oui.moment', 'oui.utils/dateHelper', 'oui.utils/error'],
    function (_, Subscribable, moment, dateHelper, OUIError) {
        

        /**
         * @event NumberModel#non-number-error
         * @param {Mixed} attempted The value that was attempted to be set on the number field.
         *  The type is anything that is not a number, null or moment object.
         * @param {moment|number} currentValue The current value of the number field.
         */
        /**
         * @event NumberModel#number-change
         * @param {moment|number|null} currentValue The current value of the number field.
         */
        /**
         * @event NumberModel#bounds-error
         * @param {moment|number} newData The value that was attempted to be set on the number field.
         */
        /**
         * @event NumberModel#required-error
         * @param {moment|number} newData The value that was attempted to be set on the number field.
         */

        var nmp;

        /**
         * @class NumberModel
         * @private
         *
         * @param {null|Number|moment} initData The initial data on this number field.
         * @param {NumberField#Config} options The options for the number field.
         */
        function NumberModel(initData, options) {
            //Super call
            Subscribable.call(this);

            this._options = options;
            this._data = null;
            this._invalidData = null;
            this._lastErrorType = null;

            if(initData === '') {
                initData = null;
            }

            if((this._options.type === 'date' || this._options.type === 'month') && initData == null) {
                this.data(moment());
            } else {
                this.data(initData);
            }
        }

        NumberModel.prototype = Object.create(Subscribable.prototype);
        nmp = NumberModel.prototype;

        nmp.constructor = NumberModel;

        nmp.inBounds = function (data) {
            var that = this,
                min = this._options.min,
                max = this._options.max;

            function above(val1, val2) {
                if (_.isNull(val1) || _.isNull(val2)) {
                    return true;
                }

                if (that._options.type === 'number' || that._options.type === 'currency') {
                    return (val1 >= val2);
                } else {
                    val1 = moment(val1);
                    val2 = moment(val2);
                }

                if (that._options.type === 'time') {
                    return dateHelper.timeComparator(val1, val2) >= 0;
                } else {
                    return val1.isAfter(val2) || val1.isSame(val2);
                }
            }

            return above(data, min) && above(max, data);
        };

        nmp.data = function (newData, eventParams) {
            if (!arguments.length) {
                //Clone the date before returning
                return moment.isMoment(this._data) ? moment(this._data) : this._data;
            }

            eventParams = eventParams || {};

            //Strings are allowed to be passed due to invalid data being a string
            //Prefer not to have to run the formatter over the passed in data though.
            if(_.isString(newData) && this._options.formatter.validate(newData)) {
                newData = this._options.formatter.format(newData);
            }

            if (!(_.isNull(newData) || _.isNumber(newData) || moment.isMoment(newData))) {
                this._invalidData = newData;
                this._lastErrorType = 'non-number';
                eventParams.name = 'non-number-error';
                this.trigger(eventParams, newData, this._data);
            } else if (_.isNull(newData) && this._options.required) {
                this._invalidData = undefined;
                this._lastErrorType = 'required';
                eventParams.name = 'required-error';
                this.trigger(eventParams, newData, this._data);
            } else if (!this.inBounds(newData)) {
                this._invalidData = newData;
                this._lastErrorType = 'bounds';
                eventParams.name = 'bounds-error';
                this.trigger(eventParams, newData);
            } else if ((newData === null && this._data !== null) ||
                       (newData !== null && this._data === null) ||
                    !this.isValid() || //Error currently - make sure we are updating
                    !(this._data <= newData && this._data >= newData)) {
                this._invalidData = null;
                this._data = newData;
                eventParams.name = 'number-change';
                this.trigger(eventParams, this._data);
            }
        };

        nmp.error = function () {
            if (!this.isValid()) {
                return new OUIError(this._lastErrorType, this._invalidData);
            } else {
                return null;
            }
        };

        nmp.isValid = function () {
            return _.isNull(this._invalidData);
        };

        nmp.canUpdate = function (value, field) {
            var incValue,
                decValue,
                step = this._options.spinner.step;

            //Make potentials and check those to the bounds
            if (moment.isMoment(value)) {
                incValue = this.data();
                decValue = this.data();

                if (field === 'half-day') {
                    field = 'hours';
                    step = 12;
                }

                incValue[field](incValue[field]() + step);
                decValue[field](decValue[field]() - step);

            } else {
                incValue = value + step;
                decValue = value - step;
            }

            return { increment : this.inBounds(incValue),
                decrement : this.inBounds(decValue) };
        };

        return NumberModel;
    });

define('oui.spinner/numberController',['oui.underscore'], function (_) {
    

    var ncp;

    function NumberController(model, options) {

        this._model = model;
        this._options = options;
    }

    ncp = NumberController.prototype;

    ncp.constructor = NumberController;

    ncp.update = function (field, value) {
        if (this._options.type === 'number' || this._options.type === 'currency') {
            var newValue = this._model.data() + value;
            if (!this._model.isValid()) {
                //TODO: look into not using private properties
                if (this._model._lastErrorType === 'bounds') {
                    if (this._model._invalidData < this._options.min) {
                        newValue = this._options.min;
                    } else {
                        newValue = this._options.max;
                    }
                }
            }

            this._model.data(newValue);

        } else {
            var potentialTime = this._model.data();

            if (field === 'half-day') {
                potentialTime.add('hours', 12);
            } else {
                potentialTime.add(field, value);
            }

            this._model.data(potentialTime);
        }
    };

    ncp.canUpdate = function (value, field) {
        return this._model.canUpdate(value, field);
    };

    ncp.increment = function (field) {
        if (this._model.canUpdate(this._model.data(), field).increment) {
            this.update(field, this._options.spinner.step);
        }
    };

    ncp.decrement = function (field) {
        if (this._model.canUpdate(this._model.data(), field).decrement) {
            this.update(field, -this._options.spinner.step);
        }
    };

    ncp.data = function (value, eventParams) {
        if (!arguments.length) {
            return this._model.data();
        }

        var newValue;
        if (_.isString(value) && this._options.formatter.validate(value)) {
            newValue = this._options.formatter.unformat(value);
        } else if (value === '' && !_.isNull(this._options.emptyValue)) {
            newValue = this._options.emptyValue;
        } else {
            newValue = value;
        }

        this._model.data(newValue, eventParams);
    };

    ncp.isValid = function () {
        return this._model.isValid();
    };

    ncp.error = function () {
        return this._model.error();
    };

    return NumberController;
});

/**
 * @module repeater
 * @private
 */
define('oui.utils/repeater',['oui.underscore'], function (_) {
    

    return function (fn, options) {
        options = _.defaults(options || {}, { initial : 500, step : 50, min : 50 });
        var time = options.initial;
        var interval = null;

        return {
            start : function recur()  {
                if (time - options.step >= options.min) {
                    time -= options.step;
                }

                //In case of undefined
                //var halt = fn() === false;
                fn();

                clearInterval(interval);
                interval = setInterval(recur, time);
            },
            stop : function () {
                clearInterval(interval);
                time = options.initial;
            }
        };
    };
});


define('oui.spinner/spinnerTemplate.htm',[],function () { return '<div class="orcl-number-field">\n    <input type="text">\n</div>\n';});


define('oui.spinner/wrapTemplate.htm',[],function () { return '<div class="orcl-number-field">\n</div>\n';});


define('oui.spinner/controlsTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="btn-group btn-group-vertical"> <button type="button" data-action="increment" title="'+
((__t=( data.locale.spinner.increment  ))==null?'':__t)+
'" class="btn btn-form btn-micro" tabindex="-1"> <i class="up"></i> </button> <button type="button" data-action="decrement" title="'+
((__t=( data.locale.spinner.decrement  ))==null?'':__t)+
'" class="btn btn-form btn-micro" tabindex="-1"> <i class="down"></i> </button></div>';
return __p;
};});

define('oui.spinner/numberView',['oui.underscore', 'oui.jquery', 'oui.moment', 'oui.base/view',
    'oui.utils/keyCode', 'oui.utils/repeater',
    'oui.template/nativeElement.htm',
    'oui.spinner/spinnerTemplate.htm', 'oui.spinner/wrapTemplate.htm', 'oui.spinner/controlsTemplate.htm',
    'oui.template/instructions.htm'],
    function (_, $, moment, View, keyCode, repeat, nativeElement, spinnerTemplate, wrapTemplate, controlsTemplate, instructions) {
        

        var nvp;

        var TIME_REGEX = /(\d+):(\d+)(?::(\d+))?(?: (\w{2}))?/;

        function NumberView(model, controller, element, options, settings) {
            var that = this;

            function fieldChangeHandler() {
                var update;
                that._cursorLoc = that._$input[0].selectionStart;

                update = that._controller.canUpdate(that._controller.data(), that._getField(that._cursorLoc, that._inputEl.value));

                that._$incButton.prop('disabled', !update.increment);
                that._$decButton.prop('disabled', !update.decrement);
            }

            View.call(this, options, element, controller, settings);

            model.on({
                'view.number-change' : this.update.bind(this),
                'view.required-error' : this.error.bind(this),
                'view.bounds-error' : this.error.bind(this),
                'view.non-number-error' : this.error.bind(this)
            });

            //Was calling find a lot so made it a variable
            this._$input = this._$el.find('input');
            this._inputEl = this._$input[0];
            this._$buttons = this._$el.find('button');
            this._$incButton = this._$buttons.filter('[data-action=increment]');
            this._$decButton = this._$buttons.filter('[data-action=decrement]');
            this.min(this._options.min);
            this.max(this._options.max);
            this._triggerChange = false;

            //Update the view if the model has some data
            if (!_.isNull(this._controller.data())) {
                this.update(this._controller.data());
            } else if(!this._controller.isValid()) {
                this.error(this._controller.error().data);
            }

            this._$input.on('change', function (event, data) {
                data = data || { update : true };
                this._triggerChange = true;

                if (data == null || data.update) {
                    that._controller.data(this.value);
                }
            });


            this._$input.on('click', function () {
                if (that._options.type !== 'number' &&
                    that._options.type !== 'currency' &&
                    that._options.spinner.controls) {
                    fieldChangeHandler();
                }
            });

            var repeaters = {
                increment : repeat(_.bind(this.increment, this)),
                decrement : repeat(_.bind(this.decrement, this))
            };

            this._$buttons.on('mousedown', function (event) {
                var repeater = repeaters[$(this).attr('data-action')];
                event.preventDefault(); //Keep focus where it is

                repeater.start();
                $('body').on('mouseup.oui-spinner mouseleave.oui-spinner', function () {
                    repeater.stop();
                    $('body').off('mouseup.oui-spinner mouseleave.oui-spinner');
                });
            });

            if (this._options.spinner.scroll) {
                var eventName = 'wheel'; //Standardized event
                if(document.onmousewheel !== undefined) {
                    eventName = 'mousewheel'; //Only needed for Safari
                }
                this._$input.on(eventName, function (event) {

                    function getWheelDelta(e) {
                        if (!(e.wheelDelta || e.detail)) {
                            e = e.originalEvent;
                        }

                        if (e.deltaY) {
                            return -e.deltaY;
                        } else if (e.detail) {
                            return e.detail * -40;
                        } else if (e.wheelDelta) {
                            return e.wheelDelta;
                        }
                    }

                    //http://youmightnotneedjquery.com/#matches
                    if(that._$input[0] === document.activeElement &&
                       !(event.shiftKey ||
                         event.ctrlKey ||
                         event.metaKey ||
                         event.altKey)) {
                        event.preventDefault();

                        if (getWheelDelta(event) < 0) {
                            that.decrement();
                        } else {
                            that.increment();
                        }
                    }
                });
            }

            this._$input.on('keydown',function (event) {
                switch (event.keyCode) {
                case keyCode.UP:
                    if (that._options.spinner.controls) {
                        event.preventDefault();
                        that.increment();
                    }
                    break;
                case keyCode.DOWN:
                    if (that._options.spinner.controls) {
                        event.preventDefault();
                        that.decrement();
                    }
                    break;
                }
            })
            .on('keypress', function(event) {
                var input = that._$input[0];
                var caps = event.shiftKey;
                var str = that._inputEl.value
                            .split('');
                var character = String.fromCharCode(event.which);

                if(!caps) {
                    character = character.toLowerCase();
                }

                str.splice(input.selectionStart,
                                input.selectionEnd - input.selectionStart,
                                character);


                str = str.join('');

                if(_.isRegExp(that._options.restrictTo) && !that._options.restrictTo.test(str)) {
                    event.preventDefault();
                    that._options.restrictTo.lastIndex = 0;
                } else if (_.isFunction(that._options.restrictTo) && !that._options.restrictTo(str)) {
                    event.preventDefault();
                }
            }).on('keyup', function (event) {
                if(that._options.spinner.controls) {
                    switch (event.keyCode) {
                    case keyCode.END:
                        event.preventDefault();
                        if(!_.isNull(that._options.min)) {
                            that._controller.data(that._options.min);
                        } else if (that._options.type === 'time') {
                            that._controller.data(that._controller.data().startOf('day'));
                        }
                        break;
                    case keyCode.HOME:
                        event.preventDefault();
                        if(!_.isNull(that._options.max)) {
                            that._controller.data(that._options.max);
                        } else if (that._options.type === 'time') {
                            that._controller.data(that._controller.data().endOf('day'));
                        }
                        break;
                    case keyCode.RIGHT:
                    case keyCode.LEFT:
                        fieldChangeHandler();
                        break;
                    }
                }
            });

            this._disabled = false;
        }

        NumberView.prototype = Object.create(View.prototype);
        nvp = NumberView.prototype;

        nvp.constructor = NumberView;

        nvp.nativeType = function(options) {
            if(options.type === 'number' || options.type === 'currency') {
                return 'number';
            } else if (options.type === 'time') {
                return options.type;
            }
            return null;
        };

        nvp.destroy = function (clean) {
            //Remove all events on destroy
            //TODO: only remove oui namespaced events
            this._$el.add(this._$input).add(this._$buttons).off();

            if (clean.self) {
                this._$el.remove();
            }
        };

        nvp._getField = function (loc, value) {
            var matches = TIME_REGEX.exec(value) || [ 0, '', '', '' ],
                hours = 0,
                minutes = 0,
                seconds = 0;

            loc = loc || 0;

            if (_.isFunction(this._options.field)) {
                return this._options.field(loc, value);
            }

            hours = matches[1].length + 1;
            minutes = hours + matches[2].length + 1;
            seconds = minutes + (matches[3] || '').length + 1;

            if (loc < hours) {
                return 'hours';
            } else if (loc < minutes) {
                return 'minutes';
            } else if (loc < seconds) {
                return 'seconds';
            } else {
                return 'half-day';
            }
        };

        nvp.buildElement = function (opts, el) {
            var $element;

            if (_.isNull(el)) {
                //Build element from scratch
                $element = $(spinnerTemplate);
            } else if (el.jquery) {
                //We were given a jquery element
                $element = el.attr('type', 'text')
                    .wrap(wrapTemplate)
                    .parent();
            }

            if (opts.spinner.controls) {
                $element.append($(controlsTemplate({ locale : opts.locale })));
                $element.addClass('spinner');
                $element.children('input').attr('role', 'spinbutton');
                $element.append(instructions({
                    instructions : opts.locale.instructions.spinner,
                    name : 'spinner'
                }));
            }

            return $element;
        };

        nvp.buildNativeElement = function (options, el) {
            var $element;

            if(_.isNull(el)) {
                $element = $(nativeElement({ type : this.nativeType(options) }));
            } else if (el.jquery) {
                $element = el.attr('type', this.nativeType(options));
            }

            $element = $element.wrap(wrapTemplate)
                .parent();

            $element.find('input')
                .attr('max', options.max)
                .attr('min', options.min)
                .attr('step', options.step);

            return $element;
        };

        nvp.init = function () {
            var that = this;
            var input = this._$el.children('input');
            var instructions = this._$el.find('.instructions');

            var describedBy = _.reduce(instructions, function (memo, elem) {
                var id = that.id() + '-instructions-' + $(elem).attr('data-name');
                $(elem).attr('id', id);
                return memo + ' ' + id;
            }, '');
            input.attr('aria-describedby', ((input.attr('aria-describedby') || '')  + ' ' + describedBy).trim());
        };

        nvp.disable = function (disabled) {
            this._disabled = disabled;
            var els = this._$buttons
                .add(this._$input)
                .toggleClass('disabled', disabled);
            //With the prophooks - have to use setAttribute and removeAttribute
            //instead of the jQuery equivalent
            _.each(els, function (element) {
                if(disabled) {
                    element.setAttribute('disabled', 'disabled');
                } else {
                    element.removeAttribute('disabled');
                }
            });

        };

        nvp.error = function (attempted) {
            var currentValue = this._inputEl.value, newValue;
            this._$el.add(this._$el.closest('.control-group')).addClass('error');

            if(_.isNumber(attempted) || moment.isMoment(attempted)) {
                newValue = this._options.formatter.format(attempted, { truncate : false });
            } else if(_.isString(attempted)){
                newValue = attempted;
            } else if (attempted == null) {
                newValue = '';
            }

            if(newValue != null && newValue !== currentValue) {
                this._inputEl.value = newValue;
            }
        };

        nvp.min = function(min) {
            if(!_.isNull(min)) {
                this._$input.attr('aria-valuemin', min);
            } else {
                this._$input.removeAttr('aria-valuemin');
            }
        };

        nvp.max = function(max) {
            if(!_.isNull(max)) {
                this._$input.attr('aria-valuemax', max);
            } else {
                this._$input.removeAttr('aria-valuemax');
            }
        };

        nvp.increment = function () {
            if (!this._disabled) {
                var input = this._$input[0];
                this._triggerChange = false;
                this._controller.increment(this._getField(this._cursorLoc, this._$input.val()));
                input.selectionStart = this._cursorLoc;
                input.selectionEnd = this._cursorLoc;
            }
        };

        nvp.decrement = function () {
            if (!this._disabled) {
                this._triggerChange = false;
                var input = this._$input[0];
                this._controller.decrement(this._getField(this._cursorLoc, this._$input.val()));
                input.selectionStart = this._cursorLoc;
                input.selectionEnd = this._cursorLoc;
            }
        };

        nvp.update = function (value) {
            var update = this._controller.canUpdate(value, this._getField(this._cursorLoc, this._inputEl.value)),
                formattedValue, currentValue = this._inputEl.value;

            this._$el.add(this._$el.closest('.control-group')).removeClass('error');


            if (_.isNull(value)) {
                this._inputEl.value = '';
            } else {
                formattedValue = this._options.formatter.format(value, { truncate : false });
                this._inputEl.value = formattedValue;

                this._$input.attr('aria-valuenow', value);
                this._$input.attr('aria-valuetext', formattedValue);
            }

            //So the value was updated from what was currently in the input
            if(this._triggerChange && currentValue !== this._inputEl.value) {
                this._$input.trigger('change', { update : false });
                this._triggerChange = false;
            }

            if (!this._disabled) {
                this._$incButton.prop('disabled', !update.increment);
                this._$decButton.prop('disabled', !update.decrement);
            }
        };

        return NumberView;
    }
);

define('oui.spinner/number',['oui.underscore', 'oui.base/input', 'oui.spinner/numberModel', 'oui.spinner/numberController', 'oui.spinner/numberView', 'oui.logger'],
    function (_, Input, NumberModel, NumberController, NumberView, logger) {
        

        var nfp;

        /**
         * @callback NumberField#format
         * @param {moment} moment - The moment object to format
         * @returns {String}
         */

        /**
         * @callback NumberField#unformat
         * @param {String} str - The string representation of a date
         * @returns {moment}
         */

        /**
         * @callback NumberField#validate
         * @param {String} str - The string representation of a date
         * @returns {Boolean}
         */

        /**
         * @typedef {Object} NumberField#Config
         *
         * @property {null|RegExp|Function} [restrictTo=null] Option to determine whether new character should be added to
         * the number field. Fired on keyDown and if it fails the test or the function returns false the event will be
         * prevented.
         * @property {String} [type='number'] The type of data that the number field contains and will represent. Valid
         * options are 'number', 'currency', 'time', 'date'.
         * @property {Boolean} [utc=false] If using a 'time' type number field, then this determines whether the time should
         * be parsed as UTC time or local time. Default is for local time. **Note**: This option is unused and reserved for
         * future use.
         * @property {Object} [spinner] Options for spinner functionality of the number field.
         * @property {Boolean} [spinner.controls=false] Whether or not the increment/decrement arrows should be drawn on the
         * component.
         * @property {Number} [spinner.step=1] The default step value when incrementing and decrementing a value.
         * @property {Boolean} [spinner.scroll=true] Whether or not the spinner should support scroll events on the control. These
         * only get fired when the component has focus.
         * @property {Object} [formatter={}] - The formatter the number field will use when attempting to convert data
         * types. Used when attempting to display the date/number in the input and when extracting the date/number from the input.
         * @property {NumberField#format} [formatter.format]
         * @property {NumberField#unformat} [formatter.unformat]
         * @property {NumberField#validate} [formatter.validate]
         * @property {moment|Number|null} [max=null] The max value the number field can be or `null` for no max.
         * @property {moment|Number|null} [min=null] The min value the number field can be or `null` for no min.
         * @property {moment|Number|null} [emptyValue=null] The value the number field should become if it is not allowed to
         * be empty. If the number field is not allowed to be empty (no null value) then this option should be specified.
         * This can be used in conjunction with the {@link NumberField#Config.required} property.
         * @property {Boolean} [required=false] Whether the spinner is a required field or not.
         * @property {Boolean} [fallback=false] Whether the control should fallback to its native control. The control will
         * always fallback to the native control on a mobile device regardless of setting. **Note**: This property is
         * deprecated and should not be set.
         * @property {Boolean} [fluid=false] Whether the control should take up as much horizontal space as possible.
         */

        var __WIDGET_NAME__ = 'number',
            __defaults__ = {
                restrictTo : null,
                type : 'number',
                utc : false,
                spinner : {
                    controls : false,
                    step : 1,
                    scroll : true
                },
                formatter : {},
                max : null,
                min : null,
                emptyValue : null,
                required : false,
                fallback : false,
                fluid : false
            };

        /**
         * This class represents a number field which can handle both dates/times, and numbers.
         *
         * @class NumberField
         * @since 1.2.0
         * 
         * @param {NumberField#Config} [options] The widgets options.
         * @param {HTMLInputElement|jQuery} [element] The JQuery element to make a number field.
         * @param {moment|Number|null} [initData] The data to initialize the widget with.
         */
        function NumberField(options, element, initData) {
            var model, view, controller;

            //Super call
            Input.call(this, options, element, initData);

            model = new NumberModel(this._initData, this._options);
            controller = new NumberController(model, this._options);
            view = new NumberView(model, controller, this._element, this._options, this._settings);

            /**
             * The view that the field calls into to get the element.
             *
             * @property _view
             * @private
             * @type {NumberView}
             */
            this._view = view;
            /**
             * The controller that the field calls into to change the value
             * of the element.
             *
             * @property _controller
             * @private
             * @type {NumberController}
             */
            this._controller = controller;

            this._setupEvents([model, view]);
        }

        NumberField.prototype = Object.create(Input.prototype);
        nfp = NumberField.prototype;

        nfp.constructor = NumberField;

        /**
         * Gets the widget name
         *
         * @method NumberField#name
         * @private
         * @return {String} 'number' - widget name.
         */
        nfp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Please use [data](#method_data).
         *
         * @deprecated
         * @method NumberField#getData
         * @return {moment|Number|String|null} The current value of the number field.
         */
        nfp.getData = function () {
            logger.warn('The use of getData (and get/set methods in general) in OUI is now deprecated. Please use the combined method "data" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            return this.data();
        };

        /**
         * Please use [data](#method_data).
         *
         * @deprecated
         * @method NumberField#setData
         * @param {moment|Number|String|null} newData The new value of the number field.
         */
        nfp.setData = function (newData) {
            logger.warn('The use of setData (and get/set methods in general) in OUI is now deprecated. Please use the combined method "data" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            this.data(newData);
        };

        /**
         * Getter/setter for the data of the number field.
         *
         * @method NumberField#data
         * @fires NumberModel#non-number-error
         * @fires NumberModel#number-change
         * @fires NumberModel#bounds-error
         * @fires NumberModel#required-error
         * @param {moment|Number|String|null} [newData] The new value of the number
         *   field.
         * @return {moment|Number|String|null} The current value of the number field.
         */
        nfp.data = function (newData) {
            if (!arguments.length) {
                if(this.isValid()) {
                    return this._controller.data();
                } else {
                    return this._controller.error().data;
                }
            }

            this._controller.data(newData);
        };

        /**
         * Getter/setter for the data of the number field.
         *
         * Similar to {{#crossLink "NumberField/data"}}{{/crossLink}} except it does not
         * trigger an event after selection.
         *
         * @method NumberField#value
         * @param {moment|Number|String|null} [newData] The new value of the number
         *  field.
         * @return {moment}
         */
        nfp.value = function(newData) {
            if(!arguments.length) {
                return this.data();
            }

            this._controller.data(newData, { silent : true });
        };

        /**
         * Getter/Setter for the formatter.
         *
         * Rerenders the view when the formatter is reset.
         *
         * @method NumberField#formatter
         * @param {Object} formatter A new formatter to use for the existing number spinner.
         * @return {Object} - current formatter used by the number spinner.
         */
        nfp.formatter = function (formatter) {
            if (!arguments.length) {
                return this._options.formatter;
            }

            this._options.formatter = formatter;
            this._view.update(this._controller.data());
            this._view.trigger('reformat');
        };

        /**
         * Getter/Setter for the min value of the spinner.
         *
         * When using the setter it will rerender the view using the new min value.
         *
         * @method NumberField#min
         * @param {moment|number} min The new min to set
         * @return {moment|number} - The current min
         */
        nfp.min = function (min) {
            if (!arguments.length) {
                return this._options.min;
            }

            this._options.min = min;
            this._view.min(min);
        };

        /**
         * Getter/Setter for the max value of the spinner.
         *
         * When using the setter it will rerender the view using the new max value.
         *
         * @method NumberField#max
         * @param {moment|number} max The new max to set
         * @return {moment|number} - The current max
         */
        nfp.max = function (max) {
            if (!arguments.length) {
                return this._options.max;
            }

            this._options.max = max;
            this._view.min(max);
        };

        /**
         * Gets the last error of the number field.
         *
         * @method NumberField#error
         * @return {OUIError|null} error The last error of the field.
         */
        nfp.error = function () {
            return this._controller.error();
        };

        /**
         * Get the validity state of the number field.
         *
         * @method NumberField#isValid
         * @return {Boolean} - Whether the field is currently in a valid state.
         */
        nfp.isValid = function () {
            return this._controller.isValid();
        };

        /**
         * Programmatically decrement the number field.
         *
         * Field is not required but is only used when the
         * number field is of type <code>'time'</code>.
         *
         * If no field is given it defaults to 'Hours'.
         *
         * Valid fields are:
         *
         *     'Hours',
         *     'Minutes',
         *     'Seconds',
         *     'half-day'
         *
         * @method NumberField#decrement
         * @param {String} [field] The field to increment.
         */
        nfp.decrement = function (field) {
            this._controller.decrement(field);
        };

        /**
         * Programmatically increment the number field.
         *
         * Field is not required but is only used when the
         * number field is of type <code>'time'</code>.
         *
         * If no field is given it will default to hours.
         *
         * Valid fields are:
         *
         *     'Hours',
         *     'Minutes',
         *     'Seconds',
         *     'half-day'
         *
         * @method NumberField#decrement
         * @param {String} [field] The field to increment.
         */
        nfp.increment = function (field) {
            this._controller.increment(field);
        };

        /**
         * @method NumberField#defaults
         * @return {Object} The default options for the widget.
         */
        nfp.defaults = function () {
            return __defaults__;
        };

        return NumberField;

    }
);

define('oui.mixins/popup',['oui.jquery', 'oui.underscore', 'oui.logger'], function($, _, logger) {
    

    /**
     * This is meant to be a mixin to provide common popup positioning.
     *
     * **NOT** meant to be constructed like the OUI objects.
     *
     * @mixin popup
     * @private
     */
    return {
        _positionPopup : function($relativeElement, parentIsBody, alignment) {
            var top, left, position,
                offset = $relativeElement.offset(),
                popupHeight = this._$popup.outerHeight(true),
                popupWidth = this._$popup.width();

            alignment = alignment || 'left';

            //Remove this if it was once positioned above the element
            this._$popup.removeClass('up');

            //Determine base relative element coordinates
            if (parentIsBody) {
                position = $relativeElement.offset();
            } else if ($relativeElement.parent().length !== 0){

                var parent = $relativeElement.parent();
                position = $relativeElement.position();

                if(!parent.offsetParent().is(this._$popup.offsetParent()) &&
                  parent.offsetParent().find(this._$popup.offsetParent()).length === 0) {
                    var additionalPosition = parent.offsetParent().position();

                    position.top = position.top + additionalPosition.top;
                    position.left = position.left + additionalPosition.left;
                }

            } else {
                position = $relativeElement.position();
            }

            //Set up base css top/left coordinates
            top = position.top + $relativeElement.outerHeight();
            left = position.left;

            if(alignment === 'right') {
                left = position.left + ($relativeElement.innerWidth() - popupWidth);
            }

            //Correct for window dimensions
            if (offset.top + popupHeight + $relativeElement.outerHeight() > window.pageYOffset + $(window).height()) {
                top -= (popupHeight + $relativeElement.outerHeight());
                this._$popup.addClass('up');
            }

            if (offset.left + popupWidth > window.pageXOffset + $(window).width()) {
                left -= (popupWidth - $relativeElement.innerWidth());
            }

            //Set the coordinates
            this._$popup.css({
                'top': top,
                'left': left
            });
        },
        _popupZindex : function($relativeElement, parentIsBody) {
            //Find the max z-index in the parent chain (only the parent chain) which
            //means this could conflict with a sibling element.
            if (parentIsBody) {
                var currentZindex = parseInt(this._$popup.css('z-index'), 10);
                var max = _.chain($relativeElement.parents())
                    .map(function (parent) {
                        return parseInt($(parent).css('z-index'), 10) || 0;
                    })
                    .max()
                    .value();

                if(!_.isNaN(currentZindex) && currentZindex <= max) {
                    this._$popup.css('z-index', max + 1);
                }
            }
        },
        /**
         * Common showing logic for a popup. It takes care of positioning, z-indexing, aria attributes.
         *
         * Defaults to below the component and is positioned above or to the side if the window dimensions are too
         * small.
         *
         * @method popup#showPopup
         * @fires popup#showing
         * @fires popup#shown
         * @param {jQuery} $relativeElement The element that the popup is to be positioned around.
         */
        showPopup : function($relativeElement, alignment) {
            if(!_.isUndefined(this._$popup)) {
                var parentIsBody;

                /**
                 * Fired before the popup is shown.
                 *
                 * @event popup#showing
                 */
                this._$popup.trigger('showing');

                //If the popup is not appended - append it
                if(this._$popup.parent().length === 0 && this._options && this._options.appendTo) {
                    $(this._options.appendTo).append(this._$popup);
                }

                parentIsBody = this._$popup.parent().is('body');

                this._$popup.attr('aria-hidden', false);

                this._$popup.show();
                this._$el.addClass('open');
                this._positionPopup($relativeElement, parentIsBody, alignment);
                this._popupZindex($relativeElement, parentIsBody);

                /**
                 * Fired after the popup is shown.
                 *
                 * @event popup#shown
                 */
                this._$popup.trigger('shown');
            } else {
                logger.warn('_$popup is not defined.');
            }
        },
        /**
         * Hides the popup and fires events for the view to listen to.
         *
         * @method popup#hidePopup
         * @fires popup#hiding
         * @fires popup#hidden
         */
        hidePopup : function() {
            if(!_.isUndefined(this._$popup)) {
                /**
                 * Fired before the popup is hidden
                 *
                 * @event popup#hiding
                 */
                this._$popup.trigger('hiding');
                this._$popup.hide();
                this._$el.removeClass('open');
                this._$popup.attr('aria-hidden', true);
                /**
                 * Fired after the popup is hidden.
                 *
                 * @event popup#hidden
                 */
                this._$popup.trigger('hidden');
            }
        },
        /**
         * @method popup#isVisible
         */
        isVisible : function() {
            return this._$el.hasClass('open');
        }
    };
});

define('oui.date/dateView',['oui.underscore', 'oui.jquery', 'oui.base/view', 'oui.utils/uniqueId',
    'oui.date/datePickerTemplate.htm', 'oui.date/buttonTemplate.htm',
    'oui.date/wrapTemplate.htm', 'oui.date/calendarTemplate.htm',
    'oui.template/nativeElement.htm',
    'oui.spinner/number', 'oui.utils/keyCode', 'oui.utils/dateHelper', 'oui.moment', 'oui.mixins/popup'],
    function (_, $, View, uniqueId, datePickerTemplate, buttonTemplate, wrapTemplate, calendarTemplate, nativeElement, NumberField, keyCode, dateHelper, moment,
              popupMixin) {
        

        var dvp;

        function extractDate(element) {
            return moment(parseInt(element.attr('data-date'), 10));
        }

        var nativeFormats = {
            month : 'YYYY-MM',
            datetime : 'YYYY-MM-DDTHH:mm:ssZ',
            date : 'YYYY-MM-DD'
        };

        /**
         * @class DateView
         * @private
         * @extends View
         * @mixes popup
         * @param {DateModel} model
         * @param {DateController} controller
         * @param {jQuery|null} element
         * @param {DateField#Config} options
         */
        function DateView(model, controller, element, options, settings) {
            var that = this;

            this._id = uniqueId('dateView');
            this._visited = false;
            //Super call
            View.call(this, options, element, controller, settings);

            //TODO: This shouldn't be necessary
            /**
             * @property _model
             * @private
             * @type {DateModel}
             */
            this._model = model;

            //Bind events on the model
            model.on({
                'view.date-change' : { fn : this.select, context : this },
                'view.invalid-error' : { fn : this.error, context : this },
                'view.required-error' : { fn : this.required, context : this},
                'view.deselect' : { fn : this.select, context : this },
                'view.bounds-error' : { fn : this.error, context : this },
                'view.unselectable-error' : { fn : this.error, context : this }
            });

            this.hideCal = true;
            this.calendar = null;
            this.spinner = null;
            this._calendarStartDate = null;
            this._triggerChange = false;

            //Useful instead of using find everywhere
            this._$button = this._$el.find('button');
            this._$input = this._$el.find('input');
            this._inputEl = this._$input[0];

            //Bind events to the elements
            this._$button
                .on('click',function (e) {
                    e.preventDefault();
                    that.toggleCalendar();
                }).hover(function () {
                    that.hideCal = false;
                }, function () {
                    that.hideCal = true;
                })
                .on('mousedown', function(event) {
                    event.preventDefault();
                    that._$input.focus();
                });

            this._$input
                .on('focus', function () {
                    if(that._options.openOnFocus) {
                        that.show(that._model.date());
                    }
                })
                .on('blur', function () {
                    if (that.hideCal && !_.isNull(that.calendar)) {
                        that.hidePopup();
                    }
                    that._visited = true;

                    if(this.value === '' && that._options.required) {
                        that._controller.date(null);
                    }
                })
                .hover(function () {
                    that.hideCal = false;
                }, function () {
                    that.hideCal = true;
                })
                .on('change', function (e, data) {
                    data = data || {update : true};
                    that._triggerChange = true;

                    if (data == null || data.update) {
                        that._controller.date(this.value);
                    }
                });

            this.addShortcuts({
                //Opens the popup
                'F2 SHIFT-F10': function() {
                    if(!that.isDisabled() && !that.isReadOnly()) {
                        that.show(that._model.date());
                    }
                },
                //Closes the popup
                'ESC' : {
                    fn : this.cancel,
                    stopPropagation : this.isVisible
                },
                'ENTER' : {
                    fn : this.handleEnter,
                    prevention : this.isVisible
                },
                'SPACE' : {
                    fn : this.handleSpace,
                    prevention : this.isVisible
                },
                'TAB' : {
                    fn : this.handleTab,
                    stopPropagation : this.isVisible,
                    prevention : this.isVisible
                },
                //Move highlighted day
                'UP' : {
                    fn : _.partial(this.move, { week : -1 }),
                    prevention : this.isVisible
                },
                'DOWN' : {
                    fn : _.partial(this.move, { week : 1 }),
                    prevention : this.isVisible
                },
                'LEFT' : {
                    fn : _.partial(this.move, { day : -1 }),
                    prevention : this.isVisible
                },
                'RIGHT' : {
                    fn : _.partial(this.move, { day : 1 }),
                    prevention : this.isVisible
                },
                'PAGEUP' : _.partial(this.move, { month : -1 }),
                'PAGEDOWN' : _.partial(this.move, { month : 1 }),
                'HOME' : {
                    fn : function () {
                        this.jumpTo('first-day');
                    },
                    prevention : this.isVisible
                },
                'END' : {
                    fn : function () {
                        this.jumpTo('last-day');
                    },
                    prevention : this.isVisible
                },
                'alt-PAGEUP' : _.partial(this.move, { year : -1 }),
                'alt-PAGEDOWN' : _.partial(this.move, { year : 1 })
            });
        }

        DateView.prototype = Object.create(View.prototype);
        dvp = DateView.prototype;

        dvp.constructor = DateView;

        _.extend(dvp, popupMixin);

        dvp.nativeType = function (options) {
            return options.type;
        };

        /**
         * Provides view specific destroy logic.
         *
         * Removes event listeners and elements.
         *
         * @method DateView#destroy
         */
        dvp.destroy = function (cleanUp) {

            //Calendar should probably be appended to the DOM on create
            if (!_.isNull(this.calendar)) {
                this.calendar.find('button td').off();
                if (!_.isNull(this.spinner)) {
                    this.spinner.destroy();
                }
                this.calendar.remove();
            }

            //Who needs to specify events when off and no parameters will remove everything
            //http://api.jquery.com/off
            this._$input.off();
            this._$button.off();
            this._$el.off();

            if (cleanUp.self) {
                this._$el.remove();
            }
        };

        /**
         * @method DateView#show
         */
        dvp.show = function (date) {
            if(this.calendar == null && !this._$el.hasClass('native')) {
                this.calendar = this._$popup = this.buildCalendar();

                $(this._options.appendTo).append(this._$popup);
            }

            date = date || this._model.date();
            if(date !== undefined) {
                this.showDate(date);
            }

            if (this._options.type === 'datetime') {
                this.spinner.value(this._model.date());
            }

            this.showPopup(this._$el);
        };

        /**
         * @method DateView#handleEnter
         */
        dvp.handleEnter = function() {
            if(this.isVisible()) {
                this._triggerChange = true;
                //Submit;
                var active = this._$popup.find('.active');
                var time = this.spinner != null ? this.spinner.data() : null;

                if(!active.hasClass('inactive')) {
                    this._controller.date(extractDate(active), time);
                }
                this.hidePopup();
            }
        };

        //TODO: look into how similar this is to ENTER
        /**
         * @method DateView#handleSpace
         */
        dvp.handleSpace = function() {
            if(this.isVisible()) {
                //Submit;
                var active = this._$popup.find('.active');
                var time = this.spinner != null ? this.spinner.data() : null;

                if(!active.hasClass('inactive')) {
                    this._controller.date(extractDate(active), time);
                }
            }
        };

        /**
         * @method DateView#handleTab
         */
        dvp.handleTab = function() {
            if(this.isVisible() && this._options.type === 'datetime') {
                this.hideCal = false;
                //Navgiate to lower views (time-spinner)
                this.calendar.find('input').focus();
            }
        };

        /**
         * @method DateView#move
         */
        dvp.move = function ( changes ) {
            if(!_.isNull(this._$popup) && this.isVisible()) {
                var active = this._$popup.find('.active'),
                    date = extractDate(active),
                    newActive;

                newActive = active;

                _.each(changes, function(changeBy, field) {
                    date.add(field, changeBy);
                });

                newActive = this._$popup.find('[data-date=' + date.valueOf() + ']');

                if(newActive.hasClass('next-month') || newActive.hasClass('prev-month')) {
                    //Shift months;
                    date = extractDate(newActive);
                    this.showDate(date, date);
                } else if (newActive.length === 0 && active.length === 1) {

                    if ( this._model.isInBounds(date) ) {
                        this.showDate(date, date);
                    }
                } else if(this._controller._model.isInBounds(extractDate(newActive))) {
                    this.active(newActive);
                }
            }
        };

        /**
         * @method DateView#jumpTo
         */
        dvp.jumpTo = function ( clz ) {
            if(!_.isNull(this._$popup) && this.isVisible()) {
                if(clz === 'first-day') {
                    this.active(this._$popup.find('.current-month').first());
                } else if (clz === 'last-day' ) {
                    this.active(this._$popup.find('.current-month').last());
                }
            }
        };

        /**
         * Constructs the view specific elements.
         *
         * @method DateView#buildElement
         * @param {Object} options
         * @param {Object} el
         */
        dvp.buildElement = function (options, el) {
            var $element;

            if (_.isNull(el)) {
                //Build element from scratch
                $element = $(datePickerTemplate)
                            .append(buttonTemplate({ locale : options.locale }));
            } else if (el.jquery) {
                $element = el.attr('spellcheck', false)
                    .wrap(wrapTemplate)
                    .parent()
                    .append(buttonTemplate({ locale : options.locale }));
            }

            $element.find('input')
                .attr('type', 'text')
                .attr('required', options.required)
                .attr('aria-required', options.required)
                .attr('aria-haspopup', true)
                .attr('aria-expanded', false)
                .attr('aria-activedescendant', $element.find('input').attr('id'));

            return $element;
        };

        /**
         * @method DateView#buildNativeElement
         */
        dvp.buildNativeElement = function (options, el) {
            var $element;

            if(_.isNull(el)) {
                $element = $(nativeElement({ type : options.type }))
                        .wrap(wrapTemplate)
                        .parent();
            } else if (el.jquery) {
                $element = el
                    .attr('type', options.type)
                    .wrap(wrapTemplate)
                    .parent();
            }

            if(!_.isNull(options.max)) {
                $element.find('input').attr('max', moment(options.max).format(nativeFormats[options.type]));
            }

            if(!_.isNull(options.min)) {
                $element.find('input').attr('min', moment(options.min).format(nativeFormats[options.type]));
            }

            return $element;
        };

        /**
         * @method DateView#required
         */
        dvp.required = function () {
            if(this._visited) {
                this.error.apply(this, arguments);
            }
        };

        /**
         * @method DateView#markVisited
         */
        dvp.markVisited = function () {
            this._visited = true;
        };

        /**
         * @method DateView#error
         */
        dvp.error = function (current, attempted) {
            var currentValue = this._inputEl.value, newValue;
            this._$el.add(this._$el.closest('.control-group')).addClass('error');

            if(moment.isMoment(attempted)) {
                newValue = this._options.formatter.format(attempted);
            } else if(_.isString(attempted)){
                newValue = attempted;
            } else if (attempted == null) {
                newValue = '';
            }

            if(newValue != null && newValue !== currentValue) {
                this._inputEl.value = newValue;
            }

        };

        /**
         * Toggles whether the calendar is visible.
         *
         * @method DateView#toggleCalendar
         */
        dvp.toggleCalendar = function () {
            if (this.isVisible()) {
                this.hidePopup();
            } else {
                this.show(this._model.date());
            }
        };

        //TODO: should these happen in the view or should the model/controller
        //fire some events saying that the buttons should be disabled.
        /**
         * Disables the next year and month as well as the previous year and month
         * when the date changes.
         *
         * @method DateView#toggleNextPrevButtons
         * @param {Date} date The date to determine whether the next/prev buttons should
         *    be disabled.
         */
        dvp.toggleNextPrevButtons = function (date) {
            var nextMonth = moment(date).add(1, 'month').startOf('month'),
                nextYear = moment(date).add(1, 'year').startOf('month'),
                prevMonth = moment(date).subtract(1, 'month').startOf('month'),
                prevYear = moment(date).subtract(1, 'year').startOf('month'),
                max, min;

            if (!_.isNull(this._options.max)) {
                max = moment(this._options.max);

                this.calendar.find('.next-buttons button:eq(0)').prop('disabled', nextMonth.isAfter(max));
                this.calendar.find('.next-buttons button:eq(1)').prop('disabled', nextYear.isAfter(max));
            } else {
                this.calendar.find('.next-buttons button:eq(0)').prop('disabled', false);
                this.calendar.find('.next-buttons button:eq(1)').prop('disabled', false);
            }

            if (!_.isNull(this._options.min)) {
                min = moment(this._options.min);
                min.startOf('month');

                this.calendar.find('.prev-buttons button:eq(1)').prop('disabled', prevMonth.isBefore(min));
                this.calendar.find('.prev-buttons button:eq(0)').prop('disabled', prevYear.isBefore(min));
            } else {
                this.calendar.find('.prev-buttons button:eq(1)').prop('disabled', false);
                this.calendar.find('.prev-buttons button:eq(0)').prop('disabled', false);
            }
        };

        /**
         * The handler for the select event that is fired by the model.
         *
         * @method DateView#select
         * @param {Date} date The date that is selected.
         */
        dvp.select = function (date) {
            this.changeValue(date);
        };

        /**
         * @method DateView#hide
         */
        dvp.hide = function () {
            this.hidePopup();
            this.hideCal = true;
        };

        /**
         * Handles updating the input value and then triggering a change.
         *
         * @method DateView#changeValue
         * @param {Date} date The date to change the input to and fire a change event.
         */
        dvp.changeValue = function (date) {
            var currentValue = this._inputEl.value;

            this._$el.add(this._$el.closest('.control-group')).removeClass('error');
            if (_.isNull(date)) {
                this._inputEl.value = '';
            } else if (this._$el.hasClass('native')) {
                this._inputEl.value = date.format(nativeFormats[this._options.type]);
            } else {

                if (!_.isNull(this.calendar) && this.isVisible()) {
                    this.calendar.find('.selected').removeClass('selected');
                    this.calendar.find('[data-date=' + date.clone().startOf('day').valueOf() + ']').addClass('selected');
                }

                if (this.spinner != null && this._options.type === 'datetime') {
                    this.spinner.value(this._model.date());
                }

                if(moment.isMoment(date)) {
                    this._inputEl.value = this._options.formatter.format(date);
                } else if (_.isString(date)) {
                    this._inputEl.value = date;
                } //Otherwise - it should never have gotten here
            }

            if(this._triggerChange && currentValue !== this._inputEl.value) {
                this._$input.trigger('change', {update : false});
                this._triggerChange = false;
            }
        };

        /**
         * Initializes the widget after it is on the DOM since a widget can
         * be created before being on the DOM.
         *
         * Handles all fallback logic to the native HTML5 element.
         *
         * @method DateView#init
         */
        dvp.init = function () {
            //TODO: Move this to constructor - since the model is already set and created when passed into the view
            this.changeValue(this._model.date());

            if(!this._model.isValid()) {
                this.error(this._model.date(), this._model.error().data);
            }

            //TODO: find a way to set the ID before building the element?
            this._$el.find('.instructions').attr('id', this.id() + '-instructions');
            this._$input.attr('aria-describedby', ((this._$input.attr('aria-describedby') || '')  + ' ' + this.id() + '-instructions').trim());
        };

        /**
         * Handles rendering each cell as well as the title.
         *
         * @method DateView#showDate
         * @param {Date} date The date to build the calendar for.
         */
        dvp.showDate = function (date, activeDate) {
            var that = this, startMonth, startYear, startDate,
                today = moment().startOf('day'),
                selDate = this._model.date(),
                max = moment(this._options.max), min = moment(this._options.min),
                $cal = this.calendar,
                activeDay, title;

            if (_.isNull(date) || (!moment.isMoment(date) && !this._model.isValid())) {
                date = moment();
            }

            if(!this._model.isValid() || !this._model.selected()) {
                selDate = null;
            }

            //Only show valid dates - if the date given is out of bounds
            //then go to the bound it exceeds
            if (this._options.min != null && date.isBefore(min)) {
                activeDate = min;
                date = min;
            }

            if (this._options.max != null && date.isAfter(max)) {
                date = max;
                activeDate = max;
            }

            activeDate = activeDate || selDate || today;

            date = moment(date).startOf('month');
            startMonth = date.month();
            startYear = date.year();
            startDate = date.clone();

            activeDay = extractDate(this._$popup.find('.active')).date();

            if (_.isFunction(this._options.title)) {
                title = this._options.title(date);
                $cal.find('.title').text(title).prop('title', title);
            } else {
                title = this._options.locale.months[startMonth] + ' ' + startYear;
                $cal.find('.title').text(title).prop('title', title);
            }

            //Might move to another month so cant be chained until after startMonth and year are set.
            date.startOf('week');

            $cal.find('tbody td').each(function (i, td) {
                var $el = $(td)
                    .removeClass('selected active')
                    .text(date.date())
                    .attr({
                        'data-date': date.valueOf(),
                        'id': 'calendar-' + that._id + '-' + i,
                        'aria-label' : date.date(),
                        'aria-selected' : false
                    });

                if (that._model.selected() && dateHelper.dateComparator(date, selDate) === 0) {
                    $el.addClass('selected');
                    $el.attr('aria-selected', true);
                }

                if(dateHelper.dateComparator(date, activeDate) === 0) {
                    that.active($el);
                }

                $el.toggleClass('selected-month', that._model.selected() && date.month() === selDate.month() && date.year() === selDate.year());

                $el.toggleClass('today', that._options.type !== 'month' && dateHelper.dateComparator(date, today) === 0);

                $el.toggleClass('inactive',
                    ((!_.isNull(max) && date.isAfter(max)) ||
                    (!_.isNull(min) && date.isBefore(min)) ||
                    (!_.isUndefined(that._options.unselectable) && dateHelper.contains(that._options.unselectable, date))));

                $el.toggleClass('current-month', date.month() === startMonth);

                $el.toggleClass('prev-month', date.month() === ((startMonth - 1) % 12 + 12) % 12 );

                $el.toggleClass('next-month', date.month() === (startMonth + 1) % 12);

                date.add('day', 1);
            });

            if(this._$popup.find('.active').length === 0) {
                //No active was added - move to day that was active of previous calendar
                startDate.date(activeDay);
                this.active(this._$popup.find('[data-date=' + startDate.valueOf() + ']'));
            }

            this.toggleNextPrevButtons(startDate);
        };

        /**
         * @method DateView#active
         */
        dvp.active = function (activeElement) {
            this.calendar.find('.active').removeClass('active');

            this._$input.attr('aria-activedescendant', activeElement.attr('id'));
            activeElement.addClass('active');
        };

        /**
         * Adds the element to the DOM after it is built.
         *
         * @method DateView#buildCalendar
         */
        dvp.buildCalendar = function () {
            var that = this, currentDate = this._model.date(), calendar;

            function addHover (clazz) {
                calendar.on('mouseenter', clazz,function () {
                    calendar.find(clazz).addClass('hover');
                }).on('mouseleave', clazz, function () {
                    calendar.find(clazz).removeClass('hover');
                });
            }

            if (_.isNull(currentDate) || (!moment.isMoment(currentDate) && !this._model.isValid())) {
                currentDate = moment();
            }

            calendar = $(calendarTemplate({
                id : 'oui-calendar-label-' + this._id,
                locale : this._options.locale ,
                type : this._options.type
            }));

            //Position and show the calendar
            if (that._options.type === 'datetime') {
                //TODO: can this be moved out of here and into the DateField and inject it into the view
                //if it is a datetime date field.
                that.spinner = new NumberField(that._options.number, calendar.find('input'), this._model.date())
                                        .init();

                calendar.find('input').on('keydown', function (event) {
                    if(event.which === keyCode.TAB) {
                        event.preventDefault();
                        event.stopPropagation();
                        that._$input.focus();
                    } else if (event.which === keyCode.ENTER) {
                        event.preventDefault();
                        that._$input.focus();
                        that.handleEnter();
                    }
                }).on('change', function () {
                    that._triggerChange = true;
                    that._controller.date(that._controller.date(), that.spinner.data());
                });
            }

            calendar.attr('id', 'oui-calendar-' + this._id);
            that._$input.attr('aria-owns', calendar.attr('id'));

            if (that._options.type === 'month') {
                calendar.addClass('month');
            }

            calendar.hover(function () {
                that.hideCal = false;
            }, function () {
                that.hideCal = true;
            });

            addHover('.current-month');
            addHover('.prev-month');
            addHover('.next-month');

            calendar.on('click', '.button-controls button', function() {
                var field = $(this).attr('data-field'),
                    value = $(this).attr('data-value');

                var changes = {};
                changes[field] = value === 'next' ? 1 : -1;
                that.move(changes);
            }).on('click', '.today-cancel .today', function () {
                var date = moment(),
                    time;

                if (that._options.type === 'datetime') {
                    time = that.spinner.data() || moment();
                }

                that._triggerChange = true;
                that._controller.date(date.valueOf(), time || null);
                that.hidePopup();
            }).on('click', '.cancel', function () {
                that.cancel();
            }).on('blur', 'input', function () {
                if (that.hideCal) {
                    that.hidePopup();
                }
            }).on('showing', function () {
                if(that._model.isValid() && that._model.selected()) {
                    currentDate = that._model.date();
                    that._setCurrentDate(currentDate);
                } else {
                    currentDate = moment().startOf('day');
                    that._setCurrentDate(null);
                }
            }).on('shown', function () {
                that._$input.attr('aria-expanded', true);
                that._$input.attr('aria-activedescendant', that._$popup.find('.active').attr('id'));
            }).on('mousedown', function(event) {
                //Prevent default on any mousedown in the calendar except for inputs
                if($(event.target).parents('.orcl-number-field').length === 0) {
                    event.preventDefault();
                    //Keep focus on the date picker input if it isnt focused
                    if(!$(document.activeElement).is(that._$input)) {
                        that._$input.focus();
                    }
                }
            }).on('hidden', function () {
                that._$input.attr('aria-activedescendant', that._$input.attr('id'));
                that._$input.attr('aria-expanded', false);
            }).on('click', 'td:not(.inactive)', function (ev) {
                var time = null, timeMoment;
                ev.preventDefault();

                if (!(that._options.type === 'month' && $(this).hasClass('selected-month'))) {
                    if (that._options.type === 'datetime') {
                        timeMoment = that.spinner.data();

                        if(!_.isNull(timeMoment)) {
                            time = timeMoment.valueOf();
                        }
                    }

                    that._triggerChange = true;
                    that._controller.date($(this).data('date'), time || null);

                    if(that._options.type !== 'datetime') {
                        that.hidePopup();
                    }
                }

            }).on('dblclick', 'td:not(.inactive)', function (ev) {
                ev.preventDefault();
                that.hidePopup();
            });

            return calendar;
        };

        /**
         * Disables the input and button controls of the date view.
         *
         * @method DateView#disable
         * @param {Boolean} disable Disable or enable the view.
         */
        dvp.disable = function (disable) {
            this._disabled = disable;
            var els = this._$input.add(this._$button).toggleClass('disabled', disable);

            //With the prophooks - have to use setAttribute and removeAttribute
            //instead of the jQuery equivalent
            _.each(els, function (element) {
                if(disable) {
                    element.setAttribute('disabled', 'disabled');
                } else {
                    element.removeAttribute('disabled');
                }
            });
        };

        /**
         * @method DateView#isDisabled
         */
        dvp.isDisabled = function () {
            return this._disabled;
        };

        /**
         * @method DateView#_setCurrentDate
         */
        dvp._setCurrentDate = function (date) {
            this._calendarStartDate = date;
        };

        /**
         * @method DateView#cancel
         */
        dvp.cancel = function () {
            this._controller.date(this._calendarStartDate);
            this.hidePopup();
        };

        return DateView;
    }
);

define('oui.date/dateModel',['oui.underscore', 'oui.utils/subscribable', 'oui.utils/error', 'oui.utils/dateHelper', 'oui.moment'],
       function (_, Subscribable, OUIError, dateHelper, moment) {
    

    /**
     * Fired when the date field is not required, the current date is not null, and the new date is null.
     *
     * @event DateModel#deselect
     * @param {null} - The new date that the date field represents.
     */
    /**
     * Fired when the date field is required, the current date is not null, and the new date is null.
     *
     * @event DateModel#required-error
     * @param {moment|null} - The current date of the date field
     * @param {String} - The data that was attempted to be set on the date field.
     */
    /**
     * Fired when the new date is not a valid moment object.
     *
     * @event DateModel#invalid-error
     * @param {moment|Null} - The current date of the date field
     * @param {String} - The data that was attempted to be set on the date field.
     */
    /**
     * Fired when the new date exists in the unselectable dates array.
     *
     * @event DateModel#unselectable-error
     * @param {moment|Null} - The current date of the date field
     * @param {moment} - The date that resulted in the error
     */
    /**
     * Fired when the new date exists outside of the defined min and max bounds.
     *
     * @event DateModel#bounds-error
     * @param {moment|Null} - The current date of the date field
     * @param {moment} - The date that resulted in the error
     */
    /**
     * Fired when a valid date is supplied and it differs from the current date.
     *
     * @event DateModel#date-change
     * @param {moment|Null} - The current date of the date field
     */

    var dmp;

    /**
     * The class that holds the data used in the DateView.
     *
     * The date will never become invalid. It will always be the last valid value.
     *
     * @class DateModel
     * @extends Subscribable
     * @private
     * @param {Date|null} initDate The date to start the control at.
     * @param {DateField#Config} options The widget options.
     */
    function DateModel(initDate, options) {
        //Super call
        Subscribable.call(this);

        /**
         * The date of the model.
         *
         * @property _date
         * @private
         * @type {moment}
         * @default null
         */
        this._date = null;
        this._invalidData = null;
        this._lastErrorType = null;
        /**
         * The widget options.
         *
         * @property _options
         * @private
         * @type {DateField#Config}
         */
        this._options = options;

        //TODO: set up invalid data if given invalid data
        if(moment.isMoment(initDate) && !initDate.isValid()) {
            this.date(initDate._i);
        } else {
            this.date(initDate);
        }

    }

    DateModel.prototype = Object.create(Subscribable.prototype);
    dmp = DateModel.prototype;

    dmp.constructor = DateModel;

    /**
     * Utility function to return the bounds validity of the supplied date.
     *
     * @method DateModel#isInBounds
     * @return {Boolean} - Whether the moment is within the bounds of the date model
     */
    dmp.isInBounds = function (date) {
        var max = moment(this._options.max),
            min = moment(this._options.min);

        return (this._options.max == null || !max.isValid() || (date.isSame(max) || date.isBefore(max))) &&
            (this._options.min == null || !min.isValid() || (date.isSame(min) || date.isAfter(min)));
    };

    /**
     * Getter/Setter for the date property.
     *
     * @method DateModel#date
     * @param {moment} date The new date moment.
     * @param {Object} eventParams The parameters to pass to trigger. Can force an event to be silent.
     * @fires DateModel#deselect
     * @fires DateModel#required-error
     * @fires DateModel#invalid-error
     * @fires DateModel#unselectable-error
     * @fires DateModel#bounds-error
     * @fires DateModel#date-change
     * @return {moment} The current date of the model.
     */
    dmp.date = function (date, eventParams) {
        if (!arguments.length) {
            //Return a clone so caller can change without affecting model
            if (this.isValid()) {
                return (_.isNull(this._date) || !this._date.isValid()) ? null : this._date.clone();
            } else {
                return this._invalidData;
            }
        }

        if (_.isNull(date) && !this._options.required) {
            this._date = null;
            this._invalidData = null;
            this.trigger(eventParams, 'deselect', null);
        } else if (_.isNull(date)) {
            this._invalidData = undefined;
            this._lastErrorType = 'required-error';
            this.trigger(eventParams, 'required-error', this._date, null);
        } else if (_.isString(date) || !date.isValid()) {
            //Trigger an invalid-error
            this._invalidData = date._i || date;
            this._lastErrorType = 'invalid-error';
            this.trigger(eventParams, 'invalid-error', this._date, this._invalidData);
        } else if (!_.isUndefined(this._options.unselectable) &&
            dateHelper.contains(this._options.unselectable, date)) {
            this._invalidData = date;
            this._lastErrorType = 'unselectable-error';
            this.trigger(eventParams, 'unselectable-error', this._date, date);
        } else if (!this.isInBounds(date)) {
            //Trigger a bounds-error
            this._invalidData = date;
            this._lastErrorType = 'bounds-error';
            this.trigger(eventParams, 'bounds-error', this._date, date);
        } else if (_.isNull(this._date) || !this.isValid() || !moment(this._date).isSame(date)) {
            this._date = date;
            this._invalidData = null;
            this.trigger(eventParams, 'date-change', this._date);
            this.trigger(eventParams, 'select', this._date);
        } else if (this.isValid() && moment(this._date).isSame(date)) {
            this.trigger('view.date-change', this._date);
        }
    };

    /**
     * Getter for the validity status of the model.
     *
     * @method DateModel#isValid
     * @return {Boolean} - Whether the model is currently valid or not
     */
    dmp.isValid = function () {
        return _.isNull(this._invalidData);
    };

    /**
     * Getter for the current error of the model.
     *
     * @method DateModel#error
     * @return {OUIError|null} - Current error of the model. Null if there is no current error.
     */
    dmp.error = function () {
        if (!this.isValid()) {
            return new OUIError(this._lastErrorType, this._invalidData);
        } else {
            return null;
        }
    };

    /**
     * Getter for the current selected status of the model.
     *
     * Returns true if the model is valid and the current date is not null.
     *
     * @method DateModel#selected
     * @return {Boolean} - Whether the model is currently selected.
     */
    dmp.selected = function () {
        return this.isValid() && !_.isNull(this._date);
    };

    return DateModel;
});

define('oui.date/dateController',['oui.underscore', 'oui.moment'], function (_, moment) {
    

    /**
     * @class DateController
     * @private
     *
     * @param {DateModel} model The model that the controller uses to set and get data.
     * @param {DateFieldConfig} options The widgets options object.
     */
    function DateController(model, options) {

        /**
         * The model for the controller.
         *
         * @property _model
         * @private
         * @type {DateModel}
         */
        this._model = model;
        /**
         * The widgets options.
         *
         * @property _options
         * @type {Oject}
         * @private
         */
        this._options = options;
    }

    DateController.prototype = {
        constructor : DateController,
        /**
         * Getter/setter of the date in the model.
         *
         * @method DateController#date
         * @return {Date} The model's date.
         * @param {Date|String|Number} dateMs The date to update to.
         * @param {Date|String|Number} [timeMs] The time object if the date field
         *   includes time.
         */
        date : function (dateMs, timeMs) {
            var date, time;
            if (!arguments.length) {
                return this._model.date();
            }

            var eventParams;
            var newDate;

            if(timeMs != null && timeMs.silent) {
                eventParams = timeMs;
                timeMs = null;
            }

            //Default it to null if it is undefined
            timeMs = timeMs || null;
            dateMs = dateMs || null;

            if (_.isString(dateMs)) {
                if (this._options.formatter.validate(dateMs)) {
                    newDate = this._options.formatter.unformat(dateMs);
                } else {
                    newDate = dateMs;
                }
            } else if (_.isNull(dateMs)) {
                newDate = dateMs;
            } else {

                //Controller should handle date and time addition and not the model.
                date = moment(dateMs);

                if (!_.isNull(date)) {

                    //Since it is not possible to select the time from the picker - force it to the
                    //beginning of the date - regardless if it was given
                    if (!_.isNull(timeMs) && this._options.type === 'datetime') {
                        date.startOf('day');
                        time = moment(timeMs);
                        date.add('hour', time.hours());
                        date.add('minute', time.minutes());
                        date.add('seconds', time.seconds());
                    } else if(this._options.type !== 'datetime') {
                        date.startOf('day');
                    }
                }

                newDate = date;
            }

            this._model.date(newDate, eventParams);
        },
        /**
         * @method DateController#isValid
         */
        isValid : function () {
            return this._model.isValid();
        },
        /**
         * @method DateController#error
         */
        error : function () {
            return this._model.error();
        }
    };

    return DateController;
});

define('oui.date/dateField',['oui.underscore', 'oui.base/input', 'oui.date/dateView', 'oui.date/dateModel', 'oui.date/dateController', 'oui.spinner/number', 'oui.logger'],
    function (_, Input, DateView, DateModel, DateController, NumberField, logger) {
        

        /**
         * Enum for valid type values in the DateField
         *
         * @typedef {String} DateField#Types
         * @property {String} 'date'
         * @property {String} 'datetime'
         */

        /**
         * @callback DateField#format
         * @param {moment} moment - The moment object to format
         * @returns {String}
         */

        /**
         * @callback DateField#unformat
         * @param {String} str - The string representation of a date
         * @returns {moment}
         */

        /**
         * @callback DateField#validate
         * @param {String} str - The string representation of a date
         * @returns {Boolean}
         */

        var __WIDGET_NAME__ = 'date',
            dfp, //Used to shorten/store DateField.prototype
            __OVERRIDES__ = {
                number : {
                    type : 'time',
                    spinner : {
                        controls : true
                    },
                    fluid : true
                }
            },
            /**
             * @typedef {Object} DateField#Config
             * @property {Boolean} [openOnFocus=false] - Whether or not the date field should open the calendar when focused.
             * @property {Object} [locale={}] - The locale of the date field. See [OUI's lang function]{@link module:oui.lang}
             * for information on setting this for OUI globally.
             * @property {moment|null} [min=null] - The minimum value of the date field. If the value is null there is no
             * minimum for the date field.
             * @property {moment|null} [max=null] - The maximum value of the date field. If the value is null there is no
             * maximum for the date field.
             * @property {DateField#Types} [type='date'] - The type of date picker to be created. Valid values are 'date' and
             * 'datetime'.
             * @property {Boolean} [fallback=false] - **Deprecated**: Whether or not the date field should fallback to the
             * native control everywhere.
             * @property {Boolean} [required=false] - Whether the date field is required. If it is required there will be an
             * error trigger if left empty.
             * @property {Boolean} [fluid=false] - Whether the date field should take up the widget of its parent element.
             * @property {String|jQuery} [appendTo='body'] - Selector or jQuery element to append the calendar to.
             * @property {NumberField#Config} [number] - The number field properties. These get used when the type='datetime'
             * @property {Boolean} [utc=false] - **Note**: Reserved for future use.
             * @property {Object} formatter - The formatter the date field will use when attempting to convert data
             * types. Used when attempting to display the date in the input and when extracting the date from the input.
             * @property {DateField#format} [formatter.format]
             * @property {DateField#unformat} [formatter.unformat]
             * @property {DateField#validate} [formatter.validate]
             */
            __DEFAULTS__ = _.extend({
                openOnFocus : false,
                locale : {},
                min : null,
                max : null,
                type : 'date',
                fallback : false,
                required : false,
                fluid : false,
                appendTo : 'body',
                number : NumberField.prototype.defaults(),
                utc : false,
                formatter : {}
            }, __OVERRIDES__);

        /**
         * The common date input element. Displays a UI element that can accept a
         * typed date as well as through a selection on the calendar popup.
         *
         * Each date field consists of an input and a calendar button wrapped by a div with class 'orcl-date-picker'. This
         * class is expected to be used as a selector and will be supported. Appending anything inside this div is not
         * supported. The styles only expect there to be an input and a button with hidden screen reader instructions. Any
         * additional labels should be placed outside the '.orcl-date-picker'.
         *
         * See {@link DateField#Config} for more information regarding options for the Date Picker.
         *
         * @example
         * <caption>Example construction (accepts all defaults - adds required options)</caption>
         *
         * var dateField = new oui.DateField({
         *    formatter : {
         *        format : function (moment) {
         *            return moment.format('LLL');
         *        },
         *        unformat : function (str) {
         *            return moment(str);
         *        },
         *        validate : function (str) {
         *            return moment(str).isValid();
         *        }
         *    }
         * });
         * $container.append(dateField.element());
         * dateField.init();
         *
         * @example
         * <caption>Example construction using existing date input</caption>
         *
         * var dateField = new oui.DateField({
         *    formatter : {
         *        format : function (moment) {
         *            return moment.format('LLL');
         *        },
         *        unformat : function (str) {
         *            return moment(str);
         *        },
         *        validate : function (str) {
         *            return moment(str).isValid();
         *        }
         *    }
         * }, $date);
         * dateField.init();
         *
         * @class DateField
         * @extends Input
         * @since 1.2.0
         *
         * @param {DateField#Config} [options] The options for the DateField. If not
         *  specified, all the defaults will be used.
         * @param {jQuery} [element] The JQuery element that this field should use.
         *  If not specified, DateField will construct its own element.
         * @param {moment|null} [initDate] The date to initialize the field with.
         */
        function DateField(options, element, initDate) {
            var model;

            //Super constructor
            Input.call(this, options, element, initDate);

            model = new DateModel(this._initData, this._options);
            /**
             * The controller object that handles the programmatic input from
             * the DateField.
             *
             * @property _controller
             * @private
             * @type {DateController}
             */
            this._controller = new DateController(model, this._options);
            /**
             * The view element.
             *
             * @property _view
             * @private
             * @type {DateView}
             */
            this._view = new DateView(model, this._controller, this._element || null, this._options, this._settings);

            Object.defineProperty(this, '_data', {
                get : this.date,
                set : this.date
            });

            this._setupEvents(model);
        }

        DateField.prototype = Object.create(Input.prototype);
        dfp = DateField.prototype;

        dfp.constructor = DateField;

        /**
         * Returns the name of the widget.
         *
         * @method DateField#name
         * @returns {String} - 'date'
         */
        dfp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Getter/Setter for the DateField's date. If no parameter is passed it
         * will be a getter otherwise it is a setter.
         *
         * @method DateField#date
         * @since 1.3.0
         * @param {moment|null} [date] The date to set for the DateField.
         * @fires DateModel#deselect
         * @fires DateModel#required-error
         * @fires DateModel#invalid-error
         * @fires DateModel#unselectable-error
         * @fires DateModel#bounds-error
         * @fires DateModel#date-change
         * @return {moment|null} date The DateField's date when used as a getter.
         */
        dfp.date = function (date) {
            if (!arguments.length) {
                return this._controller.date();
            }
            this._view.markVisited();
            this._controller.date(date);
        };

        /**
         * Get the validity state of the date field.
         *
         * @method DateField#isValid
         * @return - Whether the date field is currently in a valid state.
         */
        dfp.isValid = function () {
            return this._controller.isValid();
        };

        /**
         * Gets the last error of the date field.
         *
         * @method DateField#error
         * @return {OUIError|null} The last error of the date field.
         */
        dfp.error = function () {
            return this._controller.error();
        };

        /**
         * Getter for the date of the DateField.
         *
         * @method DateField#getDate
         * @deprecated
         * @return {moment|null} date The DateField's date
         */
        dfp.getDate = function () {
            logger.warn('The use of getDate (and get/set methods in general) in OUI is now deprecated. Please use the combined method "date" instead.');
            //TODO: Add a specific link when it is written.
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            return this.date();
        };

        /**
         * Setter for the date of the DateField.
         *
         * @method DateField#setDate
         * @deprecated
         * @param {moment|null} date The date to set for the DateField.
         */
        dfp.setDate = function (date) {
            logger.warn('The use of setDate (and get/set methods in general) in OUI is now deprecated. Please use the combined method "date" instead.');
            //TODO: Add a specific link when it is written.
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            this.date(date);
        };

        /**
         * Getter/Setter for the min property of the DateField.
         *
         * @method DateField#min
         * @since 1.3.0
         * @param {moment|null} [date] The date to set for the DateField's min.
         * @return {Date} date The DateField's minimum date when used as a getter.
         */
        dfp.min = function (date) {
            if (!arguments.length) {
                return this._options.min;
            }
            this._options.min = date;
        };

        /**
         * Setter for the min property of the DateField.
         *
         * @method DateField#setMinDate
         * @since 1.3.0
         * @param {moment|null} [date] The date to set for the DateField's min.
         * @deprecated
         */
        dfp.setMinDate = function (date) {
            logger.warn('The use of setMinDate (and get/set methods in general) in OUI is now deprecated. Please use the combined method "min" instead.');
            //TODO: Add a specific link when it is written.
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            this.min(date);
        };

        /**
         * Getter/Setter for the max property of the DateField.
         *
         * @method DateField#max
         * @since 1.3.0
         * @param {moment|null} [date] The date to set for the DateField's max.
         * @return {moment} date The DateField's maximum date when used as a getter.
         */
        dfp.max = function (date) {
            if (!arguments.length) {
                return this._options.max;
            }
            this._options.max = date;
        };

        /**
         * Setter for the max property of the DateField.
         *
         * @method DateField#setMaxDate
         * @since 1.3.0
         * @param {moment|null} [date] The date to set for the DateField's max.
         * @deprecated
         */
        dfp.setMaxDate = function (date) {
            logger.warn('The use of setMaxDate (and get/set methods in general) in OUI is now deprecated. Please use the combined method "max" instead.');
            //TODO: Add a specific link when it is written.
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            this.max(date);
        };

        /**
         * Gets the default options for the date field.
         *
         * @method DateField#defaults
         * @private
         * @return {Object} The default options for the widget
         */
        dfp.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Getter/Setter for the value of the of the Date Field
         *
         * See {@link DateField#date} for information regarding the similar method.
         *
         * **Note**: does not trigger an event notifying of a change
         *
         * @method DateField#value
         * @param {moment|String|null} [val] The current value of the date field.
         * @return - the current value of the date field.
         */
        dfp.value = function (val) {
            if(!arguments.length) {
                return this.data();
            }

            this._controller.date(val, { silent : true });
        };

        /**
         * Getter/Setter for the required option of the date field
         *
         * Rerenders the view when reset.
         *
         * @method DateField#required
         * @param {Boolean} [req] - Whether or not the date field is required.
         * @returns {Boolean} - The current value of the required option.
         */
        dfp.required = function (req) {
            if(!arguments.length) {
                return this._options.required;
            }

            this._options.required = req;
            this.value(this._controller.date());
        };

        return DateField;
    }
);


define('oui.picker/pickerTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="orcl-picker"> <div class="search '+
((__t=( data.options.search ? '' : 'hidden' ))==null?'':__t)+
'"> </div> <div class="tree-content"> </div></div>';
return __p;
};});


define('oui.picker/pickerButtonTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button type="button" class="btn btn-picker" title="'+
((__t=( data.locale.picker.select  ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.picker.select ))==null?'':__t)+
'" > <i class="pgbu-icon-small pgbu-icon-picker-dots"></i></button>';
return __p;
};});


define('oui.picker/pickerInputTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="orcl-picker-input"> <input type="text"> <button type="button" tabindex="-1" class="btn btn-picker" title="'+
((__t=( data.locale.picker.select ))==null?'':__t)+
'"> <i class="pgbu-icon-small pgbu-icon-picker-dots"></i> </button></div>';
return __p;
};});


define('oui.picker/pickerWrapInputTemplate.htm',[],function () { return '<div class="orcl-picker-input">\n</div>\n';});


define('oui.picker/titleTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<label id="'+
((__t=( data.id ))==null?'':__t)+
'" for="'+
((__t=( data.treeId ))==null?'':__t)+
'" class="control-label">'+
((__t=( data.locale.title  ))==null?'':__t)+
'</label>';
return __p;
};});


define('oui.picker/pickerModalTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="modal pgbu-modal hide fade orcl-picker-modal" tabindex="-1" role="dialog" aria-hidden="true" aria-labelledby="'+
((__t=( data.modalId ))==null?'':__t)+
'"> <div class="modal-header"> <a title="'+
((__t=( data.locale.close ))==null?'':__t)+
'" class="modal-close" href="#" data-dismiss="modal"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="'+
((__t=( data.modalId ))==null?'':__t)+
'" class="modal-title">'+
((__t=( data.locale.title  ))==null?'':__t)+
'</h3> </div> <div class="modal-body"></div> <div class="modal-footer"> ';
 if(!data.required) { 
__p+=' <button type="button" title="'+
((__t=( data.locale.remove ))==null?'':__t)+
'" class="btn btn-blue-dark modal-clear pull-left" data-dismiss="modal">'+
((__t=( data.locale.remove  ))==null?'':__t)+
'</button> ';
 } 
__p+=' <button type="button" title="'+
((__t=( data.locale.cancel ))==null?'':__t)+
'" class="btn btn-blue-dark modal-cancel" data-dismiss="modal">'+
((__t=( data.locale.cancel  ))==null?'':__t)+
'</button> <button type="button" title="'+
((__t=( data.locale.ok ))==null?'':__t)+
'" class="btn btn-action modal-accept disabled" data-dismiss="modal" disabled="disabled">'+
((__t=( data.locale.ok  ))==null?'':__t)+
'</button> </div></div>';
return __p;
};});

define('oui.picker/pickerView',['oui.underscore', 'oui.jquery', 'oui.base/view',
    'oui.picker/pickerTemplate.htm',
    'oui.picker/pickerButtonTemplate.htm',
    'oui.picker/pickerInputTemplate.htm',
    'oui.picker/pickerWrapInputTemplate.htm',
    'oui.picker/titleTemplate.htm',
    'oui.picker/pickerModalTemplate.htm', 'oui.utils/keyCode', 'oui.bootstrap'],
    function (_, $, View, pickerTemplate, pickerButtonTemplate, pickerInputTemplate, pickerWrapInputTemplate, titleTemplate, pickerModalTemplate, keyCode) {
        

        /**
         * @event PickerView#cancel
         */
        /**
         * @event PickerView#close
         */

        var pvp;

        /**
         * @class PickerView
         * @private
         * @extends View
         * @since 1.3.0
         * 
         */
        function PickerView(options, element, treePanel, typeAhead, settings) {
            var that = this, model = {
                data : [],
                init : false
            };

            this._typeAhead = typeAhead;
            this._treePanel = treePanel;

            View.call(this, options, element, null, settings);

            function showModalHandler(event) {
                event.preventDefault();
                that._$modal.modal('show');
            }

            function isDataEqual(newData) {
                var selected = newData || treePanel.selected();

                if(model.data.length !== selected.length) {
                    return false;
                }

                return _.every(selected, function(node) {
                    return _.find(model.data, function (tree) {
                        return _.isEqual(tree, _.pick(node, _.keys(tree)));
                    });
                });
            }

            function resetHandler(data) {
                var disabled = isDataEqual(data);

                that._$modal.find('.modal-accept')
                    .prop('disabled', disabled)
                    .toggleClass('disabled', disabled);
            }

            function cancelHandler() {
                that.trigger('cancel');
                treePanel.selected(model.data);
                resetHandler();
            }

            //When the typeahead issues the search event forwards that to the
            //treePanel.
            typeAhead.on('search invalidate clear', function (query) {
                if (_.isUndefined(query) || _.isEmpty(query)) {
                    query = null;
                }
                treePanel.filter(query);
                that._filter(query);
            });

            treePanel.on('filter', function (results) {
                if(_.isNull(results)) {
                    typeAhead.invalidate();
                    that._filter(null);
                }
                if(that._$modal) {
                    that._$modal.find('.modal-accept').prop('disabled', treePanel.isFiltered());
                }
            });

            typeAhead.element().find('input').attr('aria-controls', treePanel.id());

            treePanel.on('invalidate', function () {
                model.data = [];
            });

            this.element().on('oui-select', function(event, widget, eventName, selected) {
                if(widget.name() === 'typeahead') {
                    treePanel.value(selected);
                    //TODO: triggering this event quite often
                    treePanel.trigger('accept', treePanel.selected());
                }
            });

            if(options.modal || _.isObject(options.modal)){

                if(_.isObject(options.modal)) {
                    this._$modal = options.modal;
                } else {
                    this._$modal = $(pickerModalTemplate({ modalId : 'oui-' + settings.name + '-' + settings.id + '-modal-label', required : options.required, locale : options.locale }));
                    this._$modal.find('.modal-body').append($(pickerTemplate({ options : options })));
                }

                this._$modal.on('hidden', function() {
                    that._$modal.detach();
                });

                this._$modal.on('show', function() {
                    if(that._$modal.parent().length === 0) {
                        $(options.appendTo).append(that._$modal);
                    }
                });

                if (this._$el.is('button')) {
                    this._$el.on('click', showModalHandler);
                } else {
                    this._$el.find('button').on('click', showModalHandler);
                    this._$el.on('keydown', function(event) {
                        var keys = event.keyCode === keyCode.F2 ||
                                (event.keyCode === keyCode.F10 && event.shiftKey);
                        if(keys && !that._$modal.hasClass('in')) {
                            showModalHandler(event);
                        }
                    });
                }

                this._$modal.find('.tree-content').append(treePanel.element());
                this._$modal.find('.search').append(typeAhead.element());

                treePanel.on('picker.select', function (data) {
                    resetHandler(data);
                });

                treePanel.on('accept', function(data) {
                    model.data = data;

                    if(!that._$el.hasClass('orcl-type-ahead') && _.isFunction(that._options.viewFns.display)) {
                        that._$el.find('input').val(that._options.viewFns.display(data));
                    }
                });

                treePanel.on('picker.accept', function(data) {
                    model.data = data;
                    resetHandler();
                });

                this._$modal.on('click', '.modal-accept', function () {
                    treePanel.trigger('accept', treePanel.selected());
                    that._$el.trigger('change');
                    resetHandler();
                }).on('click', '.modal-clear', function () {
                    treePanel.selected([]);
                    treePanel.trigger('accept', treePanel.selected());
                }).on('click', '[data-dismiss=modal]', function (e) {
                    e.stopPropagation();
                    e.preventDefault();
                    that._$modal.modal('hide');

                    if(!$(this).hasClass('modal-accept')) {
                        cancelHandler();
                    }
                }).on('keydown', function (e) {
                    switch (e.keyCode) {
                    case keyCode.ENTER:
                        e.preventDefault();
                        if(!that._$modal.find('.modal-accept').hasClass('disabled')) {
                            treePanel.trigger('accept', treePanel.selected());
                            that._$el.trigger('change');
                            that._$modal.modal('hide');
                        }
                        resetHandler();
                        break;
                    }
                }).on('keyup', function (e) {
                    switch (e.keyCode) {
                    case keyCode.ESC:
                        e.stopPropagation();
                        that._$modal.modal('hide');
                        cancelHandler();
                        break;
                    }
                })
                .on('shown.treepanel-init', function() {
                    treePanel.init();
                    typeAhead.init();
                    that._$modal.off('shown.treepanel-init');
                });

                treePanel.on('dblclick', function () {
                    if(!that._$modal.find('.modal-accept').hasClass('disabled')) {
                        treePanel.trigger('accept', treePanel.selected());
                        resetHandler();
                        that._$modal.modal('hide');
                    }
                });
            } else {
                this.inject(treePanel.element(), '.tree-content');
                this.inject(typeAhead.element(), '.search');
            }

            return this;
        }

        PickerView.prototype = Object.create(View.prototype);
        pvp = PickerView.prototype;

        pvp.constructor = PickerView;

        pvp.buildElement = function (options, el) {
            var $element;

            if (options.modal) {
                if (_.isNull(el) && options.buttonTriggerOnly) {
                    $element = $(pickerButtonTemplate({ locale : options.locale }));
                } else if (_.isNull(el)) {
                    $element = $(pickerInputTemplate({ locale : options.locale }));
                } else if (el.jquery && el.is('input')) {
                    //Input given
                    if(el.parent().hasClass('orcl-type-ahead')) {
                        el.parent().addClass('orcl-picker-input').addClass('orcl-picker');
                    } else {
                        el.wrap(pickerWrapInputTemplate);
                    }
                    $element = el.parent();
                    $element.append($(pickerButtonTemplate({ locale : options.locale })).attr('tabindex', '-1'));
                } else if (el.jquery && (el.is('button') || el.is('a'))) {
                    //Button given
                    $element = el;
                }

                if(!$element.hasClass('orcl-type-ahead')) {
                    $element.find('input').attr('readonly', true).attr('aria-haspopup', true);
                }
            } else if (!_.isNull(el)){
                return el.addClass('orcl-picker');
            } else {
                $element = $(pickerTemplate({ options : options})).addClass('orcl-picker');

                if(_.isString(options.locale.title)) {
                    $element.find('.tree-content').prepend(titleTemplate({ id : this._treePanel.id() + '-label', treeId: this._treePanel.id(), locale : options.locale}));

                    this._treePanel.element().attr('aria-labelledby', this._treePanel.id() + '-label');
                }
            }

            $element.addClass('orcl-picker');

            return $element;
        };

        pvp.init = function () {
            if(this._$el.hasClass('orcl-picker-input')) {
                var input = this._$el.children('input');
                this._$el.find('.instructions[data-name=' + this._settings.name + ']').attr('id', this.id() + '-instructions-' + this._settings.name);
                input.attr('aria-describedby', ((input.attr('aria-describedby') || '')  + ' ' + this.id() + '-instructions-' + this._settings.name).trim());
            }
        };

        pvp.destroy = function (clean) {
            if (this._options.modal) {
                //We have to get rid of the modal
                this._$modal.one('hidden shown', function () {
                    $(this).remove();
                });

                this._$el.find('button, .instructions').remove();
                this._$el.find('input').unwrap();
                //this._$modal.remove();
                this._$modal.modal('hide');
            }

            if (clean.self) {
                this._$el.remove();
            }
        };

        pvp.disable = function (disabled) {

            function disable (els) {
                _.each(els, function (el) {
                    if(disabled) {
                        el.setAttribute('disabled', 'disabled');
                        el.disabled = true;
                    } else {
                        el.removeAttribute('disabled');
                        delete el.disabled;
                    }
                });
            }

            var elems;

            if (this._options.buttonTriggerOnly) {
                elems = this._$el.toggleClass('disabled');

                disable(elems);
            } else if (this._options.modal) {
                elems = this._$el.find('input, button, .btn')
                    .toggleClass('disabled', disabled);

                disable(elems);
            } else {
                this._typeAhead.disable(disabled);
                this._treePanel.disable(disabled);
            }
        };

        pvp._filter = function (query) {
            if(query == null || query === '') {
                this._typeAhead.element().find('input').addClass('empty');
            } else {
                this._typeAhead.element().find('input').removeClass('empty');
            }
        };

        pvp.id = function (settings) {
            var picker = this._$el, pickerInput = this._$el.hasClass('orcl-picker-input'), element = picker;

            if(pickerInput) {
                element = picker.find('input');
            }

            if(!arguments.length) {
                return element.attr('id');
            }

            if(element.length > 0 && element.attr('id') == null) {
                if(_.isObject(settings)) {
                    element.attr('id', 'oui-' + settings.name + '-' + settings.id);
                } else if (_.isString('settings')) {
                    element.attr('id', settings);
                }
            }
        };

        pvp.show = function () {
            this._$modal.modal('show');
        };

        pvp.hide = function () {
            this._$modal.modal('hide');
        };

        return PickerView;
    }
);

define('oui.tree/tree',['oui.underscore', 'oui.jquery'], function (_, $) {
    

    /**
     * A typical tree data structure where each node is itself a tree.
     *
     * @example
     *     var tree = new Tree("1");
     *
     * @example
     *     var tree = new Tree("2", "1", { hasChildren : false, code : 11101 });
     *
     * @class Tree
     * @private
     * @since 1.2.0
     * 
     * @param {String}  id             The id for this tree
     * @param {String}  [parentId]     The id of the parent tree if it has one
     * @param {Object}  [data]         The object to hold all custom data
     * @param {Array}   [subTreeData]  The data to make up all of the subtrees
     * @param {Integer} [level]        The level in the hierarchy this node appears
     */
    function Tree(id, parentId, data, subTreeData, level, generator) {
        var that = this;

        /**
         * The id of the tree.
         *
         * @property _id
         * @type {String}
         * @private
         */
        this._id = id;
        /**
         * The id of the parent tree.
         *
         * @property _parentId
         * @type {String}
         * @private
         */
        this._parentId = parentId;
        /**
         * The custom data associated with this tree node.
         *
         * Possibly holds whether the tree contains children, codes, names, etc.
         *
         * @property _data
         * @type {Object}
         * @private
         */
        this._data = data;
        /**
         * The trees that are children of this tree
         *
         * @property _children
         * @type {Array}
         * @private
         * @default []
         */
        this._children = [];
        /**
         * Whether or not this tree has been loaded.
         * Used with lazy loading to determine if we need to get the data from the server.
         *
         * @property _loaded
         * @type {Boolean}
         * @default false
         * @readOnly
         * @private
         */
        this._state = 'unloaded';
        /**
         * The level in the tree this node appears.
         *
         * @property _level
         * @type {Integer}
         * @private
         */
        this._level = level;

        if (!_.isNull(subTreeData) && !_.isUndefined(subTreeData)) {
            if (_.isObject(subTreeData)) {
                subTreeData = _.toArray(subTreeData);
            }
            _.each(subTreeData, function (node) {
                that.addSubTree(new Tree(generator(), that.id, node, node.children || null, that.level + 1, generator));
            });
        }

        return this;
    }

    Tree.prototype = {
        constructor : Tree,
        /**
         * The getter for the id property.
         * Uses the ES5 getter syntax.
         *
         * @example
         *     console.log(tree.id);
         *
         * @method Tree#id
         * @return {String} id
         */
        get id() {
            return this._id;
        },
        /**
         * The getter for the parentId property.
         * Uses the ES5 getter syntax.
         *
         * @example
         *     console.log(tree.parentId);
         *
         * @method Tree#parentId
         * @return {String} parentId
         */
        get parentId() {
            return this._parentId;
        },
        /**
         * The getter for the custom data property.
         * Uses the ES5 getter syntax.
         *
         * @example
         *     console.log(tree.data);
         *
         * @method Tree#data
         * @return {Object} data
         */
        get data() {
            return this._data;
        },
        /**
         * The getter for the children property.
         * Uses the ES5 getter syntax.
         *
         * @example
         *     console.log(tree.children);
         *
         * @method Tree#children
         * @return {Array} children
         */
        get children() {
            return this._children;
        },
        /**
         * Add a subtree to the this tree.
         * Marks this tree as being loaded as well as having children
         *
         * @method Tree#addSubTree
         * @param {Tree} subTree The subtree to add to the children property
         */
        addSubTree : function (subTree, index) {
            if(_.isUndefined(index)) {
                this._children.push(subTree);
            } else {
                this._children.splice(index, 0, subTree);
            }
            this._state = 'loaded';
            this._hasChildren = true;
        },
        /**
         * Remove a subtree and all it's children from this tree.
         * Property has children becomes false if children is 0
         *
         * @method Tree#removeSubTree
         * @param {Tree} subTree The subtree to remove from children
         */
        removeSubTree : function (subTree) {
            this._children = _.reject(this._children, function (tree) {
                return tree === subTree;
            });
            if (this._children.length === 0) {
                this._hasChildren = false;
            }
        },
        /**
         * Get an descendent node by its id.
         *
         * @method Tree#getNodeById
         * @param {String} id
         * @return {Tree} node The descendent of the current tree.
         */
        getNodeById : function (id) {
            var node = null, i;
            if (this.id === id) {
                node = this;
            } else {
                node = _.find(this._children, function (child) {
                    return child.id === id;
                });

                if (_.isUndefined(node)) {
                    node = null;
                    for (i = 0; i < this._children.length; i++) {
                        node = this._children[i].getNodeById(id);
                        if (!_.isNull(node) && node.id === id) {
                            break;
                        }
                    }
                }
            }

            return node;
        },
        /**
         * Setter to set whether the tree is loaded or not.
         * Uses ES5 setter syntax.
         *
         * @example
         *     tree.loaded = true;
         *
         * @method Tree#loaded
         * @param {Boolean} newValue whether or not the tree is loaded
         */
        set state(newValue) {
            this._state = newValue;
        },
        /**
         * Whether or not this tree is loaded or not.
         * Uses ES5 getter syntax.
         *
         * @method Tree#loaded
         * @return true if the current tree is loaded or has children
         */
        get state() {
            return this._state;
        },
        set level(l) {
            this._level = l;

            _.each(this._children, function (child) {
                child.level = l + 1;
            });
        },
        get level() {
            return this._level;
        },
        toJSON : function () {
            var ret = $.extend(true, {}, this.data || {});

            if (this.children.length > 0) {
                ret.children = [];

                _.each(this.children, function (child) {
                    ret.children.push(child.toJSON());
                });
            }

            return ret;
        }
    };

    return Tree;

});

define('oui.tree/treeModel',['oui.underscore', 'oui.utils/subscribable', 'oui.tree/tree', 'oui.logger'], function (_, Subscribable, Tree, logger) {
    

    /**
     * Fired when a node is added to the tree.
     * @event TreeModel#node-added
     *
     * @param {Object} node The node that was added.
     * @param {Boolean} lazyLoad Whether lazy load is on or off.
     */
    /**
     * Fired when a node gets marked as active.
     * @event TreeModel#active
     *
     * @param {String} id The client-side id of the active node.
     */
    /**
     * Fires when the disabled nodes list changes at all.
     *
     * @event TreeModel#disabled-node
     *
     * @param {Array} disabledNodes The list of client-side ids of the disabled nodes.
     */
    /**
     * Fires when the expanded nodes list changes at all.
     *
     * @event TreeModel#expand
     *
     * @param {Array} expandedNodes The list of client-side ids of the expanded nodes.
     */
    /**
     * Fires when the expanded nodes is reduced in size.
     *
     * @event TreeModel#collapse
     * @param {String} id The client-side generated id of the node that collapsed.
     */
    /**
     * Fires when a node is removed from the tree model.
     * @event TreeModel#node-removed
     * @param {TreeModel} model The tree model.
     */
    /**
     * Fired when a node(s) gets selected.
     * @event TreeModel#select
     * @param {Array} nodes The array of client-side ids.
     */
    /**
     * Fired when a node(s) becomes unselected.
     * @event TreeModel#selectable
     * @param {Array} nodes The array of client-side ids of the unselectable nodes.
     */

    var tmp;

    function idGenerator() {
        var treeId = 0;

        return function () {
            return (++treeId).toString();
        };
    }

    /**
     * The model used for the tree panel.
     *
     * @class TreeModel
     * @private
     * @extends Subscribable
     * @since 1.2.0
     * 
     * @param {TreePanel#Config} options The options for the Tree Panel.
     * @param {Array}   initData    The initial data before any lazy loading takes place.
     */
    function TreeModel(options, initData) {
        var that = this, generator = idGenerator();

        function updateTreeHash(node) {
            that._treeHash[node.id] = node;
            if(node.id !== null && options.storeFns && _.isFunction(options.storeFns.attrs)) {
                var attributes = options.storeFns.attrs(node.data);
                _.each(attributes, function(attr) {
                    that.getAttributeArray(attr).push(node.id);
                });

                if(!options.multiSelect) {
                    that._selectedNodes = _.last(that._selectedNodes, 1);
                }
            }
            //Context will be undefined due to using _.each
            _.each(node.children, updateTreeHash);
        }

        /**
         * The root node for this model.
         *
         * @property _root
         * @type {Tree}
         * @private
         * @default Tree(null, undefined, null, groupedData);
         */
        this._root = new Tree(null, undefined, null, initData, 0, generator);

        Subscribable.call(this);

        /**
         * The list of currently expanded nodes.
         *
         * @property _expandedNodes
         * @type {Array}
         * @private
         * @default []
         */
        this._expandedNodes = [];
        /**
         * The list of currently selected nodes.
         *
         * @property _selectedNodes
         * @type {Array}
         * @private
         * @default []
         */
        this._selectedNodes = [];
        this._nonloadedSelectedNodes = [];

        /**
         * The currently active node. The active node is a node that is
         * currently highlighted on the UI as the node key actions behave on.
         *
         * @property _activeNode
         * @type {String}
         * @private
         * @default null
         */
        this._activeNode = null;
        /**
         * The list of unselectable nodes. Currently, if the parent id is included
         * in the list then the children are automatically unselectable as well.
         *
         * @property _unselectableNodes
         * @type {Array}
         * @private
         * @default []
         */
        this._unselectableNodes = [];
        /**
         * The list of disabled nodes. These are the nodes that have been selected
         * elsewhere and it can be thought of as gray-d out.
         *
         * @property _disabledNodes
         * @type {Array}
         * @private
         * @default []
         */
        this._disabledNodes = [];
        /**
         * A map of id to Tree.
         *
         * @property _treeHash
         * @private
         * @type {Object}
         * @default []
         */
        this._treeHash = {};
        this._generator = generator;
        this._filtered = false;
        this._filterMap = {};
        this._activeQuery = null;

        this._expandedNodes.push(null);

        if (!_.contains(_.keys(this._treeHash), 'null')) {
            this._treeHash['null'] = this._root;
        }

        if (!_.isNull(initData)) {
            updateTreeHash(this._root);
        }

        this._options = options;

        return this;
    }

    TreeModel.prototype = Object.create(Subscribable.prototype);
    tmp = TreeModel.prototype;

    tmp.constructor = TreeModel;

    tmp.getAttributeArray = function(name) {
        if(name === 'unselectable') {
            return this._unselectableNodes;
        } else if(name === 'disabled') {
            return this._disabledNodes;
        } else if( name === 'expanded') {
            return this._expandedNodes;
        } else if (name === 'selected') {
            return this._selectedNodes;
        }
    };

    tmp.addSubTree = function (parentNode, data, index) {
        var that = this;

        function updateTreeHash(node) {
            that._treeHash[node.id] = node;
            if(node.id !== null && that._options.storeFns && _.isFunction(that._options.storeFns.attrs)) {
                var attributes = that._options.storeFns.attrs(node.data);
                _.each(attributes, function(attr) {
                    that.getAttributeArray(attr).push(node.id);
                });

                if(!that._options.multiSelect) {
                    that._selectedNodes = _.last(that._selectedNodes, 1);
                }
            }
            //Context will be undefined after the call using _.each
            _.each(node.children, updateTreeHash);
        }

        if(_.isNumber(data)) {
            index = data;
            data = undefined;
        }

        if (_.isArray(parentNode)) {
            data = parentNode;
            parentNode = this._root;
        }

        _.each(data, function (node) {
            var tree, nonloadedNode;
            if (!(node instanceof Tree)) {
                tree = new Tree(that._generator(), parentNode.id, _.omit(node, 'children'), node.children, null, that._generator);
            } else {
                tree = node;
            }

            if (_.isNull(parentNode)) {
                that._root.addSubTree(tree, index);
                tree.level = that._root.level + 1;
            } else {
                parentNode.addSubTree(tree, index);
                tree.level = parentNode.level + 1;
            }

            nonloadedNode = _.find(that._nonloadedSelectedNodes, function (tree) {
                return _.isEqual(tree, _.pick(node, _.keys(tree)));
            });
            //If this new tree is in the nonloaded nodes selected data - select it
            if (!_.isUndefined(nonloadedNode)) {
                //Remove it from the nonloaded nodes - since it is selected
                that._nonloadedSelectedNodes = _.reject(that._nonloadedSelectedNodes, function (selNode) {
                    return _.isEqual(selNode, _.pick(node, _.keys(selNode)));
                });
                //Add it to the selectedNodes
                that._selectedNodes.push(tree.id);
            }

            updateTreeHash(tree);
            that.trigger('node-added', tree, that._options.lazyLoad);
        });
    };

    /**
     * Removes a subtree from the model. Will remove all children associated with it.
     *
     * @method TreeModel#removeSubTree
     * @param {String} id The id to remove.
     */
    tmp.removeSubTree = function (id) {
        var node = this.node(id), that = this;

        function updateTreeHash(tree) {
            delete that._treeHash[tree.id];

            var reject = function (node) {
                return node === tree.id;
            };

            that._selectedNodes = _.reject(that._selectedNodes, reject);
            that._disabledNodes = _.reject(that._disabledNodes, reject);
            that._expandedNodes = _.reject(that._expandedNodes, reject);
            if (tree.children.length > 0) {
                //Context will be undefined after the call with _.each
                _.each(tree.children, updateTreeHash);
            }
        }

        if (!_.isNull(id) && !_.isNull(node)) {

            if (_.isNull(node.parentId)) {
                this._root.removeSubTree(node);
            } else {
                this.node(node.parentId).removeSubTree(node);
            }

            updateTreeHash(node);
            this.trigger('node-removed', this);
        }
    };

    /**
     * Getter for the root.
     *
     * @method TreeModel#tree
     * @return {Tree} root The root node.
     */
    tmp.tree = function () {
        if (!arguments.length) {
            return this._root;
        }
        //No setter
        logger.warn('"tree" does not support the setting operation.');
    };

    /**
     * Adds an id to the expanded nodes list.
     *
     * @method TreeModel#expanded
     * @param {Array} ids The id to expand.
     */
    tmp.expanded = function (ids) {
        if (!arguments.length) {
            return this._expandedNodes;
        }

        this.trigger('collapse', _.difference(this._expandedNodes, ids));

        if (_.isArray(ids)) {
            this._expandedNodes = _.union([null], ids);
        } else if (_.isNull(ids)) {
            this._expandedNodes = [null];
        }

        this.trigger('expand', this._expandedNodes);
    };

    tmp.selected = function (nodes, eventParams) {
        var tree, that = this, selectedNodes = [], nonloadedSelectedNodes = [];

        if (!arguments.length) {
            return this._selectedNodes;
        }

        if(eventParams == null) {
            eventParams = {};
        }


        if (nodes.length > 1 && !this._options.multiSelect) {
            throw new Error('Selection contains more than 1 node');
        }

        nodes = _.difference(nodes, this._disabledNodes, this._unselectableNodes);

        _.each(nodes, function (node) {
            //TODO: when filtered on the view maybe use data attributes
            if (_.isString(node) && that._filtered) {
                //getNodeByPath will do checking of the node.id in case it was already on the client. Other wise use
                // the old approach.
                if (_.isObject(node) && that._options.recordKeys.id != null) {
                    node = node[that._options.recordKeys.id];
                }

                tree = _.chain(that._filterMap)
                                    .values()
                                    .flatten()
                                    .find(function(tree) {
                                        return tree.id === node;
                                    })
                                    .value();

                if (tree.id != null) {
                    selectedNodes.push(tree.id);
                } else if (_.isNull(tree) && that._options.lazyLoad) {
                    //It has not been loaded yet put it into the nonloadedSelectedNodes
                    nonloadedSelectedNodes.push(node);
                }
            } else if (_.isString(node)) {
                //Assuming this is a client-side id.
                selectedNodes.push(node);
            } else if (_.isObject(node)) {
                tree = that.getNodeByData(node);

                if (_.isNull(tree) && that._options.lazyLoad) {
                    //It has not been loaded yet put it into the nonloadedSelectedNodes
                    nonloadedSelectedNodes.push(node);
                } else if (_.isNull(tree)) {
                    logger.error('Node not found.');
                } else {
                    selectedNodes.push(tree.id);
                }
            }
        });

        if (!_.isEqual(selectedNodes, this._selectedNodes) || !_.isEqual(nonloadedSelectedNodes, this._nonloadedSelectedNodes)) {
            var unselectedNodes = _.difference(this._selectedNodes, selectedNodes);
            this._selectedNodes = selectedNodes;
            this._nonloadedSelectedNodes = nonloadedSelectedNodes;
            eventParams.name = 'select';
            this.trigger(eventParams, this._selectedNodes, unselectedNodes);
        }
    };

    tmp.nonLoadedSelected = function () {
        return this._nonloadedSelectedNodes;
    };

    /**
     * Sets the active node.
     *
     * @method TreeModel#active
     * @param {String} id The currently active node.
     */
    tmp.active = function (id) {
        if (!arguments.length) {
            return this._activeNode;
        }

        if (this._activeNode !== id) {
            this._activeNode = id;
            this.trigger('active', id);
        }
    };

    /**
     * Determines if a node is currently expanded or not.
     *
     * @method TreeModel#isExpanded
     * @param {String} id The id to check if it is contained in the expanded nodes.
     * @return true if the list of expanded nodes contains the id.
     */
    tmp.isExpanded = function (id) {

        return _.contains(this._expandedNodes, id);
    };

    /**
     * Determines if a node is currently selected or not.
     *
     * @method TreeModel#isSelecteded
     * @param {String} id The id to check if it is contained in the selected nodes.
     * @return true if the list of selected nodes contains the id.
     */
    tmp.isSelected = function (id) {
        return _.contains(this._selectedNodes, id);
    };

    /**
     * Determines if a node is disabled.
     *
     * @method TreeModel#isDisabled
     * @param {String} id The id to check if it is disabled or not.
     * @return true if there is no intersection between any of its ancestors and the disabled list.
     */
    tmp.isDisabled = function (id) {
        var that = this, parentId, path;


        function constructPath(curPath, curId) {
            if (curId !== null && _.contains(_.keys(that._treeHash), curId)) {
                curPath.push(curId);
                parentId = that._treeHash[curId].parentId;
                return constructPath(curPath, parentId);
            }
            return curPath;
        }

        if(this._disabledNodes.length > 0) {
            path = constructPath([], id);
            return  _.intersection(path, this._disabledNodes).length !== 0;
        } else {
            return false;
        }
    };

    /**
     * Determines if a node is selectable.
     *
     * @method TreeModel#isSelectable
     * @param {String} id The id to check if it is selectable or not.
     * @return true if there is no intersection between any of its ancestors and the unselectable list.
     */
    tmp.isSelectable = function (id) {
        return  !_.contains(this._unselectableNodes, id);
    };

    tmp.disabled = function (ids) {
        if (!arguments.length) {
            return this._disabledNodes;
        }


        if (_.isNull(ids)) {
            this._disabledNodes = [];
        } else {
            this._disabledNodes = ids;
        }

        this.trigger('disabled-node', this._disabledNodes);
    };

    /**
     * Sets a node to being unselectable. Removes all children nodes from the list.
     *
     * @method TreeModel#unselectable
     * @param {Array} ids The id to add to the unselectable list.
     */
    tmp.unselectable = function (ids) {
        if (!arguments.length) {
            return this._unselectableNodes;
        }


        if (_.isNull(ids)) {
            this._unselectableNodes = [];
        } else {
            this._unselectableNodes = ids;
        }

        this.trigger('selectable', this._unselectableNodes);
    };

    /**
     * Expands all the nodes until it reaches the selected node.
     *
     * Currently only works with already loaded nodes.
     *
     * @method TreeModel#expandPathTo
     * @param {Array} selectedNodes The nodes to highlight/expand all their ancestors.
     */
    tmp.expandPathTo = function (selectedNodes) {
        //FIXME: this will not work with lazyLoading...maybe we want that
        var parentId, that = this;


        //Check if all are loaded
        selectedNodes = _.filter(selectedNodes, function (id) {
            var exists = that.node(id) != null;

            if(!exists) {
                logger.warn('Node with data=' + JSON.stringify(id) + ' not loaded yet. Skipping.');
            }

            return exists;
        });

        for (var i = 0; i < selectedNodes.length; i++) {
            parentId = that.node(selectedNodes[i]).parentId;
            while (parentId !== undefined) {
                this._expandedNodes.push(parentId);
                parentId = that.node(parentId).parentId;
            }
        }
        this._expandedNodes = _.uniq(this._expandedNodes);

        this.trigger('expand', this._expandedNodes);
    };

    //Utility methods only getters
    //TODO: rename to queryBy since getters/setters are deprecated and consistency
    tmp.node = function(id) {
        if(_.isNull(id)) {
            return this._root;
        } else if(_.isObject(id)) {
            //Return getNodeByData?
            return this.getNodeByData(id);
        } else if (_.isString(id) || _.isNumber(id)) {
            //Client-side id
            var contained = this._treeHash[id] != null;

            if(contained) {
                return this._treeHash[id];
            } else {
                var tree = _.chain(this._filterMap)
                                    .values()
                                    .flatten()
                                    .find(function(node) {
                                        return node.id === id;
                                    })
                                    .value();

                return tree || null;
            }
        }
    };

    /**
     * Gets a specific node using a more efficient path based get instead of a search.
     *
     * @method TreeModel#getNodeByPath
     * @param {String} id The id of the node to find.
     * @returns {Tree} node The node found.
     */
    tmp.getNodeByPath = function (id) {
        var contained;

        //If filtered first check the ids. If it can not find that. use the array position.
        if (this._filtered) {
            var node = _.find(this._filterMap[this._activeQuery], function (node) {
                if (node.id != null) {
                    return node.id === id;
                }
                return false;
            });
            if (node == null) {
                node = this._filterMap[this._activeQuery][id];
            }
        }
        if (_.isNull(id)) {
            contained = _.contains(_.keys(this._treeHash), 'null');
            id = 'null';
        } else {
            contained = _.contains(_.keys(this._treeHash), id);
        }

        if (contained) {
            return this._treeHash[id];
        } else {
            return this._root.getNodeById(id);
        }
    };

    tmp.getNodeByData = function (data) {
        var tree;

        tree = _.find(this._treeHash, function (tree) {
            return !_.isNull(tree.id) && _.isEqual(data, _.pick(tree.data, _.keys(data)));
        });

        if (_.isUndefined(tree)) {
            var flatMap = _.chain(this._filterMap).values().flatten().value();
            if(flatMap.length > 0) {
                tree = _.find(flatMap, function (tree) {
                    return !_.isNull(tree.id) && _.isEqual(data, _.pick(tree.data, _.keys(data)));
                });
            }

            if(_.isUndefined(tree)) {
                return null;
            } else {
                return tree;
            }
        } else {
            return tree;
        }
    };

    /**
     * Determines if a node has children or not.
     *
     * @method TreeModel#hasChildren
     * @param {String} id The id of the node to see if it has children.
     * @returns true if the node.children > 0 || node.data.hasChildren is true || model is lazyLoaded and the node is not currently loaded.
     */
    tmp.hasChildren = function (id) {
        var node;

        node = this.node(id);

        if (node.children.length > 0) {
            return true;
        }

        if (this._options.lazyLoad && node.state === 'loaded') {
            return node.children.length > 0;
        }

        if (node.data.hasChildren !== undefined) {
            return node.data.hasChildren;
        }

        return this._options.lazyLoad && (node.state === 'unloaded' || node.state === 'loading');
    };

    tmp.filtered = function (data, query, force) {
        var that = this;
        if (!arguments.length) {
            return this._filterMap[this._activeQuery] || null;
        }
        function addParents(node, parents) {
            if (node.parentId != null) {
                var parentNode = parents.push(that.node(node.parentId));
                if (parentNode != null) {
                    addParents(parentNode);
                }
            }
        }

        this._activeQuery = query;
        if (_.isNull(query)) {
            if(this._filtered || force) {
                this._filtered = false;
                this.trigger('filter', null, query);
            }
        } else {
            if (data.length) {
                if (!this._options.lazyLoad) {
                    data = _.map(data, function (value) {
                        if (_.isString(value) && that._options.recordKeys.id != null) {
                            var lookupObject = {};
                            lookupObject[that._options.recordKeys.id] = value;
                            return that.getNodeByData(lookupObject);
                        } else if(value instanceof Tree) {
                            return value;
                        }

                    });
                    var parents = [];
                    _.each(data, function (entry) {
                        addParents(entry, parents);
                    });
                    parents = _.uniq(parents);
                    this._expandedNodes = _.pluck(parents, 'id');
                    data = _.union(data, parents);
                } else {
                    data = _.map(data, function(node) {
                        var treeNode = that.getNodeByData(_.omit(node, 'children'));
                        if(treeNode == null) {
                            return new Tree(that._generator(), null, _.omit(node, 'children'), node.children, null, that._generator);
                        } else {
                            return treeNode;
                        }
                    });
                }
            }

            this._filterMap[query] = data;
            this._filtered = true;

            this.trigger('filter', data, query);
        }
    };

    tmp.clear = function () {
        var newRoot = new Tree(null, undefined, null, null, 0, this._generator);
        this._root = newRoot;

        this._selectedNodes = [];
        this._unselectableNodes = [];
        this._expandedNodes = [null];
        this._nonloadedSelectedNodes = [];
        this._disabledNodes = [];
        this._activeNode = null;

        this._treeHash = {};
        this._treeHash['null'] = this._root;

        this.trigger('invalidate');
    };

    tmp.updateNode = function (node, newValues) {
        if (node != null) {
            _.extend(node._data, newValues);
            this.trigger('node-updated', node, newValues.children != null);
        }
    };

    tmp.isFiltered = function () {
        return this._filtered;
    };

    return TreeModel;

});

/**
 * @module stringUtils
 * @private
 */
define('oui.utils/stringUtils',[],function () {
    

    return {
        /**
         * @function module:stringUtils.trim
         */
        trim : function (str) {
            return str.trim();
        },
        /**
         * @function module:stringUtils.union
         */
        union : function (str, addedString) {
            str = str || '';
            var split = str.split(' ');
            split.push(addedString);
            return split.join(' ').trim();
        },
        /**
         * @function module:stringUtils.reject
         */
        reject : function (str, removedString) {
            str = str || '';
            var split = str.split(' ');
            split.splice(split.indexOf(removedString), 1);
            return split.join(' ').trim();
        },
        /**
         * @function module:stringUtils.isEmpty
         */
        isEmpty : function (str) {
            return str === '\n' || this.trim(str) === '';
        },
        /**
         * @function module:stringUtils.pad
         */
        pad : function (str, length, ch) {
            while (str.length < length) {
                str = ch + str;
            }

            return str;
        },
        /**
         * @function module:stringUtils.regexEscape
         */
        regexEscape : function(str) {
            if(str != null) {
                return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
            } else {
                return null;
            }
        }
    };
});

define('oui.tree/treeController',['oui.underscore', 'oui.jquery', 'oui.utils/stringUtils'], function (_, $, stringUtils) {
    

    /**
     * @event TreeModel#node-moved
     * @param {Object} parent
     * @param {Object} activeNode
     */
    /**
     * @event TreeModel#loading
     * @param {String} id The id of the node that is being loaded
     */
    /**
     * @event TreeModel#node-loaded
     * @param {Object} node The node that has finished loading
     */
    /**
     * @event TreeModel#node-loaded
     * @param {Object} node The node that has finished loading
     */
    /**
     * @event TreeModel#loading
     */

    var tcp;

    /**
     * The controller used for the tree panel.
     *
     * @class TreeController
     * @private
     * @since 1.2.0
     * 
     * @param {TreeModel} model      The model that the controller will act on.
     * @param {Object}    options    The options for the tree. The controller only takes what it needs.
     */
    function TreeController(model, options) {

        /**
         * The model for the tree panel.
         *
         * @property _model
         * @type {TreeModel}
         * @private
         */
        this._model = model;
        this._options = options;

        this._loadDeferreds = {};
        this._filteredDeferreds = {};
        return this;
    }

    tcp = TreeController.prototype;

    tcp.constructor = TreeController;

    /**
     * Takes the id and expands it regardless of if it is already expanded.
     *
     * @method TreeController#expand
     * @param {Array} ids The id of the node to be expanded.
     */
    tcp.expanded = function (ids) {
        var that = this;
        if (!arguments.length) {
            return this._model.expanded();
        }

        ids = _.chain(ids)
            .reject(function (id) {
                return !that._model.hasChildren(id);
            })
            .map(function (id) {
                var state = that._model.getNodeByPath(id).state;
                if (state === 'unloaded') {
                    that.load(id);
                }

                return id;
            })
            .value();

        if (!_.isEqual(ids, this._model.expanded())) {
            this._model.expanded(ids);
        }
    };

    tcp.toggleExpand = function (id) {
        if (this._model.isExpanded(id)) {
            this.expanded(_.without(this._model.expanded(), id));
        } else {
            this.expanded(_.union(this._model.expanded(), [id]));
        }
    };

    /**
     * Takes an id and marks it as active.
     *
     * @method TreeController#active
     * @param {String} id The id of the node to mark as active.
     */
    tcp.active = function (id) {
        if(!arguments.length) {
            return this._model.active();
        }

        this._model.active(id);
    };

    tcp.moveUp = function () {
        this._move(-1);
    };

    tcp.moveDown = function () {
        this._move(1);
    };

    tcp._move = function (number) {
        var tree = this._model.tree(),
            that = this,
            selected = this.selected(),
            parent,
            activeIndex,
            swap;

        function removeUnmoveableIndices(nodes) {
            var unmoveableIndex = number < 0 ? 0 : tree.getNodeById(_.first(nodes).parentId).children.length - 1;
            var move = number < 0 ? 1 : -1;
            var indices = _.map(nodes, function (node) {
                return {
                    node : node,
                    index: _.indexOf(tree.getNodeById(node.parentId).children, node)
                };
            });

            var grouped = _.indexBy(indices, function (sortObj) {
                return sortObj.index;
            });

            while(grouped[unmoveableIndex] != null) {
                delete grouped[unmoveableIndex];
                unmoveableIndex += move;
            }

            return _.map(grouped, function (sortObj) {
                return sortObj.node;
            });
        }

        if (_.isNull(this._model.filtered())) {
            //Map id to node
            var nodes = _.map(selected, function (node) {
                return tree.getNodeById(node);
            });

            //Group by parentId
            var grouped = _.groupBy(nodes, function (node) {
                return node.parentId;
            });

            //Remove continuous first index for each group
            _.each(grouped, function (groupNodes, key, list) {
                list[key] = removeUnmoveableIndices(groupNodes);
            });

            //Sort each group
            _.each(grouped, function (groupNodes, key, list) {
                list[key] = _.sortBy(groupNodes, function (groupNode) {
                    return _.indexOf(tree.getNodeById(groupNode.parentId).children, groupNode);
                });

                if(number > 0) {
                    list[key].reverse();
                }
            });

            //Swap
            _.each(grouped, function (groupNodes) {
                _.each(groupNodes, function (node) {
                    parent = tree.getNodeById(node.parentId);
                    activeIndex = _.indexOf(parent.children, node);
                    swap = parent.children[activeIndex + number];

                    if (!_.isUndefined(swap)) {
                        parent.children[activeIndex + number] = parent.children[activeIndex];
                        parent.children[activeIndex] = swap;
                        that._model.trigger('node-moved', parent, node, number);
                    }
                });
            });

        }
    };

    /**
     * Loads a node if it is lazily loaded.
     *
     * @method TreeController#load
     * @param {String} id The id of the node to load it's children of.
     */
    tcp.load = function (id) {
        var that = this,
            node = this._model.tree().getNodeById(id) || this._model.tree();

        if (node.state === 'unloaded' && this._options.lazyLoad) {
            var loadFn = this._options.storeFns.load;
            that._model.trigger('loading', id);
            node.state = 'loading';
            var deferred = loadFn.call(this, node.id, node.data);

            if(deferred && deferred.promise && deferred.abort) {
                this._loadDeferreds[node.id] = deferred;
            }

            $.when(deferred).done(function (data) {
                that._model.addSubTree(node, data);
                node.state = 'loaded';
                that._model.trigger('node-loaded', node);
            }).fail(function(jqXHR, args) {
                if(!(args && args.filtered)) {
                    node.state = 'loaded';
                    that._model.trigger('node-loaded', node);
                } else {
                    node.state = 'unloaded';
                    that._model.expanded(_.without(that._model.expanded(), node.id));
                }
            }).always(function () {
                delete that._loadDeferreds[node.id];
            });
        }
    };

    /**
     * @method TreeController#selected
     */
    tcp.selected = function (ids, eventParams) {
        if (!arguments.length) {
            return this._model.selected();
        }

        this._model.selected(ids, eventParams);
    };

    tcp.nonLoadedSelected = function () {
        return this._model.nonLoadedSelected();
    };

    /**
     * Given an array of nodes it will expand all the ancestor nodes to make them visible.
     *
     * @method TreeController#expandPathTo
     * @param {Array} nodes The nodes to expand to.
     */
    tcp.expandPathTo = function (nodes) {
        this._model.expandPathTo(nodes);
    };

    /**
     * Takes a query and filters the data based on the fields given.
     * Currently only supports filtering already loaded data.
     *
     * @method TreeController#filter
     * @param {String} query The string to match against.
     * @param {Array} fields The fields to check.
     */
    tcp.filter = function (query, fields) {
        var regex = new RegExp(stringUtils.regexEscape(query), 'ig'), that = this;

        function createFilteredArray(list) {

            return _.chain(list)
                .filter(function (node) {
                    var data = node.data;
                    return _.some(fields, function (field) {
                        return data && data[field] && data[field].match(regex);
                    });
                })
                .flatten()
                .value();

        }

        if (_.isNull(query)) {
            this._model.filtered(null, query);

            _.each(this._filteredDeferreds, function(jqXHR) {
                jqXHR.abort({ filtered : false});
            });
        } else if (this._options.lazyLoad || !this._options.clientFilter) {
            var filterFn = this._options.storeFns.filter;
            this._model.trigger('loading', null);
            _.each(this._loadDeferreds, function(jqXHR) {
                jqXHR.abort({ filtered : true});
            });

            var deferred = filterFn.call(this, query, fields);

            if(deferred && deferred.promise && deferred.abort) {
                this._filteredDeferreds[query] = deferred;
            }
            $.when(deferred).done(function (data) {
                that._model.filtered(data, query);
            }).fail(function (jqxhr, args) {
                if(args && !args.filtered) {
                    //Force the filter to rerender the tree
                    that._model.filtered(null, null, true);
                } else {
                    that._model.filtered([], query);
                }
            }).always(function () {
                delete that._filteredDeferreds[query];
            });
        } else {
            var filterResults = createFilteredArray(this._model._treeHash);
            this._model.filtered(filterResults, query);
        }
    };

    /**
     * Takes data and adds it to the proper location in the tree.
     *
     * @method TreeController#add
     * @param {String|Object} [parentNode] The data to find the parentNode by.
     * @param {Array} data The nodes to be added.
     * @param {Number} [index] The index to add the data at.
     */
    tcp.add = function (parentNode, data, index) {
        //TODO: investigate the arguments and when parentNode is not given
        var that = this;

        function removeChildren(node) {
            var treeNode = that._model.getNodeByData(_.omit(node, 'children'));
            _.each(node.children, function (child) {
                removeChildren(child);
            });

            if (!_.isNull(treeNode)) {
                that._model.removeSubTree(treeNode.id);
            }
        }

        if(_.isNumber(data)) {
            //Index
            index = data;
            data = undefined;
        }

        _.each(data, function (node) {
            removeChildren(node);
        });

        if (_.isArray(parentNode)) {
            this._model.addSubTree(parentNode, index);
        } else {
            this._model.addSubTree(this.node(parentNode), data, index);
        }
    };

    /**
     * Removes the subtree from the model.
     *
     * @method TreeController#remove
     * @param {String|Object} data The id of the subtree to remove.
     */
    tcp.remove = function (data) {
        if (_.isObject(data)) {
            //Should be data of the node to search and remove as
            //many attributes as needed to identify the node uniquely
            //because it stops after first find.
            //Right now the data has to be the exact same as the node
            this._model.removeSubTree(this._model.getNodeByData(data).id);
        } else {
            //This should be the id of the node to remove (client-side id)
            this._model.removeSubTree(data);
        }
    };

    tcp.tree = function () {
        return this._model.tree();
    };

    tcp.node = function (id) {
        return this._model.node(id);
    };

    tcp.disabled = function (nodes) {
        if (!arguments.length) {
            return this._model.disabled();
        }

        this._model.disabled(nodes);
    };

    tcp.unselectable = function (nodes) {
        if (!arguments.length) {
            return this._model.unselectable();
        }

        var unselectable = [];

        //TODO: store this function? - so when new nodes get
        //added they can be run through it?
        if (_.isFunction(nodes)) {
            _.each(this._model._treeHash, function (node) {
                if (!_.isNull(node.id) && nodes(node.data)) {
                    unselectable.push(node.id);
                }
            });
        } else {
            unselectable = nodes;
        }
        this.selected(_.difference(this.selected(), nodes));
        this._model.unselectable(unselectable);
    };

    tcp.invalidate = function () {
        this._model.clear();

        this.load(null);
    };

    tcp.updateNodes = function (nodes) {
        var that = this;
        _.each(nodes, function (node) {
            var actualNode;
            //Sample Idea for a better way of getting the nodes
            if (that._options.recordKeys.id != null) {
                actualNode = that.node(_.pick(node, that._options.recordKeys.id));
            } else {
                actualNode = that.node(node);
            }
            that._model.updateNode(actualNode, node);
        });
    };

    tcp.isFiltered = function () {
        return this._model.isFiltered();
    };

    tcp.filtered = function () {
        return this._model.filtered();
    };

    return TreeController;

});


define('oui.tree/treeTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div role="tree" class="tree" tabindex="0"> <div class="filter-notif"></div> <i title="'+
((__t=( data.locale.loading ))==null?'':__t)+
'" class="loader pgbu-icon pgbu-icon-loading"></i> <div class="no-results">'+
((__t=( data.locale.no_results_found ))==null?'':__t)+
'</div> <ul class="root"> </ul></div>';
return __p;
};});


define('oui.tree/disableTemplate.htm',[],function () { return '<div class="disable-overlay"></div>\n';});


define('oui.tree/nodeTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<li role="treeitem" aria-selected="false" '+
((__t=( data.ariaExpanded ))==null?'':__t)+
' id="'+
((__t=( data.pageId ))==null?'':__t)+
'" data-treeid="'+
((__t=( data.id))==null?'':__t)+
'" class="'+
((__t=( data.group ? 'group' : '' ))==null?'':__t)+
'"> '+
((__t=( data.itemMarkup ))==null?'':__t)+
' <i title="'+
((__t=( data.locale.loading ))==null?'':__t)+
'" class="loader pgbu-icon pgbu-icon-loading"></i> <ul role="group" class="children        '+
((__t=( data.show ? 'show' : 'hide'  ))==null?'':__t)+
'"> '+
((__t=( data.children ))==null?'':__t)+
' </ul></li>';
return __p;
};});


define('oui.tree/nodeItemTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div    title="'+
((__t=( data.title ))==null?'':_.escape(__t))+
'"    style="padding-left: '+
((__t=( data.margin))==null?'':__t)+
'px;"    class="item node            '+
((__t=( data.selected ? 'selected' : ''  ))==null?'':__t)+
' '+
((__t=( data.disabled ? 'tree-disabled' : ''  ))==null?'':__t)+
' '+
((__t=( data.selectable ? '' : 'unselectable'  ))==null?'':__t)+
' '+
((__t=( data.active ? 'active' : ''  ))==null?'':__t)+
'"> <i title="'+
((__t=( data.collapsed ? data.locale.collapsed : data.locale.expanded))==null?'':__t)+
'" class="tree-icon        '+
((__t=( data.collapsed ? 'collapsed' : 'expanded'  ))==null?'':__t)+
' '+
((__t=( data.hidden ? 'tree-hidden' : ''  ))==null?'':__t)+
'"></i> '+
((__t=( data.iconMarkup ))==null?'':__t)+
' '+
((__t=( data.valueMarkup ))==null?'':__t)+
'</div>';
return __p;
};});


define('oui.tree/flatIconTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<li id="oui-tree-node-'+
((__t=( data.pageId ))==null?'':__t)+
'" role="treeitem" data-treeid="'+
((__t=( data.id ))==null?'':__t)+
'"> <div title="'+
((__t=( data.title ))==null?'':_.escape(__t))+
'"        class="item flat-item            '+
((__t=( data.selected ? 'selected' : ''  ))==null?'':__t)+
' '+
((__t=( data.disabled ? 'tree-disabled' : ''  ))==null?'':__t)+
' '+
((__t=( data.selectable ? '' : 'unselectable'  ))==null?'':__t)+
' '+
((__t=( data.active ? 'active' : ''  ))==null?'':__t)+
'"> '+
((__t=( data.iconMarkup  ))==null?'':__t)+
' '+
((__t=( data.valueMarkup ))==null?'':__t)+
' </div></li>';
return __p;
};});

define('oui.tree/treeView',['oui.underscore', 'oui.jquery', 'oui.base/view',
    'oui.tree/treeTemplate.htm', 'oui.tree/disableTemplate.htm',
    'oui.tree/nodeTemplate.htm',
    'oui.tree/nodeItemTemplate.htm',
    'oui.tree/flatIconTemplate.htm',
    'oui.utils/keyCode', 'oui.utils/stringUtils'],
    function (_, $, View, treeTemplate, disableTemplate, nodeTemplate, nodeItemTemplate, flatIconTemplate, keyCode, stringUtils) {
        

        var tvp;

        /**
         * The view that constructs the element and binds all the events for the TreePanel.
         *
         * @class TreeView
         * @private
         * @extends View
         * @since 1.2.0
         * 
         * @param {TreeModel} model The data for the tree panel.
         * @param {TreeController} controller The object to accept all UI calls.
         * @param {TreePanel#Config} options Object that contains all the functions that can be overrideable.
         * @param {Object} settings
         */
        function TreeView(model, controller, options, settings) {
            var that = this;

            View.call(this, options, null, controller, settings);

            //TODO: look into better way of doing this
            //need to wait until after buildElement and id are called
            this._$el.find('.no-results').attr('id', this.id() + '-no-results');

            model.on({
                'view.load' : {fn : this.render, context : this },
                'view.collapse' : {fn : this.collapse, context : this },
                'view.node-removed' : {fn : this.render, context : this },
                'view.expand' : {fn : this.expand, context : this },
                'view.disabled-node' : {fn : this.disabled, context : this },
                'view.enabled-node' : {fn : this.disabled, context : this },
                'view.selectable' : {fn : this.selectable, context : this },
                'view.select' : {fn : this.select, context : this },
                'view.deselect' : {fn : this.deselect, context : this },
                'view.active' : {fn : this.active, context : this },
                'view.node-loaded' : {fn : this.nodeLoaded, context : this },
                'view.loading' : {fn : this.loading, context : this },
                'view.node-added' : {fn : this.nodeAdded, context : this },
                'view.filter' : {fn : this.filter, context : this },
                'view.node-moved' : {fn : this.moved, context : this},
                'view.node-updated' : {fn : this.renderNode, context : this}
            });

            //TODO: this should not be here.
            //Everything should be doable through the controller
            Object.defineProperty(this, '_model', {
                writable : false,
                enumerable : false,
                value : model
            });

            //TODO: move these to a function for binding events?
            this._$el.on('click', '.tree-icon, .node.unselectable', function (e) {
                e.preventDefault();
                e.stopPropagation();
                controller.toggleExpand($(this).parents('li').attr('data-treeid'));
            });

            this._$el.on('click', '[data-trigger]', function (e) {
                e.preventDefault();
                e.stopPropagation();
                var $element = $(this),
                    id = $element.parents('li').attr('data-treeid'),
                    triggerName = $element.attr('data-trigger');
                that.trigger('data-trigger-' + triggerName, $(this), controller.node(id).data);
            });

            this._$el.on('click', '.filter-notif', function () {
                that._controller.filter(null);
            });

            this.addShortcuts({
                'UP' : _.partial(this._moveActive, -1),
                'DOWN' : _.partial(this._moveActive, 1),
                'LEFT' : that._handleCollapse,
                'RIGHT' : that._handleExpand,
                'HOME' : that._top,
                'END' : that._bottom,
                'ESC' : {
                    fn : function () {
                        if(that._controller.isFiltered()) {
                            that._controller.filter(null);
                        }
                    },
                    stopPropagation : that._controller.isFiltered.bind(that._controller)
                },
                'SPACE' : that._select,
                'shift-UP' : _.partial(this._extendSelection, -1),
                'shift-DOWN': _.partial(this._extendSelection, 1),
                'ctrl-UP' : _.partial(this._moveActive, -1),
                'ctrl-DOWN' : _.partial(this._moveActive, 1),
                'ctrl-shift-UP' : this.moveUp,
                'ctrl-shift-DOWN' : this.moveDown,
                'ctrl-SPACE' : that._toggleSelection
            });

            that._$el.on('keydown keyup', function(event) {
                if(that._options.multiSelect && event.keyCode === keyCode.SHIFT) {
                    //Put on disable select class
                    $('body').toggleClass('oui-disable-select', event.type === 'keydown');
                }
            });

            that._$el.on('blur', function() {
                $('body').removeClass('oui-disable-select');
                //Remove currently active element when blurred
                that._controller.active(null);
            });

            that._$el.on('keypress', function (event) {
                var string = String.fromCharCode(event.keyCode);
                if(string.length > 0) {
                    that._find(string);
                }
            });

            this._$el.on('click', '.item:not(".tree-disabled"):not(".unselectable")', function (event) {
                event.preventDefault();

                var id = $(this).parent('li').attr('data-treeid'), group = [];
                var $target = $(event.target);
                var prevSelected = that._$el.find('.active').parent('li'), clicked = $(this).parent('li');
                var nodes;

                var selectedIndex;
                var prevSelectedIndex;
                var selected;

                if (!$target.hasClass('tree-icon') && !$target.is('[data-trigger]')) {
                    if (event.ctrlKey) {
                        that._toggleSelection(id);
                    } else if (event.shiftKey && that._options.multiSelect) {
                        nodes = that.$('[data-treeid]:visible:not(".tree-disabled, .unselectable")');
                        selectedIndex = nodes.index(clicked);
                        prevSelectedIndex = nodes.index(prevSelected);

                        if(selectedIndex < prevSelectedIndex) {
                            selected = nodes.slice(selectedIndex, prevSelectedIndex + 1);
                        } else {
                            selected = nodes.slice(prevSelectedIndex, selectedIndex + 1);
                        }

                        _.each(selected, function(el) {
                            group.push($(el).attr('data-treeid'));
                        });

                        controller.selected(group);

                    } else {
                        controller.selected([id]);
                    }
                }
            })
            .on('dblclick', '.item:not(".tree-disabled"):not(".unselectable")', function () {
                var id = $(this).parent('li').attr('data-treeid');
                that.trigger('dblclick', $(this), controller.node(id).data);
            })
            .on('mousedown', '.item', function(event) {
                if(!(event.shiftKey || event.ctrlKey)) {
                    var id = $(this).parent('li').attr('data-treeid');
                    that._controller.active(id);
                }

                if(!($(event.target).attr('tabindex') != null ||
                   $(event.target).is('button') ||
                   $(event.target).is('a') ||
                   $(event.target).attr('draggable') != null)) {
                    event.preventDefault();
                    that._$el.focus();
                }
            })
            .on('click', '.item:not(".tree-disabled"):not(".unselectable")',function () {
                var id = $(this).parent('li').attr('data-treeid');
                that.trigger('click-selectable', $(this), controller.node(id).data);
            }).on('click', '.item.tree-disabled, .item.unselectable', function () {
                var id = $(this).parent('li').attr('data-treeid');
                that.trigger('click-unselectable', $(this), controller.node(id).data);
            });

            //TODO: remove this (use this._options.viewFns.<functionName>)
            _.extend(this, options.viewFns);

            return this;
        }

        TreeView.prototype = Object.create(View.prototype);
        tvp = TreeView.prototype;

        tvp.constructor = TreeView;

        tvp.moveUp = function () {
            if(this._options.move) {
                this._controller.moveUp();
            }
        };

        tvp.moveDown = function () {
            if(this._options.move) {
                this._controller.moveDown();
            }
        };

        tvp._extendSelection = function(delta) {
            var items = this._$el.find('[data-treeid]:visible'),
                selectionIndex = items.index(items.filter('.selection-pointer')),
                activeIndex = items.index(this._$el.find('.active').parent()),
                selected, group = [];

            if(selectionIndex === -1) {
                selectionIndex = activeIndex;
            } else {
                items.eq(selectionIndex).removeClass('selection-pointer');
            }

            if((delta < 0 && selectionIndex + delta >= 0) ||
               (delta > 0 && selectionIndex + delta < items.length)) {
                selectionIndex+= delta;
            }

            if(selectionIndex < activeIndex) {
                selected = items.slice(selectionIndex, activeIndex + 1);
            } else {
                selected = items.slice(activeIndex, selectionIndex + 1);
            }

            _.each(selected, function(el) {
                group.push($(el).attr('data-treeid'));
            });

            this._controller.selected(group);

            items.eq(selectionIndex).addClass('selection-pointer');
        };

        tvp._top = function() {
            var items = this._$el.find('.item:visible');
            if(items.length > 0) {
                this._controller.active(items.eq(0).parent('li').attr('data-treeid'));
            }
        };

        tvp._bottom = function() {
            var items = this._$el.find('.item:visible');
            if(items.length > 0) {
                this._controller.active(items.eq(items.length - 1).parent('li').attr('data-treeid'));
            }
        };

        tvp._toggleSelection = function(id) {
            var activeElementId = id || this._$el.find('.active').parent().attr('data-treeid'),
                currentSelection = this._controller.selected();

            if(_.contains(currentSelection, activeElementId)) {
                this._controller.selected(_.without(currentSelection, activeElementId));
            } else if (this._options.multiSelect) {
                this._controller.selected(_.union(currentSelection, [activeElementId]));
            }
        };

        tvp._select = function () {
            var activeElement = this._$el.find('.active');
            var activeId = activeElement.parent().attr('data-treeid');

            if(activeElement.length !== 0 &&
               !activeElement.hasClass('unselectable') && !activeElement.hasClass('tree-disabled')) {
                this._controller.selected([activeId]);
            }
        };

        tvp._find = function(letter) {
            var items = this._$el.find('.item:visible'),
                regex = new RegExp('^' + stringUtils.regexEscape(letter), 'gi'),
                activeIndex = items.index(this._$el.find('.active'));

            var nextActive = _.chain(items)
                .sortBy(function(item, index) {
                    if(index === activeIndex) {
                        return index + items.length;
                    } else if (index > activeIndex ) {
                        return index - items.length;
                    } else {
                        return index;
                    }
                })
                .find(function(item) {
                    return regex.test($(item).text().trim());
                })
                .value();

            if(nextActive != null) {
                this._controller.active($(nextActive).parent().attr('data-treeid'));
            }
        };

        tvp._moveActive = function(delta) {
            var activeElement = this._$el.find('.active'),
                items = this._$el.find('.item:visible'),
                currentIndex = items.index(activeElement),
                nextItem;

            //Moving up and there is no active
            if(currentIndex !== -1) {
                nextItem = items.eq(currentIndex + delta);
            }

            if(items.length !== 0) {
                if(nextItem != null && nextItem.length !== 0) {
                    this._controller.active(nextItem.parent().attr('data-treeid'));
                } else if (delta < 0) {
                    this._controller.active(items.eq(items.length -1).parent().attr('data-treeid'));
                } else if (delta > 0) {
                    this._controller.active(items.eq(0).parent().attr('data-treeid'));
                }
            }

        };

        tvp._handleCollapse = function() {
            var activeElement = this._$el.find('.active');

            if (activeElement.length > 0) {
                var parents = activeElement.parents('li');

                if(this._model.isExpanded(parents.attr('data-treeid'))) {
                    this._controller.expanded(_.without(this._controller.expanded(), parents.attr('data-treeid')));
                } else if (parents.length > 1){
                    this._controller.active(parents.eq(1).attr('data-treeid'));
                }
            }
        };

        tvp._handleExpand = function() {
            var activeElement = this._$el.find('.active');
            if (activeElement.length > 0) {
                var parents = activeElement.parents('li');

                if(this._model.hasChildren(parents.attr('data-treeid')) &&
                   this._model.isExpanded(parents.attr('data-treeid'))) {
                    this._controller.active(parents.eq(0).children('.children').children('li:eq(0)').attr('data-treeid'));
                } else {
                    this._controller.expanded(_.union(this._controller.expanded(), [activeElement.parents('li').attr('data-treeid')]));
                }
            }
        };

        tvp.buildElement = function (options) {
            var $el = $(treeTemplate({ locale : options.locale }));

            $el.toggleClass('picker', options.picker);

            $el.toggleClass('fill', options.fill);

            $el.attr('aria-multiselectable', options.multiSelect);

            return $el;
        };

        tvp.id = function (settings) {
            var tree = this._$el;

            if(!arguments.length) {
                return this._$el.attr('id');
            }

            if(tree.length > 0 && _.isUndefined(tree.attr('id'))) {
                if(_.isObject(settings)) {
                    tree.attr('id', 'oui-' + settings.name + '-' + settings.id);
                } else if (_.isString(settings)) {
                    tree.attr('id', settings);
                }
            }
        };

        tvp.nodeAdded = function (tree, lazyLoaded) {
            //This sometimes needs to happen but not others.
            if (!lazyLoaded || (tree.parentId == null && this._controller.node(tree.parentId).state === 'loaded')) {
                //Node loaded event will never fire - have to append this tree to proper position in DOM
                this.render(this._model);
            }
        };

        tvp.loading = function (id) {
            if (_.isNull(id)) {
                this._$el.addClass('loading');
                this._$el.find('.root').empty(); //Loading spinner should not be on the screen at the same time
            } else {
                this._$el.find('[data-treeid=' + id + ']').addClass('loading');
            }

        };

        /**
         * Method called after the collapse event has fired.
         *
         * Hides the children and changes the icon of to the collapsed icon.
         *
         * @method TreeView#collapse
         * @param {Array} ids The id of the node to collapse
         */
        tvp.collapse = function (ids) {
            var that = this;

            _.each(ids, function (id) {
                var $li = that._$el.find('[data-treeid=' + id + ']').attr('aria-expanded', false);

                $li.children('.item').children('.tree-icon').removeClass('expanded').addClass('collapsed');
                $li.children('.children')
                    .addClass('hide')
                    .removeClass('show');
            });
        };

        /**
         * Method called after the expand event has fired.
         *
         * Changes the icon to the expanded icon as well as creates the children for the node
         * if they have not been already created and then shows them.
         * If there are no children remove the collapse/expand icon.
         *
         * @method TreeView#expand
         * @param {Array} ids The id of the node to expand
         */
        tvp.expand = function (ids) {
            var model = this._model, that = this;

            _.each(ids, function (id) {
                var $li, $children, markup;
                if (!_.isNull(id)) {
                    $li = that._$el.find('[data-treeid=' + id + ']').attr('aria-expanded', true);

                    $children = $li.children('.children');
                    $li.children('.item')
                        .children('.tree-icon')
                        .removeClass('collapsed')
                        .addClass('expanded');

                    //Add each of its children and display them
                    if (model.hasChildren(id)) {
                        if ($children.children().length === 0) {
                            markup = that.createTree(model, model.getNodeByPath(id), false);
                            $children.append(markup);
                        }
                        $children.removeClass('hide');
                        $li.children('.children').addClass('show');
                    } else {
                        $li.find('.tree-icon').css('visibility', 'none');
                    }
                }
            });

        };

        /**
         * Called after the selectable event fires.
         *
         * Marks the nodes passed in as unselectable.
         *
         * @method TreeView#selectable
         * @param {Array} ids The unselectable ids to be marked as unselectable.
         */
        tvp.selectable = function (ids) {
            var that = this;

            this._$el.find('.unselectable').removeClass('unselectable').attr('aria-selected', false);
            _.each(ids, function (id) {
                that._$el.find('[data-treeid=' + id + ']').children('.item').addClass('unselectable').removeAttr('aria-selected');
            });
        };

        /**
         * Called after the enable and disable event fires.
         *
         * Marks all the nodes with a class tree-disabled - since disabled is likely used in the CSS already.
         *
         * @method TreeView#disable
         * @param {Array} ids The disabled ids to be marked.
         */
        tvp.disabled = function (ids) {
            var that = this;

            this._$el.find('.tree-disabled').removeClass('tree-disabled').parent().removeAttr('aria-disabled').attr('aria-selected', false);
            _.each(ids, function (id) {
                that._$el.find('[data-treeid=' + id + ']')
                    .attr('aria-disabled', true)
                    .removeAttr('aria-selected')
                    .find('.item').addClass('tree-disabled');
            });
        };

        tvp.disable = function (disabled) {

            if (disabled) {
                this._$el.append($(disableTemplate));
                this._$el.removeAttr('tabindex');
            } else {
                this._$el.find('.disable-overlay').remove();
                this._$el.attr('tabindex', 0);
            }
        };

        /**
         * Called after the select event fires.
         *
         * Marks each element as selected.
         *
         * @method TreeView#select
         * @param {Array} selectedIds The ids of the nodes that are selected.
         */
        tvp.select = function (selectedIds) {
            var that = this, nodes;

            this._$el.find('.selection-pointer').removeClass('selection-pointer');
            this._$el.find('.selected').removeClass('selected').attr('aria-selected', false);

            this._$el.toggleClass('has-selection', selectedIds.length > 0);

            this._$el.find('.filter-notif').text(this._options.locale.notification.replace(/{\d+}/g, function(){
                return selectedIds.length;
            }));

            _.each(selectedIds, function (id) {
                that._$el.find('[data-treeid=' + id + ']').attr('aria-selected', true).children('.item').addClass('selected');
            });

            if (!_.isNull(that._model.filtered())) {
                nodes = that._model.nonLoadedSelected();

                _.each(nodes, function (node) {
                    that._$el
                        .find('li:eq(' + _.indexOf(that._model.filtered(), node) + ')')
                        .children('.item')
                        .addClass('selected');
                });
            }
        };

        /**
         * Called after the deselect event fires.
         *
         * Marks the element as unselected.
         *
         * @method TreeView#deselect
         * @param {String} id The id of the element to deselect.
         */
        tvp.deselect = function (id) {

            this._$el.find('[data-treeid=' + id + ']').children('.item').removeClass('selected');
        };

        /**
         * Called after the active event fires.
         *
         * Marks the node as active. (The node from where all keyboard interaction takes place)
         * There can only ever be on active element.
         *
         * @method TreeView#active
         * @param {String} id The id of the element to mark as active.
         */
        tvp.active = function (id) {

            this._$el.find('.active').removeClass('active');
            var activeElement = this._$el
                .find('[data-treeid=' + id + ']')
                .children('.item')
                .addClass('active');

            //Only scroll on a tree with an active element
            //scrollIntoView scrolls tree and all scrolling container parents
            if(activeElement.length !== 0) {
                var position = activeElement.position();
                var treeHeight = this._$el.height();
                //Scroll parent container if it has one
                if (position.top + activeElement.height() >= treeHeight) {
                    this._$el.scrollTop((activeElement[0].offsetTop + activeElement.height()) - treeHeight);
                } else if (position.top < 0 || position.top >= treeHeight) {
                    this._$el.scrollTop(this._$el.scrollTop() + position.top);
                }

                this._$el.attr('aria-activedescendant', activeElement.parent().attr('id'));
            } else {
                this._$el.removeAttr('aria-activedescendant');
            }

        };

        /**
         * Called after the node-loaded event fires.
         *
         * Creates the tree representing this subtree and appends it to the parent element.
         *
         * @method TreeView#nodeLoaded
         * @param {Tree} node The node that was loaded.
         */
        tvp.nodeLoaded = function (node) {
            var model = this._model, $li;

            if (_.isNull(node.id)) {
                //Loaded the root node
                this.render(model); //Render is only used for determining whether list or tree
            } else {
                $li = this._$el.find('[data-treeid=' + node.id + ']');
                $li.removeClass('loading');
                $li.children('.children')[0].innerHTML = this.createTree(model, node, false);

                if (!model.hasChildren(node.id)) {
                    $li.children('.item').children('.tree-icon').addClass('tree-hidden');
                    $li.removeAttr('aria-expanded');
                }

                this.trigger('dom-children-added', $li.children('.children'));
            }
        };

        tvp.moved = function (node, movedNode, number) {
            var model = this._model, $li;

            if (_.isNull(node.id)) {
                //Loaded the root node
                this.render(model);
            } else {
                $li = this._$el.find('[data-treeid=' + node.id + ']');
                $li.children('.children')[0].innerHTML = this.createTree(model, node, false);

                if (!model.hasChildren(node.id)) {
                    $li.children('.item').children('.tree-icon').addClass('tree-hidden');
                }

                this.trigger('dom-children-added', $li.children('.children'));
            }

            var $selected;

            if(number < 0) {
                $selected = this._$el.find('.selected').first();
            } else {
                $selected = this._$el.find('.selected').last();
            }
            //Only scroll on a tree with an active element
            //scrollIntoView scrolls tree and all scrolling container parents
            if($selected.length !== 0) {
                var position = $selected.position();
                var treeHeight = this._$el.height();
                //Scroll parent container if it has one
                if (position.top + $selected.height() >= treeHeight) {
                    this._$el.scrollTop(($selected[0].offsetTop + $selected.height()) - treeHeight);
                } else if (position.top < 0 || position.top >= treeHeight) {
                    this._$el.scrollTop(this._$el.scrollTop() + position.top);
                }
            }
        };

        /**
         * Called after the filter event fires.
         *
         * Renders the results as a flat list.
         *
         * @method TreeView#filter
         * @param {Array} results The nodes that match the query.
         */
        tvp.filter = function (results, query) {
            var model = this._model;
            var that = this;
            var newDescribedBy;

            this._$el.toggleClass('filtered', !_.isNull(results)).toggleClass('not-found', !_.isNull(results) && results.length === 0);

            if(!_.isNull(results) && results.length === 0) {
                newDescribedBy = stringUtils.union(this._$el.attr('aria-describedby'), this._$el.find('.no-results').attr('id'));
                this._$el.attr('aria-describedby', newDescribedBy);
            } else {
                newDescribedBy = stringUtils.reject(this._$el.attr('aria-describedby'), this._$el.find('.no-results').attr('id'));

                if(newDescribedBy === '') {
                    this._$el.removeAttr('aria-describedby');
                } else {
                    this._$el.attr('aria-describedby', newDescribedBy);
                }
            }

            if(this._controller.selected().length > 0) {
                this._$el.find('.filter-notif').text(this._options.locale.notification.replace(/{\d+}/g, function(){
                    return that._controller.selected().length;
                }));
            }

            if (_.isNull(results) || (this._options.filterShowHierarchy && !this._options.lazyLoad)) {
                this.render(model, results);
            } else {
                this.renderFlat(model, results, query);
            }

            //this._controller.active(this._$el.find('.item:eq(0)').parent().attr('data-treeid') || null);
        };

        /**
         * Creates the markup that represent the tree (current node) in the current
         * state (expanded/collapsed/selected/active).
         *
         * Recursive if the node is expanded - to render all of its children.
         *
         * @method TreeView#createTree
         * @param {TreeModel} model The model to get the status of the node from.
         * @param {Tree} node The node to get the data from.
         * @param {Boolean} selfVisible Whether or not the node itself is visible.
         * @return {String} markup The HTML markup representing the tree.
         */
        tvp.createTree = function (model, node, selfVisible, filterResults) {
            var markup = '',
                children = '',
                id,
                that = this,
                data;

            id = node.id;
            filterResults = filterResults || this._controller.filtered();

            if (model.isExpanded(id) || _.isNull(id) || filterResults != null) {
                _.each(node.children, function (child) {
                    children += that.createTree(model, child, true, filterResults);
                });
            }
            var isExcluded = false;
            if (node.id != null && filterResults != null) {
                isExcluded = children.length <= 0 && _.find(filterResults, function (filteredNode) {
                    return filteredNode.id === id;
                }) == null;
            }

            if (selfVisible && !isExcluded) {
                var ariaExpanded;

                if(this._model.hasChildren(id)) {
                    if(this._model.isExpanded(id)) {
                        ariaExpanded = 'aria-expanded="true"';
                    } else {
                        ariaExpanded = 'aria-expanded="false"';
                    }
                }

                data = {
                    id : id,
                    locale : this._options.locale,
                    pageId : this.id() + '-node-' + id,
                    itemMarkup : that.makeNodeItem(node, model),
                    children : children,
                    ariaExpanded : ariaExpanded,
                    show : model.isExpanded(id),
                    group : node.data.group
                };

                markup = nodeTemplate(data);
            }
            else {
                markup = children;
            }

            return markup;
        };

        /**
         *
         * @method TreeView#makeNodeItem
         * @param {*} node
         * @param {TreeModel} model
         */
        tvp.makeNodeItem = function (node, model, query) {
            var id = node.id,
                data;

            function isSelectable(attrs, model) {
                if(!model || attrs) {
                    return false;
                }

                return true;
            }

            var el = document.createElement('span');
            var attrs = [];
            el.innerHTML = this.display(node.data, query);

            if(this._controller.isFiltered() && this._options.storeFns && this._options.storeFns.attrs) {
                attrs = this._options.storeFns.attrs(node.data);
            }

            data = {
                title : el.textContent.trim(),
                locale : this._options.locale,
                valueMarkup : el.innerHTML,
                collapsed : !model.isExpanded(id),
                disabled : _.contains(attrs, 'disabled') || model.isDisabled(id),
                selectable : isSelectable(_.contains(attrs, 'unselectable'), model.isSelectable(id)),
                selected : model.isSelected(id),
                hidden : !model.hasChildren(id),
                //TODO: using private variable here...STOP!
                active : model._activeNode === id,
                margin : 15 * (node.level - 1),
                iconMarkup : ''
            };

            if (this._options.viewFns.icons) {
                data.iconMarkup = this._options.viewFns.icons(node.data);
            }
            return nodeItemTemplate(data);
        };

        tvp.makeFlatItem = function (node, model, query) {
            var id, that = this, data, templateData;

            id = node.id;
            data = node.data || node;

            var el = document.createElement('span');
            el.innerHTML = this.display(node.data, query);

            templateData = {
                id : id,
                title : el.textContent.trim(),
                pageId : this.id() + '-node-' + id,
                valueMarkup : el.innerHTML,
                selected : model.isSelected(id),
                disabled : model.isDisabled(id),
                selectable : model.isSelectable(id),
                //TODO: again using private variable here...STOP!
                active : model._activeNode === id,
                iconMarkup : ''
            };

            if (that._options.viewFns.icons) {
                data.iconMarkup = that._options.viewFns.icons(node.data);
            }

            return flatIconTemplate(templateData);
        };

        tvp.renderNode = function (node, entireNode) {
            var model = this._model;
            var $nodeLi = this._$el.find('li[data-treeid="' + node.id + '"]');

            $nodeLi.removeClass('loading');

            if (this._options.flat) {
                $nodeLi.replaceWith($(this.makeFlatItem(node, model, model._activeQuery)));
            } else {
                if (entireNode) {
                    $nodeLi.replaceWith($(this.createTree(model, node, true)));
                }
                var $nodeItem = $nodeLi.children('.item');
                $nodeItem.replaceWith($(this.makeNodeItem(node, model)));
            }

            this.trigger('dom-nodes-rendered', $nodeLi);
        };


        /**
         * Creates a flat list of the nodes requested.
         *
         * @method TreeView#createList
         * @param {TreeModel} model The model to get the status of a node.
         * @param {Array} nodes The list of nodes to be rendered.
         * @return {String} markup The HTML markup representing the nodes.
         */
        tvp.createList = function (model, nodes, query) {
            var that = this;

            function isSelectable(attrs, model) {
                if(!model || attrs) {
                    return false;
                }

                return true;
            }

            return _.reduce(nodes, function (memo, node) {
                var id = node.id, data = node.data || node;
                var el = document.createElement('span');
                var attrs = [];

                if(that._controller.isFiltered() && that._options.storeFns && that._options.storeFns.attrs) {
                    attrs = that._options.storeFns.attrs(node.data);
                }

                el.innerHTML = that.display(data, query);
                var templateData = {
                    id : id,
                    title : el.textContent.trim(),
                    pageId : that.id() + '-node-' + id,
                    valueMarkup : el.innerHTML,
                    selected : model.isSelected(id),
                    disabled : _.contains(attrs, 'disabled') || model.isDisabled(id),
                    selectable : isSelectable(_.contains(attrs, 'unselectable'), model.isSelectable(id)),
                    //TODO: again using private variable here...STOP!
                    active : model._activeNode === id,
                    iconMarkup : ''
                };

                if (that._options.viewFns.icons) {
                    templateData.iconMarkup = that._options.viewFns.icons(data);
                }

                return memo += flatIconTemplate(templateData);
            }, '');
        };

        /**
         * Emptys the root node and renders the tree as in its current state.
         *
         * @method TreeView#render
         * @param {TreeModel} model The tree model to get the root from and pass
         *  to the appropriate markup creation method.
         */
        tvp.render = function (model, filterResults) {
            var $node = this._$el.find('ul.root');

            this._$el.removeClass('loading');

            if (this._options.flat) {
                $node[0].innerHTML = this.createList(model, model.tree().children);
            } else {
                $node[0].innerHTML = this.createTree(model, model.tree(), this._options.displayRoot, filterResults);
            }

            this.trigger('dom-nodes-added', $node);

        };

        /**
         * Emptys the root node and renders the nodes given as a flat list.
         *
         * @method TreeView#renderFlat
         * @param {TreeModel} model The model to get the status of the nodes.
         * @param {Array} nodes The nodes to render in a list.
         */
        tvp.renderFlat = function (model, nodes, query) {
            var $node = this._$el.find('ul.root');

            this._$el.removeClass('loading');

            $node[0].innerHTML = this.createList(model, nodes, query);

            if(this._controller.active() == null) {
                this._controller.active($node.find('li').first().attr('data-treeid') || null);
            }
        };

        /**
         * Hides the tree view.
         *
         * @method TreeView#hide
         */
        tvp.hide = function () {
            this._$el.hide();
        };

        /**
         * Shows the tree view.
         *
         * @method TreeView#show
         */
        tvp.show = function () {
            this._$el.show();
        };

        tvp.destroy = function (clean) {
            this._$el.off('click keydown');

            if (clean.self) {
                this._$el.remove();
            }
        };

        return TreeView;

    }
)
;

define('oui.tree/treePanel',['oui.underscore', 'oui.base/widget', 'oui.tree/treeModel', 'oui.tree/treeController', 'oui.tree/treeView', 'oui.logger'],
    function (_, Widget, TreeModel, TreeController, TreeView, logger) {
        


        /**
         * @callback TreePanel#load
         * @param {Number|String} id The nodes client-side internally generated id.
         * @param {Object} data The data the node represents.
         */

        /**
         * @callback TreePanel#filter
         * @param {String} query The query to search the node's data for.
         * @param {Array<String>} fields The fields to look over in the node's data.
         */

        /**
         * @callback TreePanel#attrs
         * @param {Object} data The node's data.
         * @return {Array<String>} The current nodes attributes. Valid strings are 'unselectable', 'disabled', 'expanded',
         * and 'selected'
         */

        /**
         * @callback TreePanel#display
         * @param {Object} data The node's data.
         * @return {String} The node's string representation to use on the DOM.
         */

        /**
         *
         * @example
         *     //Assumed this property is included in the options
         *     viewFns.icons : function (data) {
         *         // Can return markup based on data of the current node requested as well
         *         // as multiple icons
         *         return '<i class="pgbu-icon pgbu-icon-ent-user"></i>';
         *     }
         *
         * @callback TreePanel#icons
         * @param {Object} data the node's data.
         * @return {HTMLString} The markup to use for displaying the node's icon.
         */

        /**
         * @typedef {Object} TreePanel#Config
         *
         * @property {Boolean} [move=false] Whether to allow move up/down in the tree. Necessary when allowing drag and drop
         * operations within the tree.
         *
         * @property {Boolean} [lazyLoad=false] Whether or not the tree is lazy loaded. Will call the load function when
         * initialized. When false the tree will not call load as it is assumed that the data is already loaded.
         *
         * @property {Boolean} [fluid=true] Whether or not the tree should take up all available width of the parent
         * container.
         *
         * @property {Boolean} [fill=true] Whether or not the tree should take up all available height of the parent
         * container.
         *
         * @property {Boolean} [clientFilter=false] Whether or not the tree will produce its own filtered list off the data.
         *
         * @property {Boolean} [multiSelect=false] Whether or not the tree supports multiselect operations.
         *
         * @property {Boolean} [displayRoot=false] Whether or not the tree should display the root node of the tree.
         * **Deprecated**: all nodes in the model are displayed.
         *
         * @property {Boolean} [flat=false] Whether or not the tree should render as a flat list.
         *
         * @property {Boolean} [picker=false] Whether or not the tree should apply picker styling. **Note**: This is an
         * internal option.
         *
         * @property {Array<String>} [defaultFilterFields=[]] The fields the tree should filter on in the data when
         * `clientFilter=true`.
         *
         * @property {Boolean} [filterShowHierarchy=false] Whether or not the tree should show the hierarchy when in a
         * filtered state.
         *
         * @property {Object} [storeFns={}] The functions to use as callbacks when loading, filtering, and retrieving current
         * data attributes.
         *
         * @property {TreePanel#load} [storeFns.load]
         *
         * @property {TreePanel#attrs} [storeFns.attrs]
         *
         * @property {TreePanel#filter} [storeFns.filter]
         *
         * @property {Object} [viewFns={}] The functions to use as callbacks when displaying the node.
         *
         * @property {TreePanel#display} [viewFns.display]
         *
         * @property {TreePanel#icons} [viewFns.icons]
         */

        var tpp,
            __WIDGET_NAME__ = 'tree-panel',
            __DEFAULTS__ = {
                move : false,
                lazyLoad : false,
                fluid : true,
                fill : true,
                clientFilter : false,
                multiSelect : false,
                displayRoot : false,
                flat : false,
                picker : false,
                defaultFilterFields : [],
                /**
                 * @property recordKeys
                 * @type Object
                 */
                recordKeys : {
                    /**
                     * The key to get the id from the record. Used to lookup nodes
                     *
                     * @property id
                     * @type {String}
                     */
//                    id : 'id',
                    /**
                     * The key to get the parentId of the record
                     *
                     * @property parentId
                     * @type {String}
                     */
//                    parentId : 'parentId'
                },
                filterShowHierarchy : false
            };

         /**
         * Tree Panel constructs a typical tree element to be placed on the DOM.
         * It is a UI element that supports both tree-structured data as well as flat data.
         *
         * It supports typical expand/collapse, selected, keyboard navigation, disabled and unselectable nodes.
         *
         *
         * @example
         *
         *      var treePanel = new TreePanel({
         *          storeFns : {
         *              load : function(id) {
         *                  if(id === null) {
         *                      return $.ajax({
         *                          url: 'mocks/topLevel.json',
         *                          type: 'GET'
         *                      });
         *                  }else{
         *                      return $.ajax({
         *                          url: 'mocks/personel' + id + '.json',
         *                          type: 'GET'
         *                      });
         *                  }
         *              },
         *              filter : function(query) {
         *                  console.log(query);
         *              }
         *          },
         *          viewFns : {
         *              display : function(node) {
         *                  //To be filled in with custom logic
         *                  return node.data.code + ' - ' + node.data.name;
         *              },
         *              icons : function(node) {
         *                  return '<i class="pgbu-icon pgbu-icon-ent-user"></i>';
         *              }
         *          }
         *          displayRoot : false,
         *          lazyLoad : false,
         *          flat : false,
         *          multiSelect : true
         *      }, initData);
         *
         * @class TreePanel
         * @since 1.2.0
         * @extends Widget
         * 
         * @param {TreePanel#Config} options The options to override the defaults.
         * @param {Array} [initData] The initial data if it has already been loaded somewhere.
         */
        function TreePanel(options, initData) {
            var model;

            Widget.call(this, options, initData);

            model = new TreeModel(this._options, this._initData);
            /**
             * The controller object that the panel keeps a reference to.
             *
             * @property _controller
             * @private
             * @type {TreeController}
             */
            this._controller = new TreeController(model, this._options);
            /**
             * The view element that the panel keeps a reference to.
             *
             * Used for hide/show otherwise everything should go
             * through the controller.
             *
             * @property _view
             * @private
             * @type {TreeView}
             */
            this._view = new TreeView(model, this._controller, this._options, this._settings);
            model.trigger('load', model);

            Object.defineProperty(this, '_data', {
                get : this.selected,
                set : this.selected
            });

            this._setupEvents([model, this._view]);
        }

        TreePanel.prototype = Object.create(Widget.prototype);

        tpp = TreePanel.prototype;

        tpp.constructor = TreePanel;

        /**
         * Used to prefix options.
         *
         * Currently unused due to the change in options.
         *
         * Plan on this name being used for public event prefixes.
         *
         * @method TreePanel#name
         * @return {String} 'tree-panel' The name used in the widget for events
         */
        tpp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Calls initial load on the controller.
         *
         * @method TreePanel#_init
         * @private
         */
        tpp._init = function () {
            //Used for first render after append
            this._controller.load(null);
        };

        /**
         * @method TreePanel#_translate
         * @private
         */
        tpp._translate = function (name) {
            var that = this;

            function getDataForIds(arr) {
                return _.map(arr || [], function(id) {
                    return that._controller.node(id).data;
                });
            }

            if ( name === 'select' ) {
                return [ name, getDataForIds(arguments[1]) , getDataForIds(arguments[2]) ];
            } else if ( name === 'node-added' ) {
                return [ name, arguments[1].data, arguments[2] ];
            } else if ( name === 'filter' ) {

                if(arguments[1] == null) {
                    return arguments;
                }

                var filteredData = _.map(arguments[1] || [], function(node) {
                    return that._controller.node(node.id).data;
                });

                return _.union([ name, filteredData], _.rest(arguments, 2));
            }

            return arguments;
        };

        /**
         * Gets the selected elements of the tree.
         *
         * @method TreePanel#getSelected
         * @deprecated
         * @return {Array} - The array of Trees that are currently selected.
         */
        tpp.getSelected = function () {
            logger.warn('The use of getSelected (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            return {
                loaded : this._controller.selected(),
                nonloaded : this._controller.nonLoadedSelected()
            };
        };

        /**
         * Getter/Setter for the selected nodes in the TreePanel. When a parameter
         * is passed in the method will behave like a setter. If no parameter is
         * passed in then it will return the selected nodes.
         *
         * @since 1.3.0
         *
         * @method TreePanel#selected
         * @param {Array} [nodes] The nodes to set as selected in the TreePanel.
         * @return {Array} selected When used as a getter the data of the selected nodes.
         *
         * @example
         *     var treePanel;
         *
         *     //Initialization...
         *
         *     treePanel.selected(['1', '2', '3']); //Will use the setter
         *
         *     treePanel.selected(); //Returns ['1', '2', '3']
         *
         *     //If you want to add to the current selection
         *     treePanel.selected(_.union(treePanel.selected(), ['4', '5', '6']));
         */
        tpp.selected = function (nodes) {
            if (!arguments.length) {
                var selectedIds = this._controller.selected(),
                    selected = [], controller = this._controller,
                    nonLoaded = this._controller.nonLoadedSelected();

                _.each(selectedIds, function (id) {
                    selected.push(controller.node(id).data);
                });

                _.each(nonLoaded, function (data) {
                    selected.push(data);
                });

                return selected;
            }

            this._controller.selected(nodes);
            this._view.trigger('picker-accept', this.selected());
            this.trigger('picker.accept', this.selected());
        };

        /**
         * Getter/Setter for the selected nodes in the TreePanel. When a parameter
         * is passed in the method will behave like a setter. If no parameter is
         * passed in then it will return the selected nodes.
         *
         * **Note**: This function will not fire events. For more information on the alternate function see {@link
         * TreePanel#selected}.
         *
         * @method TreePanel#value
         * @param {Array} [nodes] The nodes to set as selected in the TreePanel.
         * @return {Array} selected When used as a getter the data of the selected nodes.
         *
         * @example
         *     var treePanel;
         *
         *     //Initialization...
         *
         *     treePanel.selected(['1', '2', '3']); //Will use the setter
         *
         *     treePanel.selected(); //Returns ['1', '2', '3']
         *
         *     //If you want to add to the current selection
         *     treePanel.selected(_.union(treePanel.selected(), ['4', '5', '6']));
         */
        tpp.value = function (nodes) {
            if(!arguments.length) {
                return this.selected();
            }

            this._controller.selected(nodes, { silent : true });
            this._view.trigger('picker-accept', this.selected());
            this.trigger('picker.accept', this.selected());
        };

        /**
         * Used to transform a client-side id into the JSON representation
         * of the node.
         *
         * @method TreePanel#node
         * @param {String|Object} id The client-side id or an unique attribute or set of
         *  attributes of the data.
         * @return node The JSON representation of the node. Returns null if a node was not
         *  found.
         */
        tpp.node = function (id) {
            var node = this._controller.node(id);

            if (!_.isNull(node)) {
                //Returns the node based on the string id the user has.
                return node.toJSON();
            } else {
                return node;
            }
        };

        /**
         * Transforms nodes into their client-side ids.
         *
         * @method TreePanel#ids
         * @param {Array} nodes The nodes to get the client-side ids for.
         * @return ids The client-side ids of the nodes.
         */
        tpp.ids = function (nodes) {
            var ids = [], controller = this._controller;

            _.each(nodes, function (id) {
                ids.push(controller.node(id).id);
            });

            return ids;
        };

        /**
         * Clear the selection on the tree panel.
         *
         * @method TreePanel#clearSelection
         * @deprecated
         */
        tpp.clearSelection = function () {
            logger.warn('The use of clearSelection (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.selected([]);
        };

        /**
         * Sets the selection for the tree.
         *
         * @method TreePanel#setSelection
         * @deprecated
         * @param {Array} data The data to be the new selection, overwrites the current selection.
         */
        tpp.setSelection = function (data) {
            logger.warn('The use of setSelection (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.selected(data);
        };

        /**
         * Adds nodes to the tree panel.
         *
         * @method TreePanel#addData
         * @deprecated
         * @param {Tree} node The node to add to the tree panel.
         */
        tpp.addData = function (node) {
            logger.warn('The use of addData is now deprecated. Please use the combined method "add" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.add(node);
        };

        /**
         * Adds data to either the root node or a specified node if parentNode is given.
         *
         * @since 1.3.0
         *
         * @example
         *     var tp = new TreePanel(); //Assuming correct parameters
         *
         *     //... Initialization here
         *
         *     //Adding to the root node
         *     tp.add([
         *         { name : 'Example obj1'},
         *         { name : 'Example obj2'}
         *     ]);
         *
         *     //Adding to the previously added example obj1
         *     tp.add({name : 'Example obj1'}, [
         *         { name : 'Example of child of obj1'}
         *     ]);
         *
         * @method TreePanel#add
         * @param {Object} [parentNode] The object representing the data of the node.
         * @param {Array} nodes Array of objects that represent the nodes in the tree.
         * @param {Number} [index] The index of the parent to add the new node at. Added at the end when not specified.
         */
            //TODO: should this accept a map of parentNode to children to allow
            //for multi parent adds?
        tpp.add = function (parentNode, nodes, index) {
            this._controller.add(parentNode, nodes, index);
        };

        /**
         * Removes elements from the tree panel.
         *
         * Updating this remove element to take in any attribute in data or array of attributes
         * as well as the client-side id.
         *
         * @method TreePanel#removeElement
         * @deprecated
         * @param {String|Array} data The data or id of the node to remove.
         */
        tpp.removeElement = function (data) {
            logger.warn('The use of removeElement is now deprecated. Please use the combined method "remove" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.remove(data);
        };

        /**
         * Removes elements from the tree and all their respective children.
         *
         * @method TreePanel#remove
         * @param {Array} nodes The array of nodes to be removed. Can be either client-side ids or
         *  Objects representing the data of the node to remove.
         */
        tpp.remove = function (nodes) {
            var controller = this._controller;

            _.each(nodes, function (node) {
                controller.remove(node);
            });
        };

        /**
         * Allows access to the entire tree's data.
         *
         * @method TreePanel#getTree
         * @deprecated
         * @return - The full representation of the tree.
         */
        tpp.getTree = function () {
            logger.warn('The use of getTree (and get/set methods in general) in OUI is now deprecated. Please use the combined method "tree" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            return this._controller.tree();
        };

        /**
         * Returns the JSON representation of the tree.
         *
         * @example
         *      //Example return format.
         *      { //Root node
         *          children : [
         *              {
         *                  "name" : "Example name 1", //name came back on load
         *                  //Will have whatever attributes were returned
         *                  //from the load function
         *                  "children" : [] //Can have nested objects
         *              },
         *              {
         *                  "name" : "Example name 2",
         *                  "children" : []
         *              }
         *          ]
         *      }
         *
         * @method TreePanel#tree
         * @return - The JSON representation of the tree.
         */
        tpp.tree = function () {
            if (this._options.flat) {
                return this._controller.tree().toJSON().children;
            } else {
                return this._controller.tree().toJSON();
            }
        };

        /**
         * Gets the defaults for the tree panel.
         *
         * @method TreePanel#defaults
         * @return {Object} The defaults for the tree panel.
         */
        tpp.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Filters the treePanel given the query and the array of fields.
         *
         * @method TreePanel#filter
         * @param {String} query The string to match to the fields of the nodes
         * @param {Array} fields The array of strings that match the fields to match on.
         */
        tpp.filter = function (query, fields) {
            var filterFields = fields || this._options.defaultFilterFields;

            if (_.isUndefined(query)) {
                this._controller.filter(null, filterFields);
            } else {
                this._controller.filter(query, filterFields);
            }
        };

        /**
         * Hides the view and clears the expansion and selections.
         *
         * @method TreePanel#hide
         */
        tpp.hide = function () {
            this._view.hide();
            this._controller.selected([]);
            this._controller.expanded([null]);
        };

        /**
         * Shows the view and selects the nodes given as well as expands to them.
         *
         * @method TreePanel#show
         */
        tpp.show = function (selectedNodes) {
            this._view.show();
            this._controller.selected(selectedNodes);
            if (!_.isEmpty(selectedNodes)) {
                this._controller.expandPathTo(selectedNodes);
            }
        };

        /**
         * Toggles the selectable status of the nodes given.
         *
         * @method TreePanel#toggleEnable
         * @param {Array} nodes The nodes to toggle.
         */
        tpp.toggleEnable = function (nodes) {
            logger.warn('The use of toggleEnable (and get/set methods in general) in OUI is now deprecated. Please use the combined method "disabled" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.disabled(nodes);
        };

        /**
         * Toggles whether the nodes are selectable or not.
         *
         * If a node in the array is unselectable it will become selectable.
         * If a node in the array is selectable it will become unselectable.
         *
         * This could cause issues if even the unselectable nodes are to stay
         * unselectable.
         *
         * @method TreePanel#toggleSelectable
         * @deprecated
         * @param {Array} nodes The nodes to toggle.
         */
        tpp.toggleSelectable = function (nodes) {
            logger.warn('The use of toggleSelectable (and get/set methods in general) in OUI is now deprecated. Please use the combined method "unselectable" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');

            this._controller.toggleSelectable(nodes);
        };

        /**
         * Disables the nodes that are supplied or returns the currently disabled nodes
         * if no parameter is specified.
         *
         * Overwrites currently disabled nodes with new nodes.
         *
         * The disabled attribute affects the node supplied as well as all of its children.
         *
         * Useful when shifting a node to another tree and the children are not allowed to be selected
         * after the parent has already been selected.
         *
         * @example
         *     //Example comment
         *     var treePanel = new TreePanel(); //Assuming correct construction.
         *
         *     //...Assuming initialization and loading
         *
         *     treePanel.disabled([
         *         {
         *             "name" : "example loaded name"
         *         }
         *     ]); //Will disable the node that has that name.
         *
         *     treePanel.disabled([]); //Will remove all disabled nodes.
         *
         *     treePanel.disabled(); //Will use the getter and return array of nodes' data.
         *
         *     treePanel.disabled(_.union(treePanel.disabled(), [
         *         {
         *             "name" : "additional node to disable"
         *         }
         *     ])); //Adds to the currently disabled list.
         *
         * @since 1.3.0
         * @method TreePanel#disabled
         * @param {Array} [nodes] The nodes to mark as disabled.
         * @return - The array of disabled nodes' data.
         */
        tpp.disabled = function (nodes) {
            var disabled = [], controller = this._controller;

            if (!arguments.length) {
                _.each(this._controller.disabled(), function (id) {
                    disabled.push(controller.node(id).data);
                });

                return disabled;
            }

            _.each(nodes, function (node) {
                disabled.push(controller.node(node).id);
            });

            this._controller.disabled(disabled);
        };

        /**
         * Marks the nodes that are supplied as unselectable or returns the currently disabled nodes
         * if no parameter is specified.
         *
         * Overwrites currently unselectable nodes with new nodes.
         *
         * Only works on supplied node(s). The attribute does **not** affect children in the way that the
         * disabled attribute does.
         *
         * @example
         *     //Example comment
         *     var treePanel = new TreePanel(); //Assuming correct construction.
         *
         *     //...Assuming initialization and loading
         *
         *     treePanel.unselectable([
         *         {
         *             "name" : "example loaded name"
         *         }
         *     ]); //Will mark the node that has that name as unselectable.
         *
         *     treePanel.unselectable([]); //Will remove all unselectable nodes.
         *
         *     treePanel.unselectable(); //Will use the getter and return array of nodes' data.
         *
         *     treePanel.unselectable(_.union(treePanel.unselectable(), [
         *         {
         *             "name" : "additional node to mark"
         *         }
         *     ])); //Adds to the currently unselectable list.
         *
         * @since 1.3.0
         * @method TreePanel#unselectable
         * @param {Array} [nodes] The nodes to mark as unselectable.
         * @return - The array of unselectable nodes' data.
         */
        tpp.unselectable = function (nodes) {
            var unselectable = [], controller = this._controller;
            if (!arguments.length) {

                _.each(this._controller.unselectable(), function (id) {
                    var node = controller.node(id);
                    if (!_.isNull(node)) {
                        unselectable.push(node.data);
                    }
                });
                return unselectable;
            }

            if (!_.isFunction(nodes)) {
                _.each(nodes, function (node) {
                    unselectable.push(controller.node(node).id);
                });

                this._controller.unselectable(unselectable);
            } else {
                this._controller.unselectable(nodes);
            }
        };

        /**
         * Marks the nodes that are supplied as expanded or returns the currently disabled nodes
         * if no parameter is specified.
         *
         * Overwrites currently expanded nodes with new nodes.
         *
         * @example
         *     //Example comment
         *     var treePanel = new TreePanel(); //Assuming correct construction.
         *
         *     //...Assuming initialization and loading
         *
         *     treePanel.expanded([
         *         {
         *             "name" : "example loaded name"
         *         }
         *     ]); //Will mark the node that has that name as expanded.
         *
         *     treePanel.expanded([]); //Will remove all expanded nodes.
         *
         *     treePanel.expanded(); //Will use the getter and return array of nodes' data.
         *
         *     treePanel.expanded(_.union(treePanel.expanded(), [
         *         {
         *             "name" : "additional node to mark"
         *         }
         *     ])); //Adds to the currently expanded list.
         *
         * @since 1.3.0
         * @method TreePanel#expanded
         * @param {Array} [nodes] The nodes to mark as expanded.
         * @return - The array of expanded nodes' data.
         */
        tpp.expanded = function (nodes) {
            var expanded = [], controller = this._controller;
            if (!arguments.length) {

                _.each(this._controller.expanded(), function (id) {
                    var node = controller.node(id);
                    if (!_.isNull(node) && !_.isNull(node.id)) {
                        expanded.push(node.data);
                    }
                });
                return expanded;
            }

            _.each(nodes, function (node) {
                expanded.push(controller.node(node).id);
            });

            this._controller.expanded(expanded);
        };

        /**
         * Moves the currently active node up in its parent's subtree.
         *
         * The node will not move to another tree.
         * @method TreePanel#moveUp
         */
        tpp.moveUp = function () {
            this._controller.moveUp();
        };

        /**
         * Moves the currently active node down in its parent's subtree.
         *
         * The node will not move to another subtree/parent.
         * @method TreePanel#moveDown
         */
        tpp.moveDown = function () {
            this._controller.moveDown();
        };

        /**
         * Remove all data from the model (expanded, selected, disabled,...) and reloads
         * from the root node.
         *
         * @method TreePanel#invalidate
         */
        tpp.invalidate = function () {
            this._controller.invalidate();
        };

        /**
         * Updates existing nodes with new data.
         *
         * @method TreePanel#updated
         * @param {Array} nodes The array of nodes to update
         */
        tpp.updated = function (nodes) {
            this._controller.updateNodes(nodes);
        };

        /**
         * @method TreePanel#filtered
         * @return {Array} The filtered nodes
         */
        tpp.filtered = function () {
            return _.map(this._controller.filtered(), function(tree) {
                return tree.toJSON();
            });
        };

        /**
         * @method TreePanel#isFiltered
         * @return {Boolean} Whether or not the tree panel is currently filtered
         */
        tpp.isFiltered = function () {
            return this._controller.isFiltered();
        };

        return TreePanel;
    }
);

define('oui.typeAhead/typeAheadModel',['oui.underscore', 'oui.utils/subscribable', 'oui.utils/uniqueId'],
    function (_, Subscribable, uniqueId) {
        

        /**
         * @event TypeAheadModel#result-added
         * @param {Array} results The results that match the query
         */
        /**
         * @event TypeAheadModel#loaded
         * @param {Array} nodes The nodes used internally in the typeahead model
         * @param {String} query The query for the loaded nodes
         */
        /**
         * @event TypeAheadModel#selection-error
         */
        /**
         * @event TypeAheadModel#select
         */
        /**
         * @event TypeAheadModel#invalidate
         */
        /**
         * @event TypeAheadModel#active
         * @param {Number} id The active ide
         */

        var tmp;

        /**
         * The data associated with the with the typeahead component.
         *
         * @class TypeAheadModel
         * @private
         * @since 1.3.0
         * @extends Subscribable
         * 
         * @param {TypeAhead#Config} options Options passed from the widget.
         */
        function TypeAheadModel(options) {
            Subscribable.call(this);

            /**
             * Options for typeahead
             *
             * @property _options
             * @type {Object}
             * @private
             */
            this._options = options;
            /**
             * The results map. An object that maps query to results.
             *
             * @property _resultMap
             * @type {Object}
             * @private
             * @default {}
             */
            this._resultMap = {};
            this._values = [];
            this._active = {
                query : null,
                node : null
            };

            /**
             * The selected nodes in the model.
             *
             * @property _selectedNodes
             * @type {Array}
             * @private
             * @default []
             */
            this._selectedNodes = [];
        }

        TypeAheadModel.prototype = Object.create(Subscribable.prototype);
        tmp = TypeAheadModel.prototype;

        tmp.constructor = TypeAheadModel;

        /**
         * Adds the results and sets an expire time to keep the results current.
         *
         * @method TypeAheadModel#addResult
         * @param {String} query The key to add to the result map
         * @param {Array} results The values in the map.
         */
        tmp.addResult = function (query, results) {
            var that = this,
                expires = new Date(),
                clientNodes = [],
                node;

            expires.setMinutes(expires.getMinutes() + 5);

            _.each(results, function (result) {
                node = that.getNodeByData(result);
                clientNodes.push({
                    id : _.isUndefined(node) ? uniqueId('typeahead-node') + '' : node.id,
                    data : result
                });
            });

            this._resultMap[query] = {
                results : clientNodes,
                expires : expires
            };

            this.trigger('result-added', results);
            this.trigger('loaded', clientNodes, query);
        };

        /**
         * Check to see if the current data is after the time stamp expired tag.
         *
         * @method TypeAheadModel#isExpired
         * @param {String} query The results to see if they have expired.
         * @returns {Boolean} - true if the current data is after the time stamp.
         */
        tmp.isExpired = function (query) {
            var timeStamp = this._resultMap[query].expires;

            return new Date() > timeStamp;
        };

        /**
         * Whether or not the query has been loaded.
         *
         * @method TypeAheadModel#isLoaded
         * @param {String} query The query to check to see if it is loaded.
         * @returns {Boolean} - true if the query is undefined in the result map or if they query is expired.
         */
        tmp.isLoaded = function (query) {
            return !_.isUndefined(this._resultMap[query]) && !this.isExpired(query);
        };

        /**
         * Returns the results for a given query.
         *
         * @method TypeAheadModel#getResults
         * @param {String} query The query to get the results for.
         * @return {Object} - The results.
         */
        tmp.getResults = function (query) {
            if (query === null) {
                return [];
            }
            return this._resultMap[query].results;
        };

        tmp.getNodeByData = function (data) {
            return _.find(this.getAllNodes(), function (node) {
                return _.isEqual(data, _.pick(node.data, _.keys(data)));
            });
        };

        tmp.getNodeById = function (id) {
            return _.findWhere(this.getAllNodes(), { id : id });
        };

        tmp.getNodeByValue = function (str) {
            var that = this;
            return _.find(this.getAllNodes(), function (node) {
                return that._options.viewFns.display(node.data) === str;
            });
        };

        tmp.remove = function (node) {
            this._selectedNodes = _.without(this._selectedNodes, node);
        };

        /**
         * @method TypeAheadModel#selected
         * @param {Array|Object} [nodes]
         * @param {Object} eventParams The parameters to pass to trigger
         */
        tmp.selected = function (nodes, eventParams) {
            var that = this, tpNode;

            if (!arguments.length) {
                //When calling the model directly when we say we want selected nodes
                //we mean it.
                return this._selectedNodes;
            }

            if(eventParams == null) {
                eventParams = {};
            }

            if (!this._options.multiSelect && nodes.length > 1) {
                //Fire error
                this.trigger('selection-error');
                return;
            }

            //Cant overwrite the array
            var selectedNodes = [];

            _.each(nodes, function (node) {
                if (_.isString(node) || _.isNumber(node)) {
                    //Assuming this is a client side id.
                    selectedNodes.push(that.getNodeById(node));
                } else if (_.isObject(node) && !_.isUndefined(node.data)) {
                    selectedNodes.push(node);
                } else if (_.isObject(node)) {
                    tpNode = that.getNodeByData(node);

                    if (_.isUndefined(tpNode)) {
                        var clientSide = {
                            id : uniqueId('typeahead-node') + '',
                            data : node
                        };
                        that._values.push(clientSide);
                        selectedNodes.push(clientSide);
                    } else {
                        selectedNodes.push(tpNode);
                    }
                }
            });

            selectedNodes = _.uniq(selectedNodes);

            if(!_.isEqual(this._selectedNodes, selectedNodes)) {
                this._selectedNodes = selectedNodes;
                eventParams.name = 'select';
                this.trigger(eventParams, this._selectedNodes);
            }
        };

        tmp.active = function (obj) {
            if (!arguments.length) {
                return this._active;
            }

            if (_.isObject(obj)) {
                //We have a node
                this._active.node = obj;
            } else if (_.isString(obj)) {
                this._active.query = obj;
                if (!_.isUndefined(this._resultMap[obj]) && this._resultMap[obj].results.length > 0) {
                    this._active.node = this._resultMap[obj].results[0];
                } else {
                    this._active.node = null;
                }
            } else {
                this._active.query = null;
                this._active.node = null;
            }

            this.trigger('active', this._active);
        };

        tmp.getAllNodes = function () {
            return _.chain(this._resultMap)
                .map(function (value) {
                    return value.results;
                })
                .union(this._values)
                .flatten()
                .uniq()
                .value();
        };

        tmp.invalidate = function () {
            this._resultMap = {};
            this._values = [];
            this._active.query = null;
            this._active.node = null;

            this.trigger('invalidate');

        };

        return TypeAheadModel;
    }
);

define('oui.typeAhead/typeAheadController',['oui.underscore', 'oui.jquery'], function(_, $) {
    

    var tcp;

    function getQuery (value, cursorLocation, multiSelect, delimiter) {
        if (!multiSelect || (multiSelect && value.indexOf(delimiter) === -1)) {
            return value;
        } else {
            var upperBound = value.indexOf(delimiter, cursorLocation),
                lowerBound = value.lastIndexOf(delimiter, cursorLocation - 1) + 1;
            if (upperBound === -1) { upperBound = value.length; }
            if (lowerBound === -1) { lowerBound = 0; }
            return  value.substring(lowerBound, upperBound).trim();
        }
    }

    /**
     * @class TypeAheadController
     * @private
     * @since 1.3.0
     * 
     * @param {Object} model The typeahead model to manipulate the data
     * @param {Object} controllerOptions The specific options regarding the controller.
     */
    function TypeAheadController(model, controllerOptions) {
        /**
         * The model object that the controller can manipulate.
         *
         * @property _model
         * @private
         * @type {TypeAheadModel}
         */
        this._model = model;
        this._options = controllerOptions;
    }

    tcp = TypeAheadController.prototype;

    tcp.constructor = TypeAheadController;

    /**
     * Either searches using the provided function into the options or uses
     * the results from the past search to return.
     *
     * @method TypeAheadController#search
     * @param {String} value The value of the text input
     * @param {Number} cursorLocation The index of the cursor location
     * @return {Boolean|undefined}
     */
    tcp.search = function ( value, cursorLocation ) {
        var that = this, query = getQuery(value, cursorLocation, this._options.multiSelect, this._options.delimiter);

        if((_.isEmpty(query.trim()) && !this._options.allowEmpty) || query.trim().length < this._options.minLength) {
            this._model.active(null);

            if(!(this._options.multiSelect && this._model.selected().length > 0)) {
                this._model.trigger('clear');
            }
            /**
             * Triggered when data is loaded for a search query
             *
             * @event TypeAheadModel#loaded
             * @param {Array} results an empty array
             * @param {String} query the query that triggered the results
             */
            this._model.trigger('loaded', [], '');

            if(this._options.storeFns && this._options.storeFns.clear) {
                this._options.storeFns.clear();
            }

            return false;
        }

        if (query !== this._model.active().query) {
            this._model.trigger('search', query);
        }

        if(!this._options.search) {
            this._model.active(query);
            return false;
        }

        if ( this._model.isLoaded(query) ) {
            this._model.trigger('loaded', this._model.getResults(query), query);
            this._model.active(query);
        } else {
            var searchFn = this._options.storeFns.search;
            /**
             * Triggered to tell the view that an asynchronous
             * event (ajax) request is taking place and to show the user.
             *
             * @event TypeAheadModel#loading
             */
            this._model.trigger('loading');
            $.when(searchFn.call(this, query)).done(function (data) {
                that._model.addResult(query, data);
                that._model.active(query);
            });
        }
    };

    tcp.getNodeById = function(id) {
        return this._model.getNodeById(id);
    };

    tcp.node = function (str) {
        return this._model.getNodeByValue(str);
    };

    tcp.remove = function (node) {
        this._model.remove(node);
    };

    tcp.active = function (node) {
        if(!arguments.length) {
            return this._model.active();
        }

        this._model.active(node);
    };

    /**
     * Marks the next node in the current list as active.
     *
     * @method TypeAheadController#activeNext
     */
    tcp.activeNext = function() {
        var active = this._model.active(),
            results = this._model.getResults(active.query),
            index = _.indexOf(results, active.node);

        if(results.length > 0) {
            this._model.active(results[(index + 1) % results.length]);
        }
    };

    /**
     * Marks the previous node in the current list as active.
     *
     * @method TypeAheadController#activePrevious
     */
    tcp.activePrevious = function() {
        var active = this._model.active(),
            results = this._model.getResults(active.query),
            index = _.indexOf(results, active.node);

        if(results.length > 0) {
            if(index < 0) {
                index = results.length;
            }
            this._model.active(results[index - 1]);
        }
    };

    /**
     * Selects a given node based on id and when no id is given it just
     * selects the currently active node.
     *
     * @method TypeAheadController#select
     * @param {String} id The id of the node to select
     */
    tcp.selected = function(id, eventParams) {
        if(!arguments.length) {
            return this._model.selected();
        }

        this._model.selected(id, eventParams);
    };

    tcp.invalidate = function () {
        this._model.invalidate();
    };

    return TypeAheadController;
});


define('oui.typeAhead/typeAheadTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<input type="text" placeholder="'+
((__t=( data.locale.placeholder  ))==null?'':__t)+
'" />';
return __p;
};});


define('oui.typeAhead/typeAheadSearchTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<input type="search" class="empty" placeholder="'+
((__t=( data.locale.placeholder ))==null?'':__t)+
'" title="'+
((__t=( data.locale.search  ))==null?'':__t)+
'"><button class="btn btn-picker"> <span class="search-icon"><i class="pgbu-icon-small pgbu-icon-magnifying-glass"></i></span></button><a title="'+
((__t=( data.locale.clear ))==null?'':__t)+
'" class="search-clear-icon" href="#"> <i class="pgbu-icon pgbu-icon-find-clear"></i></a>';
return __p;
};});


define('oui.typeAhead/typeAheadMenuTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<ul role="menu" class="orcl-type-ahead-menu dropdown-menu '+
((__t=( data.scroll ? 'scroll': '' ))==null?'':__t)+
'" style="max-height: '+
((__t=(data.height))==null?'':__t)+
'"> '+
((__t=( data.nodes  ))==null?'':__t)+
'</ul>';
return __p;
};});


define('oui.typeAhead/typeAheadNodeTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<li role="option" id="'+
((__t=( data.pageId ))==null?'':__t)+
'" data-id="'+
((__t=( data.id))==null?'':__t)+
'" class="orcl-type-ahead-node '+
((__t=(data.active))==null?'':__t)+
'" title="'+
((__t=(data.title))==null?'':__t)+
'"><a>'+
((__t=(data.text ))==null?'':__t)+
'</a></li>';
return __p;
};});


define('oui.typeAhead/tokenTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<span class="token-container" data-id="'+
((__t=( data.id))==null?'':__t)+
'" title="'+
((__t=(data.value ))==null?'':__t)+
'"> <span class="token"> '+
((__t=( data.value  ))==null?'':__t)+
' </span> <a class="token-remove" href="#">&times;</a></span>';
return __p;
};});


define('oui.typeAhead/tokenContainerTemplate.htm',[],function () { return '<div class="token-list-container">\n    <div class="token-list">\n\n    </div>\n</div>\n';});


define('oui.tree/spinner.template',[],function () { return '<i class="loading-spinner pgbu-icon pgbu-icon-loading"></i>\n';});

define('oui.typeAhead/typeAheadTokenView',['oui.underscore', 'oui.jquery', 'oui.base/view', 'oui.utils/stringUtils', 'oui.utils/keyCode',
    'oui.mixins/popup',
    'oui.typeAhead/typeAheadTemplate.htm',
    'oui.typeAhead/typeAheadSearchTemplate.htm',
    'oui.typeAhead/typeAheadMenuTemplate.htm',
    'oui.typeAhead/typeAheadNodeTemplate.htm',
    'oui.typeAhead/tokenTemplate.htm',
    'oui.typeAhead/tokenContainerTemplate.htm',
    'oui.tree/spinner.template'
],
    function (_, $, View, stringUtils, keyCode, popupMixin, typeAheadTemplate, typeAheadSearchTemplate, menuTemplate,
              nodeTemplate, tokenTemplate, tokenContainerTemplate, loadingTemplate) {
        

        var tvp;

        /**
         * The type ahead view is responsible for displaying the suggestions when they loaded.
         *
         * @class TypeAheadTokenView
         * @private
         * @since 2.1.0
         * @extends View
         * @mixes popup
         * 
         * @param {TypeAheadModel} model The typeahead model that is responsible for holding all the past
         *  search results
         * @param {TypeAheadController} controller The type ahead controller that will modify the model.
         * @param {Object} element jQuery element for the view
         * @param {Object} viewOpts The options for the view.
         */
        function TypeAheadView(model, controller, element, viewOpts, settings) {
            var that = this;

            View.call(this, viewOpts, element, controller, settings);

            /**
             * The menu object for this typeahead
             *
             * A JQuery object.
             *
             * @property _$menu
             * @private
             * @type {Object}
             */
            this._$menu = null;
            this._$input = this._$el.find('input');
            this._$tokListContainer = this._$el.find('.token-list-container');
            this._$tokList = this._$el.find('.token-list');
            this._padding = null;

            this._inputEl = this._$input[0];

            model.on({
                'view.loaded' : {fn : this.loaded, context : this },
                'view.active' : {fn : this.active, context : this },
                'view.loading' : {fn : this.loading, context : this },
                'view.select' : {fn : this.select, context : this },
                'view.selection-change' : {fn : this.select, context : this },
                'view.invalidate' : {fn : this.clear, context: this }
            });

            this.element().on('oui-accept oui-picker-accept', function(event, widget, name, selected) {
                if(widget.name() === 'picker') {
                    that._controller.selected(selected, { silent : true });
                }
            });

            var debounced = _.debounce(controller.search, this._options.delay);

            this._$input
                .on('click', function () {
                    that.focusText();
                })
                .on('keyup', function (event) {

                    switch (event.keyCode) {
                    case keyCode.DOWN:
                    case keyCode.UP:
                    case keyCode.LEFT:
                    case keyCode.RIGHT:
                    case keyCode.ENTER:
                    case keyCode.TAB:
                    case keyCode.SHIFT:
                    case keyCode.F2:
                        break;
                    case keyCode.BACKSPACE:
                        if (!_.isEmpty(that._inputEl.value) &&
                            that._$el.find('.token-container.selected').length === 0) {
                            debounced.call(controller, this.value, this.selectionStart);
                        } else if (_.isEmpty(that._inputEl.value)) {
                            that._$menu.remove();
                        }
                        break;
                    default:
                        debounced.call(controller, this.value, this.selectionStart);
                        if (that._controller.selected().length > 0 && that._$input.width() < that._$input[0].scrollWidth) {
                            var offset = that._$input[0].scrollWidth - that._$input.width();

                            that._$tokList.css('margin-left', parseFloat(that._$tokList.css('margin-left')) - offset);
                            that._$input.css('padding-left', parseFloat(that._$input.css('padding-left')) - offset);
                        }
                    }
                })
                .on('blur', function (event) {
                    event.preventDefault();
                    if ((_.isUndefined(that._hideMenu) || that._hideMenu) && !_.isNull(that._$menu)) {
                        that._$menu.remove();
                        that._$el.find('.token-container.selected').removeClass('selected');
                    }
                })
                .on('keydown', function (event) {
                    var selectedToken = that._$el.find('.token-container.selected');

                    if (that._options.search) {
                        switch (event.keyCode) {
                        case keyCode.TAB:
                        case keyCode.SHIFT:
                        case keyCode.F2:
                            break;
                        case keyCode.DOWN:
                            event.preventDefault();
                            controller.activeNext();
                            break;
                        case keyCode.UP:
                            event.preventDefault();
                            controller.activePrevious();
                            break;
                        case keyCode.ENTER:
                            if (!_.isNull(that._$menu) && that.isVisible() && that._$menu.find('.active').length !== 0) {
                                event.preventDefault();
                                controller.selected(_.union(controller.selected(), that._$menu.find('.active').attr('data-id')));
                            }
                            break;
                        case keyCode.LEFT:
                            if (selectedToken.length !== 0) {
                                if (!selectedToken.is(':first-child')) {
                                    that.focusToken(selectedToken.prev('.token-container'));
                                    event.preventDefault();
                                }
                            } else if (this.selectionStart === 0 && that._controller.selected().length > 0) {
                                that.focusToken(that._$el.find('.token-container:last-child'));
                                event.preventDefault();
                            }
                            break;
                        case keyCode.RIGHT:
                            if (selectedToken.length !== 0 && selectedToken.is(':last-child')) {
                                //We are moving from the last token to the text input
                                that.focusText();
                                debounced.call(controller, this.value, this.selectionStart);
                                event.preventDefault();
                            } else if (selectedToken.length !== 0) {
                                //We are moving to the next token to the right
                                that.focusToken(selectedToken.next('.token-container'));
                                event.preventDefault();
                            }
                            break;
                        case keyCode.BACKSPACE:
                            //If token is already selected - remove it and move to the next one
                            if (selectedToken.length !== 0) {
                                var id = selectedToken.attr('data-id');
                                controller.selected(_.without(controller.selected(), controller.getNodeById(id)));
                                event.preventDefault();
                            } else if (this.selectionStart === 0) {
                                that._$el.find('.token-container:last-child').addClass('selected');
                                event.preventDefault();
                            }
                            break;
                        default:
                            if (that._controller.selected().length > 0 &&
                                that._$input.width() < that._$input[0].scrollWidth) {

                                that.expandTextArea();
                            } else {
                                that.focusText();
                                debounced.call(controller, this.value, this.selectionStart);
                            }
                        }
                    }
                });
        }

        TypeAheadView.prototype = Object.create(View.prototype);
        tvp = TypeAheadView.prototype;

        tvp.constructor = TypeAheadView;

        _.extend(tvp, popupMixin);

        tvp.destroy = function () {
            this._$input.off('keyup blur focus keydown');
            this._$el.remove();
        };

        tvp.expandTextArea = function () {
            var computed = this._$input[0].scrollWidth - this._$input.width();

            this._$tokList.css('margin-left', parseFloat(this._$tokList.css('margin-left')) - computed);
            this._$input.css('padding-left', parseFloat(this._$input.css('padding-left')) - computed);
        };

        tvp.focusText = function () {
            //Move the text
            var selectedToken = this._$el.find('.token-container.selected'),
                selected = this._controller.selected(),
                symetricalPadding = this._padding,
                tokListWidth = this._$tokList.width(),
                maxWidth,
                tokListContWidth;

            selectedToken.removeClass('selected');
            if (this._$input.hasClass('no-select')) {
                this._$input.removeClass('no-select');
                //In order to get the caret to reappear
                this._$input.blur();
                this._$input.focus();
            }

            //TODO: remove hardcoded 50px
            this._$tokListContainer.css('max-width', this._$input.innerWidth() - 50);
            tokListContWidth = this._$tokListContainer.width();
            maxWidth = parseFloat(this._$tokListContainer.css('max-width'));

            if (tokListWidth > tokListContWidth && tokListWidth > maxWidth) {
                this._$tokList.css('margin-left', tokListContWidth - tokListWidth);
            } else {
                this._$tokList.css('margin-left', 0);
            }

            if (selected.length === 0) {
                this._$input.css('padding-left', symetricalPadding);
            } else if (tokListWidth < maxWidth && tokListContWidth !== tokListWidth) {
                this._$input.css('padding-left', this._$tokListContainer.outerWidth(true) - (tokListContWidth - tokListWidth));
            } else {
                this._$input.css('padding-left', this._$tokListContainer.outerWidth(true));
            }
        };

        tvp.focusToken = function ($token) {
            var newPos = $token.position().left + $token.outerWidth();
            //knows that it must move over to the right
            this._$el.find('.token-container.selected').removeClass('selected');
            $token.addClass('selected');
            this._$input.addClass('no-select');

            if (!_.isNull(this._$menu)) {
                this._$menu.remove();
            }

            if (newPos > this._$tokListContainer.width()) {
                this._$tokList.css('margin-left', parseFloat(this._$tokList.css('margin-left')) - (newPos - this._$tokListContainer.width()));
                this._$tokListContainer.css('max-width', this._$input.innerWidth() - (parseFloat(this._$tokListContainer.css('margin-left')) +
                    parseFloat(this._$tokListContainer.css('margin-right'))));
            } else if ($token.position().left < 0) {
                this._$tokList.css('margin-left', parseFloat(this._$tokList.css('margin-left')) - $token.position().left);
                this._$tokListContainer.css('max-width', this._$input.innerWidth() - (parseFloat(this._$tokListContainer.css('margin-left')) +
                    parseFloat(this._$tokListContainer.css('margin-right'))));

                if (this._$tokListContainer.outerWidth(true) - $token.position().left <= this._$input.outerWidth()) {
                    this._$input.css('padding-left', this._$tokListContainer.outerWidth(true) - $token.position().left);
                }
            }
        };

        tvp.buildElement = function (options, el) {

            if (_.isNull(el)) {
                switch (options.style) {
                case 'search':

                    el = $(typeAheadSearchTemplate({ locale : options.locale }));
                    el.find('.search-icon').on('click', function() {
                        el.find('input').focus();
                    });
                    break;
                default:
                    el = $(typeAheadTemplate({ locale : options.locale }));
                }

            } else if (el.jquery && el.is('input')) {
                el[0].type = 'text';
                el[0].placeholder = options.locale.placeholder || '';
            }

            if(el.parent().hasClass('orcl-picker')) {
                el.parent().addClass('orcl-type-ahead');
            } else {
                el.wrap('<div class="orcl-type-ahead"></div>');
            }

            el = el.parent();

            el.append(tokenContainerTemplate);
            el.find('.token-list-container')
                .css('max-width', el.find('input').width() - 50);

            el.find('input').removeAttr('readonly');

            if(options.search) {
                el.find('input')
                    .attr('role', 'combobox')
                    .attr('aria-autocomplete', 'list')
                    .attr('aria-expanded', 'false');
            }

            return el;
        };

        /**
         * Empties out the current text and replaces it with a new string built
         * by calling into the supplied display function
         *
         * @method TypeAheadTokenView#select
         */
        tvp.select = function () {
            this._selectToken();

            this._$input.trigger('change', {update : false});

            if (!_.isNull(this._$menu)) {
                this._$menu.remove();
            }
        };

        tvp._selectToken = function () {
            var that = this, tokenMarkup = '', displayFn = this._options.viewFns.display,
                selected = this._controller.selected(),
                symetricalPadding = this._padding;

            this._inputEl.value = '';

            _.each(selected, function (node) {
                tokenMarkup += tokenTemplate({ value : displayFn(node.data), id : node.id });
            });

            var $tokens = $(tokenMarkup).on('click', '.token-remove',function (event) {
                event.preventDefault();
                that._controller.selected(_.without(that._controller.selected(),
                    that._controller.getNodeById($(this).parent('.token-container').attr('data-id'))));
                that._$input.focus();
            }).on('click', ':not(.token-remove)', function (event) {
                    event.preventDefault();

                    if ($(this).hasClass('token')) {
                        var parent = $(this).parent('.token-container');

                        if (parent.position().left < 0) {
                            that.focusToken(parent);
                        } else if (parent.position().left + parent.outerWidth() > that._$tokListContainer.width()) {
                            that.focusToken(parent);
                        } else {
                            that._$el.find('.token-container.selected').removeClass('selected');
                            parent.addClass('selected');
                            that._$input.addClass('no-select');
                        }
                    }

                    that._$input.focus();
                });

            this._$el.find('.token-list-container')
                .css('margin', '0px ' + symetricalPadding + 'px')
                .children('.token-list')
                .empty()
                .append($tokens);

            this.focusText();
        };

        tvp._selectNontoken = function () {
            var that = this;

            this._inputEl.value = _.reduce(this._controller.selected(), function (memo, node, index) {
                if (!_.isUndefined(node.data)) {
                    return memo + (index > 0 ? that._options.delimiter + ' ' : '') + that._options.viewFns.display(node.data);
                } else {
                    return memo + (index > 0 ? that._options.delimiter + ' ' : '') + that._options.viewFns.display(node);
                }
            }, '');
        };

        /**
         * Displays the spinner when a search is happening but has not completed
         * yet.
         *
         * @method TypeAheadTokenView#loading
         */
        tvp.loading = function () {
            if (!_.isNull(this._$menu)) {
                this._$menu.remove();
            }

            //It needs a height because when you call spin - it tries to position itself in the center
            //TODO: this should not be here
            this._$menu = this._$popup =
                $('<ul class="orcl-type-ahead-menu dropdown-menu"><li style="height: 20px;"></li></ul>');

            $(this._options.appendTo).append(this._$menu);
            //Since JQuery objects are not really HTML elements we need to grab the element.
            this._$menu.find('li').append($(loadingTemplate));

            this.showPopup(this._$el);
        };

        /**
         * Marks the node as active and moves the viewport accordingly.
         *
         * @method TypeAheadTokenView#active
         * @param {Object} activeData The node that is active.
         */
        tvp.active = function (activeData) {
            var active;
            if (!_.isNull(this._$menu)) {
                this._$menu.find('li').removeClass('active');

                if (!_.isNull(activeData.node)) {
                    active = this._$menu.find('[data-id=' + activeData.node.id + ']');
                    active.addClass('active');

                    //Scroll parent container if it has one
                    if (active.position().top + active.height() >= this._$menu.height()) {
                        this._$menu.scrollTop(active[0].offsetTop + active.height() - this._$menu.height());
                    } else if (active.position().top < 0 ||
                        active.position().top >= this._$menu.height()) {
                        this._$menu.scrollTop(this._$menu.scrollTop() + active.position().top);
                    }
                }
            }

        };

        /**
         * Builds the markup for the menu.
         *
         * @method TypeAheadTokenView#buildMenu
         * @param {Array} data The data for the menu
         * @param {String} query The current query
         */
        tvp.buildMenu = function (data, query) {
            var that = this,
                markup = '',
                regex = new RegExp('(' + stringUtils.regexEscape(query) + ')', 'gi'),
                text, highlightText;

            function buildNode(node) {
                text = that._options.viewFns.display(node.data);
                highlightText = text.replace(regex, '<strong>$1</strong>');
                return nodeTemplate({
                    title : text,
                    text : highlightText,
                    id : node.id,
                    active : that._controller.active().node === node ? 'active' : ''
                });
            }

            function buildMenu() {
                return menuTemplate({
                    nodes : markup,
                    scroll : that._options.maxSuggestions < data.length ? 'scroll' : '',
                    //TODO: find a way to get 26 out of here (It is the height of an <li> in our dropdown
                    height : data.length < that._options.maxSuggestions ? 'auto' : (that._options.maxSuggestions * 26) + 'px'
                });
            }

            if (data.length > 0 && this._options.scroll) {
                //If it does support scrolling put them all in the markup
                _.each(data, function (node) {
                    markup += buildNode(node);
                });

                markup = buildMenu();
            } else if (data.length > 0 && !this._options.scroll) {
                //If no scrolling then only put their max suggestions in the markup
                var node;
                for (var i = 0; i < this._options.maxSuggestions; i++) {
                    node = data[i];
                    markup += buildNode(node);
                }

                markup = menuTemplate({
                    nodes : markup
                });
            }
            return markup;
        };

        /**
         * The method to take the markup and construct a DOM element from it.
         * Also since this is a new element we need to add the click events to the
         * menu.
         *
         * @method TypeAheadTokenView#render
         * @param {String} markup The markup representing the menu.
         */
        tvp.render = function (markup) {
            //Find proper z-indexing, append to the location
            var that = this, active;

            function positionScrollTop() {
                if (active.position().top + active.height() > that._$menu.height()) {
                    that._$menu.scrollTop(that._$menu.scrollTop() + active.position().top + active.height() - that._$menu.height());
                } else if (active.position().top < 0) {
                    that._$menu.scrollTop(that._$menu.scrollTop() + active.position().top);
                }
            }

            //Since it is debounced - need to make sure we are still active/focused
            //TODO: use a boolean instead of is?
            if(this._$input.is(document.activeElement)) {

                if (!_.isNull(this._$menu)) {
                    this._$menu.remove();
                }

                //TODO: fix this rendering - create the menu once and only switch the items?
                this._$menu = this._$popup = $(markup);

                //Once we have the markup we need to put on the events - each time
                this._$menu.on('click', 'li', function (event) {
                    event.preventDefault();
                    that._$input.focus();

                    if (!that._options.multiSelect) {
                        that._controller.selected($(this).attr('data-id'));
                    } else {
                        that._controller.selected(_.union(that._controller.selected(), $(this).attr('data-id')));
                    }
                });

                this._$menu.hover(function () {
                    that._hideMenu = false;
                }, function () {
                    that._hideMenu = true;
                });

                //Reposition popup after adding the nodes to it (giving it proper height to calculate with)
                this.showPopup(this._$el);

                active = this._$menu.find('.active');

                //Mess with scroll-top if it is scrollable
                if (this._options.scroll && active.length > 0) {
                    positionScrollTop();
                }
            }
        };

        /**
         * @method TypeAheadTokenView#loaded
         * @param {Array} data The data for the menu
         * @param {String} query The query used to get the data.
         */
        tvp.loaded = function (data, query) {
            //Build menu, position menu, show menu
            var markup = this.buildMenu(data, query);
            this.render(markup);
        };

        tvp.clear = function () {
            this._$el.find('.token-list').empty();
            this._$input.trigger('change');
        };

        tvp.init = function () {
            View.prototype.init.apply(this, arguments);

            this._padding = parseFloat(this._$input.css('padding-left'));
        };

        return TypeAheadView;
    }
);


define('oui.typeAhead/typeAheadSearchIcons.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button class="btn btn-picker"> <i class="pgbu-icon-small pgbu-icon-magnifying-glass"></i></button><a class="search-clear-icon" href="#"> <i class="pgbu-icon-small pgbu-icon-delete"></i></a>';
return __p;
};});

define('oui.typeAhead/typeAheadNonTokenView',['oui.underscore', 'oui.jquery', 'oui.base/view', 'oui.utils/stringUtils', 'oui.utils/keyCode',
    'oui.mixins/popup',
    'oui.typeAhead/typeAheadTemplate.htm',
    'oui.typeAhead/typeAheadMenuTemplate.htm',
    'oui.typeAhead/typeAheadNodeTemplate.htm',
    'oui.typeAhead/typeAheadSearchIcons.htm',
    'oui.tree/spinner.template'
],
    function (_, $, View, stringUtils, keyCode, popupMixin, typeAheadTemplate,
              menuTemplate, nodeTemplate, searchIcons, loadingTemplate) {
        

        var tvp;

        //Copied from typeaheadController
        function getQuery (value, cursorLocation, multiSelect, delimiter) {
            if (!multiSelect || (multiSelect && value.indexOf(delimiter) === -1)) {
                return value;
            } else {
                var upperBound = value.indexOf(delimiter, cursorLocation),
                    lowerBound = value.lastIndexOf(delimiter, cursorLocation - 1) + 1;
                if (upperBound === -1) { upperBound = value.length; }
                if (lowerBound === -1) { lowerBound = 0; }
                return  value.substring(lowerBound, upperBound).trim();
            }
        }

        /**
         * The type ahead view is responsible for displaying the suggestions when they loaded.
         *
         * @class TypeAheadNonTokenView
         * @private
         * @since 2.1.0
         * @extends View
         * @mixes popup
         * 
         * @param {TypeAheadModel} model The typeahead model that is responsible for holding all the past
         *  search results
         * @param {TypeAheadController} controller The type ahead controller that will modify the model.
         * @param {Object} element jQuery element for the view
         * @param {TypeAhead#Config} viewOpts The options for the view.
         */
        function TypeAheadView(model, controller, element, viewOpts, settings) {
            var that = this;

            View.call(this, viewOpts, element, controller, settings);

            /**
             * The menu object for this typeahead
             *
             * A JQuery object.
             *
             * @property _$menu
             * @private
             * @type {Object}
             */
            this._$menu = null;
            this._$input = this._$el.find('input');
            this._selectedMap = {};
            this._triggerChange = false;

            this._inputEl = this._$input[0];

            model.on({
                'view.loaded' : {fn : this.loaded, context : this },
                'view.active' : {fn : this.active, context : this },
                'view.loading' : {fn : this.loading, context : this },
                'view.select' : {fn : this.select, context : this },
                'view.selection-change' : {fn : this.select, context : this },
                'view.invalidate' : {fn : this.clear, context: this },
                'view.clear' : {fn : this.clear, context: this}
            });

            var debounced = _.debounce(controller.search, this._options.delay);

            this._$menu = this._$popup = $(menuTemplate({
                scroll : this._options.scroll,
                //TODO: find a way to get 26 out of here (It is the height of an <li> in our dropdown
                height : (that._options.maxSuggestions * 26) + 'px'
            }));

            //Once we have the markup we need to put on the events - each time
            this._$menu.on('click', 'li', function (event) {
                event.preventDefault();
                that._$input.focus();

                if (!that._options.multiSelect) {
                    that._controller.selected([$(this).attr('data-id')]);
                } else {
                    that._controller.selected(_.union(that._controller.selected(), $(this).attr('data-id')));
                }
            }).on('mousedown', function (event) {
                event.preventDefault();
            }).on('shown', function () {
                that._$input.attr('aria-expanded', true);
            }).on('hidden', function () {
                that._$input.removeAttr('aria-activedescendant').attr('aria-expanded', false);
            });

            this.element().on('oui-accept oui-picker-accept', function(event, widget, name, selected) {
                if(widget.name() === 'picker') {
                    that._controller.selected(selected, { silent : true });
                }
            });

            this._$el.find('.search-clear-icon').on('click', function(event) {
                event.preventDefault();
                that._inputEl.value = '';
                that._controller.search(that._inputEl.value);
                that._$input.focus();
            }).on('keydown', function(event) {
                if(event.keyCode === keyCode.ENTER) {
                    event.preventDefault();
                    that._inputEl.value = '';
                    that._controller.search(that._inputEl.value);
                    that._$input.focus();
                }
            });

            this._$input
                .on('keyup', function (event) {

                    switch (event.keyCode) {
                    case keyCode.DOWN:
                    case keyCode.UP:
                    case keyCode.TAB:
                    case keyCode.SHIFT:
                    case keyCode.ENTER:
                    case keyCode.F2:
                    case keyCode.F10:
                        break;
                    case keyCode.ESC:

                        if(that._options.escToClear && !that._$el.hasClass('empty')) {
                            event.stopPropagation();
                            that._$input.val('');
                            that._controller.search(that._$input.val());
                            that._$input.focus();
                        }
                        break;
                    case keyCode.BACKSPACE:
                        /* falls through */
                    default:
                        if(this.selectionStart === this.selectionEnd) {
                            var value = this.value;
                            if(_.isObject(that._options.autoEscape)) {
                                value = that._options.autoEscape.escape(value);
                            } else if (that._options.autoEscape) {
                                value = _.escape(value);
                            }
                            debounced.call(controller, value, this.selectionStart);
                        }
                    }
                })
                .on('blur', function (event) {
                    event.preventDefault();
                    that.hidePopup();
                    if(that._options.style !== 'search') {
                        that.updateValue();
                    }
                    that.change();
                })
                .on('input', function () {
                    that._$el.toggleClass('empty', that._inputEl.value === '');
                })
                .on('keypress', function (event) {
                    if(event.keyCode !== keyCode.ENTER) {
                        if(this.selectionStart !== this.selectionEnd) {
                            that.handleBackspace(true);
                        }
                        var value = this.value;
                        if(_.isObject(that._options.autoEscape)) {
                            value = that._options.autoEscape.escape(value);
                        } else if (that._options.autoEscape) {
                            value = _.escape(value);
                        }

                        //TODO: look into dealing with multi-character delimiters
                        if(String.fromCharCode(event.which) !== that._options.delimiter.trim()) {
                            var query = getQuery(value, this.selectionStart,
                                        that._options.multiSelect, that._options.delimiter.trim());

                            var node = that._controller.node(query);

                            //User is updating the currently selected query node
                            //remove it to allow for extensions like George to George Michael
                            if(node != null) {
                                controller.remove(node);
                            }
                        }
                        debounced.call(controller, value, this.selectionStart);
                    }
                })
                .on('keydown', function (event) {
                    if (that._options.search) {
                        switch (event.which) {
                        case keyCode.TAB:
                        case keyCode.LEFT:
                        case keyCode.RIGHT:
                        case keyCode.SHIFT:
                        case keyCode.F2: //Picker open in case it is on there
                            break;
                        case keyCode.DOWN:
                            event.preventDefault();
                            controller.activeNext();
                            break;
                        case keyCode.UP:
                            event.preventDefault();
                            controller.activePrevious();
                            break;
                        case keyCode.ENTER:
                            if (!_.isNull(that._$menu) && that.isVisible() && that._$menu.find('.active').length !== 0) {
                                event.preventDefault();
                                controller.selected(controller.selected().concat(that._$menu.find('.active').attr('data-id')));

                            } else if (!that.isVisible()) {
                                that.change();
                            }
                            break;
                        case keyCode.BACKSPACE:
                        case keyCode.DELETE:

                            if(that.handleBackspace()) {
                                event.preventDefault();
                            }
                            break;
                        }
                    }
                });
        }

        TypeAheadView.prototype = Object.create(View.prototype);
        tvp = TypeAheadView.prototype;

        tvp.constructor = TypeAheadView;

        _.extend(tvp, popupMixin);

        tvp.destroy = function () {
            this._$input.off();
            this._$el.remove();

            if(this._menu != null) {
                this._$menu.remove();
            }
        };

        tvp.handleBackspace = function (force) {
            var that = this;
            var input = this._$input[0];
            var cursor = input.selectionStart;
            var cursorEnd = input.selectionEnd;
            var value = this._inputEl.value;
            if(_.isObject(this._options.autoEscape)) {
                value = this._options.autoEscape.unescape(value);
            } else if (this._options.autoEscape) {
                value = _.unescape(value);
            }

            var indexFirstDelimiter = value.lastIndexOf(this._options.delimiter, cursor);
            var indexLastDelimiter = value.indexOf(this._options.delimiter, cursorEnd);

            var selected = this._controller.selected();
            if(selected.length > 0) {

                if(indexFirstDelimiter === -1) {
                    indexFirstDelimiter = 0;
                } else {
                    indexFirstDelimiter += this._options.delimiter.length;
                }

                if(indexLastDelimiter === -1) {
                    indexLastDelimiter = value.length;
                }

                //Move the selection ends to these indices
                if(indexFirstDelimiter !== cursor) {
                    input.selectionStart = indexFirstDelimiter;
                }

                if(indexLastDelimiter !== cursorEnd) {
                    input.selectionEnd = indexLastDelimiter;
                }

                if(force || (indexFirstDelimiter === cursor && indexLastDelimiter === cursorEnd)) {
                    var display = value.substring(indexFirstDelimiter, indexLastDelimiter);

                    var filtered = _.filter(selected, function(item) {
                        return display !== that._options.viewFns.display(item.data);
                    });

                    this._controller.selected(filtered);

                }

                return true;
            }

            return false;
        };

        tvp.buildElement = function (options, el) {
            if (_.isNull(el)) {
                el = $(typeAheadTemplate({ locale : options.locale }));
            } else if (el.jquery && el.is('input')) {
                el[0].type = 'text';
                el[0].placeholder = options.locale.placeholder || '';
            }

            if(el.parent().hasClass('orcl-picker')) {
                el.parent().addClass('orcl-type-ahead');
            } else {
                el.wrap('<div class="orcl-type-ahead"></div>');
            }

            el = el.parent();

            el.find('input')
                .removeAttr('readonly');

            if(options.style === 'search') {
                el.addClass('search');
                el.append(searchIcons);
                el.toggleClass('empty', el.find('input')[0].value === '');
                el.find('.search-icon').on('click', function() {
                    el.find('input').focus();
                });

                el.find('input').addClass('search-query');
            }

            if(options.search) {
                el.find('input')
                    .attr('role', 'combobox')
                    .attr('aria-autocomplete', 'list')
                    .attr('aria-expanded', 'false');
            }

            return el;
        };

        tvp.init = function () {
            var input = this._$el.children('input');
            this._$el.find('.instructions[data-name=' + this._settings.name + ']').attr('id', this.id() + '-instructions-' + this._settings.name);
            input.attr('aria-describedby', ((input.attr('aria-describedby') || '')  + ' ' + this.id() + '-instructions-' + this._settings.name).trim());
        };

        /**
         * Empties out the current text and replaces it with a new string built
         * by calling into the supplied display function
         *
         * @method TypeAheadNonTokenView#select
         */
        tvp.select = function () {
            this.updateValue();

            this._triggerChange = true;
            this.hidePopup();
        };

        tvp.updateValue = function () {
            var that = this;

            var value = _.reduce(this._controller.selected(), function (memo, node, index) {
                var data = node.data || node;
                return memo + (index > 0 ? that._options.delimiter : '') + that._options.viewFns.display(data);
            }, '');

            if(_.isObject(this._options.autoEscape)) {
                value = this._options.autoEscape.unescape(value);
            } else if (this._options.autoEscape) {
                value = _.unescape(value);
            }
            this._inputEl.value = value;
            this._inputEl.scrollLeft = this._inputEl.scrollWidth;
        };

        /**
         * Displays the spinner when a search is happening but has not completed
         * yet.
         *
         * @method TypeAheadNonTokenView#loading
         */
        tvp.loading = function () {
            if(this._$input.is(document.activeElement)) {
                //Since JQuery objects are not really HTML elements we need to grab the element.
                this._$menu.empty().append($(loadingTemplate));

                this.showPopup(this._$el);
            }
        };

        /**
         * Marks the node as active and moves the viewport accordingly.
         *
         * @method TypeAheadNonTokenView#active
         * @param {Object} activeData The node that is active.
         */
        tvp.active = function (activeData) {
            var active;
            this._$menu.find('.active').removeClass('active');
            this._$input.removeAttr('aria-activedescendant');

            if (!_.isNull(activeData.node)) {
                active = this._$menu.find('[data-id=' + activeData.node.id + ']');
                active.addClass('active');
                this._$input.attr('aria-activedescendant', active.attr('id'));

                //Scroll parent container if it has one
                if (active.position().top + active.height() >= this._$popup.height()) {
                    this._$menu.scrollTop(active[0].offsetTop + active.height() - this._$menu.height());
                } else if (active.position().top < 0 ||
                    active.position().top >= this._$popup.height()) {
                    this._$menu.scrollTop(this._$menu.scrollTop() + active.position().top);
                }
            }

        };

        tvp.disable = function (disabled) {
            this._$el.find('input')
                .toggleClass('disabled', disabled);

            var el = this._$input[0];

            if(disabled) {
                el.setAttribute('disabled', 'disabled');
            } else {
                el.removeAttribute('disabled');
            }
        };

        /**
         * Builds the markup for the menu.
         *
         * @method TypeAheadNonTokenView#buildMenu
         * @param {Array} data The data for the menu
         * @param {String} query The current query
         */
        tvp.buildMenu = function (data, query) {
            var that = this,
                regex = new RegExp('(' + stringUtils.regexEscape(query) + ')', 'gi'),
                text, highlightText;

            function buildNode(node) {
                text = that._options.viewFns.display(node.data);
                that._selectedMap[text] = node;
                highlightText = text.replace(regex, '<strong>$1</strong>');
                return nodeTemplate({
                    title : text,
                    text : highlightText,
                    pageId : that.id() + '-' + node.id,
                    id : node.id,
                    active : that._controller.active().node === node ? 'active' : ''
                });
            }

            if (data.length > 0 && !this._options.scroll) {
                //If no scrolling then only put their max suggestions in the markup
                data = _.take(data, this._options.maxSuggestions);
            }

            return  _.reduce(data, function (memo, node) {
                return memo + buildNode(node);
            }, '');
        };

        /**
         * The method to take the markup and construct a DOM element from it.
         * Also since this is a new element we need to add the click events to the
         * menu.
         *
         * @method TypeAheadNonTokenView#render
         * @param {String} markup The markup representing the menu.
         */
        tvp.render = function (markup) {
            //Find proper z-indexing, append to the location
            var that = this, active;

            function positionScrollTop() {
                if (active.position().top + active.height() > that._$menu.height()) {
                    that._$menu.scrollTop(that._$menu.scrollTop() + active.position().top + active.height() - that._$menu.height());
                } else if (active.position().top < 0) {
                    that._$menu.scrollTop(that._$menu.scrollTop() + active.position().top);
                }
            }

            //Since it is debounced - need to make sure we are still active/focused
            //TODO: use a boolean instead of is?
            if(this._$input.is(document.activeElement) && markup !== '') {
                this._$menu.empty().append(markup);

                //Reposition popup after adding the nodes to it (giving it proper height to calculate with)
                this.showPopup(this._$el);

                active = this._$menu.find('.active');

                //Mess with scroll-top if it is scrollable
                if (this._options.scroll && active.length > 0) {
                    positionScrollTop();
                }

            } else {
                this.hidePopup();
            }
        };

        /**
         * @method TypeAheadNonTokenView#loaded
         * @param {Array} data The data for the menu
         * @param {String} query The query used to get the data.
         */
        tvp.loaded = function (data, query) {
            //Build menu, position menu, show menu
            this.render(this.buildMenu(data, query));
        };

        tvp.clear = function () {
            this._$el.addClass('empty');
            this._inputEl.value = '';
        };

        tvp.change = function () {
            if(this._triggerChange) {
                this._$input.trigger('change', { update : false });
                this._triggerChange = false;
            }
        };

        return TypeAheadView;
    }
);

define('oui.typeAhead/typeAhead',['oui.underscore', 'oui.base/input', 'oui.typeAhead/typeAheadModel', 'oui.typeAhead/typeAheadController',
    'oui.typeAhead/typeAheadTokenView', 'oui.typeAhead/typeAheadNonTokenView',
    'oui.logger'],
    function (_, Input, TypeAheadModel, TypeAheadController, TypeAheadTokenView, TypeAheadNonTokenView, logger) {
        

        /**
         * @typedef {Object} TypeAhead#Config
         *
         * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the
         * input. When the property is an Object it should have two functions escape and unescape that take in the string and
         * provide the escaped/unescaped version of the string. When the property is true it will default to using _.escape
         * and _.unescape for the strings.
         *
         * @property {Boolean} [scroll=true] Whether the menu should be scrollable.
         *
         * @property {Number} [maxSuggestions=10] How many suggestions should be placed into the suggestion menu.
         *
         * @property {Number} [delay=250] The number, in ms, to wait before the input is used to query for suggestions.
         *
         * @property {Boolean} [escToClear=false] Whether or not the typeAhead should clear the query on ESC.
         *
         * @property {String} [delimiter=', '] Character or string to use as the delimiter between querys/inputs.
         *
         * @property {String|jQuery} [appendTo='body'] The selector or jQuery object to append the menu to. Useful when the
         * default results in a z-index issue.
         *
         * @property {Boolean} [multiSelect=false] Whether or not the typeahead component can hold multiple selections.
         *
         * @property {Boolean} [tokenize=true] Whether to use tokens in the typeahead component. Tokenize is only read when
         * the component is in multiselect mode.
         *
         * @property {Boolean} [search=true] Whether the component should perform the search itself. If false, an event is
         * fired and another component, such as a tree, should use the query and perform its own search.
         *
         * @property {Boolean} [allowEmpty=false] Whether to search when the query is the empty string.
         *
         * @property {String} [style='default'] Whether to use a standard input or a 'search' style with rounded corners and
         * a magnifying glass. Valid values are 'default' and 'search'.
         *
         * @property {Boolean} [fluid=false] Whether the input should take up all available space it can.
         *
         * @property {Number} [minLength=0] The minimum string length before a search is issued.
         */

        var tap;

        var __WIDGET_NAME__ = 'typeahead',
            __DEFAULTS__ = {
                autoEscape : true,
                scroll : true,
                maxSuggestions : 10,
                delay : 250, //ms,
                escToClear : false,
                delimiter : ', ',
                appendTo : 'body',
                multiSelect : false,
                // We set this to true, since it will be most likely that users will want tokenized if they do multiselect
                // Having tokenize:true, does not affect multiSelect:false
                tokenize : true,
                search : true,
                allowEmpty : false,
                style : 'default',
                fluid : false,
                minLength : 0
            };

        //TODO: Update the picker options docs when these change.
        /**
         * The typeahead (aka autocomplete) component that recommends matches based on the
         * input to the component.
         *
         * It is navigable with the arrow keys and items can be selected with the enter key.
         *
         * @class TypeAhead
         * @since 1.3.0
         * @extends Input
         * 
         * @param {TypeAhead#Config} [options] The options to override defaults
         * @param {jQuery} [element] The jQuery element to use for typeAhead
         * @param {Object} [initData] The initial data passed to the typeAhead
         */
        //TODO: default valid structure for initData
        function TypeAhead(options, element, initData) {
            var model;

            //Super call
            Input.call(this, options, element, initData);

            model = new TypeAheadModel(this._options);
            /**
             * The controller that the component keeps a reference to.
             *
             * @property _controller
             * @private
             * @type {TypeAheadController}
             */
            this._controller = new TypeAheadController(model, this._options);
            //Setup proper view for the type ahead. Is it tokens or plain text.
            /**
             * The view element that the component keeps a reference to.
             *
             * @property _view
             * @private
             * @type {TypeAheadView}
             */
            if (this._options.tokenize && this._options.multiSelect) {
                this._view = new TypeAheadTokenView(model, this._controller, this._element, this._options, this._settings);
            } else {
                this._view = new TypeAheadNonTokenView(model, this._controller, this._element, this._options, this._settings);
            }

            Object.defineProperty(this, '_data', {
                get : this.selected,
                set : this.selected
            });

            this._setupEvents(model);
        }

        TypeAhead.prototype = Object.create(Input.prototype);
        tap = TypeAhead.prototype;

        tap.constructor = TypeAhead;

        /**
         * Returns the name of the widget that is used in eventing and id generation
         *
         * @method TypeAhead#name
         * @return - the widgets name
         */
        tap.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * @method TypeAhead#_translate
         * @private
         */
        tap._translate = function (name) {
            if( name === 'select' ) {
                var data = arguments[1];

                return [ name, _.pluck(data, 'data') ];
            }
            return arguments;
        };

        /**
         * Please use combined [selected](#method_selected).
         *
         * @method TypeAhead#getSelected
         * @deprecated
         * @return {Array} The selected objects in the typeahead.
         */
        tap.getSelected = function () {
            logger.warn('The use of getSelected (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            return this.selected();
        };

        /**
         * Combined getter/setter for the typeahead component.
         *
         * @method TypeAhead#selected
         * @fires TypeAheadModel#selection-error
         * @fires TypeAheadModel#select
         * @param {Array} id The nodes to select in the typeahead component.
         * @return {Array} The selected objects in the typeahead.
         */
        tap.selected = function (id) {
            if (!arguments.length) {
                //Public get should not return the whole node. Just the part they care about.
                return _.pluck(this._controller.selected(), 'data');
            }
            this._controller.selected(id);
        };

        /**
         * Combined getter/setter for the selected nodes of the typeahead component.
         *
         * **Note**: This function is silent. For more information regarding its counterpart see {@link TypeAhead#selected}
         *
         * @method TypeAhead#selected
         * @param {Array} selected The nodes to select in the typeahead component.
         * @return {Array} The selected objects in the typeahead.
         */
        tap.value = function (selected) {
            if(!arguments.length) {
                return this.selected();
            }

            this._controller.selected(selected, { silent : true });
        };

        /**
         * Ability to retrieve the widgets default options for use in merging with passed in options as
         * well as with composed widgets such as the date and time picker.
         *
         * @method TypeAhead#defaults
         * @return - the widgets default options
         */
        tap.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Empties the typeahead's model and empties the view.
         *
         * @method TypeAhead#invalidate
         */
        tap.invalidate = function () {
            this._controller.invalidate();
        };

        return TypeAhead;
    }
);

define('oui.picker/picker',['oui.underscore', 'oui.base/widget', 'oui.picker/pickerView', 'oui.tree/treePanel', 'oui.typeAhead/typeAhead', 'oui.logger'],
    function (_, Widget, PickerView, TreePanel, TypeAhead, logger) {
        

        /**
         * @callback Picker#display
         * @param {Array} nodes The nodes to generate the string representation for.
         * @return {String} - The display value
         */

        /**
         * @typedef Picker#Config
         * @property {Object} viewFns
         * @property {Picker#display} viewFns.display
         * @property {Boolean|jQuery|String} [modal=false] Whether or not the picker should be displayed in a modal popup or
         * not. When passed a jQuery Object or a string it is used as the modal for the picker. There are two requirements to
         * the DOM object when using a custom modal: 1. The element that is going to hold the tree panel has class
         * 'tree-content' 2. The element that is going to hold the type ahead has class 'search'. This property assumes the
         * condition is true when passed in a jQuery object or a string.
         * @property {Object} [locale={}] The specific strings that are used by the widget.
         * @property {Boolean} [required=false] Whether or not the field is required. If the field is not required then a
         * clear button will show up in the modal and allow a user to deselect all of their previous selection.
         * @property {Boolean} [buttonTriggerOnly=false] Whether or not the widget is only a button instead of an input
         * element with a button.
         * @property {String|jQuery} [appendTo='body'] Where to append the modal.
         * @property {Boolean} [fluid=false] Whether the input element should take up the widget of its container.
         * @property {Boolean} [search=false] Whether or not to show the search in the picker overridden if the Tree configuration contains a filter function.
         * @property {TreePanel#Config} [tree] The tree options.
         * @property {TypeAhead#Config} [typeahead] The typeahead options.
         */

        var __WIDGET_NAME__ = 'picker',
            __OVERRIDES__ = {
                tree : {
                    picker : true
                },
                typeAhead : {
                    fluid : true,
                    escToClear : true,
                    style : 'search',
                    search : false
                }
            },
            __DEFAULTS__ = _.extend({
                viewFns : {
                    display : function (nodes) {
                        return _.map(nodes, function (node) {
                            return node.name || node.value;
                        }).join(', ');
                    }
                },
                modal : false,
                locale : {},
                required : false,
                buttonTriggerOnly : false,
                appendTo : 'body',
                fluid : false,
                tree : TreePanel.prototype.defaults(),
                search : false,
                typeAhead : TypeAhead.prototype.defaults()
            }, __OVERRIDES__),
            pp;

        /**
         * The element that represents a Picker. It is a container around the tree and typeahead
         * widgets.
         *
         * For more information on used classes
         *
         * * {@link TreePanel#Config}
         * * {@link TypeAhead#Config}
         *
         * @class Picker
         * @extends Widget
         * @since 1.3.0
         * 
         * @param {Picker#Config} options The options for the picker.
         * @param {Object} [element] The element to use for the picker. It will create an element if none is
         *    given.
         * @param {Array} [initData] The data to use before any loading takes place.
         */
        function Picker(options, element, initData) {
            //Super call
            Widget.call(this, options, element, initData);

            this._options.search = this._options.search ||
                (this._options.tree && this._options.tree.storeFns && _.isFunction(this._options.tree.storeFns.filter)) ||
                (this._options.tree && this._options.tree.lazyLoad === false && this._options.tree.clientFilter);

            /**
             * The tree panel component that makes up a picker.
             *
             * @property _treePanel
             * @type {TreePanel}
             * @private
             */
            this._treePanel = new TreePanel(this._options.tree, this._initData);

            /**
             * The type ahead component that combined with a tree makes up the entire
             * picker.
             *
             * @property _typeAhead
             * @type {TypeAhead}
             * @private
             */
            this._typeAhead = new TypeAhead(this._options.typeAhead);

            /**
             * The view for the picker.
             *
             * @property _view
             * @type {PickerView}
             * @private
             */
            this._view = new PickerView(this._options, this._element, this._treePanel, this._typeAhead, this._settings);

            Object.defineProperty(this, '_data', {
                get : this.selected,
                set : this.selected
            });

            this._setupEvents([this._treePanel, this._view]);
        }

        Picker.prototype = Object.create(Widget.prototype);
        pp = Picker.prototype;

        pp.constructor = Picker;

        /**
         * Gets the widget name. Used to prefix some events.
         *
         * @method Picker#name
         * @return {String} 'picker'
         */
        pp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Initialize the picker by initializing its subcomponents.
         *
         * @method Picker#init
         * @private
         */
        pp._init = function () {

            if(!(this._options.modal || _.isObject(this._options.modal))) {
                this._treePanel.init();
                this._typeAhead.init();
            }

            this._treePanel.removeOff();
            this._typeAhead.removeOff();
        };

        /**
         * Getter/Setter for the selected nodes in the picker.
         *
         * If no parameter is given then the method behaves like a getter
         * otherwise it will overwrite the currently selected nodes with
         * the given nodes.
         *
         * @example
         *
         *     var picker = new oui.Picker(); //Assuming valid options
         *
         *     //Initialization
         *
         *     //Finds the first nodes that match the fields given.
         *     picker.selected([
         *         { name : "Test name" },
         *         { id : "Test id" }
         *     ]);
         *
         *     picker.selected(); //Returns [
         *         {
         *             name : "Test name"
         *             //Any additional data that is in the node
         *         },
         *         {
         *             id : "Test id"
         *             //Any additional data that is in the node
         *         }
         *     ]
         *
         *     picker.selected([]); //Removes all selected nodes
         *
         * @method Picker#selected
         * @fires TreePanel#picker-accept
         * @param {Array} data The data to search for to make selected
         *   in the tree component.
         * @return {Array} The selected nodes of the picker.
         */
        pp.selected = function (data) {
            if (!arguments.length) {
                return this._treePanel.selected();
            }

            this._treePanel.selected(data);
        };

        /**
         * Getter/Setter for the selected nodes in the picker.
         *
         * Similar to {{#crossLink "Picker/selected"}}{{/crossLink}} except it does
         * not trigger an event after selection.
         *
         * @method Picker#value
         * @param {Array} data The selected data of the tree component.
         * @return {Array} The selected nodes of the picker.
         */
        pp.value = function (data) {
            if (!arguments.length) {
                return this._treePanel.selected();
            }

            this._treePanel.value(data);
        };

        /**
         * Getter/Setter for the unselectable nodes in the picker.
         *
         * If no parameter is given then the method behaves like a getter
         * otherwise it will overwrite the currently unselectable nodes with
         * the given nodes.
         *
         * @example
         *
         *     var picker = new oui.Picker(); //Assuming valid options
         *
         *     //Initialization
         *
         *     //Finds the first nodes that match the fields given.
         *     picker.unselectable([
         *         { name : "Test name" },
         *         { id : "Test id" }
         *     ]);
         *
         *     picker.unselectable(); //Returns [
         *         {
         *             name : "Test name"
         *             //Any additional data that is in the node
         *         },
         *         {
         *             id : "Test id"
         *             //Any additional data that is in the node
         *         }
         *     ]
         *
         *     picker.unselectable([]); //Removes all unselecable nodes
         *
         * @method Picker#unselectable
         * @param {Array} data The data to search for to make unselectable
         *   in the tree component.
         * @return {Array} The unselectable nodes of the picker.
         */
        pp.unselectable = function (data) {
            if (!arguments.length) {
                return this._treePanel.unselectable();
            }

            this._treePanel.unselectable(data);
        };

        /**
         * Getter/Setter for the expanded nodes in the picker.
         *
         * If no parameter is given then the method behaves like a getter
         * otherwise it will overwrite the currently expanded nodes with
         * the given nodes.
         *
         * @example
         *
         *     var picker = new oui.Picker(); //Assuming valid options
         *
         *     //Initialization
         *
         *     //Finds the first nodes that match the fields given.
         *     picker.expanded([
         *         { name : "Test name" },
         *         { id : "Test id" }
         *     ]);
         *
         *     picker.expanded(); //Returns [
         *         {
         *             name : "Test name"
         *             //Any additional data that is in the node
         *         },
         *         {
         *             id : "Test id"
         *             //Any additional data that is in the node
         *         }
         *     ]
         *
         *     picker.expanded([]); //Removes all expanded nodes from the
         *                          //list of expanded nodes
         *
         * @method Picker#expanded
         * @param {Array} data The data to search for to make expanded
         *   in the tree component.
         * @return {Array} The expanded nodes of the picker.
         */
        pp.expanded = function (data) {
            if (!arguments.length) {
                return this._treePanel.expanded();
            }

            this._treePanel.expanded(data);
        };

        /**
         * Getter/Setter for the disabled nodes in the picker.
         *
         * If no parameter is given then the method behaves like a getter
         * otherwise it will overwrite the currently disabled nodes with
         * the given nodes.
         *
         * @example
         *
         *     var picker = new oui.Picker(); //Assuming valid options
         *
         *     //Initialization
         *
         *     //Finds the first nodes that match the fields given.
         *     picker.disabled([
         *         { name : "Test name" },
         *         { id : "Test id" }
         *     ]);
         *
         *     picker.disabled(); //Returns [
         *         {
         *             name : "Test name"
         *             //Any additional data that is in the node
         *         },
         *         {
         *             id : "Test id"
         *             //Any additional data that is in the node
         *         }
         *     ]
         *
         *     picker.disabled([]); //Removes all disabled nodes from the
         *                          //list of disabled nodes
         *
         * @method Picker#disabled
         * @param {Array} data The data to search for to make disabled
         *   in the tree component.
         * @return {Array} The disabled nodes of the picker.
         */
        pp.disabled = function (data) {
            if (!arguments.length) {
                return this._treePanel.disabled();
            }

            this._treePanel.disabled(data);
        };

        /**
         * Please use [selected](#method_selected).
         *
         * @method Picker#getSelected
         * @deprecated
         */
        pp.getSelected = function () {
            logger.warn('The use of getSelected (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            return this._treePanel.getSelected();
        };

        /**
         * Please use [selected](#method_selected).
         *
         * @method Picker#select
         * @deprecated
         */
        pp.select = function (data) {
            logger.warn('The use of select (and get/set methods in general) in OUI is now deprecated. Please use the combined method "selected" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.us.oracle.com');
            this._treePanel.setSelection(data);
        };

        /**
         * Provides custom logic for the picker widget when destroy is called.
         *
         * @method Picker#_destroy
         * @private
         */
        pp._destroy = function () {
            this._treePanel.destroy();
            this._typeAhead.destroy();
        };

        /**
         * @method Picker#defaults
         * @return {Object} The defaults for the picker.
         */
        pp.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Returns the JSON representation of the Tree component.
         *
         * @method Picker#tree
         * @return - The JSON representation of the tree.
         */
        pp.tree = function () {
            return this._treePanel.tree();
        };

        /**
         * Locks and disables the view as well as remove any data from the model.
         *
         * @method Picker#lock
         * @param {Boolean} locked Whether to render the view as locked or not
         */
        pp.lock = function (locked) {
            if (this._options.modal && !this._options.buttonTriggerOnly) {
                this._view.lock(locked);
                this._view.disable(locked);
                //TODO: remove all data from tree and typeahead.
            }
        };

        /**
         * Gets the tree panel used by the picker.
         *
         * @method Picker#treePanel
         * @return {TreePanel} treePanel The tree panel used by the picker.
         */
        pp.treePanel = function() {
            return this._treePanel;
        };

        /**
         * Get the type ahead used by the picker.
         *
         * @method Picker#typeAhead
         * @return {TypeAhead} ta The typeAhead used by the picker.
         */
        pp.typeAhead = function() {
            return this._typeAhead;
        };

        /**
         * Remove all data from the model (expanded, selected, disabled,...) and reloads
         * from the root node.
         *
         * @method Picker#invalidate
         */
        pp.invalidate = function () {
            this._treePanel.invalidate();
            this._typeAhead.invalidate();
        };

        /**
         * If the picker has the modal option set it will display the modal.
         *
         * @method Picker#show
         */
        pp.show = function () {
            if(this._options.modal) {
                this._view.show();
            }
        };

        /**
         * If the picker has the modal option set it will hide the modal.
         *
         * @method Picker#hide
         */
        pp.hide = function () {
            if(this._options.modal) {
                this._view.hide();
            }
        };

        return Picker;
    }
);

define('oui.url/urlModel',['oui.utils/subscribable'], function (Subscribable) {
    

    var ump;

    /**
     * @event UrlModel#url-change
     */
    /**
     * @event UrlModel#state-change
     */
    /**
     * @event UrlModel#validation-state-change
     */

    /**
     * Data associated with the UrlField component.
     *
     * @class UrlModel
     * @private
     * @since 1.3.0
     * @extends Subscribable
     *
     * @param {String} initData Initial URL for the widget
     */
    function UrlModel(initData) {
        // Super call
        Subscribable.call(this);

        /**
         * The URL stored in the text field
         *
         * @property _url
         * @type {String}
         * @private
         */
        this._url = initData;
        /**
         * The state of the URL validation
         *
         * @property _state
         * @type {String}
         * @private
         */
        this._state = 'none';
    }

    UrlModel.prototype = Object.create(Subscribable.prototype);
    ump = UrlModel.prototype;

    ump.constructor = UrlModel;

    /**
     * Gets or sets the URL stored in the widget.
     *
     * @method UrlModel#url
     * @protected
     * @param {String} [url] If provided, the new URL to store in the model
     * @returns {String|undefined}
     */
    ump.url = function (url) {
        if (!arguments.length) {
            return this._url;
        }

        if(this._url !== url) {
            this._url = url;
            this.trigger('url-change', this._url);
        }
    };

    /**
     * Gets or sets the current validation state of the widget.
     *
     * States are: 'none', 'valid', 'invalid', 'testing'
     *
     * @method UrlModel#state
     * @protected
     * @param {String} [state] If provided, sets the state to the specified value
     * @returns {String|undefined}
     */
    ump.state = function (state) {
        if (!arguments.length) {
            return this._state;
        }

        if(this._state !== state) {
            this._state = state;
            this.trigger('state-change', this._state);
            this.trigger('validation-state-changed', this._state);
        }
    };

    return UrlModel;
});

define('oui.url/urlController',[], function () {
    

    var ucp;

    /**
     * @class UrlController
     * @private
     * @since 1.3.0
     * 
     * @param {UrlModel} model The UrlModel to manipulate the data.
     * @param {Object} options The specific options regarding the controller.
     */
    function UrlController(model, options) {

        this._model = model;
        this._options = options;
    }

    ucp = UrlController.prototype;

    ucp.constructor = UrlController;

    /**
     * Gets or sets the URL to be stored in the widget.
     *
     * @method UrlController#url
     * @protected
     * @param {String} [url] The url to set the widget to
     * @returns {String|undefined}
     */
    ucp.url = function (url) {
        if (!arguments.length) {
            return this._model.url();
        }
        this._model.url(url);
    };

    /**
     * Tests the URL stored in the widget for validity and availability.
     *
     * @method UrlController#testUrl
     * @protected
     */
    ucp.testUrl = function () {
        var that = this;
        this._model.state('testing');

        this._options.testFn(this._model.url(), function (isValid) {
            if (isValid) {
                that._model.state('valid');
            } else {
                that._model.state('invalid');
            }
        });
    };

    return UrlController;
});


define('oui.url/urlTemplate.htm',[],function () { return '<div class="orcl-url-field">\n    <div class="input-container">\n        <input type="url" />\n    </div>\n</div>\n';});


define('oui.url/wrapTemplate.htm',[],function () { return '<div class="orcl-url-field">\n    <div class="input-container">\n    </div>\n</div>\n';});


define('oui.url/controlsTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="url-controls"> <button type="button" aria-label="'+
((__t=( data.test ))==null?'':__t)+
'" title="'+
((__t=( data.test ))==null?'':__t)+
'" class="btn btn-blue-light url-test-button">'+
((__t=( data.test ))==null?'':__t)+
'</button></div>';
return __p;
};});


define('oui.template/loadingTemplate.htm',[],function () { return '<i class="pgbu-icon pgbu-icon-loading"></i>\n';});

define('oui.url/urlView',['oui.underscore', 'oui.jquery', 'oui.base/view',
    'oui.url/urlTemplate.htm', 'oui.url/wrapTemplate.htm', 'oui.url/controlsTemplate.htm',
    'oui.template/loadingTemplate.htm',
    'oui.utils/detectMobile'],
    function (_, $, View, urlTemplate, wrapTemplate, controlsTemplate, loadingTemplate) {
        

        var uvp;

        /**
         * The type ahead view is responsible for displaying the suggestions when they loaded.
         *
         * @class UrlView
         * @private
         * @since 1.3.0
         * @extends View
         * 
         * @param {UrlModel} model The UrlModel that is responsible for this view.
         * @param {UrlController} controller The UrlController that will modify the model.
         * @param {Object} element The default HTML input element that users can interact with.
         * @param {Object} options View specific options.
         */
        function UrlView(model, controller, element, options, settings) {
            var that = this;
            // Super call
            View.call(this, options, element, controller, settings);

            //TODO: this should not be necessary in the view - should use controller
            /**
             * The model for this view to use.
             *
             * @property _model
             * @private
             * @type {UrlModel}
             */
            this._model = model;

            model.on({
                'view.url-change' : { fn : this.updateValue, context : this },
                'view.state-change' : { fn : this.updateState, context : this }
            });

            this._inputEl = this._$el.find('input')[0];

            this.updateValue(model.url());

            this._$el.find('input')
                .on('change', function () {
                    var value = this.value;
                    if(_.isObject(that._options.autoEscape)) {
                        value = that._options.autoEscape.escape(value);
                    } else if (that._options.autoEscape) {
                        value = _.escape(value);
                    }
                    that._controller.url(value);
                })
                .on('input', function () {
                    that.toggleTest(this.value);
                })
                .on('keydown', function () {
                    that.updateState('none');
                });

            this._$el.find('.url-test-button').on('click', $.proxy(this._controller.testUrl, this._controller));
        }

        UrlView.prototype = Object.create(View.prototype);
        uvp = UrlView.prototype;

        uvp.constructor = UrlView;

        uvp.destroy = function (cleanUp) {
            this._$el.find('input').off('change keydown');

            if (cleanUp.self) {
                this._$el.remove();
            }
        };

        uvp.buildElement = function (opts, el) {
            var $element;

            if (_.isNull(el)) {
                //Build element from scratch
                $element = $(urlTemplate);
            } else if (el.jquery) {
                //We were given a jquery element
                el.wrap(wrapTemplate);
                $element = el.closest('.orcl-url-field');
            }
            $element.find('.input-container').append(loadingTemplate);

            $element.append($(controlsTemplate(opts.locale)));

            return $element;
        };

        /**
         * Disables the view.
         *
         * @method UrlView#disable
         * @protected
         * @param {Boolean} disabled True to disable, false to enable
         */
        uvp.disable = function (disabled) {
            var els = this._$el.find('input, .url-test-button')
                .toggleClass('disabled', disabled);

            _.each(els, function (elem) {
                if(disabled) {
                    elem.setAttribute('disabled', 'disabled');
                } else {
                    elem.removeAttribute('disabled');
                }
            });
        };

        /**
         * Updates the URL on the user visible text box.
         *
         * @method UrlView#updateValue
         * @protected
         * @param {String} value The new URL
         */
        uvp.updateValue = function (value) {
            if(_.isObject(this._options.autoEscape)) {
                value = this._options.autoEscape.unescape(value);
            } else if (this._options.autoEscape) {
                value = _.unescape(value);
            }

            this._inputEl.value = value;
            this.updateState('none');
            this.toggleTest(value);
        };

        uvp.toggleTest = function (value) {
            var disabled = _.isNull(value) || value === '';

            this._$el.find('.url-test-button')
                .prop('disabled', disabled)
                .toggleClass('disabled', disabled);
        };

        /**
         * Updates the state of the user visible information.
         *
         * @method UrlView#updateState
         * @protected
         * @param {String} state The new state
         */
        uvp.updateState = function (state) {
            var $controlGroup = this._$el.add(this._$el.closest('.control-group')).removeClass('error success loading');

            switch (state) {
            case 'valid':
                $controlGroup.addClass('success');
                break;
            case 'invalid':
                $controlGroup.addClass('error');
                break;
            case 'testing':
                this._$el.addClass('loading');
                break;
            }
        };

        uvp.init = function () {

        };

        return UrlView;
    }
);

define('oui.url/url',['oui.base/input', 'oui.url/urlModel', 'oui.url/urlController', 'oui.url/urlView'],
    function (Input, UrlModel, UrlController, UrlView) {
        

        /**
         *
         * @example
         *
         *     function (url, isValid) {
         *         $.get(url)
         *             .done(function () { isValid(true);  })
         *             .fail(function () { isValid(false); });
         *     }
         *
         * @callback UrlField#testFn
         * @param {String} url The url to validate.
         * @param {Function} isValid The function to use to validate the url. The callback function called with the result of
         * the validation.
         */

        /**
         * @typedef {Object} UrlField#Config
         *
         * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the
         * input. When the property is an Object it should have two functions escape and unescape that take in the string and
         * provide the escaped/unescaped version of the string. When the property is true it will default to using _.escape
         * and _.unescape for the strings.
         *
         * @property {Object} [locale={}] The localized values for strings used in the UrlField widget.
         *
         * @property {UrlField#testFn} testFn The function used to validate a URL.
         *
         * @property {Boolean} [browserValidation=false] Remove browser regular expression checking or URL syntax. **Note**:
         * This option is currently not supported.
         */

        //Used to hold the UrlField.prototype to allow
        //for easier additions of functions.
        var ufp;

        var __DEFAULTS__ = {
            autoEscape : true,
            locale : {},
            browserValidation : false
        }, __WIDGET_NAME__ = 'url';

        /**
         * The UrlField widget that allows the entry of URLs into a text box.
         * The user can then click a test button to make sure that the entered URL is valid and working.
         *
         * @param {UrlField#Config} options Configurable options for the UrlField widget
         * @param {HTMLInputElement|jQuery} element The default `<input type="text">` element
         *                 to wrap into a UrlField widget. If one is not provided, it will be created.
         * @param {String} initData The initial URL to display.
         *
         * @class UrlField
         * @since 1.3.0
         * @extends Input
         *
         */
        function UrlField(options, element, initData) {
            var model;

            //Super call
            Input.call(this, options, element, initData);

            model = new UrlModel(this._initData, this._options);
            this._controller = new UrlController(model, this._options);
            this._view = new UrlView(model, this._controller, this._element, this._options, this._settings);

            this._setupEvents(model);
        }

        //Start with the base Input as the prototype for the url field.
        UrlField.prototype = Object.create(Input.prototype);
        ufp = UrlField.prototype;

        ufp.constructor = UrlField;

        /**
         * Returns the name of the widget that is used in eventing and id generation
         *
         * @method UrlField#name
         * @return {String} The widget name.
         */
        ufp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Gets or sets the currently entered URL from the text box.
         *
         * @method UrlField#url
         * @fires UrlModel#url-change
         * @param {String} [url] The new URL to set the widget to
         * @returns {String|undefined}
         */
        ufp.url = function (url) {
            if (!arguments.length) {
                return this._controller.url();
            }
            this._controller.url(url);
        };

        /**
         * Ability to retrieve the widgets default options for use in merging with passed in options as
         * well as with composed widgets such as the date and time picker.
         *
         * @method UrlField#defaults
         * @return - the widgets default options
         */
        ufp.defaults = function () {
            return __DEFAULTS__;
        };

        return UrlField;

    }
);

define('oui.icon/iconModel',['oui.underscore', 'oui.utils/subscribable'], function (_, Subscribable) {
    

    /**
     * @event IconModel#icon-changed
     * @param {String} selectedIcon The newly selected color of the color picker.
     */
    /**
     * @event IconModel#color-changed
     * @param {String} selectedIcon The newly selected color of the color picker.
     */
    /**
     * @event IconModel#deselect
     */


    /**
     * This class represents the model for the icon picker. It stores the
     * available icons as well as the selected.
     *
     * @class IconModel
     * @private
     * 
     * @param {IconPicker#Config} options The options for the icon picker.
     * @param {String} initData The icon to start the picker off at.
     */
    function IconModel(options, initData) {

        Subscribable.call(this);

        /**
         * The currently selected icon.
         *
         * @property _selectedIcon
         * @private
         * @type {String}
         */
        this._selectedIcon = initData;
        /**
         * The available icons in the model.
         *
         * @property _avilableIcons
         * @private
         * @type {Array}
         */
        this._availableIcons = options.map;
    }

    IconModel.prototype = Object.create(Subscribable.prototype);

    IconModel.prototype.constructor = IconModel;

    /**
     * The getter/setter for the model.
     *
     * @method IconModel#icon
     * @param {String} [icon] The icon to set.
     * @return {String} The selected icon.
     */
    IconModel.prototype.icon = function (icon) {
        if (!arguments.length) {
            return this._selectedIcon;
        }

        if(_.isNull(icon)) {
            this._selectedIcon = icon;
            this.trigger('deselect');
        } else if (icon !== this._selectedIcon) {
            this._selectedIcon = icon;
            this.trigger('icon-changed', this._selectedIcon);
        }

    };

    /**
     * Getter for the available icons in the model.
     *
     * @method IconModel#availableIcons
     * @return {Array} The icons available to choose from.
     */
    IconModel.prototype.availableIcons = function () {
        if (!arguments.length) {
            return this._availableIcons;
        }
    };

    return IconModel;
});


define('oui.icon/iconButtonTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button type="button" class="input orcl-icon-picker" tabindex="0" role="combobox" aria-haspopup="true" aria-expanded="false"> <div class="icon-box"> <i class="thumbnail pgbu-icon-small"> <span class="spoken-text"></span> </i> </div> <div class="btn btn-picker" title="'+
((__t=( data.locale.icon.select ))==null?'':__t)+
'"> <i class="caret"></i> </div></button>';
return __p;
};});


define('oui.icon/paletteTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div role="listbox" id="'+
((__t=( data.id ))==null?'':__t)+
'" class="palette"> '+
((__t=( data.markup  ))==null?'':__t)+
'</div>';
return __p;
};});


define('oui.icon/rowTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="row '+
((__t=( data.cls ? data.cls : ''  ))==null?'':__t)+
'"> '+
((__t=( data.row  ))==null?'':__t)+
'</div>';
return __p;
};});


define('oui.icon/clearButton.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button type="button" class="btn btn-blue-light btn-small clear">'+
((__t=( data.locale.clear  ))==null?'':__t)+
'</button>';
return __p;
};});


define('oui.icon/iconTileTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div role="option" id="'+
((__t=( data.id ))==null?'':__t)+
'" class="tile" title="'+
((__t=( data.locale[data.icon] ))==null?'':__t)+
'" data-icon="'+
((__t=( data.icon  ))==null?'':__t)+
'"> <i class="icon-preview pgbu-icon-small '+
((__t=( data.cls  ))==null?'':__t)+
'"></i></div>';
return __p;
};});

define('oui.icon/iconView',[
    'oui.underscore', 'oui.jquery', 'oui.base/view',
    'oui.icon/iconButtonTemplate.htm',
    'oui.icon/paletteTemplate.htm',
    'oui.icon/rowTemplate.htm',
    'oui.icon/clearButton.htm',
    'oui.icon/iconTileTemplate.htm',
    'oui.mixins/popup', 'oui.logger'
], function (_, $, View, buttonTemplate, paletteTemplate, rowTemplate, clearButton, tileTemplate, popupMixin, logger) {
        

        var cvp;

        /**
         * @class IconView
         * @private
         * @extends View
         * @mixes popup
         *
         * @param {IconModel} model
         * @param {IconController} controller
         * @param {jQuery|null} element
         * @param {IconPicker#Config} options
         * @param {Object} settings
         */
        function IconView(model, controller, element, options, settings) {
            var that = this;

            //TODO: these functions get created for all icon views - move to a static function with a .call(this)?
            function iconToClass (key) {
                if(that._options.viewFns && that._options.viewFns.iconClass) {
                    return that._options.viewFns.iconClass(key);
                } else {
                    return key;
                }
            }

            function buildRow(colors, rowNum) {
                return rowTemplate({ cls : '', row : _.reduce(colors, function(memo, color, index) {
                    return memo + that.tileTemplate({ id : that.id() + '-' + rowNum + '-' + index, icon : color, cls : iconToClass(color), locale : that._options.locale });
                }, '')});
            }

            function buildPalette() {
                var rows = _.groupBy(options.map, function(item, index) {
                    return Math.floor((index / options.columns)) % options.rows;
                });
                var markup = _.reduce(rows, function(memo, row, index) {
                    return memo + buildRow(row, index);
                }, '');

                if(options.clearable) {
                    markup += rowTemplate({ cls : 'clear-row', row : clearButton({ locale : options.locale }) });
                }

                return $(paletteTemplate({ id : 'oui-' + settings.name + '-' + settings.id + '-palette', markup : markup }));
            }

            //Super call
            View.call(this, options, element, controller, settings);

            this._$el.find('.thumbnail').attr('id', this.id() + '-thumbnail');
            this._$el.attr('aria-describedby', this.id() + '-thumbnail');

            /**
             * The JQuery object that represents the pop-up palette
             *
             * @property _palette
             * @private
             * @type {Object}
             */
            this._palette = this._$popup = buildPalette();

            //Setting up the events that the view will listen to on the model
            model.on({
                'view.icon-changed' : { fn : this.update, context : this },
                'view.deselect' : { fn : this.deselect, context : this }
            });

            this.update(model.icon());

            //Binding events to the elements
            this._$el.on('mousedown', function (e) {
                if(!$(this).hasClass('disabled') && !$(this).hasClass('native')) {
                    e.stopPropagation(); // prevent it from hiding itself (see this._hide)
                    if (that.isVisible()) {
                        that._hide();
                    } else {
                        that._show();
                    }
                }
            })
            .on('blur', function() {
                that._hide();
            })
            .on('focus', function() {
                if(that._options.openOnFocus) {
                    that._show();
                }
            });

            this.addShortcuts({
                'ESC' : {
                    fn : this._hide,
                    stopPropagation : this.isVisible.bind(this)
                },
                'LEFT' : this._switchActive.bind(this, -1, 0),
                'RIGHT' : this._switchActive.bind(this, 1, 0),
                'UP' : this._switchActive.bind(this, 0, -1),
                'DOWN' : function() {
                    if(!that.isVisible()) {
                        that._show();
                    }
                    that._switchActive(0, 1);
                },
                'SPACE ENTER' : function() {
                    if(!this.isVisible()) {
                        that._show();
                    }
                },
                'ENTER' : {
                    up : true,
                    fn : function() {
                        var $activeTile = that._palette.find('.active');
                        if ($activeTile.length) {
                            that._controller.icon($activeTile.attr('data-icon'));
                            that._hide();
                        }
                    }
                },
            });

            this._palette
                .on('mousedown', function(event) {
                    event.preventDefault();
                })
                .on('click', '.tile', function () {
                    that._hide();
                    that._controller.icon($(this).attr('data-icon'));
                })
                .on('click', '.clear', function () {
                    that._hide();
                    that._controller.icon(null);
                });
        }

        IconView.prototype = Object.create(View.prototype);
        cvp = IconView.prototype;

        cvp.constructor = IconView;

        _.extend(cvp, popupMixin);

        cvp.tileTemplate = tileTemplate;

        cvp.destroy = function (cleanUp) {
            this._palette.off();
            this._palette.remove();

            this._$el.off();
            if (cleanUp.self) {
                this._$el.remove();
            }
        };

        cvp.update = function (color) {
            if (this._options.thumbnail) {
                this._$el.find('.thumbnail')
                    .attr('class', 'thumbnail pgbu-icon-small ' + this._options.viewFns.iconClass(color))
                    .attr('title', this._options.locale[color]);

                this._$el.find('.spoken-text:not(.instructions)').text(this._options.locale[color]);
            }
        };

        cvp.buildElement = function (options, el) {
            this._existingElement = false; // Whether we should toggle active on the element when clicked

            if (_.isNull(el)) {
                //create our own element
                el = $(buttonTemplate({ locale : options.locale }));
            } else if (el.jquery) {
                //Got a jquery element
                if (el.is('button')) {
                    el.addClass('orcl-icon-picker');
                    this._existingElement = true;
                }
            }

            return el;
        };

        cvp.deselect = function () {
            if (this._options.thumbnail) {
                this._$el.find('.thumbnail').addClass('empty');
            }
        };

        cvp.disable = function (disabled) {
            function disable (els) {
                _.each(els, function (el) {
                    if(disabled) {
                        el.setAttribute('disabled', 'disabled');
                        el.disabled = true;
                    } else {
                        el.removeAttribute('disabled');
                        el.disabled = false;
                    }
                });
            }

            if (!arguments.length) {
                //Error here?
                logger.warn('Disable does not have a getter');
            }

            var toDisable = this._$el.add(this.$('input, button, .btn'));

            toDisable.toggleClass('disabled', disabled);
            disable(toDisable);

            if (disabled) {
                this._$el.removeAttr('tabindex');
            } else {
                this._$el.attr('tabindex', 0);
            }
        };

        cvp._show = function () {
            var that = this;

            if(!this.isVisible()) {

                if ($(this._options.appendTo).children().filter(this._palette).length === 0) {
                    //Not on the DOM - append the menu;
                    $(this._options.appendTo).append(this._palette);
                    this._$el.attr('aria-owns', this._palette.attr('id'));
                }

                //TODO: maybe make this a configuration option?
                this.showPopup(this._$el, 'right');
                this._$el.focus();

                this._$el.attr('aria-expanded', true);

                if (that._existingElement) {
                    that._$el.addClass('active');
                }
            }
        };

        cvp._hide = function () {
            this.hidePopup();
            this._$el.attr('aria-expanded', false);
            this._$el.removeClass('active');
            this._palette.find('.tile').removeClass('active');
            this._$el.removeAttr('aria-activedescendant');
        };

        cvp._switchActive = function (dx, dy) {
            var $tiles = this._palette.find('.tile'),
                $current = $tiles.filter('.active');

            $current.removeClass('active');

            if($current.length === 0 && (dy > 0 || dx > 0)) {
                $tiles.first().addClass('active');
            } else if($current.length === 0 && (dy < 0 || dx < 0)) {
                $tiles.last().addClass('active');
            } else if ($current.length > 0) {
                var currentIndex = $current.index(),
                    rows = this._palette.find('.row'),
                    currentRowIndex = rows.index($current.closest('.row')),
                    tiles = $current.siblings('.tile').add($current);

                var nextRowIndex = ((currentRowIndex + dy) % rows.length + rows.length) % rows.length;
                var nextColumnIndex = ((currentIndex + dx) % tiles.length + tiles.length) % tiles.length;

                rows.eq(nextRowIndex).find('.tile').eq(nextColumnIndex).addClass('active');
            }

            this._$el.attr('aria-activedescendant', this._palette.find('.active').attr('id'));

        };

        cvp.id = function (settings) {
            if(!arguments.length) {
                return this._$el.attr('id');
            }


            if(_.isUndefined(this._$el.attr('id'))) {
                if(_.isObject(settings)) {
                    this._$el.attr('id', 'oui-' + settings.name + '-' + settings.id);
                } else if (_.isString(settings)) {
                    this._$el.attr('id', settings);
                }
            }
        };

        return IconView;
    }
);


define('oui.color/buttonTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<button type="button" class="input orcl-color-picker" tabindex="0" role="combobox" aria-haspopup="true" aria-expanded="false"> <div class="color-box"> <i class="thumbnail"> <span class="spoken-text"></span> </i> </div> <div class="btn btn-picker" title="'+
((__t=( data.locale.color.select  ))==null?'':__t)+
'"> <i class="pgbu-icon-small pgbu-icon-paint-bucket"></i> </div></button>';
return __p;
};});


define('oui.color/tileTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div role="option" id="'+
((__t=( data.id ))==null?'':__t)+
'" class="tile" title="'+
((__t=( data.locale.color[data.cls.toUpperCase()] ))==null?'':__t)+
'" data-icon="#'+
((__t=( data.cls  ))==null?'':__t)+
'" > <div class="color-preview" style="background-color:#'+
((__t=( data.cls  ))==null?'':__t)+
'"></div></div>';
return __p;
};});


define('oui.color/nativeElement.htm',[],function () { return '<div class="orcl-color-picker native">\n    <input type="color" />\n</div>\n';});

define('oui.color/colorView',[
    'oui.underscore',
    'oui.jquery',
    'oui.icon/iconView',
    'oui.color/buttonTemplate.htm',
    'oui.color/tileTemplate.htm',
    'oui.color/nativeElement.htm',
    'oui.logger'
], function(_, $, View, buttonTemplate, tileTemplate, nativeElement, logger) {
    

    var cvp;

    /**
     * @class ColorView
     * @private
     * @extends IconView
     *
     * Does not use the parameters. Forwards them with an apply call.
     *
     * @param {IconModel} model
     * @param {IconController} controller
     * @param {jQuery|null} element
     * @param {ColorPicker#Config} options
     * @param {Object} settings
     */
    function ColorView() {
        var that = this;

        View.apply(this, arguments);

        if(this._$el.hasClass('native')) {
            this._$el.on('change', 'input', function () {
                that._controller.color(this.value);
            });
        }
    }

    ColorView.prototype = Object.create(View.prototype);
    cvp = ColorView.prototype;

    /**
     * @property ColorView#tileTemplate
     */
    cvp.tileTemplate = tileTemplate;

    /**
     * @method ColorView#nativeType
     * @return {String} 'color' The native type to create
     */
    cvp.nativeType = function () {
        return 'color';
    };

    /**
     * @method ColorView#buildElement
     */
    cvp.buildElement = function (options, el) {
        this._existingElement = false; // Whether we should toggle active on the element when clicked

        if (_.isNull(el)) {
            //create our own element
            el = $(buttonTemplate( { locale : options.locale }));
        } else if (el.jquery) {
            //Got a jquery element
            if (el.is('button')) {
                el.addClass('orcl-color-picker');
                this._existingElement = true;
            }
        }

        return el;
    };

    /**
     * @method ColorView#buildNativeElement
     */
    cvp.buildNativeElement = function (options, el) {
        if(_.isNull(el)) {
            el = $(nativeElement);
        } else if (el.jquery) {
            //Got a jquery element
            if (el.is('button')) {
                el.addClass('orcl-color-picker');
                this._existingElement = true;
            } else {
                //Wrap with orcl-color-picker div and make sure type of input is 'color'
                logger.info('Given an element which is not a button.');
            }
        }

        return el;
    };

    /**
     * @method ColorView#update
     */
    cvp.update = function(color) {
        if (this._options.thumbnail) {
            var title;
            if(color == null) {
                title = this._options.locale.color.empty;
            } else {
                title = this._options.locale.color[color.substring(1).toUpperCase()];
            }

            this._$el.find('.thumbnail')
                .removeClass('empty')
                .css('background-color', color)
                .attr('title', title);

            this._$el.find('.spoken-text:not(.instructions)').text(title);

        }

        var input = this.$('input');
        if (input.length > 0) {
            input[0].value = color;
        }
    };

    return ColorView;
});

define('oui.icon/iconController',[],function () {
    

    var ccp;

    /**
     * This class represents the controller for the icon picker.
     *
     * @class IconController
     * @private
     * 
     * @param {IconModel} model The model that the controller will call into
     *    to get/set data.
     * @param {IconPicker#Config} options The options object used for the widget.
     */
    function IconController(model, options) {

        /**
         * The model used in the controller.
         *
         * @property _model
         * @private
         * @type {iconModel}
         */
        this._model = model;
        /**
         * The options for the widget.
         *
         * @property _options
         * @private
         * @type {Object}
         */
        this._options = options;
    }

    ccp = IconController.prototype;
    ccp.constructor = IconController;

    /**
     * Getter/setter for the icon of the model.
     *
     * @method IconController#icon
     * @param {String} [icon] The icon to set on the model.
     * @return {String} The selected icon of the model.
     */
    ccp.icon = function (icon) {
        if (!arguments.length) {
            return this._model.icon();
        }

        this._model.icon(icon);
    };

    return IconController;
});

define('oui.color/colorPicker',['oui.underscore', 'oui.base/widget', 'oui.icon/iconModel', 'oui.color/colorView', 'oui.icon/iconController', 'oui.logger'],
    function (_, Widget, Model, View, Controller, logger) {
        

        /**
         * @typedef {Object} ColorPicker#Config
         *
         * @property {Boolean} [openOnFocus=false] Whether or not the color picker should open the popup when focused.
         * @property {Array<String>} [map] Map of colors (hex values) to use for the palette.
         * @property {Boolean} [fluid=false] Whether or not the component will take all available horizontal space of the
         * parent.
         * @property {String|jQuery} [appendTo='body'] The selector to append the palette to.
         * @property {Boolean} [thumbnail=true] Whether or not to display a preview/thumbnail of the selected color.
         * @property {Number} [rows=5] Specify how many rows the color picker should have. The color picker will not generate
         * empty rows.
         * @property {Number} [columns=8] Specify how many columns the color picker should have. The color picker uses this
         * property as the primary objected to match.
         * @property {Object} [locale={}] Map of strings to use for the color picker for localization.
         * @property {Boolean} [clearable=false] Whether the color picker can accept a null value. When true it renders the
         * palette with a clear button to allow a user to remove their selection.
         */

        var __DEFAULTS__ = {
                openOnFocus : false,
                map : [
                    '000000', '993300', '333300', '003300', '003366', '000080', '333399', '333333', //Row 1
                    '800000', 'ff6600', '808000', '008000', '008080', '0000ff', '666699', '808080', //Row 2
                    'ff0000', 'ff9900', '99cc00', '339966', '33cccc', '3366ff', '800080', '969696', //Row 3
                    'ff00ff', 'ffcc00', 'ffff00', '00ff00', '00ffff', '00ccff', '993366', 'c0c0c0', //Row 4
                    'ff99cc', 'ffcc99', 'ffff99', 'ccffcc', 'ccffff', '99ccff', 'cc99ff', 'ffffff'  //Row 5
                ],
                fluid : false,
                appendTo : 'body',
                thumbnail : true,
                rows : 5,
                columns : 8,
                locale : {},
                clearable : false
            };

        /**
         * This class creates a customizeable swatch of colors.
         *
         * See {@link ColorPicker#Config} for details regarding the options for the Color Picker
         *
         * @example
         * <caption>Creating a Color Picker (accepts all defaults)</caption>
         *
         * var color = new oui.ColorPicker();
         * //Assumes parent container element is called $container
         * $container.append(color.element());
         * color.init();
         *
         * @example
         * <caption>Creating a Color Picker with existing element</caption>
         *
         * //Assumes element being used is already a jQuery object called $colorEl
         * var color = new oui.ColorPicker($colorEl);
         * color.init();
         *
         * @class ColorPicker
         * @extends Widget
         * @param {ColorPicker#Config} [options] The options object to use throughout the widget.
         * @param {HTMLInputElement|jQuery} [element] JQuery element to use as color picker on the DOM.
         * @param {String} [initColor='#FFFFFF'] The color to initialize the color picker with.
         */
        function ColorPicker(options, element, initColor) {
            var model;

            //Super call
            Widget.call(this, options, element, initColor);

            //Set up default _initData for the ColorPicker
            this._initData = this._initData || '#FFFFFF';

            model = new Model(this._options, this._initData);
            /**
             * The controller object for the widget.
             *
             * @property _controller
             * @private
             * @type {ColorController}
             */
            this._controller = new Controller(model, this._options);
            /**
             * The view for the widget.
             *
             * @property _view
             * @private
             * @type {ColorView}
             */
            this._view = new View(model, this._controller, this._element, this._options, this._settings);

            Object.defineProperty(this, '_data', {
                get : this.color,
                set : this.color
            });

            this._setupEvents(model);
        }

        ColorPicker.prototype = Object.create(Widget.prototype);
        var cpp = ColorPicker.prototype;

        cpp.constructor = ColorPicker;

        /**
         * Returns the name of the widget that is used in eventing and id generation
         *
         * @method ColorPicker#name
         * @return {String} 'color' - the widgets name
         */
        cpp.name = function () {
            return 'color';
        };

        /**
         * Translate events to proper public facing events.
         *
         *
         * @method ColorPicker#_translate
         * @private
         * @param {String} name The name of the event to translate
         * @param {...*} args The event parameters to translate
         * @return {Array|arguments} The translated event parameters.
         */
        cpp._translate = function (name) {
            var args = _.toArray(arguments);

            if(name === 'icon-changed') {
                return _.union(['color-changed'], _.rest(args, 1));
            }

            return arguments;
        };

        /**
         * @method ColorPicker#setColor
         * @deprecated
         * @param {String} rgbValue The color to select.
         */
        cpp.setColor = function (rgbValue) {
            logger.warn('The use of setColor (and get/set methods in general) in OUI is now deprecated. Please used the combined method "color" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.use.oracle.com');

            this._controller.icon(rgbValue);
        };

        /**
         * @method ColorPicker#getColor
         * @deprecated
         * @return {String} The selected color.
         */
        cpp.getColor = function () {
            logger.warn('The use of getColor (and get/set methods in general) in OUI is now deprecated. Please used the combined method "color" instead.');
            logger.warn('For more information see YUIDoc on pgbulwf.use.oracle.com');

            return this._controller.icon();
        };

        /**
         * Getter and setter for the selected color of the picker.
         *
         * @example
         *     var icon;
         *
         *     //Initialized
         *
         *     //Setter
         *     icon.color('000000');
         *
         *     //Getter
         *     icon.color(); //Returns '000000'
         *
         * @method ColorPicker#color
         * @fires IconModel#color-changed
         * @fires IconModel#deselect
         * @param {String} [color] The hex value to set.
         * @return {String} The selected color.
         */
        cpp.color = function (color) {
            if (!arguments.length) {
                return this._controller.icon();
            }

            this._controller.icon(color);
        };

        /**
         * Programmaticaly open the color picker select popup.
         *
         * @method ColorPicker#show
         */
        cpp.show = function () {
            this._view._show();
        };

        /**
         * Programmatically close the color picker select popup.
         *
         * @method ColorPicker#hide
         */
        cpp.hide = function () {
            this._view._hide();
        };

        /**
         * Ability to retrieve the widgets default options for use in merging with passed in options as
         * well as with composed widgets such as the date and time picker.
         *
         * @method ColorPicker#defaults
         * @return - the widgets default options
         */
        cpp.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Getter or setter for the disable state of the color picker
         *
         * @method ColorPicker#disable
         * @param {Boolean} [disabled] Whether to disable the view or now. True to disable and false to enable
         * @return {Boolean} - Whether the view is currently disabled
         */
        cpp.disable = function (disabled) {
            if (!arguments.length) {
                return this._view.disable();
            }

            this._view.disable(disabled);
        };

        return ColorPicker;
    }
);

define('oui.icon/iconPicker',['oui.base/widget', 'oui.icon/iconModel', 'oui.icon/iconView', 'oui.icon/iconController'],
    function (Widget, Model, View, Controller) {
        

        /**
         * Function to get the CSS class representation of the icon.
         *
         * @callback IconPicker#iconClass
         * @param {String} key The key to return the CSS class for
         * @return {String} - The css to use for the icon.
         */

        /**
         * @typedef {Object} IconPicker#Config
         *
         * @property {Boolean} [openOnFocus=false] Whether or not the color picker should open the popup when focused.
         * @property {Array<String>} [map] Map of colors (hex values) to use for the palette.
         * @property {Boolean} [fluid=false] Whether or not the component will take all available horizontal space of the
         * parent.
         * @property {String|jQuery} [appendTo='body'] The selector to append the palette to.
         * @property {Boolean} [thumbnail=true] Whether or not to display a preview/thumbnail of the selected color.
         * @property {Number} [rows=5] Specify how many rows the color picker should have. The color picker will not generate
         * empty rows.
         * @property {Number} [columns=8] Specify how many columns the color picker should have. The color picker uses this
         * property as the primary objected to match.
         * @property {Object} [locale={}] Map of strings to use for the color picker for localization.
         * @property {Boolean} [clearable=false] Whether the color picker can accept a null value. When true it renders the
         * palette with a clear button to allow a user to remove their selection.
         * @property {Object} viewFns
         * @property {IconPicker#iconClass} viewFns.iconClass
         */

        var __WIDGET_NAME__ = 'icon',
            __DEFAULTS__ = {
                openOnFocus : false,
                map : [],
                fluid : false,
                appendTo : 'body',
                thumbnail : true,
                locale : {},
                clearable : false
            },
            ipp;

        /**
         * This class creates a palette of icons to choose from.
         *
         * There is no default map option so all the icons must be given.
         *
         * See {@link IconPicker#Config} for more information about options allowed by Icon Picker.
         *
         * @class IconPicker
         * @extends Widget
         *
         * @param {IconPicker#Config} [options] The options object to use throughout the widget.
         * @param {Object} [element] JQuery element to use as picker on the DOM.
         * @param {String} [initIcon] The icon to initialize the picker with.
         */
        function IconPicker(options, element, initIcon) {
            var model;

            //Super call
            Widget.call(this, options, element, initIcon);

            model = new Model(this._options, this._initData);
            /**
             * The controller object for the widget.
             *
             * @property _controller
             * @private
             * @type {IconController}
             */
            this._controller = new Controller(model, this._options);
            /**
             * The view for the widget.
             *
             * @property _view
             * @private
             * @type {IconView}
             */
            this._view = new View(model, this._controller, this._element, this._options, this._settings);

            Object.defineProperty(this, '_data', {
                get : this.icon,
                set : this.icon
            });

            this._setupEvents(model);
        }

        IconPicker.prototype = Object.create(Widget.prototype);
        ipp = IconPicker.prototype;

        ipp.constructor = IconPicker;

        /**
         * Returns the name of the widget that is used in eventing and id generation
         *
         * @method IconPicker#name
         * @return - the widgets name
         */
        ipp.name = function () {
            return __WIDGET_NAME__;
        };

        /**
         * Ability to retrieve the widgets default options for use in merging with passed in options as
         * well as with composed widgets such as the date and time picker.
         *
         * @method IconPicker#defaults
         * @return - the widgets default options
         */
        ipp.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Combined getter/setter for the disabled state of the icon picker.
         *
         * If an argument is supplied it will attempt to set the disabled status of the component.
         * If there is no argument the current status will be returned.
         *
         * @method IconPicker#disable
         * @param {Boolean} [disabled] whether or not the component should be disabled
         * @return - disabled state of the icon picker.
         */
        ipp.disable = function (disabled) {
            if (!arguments.length) {
                return this._view.disable();
            }

            this._view.disable(disabled);
        };

        /**
         * Programmaticaly open the color picker select popup.
         *
         * @method IconPicker#show
         */
        ipp.show = function () {
            this._view._show();
        };

        /**
         * Programmatically close the color picker select popup.
         *
         * @method IconPicker#hide
         */
        ipp.hide = function () {
            this._view._hide();
        };

        /**
         * Getter and setter for the selected icon of the picker.
         *
         * @example
         *     var icon;
         *
         *     //Initialized
         *
         *     //Setter
         *     icon.icon(key);
         *
         *     //Getter
         *     icon.icon(); //Returns key
         *
         * @method IconPicker#color
         * @fires IconModel#icon-changed
         * @fires IconModel#deselect
         * @param {String} [newIcon] The hex value to set.
         * @return {String} The selected color.
         */
        ipp.icon = function (newIcon) {
            if (!arguments.length) {
                return this._controller.icon();
            }

            this._controller.icon(newIcon);
        };

        return IconPicker;
    }
);

define('oui.combo/comboBoxModel',['oui.underscore', 'oui.utils/subscribable', 'oui.utils/uniqueId'], function (_, Subscribable, uniqueId) {
    

    /**
     * @event ComboBoxModel#values-change
     * @param {Array} values The new values the combobox is now using to build the menu.
     */
    /**
     * @event ComboBoxModel#active-change
     * @param {Number} id The new id of the node to be marked as active
     */
    /**
     * @event ComboBoxModel#selection-change
     * @param {Number} id The new id of the node that is selected
     */
    /**
     * @event ComboBoxModel#invalid-selection
     */
    /**
     * @event ComboBoxModel#required-error
     */
    /**
     * @event ComboBoxModel#valueSelected
     * @deprecated
     * @param {Number} id The new id of the node that is selected
     */

    var cmp;

    function createValues(vals) {
        return _.map(vals, function (value) {
            if (value.children) {
                return { _id : uniqueId('combobox-id'), _data : value, _children : createValues(value.children) };
            } else {
                return { _id : uniqueId('combobox-id'), _data : value};
            }
        });
    }

    function flatten(map, values) {
        map = map || {};

        _.each(values, function (val) {
            map[val._id] = val;

            flatten(map, val._children);
        });

        return map;
    }

    /**
     * Data associated with the ComboBox component.
     *
     * @class ComboBoxModel
     * @private
     * @since 1.3.0
     * @extends Subscribable
     *
     * @param {String|null} initData Initial data for the widget
     * @param {ComboBox#Config} options Initial options passed to widget
     */
    function ComboBoxModel(initData, options) {
        // Super call
        Subscribable.call(this);

        var values = createValues(initData);
        var store;
        var selectedNode;

        if(options.emptyValue) {
            values = _.union([options.emptyValue], values);
        }

        if(!options.allowEmpty) {
            selectedNode = _.find(store, function(node) {
                return !node._children;
            });
        }

        store = flatten({}, values);

        /**
         * The acceptable values for the widget
         *
         * @property _values
         * @type {Array}
         * @private
         */
        this._values = values;
        this._customValues = [];
        /**
         * The initial widget options
         *
         * @property _options
         * @type {Object}
         * @private
         */
        this._options = options;
        /**
         * The selected value for the widget
         *
         * @property _selected
         * @type {Object}
         * @private
         */
        this._selected = [];
        this._active = null;
        this._store = store;
        this._invalidData = undefined;

        if(!this._options.multiSelect) {
            var firstNode = _.find(store, function(node) {
                return !node._children;
            });

            if(firstNode != null) {
                this.selected([firstNode._id]);
            }
        }

    }

    ComboBoxModel.prototype = Object.create(Subscribable.prototype);
    cmp = ComboBoxModel.prototype;

    cmp.constructor = ComboBoxModel;

    /**
     * Gets or sets the selected value in the widget.
     *
     * @method ComboBoxModel#selected
     * @protected
     * @param {Object|String} [selected] The new selected value for the widget
     * @returns {Object|undefined}
     */
    cmp.selected = function (selected, eventParams) {
        var that = this;
        if (!arguments.length) {
            return this._selected;
        }

        var valueIds = _.map(this._values, function(node) {
            return node._id;
        });
        if(this._options.multiSelect) {
            if(!_.isArray(selected)) {
                selected = [ selected ];
            }
            selected = _.chain(selected)
            .map(function(node) {
                var convert = that.node(node);

                if(convert) {
                    return convert._id;
                } else {
                    return null;
                }
            })
            .sortBy(function(node) {
                return _.indexOf(valueIds, node);
            })
            .value();
        } else {
            var convert = that.node(_.isArray(selected) ? _.first(selected) : selected);

            if(convert) {
                selected = [convert._id];
            } else {
                selected = [ null ];
            }
        }


        if(_.contains(selected, null)) {
            if(!this._options.ignoreInvalid && !that._options.keyboardInput) {
                this._invalidData = null;
                this.trigger(eventParams, 'invalid-selection');
            }
        } else if (this._options.required && (selected.length === 0 ||
                    (selected.length === 1 && (this._options.storeFns.value(this.node(selected[0])._data) === null ||
                                                this._options.storeFns.value(this.node(selected[0])._data) === '')))) {
            this._selected = selected;
            this._invalidData = selected;
            this.trigger(eventParams, 'required-error');
        }else if(selected.length > 0 && !_.isEqual(selected, this._selected)) {
            this._selected = selected;
            this._invalidData = undefined;

            this.trigger(eventParams, 'selection-change', this._selected);
            this.trigger(eventParams, 'valueSelected', this._selected);
        }
    };

    /**
     * @method ComboBoxModel#active
     */
    cmp.active = function (active) {
        if (!arguments.length) {
            return this._active;
        }

        this._active = active || null;
        this.trigger('active-change', this._active);
    };

    /**
     * Gets or sets the accepted values for the widget.
     *
     * @method ComboBoxModel#values
     * @protected
     * @param {Array|String} [values] The new acceptable values for the widget
     * @returns {Array|undefined}
     */
    cmp.values = function (values) {
        if (!arguments.length) {
            return this._values;
        }

        if(this._options.emptyValue) {
            values = _.union([ this._options.emptyValue ], values);
        }

        this._values = createValues(values);
        this._store = flatten({}, this._values);
        var selectedNode;

        if(!this._options.allowEmpty) {
            selectedNode = _.find(this._store, function(node) {
                return !node._children;
            });
        }

        this.trigger('values-change', this._values);
        if(!this._options.allowEmpty) {
            if(_.isUndefined(selectedNode)) {
                this._selected = [];
                this.trigger('view.selection-change', this._selected);
            } else if (this._selected !== selectedNode._id) {
                this._selected = [selectedNode._id];
                this.trigger('view.selection-change', this._selected);
            }
        }
    };

    /**
     * @method ComboBoxModel#node
     */
    cmp.node = function (id) {
        if (_.isNull(id)) {
            return null;
        } else if (_.isString(id)) {
            return this.getNodeByDisplay(id);
        } else if (_.isNumber(id)) {
            return this.getNodeById(id);
        } else {
            return this.getNodeByData(id);
        }
    };

    /**
     * @method ComboBoxModel#getNodeByDisplay
     */
    cmp.getNodeByDisplay = function (display) {
        var that = this,
            node;

        node = _.find(this._store, function (val) {
            return _.isEqual(display, that._options.viewFns.display(val._data));
        });

        if (_.isUndefined(node) && this._options.allowCustomValues) {
            node = _.find(this._customValues, function (val) {
                return _.isEqual(display, that._options.viewFns.display(val._data));
            });
        }

        return node || null;
    };

    /**
     * @method ComboBoxModel#getNodeById
     */
    cmp.getNodeById = function (id) {
        var node;

        node = this._store[id];

        if (_.isUndefined(node) && this._options.allowCustomValues) {
            //Search through custom values;
            node = _.find(this._customValues, function (custom) {
                return custom._id === id;
            });
        }

        return node || null;
    };

    /**
     * @method ComboBoxModel#getNodeByData
     */
    cmp.getNodeByData = function (data) {
        var node;

        node = _.find(this._store, function (val) {
            return _.isEqual(data, _.pick(val._data, _.keys(data)));
        });

        if (_.isUndefined(node) && this._options.allowCustomValues) {
            node = _.find(this._customValues, function (val) {
                return _.isEqual(data, _.pick(val._data, _.keys(data)));
            });
        }

        return node || null;
    };

    /**
     * @method ComboBoxModel#isValid
     */
    cmp.isValid = function () {
        return this._invalidData === undefined;
    };

    return ComboBoxModel;
});

define('oui.combo/comboBoxController',['oui.jquery', 'oui.underscore'], function ($, _) {
    

    var ccp;

    /**
     * @class ComboBoxController
     * @private
     * @since 1.3.0
     * 
     * @param {ComboBoxModel} model The ComboBoxModel to manipulate the data.
     * @param {ComboBox#Config} options The specific options regarding the controller.
     */
    function ComboBoxController(model, options) {

        this._model = model;
        this._options = options;
    }

    ccp = ComboBoxController.prototype;

    ccp.constructor = ComboBoxController;

    ccp.node = function (data) {
        return this._model.node(data);
    };

    /**
     * Gets or sets the selected value in the widget.
     *
     * @method ComboBoxController#selected
     * @protected
     * @param {Object|String} [selected] The new selected value for the widget
     * @returns {Object|undefined}
     */
    ccp.selected = function (selected, eventParams) {
        if (!arguments.length) {
            return this._model.selected();
        }
        this._model.selected(selected, eventParams);
    };

    ccp.active = function (active) {
        if (!arguments.length) {
            return this._model.active();
        }
        this._model.active(active);
    };

    /**
     * Gets or sets the accepted values for the widget.
     *
     * @method ComboBoxController#values
     * @protected
     * @param {Array|String} [values] The new acceptable values for the widget
     * @returns {Array|undefined}
     */
    ccp.values = function (values) {
        if (!arguments.length) {
            return this._model.values();
        }
        this._model.values(values);
    };

    ccp.convert = function ($el) {
        var optGroups = $el.children('optgroup'), vals = [];

        function buildChildren(group) {
            return _.map($(group).children('option'), function (child) {
                return { name : $(child).text(), value : child.value };
            });
        }

        if (optGroups.length > 0) {
            //We have headers
            vals = _.map(optGroups, function (group) {
                return {
                    name : $(group).attr('label'),
                    children : buildChildren(group)
                };
            });

            vals = _.union(vals, buildChildren($el));
        } else {
            vals = buildChildren($el);
        }

        return vals;
    };

    ccp.isValid = function () {
        return this._model.isValid();
    };

    return ComboBoxController;
});


define('oui.combo/comboBoxListTemplate.htm',[],function () { return '<div class="orcl-combo-box combo-box-list">\n    <button type="button" role="combobox" aria-expanded="false" class="btn btn-form combo-button combo-value">\n        <span></span>\n    </button>\n</div>\n';});


define('oui.combo/listItemMarkup.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<li role="'+
((__t=( data.children ? 'group' : 'option'))==null?'':__t)+
'" title="'+
((__t=( data.text ))==null?'':_.escape(__t))+
'" id="'+
((__t=( data.pageId ))==null?'':__t)+
'"    ';
 if (data.speakAs != null && data.speakAs !== '') { 
__p+='        aria-label="'+
((__t=( data.speakAs ))==null?'':__t)+
'"    ';
 } 
__p+='    class="'+
((__t=( data.header ? 'combo-item-header' : 'combo-item' ))==null?'':__t)+
' '+
((__t=( data.selected ? 'checked' : '' ))==null?'':__t)+
' '+
((__t=( data.active ? 'active' : '' ))==null?'':__t)+
'"    data-combo-id="'+
((__t=( data.id  ))==null?'':__t)+
'"> <a href="#"> <i class="check"></i> '+
((__t=( data.display ))==null?'':__t)+
' </a> ';
 if(data.children) { 
__p+=' <ul class="children"> '+
((__t=( data.children ))==null?'':__t)+
' </ul> ';
 } 
__p+='</li>';
return __p;
};});


define('oui.combo/menuMarkup.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<ul role="listbox" aria-label="'+
((__t=( data.locale.menu ))==null?'':__t)+
'" class="dropdown-menu combo-dropdown-menu"></ul>';
return __p;
};});

/**
 * @module inlineStyle
 * @private
 */
define('oui.utils/inlineStyle',[], function () {
    

    /**
     * @param {jQuery} $el The element to grab the inline style property off of
     * @param {String} prop The property to look for in the attribute
     *
     * @return {String|undefined} The value of the property in the style attribute.
     */
    return function ($el, prop) {
        var styles = $el.attr('style'),
            value;
        if (styles) {
            styles.split(';').forEach(function (e) {
                var style = e.split(':');
                if (style[0].trim() === prop) {
                    value = style[1];
                }
            });
        }
        return value;
    };
});

define('oui.combo/comboBoxButtonView',['oui.underscore', 'oui.jquery', 'oui.base/view',
       'oui.utils/stringUtils',
    'oui.combo/comboBoxListTemplate.htm', 'oui.combo/listItemMarkup.htm',
    'oui.combo/menuMarkup.htm',
    'oui.utils/inlineStyle',
    'oui.mixins/popup'],
    function (_, $, View, stringUtils, comboBoxListTemplate, listItemTemplate, menuMarkup, inlineStyle, popupMixin) {
        

        var cvp;

        /**
         * This view displays the combobox as a dropdown without an input element, so no text can be typed
         *
         * @class ComboBoxButtonView
         * @private
         * @since 1.3.0
         * @extends View
         * @mixes popup
         * 
         * @param {ComboBoxModel} model The ComboBoxModel that is responsible for this view.
         * @param {ComboBoxController} controller The ComboBoxController that will modify the model.
         * @param {jQuery|null} element The default HTML input element that users can interact with.
         * @param {ComboBox#Config} options View specific options.
         */
        function ComboBoxButtonView(model, controller, element, options, settings) {
            // Super call
            View.call(this, options, element, controller, settings);

            this._$menu = this._$popup = $(menuMarkup({ locale : options.locale }))
                .attr('aria-multiselectable', options.multiSelect)
                .toggleClass('multi-select', options.multiSelect)
                .attr('id', this.id() + '-menu');

            this._$originalElement = this._$el.children('.combo-original-element');
            this._$dropDownButton = this._$el.children('.combo-button');
            this._$valueEl = this._$el.children('.combo-value');

            model.on({
                'view.selection-change' : this.updateValue.bind(this),
                'view.active-change' : this.updateActive.bind(this),
                'view.invalid-selection' : this.invalid.bind(this),
                'view.required-error' : this.invalid.bind(this),
                'view.values-change' : this._createDropdown.bind(this)
            });

            this._setupCustom();
            this._initListeners();

            this._createDropdown(this._controller.values());

            if(this._controller.selected().length > 0) {
                this.updateValue(this._controller.selected());
            }
        }

        ComboBoxButtonView.prototype = Object.create(View.prototype);
        cvp = ComboBoxButtonView.prototype;

        cvp.constructor = ComboBoxButtonView;

        _.extend(cvp, popupMixin);

        /**
         * Disables the view.
         *
         * @method ComboBoxButtonView#disable
         * @protected
         * @param {Boolean} disabled True to disable, false to enable
         */
        cvp.disable = function (disabled) {
            var els = this._$dropDownButton.add(this._$valueEl)
                .toggleClass('disabled', disabled);

            _.each(els, function (element) {
                if(disabled) {
                    element.setAttribute('disabled', 'disabled');
                } else {
                    element.removeAttribute('disabled');
                }
            });
        };

        cvp.invalid = function () {
            this._$el.add(this._$el.closest('.control-group')).addClass('error');

            this.updateValue(this._controller.selected());
        };

        cvp._setupCustom = function () {
            var originalId = this._$originalElement.attr('id');

            this._$dropDownButton.attr('title', this._options.titleText);

            // Swap IDs on the button so that corresponding labels still work
            this._$originalElement
                .attr('data-original-id', originalId)
                .removeAttr('id');
            this._$dropDownButton.attr('id', originalId);
        };

        cvp.template = function () {
            return comboBoxListTemplate;
        };

        cvp.buildElement = function (opts, el) {
            var $element = $(this.template());
            var $value = $element.children('.combo-value');

            // We were given a jquery element
            // Whether it's an <input> or a <select> we need to replace it with our own thing
            if (!_.isNull(el) && el.jquery) {
                el.after($element);

                var attrs = {};
                if(el.length > 0) {
                    _.each(el[0].attributes, function(key) {
                        if(!_.contains(['id', 'style', 'type', 'name'], key.nodeName)) {
                            attrs[key.nodeName] = key.value;
                        }
                    });
                }

                _.each(attrs, function(value, key) {
                    if(key !== 'class') {
                        $value.attr(key, value);
                        el.removeAttr(key);
                    }
                    el.attr('data-original-' + key, value);
                });

                // Move the element into the wrapper and hide it
                el.prependTo($element).css('display', 'none').addClass('combo-original-element');

                // Set the new element to the same width as the old one if it was applied specifically
                if (inlineStyle(el, 'width')) {
                    $element.css('width', el.css('width'));
                }

                if(opts.copyClasses && attrs['class']) {
                    $element.attr('class', $element.attr('class') + ' ' + attrs['class']);
                }
            }

            return $element;
        };

        cvp.init = function () {
            var $origElement = this._$el.find('.combo-original-element');

            if ($origElement.is('select') && this._controller.values().length === 0) {
                this._controller.values(this._controller.convert($origElement));

                var value = $origElement[0].value;
                var option = _.find($origElement.find('option'), function(option) {
                    return option.value === value;
                });

                if(option != null) {
                    this._controller.selected($(option).text());
                }
            }

            //TODO: find a way to set the ID before building the element?
            this._$el.find('.instructions').attr('id', this.id() + '-instructions');
            this._$valueEl.attr('aria-describedby', ((this._$valueEl.attr('aria-describedby') || '')  + ' ' + this.id() + '-instructions').trim());
        };

        cvp.destroy = function (clean) {
            var that = this;
            this._$originalElement
                .insertAfter(this._$el)
                .css('display', '');

            this._$menu.remove();

            var attrs = {};

            if(this._$originalElement[0] != null) {
                _.each(this._$originalElement[0].attributes, function(key) {
                    if(key.nodeName.match('data-original-')) {
                        attrs[key.nodeName.replace('data-original-', '')] = key.value;
                    }
                });

                _.each(attrs, function(value, key) {
                    that._$originalElement.attr(key, value);
                    that._$originalElement.removeAttr('data-original-' + key);
                });
            }

            if (clean.self) {
                this._$el.remove();
            }
        };

        cvp.updateValue = function (selected) {
            var that = this, displayValue, title;

            var span = document.createElement('span');

            this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid());

            this.active(this._controller.active());

            if (_.isNull(selected)) {
                this._$valueEl.children('span').text('');
                this._$valueEl.attr('title', '');
            } else {
                displayValue = _.map(selected, function(id) {
                    var intermValue = that._options.viewFns.display(that._controller.node(id)._data);
                    span.innerHTML = intermValue;
                    displayValue = span.textContent;

                    if(that._options.viewFns.text) {
                        intermValue = that._options.viewFns.text(that._controller.node(id)._data);
                    }

                    return intermValue;
                }).join(that._options.delimiter);

                if(this._options.viewFns.speak) {
                    title = _.map(selected, function(id) {
                        var intermValue = that._options.viewFns.speak(that._controller.node(id)._data);

                        if((intermValue == null || intermValue === '') && that._options.viewFns.text) {
                            intermValue = that._options.viewFns.text(that._controller.node(id)._data);
                        }

                        if(intermValue == null || intermValue === '') {
                            intermValue = that._options.viewFns.display(that._controller.node(id)._data);
                            span.innerHTML = intermValue;
                            displayValue = span.textContent;
                        }
                        return intermValue;
                    }).join(that._options.delimiter);
                }

                var domNode = this._$valueEl.children('span').html(displayValue)[0];

                if(title != null) {
                    this._$valueEl.attr('title', title);
                } else if(domNode != null) {
                    this._$valueEl.attr('title', domNode.textContent.trim());
                } else {
                    this._$valueEl.attr('title', displayValue);
                }
            }

            this._$menu.find('.checked').removeClass('checked').attr('aria-selected', false);
            _.each(selected, function(id) {
                that._$menu.find('[data-combo-id=' + id + ']')
                    .addClass('checked')
                    .attr('aria-selected', true);
            });

            return displayValue; //Used by subclass;
        };

        cvp.updateActive = function (active) {
            if(this._options.multiSelect) {
                this.active(this._controller.active());
            } else {
                this.updateValue(_.compact([active]));
            }
        };

        cvp.active = function (id) {
            var pageId = this._$menu.find('.combo-item').removeClass('active')
                .filter('[data-combo-id="' + id + '"]').addClass('active').attr('id');

            if(this.isVisible()) {
                this._$valueEl.attr('aria-activedescendant', pageId);
            }

            var active = this._$menu.find('.active');

            if (active.length > 0) {
                if (active.position().top + active.height() > this._$menu.height()) {
                    this._$menu.scrollTop(this._$menu.scrollTop() + active.position().top + active.height() - this._$menu.height());
                } else if (active.position().top < 0) {
                    this._$menu.scrollTop(this._$menu.scrollTop() + active.position().top);
                }
            }
        };

        /**
         * The function that handles the keyup, keydown, mousedown, mousein, and mouseout events of the ComboBox
         *
         * @private
         * @method ComboBoxButtonView#_initListeners
         */
        cvp._initListeners = function () {
            var that = this, cancelBlur;

            this._$dropDownButton.on('mousedown', function (event) {
                //don't lose focus from anything
                event.preventDefault();

                if (!that.isVisible()) {
                    that._expand();
                    that._$valueEl.focus();
                } else {
                    that._collapse();
                }

            });

            this._$valueEl.on('blur', function (event) {
                event.preventDefault();
                if (!cancelBlur && that.isVisible()) {
                    that._collapse();
                    that.updateValue(that._controller.selected());
                }
            }).on('focus', function () {
                if(that._options.openOnFocus) {
                    that._expand();
                }
            });

            this.addShortcuts({
                'ESC' : {
                    fn : function() {
                        that._collapse();
                        that.updateValue(that._controller.selected());

                        if(!that._options.multiSelect && that._controller.selected().length > 0) {
                            that._controller.active(that._controller.selected()[0]);
                        }
                    },
                    stopPropagation : function() {
                        return that.isVisible();
                    }
                },
                'ENTER' : {
                    fn : function() {
                        if (!that.isVisible()) {
                            that._expand();
                            that._$valueEl.focus();
                        } else {
                            that._handleEnter();
                        }
                    },
                    prevention : true
                },
                'SPACE' : {
                    fn : function() {
                        if (!that._$el.hasClass('combo-box-input') && !that.isVisible()) {
                            that._expand();
                            that._$valueEl.focus();
                        } else if(that.isVisible() && that._options.multiSelect) {
                            that._controller.selected([that._controller.active()]);
                            that._collapse();
                        }
                    },
                    prevention : function () {
                        return !that._$el.hasClass('combo-box-input') && !that.isVisible() ||
                            (that.isVisible() && that._options.multiSelect);
                    }
                },
                'shift-UP' : function() {
                    that._moveUp();
                    that._select(that._$menu.find('.active'), true, false);
                },
                'shift-DOWN' : function() {
                    that._moveDown();
                    that._select(that._$menu.find('.active'), true, false);
                },
                'SHIFT' : function() {
                    var active = that._findMenuItem(that._controller.active());
                    active.addClass('axis');
                },
                'ctrl-SPACE' : function() {
                    if(that.isVisible() && that._options.multiSelect) {
                        that._toggleSelect(that._controller.active());
                    }
                },
                'UP' : {
                    fn : function() {
                        if (that.isVisible()) {
                            that._moveUp();
                        } else {
                            that._expand();
                            that._$valueEl.focus();
                        }
                    },
                    prevention : true
                },
                'DOWN' : {
                    fn : function() {
                        if (that.isVisible()) {
                            that._moveDown();
                        } else {
                            that._expand();
                            that._$valueEl.focus();
                        }
                    },
                    prevention : true
                }
            });

            //TODO: look into shortcut mixin if it can have an up/down function
            this.addShortcut('SHIFT', {
                up : true,
                fn : function() {
                    that._$menu.find('.axis').removeClass('axis');
                }
            });

            if(!this._options.keyboardInput) {
                //http://api.jquery.com/keypress/
                //jQuery normalizes the .which property for us
                that._$el.on('keypress', function (event) {
                    var string = String.fromCharCode(event.which);
                    if(string.length > 0) {
                        that._find(string);
                    }
                });
            }

            //set click behavior for list items
            that._$menu.on('click', '.combo-item', function (event) {
                event.preventDefault();

                that._select($(this), event.shiftKey, event.ctrlKey);
            });

            that._$menu.on({
                'mouseover mouseenter': function () {
                    cancelBlur = true;
                },
                'mousedown': function (event) {
                    event.preventDefault();
                    that._$valueEl.focus();
                },
                'mouseleave': function () {
                    cancelBlur = false;
                }
            });

            //IE9 and potentially IE10 related
            //Super janky but need to refocus the element
            //when the scrollbar is mousedown'd.
            $(window).on('mousedown', function(event) {
                if($(event.target).is(that._$menu)) {
                    event.preventDefault();
                    setTimeout(function() {
                        that._$valueEl.focus();
                    }, 0);
                }
            });

        };

        cvp._toggleSelect = function(id) {

            if(!this._options.multiSelect) {
                this._controller.selected(id);
            } else {
                var selected = this._controller.selected();

                if(_.contains(selected, id)) {
                    this._controller.selected(_.without(selected, id));
                } else {
                    this._controller.selected(_.union(selected, [id]));
                }
            }
        };

        cvp._findMenuItem = function (id) {
            return this._$menu.find('[data-combo-id=' + id + ']');
        };

        cvp._select = function(item, shiftKey, ctrlKey) {
            var that = this;
            var id = this._getId(item);

            this._controller.active(id);

            if(this._options.multiSelect && (ctrlKey || shiftKey)) {
                if (ctrlKey) {
                    this._toggleSelect(id);
                } else if (shiftKey) {
                    var comboItems = this._$menu.find('.combo-item');
                    var selectedIndex = comboItems.index(item);
                    var prevSelectedIndex = comboItems.index(this._$menu.find('.axis'));
                    var selected;

                    if(selectedIndex < prevSelectedIndex) {
                        selected = comboItems.slice(selectedIndex, prevSelectedIndex + 1);
                    } else {
                        selected = comboItems.slice(prevSelectedIndex, selectedIndex + 1);
                    }

                    selected = _.map(selected, function(el) {
                        return that._getId($(el));
                    });

                    this._controller.selected(selected);
                }
            } else {
                this._controller.selected(id);
                this._collapse();
            }
        };

        cvp._find = function (searchString) {
            var display = this._options.viewFns.display,
                regex = new RegExp('^' + stringUtils.regexEscape(searchString), 'gi'),
                active = this._controller.active(),
                activeIndex, found,
                vals =  _.chain(this._controller.values())
                    .map(function(val) {
                        if(val._children) {
                            return val._children;
                        }
                        return val;
                    })
                    .flatten()
                    .value();

            active = this._controller.active();

            activeIndex = _.indexOf(vals, this._controller.node(active));

            if(active != null) {
                vals = _.sortBy(vals, function(val, index) {
                    if(index === activeIndex) {
                        return index + vals.length;
                    } else if (index > activeIndex ) {
                        return index - vals.length;
                    } else {
                        return index;
                    }
                });
            }

            found = _.find(vals, function(val) {
                return regex.test(display(val._data));
            });

            if(found != null) {
                if(this._options.multiSelect) {
                    this._expand();
                    this._controller.active(found._id);
                } else {
                    this._toggleSelect(found._id);
                }
            }
        };

        /**
         * The function that creates (but does not actually reveal) the contents of the ComboBox dropdown.
         *
         * @private
         * @method ComboBoxButtonView#_createDropdown
         */
        cvp._createDropdown = function (values, filter) {
            var that = this, regex;

            if (_.isString(filter)) {
                regex = new RegExp('(' + stringUtils.regexEscape(filter) + ')', 'gi');
            }

            var span = document.createElement('span');

            function buildMarkup(vals) {
                return _.reduce(vals, function (memo, val) {
                    var display = that._options.viewFns.display(val._data);
                    span.innerHTML = display;
                    var text = span.textContent;
                    var speak;
                    if(that._options.viewFns.text) {
                        text = that._options.viewFns.text(val._data);
                    }

                    if(that._options.viewFns.speak) {
                        speak = that._options.viewFns.speak(val._data);
                    }

                    if (regex != null) {
                        display = text.replace(regex, '<strong>$1</strong>');
                    }

                    memo += listItemTemplate({
                        id : val._id,
                        pageId : that.id() + '-' + val._id,
                        speakAs : speak,
                        display : display,
                        text : text,
                        selected : _.contains(that._controller.selected(), val._id),
                        active : that._controller.active() != null && val._id === that._controller.active()._id,
                        header : val._children !== undefined,
                        children : buildMarkup(val._children)
                    });

                    return memo;
                }, '');
            }

            that._$menu[0].innerHTML = buildMarkup(values);
            that._$menu.find('.combo-item').attr('aria-selected', false);
        };

        /**
         * Expands the ComboBox drop down
         *
         * @private
         * @method ComboBoxButtonView#_expand
         */
        cvp._expand = function () {
            this._$menu.width(this._$el.innerWidth())
                .children('.combo-item-hovered').removeClass('combo-item-hovered');

            this._$valueEl.attr('aria-expanded', true);

            this._$valueEl.attr('aria-activedescendant', this._$menu.find('.active').attr('id'));

            this.showPopup(this._$el);
            this._$valueEl.attr('aria-owns', this._$menu.attr('id'));

            if(this._controller.active() == null) {
                if(this._controller.selected().length === 0) {
                    this._controller.active(this._getId(this._$menu.find('.combo-item:eq(0)')));
                } else {
                    this._controller.active(_.first(this._controller.selected()));
                }
            } else if (!this._options.multiSelect && this._controller.active() !== _.first(this._controller.selected())) {
                this._controller.active(_.first(this._controller.selected()));
            } else {
                this.updateActive(this._controller.active());
            }
        };

        cvp._collapse = function () {
            this.hidePopup();
            this._$menu.trigger('mouseleave');
            this._$valueEl.removeAttr('aria-activedescendant');
            this._$valueEl.attr('aria-expanded', false);
        };

        cvp._getId = function ($el) {
            return parseInt($el.attr('data-combo-id'), 10);
        };

        cvp._moveDown = function () {
            //move the selection down one - if nothing is selected then select top
            //but start at the selected item
            var $active, $comboItems, activeIndex;

            $active = this._$menu.find('.combo-item.active');
            $comboItems = this._$menu.find('.combo-item');
            activeIndex = $comboItems.index($active);

            if (activeIndex < $comboItems.length - 1) {
                //Select next
                this._controller.active(this._getId($comboItems.eq(activeIndex + 1)));
            } else {
                this._controller.active(this._getId(this._$menu.find('.combo-item').first()));
            }
        };

        cvp._moveUp = function () {
            //move the selection down one - if nothing is selected then select top
            //but start at the selected item
            var $active, $comboItems, activeIndex;

            $active = this._$menu.find('.combo-item.active');
            $comboItems = this._$menu.find('.combo-item');
            activeIndex = $comboItems.index($active);

            if (activeIndex > 0) {
                //Select next
                this._controller.active(this._getId($comboItems.eq(activeIndex - 1)));
            } else {
                this._controller.active(this._getId(this._$menu.find('.combo-item').last()));
            }
        };

        cvp._handleEnter = function () {
            this._controller.selected(this._getId(this._$menu.find('.combo-item.active')));

            //If opened - close the menu
            if (this.isVisible()) {
                this._collapse();
            }
        };

        cvp.id = function (settings) {
            var input = this._$el.find('.combo-value');

            if(!arguments.length) {
                return input.attr('id');
            }

            if(input.length > 0 && _.isUndefined(input.attr('id'))) {
                if(_.isObject(settings)) {
                    input.attr('id', 'oui-' + settings.name + '-' + settings.id);
                } else if (_.isString(settings)) {
                    input.attr('id', settings);
                }
            }
        };


        return ComboBoxButtonView;
    }
);


define('oui.combo/comboBoxInputTemplate.htm',[],function () { return '<div class="orcl-combo-box combo-box-input">\n    <input role="combobox" aria-expanded="false" aria-autocomplete="list" class="combo-value" type="text" />\n    <button type="button" class="btn btn-picker combo-button" tabindex="-1"><span class="caret"></span></button>\n</div>\n';});

define('oui.combo/comboBoxInputView',['oui.underscore', 'oui.combo/comboBoxButtonView',
    'oui.utils/stringUtils',
    'oui.combo/comboBoxInputTemplate.htm', 'oui.utils/keyCode'],
    function (_, View, stringUtils, comboBoxInputTemplate, keyCode) {
        

        var cvp;

        /**
         * This view displays the combobox as a dropdown without an input element, so no text can be typed
         *
         * @class ComboBoxInputView
         * @private
         * @since 1.3.0
         * @extends ComboBoxButtonView
         * 
         * @param {ComboBoxModel} model The ComboBoxModel that is responsible for this view.
         * @param {ComboBoxController} controller The ComboBoxController that will modify the model.
         * @param {jQuery|null} element The default HTML input element that users can interact with.
         * @param {ComboBoxConfig} options View specific options.
         */
        function ComboBoxInputView(model, controller, element, options, settings) {
            // Super call
            View.call(this, model, controller, element, options, settings);

            //TODO: add properties only used in the input combobox
            model.on({
                'view.active-change' : { fn : this.updateActive, context : this }
            });
        }

        ComboBoxInputView.prototype = Object.create(View.prototype);
        cvp = ComboBoxInputView.prototype;

        cvp.constructor = ComboBoxInputView;

        cvp.template = function () {
            return comboBoxInputTemplate;
        };

        cvp.updateActive = function () {
            this.active(this._controller.active());
        };

        cvp._setupCustom = function () {
            var originalId = this._$originalElement.attr('id');

            this._$dropDownButton.attr('title', this._options.titleText);

            // Swap IDs on the button so that corresponding labels still work
            this._$originalElement
                .attr('data-original-id', originalId)
                .removeAttr('id');
            this._$valueEl.attr('id', originalId);
        };

        cvp.updateValue = function (id) {
            //Super call
            var value = View.prototype.updateValue.call(this, id);

            if(_.isObject(this._options.autoEscape)) {
                value = this._options.autoEscape.unescape(value);
            } else if (this._options.autoEscape) {
                value = _.unescape(value);
            }
            this._$valueEl[0].value = value;
        };

        cvp._initListeners = function () {
            var that = this, cancelInput = false;
            //Super call
            View.prototype._initListeners.call(this);

            //Setup input only listeners
            this._$el.on('keyup', function (event) {
                switch (event.keyCode) {
                case keyCode.BACKSPACE:
                    var value = that._$valueEl[0].value;

                    if(_.isObject(that._options.autoEscape)) {
                        value = that._options.autoEscape.escape(value);
                    } else if (that._options.autoEscape) {
                        value = _.escape(value);
                    }
                    that._autoComplete(value);
                    cancelInput = false;
                }
            })
            .on('keydown', function (event) {
                switch (event.keyCode) {
                case keyCode.BACKSPACE:
                    cancelInput = true;
                }
            })
            //Grid triggers input event from the column to trigger the auto complete after
            //the column editor was shown from a character key
            .on('input', function () {
                if(!cancelInput) {
                    var value = that._$valueEl[0].value;

                    if(_.isObject(that._options.autoEscape)) {
                        value = that._options.autoEscape.escape(value);
                    } else if (that._options.autoEscape) {
                        value = _.escape(value);
                    }
                    that._autoComplete(value);
                }
                cancelInput = false;
            })
            .on('change', 'input', function() {
                if(!that._$el.hasClass('open')) {
                    that._handleEnter();
                }
            });

        };

        cvp._autoComplete = function (text) {
            var that = this;

            function flatten(values) {
                var filterableNodes = [];

                _.each(values, function(value) {
                    if(value._children != null && value._children.length > 0) {
                        filterableNodes = _.union(filterableNodes, flatten(value._children));
                    } else {
                        filterableNodes.push(value);
                    }
                });

                return filterableNodes;
            }

            if (text.trim() !== '') {
                var allValues = flatten(this._controller.values());

                var vals = _.compact(_.map(allValues, function (value) {
                    if (that._options.viewFns.display(value._data).match(new RegExp(stringUtils.regexEscape(text), 'ig'))) {
                        return value;
                    }
                }));

                if (vals.length > 0) {
                    this._createDropdown(vals, text);
                    this._expand();
                    this._controller.active(this._getId(this._$menu.find('.combo-item:eq(0)')));
                } else {
                    this._collapse();
                }
            } else {
                //Display entire dropdown menu
                this._createDropdown(this._controller.values());
                this._expand();
                this.active(this._controller.selected());
            }
        };

        cvp._collapse = function () {
            //Super cal
            View.prototype._collapse.call(this);

            //TODO: find better way to reset drop down menu to non-filtered state
            this._createDropdown(this._controller.values());

            var active = this._controller.active();

            //Restore active if there was one
            if (!_.isNull(active)) {
                this._$menu.find('[data-combo-id=' + active + ']').addClass('active');
            }
        };

        cvp._handleEnter = function () {
            if (this._$el.hasClass('open') && this._$menu.find('.active').length > 0) {
                //Something in the menu was selected - select that
                View.prototype._handleEnter.call(this);
            } else {
                var value = this._$valueEl[0].value;

                if(_.isObject(this._options.autoEscape)) {
                    value = this._options.autoEscape.escape(value);
                } else if (this._options.autoEscape) {
                    value = _.escape(value);
                }
                //Attempt to use value in the input as the value
                this._controller.selected(value);
            }
        };

        return ComboBoxInputView;
    }
);

define('oui.combo/comboBoxView',['oui.combo/comboBoxButtonView', 'oui.combo/comboBoxInputView'], function (ComboBoxButtonView, ComboBoxInputView) {
    

    /*
     * Convenience function to automatically pick the proper view type based on the initial options
     */
    return function (model, controller, element, options, settings) {
        var CorrectConstructor = options.keyboardInput ? ComboBoxInputView : ComboBoxButtonView;
        return new CorrectConstructor(model, controller, element, options, settings);
    };
});

define('oui.combo/comboBox',['oui.underscore', 'oui.base/input', 'oui.combo/comboBoxModel', 'oui.combo/comboBoxController', 'oui.combo/comboBoxView', 'oui.logger'],
    function (_, Input, ComboBoxModel, ComboBoxController, ComboBoxView, logger) {
        

        /**
         * @callback ComboBox#display
         * @param {Object} data The nodes data to be used to construct a string
         * @return {String|HTML} The string representation of the node.
         */

        /**
         * @callback ComboBox#text
         * @param {Object} data The nodes data to be used to construct a string
         * @return {String} The string representation of the node.
         */

        /**
         * @callback ComboBox#speak
         * @param {Object} data The nodes data to be used to construct a string
         * @return {String} The string representation of the node to speak.
         */

        /**
         * @callback ComboBox#value
         * @param {Object} data The nodes data to be used to construct a string
         * @return {String} The unique identifier to use for the node.
         */

        /**
         * @typedef {Object} ComboBox#Config
         *
         * @property {Boolean} [ignoreInvalid=true] Allow combobox to ignore selections that result in invalid entries
         * @property {Boolean} [required=false] Whether or not the combobox is required. Much like the native select element
         * an element with a value of '' will result in a selection that will trigger an error.
         * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the input
         * element. When the property is an Object it should have two functions escape and unescape that take in the string
         * and provide the escaped or unescaped version of the string. When the property is a Boolean it will just used
         * underscores escape/unescape function.
         * @property {Boolean} [openOnFocus=false] Whether or not the combobox should open the list of values when focused.
         * @property {String|jQuery} [appendTo='body'] The jQuery object or selector to append the dropdown menu to.
         * @property {String} [titleText=''] The hover text display when mousing over the combobox.
         * @property {Boolean} [keyboardInput=true] Whether this combobox will allow the user to enter values with the
         * keyboard instead of just picking from a list. Setting to `true` will allow the user to type/auto-complete their
         * entries; setting to `false` means the user will need to search through a list in order to find the proper value.
         * @property {Boolean} [allowCustomValues=false] Whether the combo box allows custom values. This property is only
         * relevant if the keyboardInput property is set to `true`. Setting to `true` will allow a user to provide their own
         * value to the dropdown instead of just selecting from the provided options. **Currently Unused** **Note**: When
         * `false`, if the user attempts to provide a custom value, the value will revert to the old value when the focus is
         * lost.
         * @property {Boolean} [fluid=false] Whether the combo box takes up all available space inside the parent element.
         * @property {Boolean} [allowEmpty=false] Whether or not the combo box can be constructed with an empty value. If
         * false, the combo box will use the first selectable value in the values as the currently selected value. If true,
         * the combo box will allow the combobox to start as empty but once the user makes a selection they may not be able
         * to return to an empty state as there is no option that represents the empty state by default.
         * @property {Boolean} [multiSelect=false] Whether or not the combo box supports multi-selecting.
         * @property {String} [delimiter=', '] The delimiter to use for multi selection rendering behavior.
         * @property {Object} viewFns The functions used during rendering of the component.
         * @property {ComboBox#display} viewFns.display
         * @property {ComboBox#text} [viewFns.text]
         * @property {ComboBox#speak} [viewFns.speak]
         * @property {Object} storeFns The functions used during manipulation of data in the model.
         * @property {ComboBox#value} [storeFns.value]
         * @property {Boolean} [copyClasses=false] Whether the combobox should take the classes from the original element and
         * place them on the combobox (.orcl-combobox)
         */

        var __DEFAULTS__ = {
            ignoreInvalid : true,
            required : false,
            autoEscape : true,
            openOnFocus : false,
            appendTo : 'body',
            titleText : '',
            keyboardInput : true,
            allowCustomValues : false,
            fluid : false,
            allowEmpty : false,
            copyClasses : false,
            viewFns : {
                display : function (data) {
                    if (!data.name) {
                        if (!_.isUndefined(data.display)) {
                            return data.display;
                        }
                        return '';
                    }
                    return data.name;
                }
            },
            storeFns : {
                value : function (data) {
                    if(!data.value) {
                        if(data.id != null) {
                            return data.id;
                        }
                        return '';
                    }

                    return data.value;
                }
            },
            multiSelect : false,
            delimiter : ', '
        };

        /**
         * The ComboBox widget allows the user to pick from a pre-determined list of options, and optionally provide
         * their own values.
         *
         * See {@link ComboBox#Config} for more information regarding options.
         *
         * @example
         * <caption>Creating a Combo Box (accepts all defaults, with no values)</caption>
         *
         * var combo = new oui.ComboBox();
         * $container.append(combo.element());
         * combo.init();
         *
         * @example
         * <caption>Creating a Combo Box (from `select` element)</caption>
         *
         * var combo = new oui.ComboBox($select);
         * combo.init();
         *
         * @example
         * <caption>Creating a Combo Box (with values)</caption>
         *
         * var combo = new oui.ComboBox([
         *     { name : 'Subaru', value : 'subaru' },
         *     { name : 'Ford', value : 'ford' },
         *     { name : 'Toyota', value : 'toyota' },
         *     { name : 'Honda', value : 'honda' },
         *     { name : 'Jeep', value : 'jeep' },
         * ]);
         * $container.append(combo.element());
         * combo.init();
         *
         * @param {ComboBox#Config} [options] Configurable options for the ComboBox widget
         * @param {HTMLInputElement|HTMLSelectElement|jQuery} [element] The default `<input type="text">` or
         *                 `<select>` element to wrap into a ComboBox widget.
         * @param {Array} [initData] The values to populate the combobox with. The objects by default should be
         *                  in the form of { name : '', value : '' }. Of course, the display function can be updated
         *                  if the data is not keyed with name and value.
         *
         * @tutorial using-combobox
         *
         * @class ComboBox
         * @since 1.3.0
         * @extends Input
         */
        function ComboBox(options, element, initData) {
            var model;

            if (!_.isUndefined(options.values)) {
                logger.warn('options.values is deprecated, please move it to the initData parameter and the selected item to a method call');
                delete options.values;
            }

            //New options called type;
            options.type = options.keyboardInput ? 'combo-input' : 'combo-button';

            Input.call(this, options, element, initData);

            model = new ComboBoxModel(this._initData, this._options);
            this._controller = new ComboBoxController(model, this._options);
            this._view = new ComboBoxView(model, this._controller, this._element, this._options, this._settings);

            Object.defineProperty(this, '_data', {
                get : this.selected,
                set : this.selected
            });

            this._setupEvents(model);
        }

        ComboBox.prototype = Object.create(Input.prototype);
        var cbp = ComboBox.prototype;

        cbp.constructor = ComboBox;

        /**
         * Gets the widget's name.
         *
         * @method ComboBox#name
         * @return {String} 'combobox'
         */
        cbp.name = function () {
            return 'combobox';
        };

        /**
         * @method ComboBox#_translate
         * @private
         *
         * @param {String} name The event to translate
         * @param {...*} args The arguments to translate
         * @return {arguments|Array} The translated data
         */
        cbp._translate = function (name ) {
            var that = this;

            if(name === 'selection-change' ||
               name === 'valueSelected') {
                var selected = arguments[1];

                if(selected.length === 1) {
                    return [ name, this._controller.node(selected[0])._data];
                } else {
                    return [ name, _.map(selected, function(node) {
                        return that._controller.node(node)._data;
                    })];
                }
            }
            return arguments;
        };

        /**
         * Gets or sets the currently selected value from the ComboBox.
         *
         * When multiSelect is true this method accepts an Array of Strings or Objects otherwise it only accepts a single
         * item as the new value.
         *
         * When passing in a String it should be its display value. When passing in an Object the data should be a unique
         * identifier of the node that is selected.
         *
         * (A synonym for data)
         *
         * For the silent version of this method see {@link ComboBox#value}
         *
         * @fires ComboBoxModel#invalid-selection
         * @fires ComboBoxModel#required-error
         * @fires ComboBoxModel#selection-change
         * @fires ComboBoxModel#valueSelected
         *
         * @method ComboBox#selected
         * @param {Object|String|Array} [newSelected] The new value(s) to set in the ComboBox. Array may only be used when
         * `multiSelect` is true.
         * @returns {Object|Array<Object>}
         */
        cbp.selected = function (newSelected) {
            if (!arguments.length) {
                var that = this;
                var selected = this._controller.selected();

                if(!this._options.multiSelect && selected.length === 1) {
                    return that._controller.node(selected[0])._data;
                }

                var select = _.map(selected, function(node) {
                    return that._controller.node(node)._data;
                });

                if (select.length > 0) {
                    return select;
                } else {
                    return null;
                }
            }

            this._controller.selected(newSelected);
        };

        /**
         * Gets or sets the currently selected value from the ComboBox.
         *
         * See {@link ComboBox#selected} for more details about this method.
         *
         * Differences from selected:
         *     does not trigger public event (selection-changed)
         *
         * @method ComboBox#value
         * @param {Object|String|Array} [newSelected] The new value to set in the ComboBox
         * @returns {Object|Array<Object>}
         */
        cbp.value = function (newSelected) {
            if (!arguments.length) {
                return this.selected();
            }

            this._controller.selected(newSelected, { silent : true });
        };

        /**
         * Gets or sets the accepted values for the ComboBox.
         *
         * @fires ComboBoxModel#values-change
         *
         * @method ComboBox#values
         * @param {Array<Object>} [newValues] The new values list for the ComboBox
         * @returns {Array<Object>} The data used for the combobox nodes
         */
        cbp.values = function (newValues) {
            if (!arguments.length) {
                return _.pluck(this._controller.values(), '_data');
            }
            this._controller.values(newValues);
        };

        /**
         * @method ComboBox#defaults
         * @return {Object}
         */
        cbp.defaults = function () {
            return __DEFAULTS__;
        };

        return ComboBox;

    }
);


define('oui.editor/linkDialogTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="modal pgbu-modal hide fade" aria-role="dialog" aria-labelledby="linkDialogHeader"> <div class="modal-header"> <a class="modal-close" href="#" data-dismiss="modal" aria-hidden="true"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="linkDialogHeader">'+
((__t=( data.locale.editor.insertLink ))==null?'':__t)+
'</h3> </div> <div class="modal-body"> <div class="control-group"> <label class="control-label" for="modal-text-input">'+
((__t=( data.locale.editor.linkText ))==null?'':__t)+
'</label> <div class="controls"> <input class="link-text" type="text" id="modal-text-input"/> </div> </div> <div class="control-group"> <label class="control-label" for="modal-url-input">'+
((__t=( data.locale.editor.url ))==null?'':__t)+
'</label> <div class="controls"> <input class="link-url" type="text" id="modal-url-input"/> </div> </div> <div class="control-group"> <label class="control-label" for="link-location"> '+
((__t=( data.locale.editor.openIn ))==null?'':__t)+
' </label> <div class="controls"> <select id="link-location"> <option value="_blank">'+
((__t=( data.locale.editor.openInNew ))==null?'':__t)+
'</option> <option value="_self">'+
((__t=( data.locale.editor.openInCurrent ))==null?'':__t)+
'</option> </select> </div> </div> </div> <div class="modal-footer"> <button type="button" class="btn btn-blue-dark remove-link pull-left">'+
((__t=( data.locale.editor.removeLink ))==null?'':__t)+
'</button> <button type="button" class="btn btn-blue-dark cancel" data-dismiss="modal" aria-hidden="true">'+
((__t=( data.locale.cancel ))==null?'':__t)+
'</button> <button type="button" class="btn btn-action accept">'+
((__t=( data.locale.editor.insertLink ))==null?'':__t)+
'</button> </div></div>';
return __p;
};});

define('oui.editor/linkDialog',['oui.jquery', 'oui.underscore', 'oui.combo/comboBox',
       'oui.editor/linkDialogTemplate.htm', 'oui.bootstrap'], function ($, _, ComboBox, dialogTemplate) {
    

    var __DEFAULTS__ = {
        'cancel' : $.noop,
        'accept' : $.noop,
        'remove' : $.noop
    };

    function LinkDialog(options) {
        var that = this;

        options = _.defaults(options, __DEFAULTS__);

        this._$el = $(dialogTemplate({ locale : options.locale}));

        this._combo = new ComboBox({
            keyboardInput : false
        }, this._$el.find('#link-location')).init();

        this._$el.on('click', '.accept',function () {
            options.accept({
                text : that._$el.find('.link-text').val(),
                link : that._$el.find('.link-url').val(),
                location : that._combo.value().value
            });
            that._$el.modal('hide');
        }).on('click', '[data-dismiss=modal]',function (event) {
            event.stopPropagation();
            options.cancel();
        }).on('click', '.remove-link',function () {
            that._$el.modal('hide');
            options.remove();
        }).on('change input populated.oui', function () {
            var disabled = that._$el.find('.link-text').val() === '' ||
                            that._$el.find('.link-url').val() === '';
            that._$el.find('.accept')
                .toggleClass('disabled', disabled)
                .prop('disabled', disabled);
        });
    }

    LinkDialog.prototype = {
        show : function () {
            this._$el.modal('show');
        },
        populate : function (linkData) {
            this._$el.find('.link-text').val(linkData.text);
            this._$el.find('.link-url').val(linkData.link);
            if(linkData.location !== '') {
                this._combo.value({ value : linkData.location});
            }
            this._$el.trigger('populated.oui');
        },
        element : function () {
            return this._$el;
        }
    };

    return LinkDialog;
});


define('oui.editor/editorTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="orcl-editor"> <div class="toolbar-container"> </div> <div class="iframe-container"> <iframe src="about:blank" seamless="true" class="editor-frame" title="'+
((__t=( data.locale.editor.title ))==null?'':__t)+
'"> </iframe> </div></div>';
return __p;
};});


define('oui.editor/toolbarTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="editor-toolbar btn-toolbar"> ';
 if (data.hidden.formatting == null || data.hidden.formatting) { 
__p+=' <div class="btn-group formatting"> <select class="format"> <option value="h1">Heading 1</option> <option value="h2">Heading 2</option> <option value="h3">Heading 3</option> <option value="h4">Heading 4</option> <option value="h5">Heading 5</option> <option value="h6">Heading 6</option> <option value="p">Paragraph</option> <option value="blockquote">Block Quote</option> <option value="div">Normal</option> </select> </div> ';
 } 
__p+=' ';
 if (data.hidden.fontName == null || data.hidden.fontName) { 
__p+=' <div class="btn-group font-name"> <select> <option value="Times New Roman">Times New Roman</option> <option value="Arial">Arial</option> <option value="Georgia">Georgia</option> <option value="Verdana">Verdana</option> <option value="Courier New">Courier New</option> </select> </div> ';
 } 
__p+=' ';
 if (data.hidden.fontSize == null || data.hidden.fontSize) { 
__p+=' <div class="btn-group font-size"> <select> <option value="1">8</option> <option value="2">10</option> <option value="3">12</option> <option value="4">14</option> <option value="5">18</option> <option value="6">24</option> <option value="7">36</option> </select> </div> ';
 } 
__p+=' ';
 if (data.hidden.color == null || data.hidden.color) { 
__p+=' <div class="btn-group color"> <button type="button" title="'+
((__t=( data.locale.editor.fontColor ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.fontColor ))==null?'':__t)+
'" class="btn btn-blue-light font-color"> <i class="pgbu-icon-small pgbu-icon-font-color"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.backgroundColor ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.backgroundColor ))==null?'':__t)+
'" class="btn btn-blue-light background-color"> <i class="pgbu-icon-small pgbu-icon-background-color"></i> </button> </div> ';
 } 
__p+=' ';
 if (data.hidden.style == null || data.hidden.style) { 
__p+=' <div class="btn-group style" data-toggle="buttons-checkbox"> <button type="button" title="'+
((__t=( data.locale.editor.bold ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.bold ))==null?'':__t)+
'" data-command="bold" class="btn btn-blue-light bold"> <i class="pgbu-icon-small pgbu-icon-bold"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.italic ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.italic ))==null?'':__t)+
'" data-command="italic" class="btn btn-blue-light italic"> <i class="pgbu-icon-small pgbu-icon-italic"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.underline ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.underline ))==null?'':__t)+
'" data-command="underline" class="btn btn-blue-light underline"> <i class="pgbu-icon-small pgbu-icon-underline"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.strikeThrough ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.strikeThrough ))==null?'':__t)+
'" data-command="strikeThrough" class="btn btn-blue-light strikethrough"> <i class="pgbu-icon-small pgbu-icon-strikeout"></i> </button> </div> ';
 } 
__p+=' ';
 if (data.hidden.script == null || data.hidden.script) { 
__p+=' <div class="btn-group script" data-toggle="buttons-checkbox"> <button type="button" title="'+
((__t=( data.locale.editor.subscript ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.subscript ))==null?'':__t)+
'" data-command="subscript" class="btn btn-blue-light subscript"> <i class="pgbu-icon-small pgbu-icon-subscript"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.superscript ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.superscript ))==null?'':__t)+
'" data-command="superscript" class="btn btn-blue-light superscript"> <i class="pgbu-icon-small pgbu-icon-superscript"></i> </button> </div> ';
 } 
__p+=' ';
 if (data.hidden.alignment == null || data.hidden.alignment) { 
__p+=' <div class="btn-group alignment" data-toggle="buttons-radio"> <button type="button" title="'+
((__t=( data.locale.editor.justifyLeft ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.justifyLeft ))==null?'':__t)+
'" data-command="justifyLeft" class="btn btn-blue-light left"> <i class="pgbu-icon-small pgbu-icon-align-left"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.justifyCenter ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.justifyCenter ))==null?'':__t)+
'" data-command="justifyCenter" class="btn btn-blue-light center"> <i class="pgbu-icon-small pgbu-icon-align-center"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.justifyRight ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.justifyRight ))==null?'':__t)+
'" data-command="justifyRight" class="btn btn-blue-light right"> <i class="pgbu-icon-small pgbu-icon-align-right"></i> </button> </div> ';
 } 
__p+=' ';
 if (data.hidden.paragraph == null || data.hidden.paragraph) { 
__p+=' <div class="btn-group list-indent"> <button type="button" title="'+
((__t=( data.locale.editor.orderedList ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.orderedList ))==null?'':__t)+
'" data-command="insertOrderedList" class="btn btn-blue-light ordered-list"> <i class="pgbu-icon-small pgbu-icon-list-ordered"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.unorderedList ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.unorderedList ))==null?'':__t)+
'" data-command="insertUnorderedList" class="btn btn-blue-light unordered-list"> <i class="pgbu-icon-small pgbu-icon-list-unordered"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.increaseIndent ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.increaseIndent ))==null?'':__t)+
'" data-command="indent" class="btn btn-blue-light indent"> <i class="pgbu-icon-small pgbu-icon-indent"></i> </button> <button type="button" title="'+
((__t=( data.locale.editor.decreaseIndent ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.decreaseIndent ))==null?'':__t)+
'" data-command="outdent" class="btn btn-blue-light outdent"> <i class="pgbu-icon-small pgbu-icon-outdent"></i> </button> </div> ';
 } 
__p+=' ';
 if (data.hidden.link == null || data.hidden.link) { 
__p+=' <div class="btn-group link"> <button type="button" title="'+
((__t=( data.locale.editor.insertLink ))==null?'':__t)+
'" aria-label="'+
((__t=( data.locale.editor.insertLink ))==null?'':__t)+
'" class="btn btn-blue-light link"> <i class="pgbu-icon-small pgbu-icon-link"></i> </button> </div> ';
 } 
__p+='</div>';
return __p;
};});


define('oui.editor/linkTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<a href="'+
((__t=( data.link ))==null?'':__t)+
'" target="'+
((__t=( data.location ))==null?'':__t)+
'">'+
((__t=( data.text  ))==null?'':__t)+
'</a>';
return __p;
};});

define('oui.editor/editorView',['oui.underscore', 'oui.jquery', 'oui.base/view', 'oui.combo/comboBox', 'oui.color/colorPicker', 'oui.editor/linkDialog', 'oui.editor/editorTemplate.htm', 'oui.editor/toolbarTemplate.htm', 'oui.editor/linkTemplate.htm'],
    function (_, $, View, ComboBox, ColorPicker, LinkDialog, editorMarkup, toolbarTemplate, linkTemplate) {
        

        var HEADING_TO_VALUE = {
            'Heading 1' : 'h1',
            'Heading 2' : 'h2',
            'Heading 3' : 'h3',
            'Heading 4' : 'h4',
            'Heading 5' : 'h5',
            'Heading 6' : 'h6'
        };

        var evp;

        function EditorView(model, controller, element, options, settings) {
            var that = this;

            function updateDocument () {
                that._$el.closest('.control-group')
                    .add(that._$el)
                    .removeClass('error');
                if(that._$editorBody != null && that._$editorBody.html() !== that._controller.document()) {
                    that._$editorBody.html(that._controller.document());
                    that._updateToolbar();

                }
            }

            View.call(this, options, element, controller, settings);

            this._$document = null;
            this._$editorBody = null;
            this._$toolbar = this._$el.find('.toolbar-container');
            this._ranges = [];

            //Initialize comboboxes from selects
            this._linkDialog = new LinkDialog({
                locale : options.locale,
                accept : function (linkData) {
                    function replaceSelection() {
                        var selection = that._$document[0].getSelection();
                        if(selection.isCollapsed) {
                            selection.removeAllRanges();
                            _.each(that._ranges, function (range) {
                                selection.addRange(range);
                            });
                        }
                        that._$editorBody.focus();
                    }
                    function replaceSelectedText(replacementText) {
                        var sel, range;
                        if (that._$document[0].getSelection) {
                            sel = that._$document[0].getSelection();
                            if (sel.rangeCount) {
                                range = sel.getRangeAt(0);
                            } else if (that._ranges.length > 0) {
                                range = that._ranges[0];
                            }
                            range.deleteContents();
                            var textNode = that._$document[0].createTextNode(replacementText);
                            range.insertNode(textNode);

                            //Set cursor to end of new text node
                            range.setStartBefore(textNode);
                            range.setEndAfter(textNode);
                            sel.removeAllRanges();
                            sel.addRange(range);

                            return textNode;
                        } else if (that._$document[0].selection && that._$document[0].selection.createRange) {
                            range = that._$document[0].selection.createRange();
                            range.text = replacementText;

                            return range.htmlText;
                        }
                    }

                    if ((that._ranges.length > 0 && !that._ranges[0].collapsed) ||
                        !that._$document[0].queryCommandSupported('insertHTML')) {

                        replaceSelection();
                        //Need to unlink the current link if it exists otherwise browsers
                        //will keep the old link <a> in the document with no text
                        that._$document[0].execCommand('unlink', false);
                        var text = replaceSelectedText(linkData.text);
                        that._$document[0].execCommand('createLink', false, linkData.link);
                        var link = $(text).closest('a')[0];
                        link.target = linkData.location;
                        that._controller.document(that._$editorBody.html());
                    } else if(that._$document[0].queryCommandSupported('insertHTML')) {
                        that._$document[0].execCommand('insertHTML', false, linkTemplate(linkData).trim());
                        that._controller.document(that._$editorBody.html());
                    }

                    if(that._$editorBody[0].normalize) {
                        that._$editorBody[0].normalize();
                    }
                    that._$editorBody.focus();
                },
                remove : function () {
                    that._$document[0].execCommand('unlink');
                    that._$editorBody.focus();
                    that._controller.document(that._$editorBody.html());
                }
            });

            //Set up events from buttons
            that._setupToolbar();

            model.on({
                'document-changed': function () {
                    updateDocument();
                },
                'character-length-error': function () {
                    that.error();
                }
            });

            var iframe = this._$el.find('iframe')[0];

            //Set up listeners for the buttons that are active (queryState)
            //If iframe is completely loaded just call the functions
            //otherwise wait until it has loaded
            if (!_.isNull(iframe.contentWindow) && iframe.contentWindow.document.readyState === 'complete') {
                this.setupIframe(iframe.contentWindow);
                this._updateToolbar();
            } else {
                this._$el.find('.editor-frame').one('load', function() {
                    that.setupIframe(this.contentWindow);
                    that._updateToolbar();
                    updateDocument();
                });
            }
        }

        EditorView.prototype = Object.create(View.prototype);
        evp = EditorView.prototype;

        evp.constructor = EditorView;

        evp.setupIframe = function (iframeWindow) {
            var that = this,
                iframeDoc = iframeWindow.document;

            function sameSelection() {

                var ranges = iframeDoc.getSelection().rangeCount;

                if (ranges !== that._ranges.length) {
                    return false;
                }

                for (var i = 0; i < ranges; i++) {
                    var same = false;
                    for (var j = 0; j < that._ranges.length; j++) {
                        if (that._ranges[j].compareBoundaryPoints(window.Range.END_TO_END, iframeDoc.getSelection().getRangeAt(i)) === 0 &&
                            that._ranges[j].compareBoundaryPoints(window.Range.START_TO_START, iframeDoc.getSelection().getRangeAt(i)) === 0) {
                            same = true;
                        }
                    }

                    if (!same) {
                        return false;
                    }
                }

                return true;
            }

            this._$document = $(iframeDoc);
            this._$el.after(this._linkDialog.element());

            this._$editorBody = this._$document.find('body');
            this._$editorBody[0].contentEditable = true;
            this._$editorBody.html(this._controller.document());

            //Since this is an iframe with no external access to stylesheets
            //just use css attributes inline (shouldnt need anymore than these)
            this._$editorBody.css({
                'margin' : 0,
                'padding' : '8px',
                'font-family' : that._options.defaultFont != null ? that._options.defaultFont : '',
                'box-sizing' : 'border-box'
            });

            var iframe = this._$el.find('.editor-frame'),
                parent = iframe.parent();

            if(iframe.outerHeight() !== parent.height()) {
                iframe.outerHeight(parent.height());

                $(window).on('resize', function() {
                    iframe.hide();
                    iframe.height(parent.innerHeight() - 10);
                    iframe.show();
                });
            } else if(!parent.is(':visible')){

                $(window).on('resize', function() {
                    if(parent.height() !== iframe.height()) {
                        iframe.hide();
                        iframe.height(parent.innerHeight() - 10);
                        iframe.show();
                    }
                });
            }

            //Setup Editor Selection event handlers
            this._$document.on('beforedeactivate', function () {
                var selection = iframeDoc.getSelection();
                that._ranges = [];
                for (var i = 0; i < selection.rangeCount; i++) {
                    that._ranges[i] = selection.getRangeAt(i).cloneRange();
                }
            }).on('selectionchange toolbar-update.oui', function () {
                //Update the toolbar
                var selection = iframeDoc.getSelection();
                that._ranges = [];
                for (var i = 0; i < selection.rangeCount; i++) {
                    that._ranges[i] = selection.getRangeAt(i).cloneRange();
                }

                that._updateToolbar();
            }).on('mousedown', function () {
                //Firefox does not support the select event...so fake it.
                that._$document.on('mousemove mouseup mouseleave', function () {
                    if (!sameSelection()) {
                        that._$document.trigger('toolbar-update.oui');
                    }
                }).on('mouseup mouseleave', function () {
                    that._$document.off('mouseup mousemove mouseleave');
                });

            }).on('keydown keyup', function () {
                that._$document.trigger('toolbar-update.oui');
            }).on('blur change input keyup keydown', function () {
                //IE10-11 does not support input on contenteditable divs
                //http://connect.microsoft.com/IE/feedback/details/794285/ie10-11-input-event-does-not-fire-on-div-with-contenteditable-set
                that._controller.document($(this).find('body').html());
            });
        };

        evp.error = function (error) {
            if(error == null) {
                error = true;
            }

            this._$el.closest('.control-group')
                .add(this._$el)
                .toggleClass('error', error);
        };

        evp._updateToolbar = function () {
            var that = this, doc = this._$document[0];

            if (that._ranges.length === 0 && doc.getSelection().rangeCount === 0 && doc.getSelection().isCollapsed) {
                doc.getSelection().removeAllRanges();
                var range = doc.createRange();
                doc.getSelection().addRange(range);
            }

            _.each(this._$toolbar.find('.btn[data-command]'), function (element) {
                if (doc.queryCommandState($(element).attr('data-command'))) {
                    $(element).addClass('active');
                } else {
                    $(element).removeClass('active');
                }
            });

            var fontName = doc.queryCommandValue('fontName');
            if (fontName == null || fontName === '') {
                fontName = 'Times New Roman';
            }
            that._fontName.value({ value : fontName.split('\'').join('')});

            var fontSize = doc.queryCommandValue('fontSize');

            if (fontSize == null || fontSize === '' || fontSize === 0) {
                fontSize = '3';
            }
            that._fontSize.value({ value : fontSize.toString()});
            var formatBlock = doc.queryCommandValue('formatBlock');

            if(formatBlock === '' || formatBlock === 'Normal' || formatBlock === 'x') {
                formatBlock = 'div';
            }

            var trueValue = HEADING_TO_VALUE[formatBlock] || formatBlock;

            that._format.value({ value : trueValue});
        };

        evp._setupToolbar = function () {
            var that = this, options = this._options;

            function replaceSelection() {
                var selection = that._$document[0].getSelection();
                if(selection.isCollapsed) {
                    selection.removeAllRanges();
                    _.each(that._ranges, function (range) {
                        selection.addRange(range);
                    });
                }
                that._$editorBody.focus();
            }

            this._$toolbar.on('click', '.btn[data-command]', function () {
                replaceSelection();
                that._$document[0].execCommand($(this).attr('data-command'), false);
                that._controller.document(that._$editorBody.html());
            })
            .on('click', '[data-trigger]', function(event) {
                that.trigger($(this).attr('data-trigger'), event);
            });

            this._fontName = new ComboBox({
                fluid : true,
                titleText : options.locale.editor.fontName
            }, this._$toolbar.find('.font-name select'))
                .init()
                .on('valueSelected', function (value) {
                    replaceSelection();
                    that._$document[0].execCommand('fontName', false, value.value);
                    that._controller.document(that._$editorBody.html());
                });

            this._fontName.element().find('.btn-form').removeClass('btn-form').addClass('btn-blue-light');

            this._fontSize = new ComboBox({
                fluid : true,
                titleText : options.locale.editor.fontSize
            }, this._$toolbar.find('.font-size select'))
                .init()
                .on('valueSelected', function (value) {
                    replaceSelection();
                    that._$document[0].execCommand('fontSize', false, value.value);
                    that._controller.document(that._$editorBody.html());
                });

            this._fontSize.element().find('.btn-form').removeClass('btn-form').addClass('btn-blue-light');

            this._format = new ComboBox({
                fluid : true,
                titleText : options.locale.editor.fontFormat
            }, this._$toolbar.find('.formatting select'))
                .init()
                .on('valueSelected', function (value) {
                    replaceSelection();
                    var trueValue = value.value;

                    if(value.value === 'div')  {
                        trueValue = 'Normal';
                    }

                    that._$document[0].execCommand('removeFormat', false);
                    that._$document[0].execCommand('formatBlock', false, trueValue);

                    if(that._$document[0].queryCommandValue('formatBlock') !== trueValue) {
                        trueValue = '<' + trueValue + '>';
                        that._$document[0].execCommand('formatBlock', false, trueValue);
                        that._controller.document(that._$editorBody.html());
                    }
                });

            this._format.element().find('.btn-form').removeClass('btn-form').addClass('btn-blue-light');

            this._background = new ColorPicker({}, this._$toolbar.find('.background-color'))
                .init()
                .on('color-changed', function (selectedColor) {
                    replaceSelection();
                    that._$document[0].execCommand('backColor', false, selectedColor);
                    that._controller.document(that._$editorBody.html());
                });

            this._highlight = new ColorPicker({}, this._$toolbar.find('.font-color'))
                .init()
                .on('color-changed', function (selectedColor) {
                    replaceSelection();
                    that._$document[0].execCommand('foreColor', false, selectedColor);
                    that._controller.document(that._$editorBody.html());
                });

            function getLinkData() {
                var linkData = {
                    text : '',
                    location : '',
                    link : ''
                };

                //If the range is collapsed start with an empty link dialog
                if (that._ranges.length > 0 && !that._ranges[0].collapsed) {
                    //We have something selected!
                    var anchor = $(that._ranges[0].commonAncestorContainer).closest('a');

                    //Needed for IE due to commonAncestorContainer not containing the <a>
                    //but may cause issues with too much selected
                    if (anchor.length === 0) {
                        anchor = $(that._ranges[0].commonAncestorContainer).children('a:eq(0)');
                    }

                    if(anchor.length > 0) {
                        //We have an anchor that the text is a part of!
                        linkData.link = anchor.attr('href');
                        linkData.location = anchor.attr('target');
                    }

                    linkData.text = that._ranges[0].toString();

                    //IE does not properly return the string representation on toString
                    if(linkData.text === '') {
                        var clone = that._ranges[0].cloneContents();
                        var div = document.createElement('div');
                        div.appendChild(clone);
                        linkData.text = div.textContent;
                    }
                }

                return linkData;
            }

            this._$el.find('.btn.link').on('click', function () {
                that._linkDialog.show();
                //Get name and link and send that to the dialog
                that._linkDialog.populate(getLinkData());
            });
        };

        evp.insert = function(text) {
            var that = this;

            function replaceSelection() {
                var selection = that._$document[0].getSelection();
                if(selection.isCollapsed) {
                    selection.removeAllRanges();
                    _.each(that._ranges, function (range) {
                        selection.addRange(range);
                    });
                }
                that._$editorBody.focus();
            }
            function replaceSelectedText(replacementText) {
                var sel, range;
                if (that._$document[0].getSelection) {
                    sel = that._$document[0].getSelection();
                    if (sel.rangeCount) {
                        range = sel.getRangeAt(0);
                    } else if (that._ranges.length > 0) {
                        range = that._ranges[0];
                    }
                    range.deleteContents();
                    var el = that._$document[0].createElement('div');
                    el.innerHTML = replacementText;

                    var frag = that._$document[0].createDocumentFragment(), node, lastNode;
                    while ( (node = el.firstChild) ) {
                        lastNode = frag.appendChild(node);
                    }
                    range.insertNode(frag);

                    if(lastNode) {
                        //Set cursor to end of new text node
                        range.setStartAfter(lastNode);
                        range.setEndAfter(lastNode);
                        sel.removeAllRanges();
                        sel.addRange(range);
                    }
                } else if (that._$document[0].selection && that._$document[0].selection.createRange) {
                    range = that._$document[0].selection.createRange();
                    range.text = replacementText;
                }
            }
            if (!this._$document[0].queryCommandSupported('insertHTML')) {
                replaceSelection();
                replaceSelectedText(text);
            } else {
                this._$document[0].execCommand('insertHTML', false, text);
            }

            this._controller.document(that._$editorBody.html());
        };

        evp.buildElement = function (options, el, controller) {
            var $element = $(editorMarkup({ locale : options.locale})),
                $toolbar,
                existingMarkup;

            if(options.toolbar.jquery) {
                $toolbar = options.toolbar;
            } else {
                $toolbar = $(toolbarTemplate({ hidden : options.toolbar, locale : options.locale }));
            }

            if (!_.isNull(el) && el.jquery) {
                existingMarkup = el.html();
                if (existingMarkup.trim().length !== 0) {
                    controller.document(existingMarkup);
                }
                el.empty();
                $element = el.append($element.children());
                $element.addClass('orcl-editor');
            }

            $element.find('.toolbar-container').append($toolbar);

            $element.toggleClass('fill', options.fill);

            return $element;
        };

        evp.disable = function (disabled) {
            if (!arguments.length) {
                return this._$el.hasClass('disabled');
            }

            if (disabled) {
                this._$el.addClass('disabled');
                this._fontSize.disable(disabled);
                this._fontName.disable(disabled);
                this._format.disable(disabled);

                this._$editorBody[0].contentEditable = false;

                this._$el.find('.editor-frame').addClass('disabled');

                this._$toolbar.find('.btn-group > .btn').addClass('disabled').attr('disabled', 'disabled');
            } else {
                this._$el.removeClass('disabled');
                this._fontSize.disable(disabled);
                this._fontName.disable(disabled);
                this._format.disable(disabled);

                this._$editorBody[0].contentEditable = true;

                this._$el.find('.editor-frame').removeClass('disabled');
                this._$toolbar.find('.btn-group > .btn').removeClass('disabled').removeAttr('disabled');
            }
        };

        evp.init = function () {
            $(this._$el.parent()).append(this._linkDialog.element());
        };

        return EditorView;
    }
);

define('oui.editor/editorController',[], function () {
    

    var ecp;

    function EditorController(model, options) {
        this._model = model;
        this._options = options;
    }

    ecp = EditorController.prototype;

    ecp.constructor = EditorController;

    ecp.document = function (doc) {
        if (!arguments.length) {
            return this._model.document();
        }

        this._model.document(doc);
    };

    return EditorController;
});

define('oui.editor/editorModel',['oui.utils/subscribable'],
    function (Subscribable) {
        

        /**
         * Fired when the document is updated
         *
         * @event EditorModel#document-changed
         * @param {String} document The current document in the editor.
         */
        /**
         * Fired when the attempted document is longer than max characters
         * specified in the options.
         *
         * @event EditorModel#character-length-error
         * @param {String} doc The attempted document
         * @param {String} document The current document
         */

        var emp;

        /**
         * @class EditorModel
         * @extends Subscribable
         * @private
         *
         * 
         * @param {EditorOptions} options The options for the editor
         * @param {String} initData The initial document to set on the editor
         */
        function EditorModel(options, initData) {

            //Super call
            Subscribable.call(this);

            this._document = initData || '';
            this._options = options;

            this._lastErrorType = null;
            this._invalidData = null;
        }

        EditorModel.prototype = Object.create(Subscribable.prototype);
        emp = EditorModel.prototype;

        emp.constructor = EditorModel;

        emp.document = function (doc) {
            if (!arguments.length) {
                return this.isValid() ? this._document : this._invalidData;
            }

            var docLength = this.characters(doc);

            if(this._options.maxCharacters && docLength > this._options.maxCharacters) {
                this._lastErrorType = 'character-length-error';
                this._invalidData = doc;
                this.trigger(this._lastErrorType, doc, this._document);
            } else if(this._document !== doc || !this.isValid()) {
                this._document = doc;
                this._invalidData = null;

                this.trigger('document-changed', doc);
            }
        };

        emp.characters = function (string) {
            string = string || this._document;
            return string.replace(/<[^>]*>/g, '').replace(/&[^;]*;/g, 'a').length;
        };

        emp.isValid = function () {
            return this._invalidData == null;
        };

        return EditorModel;
    }
);

define('oui.editor/editor',['oui.base/widget', 'oui.editor/editorView', 'oui.editor/editorController', 'oui.editor/editorModel'], function (Widget, View, Controller, Model) {
    

    /**
     * @typedef {Object} Editor#ToolbarConfig
     * @property {String} formatting
     * @property {String} fontSize
     * @property {String} fontName
     * @property {String} color
     * @property {String} style
     * @property {String} script
     * @property {String} alignment
     * @property {String} paragraph
     * @property {String} link
     */

    /**
     * @typedef {Object} Editor#Config
     * @property {String|null} [defaultFont=null] Font to use in the editor iframe. Useful when there is no font selector in
     * the toolbar and the font should be constant within the editor. Should be a font-family value.
     * @property {Boolean} [showToolbar=true] Whether or not to show the toolbar for the component.
     * @property {Editor#ToolbarConfig|jQuery} [toolbar={}] Object that represents which components should be on the toolbar. If a
     * key is undefined it defaults to visible. If the property is a jQuery object it is appended in the toolbar's spot. This
     * allows for custom events.
     * @property {Boolean} [fluid=false] Whether or not the editor should take up as much horizontal space as it can.
     * @property {Boolean} [fill=false] Whether or not the editor should take up as much vertical space as it can.
     * @property {Number|null} [maxCharacter=null] The maximum number of characters allow in the editor.
     */


    var ep,
    __DEFAULTS__ = {
        defaultFont : null,
        showToolbar : true,
        toolbar : {},
        fluid : false,
        fill : false,
        maxCharacters : null
    }, __NAME__ = 'editor';


    /**
     * The common rich text editor. Displays a UI element that can build a text document with
     * rich text features.
     *
     * @example
     * <caption>Constructing an Editor (accepts all defaults)</caption>
     *
     * var editor = new oui.Editor();
     * $container.append(editor.element());
     * editor.init();
     *
     * @example
     * <caption>Constructing an Editor with custom toolbar</caption>
     *
     * var editor = new oui.Editor({
     *     toolbar : $toolbar
     * });
     * $container.append(editor.element());
     * editor.init();
     *
     * @class Editor
     * @extends Widget
     * @since 1.3.0
     * @tutorial advanced-editor-usage
     *
     * @param {Editor#Config} [options] The options for the Editor. If not
     *  specified, all the defaults will be used.
     * @param {HTMLDivElement|jQuery} [element] The JQuery element that this field should use.
     *  If not specified, Editor will construct its own element.
     * @param {String} [initData] The document to use.
     */
    function Editor(options, element, initData) {
        var model;

        //Super call
        Widget.call(this, options, element, initData);

        model = new Model(this._options, this._initData);
        this._controller = new Controller(model, this._options);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);
    }

    Editor.prototype = Object.create(Widget.prototype);
    ep = Editor.prototype;

    ep.constructor = Editor;

    /**
     * Returns the widget name.
     *
     * @method Editor#name
     * @return {String} 'editor'
     */
    ep.name = function () {
        return __NAME__;
    };

    /**
     * Gets/Sets the document in the editor.
     *
     * @method Editor#document
     * @fires EditorModel#document-changed
     * @fires EditorModel#character-length-error
     * @param {String} [doc] HTML document fragment to set on the editor.
     * @return doc The document that is rendered in the editor.
     */
    ep.document = function (doc) {
        if (!arguments.length) {
            return this._controller.document();
        }

        this._controller.document(doc);
    };

    /**
     * Inserts an HTML fragment into the editor document's current cursor location.
     *
     * @method Editor#insert
     * @fires EditorModel#document-changed
     * @fires EditorModel#character-length-error
     * @param {String} text The HTML fragment to set at the current editors cursor location.
     */
    ep.insert = function (text) {
        this._view.insert(text);
    };

    /**
     * Renders the view as an error state.
     *
     * @method Editor#error
     * @param {Boolean} error Whether or not the editor is in an error state.
     */
    ep.error = function (error) {
        this._view.error(error);
    };

    /**
     * @method Editor#defaults
     * @return {Object} The defaults for the Editor.
     */
    ep.defaults = function () {
        return __DEFAULTS__;
    };

    return Editor;

});

/**
 * @module is
 * @private
 */
define('oui.utils/is',['oui.jquery', 'oui.underscore'], function ($, _) {
    

    var OUI_CLASSES = [
        'orcl-date-picker',
        'orcl-picker',
        'orcl-picker-input',
        'orcl-typeahead',
        'tree',
        'orcl-number-field',
        'orcl-editor',
        'orcl-file-picker',
        'orcl-combo-box',
        'combo-box-list',
        'orcl-color-picker',
        'orcl-icon-picker',
        'orcl-url-field'
    ];

    return function (elem) {
        if (!elem.jquery) {
            elem = $(elem);
        }


        if(elem.length > 1) {
            //TODO: WARN there are multiple elements requested?
            return false;
        } else {
            var hasClass = _.bind(elem.hasClass, elem);
            var parentHasClass = _.bind(elem.hasClass, elem.parent());
            return _.any(OUI_CLASSES, hasClass) ||
                _.any(OUI_CLASSES, parentHasClass);
        }
    };
});

/**
 * @namespace orcl.picker
 */
define('oui.picker/pickerWidget',['oui.jquery', 'oui.picker/picker', 'oui.jqueryui'],
    function ($, Picker) {
        

        $.widget('orcl.picker', {

            _create : function () {
                var widget = this;

                this.widgetEventPrefix = this.widgetEventPrefix + '-';

                this._picker = new Picker(this.options, this.element);
                this._picker.init();

                this._picker.on('select', function (node) {
                    widget._trigger('select', null, node);
                });

                this._picker.on('accept', function (data) {
                    //TODO: look into why the jquery event
                    //with multiselect unwraps the array to make a new
                    //array which then loses all but the first element
                    widget._trigger('accept', null, [data]);
                });
            },

            /**
             *
             * @method orcl.picker#getSelected
             * @deprecated
             */
            getSelected : function () {
                return this._picker.getSelected();
            },

            /**
             * @method orcl.picker#selected
             * @param {Array} data The array of data to be selected in the picker.
             * @return {Array} - The array of selected nodes in the picker.
             */
            selected : function (data) {
                if (!arguments.length) {
                    return this._picker.selected();
                }

                this._picker.selected(data);
            },

            select : function (data) {
                return this._picker.select(data);
            },

            unselectable : function (data) {
                if (!arguments.length) {
                    return this._picker.unselectable();
                }

                this._picker.unselectable(data);
            },

            expanded : function (data) {
                if (!arguments.length) {
                    return this._picker.expanded();
                }

                this._picker.expanded(data);
            },

            disable : function (disabled) {
                this._picker.disable(disabled);
            },

            disabled : function (data) {
                if (!arguments.length) {
                    return this._picker.disabled();
                }

                this._picker.disabled(data);
            },

            _destroy : function () {
                this._picker.destroy({ self : false});
                delete this._picker;
            }
        });

    }
);


define('oui.splitPane/controlsTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="splitter"> <div class="splitter-inner"> <div class="collapse-handle collapse-handle-horizontal collapse-handle-up" title="'+
((__t=( data.collapse.up ))==null?'':__t)+
'"></div> <div class="collapse-handle collapse-handle-vertical collapse-handle-left" title="'+
((__t=( data.collapse.left ))==null?'':__t)+
'"></div> <div class="resize-handle" title="'+
((__t=( data.resize ))==null?'':__t)+
'"></div> <div class="collapse-handle collapse-handle-horizontal collapse-handle-down" title="'+
((__t=( data.collapse.down ))==null?'':__t)+
'"></div> <div class="collapse-handle collapse-handle-vertical collapse-handle-right" title="'+
((__t=( data.collapse.right ))==null?'':__t)+
'"></div> </div></div>';
return __p;
};});

/**
 * A basic split pane widget that allows dynamic resizing of two HTML elements in a container
 *
 * @namespace orcl.splitPane
 */
define('oui.splitPane/splitPaneWidget',['oui.jquery', 'oui.underscore', 'oui.utils/inlineStyle', 'oui.splitPane/controlsTemplate.htm', 'oui.utils/language', 'oui.logger', 'oui.jqueryui'],
    function ($, _, inlineStyle, splitterTemplate, language, logger) {
        

        /**
         * Returns the specified CSS property as a pixel value
         *
         * @method orcl.splitPane~intCssProperty
         * @private
         * @param {*|HTMLElement} elem The element to operate on
         * @param {String} property The CSS property to retrieve
         * @returns {Number} the pixel size of the CSS property
         */
        function intCssProperty(elem, property) {
            return parseInt(elem.css(property).slice(0, -2), 10);
        }

        /**
         * Forces the removal of a CSS property that cannot be removed with `$(...).css(property, '');`
         *
         * This can happen when a property is applied by manipulating the `style` attribute or using the `!important`
         * modifier.
         *
         * @method orcl.splitPane~forceRemoveProperty
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property the CSS property to remove
         */
        function forceRemoveProperty(element, property) {
            element[0].style.removeProperty(property);
        }

        /**
         * Forces the removal of a CSS property that cannot be removed with `$(...).css(property, '');`
         *
         * This can happen when a property is applied by manipulating the `style` attribute or using the `!important`
         * modifier.
         *
         * @method orcl.splitPane~setCalcProperty
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property The property to add a calc'd field for
         * @param {String} value The calculated size, generally in the form "XX% +/- YYpx"
         */
        function setCalcProperty(element, property, value) {
            var propStrings = [
                'calc(' + value + ')',
                '-moz-calc(' + value + ')',
                '-webkit-calc(' + value + ')'
            ];

            element.each(function () {
                var $this = $(this), that = this;
                forceRemoveProperty($this, property);
                _.each(propStrings, function (str) {
                    if(that.style[property] === '') {
                        that.style[property] = str;
                    }
                });
            });
        }

        /**
         * Removes a CSS property from an object and stores it to be retrieved later.
         *
         * @method orcl.splitPane~saveOriginalCSSValue
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property The property to store
         */
        function saveOriginalCSSValue(element, property) {
            element.data('sp-original-' + property, inlineStyle(element, property));
        }

        /**
         * Restores a CSS property for an object that had its property stored earlier.
         *
         * @method orcl.splitPane~restoreOriginalCSSValue
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property The property to retrieve
         */
        function restoreOriginalCSSValue(element, property) {
            element.css(property, element.data('sp-original-' + property) || '');
            $.removeData(element, 'sp-original-' + property);
        }

        /**
         * Checks to see if a CSS property is stored.
         *
         * @method orcl.splitPane~hasOriginalCSSValue
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property The property to store
         */
        function hasOriginalCSSValue(element, property) {
            return !_.isUndefined(element.data('sp-original-' + property));
        }

        $.widget('orcl.splitPane', {

            /**
             * @typedef {Object} orcl.splitPane~Config
             *
             * @property {String} [orientation='horizontal'] The direction of the split pane. "horizontal" split panes can be
             * resied up or down; "vertical" split panes can be resized left or right.
             *
             * @property {Boolean} [collapseFirst=false] Determines whether the left or top child of the split pane can be
             * collapsed.
             *
             * @property {Boolean} [collapseSecond=false] Determines whether the right or bottom child of the split pane can
             * be collapsed.
             *
             * @property {Number} [firstSplitSize=0.5] The split ratio of the split pane, represented as the relative size of
             * the first child pane to the size of the container. For example, a 50:50 split would have a `firstSplitSize` of
             * 0.5. A 1:2 split would have a `firstSplitSize` of 0.33333333333.
             *
             * @property {Number} [throttleSpeed=500] The rate at which the secondary resize event, 'split-resize-throttled'
             * is fired, in milliseconds.
             *
             * @property {Object} [locale={}] Localization information for the split pane.
             *
             * @property {Boolean} [enabled=true] Whether or not the splitter and it's components are enabled
             */

            /**
             * @property {orcl.splitPane~Config} orcl.splitPane~options
             */
            options : {
                orientation : 'horizontal',
                collapseFirst : false,
                collapseSecond : false,
                firstSplitSize : 0.5,
                throttleSpeed : 500,
                locale : language.lang(),
                enabled : true
            },

            /**
             * Makes sure that a new split pane only contains two children.
             *
             * @method orcl.splitPane#_validateElement
             * @private
             */
            _validateElement : function () {
                if (this.element.children().length !== 2) {
                    var message = 'A split pane container must only contain two child elements';
                    logger.fatal(message);
                    throw new Error(message);
                }
            },

            /**
             * Returns the CSS property that matches the direction the split pane resizes.
             *
             *     horizontal => height
             *     vertical   => width
             *
             * @method orcl.splitPane#_getSplitSizeProperty
             * @returns {String}
             * @private
             */
            _getSplitSizeProperty : function () {
                return this.options.orientation === 'horizontal' ? 'height' : 'width';
            },

            /**
             * Returns the CSS property that is the opposite direction that the split pane resizes.
             *
             *     horizontal => width
             *     vertical   => height
             *
             * @method orcl.splitPane#_getOppositeSizeProperty
             * @returns {string}
             * @private
             */
            _getOppositeSizeProperty : function () {
                return this.options.orientation === 'horizontal' ? 'width' : 'height';
            },

            /**
             * Creates the splitter `<div>`
             *
             * @method orcl.splitPane#_createSplitter
             * @returns {*|HTMLElement}
             * @private
             */
            _createSplitter : function () {
                return $(splitterTemplate(this.options.locale));
            },

            /**
             * @method orcl.splitPane#_removeMouseHandlers
             * @private
             */
            _removeMouseHandlers : function () {
                $(document).off( 'mousemove' + this.mouseNamespace)
                    .off('mouseleave' + this.mouseNamespace)
                    .off('mouseup' + this.mouseNamespace);
            },

            /**
             * Sets up the event listeners for the splitter resizing
             *
             * @method orcl.splitPane#_setupHandlers
             * @private
             */
            _setupHandlers : function () {
                var that = this,
                    lastX, lastY;

                this.$splitter
                    .on('mousedown', function (event) {
                        if (that.options.enabled) {
                            $('body').addClass('oui-disable-select');
                            lastX = event.pageX;
                            lastY = event.pageY;

                            var moveHandler = function (event) {
                                if (that.options.orientation === 'horizontal') {
                                    that._mouseSetSplit(event.pageY - lastY);
                                } else if (that.options.orientation === 'vertical') {
                                    that._mouseSetSplit(event.pageX - lastX);
                                }

                                if (!that._splitterEdge) {
                                    lastX = event.pageX;
                                    lastY = event.pageY;
                                }
                            };

                            $(document)
                                .on('mousemove' + that.mouseNamespace, moveHandler)
                                .on('mouseleave' + that.mouseNamespace, function() {
                                    $('body').removeClass('oui-disable-select');
                                    that._removeMouseHandlers();
                                })
                                .on('mouseup' + that.mouseNamespace, function () {
                                    $('body').removeClass('oui-disable-select');
                                    that._removeMouseHandlers();
                                });
                        }

                    });


                this.$splitter.find('.collapse-handle-up, .collapse-handle-left')
                    .on('click.collapse-handler', function () {
                        if (that.options.enabled) {
                            that.collapseFirst();
                        }
                    });

                this.$splitter.find('.collapse-handle-down, .collapse-handle-right')
                    .on('click.collapse-handler', function () {
                        if (that.options.enabled) {
                            that.collapseSecond();
                        }
                    });
            },

            /**
             * Sets the proper classes on the collapse handles on the splitter bar
             *
             * @method orcl.splitPane#_setupCollapseHandles
             * @private
             */
            _setupCollapseHandles : function () {

                this.$splitter.find('.collapse-handle').addClass('unavailable');

                if (this.options.collapseFirst) {
                    this.$splitter.find('.collapse-handle-up, .collapse-handle-left').removeClass('unavailable');
                }

                if (this.options.collapseSecond) {
                    this.$splitter.find('.collapse-handle-down, .collapse-handle-right').removeClass('unavailable');
                }

            },

            /**
             * Updates the DOM for the splitter
             *
             * @method orcl.splitPane#_setupSplit
             * @private
             */
            _setupSplit : function () {
                var splitSizeProp = this._getSplitSizeProperty();

                this.$splitContainer.append(this.$firstChild.addClass('split-child'));
                this.$splitContainer.append(this.$secondChild.addClass('split-child'));
                this.$splitter.insertAfter(this.$firstChild);

                saveOriginalCSSValue(this.$firstChild, splitSizeProp);
                saveOriginalCSSValue(this.$secondChild, splitSizeProp);
                saveOriginalCSSValue(this.$firstChild, 'max-' + splitSizeProp);
                saveOriginalCSSValue(this.$secondChild, 'max-' + splitSizeProp);

                forceRemoveProperty(this.$firstChild, 'max-' + splitSizeProp);
                forceRemoveProperty(this.$secondChild, 'max-' + splitSizeProp);

                this._setupCollapseHandles();

                this._setupHandlers();

                this.setSplit();

            },

            /**
             * A common location to update the call to the event 'split-resize-throttled' so that both the constructor
             * and subsequest 'option' calls can modify it.
             *
             * @method orcl.splitPane#_updateSplitResizeThrottle
             * @private
             */
            _updateSplitResizeThrottle : function () {
                var that = this;
                this._splitResizeThrottle = _.throttle(function () {
                    /**
                     * A rate-limited splitter resize event. Its frequency is determined by the `throttleSpeed`
                     * property
                     *
                     * @event orcl.splitPane#split-resize-throttled
                     */
                    that._trigger('split-resize-throttled');
                }, this.options.throttleSpeed);
            },

            /**
             * Overridden jQuery UI constructor function
             *
             * @method orcl.splitPane#_create
             * @override
             * @private
             */
            _create : function () {

                this.eventNamespace = this.eventNamespace + '-';

                this._uniqueId = _.uniqueId('orcl-sp-');
                this.mouseNamespace = '.sp-drag-handler' + this._uniqueId;

                this._validateElement();

                this._updateSplitResizeThrottle();

                this.$firstChild = this.element.children(':eq(0)');
                this.$secondChild = this.element.children(':eq(1)');

                this.$splitContainer = $('<div class="split-container"></div>').addClass(this.options.orientation).appendTo(this.element);
                this.$splitter = this._createSplitter().appendTo(this.$splitContainer);

                this._setupSplit();
                this.setSplit(); // Second call to reflow after scrollbars go away!

                this.disable(!this.options.enabled);
            },

            /**
             * Internal convenience method to allow the splitter to change orientations
             *
             * @method orcl.splitPane#_reorient
             * @private
             */
            _reorient : function () {
                var oppositeSizeProperty = this._getOppositeSizeProperty();
                this.$splitContainer.removeClass('vertical').removeClass('horizontal').addClass(this.options.orientation);
                forceRemoveProperty(this.$firstChild, oppositeSizeProperty);
                forceRemoveProperty(this.$secondChild, oppositeSizeProperty);
                forceRemoveProperty(this.$splitter, this.options.orientation === 'vertical' ? 'top' : 'left');
            },

            /**
             * Overridden jQuery UI _setOption method.
             *
             * @method orcl.splitPane#_setOption
             * @private
             * @param {String} key Option key
             * @param {*} value Option value
             */
            _setOption : function (key, value) {
                if (this._superApply) {
                    this._superApply(arguments);
                } else {
                    this.options[key] = value;
                }

                switch (key) {
                case 'firstSplitSize':
                    this.setSplit();
                    break;
                case 'orientation':
                    this._reorient();
                    this.setSplit();
                    break;
                case 'collapseFirst':
                    this._setupCollapseHandles();
                    break;
                case 'collapseSecond':
                    this._setupCollapseHandles();
                    break;
                case 'throttleSpeed':
                    this._updateSplitResizeThrottle();
                    break;
                }
            },

            /**
             * Internal method called to resize splitter by a relative amount
             *
             * @method orcl.splitPane#_mouseSetSplit
             * @private
             * @param {Number} delta Amount to resize the splitter by, in **pixels**.
             */
            _mouseSetSplit : function (delta) {
                var available = intCssProperty(this.$splitContainer, this._getSplitSizeProperty());
                this.setSplit(this.options.firstSplitSize + (delta / available));
            },

            /**
             * Internal method to revert the removed min-values when the splitter is collapsed
             *
             * @method orcl.splitPane#_restoreMinValues
             * @private
             * @param {Boolean} collapse Is the splitter being collapsed
             * @param {String} splitSizeProperty The property to restore values
             */
            _restoreMinValues : function (collapse, splitSizeProperty) {
                if (hasOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty)) {
                    restoreOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty);
                }

                if (hasOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty)) {
                    restoreOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty);
                }

                if (collapse) {
                    saveOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty);
                    saveOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty);
                    this.$firstChild.css('min-' + splitSizeProperty, '');
                    this.$secondChild.css('min-' + splitSizeProperty, '');
                }
            },

            /**
             * Toggles enablement of the splitter.
             *
             * @method orcl.splitPane#disable
             * @param {Boolean} disable Whether or not to disable the splitter.
             */
            disable : function(disable) {
                this._setOption('enabled', !disable);
                if (disable) {
                    this.$splitter.addClass('disabled');
                } else {
                    this.$splitter.removeClass('disabled');
                }
            },

            /**
             * Sets the splitter division based on a pixel value.
             *
             * @method orcl.splitPane#setSplitPixels
             * @param {Number} newSplitSize The size (in **pixels**) of the first splitter child.
             */
            setSplitPixels : function (newSplitSize) {
                var available = intCssProperty(this.$splitContainer, this._getSplitSizeProperty());
                this.setSplit(newSplitSize / available);
            },

            /**
             * Sets the splitter division based on a percentage of the container.
             *
             * @method orcl.splitPane#setSplit
             * @param {Number} newSplitSize The size (as a **percentage**) of the split container
             * @param {Boolean} collapse (Internal) Used to signal that this is a collapse operation
             */
            setSplit : function (newSplitSize, collapse) {
                var splitSizeProperty = this._getSplitSizeProperty(),
                    available = intCssProperty(this.$splitContainer, splitSizeProperty),
                    splitterSize = intCssProperty(this.$splitter, splitSizeProperty),
                    splitterSizeHalf = splitterSize / 2,
                    absMin = splitterSizeHalf / available,
                    firstMin = collapse ?
                        absMin :
                        (Math.max(intCssProperty(this.$firstChild, 'min-' + splitSizeProperty) || 0, absMin) + splitterSizeHalf) / available,
                    secondMin = collapse ?
                        absMin :
                        (Math.max(intCssProperty(this.$secondChild, 'min-' + splitSizeProperty) || 0, absMin) + splitterSizeHalf) / available,
                    firstPercent, secondPercent;

                if (!_.isUndefined(newSplitSize)) { // Need to handle setting to 0.0
                    this.options.firstSplitSize = newSplitSize;
                }

                this._restoreMinValues(collapse, splitSizeProperty);

                this._splitterEdge = false;

                if (this.options.firstSplitSize < firstMin) {
                    this.options.firstSplitSize = firstMin;
                    this._splitterEdge = true;
                } else if (this.options.firstSplitSize > (1 - secondMin)) {
                    this.options.firstSplitSize = (1 - secondMin);
                    this._splitterEdge = true;
                }

                firstPercent = this.options.firstSplitSize * 100;
                secondPercent = 100 - firstPercent;

                var start = intCssProperty(this.$firstChild, splitSizeProperty);
                setCalcProperty(this.$firstChild, splitSizeProperty, firstPercent + '% - ' + splitterSizeHalf + 'px');
                setCalcProperty(this.$secondChild, splitSizeProperty, secondPercent + '% - ' + splitterSizeHalf + 'px');
                setCalcProperty(this.$splitter, this.options.orientation === 'vertical' ? 'left' : 'top', firstPercent + '% - ' + splitterSizeHalf + 'px');
                var finish = intCssProperty(this.$firstChild, splitSizeProperty);

                //Do not trigger event if it is the same as when we started
                if(start !== finish) {
                    /**
                     * Triggered when the splitter is resized.
                     *
                     * @event orcl.splitPane#split-resize
                     */
                    this._trigger('split-resize');
                    this._splitResizeThrottle();
                }
            },

            /**
             * Collapses the first child to 0 height.
             * If pixelSize is specified, this becomes a synonym for `setSplitPixels`.
             *
             * @method orcl.splitPane#collapseFirst
             * @param {Number} [pixelSize] The size in pixels of the first split child.
             */
            collapseFirst : function (pixelSize) {
                if (!_.isUndefined(pixelSize)) {
                    this.setSplitPixels(pixelSize);
                } else if (this.options.collapseFirst) {
                    this.setSplit(0.0, true);

                    /**
                     * Triggered when the splitter is collapsed in either direction.
                     *
                     * @event orcl.splitPane#collapsed
                     * @param {Object} collapsing Has one property set depending on which child is collapsing. For example,
                     *                            if the first child is collapsed, the value will be `{ first : true }`
                     */
                    this._trigger('collapsed', { first : true });

                    /**
                     * Triggered when the first child in the splitter is collapsed.
                     *
                     * @event orcl.splitPane#collapsed-first
                     */
                    this._trigger('collapsed-first');
                }
            },

            /**
             * Collapses the second child to 0 height.
             * If pixelSize is specified, this becomes a synonym for `setSplitPixels` with a calculation based off of the
             * second child's new height.
             *
             * @method orcl.splitPane#collapseSecond
             * @param {Number} [pixelSize] The size in pixels of the second split child.
             */
            collapseSecond : function (pixelSize) {
                if (!_.isUndefined(pixelSize)) {
                    var available = intCssProperty(this.$splitContainer, this._getSplitSizeProperty());
                    this.setSplitPixels(available - pixelSize);
                } else if (this.options.collapseSecond) {
                    this.setSplit(1.0, true);
                    this._trigger('collapsed', { second : true });

                    /**
                     * Triggered when the second child in the splitter is collapsed.
                     *
                     * @event orcl.splitPane#collapsed-second
                     */
                    this._trigger('collapsed-second');
                }
            },

            /**
             * Removes the splitter widget from it's container.
             *
             * @method orcl.splitPane#_destroy
             * @private
             */
            _destroy : function () {
                var splitSizeProperty = this._getSplitSizeProperty();

                this.element.append(this.$firstChild.removeClass('split-child'));
                this.element.append(this.$secondChild.removeClass('split-child'));

                restoreOriginalCSSValue(this.$firstChild, splitSizeProperty);
                restoreOriginalCSSValue(this.$secondChild, splitSizeProperty);
                restoreOriginalCSSValue(this.$firstChild, 'max-' + splitSizeProperty);
                restoreOriginalCSSValue(this.$secondChild, 'max-' + splitSizeProperty);

                $(document).off('.sp-drag-handler-' + this._uniqueId);

                this.$splitContainer.remove();

            }

        });

    });

/**
 * A widget that extends the JQuery-UI default widget and provides
 * a date picker ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').datePicker();
 *
 * @namespace orcl.datePicker
 *
 * @tutorial constructing-widget
 */
define('oui.date/dateWidget',['oui.jquery', 'oui.date/dateField', 'oui.logger',
    'oui.jqueryui'],
    function ($, DateField, logger) {
        

        $.widget('orcl.datePicker', {

            _create : function () {
                var that = this;

                logger.debug('"calling datePicker _create"');
                this.widgetEventPrefix = this.widgetEventPrefix + '-';
                logger.info('datePicker\'s event prefix: ' + this.widgetEventPrefix);

                this._dateField = new DateField(this.options, this.element, this.element[0].value);
                this._dateField.init();

                this._dateField.on('select', function (date) {
                    that._trigger('select', null, {date : date});
                });
            },

            disable : function (disabled) {
                if (!arguments.length) {
                    return this._dateField.disabled();
                }
                this._dateField.disable(disabled);
            },

            setDate : function (date) {
                this._dateField.setDate(date);
            },

            getDate : function () {
                return this._dateField.getDate();
            },

            date : function (date) {
                if (!arguments.length) {
                    return this._dateField.date();
                }
                this._dateField.date(date);
            },

            value : function (date) {
                if (!arguments.length) {
                    return this._dateField.value();
                }
                this._dateField.value(date);
            },

            setMinDate : function (date) {
                this._dateField.setMinDate(date);
            },

            min : function (date) {
                if (!arguments.length) {
                    return this._dateField.min();
                }
                this._dateField.min(date);
            },

            setMaxDate : function (date) {
                this._dateField.setMaxDate(date);
            },

            max : function (date) {
                if (!arguments.length) {
                    return this._dateField.max();
                }
                this._dateField.max(date);
            },

            /**
             * This method is called when new options are supplied. An example of this can be:
             *
             *      $('.some_class').datePicker({min : newMin});
             *
             * @private
             * @method orcl.datePicker#_setOption
             * @param {String} key The option key to update
             * @param {String} value The value the new key should hold
             */
            _setOption : function (key, value) {
                this._dateField.option(key, value);
            },

            /**
             * Widget specific clean up and removal of element
             *
             * @private
             * @method orcl.datePicker# _destroy
             */
            _destroy : function () {
                this._dateField.destroy({ self : false});
            }

        });

    });

/**
 * A widget that extends the JQueryUI default widget and provides
 * a spinner ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').number();
 * @namespace orcl.number
 */
define('oui.spinner/spinnerWidget',['oui.jquery', 'oui.spinner/number', 'oui.jqueryui'],
    function ($, NumberField) {
        

        $.widget('orcl.number', {

            options : {
                type : 'number', //Valid options are number, currency, time
                max : null,
                min : null,
                fallback : false,
                fluid : false,
                spinner : {
                    controls : false,
                    step : 1,
                    scroll : true
                },
                formatter : {}
            },

            _create : function () {
                var value = this.element[0].value;
                this.widgetEventPrefix = this.widgetEventPrefix + '-';

                if (value === '') {
                    value = null;
                }
                this._number = new NumberField(this.options, this.element, value);
                this._number.init();
            },

            /**
             * Please use the "data" method
             *
             * @method orcl.number#setValue
             * @deprecated
             * @param {*} value
             */
            setValue : function (value) {
                this._number.setData(value);
            },

            /**
             * Please use the "data" method
             *
             * @method orcl.number#getValue
             * @deprecated
             */
            getValue : function () {
                this._number.getData();
            },

            setData : function (data) {
                this._number.setData(data);
            },

            getData : function () {
                this._number.getData();
            },

            data : function (data) {
                if (!arguments.length) {
                    return this._number.data();
                }
                this._number.data(data);
            },

            value : function (data) {
                if (!arguments.length) {
                    return this._number.value();
                }
                this._number.value(data);
            },

            increment : function (field) {
                this._number.increment(field);
            },

            decrement : function (field) {
                this._number.decrement(field);
            },

            _setOptions : function (options) {
                this._number.updateOptions(options);
            },

            _setOption : function (key, value) {
                this._number.updateOption(key, value);
            },

            disable : function (disabled) {
                this._number.disable(disabled);
            },

            _destroy : function () {
                // Destroys the underlying OUI number field
                this._number.destroy({ self : false });
            }

        });
    }
);

/**
* A widget that extends the JQuery-UI default widget and provides
* a Url testing field ui element using an input html element.
*
* Example creation with default options: (Note: `testFn` is required!)
*
*     $('selector').url({testFn: function (url, isValid) { ... }});
*
* @namespace orcl.url
*/
define('oui.url/urlWidget',['oui.jquery', 'oui.url/url', 'oui.logger', 'oui.jqueryui'],
    function ($, UrlField, logger) {
        

        $.widget('orcl.url', {

            options : UrlField.prototype.defaults(),

            _create : function () {
                var value = this.element[0].value;
                this.widgetEventPrefix = this.widgetEventPrefix + '-';

                if (value === '') {
                    value = null;
                }
                this._url = new UrlField(this.options, this.element, value);
                this._url.init();
            },

            url : function (url) {
                if (!arguments.length) {
                    return this._url._controller.url();
                }
                this._url._controller.url(url);
            },

            disable : function (disabled) {
                this._url._view.disable(disabled);
            },

            _setOption : function () {
                //TODO: implement options function
                logger.warn('url._setOption: You cannot currently change any options in this widget.');
            },

            _destroy : function () {
                this._url.destroy({ self : false});
            }

        });
    }
);

define('oui.popover/popoverWidget',['oui.jquery', 'oui.underscore', 'oui.utils/keyCode', 'oui.bootstrap'], function ($, _, keyCode) {
    

    var popoverClass = '.orcl-popover',
        $arrow = $('<div class="arrow" />'),
        $closeButton = $('<div class="orcl-popover-close" />'),
        Popover = function (el) {
            var $el = $(el);

            $el.find('.orcl-popover-outer').on('click.orcl-popover', this.toggle);
            $('html').on('click.orcl-popover', function () {
                $el.removeClass('open');
            });
        };

    function createPopoverComponents($this) {
        if ($this.find('.arrow').length === 0) {
            $this.find('.orcl-popover-outer').append($arrow.clone());
        }

        if ($this.find('.orcl-popover-close').length === 0) {
            $this.find('.orcl-popover-inner').append($closeButton.clone());
        }
    }

    function positionPopover($this) {
        var $outer = $this.find('.orcl-popover-outer'),
            $inner = $this.find('.orcl-popover-inner');

        if ($this.hasClass('temporary-position')) {
            $this.removeClass('left right temporary-position'); // We only attempt to auto position left or right
        }

        if (!($this.hasClass('left') || $this.hasClass('right') || $this.hasClass('top') || $this.hasClass('bottom'))) {
            $this.addClass(determineOrientation($this));
            $this.addClass('temporary-position');
        }

        var position = $this.offset();

        position.top = position.top - $(window).scrollTop();
        position.left = position.left - $(window).scrollLeft();

        //Right - default
        $inner.css('left', position.left + ($outer.outerWidth()));
        $inner.css('top', position.top - ($inner.outerHeight(true) / 2) + ($this.outerHeight() / 2));

        if ($this.hasClass('left')) {
            $inner.css('left', position.left - $inner.outerWidth(true));
        }

        if ($this.hasClass('top')) {
            $inner.css('left', position.left - ($inner.outerWidth(true) / 2) + ($this.outerWidth() / 2));
            $inner.css('top', position.top - ($inner.outerHeight(true)));
        }

        if ($this.hasClass('bottom')) {
            $inner.css('left', position.left - ($inner.outerWidth(true) / 2) + ($this.outerWidth() / 2));
            $inner.css('top', position.top + $outer.height());
        }
    }

    function determineOrientation($this) {
        var $arrow = $this.find('.arrow'),
            $inner = $this.find('.orcl-popover-inner'),
            rightEdge = $this.offset().left + $this.width() + $arrow.innerWidth() + $inner.innerWidth(),
            pageWidth = $(window).width();

        return (rightEdge < pageWidth) ? 'right' : 'left';
    }

    Popover.prototype = {
        constructor : Popover,
        focus : function(forward) {
            var focusable = $(this).find('a, button, input, [tabindex]:not([tabindex=-1])')
                .filter(':visible'),
                length = focusable.length;

            if($.contains(this, document.activeElement)) {
                //Go on to the next one;
                var index = focusable.index(document.activeElement);
                if(forward) {
                    index = (index + 1) % length;
                } else {
                    index = ((index - 1) % length + length) % length;
                }

                focusable.get(index).focus();
            } else if (forward && focusable.length > 0){
                focusable.get(0).focus();
            } else if (!forward && focusable.length > 0) {
                focusable.get(focusable.length - 1).focus();
            }
        },
        close : function () {
            var $this = $(this);

            if ($this.is('.disabled, :disabled')) { return; }

            closeAllPopovers();

            $this.removeClass('open');

            $this.focus();

            return false;
        },
        toggle : function () {
            var $this = $(this),
                isActive;

            if ($this.is('.disabled, :disabled')) { return; }

            isActive = $this.hasClass('open');

            closeAllPopovers();

            if (!isActive) {
                createPopoverComponents($this);
                positionPopover($this);
                $this.toggleClass('open');
            }

            $this.focus();

            return false;
        }
    };

    function closeAllPopovers() {
        $(popoverClass).each(function () {
            $(this).removeClass('open');
        });
    }

    $.fn.orclPopover = function (option) {
        return this.each(function () {
            var $this = $(this),
                data = $this.data('orclPopover');
            $(this).attr('tabindex', 0);
            if (!data) { $this.data('orclPopover', (data = new Popover(this))); }
            if (typeof option === 'string') { data[option].call($this); }
        });
    };

    $.fn.orclPopover.Constructor = Popover;

    $(document)
        .on('click.orcl-popover touchstart.orcl-popover', '.orcl-popover-inner *:not(.orcl-popover-close)', function (e) {
            e.stopPropagation();
        })
        .on('click', closeAllPopovers)
        .on('keyup', popoverClass, function(event) {
            switch(event.keyCode) {
            case keyCode.ESC:
                if($(this).hasClass('open')) {
                    event.stopPropagation();
                    Popover.prototype.close.call(this);
                }
                break;
            }
        })
        .on('keydown', popoverClass, function(event) {
            switch(event.keyCode) {
            case keyCode.TAB:
                if($(this).hasClass('open')) {
                    Popover.prototype.focus.call(this, !event.shiftKey);
                    event.preventDefault();
                    event.stopPropagation();
                }
                break;
            case keyCode.ENTER:
                Popover.prototype.toggle.call(this);
                break;
            }
        })
        .on('click.orcl-popover touchstart.orcl-popover', popoverClass, Popover.prototype.toggle);

    $(window).on('resize scroll', _.debounce(closeAllPopovers, 250, { leading : true }));
});


define('oui.file/pickerTemplate.htm', ['oui.underscore'], function (_) { return function(data){
var __t,__p='',__j=Array.prototype.join,print=function(){__p+=__j.call(arguments,'');};
__p+='<div class="orcl-file-picker"> <input type="text" disabled aria-hidden="true"/> <label class="file-label"> <span class="btn btn-picker" aria-hidden="true" title="'+
((__t=( data.locale.browse ))==null?'':__t)+
'"> <i class="pgbu-icon-small pgbu-icon-picker-dots"></i> </span> </label></div>';
return __p;
};});


define('oui.file/wrapTemplate.htm',[],function () { return '<div class="orcl-file-picker">\n</div>\n';});

define('oui.file/fileView',['oui.jquery', 'oui.underscore', 'oui.base/view', 'oui.file/pickerTemplate.htm',
'oui.template/nativeElement.htm', 'oui.file/wrapTemplate.htm'], function($, _, View, pickerTemplate, nativeTemplate, wrapTemplate) {
    

    var fvp;

    function FileView(options, element, settings) {
        var that = this;
        View.call(this, options, element, null, settings);

        this._$el.find('label').attr('id', this.id() + '-label');

        this._$el.find('[type=file]').on('change', function () {
            var fileListString;

            if(this.files) {
                fileListString = Array.prototype.slice.call(this.files).map(function(file) { return file.name; }).join(', ');
            } else {
                fileListString = Array.prototype.slice.call(this.value.split(',')).map(function(file) {
                    var index;
                    if(file.lastIndexOf('\\') !== -1) {
                        index = file.lastIndexOf('\\') + 1;
                    } else {
                        index = file.lastIndexOf('/') + 1;
                    }
                    return file.substring(index);
                }).join(', ');
            }

            that._$el.find('[type=text]')[0].value = fileListString;
            //Once the dialog is closed and the element changes make sure it retains focus
            $(this).focus();
        });

        this._$el.find('[type=file]').on('focus', function() {
            that._$el.addClass('focus');
        }).on('blur', function() {
            that._$el.removeClass('focus');
        });
    }

    FileView.prototype = Object.create(View.prototype);
    fvp = FileView.prototype;

    fvp.nativeType = function () {
        return 'file';
    };

    fvp.buildNativeElement = function(options, el) {
        var $element;

        if(_.isNull(el)) {
            $element = $(nativeTemplate({ type : this.nativeType() })).wrap(wrapTemplate).parent();
        } else if (el.jquery) {
            $element = $(el).wrap(wrapTemplate).parent();
        }

        return $element;
    };

    fvp.buildElement = function(options, el) {
        var $element = $(pickerTemplate({ locale : options.locale }));
        var fileInput;

        if(_.isNull(el)) {
            $element.find('label').prepend(nativeTemplate({ type : this.nativeType() }));
        } else if (el.jquery) {
            if(el.parent().length > 0) {
                $element.insertAfter(el);
            }
            $element.find('label').prepend(el);
        }

        fileInput = $element.find('[type=file]');
        fileInput.prop('multiple', options.multiple);

        return $element;
    };

    fvp.id = function (settings) {
        var input = this._$el.find('input[type=file]').first();

        if(!arguments.length) {
            return input.attr('id');
        }

        if(input.length > 0 && _.isUndefined(input.attr('id'))) {
            if(_.isObject(settings)) {
                input.attr('id', 'oui-' + settings.name + '-' + settings.id);
            } else if (_.isString(settings)) {
                input.attr('id', settings);
            }
        }
    };

    return FileView;
});

define('oui.file/filePicker',['oui.base/input', 'oui.file/fileView'], function(Input, FileView) {
    

    /**
     * @typedef {Object} FilePicker#Config
     * @property {Boolean} [fluid=false] Whether or not the file picker should take available widget.
     * @property {Boolean} [multiple=true] Whether or not the file picker should allow multiple files to be selected.
     */

    var fpp;
    var __WIDGET_NAME__ = 'file';

    var __DEFAULTS__ = {
        fluid : false,
        multiple : true
    };

    /**
     * The File Picker widget allows the user to select a file.
     *
     * Uses a native file input with additional styling over it to provide the look and feel.
     * The native file input must be used for security reasons. If at any point the browser believes the
     * user is being tricked into picking a file it will prevent that action from taking place.
     *
     * **Note**: File Picker does not allow setting of file programmatically due to security concerns. The browsers limit the
     * interaction of file inputs because a user may be tricked into selecting something faulty.
     *
     * @class FilePicker
     * @extends Input
     * @param {FilePicker#Config} [options]
     * @param {jQuery} [element]
     */
    function FilePicker(options, element) {
        var view;

        Input.call(this, options, element);

        view = new FileView(this._options, this._element, this._settings);

        this._view = view;

        this._setupEvents(view);
    }

    FilePicker.prototype = Object.create(Input.prototype);
    fpp = FilePicker.prototype;

    /**
     * Gets the name of the widget.
     *
     * @method FilePicker#name
     * @return {String} 'file'
     */
    fpp.name = function () {
        return __WIDGET_NAME__;
    };

    /**
     * Gets the defaults for the widget.
     *
     * @method FilePicker#defaults
     * @return {Object} The defaults for the widget.
     */
    fpp.defaults = function () {
        return __DEFAULTS__;
    };

    return FilePicker;
});

/**
* Custom styled file input.
*
* Hides the built-in file input but allows screen readers to read it as a
* file input.
*
* @namespace orcl.filePicker
* 
*/
define('oui.file/fileWidget',['oui.jquery', 'oui.file/filePicker', 'oui.jqueryui'],
    function ($, FilePicker) {
        

        $.widget('orcl.filePicker', {
            options : {
                /**
                 * @property fluid
                 * @default false
                 * @type {Boolean}
                 */
                fluid : false
            },
            _create : function () {
                this._filePicker = new FilePicker(this.options, this.element).init();
            },

            _destroy : function () {
                this._filePicker.destroy({ self : false });
            }
        });
    });

/**
* A widget that extends the JQuery-UI default widget and provides
* a ComboBox ui element using an input or select html element.
*
* Example creation with default options
*
*     $('selector').comboBox();
*
* @namespace orcl.comboBox
*/
define('oui.combo/comboBoxWidget',['oui.jquery', 'oui.combo/comboBox', 'oui.logger', 'oui.jqueryui'],
    function ($, ComboBox, logger) {
        

        $.widget('orcl.comboBox', {

            options : ComboBox.prototype.defaults(),

            _create : function () {
                var that = this;
                var value = this.options.defaultValue || this.element[0].value;

                // Don't want to pass this into the actual ComboBox widget
                delete this.options.defaultValue;

                this.widgetEventPrefix = this.widgetEventPrefix + '-';

                if (this.element.is('input') || this.element.is('select')) {
                    this._comboBox = new ComboBox(this.options, this.element);
                    this._comboBox.init();
                    this._comboBox.value({ 'value' : value });
                } else {
                    var message = 'comboBox._create: Needs to be called on either an <input> or a <select>';
                    logger.fatal(message);
                    throw new Error(message);
                }

                // Attach widget to new root ComboBox element
                this._comboBox.element().data('orclComboBox', this);

                // Pass on the valueSelected event again!
                this._comboBox.on('valueSelected', function (selected) {
                    that._trigger('valueSelected', null, selected);
                });
            },

            data : function (newValue) {
                if (!arguments.length) {
                    return this.selected();
                }
                this.selected(newValue);
            },

            selected : function (newValue) {
                if (!arguments.length) {
                    return this._comboBox.selected();
                }
                this._comboBox.selected(newValue);
            },

            values : function (newValues) {
                if (!arguments.length) {
                    return this._comboBox.values();
                }
                this._comboBox.values(newValues);
            },

            value : function (newValue) {
                if (!arguments.length) {
                    return this._comboBox.value();
                }
                this._comboBox.value(newValue);
            },

            disable : function (disabled) {
                this._comboBox.disable(disabled);
            },

            _setOption : function (key, value) {
                switch (key) {
                case 'values':
                    this._comboBox.values(value);
                    break;
                default:
                    logger.warn('url._setOption: You cannot currently change any options in this widget.');
                }
            },

            _destroy : function () {
                this._comboBox.destroy({ self : false });
            }

        });
    }
);

/**
 * A widget that extends the JQuery-UI default widget and provides
 * an editor ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').editor();
 *
 * @namespace orcl.editor
 */
define('oui.editor/editorWidget',['oui.jquery', 'oui.editor/editor', 'oui.logger',
    'oui.jqueryui'],
    function ($, Editor, logger) {
        

        $.widget('orcl.editor', {

            _create : function () {
                this.widgetEventPrefix = this.widgetEventPrefix + '-';

                this._editor = new Editor(this.options, this.element, this.element[0].value)
                    .init();
            },

            disable : function (disabled) {
                if (!arguments.length) {
                    return this._editor.disable();
                }
                this._editor.disable(disabled);
            },

            text : function (text) {
                if (!arguments.length) {
                    return this._editor.document();
                }

                this._editor.document(text);
            },

            setText : function (text) {
                logger.warn('The use of setText (and get/set methods in general) in OUI is now deprecated. Please use the combined method "text" instead.');
                logger.warn('For more information see YUIDoc on pgbulwf.use.oracle.com');
                this.text(text);
            },

            getText : function () {
                logger.warn('The use of getText (and get/set methods in general) in OUI is now deprecated. Please use the combined method "text" instead.');
                logger.warn('For more information see YUIDoc on pgbulwf.use.oracle.com');
                return this.text();
            },

            /**
             * This method is called when new options are supplied. An example of this can be:
             *
             *      $('.some_class').datePicker({min : newMin});
             *
             * @private
             * @method orcl.editor#_setOption
             * @param {String} key The option key to update
             * @param {String} value The value the new key should hold
             */
            _setOption : function (key, value) {
                this._editor.option(key, value);
            },

            /**
             * Widget specific clean up and removal of element
             *
             * @private
             * @method orcl.editor#_destroy
             */
            _destroy : function () {
                this._editor.destroy({ self : false });
            }

        });

    });

/**
 * @module valHooks
 * @private
 */
define('oui.utils/valHooks',['oui.jquery', 'oui.utils/defaults', 'oui.utils/is', 'oui.utils/dom'], function($, defaults, is, dom) {
    

    var oldTextHooks = $.valHooks.text || { get : function () {}, set : function () {} };

    /**
     * @namespace $.valHooks.text
     */
    $.valHooks.text = {
        get : function (elem) {
            if(defaults().hooks && is(elem)) {
                return dom.get(elem).data();
            }

            return oldTextHooks.get.call(this, elem);
        },
        set : function (elem, val) {
            //TODO: this picker condition is a bit janky
            //can be removed when the picker controls the value of the input
            if(defaults().hooks &&
               is(elem) && !($(elem).parent().hasClass('orcl-picker') && !$(elem).parent().hasClass('orcl-typeahead'))) {
                dom.get(elem).data(val);
                return true;
            }

            return oldTextHooks.set.call(this, elem, val);
        }
    };
});

/**
 * @module propHooks
 * @private
 */
define('oui.utils/propHooks',[
    'oui.jquery',
    'oui.utils/defaults',
    'oui.utils/is',
    'oui.utils/dom'
], function($, defaults, is, dom) {
    

    var oldDisableHooks = $.propHooks.disabled || { get : function () { return null; }, set : function () { return undefined;} };

    /**
     * @namespace $.propHooks.disabled
     * @memberof $.propHooks
     */
    $.propHooks.disabled = {
        get : function (elem) {
            var widget = dom.get(elem);
            if(defaults().hooks && is(elem) && widget != null) {
                return widget.disable() || null;
            }

            return oldDisableHooks.get.call(this, elem);
        },
        set : function (elem, val) {
            var widget = dom.get(elem);
            if(defaults().hooks && is(elem) && widget != null) {
                widget.disable(val);
                return true;
            }

            return oldDisableHooks.set.call(this, elem, val);
        }
    };
});

/**
 * Provides a single point of entry, 'oui', to require the inputs library.
 *
 * Returns an object with some useful fields:
 *
 *     define(['oui'], function (oui) {
 *         console.log(oui.version); // Would print out oui version, e.g. '1.2' or '1.3'.
 *     });
 *
 * @module oui
 **/
define('oui', [
        'oui.textField/textField',
        'oui.date/dateField',
        'oui.picker/picker',
        'oui.spinner/number',
        'oui.url/url',
        'oui.color/colorPicker',
        'oui.icon/iconPicker',
        'oui.combo/comboBox',
        'oui.tree/treePanel',
        'oui.typeAhead/typeAhead',
        'oui.editor/editor',
        'oui.utils/language',
        'oui.utils/dom',
        'oui.utils/defaults',
        'oui.utils/is',
        'oui.picker/pickerWidget',
        'oui.splitPane/splitPaneWidget',
        'oui.date/dateWidget',
        'oui.spinner/spinnerWidget',
        'oui.url/urlWidget',
        'oui.popover/popoverWidget',
        'oui.file/fileWidget',
        'oui.combo/comboBoxWidget',
        'oui.editor/editorWidget',
        'oui.utils/valHooks',
        'oui.utils/propHooks'
    ],
    function (TextField, DateField, Picker, NumberField, UrlField, ColorPicker, IconPicker, ComboBox, TreePanel, TypeAhead, Editor, language, dom, defaults, is) {
        

        /**
         * @exports oui
         */
        return {
            /**
             * The exported reference to the TextField constructor.
             *
             * See [TextField]{@link oui.TextField} for more information.
             *
             * @property TextField
             * @type {TextField}
             */
            TextField   : TextField,
            /**
             * The exported reference to the DateField constructor.
             *
             * See [DateField]{@link oui.DateField} for more information.
             *
             * @property DateField
             * @type {DateField}
             */
            DateField   : DateField,
            /**
             * The exported reference to the IconPicker constructor.
             *
             * See {@link IconPicker} for more information.
             *
             * @property IconPicker
             * @type {IconPicker}
             */
            IconPicker  : IconPicker,
            /**
             * The exported reference to the Picker constructor.
             *
             * See {@link Picker} for more information.
             *
             * @property Picker
             * @type {Picker}
             */
            Picker      : Picker,
            /**
             * The exported reference to the NumberField constructor.
             *
             * See {@link NumberField} for more information.
             *
             * @property NumberField
             * @type {NumberField}
             */
            NumberField : NumberField,
            /**
             * The exported reference to the UrlField constructor.
             *
             * See {@link UrlField} for more information.
             *
             * @property UrlField
             * @type {UrlField}
             */
            UrlField    : UrlField,
            /**
             * The exported reference to the ColorPicker constructor.
             *
             * See {@link ColorPicker} for more information.
             *
             * @property ColorPicker
             * @type {ColorPicker}
             */
            ColorPicker : ColorPicker,
            /**
             * The exported reference to the ComboBox constructor.
             *
             * See {@link ComboBox} for more information.
             *
             * @property ComboBox
             * @type {ComboBox}
             */
            ComboBox    : ComboBox,
            /**
             * The exported reference to the TreePanel constructor.
             *
             * See {@link TreePanel} for more information.
             *
             * @property TreePanel
             * @type {TreePanel}
             */
            TreePanel   : TreePanel,
            /**
             * The exported reference to the TypeAhead constructor.
             *
             * See {@link TypeAhead} for more information.
             *
             * @property TypeAhead
             * @type {TypeAhead}
             */
            TypeAhead   : TypeAhead,
            /**
             * The exported reference to the Editor constructor.
             *
             * See {@link Editor} for more information.
             *
             * @property Editor
             * @type {Editor}
             */
            Editor      : Editor,
            /**
             * A function to set the default language for use within OUI and all widget locale options.
             *
             * @method lang
             * @static
             */
            lang        : language.lang,
            /**
             * A single-use function to set the module-wide default options to use for widget constructions.
             *
             * Useful for default formatters or locale objects.
             *
             * @method defaults
             * @static
             */
            defaults    : defaults,
            /**
             *
             * Function to get the OUI widget associated with the element.
             *
             * @method get
             * @static
             *
             * @param {HTMLElement} element
             * @returns Widget
             */
            get         : dom.get,
            /**
             * Function to determine whether an element is an OUI controlled element.
             *
             * This function is to be used by [get]{@link module:oui.get} after it is determined to be an OUI element.
             *
             * @example
             *
             *  var value;
             *  var $elem = $('input:eq(0)');
             *  if(oui.isOUI($elem)) {
             *      value = oui.get($elem).value();
             *  }
             *
             * @method isOUI
             * @static
             * @param {HTMLElement} element
             * @returns {Boolean}
             */
            isOUI       : is,
            /**
             * The version of OUI.
             *
             * @property version
             * @type {String}
             */
            version     : '2.22.3-a2579c97540efadbf583c05d6e65f0c9996f18a3'
        };
    }
);



// End OUI