//!
//! Built on: Fri Nov 14 2014 08:35:45
//! Library version: 2.10.0
//! Branch: master  |  Commit: 32f8709
//!

define('primutils/i18n/Localizer',['primutils.lodash',
        'primutils.logger'
], function (_, logger) {
    

    var __noSuchKey__ = 'KEY NOT FOUND: ',
        __pathDelimiter__ = '.';

    function findKey (stringMap, path) {
        if (stringMap == null || path == null) {
            return null;
        }

        // get last part of path
        var depthIndex = path.lastIndexOf(__pathDelimiter__);
        if (depthIndex === -1) {
            // return the value
            return stringMap[path];
        }
        // move down a level in the object
        var firstLevelIndex = path.indexOf(__pathDelimiter__);
        var depth = path.slice(0, firstLevelIndex);
        return findKey(stringMap[depth], path.slice(firstLevelIndex + 1));
    }

    /**
     * Allows for retrieving localized strings from a given string map of any depth.
     *
     * Also contains helper methods for formatting strings.
     *
     * -----
     *
     * Notice for developers in UP: the `'localizer'` module in UP uses this internally, but does not expose the `'setStringMap'` method.
     * <a href="i18n.emptyStrings.html">emptyStrings</a> talks about expectations for application-level developers in providing a set of strings for primutils.
     *
     * @class i18n.Localizer
     * @module i18n
     * @param [stringMap] {Object} A map of localization "keys" to their localized representation
     * @constructor
     */
    function Localizer(stringMap) {
        logger.assert(this && this instanceof Localizer, 'Localizer must be instantiated with new');
        this.setStringMap(stringMap);
        this.setStringMap = this.setStringMap.bind(this);
        this.getString = this.getString.bind(this);
        this.getFormattedString = this.getFormattedString.bind(this);
        this.formatString = this.formatString.bind(this);
    }

    Localizer.prototype = {
        /**
         * Changes which string map is used to find localized values
         *
         * @method setStringMap
         * @param stringMap {Object} A map of localization "keys" to their localized representation
         */
        setStringMap : function (stringMap) {
            this.stringMap = stringMap;
        },

        /**
         * Retrieves a localized string from the map.
         *
         * The key is a "."-delimited string. Each part of the path represents on level of objects in the string map.
         * The map can be of relatively unlimited depth.
         *
         * For example, the key "foo.bar" would return "Hello, World!" in the following map
         *
         *     {
         *         "foo" : {
         *             "bar" : "Hello, World!"
         *         }
         *     }
         *
         * @method getString
         * @param key {String} The string to find
         * @returns {String} The localized value
         */
        getString: function (key) {
            var node = findKey(this.stringMap, key);

            if (!node) {
                logger.warn('Key: \'' + key + '\' not found.');
                return __noSuchKey__ + key;
            }

            return node;
        },

        /**
         * Formats a string with placeholders.
         *
         * A placeholder takes the form of `{n}` where `n` is a decimal number starting with 0.
         *
         *     localizer.formatString("Testing {1} or {0} stuff", "fun", "exciting") === "Testing exciting or fun stuff";
         *
         * If there are more placeholders than passed arguments, the remaining placeholders will be replaced with the
         * empty string.
         *
         * @method formatString
         * @param node {String} The format string
         * @param ... {String} Each value to replace in the format string, starting with index 0
         * @returns {String} The string with the format values replaced
         */
        formatString: function (node/*, arg1, arg2, ...*/) {
            var params = _.toArray(arguments);

            return node.replace(/\{([0-9]+)\}/g, function (match, g1) {
                var value = params[parseInt(g1, 10) + 1];
                return (value != null ? value : '');
            });
        },

        /**
         * Formats a string given by the key with placeholders.
         *
         * See <a href="#method_formatString">formatString</a> for more on placeholders.
         *
         * @method getFormattedString
         * @param key {String} The string to find
         * @param ... {String} Each value to replace in the format string, starting with index 0
         * @returns {String} The string with the format values replaced
         */
        getFormattedString: function (key/*, arg1, arg2, ...*/) {
            var node = this.getString(key);

            if (node.indexOf(__noSuchKey__) === 0) {
                return node;
            }

            var params = _.toArray(arguments);
            params[0] = node;

            return this.formatString.apply(null, params);
        }
    };

    return Localizer;

});


define('primutils/formatters/helpers/unformatIgnore',['primutils.lodash'],
        function (_) {

    

    return function (options, input) {
        var ignored = options.unformatIgnore;
        if (_.isArray(ignored)) {
            _.each(ignored, function (ignoreRule) {
                input = input.replace(ignoreRule, '');
            });
        }
        else if (_.isRegExp(ignored) || _.isString(ignored)) {
            input = input.replace(ignored, '');
        }
        return input;
    };

});


define('primutils/formatters/helpers/roundValue',[], function () {
    

    return function roundValue(options, curValue, nextDigit, wholeDigits, decimalDigits) {
        // If we end up rounding curValue, these two will be the replacement strings
        var newWholeDigits = false;
        var newDecimalDigits = false;

        if (nextDigit) {
            // Displaying decimals, rounding them up (and possibly rolling up that round to the
            //  whole values)
            if (decimalDigits && parseInt(nextDigit, 10) >= 5) {
                var oldDecimals = parseInt(decimalDigits, 10);
                var newDecimals = oldDecimals + 1;

                // e.g. if we changed from "999" to "1000", that means that the non-decimal
                //  portion of the Number should be incremented
                if (newDecimals.toString().length > oldDecimals.toString().length) {
                    newWholeDigits = String(parseInt(wholeDigits, 10) + 1);
                    newDecimalDigits = decimalDigits.replace(/./g, '0');
                }
                // e.g. if we changed from "005" to "006", we don't need to increment the
                //  non-decimal portion, but have to preserve those two zeroes in the front
                else {
                    var newDigits = String(parseInt(decimalDigits, 10) + 1);
                    while (newDigits.length < decimalDigits.length) {
                        newDigits = '0' + newDigits;
                    }
                    newDecimalDigits = newDigits;
                }
            }
            // Not displaying decimals, but the decimal portion of the value causes us to
            //  round up anyway
            else if (parseInt(nextDigit, 10) >= 5) {
                newDecimalDigits = decimalDigits.replace(/./g, '0');
                newWholeDigits = String(parseInt(wholeDigits, 10) + 1);
            }
        }

        if (newWholeDigits) {
            curValue = curValue.replace(/\d+/, newWholeDigits);
        }
        if (newDecimalDigits) {
            var decimalSeparator = options.decimalSeparator === '.' ? '\\.' : options.decimalSeparator;
            var reDecimal = new RegExp('(' + decimalSeparator + ')(\\d+)');
            curValue = curValue.replace(reDecimal, '$1' + newDecimalDigits);
        }

        return curValue;
    };

});


define('primutils/formatters/baseFormatter',['primutils.lodash',
        'primutils.logger'
], function(_, logger) {
    

    function maybeWrap(fn, wrapperName, opt) {
        return opt[wrapperName] ? _.wrap(fn, opt[wrapperName]) : fn;
    }

    return function (defaultOptions, serializeFn, unserializeFn, formatFn,
                     unformatFn, validateFn, utcCheck) {

        var formatterFactory = function formatterFactory(newOptions) {
            var formatter = {};

            var options = _.defaults({}, newOptions || {}, defaultOptions);

            formatter.withOptions = function (newOptions) {
                var newFormatterOptions = _.defaults({}, newOptions || {}, options);
                return formatterFactory(newFormatterOptions);
            };

            var wrappedSerialize = maybeWrap(serializeFn, 'serializeWrapper', options);
            formatter.serialize = function (val) {
                return wrappedSerialize(options, val);
            };

            var wrappedUnserialize = maybeWrap(unserializeFn, 'unserializeWrapper', options);
            formatter.unserialize = function (valStr) {
                return wrappedUnserialize(options, valStr);
            };

            formatter.unserializeAndFormat = function (valStr) {
                return wrappedFormat(options, wrappedUnserialize(options, valStr));
            };

            var wrappedFormat = maybeWrap(formatFn, 'formatWrapper', options);
            formatter.format = function (val, formatOptions) {
                return wrappedFormat(options, val, formatOptions);
            };

            var wrappedUnformat = maybeWrap(unformatFn, 'unformatWrapper', options);
            formatter.unformat = function (valStr, unformatOptions) {
                return wrappedUnformat(options, valStr, unformatOptions);
            };

            var wrappedValidate = maybeWrap(validateFn, 'validateWrapper', options);
            formatter.validate = function (valStr, isUTC) {
                if (utcCheck && isUTC) {
                    logger.warn('The isUTC flag is no longer respected.  Ignoring.');
                }

                return wrappedValidate(options, valStr);
            };

            return formatter;
        };

        formatterFactory.withOptions = function (newOptions) {
            return formatterFactory(newOptions);
        };

        formatterFactory.setDefaultOptions = function setDefaultOptions(newOptions) {
            defaultOptions = _.defaults({}, newOptions || {}, defaultOptions);
        };

        formatterFactory.serialize = function (val, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            var wrappedFn = maybeWrap(serializeFn, 'serializeWrapper', options);
            return wrappedFn(options, val);
        };

        formatterFactory.unserialize = function (valStr, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            var wrappedFn = maybeWrap(unserializeFn, 'unserializeWrapper', options);
            return wrappedFn(options, valStr);
        };

        formatterFactory.format = function (val, newOptions, formatOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            var wrappedFn = maybeWrap(formatFn, 'formatWrapper', options);
            return wrappedFn(options, val, formatOptions);
        };

        formatterFactory.unformat = function (valStr, newOptions, unformatOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            var wrappedFn = maybeWrap(unformatFn, 'unformatWrapper', options);
            return wrappedFn(options, valStr, unformatOptions);
        };

        formatterFactory.validate = function (valStr, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            var wrappedFn = maybeWrap(validateFn, 'validateWrapper', options);
            return wrappedFn(options, valStr);
        };

        formatterFactory.unserializeAndFormat = function (valString, newOptions) {
            return formatterFactory.format(formatterFactory.unserialize(valString, newOptions), newOptions);
        };

        return formatterFactory;

    };

});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The options available for number formatters are documented in the Properties tab, below.
 *
 * The number formatter uses a custom format flag specification, best illuminated by examples:
 *
 *     define(['primutils'], function(primutils) {
 *         var numberFormatter = primutils.numberFormatter;
 *
 *         formatter = numberFormatter({
 *             format : '0' //This says to print out a number
 *                          //with no decimal separator, currency symbol,
 *                          //thousands separator and no digits following the decimal
 *         });
 *
 *         formatter.format(0); //Returns '0'
 *         formatter.format(100); //Returns '100'
 *         formatter.format(100.00); //Also returns '100'
 *
 *         //Currency
 *         currencyFormatter = numberFormatter({
 *             format : '$0,0.00',
 *             symbol : '$'
 *         });
 *
 *         currencyFormatter.format(0); //Returns '$0.00'
 *         currencyFormatter.format(10000.01); //Returns '$10,000.01'
 *
 *         //Append suffix to number for presenting hours and days
 *         naiveDurationFormatter = numberFormatter({
 *             format : '0h'
 *         });
 *
 *         naiveDurationFormatter.format(0); //Returns '0h'
 *         naiveDurationFormatter.format(7); //Returns '7h'
 *     });
 *
 * @uses formatters.FormatterTheory
 * @class formatters.numberFormatter
 * @module formatters
 */

define('primutils/formatters/numberFormatter',['primutils.lodash',
        'primutils.logger',
        'primutils/formatters/helpers/unformatIgnore',
        'primutils/formatters/helpers/roundValue',
        'primutils/formatters/baseFormatter'],

        function(_, logger, unformatIgnore, roundValue, baseFormatter) {

    

    var defaults = {
        /**
         * The maximum number of decimals that a user can enter as decimals.  For example,
         *  if the value is 5, the user typing in "5.555555" will be unformatted to the number
         *  5.55556
         *
         * @default 5
         * @property workableDecimalPlaces {Number}
         * @optional
         */
        workableDecimalPlaces: 5,
        /**
         * Whether or not digits after the decimal separator are optional.  If true and using a
         *  format of '0.00', for example, 5 <=> '5' but 5.1 <=> '5.10'
         *
         * @default false
         * @property decimalsOptional {Boolean}
         * @optional
         */
        decimalsOptional: false,
        /**
         * The character to use as a decimal separator.  In the US, for example, this is a period.
         *
         * @property decimalSeparator {String}
         * @optional
         * @default '.'
         */
        decimalSeparator : '.',
        /**
         * The character to use to separate groups of digits.  In the US, for example, a comma is
         *  used for the number one thousand: '1,000'.
         *
         * @property thousandsSeparator {String}
         * @optional
         * @default ','
         */
        thousandsSeparator : ',',
        /**
         * The currency symbol, if any, to use.  Unused by default.
         *
         * @property symbol {String}
         * @optional
         */
        symbol : '',
        /**
         * A symbol to use to indicate negative numbers.  In the US, for example, this is a '-'
         *  sign, where negative one hundred gets written as: '-100'
         *
         * Only used with negative numbers, when no negativeFormat option has been specified.
         *
         * @property negativeSymbol {String}
         * @optional
         * @default '-'
         */
        negativeSymbol : '-',
        /**
         * The format flags used for displaying numbers as positive Strings to the user.  See
         *  examples in the class description for more information.
         *
         * See also {{#crossLink "formatters.numberFormatter/negativeFormat:property"}}{{/crossLink}}
         *
         * @property format
         * @type String
         * @optional
         * @default '0'
         */
        format : '0',
        /**
         * The format flags used for displaying negative numbers as Strings to the user.
         *
         * See also {{#crossLink "formatters.numberFormatter/format:property"}}{{/crossLink}}
         *
         * If not specified, will default to the 'format' option.
         *
         * When specified, an explicit '-' sign will not be added to the number.
         *
         * As examples.  If using a format of '0,0.00' and no negativeFormat,
         *  0.5 <=> '0.50' and -0.5 <=> '-0.50'.
         *  If using a format of '0,0.00' and a negativeFormat of '(0,0.00)',
         *  0.5 <=> '0.50' and -0.5 <=> '(0.50)'.
         *
         * @property negativeFormat
         * @type String
         * @optional
         **/
        negativeFormat: null,
        /**
         * A multipler to use when formatting/unformatting values.  If the multiplier is 100 and
         *  using a format of '0,0.0$' and a symbol of '%', for example, 0.5 <=> '50.0%' and
         *  1.0 <=> '100.0%'.
         *
         * @property formattedMultiplier
         * @type Number
         * @optional
         */
        formattedMultiplier: undefined,

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    var _maxPrecision_ = null;
    function digitFlag(value, group, decimal, printDecimals) {
        if (!group && !decimal) {
            //First 0 - print all numbers before decimal
            if(value >= 0) {
                return parseInt(value, 10);
            } else {
                return parseInt(value, 10).toString().substring(1);
            }
        } else if (decimal && printDecimals) {
            //These are the 0s after the decimal separator
            if(_maxPrecision_ && _maxPrecision_.length > 0){
                var nextDigit = _maxPrecision_.slice(0, 1);
                _maxPrecision_ = _maxPrecision_.substring(1);
                return nextDigit;
            } else {
                return '0';
            }
        }

        return '';
    }

    function serialize(options, val) {
        if (val == null) {
            return null;
        }

        if (!_.isNumber(val)) {
            logger.error('serialize: transfer form is a Number');
            return null;
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return null;
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return null;
        }

        var modifier = Math.pow(10, options.workableDecimalPlaces);
        return Math.round(modifier * val) / modifier;
    }

    function unserialize(options, val) {
        if (val == null || val === '') {
            return null;
        }

        if (!_.isNumber(val)) {
            logger.error('unserialize: transfer form is a Number');
        }

        var modifier = Math.pow(10, options.workableDecimalPlaces);
        return Math.round(modifier * val) / modifier;
    }

    function groupThousands(options, formattedValue, shouldGroup) {
        if (shouldGroup) {
            if (options.thousandsSeparator === '\\.') {
                options.thousandsSeparator = '.';
            }
            var portionBeforeDecimal = formattedValue.split(options.decimalSeparator)[0];
            portionBeforeDecimal = portionBeforeDecimal.replace(/(\d)(?=(\d{3})+(?!\d))/g, '$1' + options.thousandsSeparator);
            var portionAfterDecimal = formattedValue.split(options.decimalSeparator)[1];

            formattedValue = portionBeforeDecimal;
            if (portionAfterDecimal) {
                formattedValue += options.decimalSeparator + portionAfterDecimal;
            }
        }

        return formattedValue;
    }

    var reDecimalsInFormat = /\.0+/;
    function countDecimalsInFormat(format) {
        var countInFormat = reDecimalsInFormat.exec(format);
        if (countInFormat) {
            return countInFormat[0].length - 1;
        }
        return 0;
    }

    function formatNonTruncated(options, usedFormat, value) {
        var formattedValue = '', includedNumber = false;
        var valDecimals = (String(value).split('.')[1] || '').length;
        var minDecimals = countDecimalsInFormat(usedFormat);

        if(!_.isNull(usedFormat)) {
            //Use string format to parse
            _.each(usedFormat, function(ch) {
                switch(ch) {
                case '0' :
                    if (!includedNumber) {
                        includedNumber = true;
                        var stringified = value.toFixed(Math.max(valDecimals, minDecimals));
                        if (valDecimals > 0 || minDecimals > 0) {
                            stringified = stringified.replace('.', options.decimalSeparator);
                        }
                        formattedValue += stringified;
                    }
                    break;
                case '$' :
                    //print out currency symbol
                    formattedValue += options.symbol;
                    break;
                case '.':
                case ',':
                    break;
                default :
                    //just display the character we were given
                    formattedValue += ch;
                }
            });

            formattedValue = groupThousands(options, formattedValue, true);

            // Insert a negative symbol if we should have one, but didn't insert one already
            if (value < 0 && options.negativeFormat == null && formattedValue.indexOf(options.negativeSymbol) === -1) {
                formattedValue = formattedValue.replace(/(\d)/, options.negativeSymbol + '$1');
            }
        }

        return formattedValue;
    }

    function format(options, value, formatOptions) {
        if(value == null) {
            return '';
        }

        if(!_.isNumber(value)) {
            logger.fatal('format: workable form for numberFormatter is a Number');
            throw new Error('Must be a number');
        }

        if (_.isNaN(value)) {
            logger.error('numberFormatter should not be passed NaN');
            return '';
        }
        if (value === Number.POSITIVE_INFINITY || value === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return '';
        }

        if (options.formattedMultiplier != null) {
            value = value * options.formattedMultiplier;
        }

        var usedFormat = options.format;
        if (value < 0 && options.negativeFormat != null) {
            usedFormat = options.negativeFormat;
        }

        if (formatOptions && !formatOptions.truncate) {
            return formatNonTruncated(options, usedFormat, value);
        }

        var valueHasDecimals = Math.floor(value) !== value;
        var decimalsInFormat = countDecimalsInFormat(usedFormat);

        _maxPrecision_ = '';
        if (decimalsInFormat !== 0) {
            //This will handle rounding when you are given .99999999999999
            var valueFixed1 = value.toFixed(decimalsInFormat + 1);
            _maxPrecision_ = valueFixed1.split('.')[1];
            value = parseFloat(valueFixed1);
        }
        else if (valueHasDecimals) {
            var valueFixed2 = value.toFixed(1);
            _maxPrecision_ = valueFixed2.split('.')[1];
            value = parseInt(valueFixed2, 10);
        }

        var printDecimals = !options.decimalsOptional || parseInt(_maxPrecision_, 10) > 0;
        var formattedValue = '', group = false, decimal = false;
        var wholeDigits = '';
        var decimalDigits = '';

        if(!_.isNull(usedFormat)) {
            //Use string format to parse
            _.each(usedFormat, function(ch) {
                switch(ch) {
                case '0' :
                    //print out all digits
                    var nextDigit = digitFlag(value, group, decimal, printDecimals);
                    formattedValue += nextDigit;
                    if (decimal) {
                        decimalDigits += nextDigit;
                    }
                    else {
                        wholeDigits += nextDigit;
                    }
                    break;
                case '.' :
                    //print decimal separator
                    if (printDecimals) {
                        formattedValue += options.decimalSeparator;
                    }
                    decimal = true;
                    break;
                case '$' :
                    //print out currency symbol
                    formattedValue += options.symbol;
                    break;
                case ',' :
                    //group all digits with thousands separator
                    group = true;
                    break;
                default :
                    //just display the character we were given
                    formattedValue += ch;
                }
            });

            var roundingDigit = _maxPrecision_ && _maxPrecision_[0];
            formattedValue = roundValue(options, formattedValue, roundingDigit, wholeDigits, decimalDigits);
            formattedValue = groupThousands(options, formattedValue, group);

            // Insert a negative symbol if we should have one, but didn't insert one already
            if (value < 0 && options.negativeFormat == null) {
                formattedValue = formattedValue.replace(/(\d)/, options.negativeSymbol + '$1');
            }
        }

        return formattedValue;
    }

    function baseReplace(options, valStr) {
        if (options.thousandsSeparator === '.') {
            options.thousandsSeparator = '\\.';
        }

        return valStr.replace(new RegExp(options.thousandsSeparator, 'g'), '')
                              .replace(options.decimalSeparator, '.')
                              .replace(options.symbol, '');
    }

    function patternFromString (options, valStr ) {
        var pattern = valStr.replace(/\d+/g, '0')
            .replace(options.symbol, '$')
            .replace(options.thousandsSeparator, ',')
            .replace(options.decimalSeparator, '.')
            .replace(/(?:0,)+/, '0,');
        return pattern;
    }

    function reducedPattern (options, pattern) {
        if (options.negativeFormat == null) {
            pattern = pattern.replace(/-/, '');
        }
        return pattern.replace(/,/g, '')
            .replace('$', '')
            .replace('.', '')
            .replace('+', '')
            .replace(/ /g, '') // TODO do we really want to ignore spaces?
            .replace(/\d+/g, '0');
    }

    function fuzzyMatch ( options, patternToMatchAgainst, pattern ) {
        if (patternToMatchAgainst == null) {
            return false;
        }
        return patternToMatchAgainst === pattern ||
            reducedPattern(options, patternToMatchAgainst) === reducedPattern(options, pattern);
    }

    function unformat(options, valStr, unformatOptions) {
        if (valStr == null || valStr === '') {
            return null;
        }
        valStr = unformatIgnore(options, valStr);
        if (valStr === '') {
            return null;
        }

        // TODO good idea, or match behavior with validate() function?
        var rawNumberString = baseReplace(options, valStr).replace(/[^\d\.\-\+]/g, '');
        if (rawNumberString === '') {
            return NaN;
        }

        var baseValue = parseFloat(rawNumberString);

        if (options.formattedMultiplier != null) {
            baseValue = baseValue / options.formattedMultiplier;
        }

        if (options.negativeFormat != null &&
                fuzzyMatch(options, options.negativeFormat, patternFromString(options, valStr))) {
            baseValue = -Math.abs(baseValue);
        }

        if (!unformatOptions || unformatOptions.truncate) {
            var modifier = Math.pow(10, options.workableDecimalPlaces);
            return Math.round(modifier * baseValue) / modifier;
        }
        else {
            return baseValue;
        }
    }

    function countMatches(str, toMatch) {
        var match = str.match(toMatch);
        return match ? match.length : 0;
    }

    function validate (options, valueString) {
        if (valueString == null || valueString === '') {
            return true;
        }
        valueString = unformatIgnore(options, valueString);
        if (valueString === '') {
            return true;
        }

        var hasMultipleSymbols = options.symbol && countMatches(valueString, new RegExp('\\' + options.symbol, 'g')) > 1;
        var checkDollarChar = options.symbol !== '$';
        var hasDollarCharacter = countMatches(valueString, /\$/g) > 0;
        if (hasMultipleSymbols || (checkDollarChar && hasDollarCharacter)) {
            return false;
        }

        var unlocalizedNumericString = baseReplace(options, valueString).replace(/[^\d\.\-\+]/g, '');
        // Only allow one decimal
        //  TODO: same for '-', '+' chars?  user could sneak them on at the end if they wanted
        if (countMatches(unlocalizedNumericString, /\./g) > 1) {
            return false;
        }

        var parsedVal = parseFloat(unlocalizedNumericString);
        if (!_.isNumber(parsedVal) || _.isNaN(parsedVal)) {
            return false;
        }

        var enteredPattern = patternFromString(options, baseReplace(options, valueString));
        var checkNegativeFormat = options.negativeFormat != null;
        var matchesNegativeFormat = fuzzyMatch(options, options.negativeFormat, enteredPattern);
        var matchesPositiveFormat = fuzzyMatch(options, options.format, enteredPattern);
        if (matchesPositiveFormat || (checkNegativeFormat && matchesNegativeFormat)) {
            return true;
        }

        if (parsedVal < 0 && fuzzyMatch(options, '-0', enteredPattern)) {
            return true;
        }

        return false;
    }

    return baseFormatter(defaults, serialize, unserialize, format, unformat, validate);

});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The date formatter uses [MomentJS format flags](http://momentjs.com/docs/#/displaying/format/)
 *
 * Date formatters have three primary configuration options; format, unserializeFormat, and
 *  serializedFormat.  These options are documented in the Properties tab, below.
 *
 * *Note*: for developers in the PGBU, unserializeFormat and serializeFormat are not specified
 *  by most developers.  Instead, import the 'userFormatters' module, and if you need to override
 *  the display format option, create a new formatter using the 'withOptions' function.
 *  The documentation in the UP project has more information.
 *
 * @uses formatters.FormatterTheory
 * @class formatters.dateFormatter
 * @module formatters
 */
define('primutils/formatters/dateFormatter',['primutils.moment',
        'primutils.logger',
        'primutils/formatters/helpers/unformatIgnore',
        'primutils/formatters/baseFormatter'
], function(moment, logger, unformatIgnore, baseFormatter) {
    

    var defaults = {
        /**
         * The format flags used for displaying dates as Strings to the user.  Please note that
         *  the date formatter uses
         *  [MomentJS format flags](http://momentjs.com/docs/#/displaying/format/).
         *
         * @property format
         * @type String
         * @optional
         * @default 'MM-DD-YYYY'
         */
        format: 'MM-DD-YYYY',

        // TODO bad format, probably, ever
        // Example: 2013-05-14T00:00:00.000-0400
        /**
         * The format used for unserialization of dates sent by the server.
         *
         * @property unserializeFormat
         * @type String
         * @optional
         * @default 'YYYY-MM-DD[T]HH:mm:ss.ZZ'
         */
        unserializeFormat: 'YYYY-MM-DD[T]HH:mm:ss.ZZ',

        /**
         * The format used for serialization of dates sent to the server.
         *
         * @property serializeFormat
         * @type String
         * @optional
         * @default 'YYYY-MM-DD[T]HH:mm:ss.ZZ'
         */
        serializeFormat: 'YYYY-MM-DD[T]HH:mm:ss.ZZ',

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    var _momentTester = moment();
    function checkFauxVariant(dt) {
        return dt.__pgbu_original_format_fn != null;
    }

    function checkMoment(dt) {
        if (dt != null && moment.isMoment(dt)) {
            return true;
        }
        else {
            logger.fatal('dateFormatter only works with Moment.js objects');
            throw new Error('dateFormatter only works with Moment.js objects');
        }
    }

    function serialize(options, dt) {
        if (dt == null) {
            return null;
        }

        checkMoment(dt);

        if (!dt.isValid()) {
            logger.error('dateFormatter should only be passed Moment objects that pass isValid()');
            return null;
        }

        // ...wut?  Allows dateFormatter to serialize properly with FAUX variant
        if (checkFauxVariant(dt)) {
            return dt.__pgbu_original_format_fn(options.serializeFormat);
        }
        else {
            return dt.format(options.serializeFormat);
        }
    }

    function unserialize(options, serializedDateString) {
        if (serializedDateString == null || serializedDateString === '') {
            return null;
        }

        return moment(serializedDateString, options.unserializeFormat);
    }

    function format(options, dt) {
        if (dt == null) {
            return '';
        }

        checkMoment(dt);

        if (!dt.isValid()) {
            logger.error('dateFormatter should only be passed Moment objects that pass isValid()');
            return '';
        }

        return dt.format(options.format);
    }

    var FAUX_PREFIX = 'X\'X_';
    var FAUX_SUFFIX = '_X\u00A5X\u59CB';
    var reFaux = new RegExp(FAUX_PREFIX + '|' + FAUX_SUFFIX, 'g');
    function correctFauxVariant(ds) {
        return ds.replace(reFaux, '');
    }

    function genForgivingZeroFormat(format) {
                if (format.match(/MMM|DDD|ddd|\[.*(MM|DD|dd|ww|HH|hh|mm|ss).*\]/)) {
            logger.warn('Format given for dates includes a pattern which may not work correctly with the "forgiveLeadingZeros" option.  File a request if fixing this is necessary and you\'re seeing issues.');
        }
                return format.replace(/MM(?!M)|DD(?!D)|dd(?!d)|ww|HH|hh|mm|ss/g, function (match) {
            return match[0];
        });
    }

    function unformat(options, dateString) {
        if (dateString == null || dateString === '') {
            return null;
        }

        dateString = unformatIgnore(options, dateString);
        if (checkFauxVariant(_momentTester)) {
            dateString = correctFauxVariant(dateString);
        }

        if (dateString === '') {
            return null;
        }

        var dt = moment(dateString, options.format, true);
        if (dt.isValid()) {
            return dt;
        }
        else if (options.forgiveLeadingZeros) {
            var format = genForgivingZeroFormat(options.format);
            dt = moment(dateString, format, true);
            if (dt.isValid()) {
                return dt;
            }
        }

        return false;
    }

    function validate(options, dateString) {
        if (dateString == null || dateString === '') {
            return true;
        }

        dateString = unformatIgnore(options, dateString);
        if (checkFauxVariant(_momentTester)) {
            dateString = correctFauxVariant(dateString);
        }

        if (dateString === '') {
            return true;
        }

        var dt = moment(dateString, options.format, true);
        if (dt.isValid()) {
            return true;
        }
        else if (options.forgiveLeadingZeros) {
            var format = genForgivingZeroFormat(options.format);
            dt = moment(dateString, format, true);
            if (dt.isValid()) {
                return dt;
            }
        }

        return false;
    }

    return baseFormatter(defaults, serialize, unserialize, format, unformat, validate, true);

});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The options for duration formatters are documented in the Properties tab below.
 *
 * The duration formatter formats durations according to how the PGBU deals with durations.
 *  Examples:
 *
 *     define(['primutils'], function(primutils) {
 *         var durationFormatter = primutils.durationFormatter;
 *
 *         durationFormatter.format(5); //Returns '5h'
 *         durationFormatter.format(100); //Returns '100h'
 *
 *         durationFormatter.unformat('1y 1m 1w 1d 1h'); //Returns 2221
 *         durationFormatter.unformat('1.2y'); //Returns 2400
 *
 *         durationFormatter.validate('1h'); //Returns true
 *         durationFormatter.validate('pie'); //Returns false
 *     });
 *
 * @uses formatters.FormatterTheory
 * @class formatters.durationFormatter
 * @module formatters
 */
define('primutils/formatters/durationFormatter',['primutils.lodash',
        'primutils.logger',
        'primutils/formatters/helpers/unformatIgnore',
        'primutils/formatters/baseFormatter'
], function(_, logger, unformatIgnore, baseFormatter) {
    

    var defaults = {
        // PGBU application settings
        /**
         * The units available to the duration formatter.
         *
         * Units are broken down by their MomentJS short names:
         * * n - Minutes
         * * h - Hours
         * * d - Days
         * * w - Weeks
         * * m - Months
         * * y - Years
         *
         * Each unit defines its `scale` relative to hours and a `sub` which is the child unit
         *
         *     units : {
         *         h : {
         *             scale : 1, // 1 hour == 1 hour
         *             sub : 'n'  // Hours break down into minutes
         *         },
         *         d : {
         *             scale : 8, // 1 day == 8 hours (workday)
         *             sub : 'h'  // Days break down into minutes
         *         }
         *     }
         *
         * @property units
         * @type Object
         * @optional
         * @default 8 hours/day, 40 hours/week, 172 hours/month, 2000 hours/year
         */
        units: {
            n: { // minutes
                scale: 1 / 60,
                sub: null
            },
            h: { // hours
                scale: 1,
                sub: 'n'
            },
            d: { // days
                scale: 8,
                sub: 'h'
            },
            w: { // weeks
                scale: 40,
                sub: 'd'
            },
            m: { // months
                scale: 172,
                sub: 'w'
            },
            y: { // years
                scale: 2000,
                sub: 'm'
            }
        },
        /**
         * @property flexibleScale
         * @type Boolean
         * @optional
         * @default false
         */
        flexibleScale : false,
        // PGBU user pref
        /**
         * Which unit is the base unit to display values in
         *
         * @property showIn
         * @type String
         * @optional
         * @default 'h' (Hours)
         */
        showIn: 'h',              // any of the above units
                                  //  PGBU h/d/w/m/y only

        /**
         * Should sub units be shown
         *
         * @property showSubUnits
         * @type Boolean
         * @optional
         * @default false
         */
        showSubUnits: false,
        /**
         * Should units be shown as abbreviations
         *
         * @property showAbbr
         * @type Boolean
         * @optional
         * @default true
         */
        showAbbr: true,
        /**
         * The number of decimal places to show
         *
         * @property decimalsToShow
         * @type Number (0-2)
         * @optional
         * @default 1
         */
        decimalsToShow: 1,
        /**
         * The character to use to separate groups of digits.  In the US, for example, a comma is
         *  used for the number one thousand: '1,000'.
         *
         * @property thousandsSeparator {String}
         * @optional
         * @default ','
         */
        thousandsSeparator: ',',
        /**
         * The character to use as a decimal separator.  In the US, for example, this is a period.
         *
         * @property decimalSeparator {String}
         * @optional
         * @default '.'
         */
        decimalSeparator: '.',
        /**
         * The maximum number of decimals that a user can enter as decimals in the base unit.  For example,
         *  if the value is 5, the user typing in "5.555555" will be unformatted to the number
         *  5.55556
         *
         * @default 5
         * @property workableDecimalPlaces {Number}
         * @optional
         */
        workableDecimalPlaces: 6,
        /**
         * Should all non-zero sub units be shown
         *
         * @property showAllNonzeroSubUnits
         * @type Boolean
         * @optional
         * @default false
         */
        showAllNonzeroSubUnits: false,

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    // TODO add validation options on create; showSubUnits: false + showAllNonzeroSubUnits: true
    //  makes no sense.  Probably other such things we could do as well, like check
    //  showIn/flexibleScale validity

    function serialize(options, val) {
        if (val == null) {
            return null;
        }

        if (!_.isNumber(val)) {
            logger.error('serialize: transfer form is a Number');
            return null;
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return null;
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return null;
        }

        return val;
    }

    function unserialize(options, valStr) {
        if (valStr == null || valStr === '') {
            return null;
        }

        if (!_.isNumber(valStr)) {
            logger.error('unserialize: transfer form is a Number');
        }

        return valStr;
    }

    function fixDecimal(val, places) {
        var multiplier = Math.pow(10, places);
        return (Math.round(multiplier * val) / multiplier).toFixed(places);
    }

    function stringy(options, value) {
        var base = String(value);

        var portionBeforeDecimal = base.split('.')[0];
        portionBeforeDecimal = portionBeforeDecimal.replace(/(\d)(?=(\d{3})+(?!\d))/g, '$1' + options.thousandsSeparator);
        var portionAfterDecimal = base.split('.')[1];

        var grouped = portionBeforeDecimal;
        if (portionAfterDecimal) {
            grouped += options.decimalSeparator + portionAfterDecimal;
        }
        return grouped;
    }

    function format(options, val) {
        if (val == null) {
            return '';
        }

        if (!_.isNumber(val)) {
            logger.fatal('format: workableForm for durationFormatter is a Number');
            throw new Error('Must be a number');
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return '';
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return '';
        }


        if (options.thousandsSeparator === '\\.') {
            options.thousandsSeparator = '.';
        }

        var showIn = options.showIn;
        if(options.flexibleScale){
            showIn = null;
            var descendingUnitPairs = _.sortBy(_.pairs(options.units), function(unitSetting){
                return -unitSetting[1].scale;
            });
            var bestFit = _.find(descendingUnitPairs, function(setting){
                return Math.floor(val / setting[1].scale) > 0;
            });
            if(bestFit != null){
                showIn = bestFit[0];
            }
            else {
                showIn = _.last(descendingUnitPairs)[0];
            }
        }

        var units = options.units;
        if (!units[showIn]) {
            var error = 'No unit specified for the major scale ' + showIn + ', did you forget to specify the showIn or flexileScale option?';
            logger.fatal(error);
            throw new Error(error);
        }

        var majorScale = units[showIn].scale;
        var scaledValue = val / majorScale;
        if (!options.showAbbr) {
            return stringy(options, fixDecimal(scaledValue, options.decimalsToShow));
        }
        else if (!options.showSubUnits) {
            return stringy(options, fixDecimal(scaledValue, options.decimalsToShow)) + showIn;
        }
        else if (options.showAllNonzeroSubUnits) {
            var portionVal = Math.floor(scaledValue);
            var portionUnit = showIn;
            var portionScale = majorScale;
            var nextPortionVal = val - (portionVal * portionScale);
            var ret = '';

            if (units[portionUnit].sub) {
                ret = stringy(options, portionVal) + portionUnit;
            }
            else {
                ret = stringy(options, fixDecimal(scaledValue, options.decimalsToShow)) + portionUnit;
            }
            while (units[portionUnit].sub && nextPortionVal > 0) {
                portionUnit = units[portionUnit].sub;
                portionScale = units[portionUnit].scale;
                portionVal = Math.floor(nextPortionVal / portionScale);
                nextPortionVal -= (portionVal * portionScale);

                if (portionVal > 0) {
                    if (units[portionUnit].sub) {
                        ret += ' ' + stringy(options, portionVal) + portionUnit;
                    }
                    else {
                        ret += ' ' + stringy(options, fixDecimal(portionVal, options.decimalsToShow)) + portionUnit;
                    }
                }
            }
            return ret;
        }
        else {
            var mathFn = (scaledValue >= 0) ? Math.floor : Math.ceil;
            var majorVal = mathFn(scaledValue);
            var result = stringy(options, mathFn(majorVal)) + showIn;

            if (!units[showIn].sub) {
                return stringy(options, fixDecimal(scaledValue, options.decimalsToShow)) + showIn;
            }

            if (majorVal !== scaledValue) {
                var subKey = units[showIn].sub;
                var minorScale = units[subKey].scale;
                var subVal = Math.abs(((scaledValue - majorVal) * majorScale) / minorScale);
                result += ' ' + stringy(options, fixDecimal(subVal, options.decimalsToShow)) + subKey;
            }

            return result;
        }
    }

    function clipResult(val, places) {
        var modifier = Math.pow(10, places);
        return Math.round(modifier * val) / modifier;
    }

    function unformat(options, valStr) {
        if (valStr == null || valStr === '') {
            return null;
        }

        if (!_.isString(valStr)) {
            logger.fatal('unformat: typed form for all formatters is a String');
            throw new Error('Must be a string');
        }

        if (!options.showAbbr) {
            var val = Number(valStr);
            var majorScale = options.units[options.showIn].scale;
            return clipResult(val * majorScale, options.workableDecimalPlaces);
        }

        if (options.thousandsSeparator === '.') {
            options.thousandsSeparator = '\\.';
        }

        valStr = valStr.trim()
                       .replace(new RegExp(options.thousandsSeparator, 'g'), '')
                       .replace(options.decimalSeparator, '.');

        valStr = unformatIgnore(options, valStr);

        if (valStr === '') {
            return null;
        }

        var reNextChunk = /([+-]?[\d\.]+)\s*([^\d\s]*)/g;
        var foundChunk = NaN;
        var match = false;
        var baseValue = 0;
        var previousKeys = [];
        while ((match = reNextChunk.exec(valStr)) != null) {
            foundChunk = true;
            var partVal = Number(match[1]);
            var partKey = match[2];

            //use the default type if none is set
            if(partKey.length < 1){
                partKey = options.showIn;
            }

            if (previousKeys.indexOf(partKey) !== -1) {
                return NaN;
            }
            previousKeys.push(partKey);

            if (!_.isNaN(partVal) && options.units[partKey]) {
                baseValue += partVal * (options.units[partKey].scale);
            }
            else {
                return NaN;
            }
        }

        return foundChunk && clipResult(baseValue, options.workableDecimalPlaces);
    }

    function validate(options, valStr) {
        if (valStr == null || valStr === '') {
            return true;
        }

        if (!options.allowNegative && valStr.indexOf('-') !== -1 ) {
            return false;
        }

        var wrappedUnformat = options.unformatWrapper ? _.wrap(unformat, options.unformatWrapper) : unformat;
        var unformatted = wrappedUnformat(options, valStr);

        if (unformatted == null) {
            return true;
        }
        if (_.isNaN(wrappedUnformat(options, valStr))) {
            return false;
        }

        return true;
    }

    return baseFormatter(defaults, serialize, unserialize, format, unformat, validate);

});


define('primutils/validators/dateValidator',['primutils.lodash', 'primutils.moment'], function (_, moment) {
    

    return function (opts) {

        var defaults = {
            minValue : null,
            maxValue : null
        }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'Invalid date';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE Date must be between ' + options.minValue + ' and ' + options.maxValue;
            } else if (options.maxValue != null) {
                message = 'LOCALIZE Date may be before ' + options.maxValue;
            } else if (options.minValue != null) {
                message = 'LOCALIZE Date must be after ' + options.minValue;
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (date) {
                if (date == null) {
                    return true;
                }
                if (moment.isMoment(date)) {
                    if (options.minValue != null && date.isBefore(options.minValue)) {
                        return false;
                    }
                    if (options.maxValue != null && date.isAfter(options.maxValue)) {
                        return false;
                    }
                    return true;
                } else {
                    return false;
                }

            },

            message : options.message
        };
    };
});

define('primutils/validators/numberValidator',['primutils.lodash'], function (_) {
    


    return function (opts) {

        var defaults = {
            minValue : null,
            maxValue : null
        }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'Invalid number';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE Number must be between ' + options.minValue + ' and ' + options.maxValue;
            } else if (options.maxValue != null) {
                message = 'LOCALIZE Number may be at most ' + options.maxValue;
            }else if (options.minValue != null) {
                message = 'LOCALIZE Number must be at least ' + options.minValue;
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (number) {
                if (number == null) {
                    return true;
                }
                if (isNaN(number)) {
                    return false;
                }
                if (options.minValue != null && number < options.minValue) {
                    return false;
                }
                if (options.maxValue != null && number > options.maxValue) {
                    return false;
                }
                return true;
            },

            message : options.message
        };
    };
});

define('primutils/validators/stringValidator',['primutils.lodash'], function (_) {
    


    return function (opts) {

        var defaults = {
            minValue : null,
            maxValue : null
        }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'Invalid string';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE String must be between ' + options.minValue + ' and ' + options.maxValue + ' characters.';
            } else if (options.maxValue != null) {
                message = 'LOCALIZE String may be no greater than ' + options.maxValue + ' characters.';
            } else if (options.minValue != null) {
                message = 'LOCALIZE String must be at least ' + options.minValue + ' characters.';
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (string) {
                if (string == null) {
                    return true;
                }
                var strLen = _.unescape(string).length;
                if (!_.isString(string)) {
                    return false;
                }
                if (options.minValue != null && strLen < options.minValue) {
                    return false;
                }
                if (options.maxValue != null && strLen > options.maxValue) {
                    return false;
                }
                return true;
            },

            message : options.message

        };
    };
});

define('primutils/validators/requiredValidator',['primutils.lodash'], function (_) {
    

    return function (options) {
        var defaultOptions = {
            message : 'Field is required localizeme in options'
        };
        options = _.defaults(options || {}, defaultOptions);
        return {
            options : options,
            validate : function (value) {
                if (value == null) {
                    return false;
                }
                if (_.isString(value)) {
                    return value.trim().length > 0;
                } else if (_.isArray(value)) {
                    return value.length > 0;
                } else if (isNaN(value)) {
                    return false;
                }
                return true;
            },

            message : options.message
        };
    };
});


define('primutils/validators/uniquenessValidator',['primutils.lodash'], function (_) {
    

    /**
     * @class validators.uniquenessValidator
     * @module validation
     */
    return function (opts) {

        var defaults = {
            caseSensitive : false,
            getValueFromRecord : function (record, key) {
                return record != null ? record[key] : null;
            }
        }, globalOptions = _.defaults(opts || {}, defaults);

        return {
            /**
             * Check the array for records with duplicate values
             *
             * @method checkRecordUniqueness
             * @param {Object} newValue The new value to check for uniqueness
             * @param {Object} record The record to validate
             * @param {Object} key The field to load
             * @param {Array} arrayOfRecords The array of records to check
             * @return Array of records with the same non null value for a field
             */
            checkRecordUniqueness : function (newValue, record, key, arrayOfRecords) {
                var newValueComparable = newValue;

                if (!globalOptions.caseSensitive && _.isString(newValueComparable)) {
                    newValueComparable = newValueComparable.toLocaleLowerCase();
                }

                var matchingRecords = [];
                for (var i = 0; i < arrayOfRecords.length; i++) {
                    var recordToCheck = arrayOfRecords[i];
                    if (recordToCheck !== record) {
                        var valueToCheck = globalOptions.getValueFromRecord(recordToCheck, key);
                        if (valueToCheck != null && valueToCheck !== '') {
                            if (!globalOptions.caseSensitive && _.isString(valueToCheck)) {
                                valueToCheck = valueToCheck.toLocaleLowerCase();
                            }

                            if (valueToCheck === newValueComparable) {
                                matchingRecords.push(recordToCheck);
                            }
                        }
                    }
                }
                return matchingRecords;
            },

            checkStoreUniqueness : function (key, arrayOfRecords) {
                var valueToRecordMapping = {};
                var matchingRecords = [];
                for (var i = 0; i < arrayOfRecords.length; i++) {
                    var recordToCheck = arrayOfRecords[i];
                    var valueToCheck = globalOptions.getValueFromRecord(recordToCheck, key);
                    if (valueToCheck != null) {
                        if (!globalOptions.caseSensitive && _.isString(valueToCheck)) {
                            valueToCheck = valueToCheck.toLocaleLowerCase();
                        }

                        if (valueToRecordMapping[valueToCheck] == null) {
                            valueToRecordMapping[valueToCheck] = [];
                        }
                        valueToRecordMapping[valueToCheck].push(recordToCheck);
                    }
                }
                var matches = _.values(valueToRecordMapping);
                for (var j = 0; j < matches.length; j++) {
                    if (matches[j].length > 1) {
                        matchingRecords.push(matches[j]);
                    }
                }

                return matchingRecords;
            }
        };
    };
});


define('primutils/validators/validatorFactory',['primutils.jquery',
    'primutils.lodash',
    'primutils.logger',
    'primutils/validators/dateValidator',
    'primutils/validators/numberValidator',
    'primutils/validators/stringValidator',
    'primutils/validators/requiredValidator',
    'primutils/validators/uniquenessValidator'],
    function ($, _, logger, DateValidator, NumberValidator, StringValidator, RequiredValidator, UniquenessValidator) {
        

                var __loggedTypeWarnings = {};
        var __hasLoggedInitializationError = false;
        
        var knownTypeMap = {
            string: {
                type: 'string',
                cons: StringValidator,
            },
            number: {
                type: 'number',
                cons: NumberValidator,
            },
            date: {
                cons: DateValidator
            },
            duration: {
                cons: NumberValidator
            },
        };
        knownTypeMap.text = knownTypeMap.string;
        knownTypeMap.long = knownTypeMap.double = knownTypeMap.int = knownTypeMap.integer = knownTypeMap.percent = knownTypeMap.number;
        knownTypeMap.system_duration = knownTypeMap.duration;

        // We don't really know anything about them from a validation perspective...but don't want to uselessly error either
        // TODO should we include validations around these?
        var silentTypes = ['hour', 'unit', 'enum', 'key', 'cost', 'boolean', 'timestamp6', 'start_date', 'finish_date', 'picker', 'localized_date', 'key', 'blob'];
        _.each(silentTypes, function (type) {
            knownTypeMap[type] = {};
        });

        var defaultOptions = {
            /**
             * The data type of the field being validated
             *
             * Supports: 'string', 'date', 'number'
             *
             * @property type {String}
             * @default 'string'
             */
            type : 'string',
            /**
             * The minimum value for the field being validated
             *
             * For strings, this is the minimum number of characters required
             *
             * @property minValue {*|Number}
             * @optional
             */
            minValue : null,
            /**
             * The maximum value for the field being validated
             *
             * For strings, this is the maximum number of characters allowed
             *
             * @property maxValue {*|Number}
             * @optional
             */
            maxValue : null,

            /**
             * Whether to perform a check that the field is not null
             *
             * @property required {Boolean}
             * @default false
             */
            required : false,
            /**
             * The message returned when the field is required but is null
             *
             * @property requiredMessage {String}
             */
            requiredMessage : 'options.requiredMessage Field is required',

            /**
             * Used to turn the value into an actual object.
             * If no formatter, the assumption is the value is already correct.
             *
             * The formatter should provide two functions:
             *   * validate: Takes a string to verify that it can be unformatted
             *   * unformat: Turn a string user entered value into a workable form
             *
             * @property formatter {Object}
             * @optional
             */
            formatter : null,
            /**
             * The message returned when the format is not valid
             *
             * @property formatMessage {String}
             */
            formatMessage : 'options.formatMessage Invalid Format',

            /**
             * A default validation message.
             * Messages that will be passed into the type validators, these should be localized.
             *
             * @property message {String}
             * @optional
             */
            message : null,

            /**
             * Whether to perform a check for uniqueness
             *
             * @property unique {Boolean}
             * @default false
             */
            unique : false,
            /**
             * Used for uniqueness tests.
             * record[dataIndex] is how it will pull the data off each record in the store
             *
             * @property dataIndex {String}
             * @optional
             */
            dataIndex : null,
            /**
             * Used for uniqueness.
             * This is if you have an underlying store you can give it to check; if not pass in getUniqueStore.
             *
             * @property uniqueStore {Object}
             * @optional
             */
            uniqueStore : null,
            /**
             * Used for uniqueness.
             * Callback to get the store for uniqueness checking.
             *
             * @property getUniqueStore {Function}
             * @optional
             */
            getUniqueStore : $.noop,
            /**
             * The message returned when the value is not unique
             *
             * @property uniqueMessage {String}
             */
            uniqueMessage : 'options.uniqueMessage Field value already exists. Enter a unique value.',
            /**
             * Callback to process unique response
             *
             * @propery processUniqueResponse {Function}
             * @optional
             */
            processUniqueResponse : $.noop,
            /**
             * Check entire store when checking uniqueness
             *
             * @property checkEntireStore {Boolean}
             * @default true
             */
            checkEntireStore : true
        };

        /**
         * This class generates a validator that will run through all possible permutations of standard required, formatting, type bounds checking and uniqueness.
         *
         * 1. Check required if option is set to true.
         * 2. Check the format is valid via formatter.validate(value)
         * 3. Check any bounds error based on minValue, and maxValues set
         * 4. Run uniqueness tests if unique is true.
         * 5. Call processUniqueResponse
         * 6. Return validationResponse with the errorType, message and valid
         *
         * If it fails at any step it will stop processing the data.
         *
         * Sample use:
         *
         *      var validator = validatorFactory({
         *          type : 'number',
         *          minValue : 100,
         *          message : 'Total Cost must be at least 100',
         *      });
         *      validator.validate(200); // would return {valid : true}
         *      validator.validate(50);  // would return {valid : false, errorType : 'type', message : 'Total Cost must be at least 100'}
         *
         * @class validators.validatorFactory
         * @module validation
         */
        var factory = function (options) {
                        if (!_.isObject(options)) {
                logger.error('validatorFactory | options passed in must be an object');
            }
            if (_.isObject(options)) {
                if (!__hasLoggedInitializationError && (!options.requiredMessage || !options.uniqueMessage || !options.formatMessage)) {
                    __hasLoggedInitializationError = true;
                    logger.error('validatorFactory | the options "requiredMessage", "uniqueMessage", and "formatMessage" are required.  Otherwise, validations will be printed out incorrectly.');
                }
            }
                        _.defaults(options, defaultOptions);
            var required, type, unique;
            if (options.required) {
                required = new RequiredValidator({message : options.requiredMessage});
            }

            var typeKey = options.type.toLowerCase();
            var typeInfo = knownTypeMap[typeKey];

            if (typeInfo) {
                if (typeInfo.cons) {
                    var Cons = typeInfo.cons;
                    type = new Cons(options);

                    if (typeInfo.type) {
                        options.type = typeInfo.type; // Reset to the base known type
                    }
                }

                                else {
                    if (!__loggedTypeWarnings[typeKey]) {
                        logger.debug('No known type validator for type : ' + options.type + '.  No type-specific validation rules will be run, though generic rules will still run.');
                        __loggedTypeWarnings[typeKey] = true;
                    }
                }
                            }
                        else {
                logger.warn('Unrecognized type for validatorFactory: ' + options.type + '.  Known types are: ' + _.keys(knownTypeMap));
            }
            
            if (options.unique) {
                unique = new UniquenessValidator();
            }

            return {
                options : options,
                validators : {
                    type : type,
                    unique : unique,
                    required : required
                },
                /**
                 * (This is a function available on the returned validator)
                 *
                 * Validates the value
                 *
                 * Returns a validation result with the following format:
                 *
                 *     {
                 *         valid: {Boolean},
                 *         errorType: {String},
                 *         message: {String},
                 *         matchingRecords: {Array}
                 *     }
                 *
                 * @method validate
                 * @param rawValue The value to validate
                 * @param [record] The whole record
                 * @returns {Object}
                 */
                validate : function (rawValue, record) {
                    var response = {
                        valid : true,
                        errorType : null,
                        message : null,
                        matchingRecords : null
                    };
                    var value = rawValue;

                    if (required) {
                        if (!required.validate(rawValue)) {
                            response.errorType = 'required';
                            response.message = required.message;
                            response.valid = false;
                            return response;
                        }
                    }

                    if (options.formatter != null) {
                        if (options.formatter.validate(rawValue)) {
                            value = options.formatter.unformat(rawValue);
                        } else {
                            response.errorType = 'format';
                            response.message = options.formatMessage;
                            response.valid = false;
                            return response;
                        }
                    }

                    if (type && !type.validate(value)) {
                        response.errorType = 'type';
                        response.message = type.message;
                        response.valid = false;
                    }

                    if (unique) {
                        var store = options.uniqueStore;
                        if (store == null && options.getUniqueStore !== $.noop) {
                            store = options.getUniqueStore();
                        }
                        if (store != null) {
                            var matchingRecords = [];
                            if(options.checkEntireStore){
                                matchingRecords = unique.checkStoreUniqueness(options.dataIndex, store);
                            } else {
                                matchingRecords = unique.checkRecordUniqueness(value, record, options.dataIndex, store);
                            }
                            if (matchingRecords.length > 0) {
                                response.errorType = 'unique';
                                response.valid = false;
                                response.matchingRecords = matchingRecords;
                                response.message = options.uniqueMessage;
                            }

                            if (options.processUniqueResponse !== $.noop) {
                                options.processUniqueResponse(matchingRecords);
                            }
                        }
                    }

                    return response;
                }
            };

        };

        return factory;
    });


define('primutils/utils/defaultStrings',[], function () {
    

    /**
     * An empty set of strings.
     * For the set of strings that primutils expects in production, see <a href="http://10.140.183.183:8080/job/primutils/javadoc/files/assets_primutils_utils_defaultStrings.js.html">the source of defaultStrings.js</a>, which is only used for primutils development.
     *
     * It is expected that someone, at an application level, calls something like the following:
     *
     *     require(['primutils'], function (primutils) {
     *         primutils.lang({
     *             // lots of localized strings for the current locale
     *         });
     *     });
     *
     * @class i18n.emptyStrings
     * @module i18n
     * @private
     */
    return {};
});

define('primutils/utils/language',['primutils/i18n/Localizer', 'primutils/utils/defaultStrings'], function (Localizer, defaultStrings) {
    

    return new Localizer(defaultStrings);
});


/**
 * A shared map of key names to their keycode values.
 * See the <a href="http://10.140.183.183:8080/job/primutils/javadoc/files/assets_primutils_utils_keyCode.js.html">source</a> for available keycodes.
 *
 * @class utils.keyCode
 * @module utils
 */
define('primutils/utils/keyCode',{
    UP : 38,
    DOWN : 40,
    LEFT : 37,
    RIGHT : 39,

    SPACE : 32,
    TAB : 9,
    BACKSPACE : 8,
    ENTER : 13,
    ESC : 27,

    SHIFT : 16,
    CTRL : 17,
    ALT : 18,

    PAUSE : 19,
    CAPS : 20,
    PAGEUP : 33,
    PAGEDOWN : 34,
    END : 35,
    HOME : 36,
    INSERT : 45,
    DELETE : 46,

    F1 : 112,
    F2 : 113,
    F3 : 114,
    F4 : 115,
    F5 : 116,
    F6 : 117,
    F7 : 118,
    F8 : 119,
    F9 : 120,
    F10 : 121,
    F11 : 122,
    F12 : 123,

    NUMLOCK : 144,
    SCROLLLOCK : 145,

    LEFTSUPER : 91,
    RIGHTSUPER : 92
});


define('primutils/a11y/collapsibleDropdown',['primutils.jquery', 'primutils/utils/keyCode'], function ($, keyCode) {
    

    /**
     * What this will do is allow for left/right to open/close submenus and up/down will skip anchors in currently
     * hidden submenus.
     *
     * Warning: will only support one level of submenus
     * Warning: liberally uses "> a" selector to find direct focusable children
     *
     * In general, we try not to duplicate some of bootstrap's work here (such as up/down opening up the hidden
     * dropdown in the first place), and instead only replace behavior that could be problematic and stopPropagation().
     *
     * @class a11y.collapsibleDropdown
     * @module a11y
     * @param $menu {jQuery Object} The menu to wrap - e.g. `$('.dropdown-menu').parent()` - (though for performance and
     *   simplicity you should grab the parent directly as opposed to using `.parent()`)
     * @constructor
     */
    function CustomDropdown($menu) {
        this.$menu = $menu;
        this._start();
    }

    CustomDropdown.prototype = {
        _start: function () {
            this.$menu.on('keydown', this._handleMove.bind(this));
        },

        _move: function (moveDir, $node, ev) {
            if (this.$menu.is('.open') && !$node.is('.dropdown-toggle') && $node.is('a')) {
                // "> a" lets us be a bit more careful than bootstrap - don't want subitems of the dropdown-submenu that are
                //  invisible matching this
                var $nodes = this.$menu.find('li:visible > a');
                var index = $nodes.index($node);
                var newIndex = index + moveDir;
                if (newIndex >= 0 && newIndex < $nodes.length) {
                    ev.stopPropagation();
                    var $newNode = $nodes.eq(newIndex).focus();

                    // If up/down moved us out of the submenu, hide it
                    if ($node.parents('.dropdown-submenu').length &&
                        (!$newNode.parents('.dropdown-submenu').length || $newNode.parent().is('.dropdown-submenu'))) {
                        $node.parents('.dropdown-submenu').removeClass('open');
                    }
                }
            }
        },

        _maybeToggleSubmenu: function ($node, ev) {
            var $parent = $node.parent();
            if (!$parent.is('.dropdown-submenu') && $node.parents('.dropdown-submenu').length) {
                ev.stopPropagation();

                var $submenu = $node.parents('.dropdown-submenu');
                $submenu.removeClass('open');
                $submenu.find('> a').first().focus();
            }
            else if ($parent.is('.dropdown-submenu')) {
                ev.stopPropagation();

                var isOpen = $parent.hasClass('open');
                if (!isOpen) {
                    $parent.addClass('open');
                    $parent.find('> .dropdown-menu > li > a').first().focus();
                }
                else {
                    $parent.removeClass('open');
                }
            }
        },

        _handleMove: function (ev) {
            var $node = $(ev.target);
            switch (ev.keyCode) {
                case keyCode.UP:
                    this._move(-1, $node, ev);
                    break;
                case keyCode.DOWN:
                    this._move(1, $node, ev);
                    break;
                case keyCode.LEFT:
                case keyCode.RIGHT:
                    this._maybeToggleSubmenu($node, ev);
                    break;
                default:
                    return;
            }
        }
    };

    return function($menu) {
        return new CustomDropdown($menu);
    };

});


define('primutils/a11y/accessibilityUtils',['primutils.jquery', 'primutils.lodash',
    'primutils/a11y/collapsibleDropdown'
], function ($, _, collapsibleDropdown) {
    

    /**
     * Defines a shared set of functionality around accessibility.
     *
     * Some functions are also available on jQuery. Generally those that take a jQuery object as their first
     * parameter are also bound to jQuery. You can access functions bound to jQuery with the prefix 'prim_'.
     *
     *     require(['jquery', 'primutils/a11y/accessibilityUtils'], function ($) {
     *         
     *         $('li > a.someItem').prim_toggleListItemAnchor(true);
     *     });
     *
     * This is exactly equivalent to
     *
     *     require(['jquery', 'primutils/a11y/accessibilityUtils'], function ($, accessibilityUtils) {
     *         
     *         var $item = $('li > a.someItem');
     *         accessibilityUtils.toggleListItemAnchor($item, true);
     *     });
     *
     * @class a11y.accessibilityUtils
     * @module a11y
     */

    var jqueryCompatibleUtils = {
        /**
         * Toggles the state of an `<a>` in a dropdown list item so that it gets the proper Bootstrap styling and
         * aria attributes.
         *
         * Warning: This does not prevent the user from triggering clicks on the element. This feature is pending.
         *
         * (Also available from jQuery)
         *
         * @method toggleListItemAnchor
         * @param $anchor {jQuery Object} The `<a>` to toggle
         * @param enable {Boolean} Whether the anchor should appear enabled
         */
        toggleListItemAnchor : function ($anchor, enable) {
            var $li = $anchor.parent('li');
            $li.toggleClass('disabled', !enable);
            $anchor.attr('aria-disabled', !enable).attr('tabindex', enable ? 0 : -1);
        },
        /**
         * Creates a collapsible dropdown from the given menu.
         * (See <a href="a11y.CustomDropdown.html">a11y.CustomDropdown</a> for more information.)
         *
         * (Also available from jQuery)
         *
         * @method collapsibleDropdown
         * @param $menu {jQuery Object} The menu to bind to
         */
        collapsibleDropdown : function ($menu) {
            return collapsibleDropdown($menu);
        },
        /**
         * Appends the given modal to the highest visible modal currently in the DOM, if any, to prevent `focus` competitions from causing an infinite loop and crashing the browser.
         *
         * Be aware that calling this will place your modal "inside" the DOM of the other modal.
         * If that parent modal is closed, your modal will be closed also.
         * This method should only be called if that is your expected behavior or you are an alert modal which locks out additional user activity until action is taken.
         *
         * @method modalAppend
         * @param $modal {jQuery Object} The modal to append
         * @param options {Object} if `options.allowClosingParent` is true, will allow 'hide' and 'hidden' events to propagate to the parent modal.
         *   By default, these two events WILL NOT propagate to the parent modal.
         */
        modalAppend : function ($modal, options) {
            options = options || {};

            // To avoid causing a infinite call stack by modals trying to compete for focus,
            //  we'll append to the topmost visible modal onscreen, if there currently is one.
            var appendPoint = document.body;
            var possibleModalAppend = $('.modal.in').last();
            if (possibleModalAppend.length) {
                appendPoint = possibleModalAppend;

                if (!options.allowClosingParent) {
                    jqueryCompatibleUtils.stopModalHidePropagation($modal);
                }
            }
            $modal.appendTo(appendPoint);
        },
        /**
         * Prevents `'hide'` and `'hidden'` events from propagating from the given modal to any parent modals.
         *   Useful when using <a href="#method_modalAppend">modalAppend</a>, which will call this method by default.
         *
         * @method stopModalHidePropagation
         * @param $modal {jQuery Object} the modal to suppress events on
         */
        stopModalHidePropagation : function ($modal) {
            $modal.one('hide hidden', function (e) {
                e.stopPropagation();
            });
        },

        filterFocusable : function ($collection) {
            return $collection
                .filter('a[href], button:not([disabled]), input:not([disabled]), textarea:not([disabled]), select:not([disabled]), [tabindex]')
                .filter(':visible:not([tabindex=-1]), input[type="file"]');
        },

        findFocusable : function ($parent) {
            return $parent
                .find('a[href], button:not([disabled]), input:not([disabled]), textarea:not([disabled]), select:not([disabled]), [tabindex]')
                .filter(':visible:not([tabindex=-1]), input[type="file"]');
        }
    };

    var utils = {

    };

    function makeJQueryFn(fn) {
        return function () {
            var $this = $(this);

            var args = _.toArray(arguments);
            args.unshift($this);

            var result = fn.apply(undefined, args);

            if (result === undefined) {
                return $this;
            } else {
                return result;
            }
        };
    }

    // Extend jQuery with jquery compatible fns
    $.fn.extend(_.reduce(jqueryCompatibleUtils, function (res, fn, name) {
            res['prim_' + name] = makeJQueryFn(fn);
            return res;
        }, {}));

    return _.extend({}, jqueryCompatibleUtils, utils);
});


define('primutils/templates/helpers/localizer',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/i18n/Localizer'
], function (logger, _, Localizer) {
    

    var localizer;

    /**
     * Gets a localized string from the registered localizer
     *
     * Available as: `\{{lz}}`.
     *
     * Localizer is a helper that does not have inner content.
     *
     * If only one argument is passed, it will be assumed that your string is not formatted.
     * If multiple arguments are passed, the will be inserted into the format placeholders.
     *
     * Example:
     *
     *     \{{lz 'global.hello_world'}}
     *     \{{lz 'global.hello_x' 'World'}}
     *
     * You can also use the localizer helper as a parameter to another helper.
     * Just use parentheses `()` instead of staches, for example:
     *
     *     \{{tab (lz 'global.my_tab_name') 'tab_id'}}
     *
     * To set the localizer used by this helper use the following method.
     * `myLocalizer` should be an instance of the `Localizer` class.
     *
     *     primutils.setLocalizerForLocalizerHelper(myLocalizer);
     *
     * See also: <a href="i18n.Localizer.html">i18n.Localizer</a>
     *
     * @class templates.helpers.lz
     * @module templates.helpers
     * @constructor
     * @param key {String} Localization key to read from the localizer
     * @param [...] {String} Values to insert into format placeholders
     * @param [options] {Object} Handlebars options
     * @returns {string} Localized string
     */
    var lz = function (key) {
        logger.assert(!!localizer, 'Templates: Cannot use localizer helper until a localizer has been configured');
        logger.assert(arguments.length >= 1, 'Templates: Localizer helper requires a "key" argument');
        if (arguments.length === 1) {
            return localizer.getString(key);
        } else {
            var localizerArgs = _.initial(_.toArray(arguments)); // Take everything but what would be the options arg
            return localizer.getFormattedString.apply(localizer, localizerArgs);
        }
    };

    /**
     * Sets the localizer used by the localizer helper
     * @method useLocalizer
     * @static
     * @param newLocalizer {Localizer} The localizer to use when retrieving keys
     */
    lz.useLocalizer = function (newLocalizer) {
        logger.assert(newLocalizer instanceof Localizer, 'Templates: Localizer must be an instance of the type Localizer');
        localizer = newLocalizer;
    };

    lz.handlebarsType = 'helper';
    lz.handlebarsName = 'lz';

    return lz;
});


define('primutils/templates/utils/utils',['primutils.lodash'], function (_) {
    

    var GLOBAL_ATTRIBUTES = ['accesskey', /*'class',*/ 'dir', 'dropzone', 'id', 'lang', 'spellcheck', 'style', 'tabindex', /*'title',*/ 'translate'],
        GLOBAL_BOOL_ATTRIBUTES = [/*'contenteditable',*/ 'draggable', 'hidden'];

    /**
     * Provides utility functions that can be used inside helpers and partials
     *
     * @class templates.utils.utils
     * @module templates.utils
     */

    var utils = {
        /**
         * Filters the Handlebars hash to a list of attributes that can be applied to the generated element.
         * The user can specify additional attributes that are valid for their element in addition to the globally allowed attributes.
         *
         * Attributes are broken into two categories: string and boolean.
         * String attributes the pass the filter are written to the template exactly as passed.
         * Boolean attributes ignore the value and are always written as attribute="attribute" to ensure backwards compatibility.
         * This makes it easy for the template author to specify something like `{{myHelper required=true}}` and have it interpreted correctly.
         *
         * The default list of attributes is: accesskey, dir, draggable, dropzone, hidden, id, lang, spellcheck, style, tabindex, and translate.
         *
         * If you need to prevent users from specifying these attributes but still want to call `getElementAttributes`, delete them from the hash before passing it to this method.
         *
         * Some global attributes are missing from the default list, namely: class, contenteditable, and title.
         * Generally you do not want the user specifying these herself.
         * For 'class', consider using the <a href="#method_mergeClasses">mergeClasses</a> method passed to the <a href="templates.utils.renderer.html#method_renderElement">renderer.renderElement</a> method.
         * For the others, you can explicitly pass them in the first attributes parameter to the <a href="templates.utils.renderer.html#method_renderElement">renderer.renderElement</a> method or list them in the `passedAttrs` or `booleanPassedAttrs` parameters in this function.
         *
         * @method getElementAttributes
         * @param hash {Object} The Handlebars hash of additional properties passed to the helper in a template
         * @param [passedAttrs] {Array} List of attributes that should be kept when rendering an element
         * @param [booleanPassedAttrs] {Array} List of boolean attributes that should be kept when rendering an element
         * @returns {Object} A filtered key/value pair collection valid for converting into HTML attributes
         */
        getElementAttributes : function (hash, passedAttrs, booleanPassedAttrs) {
            return _.reduce(hash, function(memo, value, key) {
                var lowerKey = key.toLowerCase();
                if (_.contains(passedAttrs, lowerKey) || _.contains(GLOBAL_ATTRIBUTES, lowerKey)) {
                    memo[key] = value;
                } else if ((_.contains(booleanPassedAttrs, lowerKey) || _.contains(GLOBAL_BOOL_ATTRIBUTES, lowerKey)) && value !== false) {
                    // Do not match boolean attributes when they are explicitly set to false
                    memo[key] = key;
                } else if(key.match(/^data-/g)) {
                    memo[key] = value;
                } else if (key.match(/^aria-/g)) {
                    memo[key] = value;
                }
                return memo;
            }, {});
        },

        /**
         * Merges an array of class names with a string separated list of classes and converts it into a space separated list of classes.
         *
         * @method mergeClasses
         * @param defaults {Array} List of classes provided by the helper or partial author
         * @param [classes] {String} Space separated list of classes usually provided by the template author
         * @returns {string} Space separated list of classes
         */
        mergeClasses : function (defaults, classes) {
            return classes ? _.union(defaults, classes.split(' ')).join(' ') : defaults.join(' ');
        },

        /**
         * Converts a space separated list of classes into an array
         *
         * @method classList
         * @private
         * @param classes {String} Space separated list of classes
         * @returns {Array} Array of classes
         */
        classList : function (classes) {
            return (classes||'').split(' ');
        },

        /**
         * Checks to see if a particular class is provided in a space separated list of classes
         *
         * @method hasClass
         * @private
         * @param classes {String} Space separated list of classes
         * @param clz {String} Class name to find
         * @returns {Boolean} Whether the class exists inside the classes string
         */
        hasClass : function (classes, clz) {
            return _.contains(utils.classList(classes), clz);
        }
    };

    return utils;
});


define('primutils/templates/utils/renderer',['primutils.lodash'], function (_) {
    

    /**
     * Provides a set of methods for writing various HTML tags.
     *
     * Generally, the render methods provide two separate "attributes" parameters.
     * The first is used for HTML attributes that are provided by the helper.
     * The second is for attributes provided by the template author.
     * Both are always required, but you can pass `null` for either when a helper does not require them.
     *
     * There is currently no effort to de-dupe the attributes because there is no way to know whether the template author's or the helper author's attribute should be preferred.
     * If you need to merge the helper value with an author provided value, it is suggested that you do the merge, place it in the first attribute object, and delete the author's value from the hash.
     *
     * @class templates.utils.renderer
     * @module templates.utils
     */

    var renderer = {
        /**
         * Converts a key/value collection of attributes into a string of attributes to append to an HTML element
         *
         * @method renderAttributesString
         * @protected
         * @param attributes {Object} key/value collection of attributes
         * @returns {String} HTML attributes string
         */
        //TODO escape quotes (") in value param
        renderAttributesString : function (attributes) {
            return _.reduce(attributes, function (memo, value, key) {
                if (value != null) {
                    return memo + ' ' + key + '="' + value + '" ';
                } else {
                    return memo;
                }
            }, '');
        },

        /**
         * Renders an HTML element with inner text.
         *
         * @method renderElementWithContent
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param dataAttributes {Object|null} Key/value collection of attributes provided by the template author
         * @param text {String} Text/HTML to place inside the element
         * @returns {string} The built HTML element
         */
        renderElementWithContent : function (tag, attributes, dataAttributes, text) {
            return '<' + tag + ' ' + renderer.renderAttributesString(attributes) + renderer.renderAttributesString(dataAttributes) + ' >' + text + '</' + tag + '>';
        },

        /**
         * Renders an HTML element as fully closed.
         *
         * This should be used for elements that do not contain inner text, but must have a proper closing tag.
         * Examples include empty `<div>`s (e.g. that you plan to fill later with jQuery) and icons `<i></i>`.
         *
         * @method renderElement
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param dataAttributes {Object|null} Key/value collection of attributes provided by the template author
         * @returns {string} The built HTML element
         */
        renderElement : function (tag, attributes, dataAttributes) {
            return renderer.renderElementWithContent(tag, attributes, dataAttributes, '');
        },

        /**
         * Renders an HTML element that is self closed.
         *
         * This should be used for elements that do not contain inner text and are always self closed.
         * Examples include image tags `<img />`.
         *
         * @method renderSelfClosingElement
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param dataAttributes {Object|null} Key/value collection of attributes provided by the template author
         * @returns {string} The built HTML element
         */
        renderSelfClosingElement : function (tag, attributes, dataAttributes) {
            return '<' + tag + ' ' + renderer.renderAttributesString(attributes) + renderer.renderAttributesString(dataAttributes) + ' />';
        }
    };

    return renderer;
});


define('primutils/templates/helpers/button',[
    'primutils.logger',
    'primutils.lodash',
    'primutils.jquery',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/utils/language'
], function (logger, _, $, utils, renderer, lang) {
    

    var BUTTON_TYPE_MAP = {
        'default' : 'blue-light',
        'primary' : 'action',
        'secondary' : 'blue-dark',
        'page' : 'primary',
        'grid' : 'blue-light',
        'toolbar' : 'blue-light',
        'form' : 'form',
        'not-btn' : 'not-btn'
    };

    var BUTTON_ATTRIBUTES = ['form', 'formaction', 'formenctype', 'formmethod', 'formtarget', 'name', /*'type',*/ 'value'];
    var BUTTON_BOOL_ATTRIBUTES = ['autofocus', 'disabled', 'formnovalidate'];

    function getType(type) {
        var newType = (BUTTON_TYPE_MAP[type] || type);
        return newType !== 'not-btn' ? 'btn-' + newType : newType;
    }

    /**
     * Allows the button to control the related dropdown menu.
     *
     * The button should either appear inside a <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a> next to a <a href="templates.helpers.dropdownMenu.html">\{{#dropdownMenu}}</a> or
     * should specify a `data-target` attribute which is a jQuery selector to a dropdown menu.
     *
     * Example:
     *
     *     \{{#button 'toolbar' dropdownToggle=true data-target="#myDropdown"}}Menu\{{/button}}
     *
     *     \{{#buttonGroup}}
     *         \{{#button 'toolbar' dropdownToggle=true}}Menu\{{/button}}
     *         \{{#dropdownMenu}} ... \{{/dropdownMenu}}
     *     \{{/buttonGroup}}
     *
     * @property dropdownToggle
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a button element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#button}}\{{/button}}`.
     *
     * Button is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#button 'primary' 'My Button Title'}}My Button\{{/button}}
     *
     * There are eight types of buttons: 'default', 'primary', 'secondary', 'page', 'grid', 'toolbar', 'form', and 'not-btn'.
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/button#HTML_Attributes">W3C Button Attributes</a>.
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>, <a href="templates.helpers.toolbar.html">\{{#toolbar}}</a>
     *
     * @class templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var button = function (type, title, options) {
        var dropdownClass, dataToggle, role;

        if(arguments.length === 2) {
            options = title;
            title = null;
        } else if (arguments.length === 1) {
            options = type;
            type = null;
            title = null;
        }

        logger.assert(!!options.fn, 'Templates: Button helper must be used as a block helper.');

        if (type != null && type !== '') { // Allow for empty types to just give 'btn'
            type = getType(type);
        }

        if (options.hash.dropdownToggle) {
            dropdownClass = 'dropdown-toggle';
            dataToggle = 'dropdown';
            role = 'button';
        }

        this.__buttonContext__ = {
            isDropdownToggle : options.hash.dropdownToggle
        };

        var text = options.fn(this).trim(),
            hasCaret = !!this.__buttonContext__.hasCaret;

        delete this.__buttonContext__;

        return renderer.renderElementWithContent('button', {
            type : options.hash.type || 'button',
            'class' : utils.mergeClasses([type !== 'not-btn' ? 'btn' : '', type, dropdownClass], options.hash['class']),
            title  : title != null ? title : $('<div>'+text+'</div>').text().trim(),
            'data-toggle' : dataToggle,
            'aria-pressed' : utils.hasClass(options.hash['class'], 'active') ? 'true' : undefined,
            'aria-expanded' : options.hash.dropdownToggle ? 'false' : undefined,
            'role' : role,
            'aria-haspopup' : role ? 'true' : undefined,
            'aria-owns' : role ? '{{menuId}}' : undefined,
            'aria-label' : (!options.hash['aria-label'] && hasCaret) ? lang.getString('templates.moreOptions') : undefined
        }, utils.getElementAttributes(options.hash, BUTTON_ATTRIBUTES, BUTTON_BOOL_ATTRIBUTES), text);
    };

    button.handlebarsType = 'helper';
    button.handlebarsName = 'button';
    button.getType = getType;

    return button;
});


define('primutils/templates/helpers/link',[
    'primutils.logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/button'
], function(logger, utils, renderer, button) {
    

    var LINK_ATTRIBUTES = ['download', /*'href',*/ 'hreflang', 'media', 'name', 'rel', 'target', 'type'];

    /**
     * Allows the link to appear as a button.
     * You can specify any of the valid <a href="templates.helpers.button.html">button types</a>.
     *
     * @property asButton
     * @type {String|null}
     * @default null
     */

    /**
     * Creates a link element.
     *
     * Available as: `\{{#link}}\{{/link}}`.
     *
     * Link is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#link 'http://www.google.com/' 'Google'}}Go To Google\{{/link}}
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/a#HTML_Attributes">W3C Anchor Attributes</a>.
     *
     * If only one arguments are passed, it is assumed to be href.
     * If you want to specify just the item's title, you should pass '#' as the first argument.
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.link
     * @module templates.helpers
     * @constructor
     * @param href {String} The end point of the anchor
     * @param [title] {String} The title text of the link
     * @param [options] {Object} Handlebars options
     * @returns {string} Link element
     */
    var link = function (href, title, options) {
        logger.assert(arguments.length >= 2, 'Templates: Link helper requires an "href" argument');

        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        logger.assert(!!options.fn, 'Templates: Link helper must be used as a block helper.');

        var type;
        if (options.hash.asButton) {
            type = button.getType(options.hash.asButton);
        }

        return renderer.renderElementWithContent('a', {
                href : href,
                title : title,
                class : options.hash.asButton ?
                    utils.mergeClasses([type !== 'not-btn' ? 'btn' : '', type], options.hash['class']) :
                    options.hash['class'],
                role : options.hash.asButton ? 'button' : options.hash._internalRole
            },
            utils.getElementAttributes(options.hash, LINK_ATTRIBUTES), options.fn(this));
    };

    link.handlebarsType = 'helper';
    link.handlebarsName = 'link';

    return link;
});


define('primutils/templates/helpers/accordion',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link'
], function (logger, _, utils, renderer, link) {
    

    var accordionHeader = function (name, headerId, bodyId, parent, open) {
        return renderer.renderElementWithContent('div', {
            'class' : 'accordion-heading'
        }, null, link.call(this, '#' + bodyId, null, {
            hash : {
                'id' : headerId,
                'aria-expanded' : open ? 'true' : 'false',
                'aria-controls' : bodyId,
                'aria-selected' : open ? 'true' : 'false',
                'role' : 'tab',
                'class' : 'accordion-toggle',
                'data-toggle' : 'collapse',
                'data-parent' : parent ? '#' + parent : undefined
            },
            fn : function () {
                return name;
            }
        }));
    };

    var accordionBody = function (headerId, options) {
        var innerText = options.fn(this);

        return renderer.renderElementWithContent('div', {
            'class' : 'accordion-body collapse' + (options.hash.open ? ' in' : ' hidden')
        }, utils.getElementAttributes(options.hash), renderer.renderElementWithContent('div', {
            'role' : 'tabpanel',
            'aria-labelledby' : headerId,
            'aria-hidden' : options.hash.open ? 'false' : 'true',
            'class' : 'accordion-inner'
        }, null, innerText));
    };

    /**
     * Opens this accordion section initially.
     * @property open
     * @for templates.helpers.accordionSection
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates an accordion section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordionSection}}\{{/accordionSection}}`.
     *
     * Accordion section is a block helper and requires inner content.
     *
     * Accordion section can only be used inside of an <a href="templates.helpers.accordion.html">accordion</a>.
     *
     * @class templates.helpers.accordionSection
     * @module templates.helpers
     * @protected
     * @constructor
     * @param name {String} The name of the section
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion section element
     */
    var accordionSection = function (name, options) {
        logger.assert(this.__accordionContext__, 'Templates: Accordion section helper must be called from within a {{#accordion}}');
        logger.assert(arguments.length === 2, 'Templates: Accordion section helper requires "name" argument');
        logger.assert(!!options.fn, 'Templates: Accordion section helper must be used as a block helper.');

        var parent = this.__accordionContext__.independent ? null : this.__accordionContext__.id,
            headerId = _.uniqueId('pgbu-accordion-header-');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-accordion-section-');

        return renderer.renderElementWithContent('div', {
            'class' : 'accordion-group' + (options.hash.open ? ' open' : '')
        }, null, accordionHeader.call(this, name, headerId, options.hash.id, parent, options.hash.open) + accordionBody.call(this, headerId, options));
    };

    accordionSection.handlebarsType = 'helper';
    accordionSection.handlebarsName = 'accordionSection';

    /**
     * Makes the accordion "twistie" style instead of the Bootstrap default.
     * @property twistie
     * @for templates.helpers.accordion
     * @type {Boolean}
     * @default false
     */

    /**
     * Makes all of the accordion sections independent of one another.
     * Normally, only one section can be open at a time.
     * When true, this allows multiple sections to be open at once.
     * @property independent
     * @for templates.helpers.accordion
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates an accordion component.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordion}}\{{/accordion}}`.
     *
     * Accordion is a block helper and requires inner content.
     *
     * Accordion requires some sub-helpers:
     *  - <a href="templates.helpers.accordionSection.html">accordionSection</a>
     *
     * Example:
     *
     *     \{{#accordion twistie=true}}
     *         \{{#accordionSection "Section 1" open=true}}
     *             ... Section Content ...
     *         \{{/accordionSection}}
     *     \{{/accordion}}
     *
     * @class templates.helpers.accordion
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion element
     */
    var accordion = function (options) {
        logger.assert(!!options.fn, 'Templates: Accordion helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-accordion-');

        var accordionContext = {
            id : options.hash.id,
            independent : options.hash.independent
        };

        this.__accordionContext__ = accordionContext;

        var innerText = options.fn(this);

        delete this.__accordionContext__;

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['accordion', options.hash.twistie ? 'toggle-accordion' : ''], options.hash['class']),
            'role' : 'tablist',
            'multiselectable' : accordionContext.independent ? 'true' : 'false'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    accordion.handlebarsType = 'helper';
    accordion.handlebarsName = 'accordion';

    return {
        accordion : accordion,
        accordionSection : accordionSection
    };

});


define('primutils/templates/helpers/buttonGroup',[
    'primutils.logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils.handlebars'
], function (logger, utils, renderer, Handlebars) {
    

    /**
     * Creates a button group.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#buttonGroup}}\{{/buttonGroup}}`.
     *
     * Button group is a block helper and requires inner content, generally <a href="templates.helpers.button.html">\{{#button}}</a> and <a href="templates.helpers.dropdownMenu.html">\{{#dropdownMenu}}</a>.
     *
     * Using `data-` attributes, you can activate the <a href="http://getbootstrap.com/2.3.2/javascript.html#buttons">Bootstrap</a> checkbox or radio functionality.
     *
     * Example:
     *
     *     \{{#buttonGroup}}
     *         \{{#button 'toolbar'}}Edit\{{/button}}
     *     \{{/buttonGroup}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>, <a href="templates.helpers.toolbar.html">\{{#toolbar}}</a>
     *
     * @class templates.helpers.buttonGroup
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Button group element
     */
    var buttonGroup = function (options) {
        logger.assert(!!options.fn, 'Templates: Button group helper must be used as a block helper.');

        var buttonGroupContext = this.__buttonGroupContext__ = {};

        var innerText = options.fn(this);

        delete this.__buttonGroupContext__;

        if (buttonGroupContext.hasDropdownMenu) {
            innerText = innerText.replace(/\{\{menuId\}\}/g, buttonGroupContext.menuId);
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-group'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), innerText);
    };

    buttonGroup.handlebarsType = 'helper';
    buttonGroup.handlebarsName = 'buttonGroup';

    return buttonGroup;
});


define('primutils/templates/helpers/caret',['primutils/templates/utils/renderer'], function (renderer) {
    

    /**
     * Creates a down-facing caret.
     *
     * Generally used as the content of a dropdown toggle button.
     *
     * Available as: `\{{caret}}`.
     *
     * Caret is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{caret}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.caret
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Caret element
     */
    var caret = function () {
        if (this.__buttonContext__ && this.__buttonContext__.isDropdownToggle) {
            this.__buttonContext__.hasCaret = true;
        }
        return renderer.renderElement('span', { 'class' : 'caret' });
    };

    caret.handlebarsType = 'helper';
    caret.handlebarsName = 'caret';

    return caret;
});


define('primutils/templates/helpers/controlGroup',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function(logger, _, utils, renderer) {
    

    /**
     * Marks the control group as required.
     *
     * The property may be set automatically by contained `\{{input}}`s.
     * If the control group contains an `\{{input required=true}}`, the control group will also automatically be marked as required.
     * However, the reverse is not true.
     * If the control group is marked as required, contained `\{{input}}`s do not automatically become required as well.
     *
     * @property required
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies that the contents on the control group is only plain text.
     * If you are only displaying a value and not creating an input, you should set this property to true.
     * This should **not** be set if the control group contains an input that has been set to readonly.
     *
     * See the difference between temporarily disabled inputs and "view-only fields" on the <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=InputTypes&subtopic=ViewOnly">Blue Sky UX Guide</a>.
     *
     * @property readOnly
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies that the contained inputs are a group of radio buttons that should be on one line instead of on top of each other.
     *
     * @property inlineRadio
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies the HTML `name` attribute to be given to all of the contained inputs.
     * If any underlying inputs do not have their `name` attribute specified, they will get this one automatically.
     * Generally used when the contained inputs are all checkboxes or radio buttons.
     *
     * @property name
     * @type {String|null}
     * @default null
     */

    /**
     * Specifies the ID attribute of the generated label.
     * Used in cases where the page author needs to access the `<label>` once its created.
     *
     * If not specified, it will receive an auto-incrementing ID so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property labelId
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Specifies the for attribute of the generated label.
     * Used in cases when the input is created after the `<label>` is created.
     *
     * If not specified, it will receive the ID of the first input in the control group so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property for
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Renders the control as inline with the label right next to the control.
     *
     * Useful for toolbars.
     *
     * @property inline
     * @type {Boolean}
     * @default false
     */

    /**
     * Places icons in radio buttons and checkboxes on the left side of the input.
     *
     * @property leftIcons
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a control group section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms&subtopic=LabelandFieldLayout">Forms on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#controlGroup}}\{{/controlGroup}}`.
     *
     * Control group is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#controlGroup 'My Input'}}
     *         \{{input 'text'}}
     *     \{{/contorlGroup}}
     *
     * See also: <a href="templates.helpers.input.html">\\{{input}}</a>
     *
     * @class templates.helpers.controlGroup
     * @module templates.helpers
     * @constructor
     * @param label {String} The control group's label text
     * @param [options] {Object} Handlebars options
     * @returns {string} Control group element
     */
    var controlGroup = function (label, options) {
        logger.assert(arguments.length === 2, 'Templates: Control group helper requires a "label" argument.');
        logger.assert(!!options.fn, 'Templates: Control group helper must be used as a block helper.');

        options.hash.labelId = options.hash.labelId || _.uniqueId('pgbu-label-');

        var controlGroupContext = {
            labelId : options.hash.labelId,
            inputRequired : !!options.hash.required,
            inputs : [],
            name : options.hash.name,
            leftIcons : !!options.hash.leftIcons
        };

        this.__controlGroupContext__ = controlGroupContext;

        var innerText = options.fn(this);

        delete this.__controlGroupContext__;

        var labelContent = renderer.renderElementWithContent('label', {
            'class' : 'control-label',
            for : controlGroupContext.inputs.length > 0 ? controlGroupContext.inputs[0] : options.hash.for,
            id : options.hash.labelId,
            title : label
        }, null, label);

        var controls = renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['controls'], options.hash.readOnly ? 'controls-view-only' : '')
        }, null, innerText);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['control-group',
                controlGroupContext.inputRequired ? 'required' : '',
                options.hash.inlineRadio ? 'controls-yes-no' : '',
                options.hash.inline ? 'control-group-inline' : ''
            ], options.hash['class']),
            'role' : controlGroupContext.isRadioGroup ? 'radiogroup' : undefined
        }, utils.getElementAttributes(options.hash), labelContent + controls);
    };

    controlGroup.handlebarsType = 'helper';
    controlGroup.handlebarsName = 'controlGroup';

    return controlGroup;
});


define('primutils/templates/helpers/icon',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, utils, renderer) {
    

    var baseIconName = 'pgbu-icon',
        baseIconPrefix = baseIconName + '-';

    try {
        /*global requirejs:true*/
        var requireConfig = requirejs.s.contexts._.config;
        baseIconName = requireConfig.markup.baseIconName || baseIconName;
        baseIconPrefix = requireConfig.markup.baseIconPrefix || baseIconName + '-';
    } catch (e) {}

    function spokenText(title) {
        return renderer.renderElementWithContent('span', {
            'class': 'spoken-text'
        }, null, title);
    }

    var customIcons = {
        'throbber' : function (fixedIconClass, title, options, size) {
            var dot = renderer.renderElement('div', { 'class' : 'dot' }, {});
            return renderer.renderElementWithContent('i', {
                'class' : utils.mergeClasses([size, 'throbber'], options.hash['class']),
                title : title
            }, utils.getElementAttributes(options.hash), dot + dot + dot + spokenText(title));
        }
    };

    /**
     * Creates an icon element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=GraphicDesign&subtopic=Icons">Icons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{icon}}`.
     *
     * Icon is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{icon 'email' 'Email' size="pgbu-icon-small"}}
     *
     * As shown, when specifying the icon type, you do not need to prepend 'pgbu-icon-'.
     * When specifying the icon size you must specify the full size class, e.g. "pgbu-icon-large".
     *
     * The base icon name can be controlled by specifying the following options in the require config:
     *
     *     markup : {
     *         baseIconName : 'pgbu-icon',
     *         baseIconPrefix : 'pgbu-icon-' // Optional, inferred from baseIconName if omitted
     *     }
     *
     * @class templates.helpers.icon
     * @module templates.helpers
     * @constructor
     * @param iconClass {String} The type of the icon
     * @param title {String} The title text of the icon as well as the text spoken by the screen reader
     * @param [options] {Object} Handlebars options
     * @returns {string} Icon element
     */
    var icon = function (iconClass, title, options) {
        logger.assert(arguments.length === 3, 'Templates: Link helper requires a "iconClass" and "title" arguments');

        var size = options.hash.size || baseIconName,
            fixedIconClass = baseIconPrefix + iconClass;

        if (_.has(customIcons, iconClass)) {
            return customIcons[iconClass].call(this, fixedIconClass, title, options, size);
        }

        return renderer.renderElementWithContent('i', {
            'class' : utils.mergeClasses([size, fixedIconClass], options.hash['class']),
            title : title
        }, utils.getElementAttributes(options.hash), spokenText(title));
    };

    icon.handlebarsType = 'helper';
    icon.handlebarsName = 'icon';

    return icon;
});


define('primutils/templates/helpers/dock',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/icon',
    'primutils/utils/language'
], function (logger, _, utils, renderer, link, button, icon, lang) {
    

    var VALID_ORIENTATIONS = ['left', 'right', 'top', 'bottom'],
        SUPPORTS_ORIENTATION_SWITCHING = ['right', 'bottom'],
        ARIA_ORIENTATIONS = {
            left : 'vertical',
            right : 'vertical',
            top : 'horizontal',
            bottom : 'horizontal'
        };

    var ORIENTATION_LANG_STRINGS = {
        'vertical' : 'dock.dockRight',
        'horizontal' : 'dock.dockBottom'
    };

    var orientButton = function (orientation, active) {
        return button.call(this, '', lang.getString(ORIENTATION_LANG_STRINGS[orientation]), {
            hash : { 'class' : active ? 'active' : '', 'data-orientation' : orientation },
            fn : function () {
                return icon.call(this, 'details-' + orientation, lang.getString(ORIENTATION_LANG_STRINGS[orientation]), {
                    hash : { size : 'pgbu-icon-small' }
                });
            }
        });
    };

    var dockHeader = function (title, titleId, noOrientation, initialOrientation) {
        var titleBlock = renderer.renderElementWithContent('div', {
                'class' : 'dock-title',
                id : titleId
            }, null, title),
            dockControls = noOrientation ? '' : renderer.renderElementWithContent('div', {
                'class' : 'dock-controls',
                'data-toggle' : 'buttons-radio'
            }, null, orientButton.call(this, 'horizontal', initialOrientation === 'bottom') + orientButton.call(this, 'vertical', initialOrientation === 'right'));

        return renderer.renderElementWithContent('div', {
            'class' : 'dock-header'
        }, null, titleBlock + dockControls);
    };

    /**
     * Prevents the user from changing the orientation of the dock.
     *
     * This option only applies to docks that are docked right or bottom.
     * Docks that are docked left or top do not support reorientation.
     *
     * @property noOrientation
     * @type {Boolean}
     * @default false
     */

    /**
     * Prevents the user from collapsing the dock.
     *
     * @property noCollapse
     * @type {Boolean}
     * @default false
     */

    /**
     * Prevents the user from manually resizing the dock.
     *
     * @property noResize
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a dock component.
     *
     * Available as: `\{{#dock}}\{{/dock}}`.
     *
     * Dock is a block helper and requires inner content.
     *
     * In order for a dock to work properly, it must be placed inside of an element with the ".dock-parent" class.
     *
     * Example:
     *
     *     <div class="dock-parent">
     *         ... content that should be tabbed through before the contents of the dock ...
     *
     *         \{{#dock 'My Dock' 'left'}}
     *             ... dock inner content ...
     *         \{{/dock}}
     *
     *         ... content that should be tabbed through after the contents of the dock ...
     *     </div>
     *
     * @class templates.helpers.dock
     * @module templates.helpers
     * @constructor
     * @param title {String} The title of the dock
     * @param [initialOrientation] {String} The initial location of the dock (Default: 'right')
     * @param [options] {Object} Handlebars options
     * @returns {string} Dock element
     */
    var dock = function (title, initialOrientation, options) {
        logger.assert(arguments.length >= 2, 'Templates: Dock helper requires a "title" argument');

        if (arguments.length === 2) {
            options = initialOrientation;
            initialOrientation = 'right';
        }

        logger.assert(!!options.fn, 'Templates: Dock helper must be used as a block helper.');
        logger.assert(_.contains(VALID_ORIENTATIONS, initialOrientation), 'Templates: Dock initialOrientation must be: left, right, top, or bottom');

        var collapsed = _.contains(utils.classList(options.hash['class']), 'collapsed');

        if (!_.contains(SUPPORTS_ORIENTATION_SWITCHING, initialOrientation)) {
            options.hash.noOrientation = true;
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-dock-');
        var titleId = _.uniqueId('pgbu-dock-title-');

        var orientationClass = 'dock-' + initialOrientation,
            innerText = options.fn(this),
            resizeHandle = options.hash.noResize ? '' : renderer.renderElement('div', {
                'role' : 'separator',
                'aria-controls' : options.hash.id,
                'aria-orientation' : ARIA_ORIENTATIONS[initialOrientation],
                tabindex : '0',
                'class' : 'resize-handle'
            }, null),
            collapseHandleString = collapsed ? lang.getString('dock.expand') : lang.getString('dock.collapse'),
            collapseHandle = options.hash.noCollapse ? '' : link.call(this, '#', collapseHandleString, {
                hash : { 'class' : 'collapse-handle', 'aria-labelledby' : titleId },
                fn : function () { return renderer.renderElementWithContent('span', { 'class' : 'spoken-text' }, null, collapseHandleString); }
            }),
            header = dockHeader.call(this, title, titleId, options.hash.noOrientation, initialOrientation),
            body = renderer.renderElementWithContent('div', {'class' : 'dock-body'}, null, innerText);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['dock', orientationClass], options.hash['class']),
            'aria-expanded' : !collapsed,
            'aria-labelledby' : titleId
        }, utils.getElementAttributes(options.hash), resizeHandle + collapseHandle + header + body);
    };

    dock.handlebarsType = 'helper';
    dock.handlebarsName = 'dock';

    return dock;

});


define('primutils/templates/helpers/dropdown',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link'
], function (logger, _, utils, renderer, link) {
    

    /**
     * Creates a divider for a dropdown menu.
     *
     * Available as: `\{{divider}}`.
     *
     * Divider is a helper that does not have inner content.
     *
     * Dividers should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * @class templates.helpers.divider
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Divider element
     */
    var divider = function (options) {
        return renderer.renderElement('li', {
            'class' : utils.mergeClasses(['divider'], options.hash['class']),
            'role' : 'separator'
        }, utils.getElementAttributes(options.hash));
    };

    divider.handlebarsType = 'helper';
    divider.handlebarsName = 'divider';

    /**
     * Creates a menu item in a dropdown menu.
     *
     * Available as: `\{{#dropdownItem}}\{{/dropdownItem}}`.
     *
     * Dropdown item is a block helper and requires inner content.
     *
     * Dropdown items should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * Additional attributes passed to the dropdownItem will be applied to the generated <a href="templates.helpers.link.html">Link `<a>`</a> and not the `<li>` with the exception of `disabled`.
     * If you pass `disabled=true`, the `<li>` will have the class 'disabled'.
     * This still does not prevent the menu item from being triggered.
     * This must still be controlled in code.
     * (There is an open issue for handling this automatically in primutils.)
     *
     * If only one arguments are passed, it is assumed to be href.
     * If you want to specify just the item's title, you should pass '#' as the first argument.
     *
     * @class templates.helpers.dropdownItem
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [href] {String} The link where the menu will take the user (Default: '#' - expecting programatic changes)
     * @param [title] {String} The HTML title attribute
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown item element
     */
    var dropdownItem = function (href, title, options) {
        if (arguments.length === 2) {
            options = title;
            title = null;
        }
        if (arguments.length === 1) {
            options = href;
            title = null;
            href = '#';
        }

        logger.assert(!!options.fn, 'Templates: Dropdown item helper must be used as a block helper.');

        var disabled = !!options.hash.disabled,
            linkText = link.call(this, href, title, options);

        return renderer.renderElementWithContent('li', {
            'class' : disabled ? 'disabled' : undefined,
            'role' : 'menuitem'
        }, null, linkText);
    };

    dropdownItem.handlebarsType = 'helper';
    dropdownItem.handlebarsName = 'dropdownItem';

    /**
     * Creates a submenu inside a dropdown menu.
     *
     * Available as: `\{{#dropdownSubmenu}}\{{/dropdownSubmenu}}`.
     *
     * Dropdown submenu item is a block helper and requires inner content.
     *
     * Dropdown submenus should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * Additional attributes passed to the dropdownItem will be applied to the generated <a href="templates.helpers.link.html">Link `<a>`</a> and not the `<li>`.
     *
     * Dropdown submenus to not currently support linking to external pages.
     *
     * @class templates.helpers.dropdownSubmenu
     * @module templates.helpers
     * @protected
     * @constructor
     * @param name {String} The text that will appear on the submenu
     * @param [title] {String} The HTML title attribute
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown submenu element
     */

    /**
     * Causes the submenu to open to the left instead of to the right.
     *
     * @property openLeft
     * @type {Boolean}
     * @default false
     */

    /**
     * Allows the developer to add classes to the generated `<li>` instead of the `<a>`.
     *
     * @property liClass
     * @type {String}
     * @default ''
     */

    var dropdownSubmenu = function (name, title, options) {
        logger.assert(arguments.length >= 2, 'Templates: Dropdown submenu helper requires a "name" argument');

        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-submenu-item-');
        options.hash['aria-hasPopup'] = 'true';

        var submenuId = _.uniqueId('pgbu-dropdown-menu-');
        options.hash['aria-owns'] = submenuId;

        logger.assert(!!options.fn, 'Templates: Dropdown submenu helper must be used as a block helper.');

        var that = this,
            linkText = link.call(this, '#', title, {
                hash : _.omit(options.hash, 'openLeft', 'liClass'),
                fn : function () {
                    return name;
                }
            }),
            listText = dropdownMenu.call(this, {
                hash : {
                    'id' : submenuId,
                    'aria-labelledby' : options.hash.id
                },
                fn : function () {
                    return options.fn(that);
                }
            }),
            liClasses = (options.hash.liClass || '') + (options.hash.openLeft ? ' pull-left' : '');

        return renderer.renderElementWithContent('li', {
            'class' : utils.mergeClasses(['dropdown-submenu'], liClasses),
            'role' : 'menuitem'
        }, null, linkText + listText);
    };

    dropdownSubmenu.handlebarsType = 'helper';
    dropdownSubmenu.handlebarsName = 'dropdownSubmenu';

    /**
     * Creates a dropdown menu element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#dropdownMenu}}\{{/dropdownMenu}}`.
     *
     * Dropdown menu is a block helper and requires inner content.
     *
     * Dropdown can have several inner elements:
     *  - <a href="templates.helpers.dropdownItem.html">dropdownItem</a>
     *  - <a href="templates.helpers.dropdownSubmenu.html">dropdownSubmenu</a>
     *  - <a href="templates.helpers.divider.html">divider</a>
     *
     * Example:
     *
     *     \{{#dropdownMenu}}
     *         \{{#dropdownItem}}My Item\{{/dropdownItem}}
     *         \{{divider}}
     *         \{{#dropdownSubmenu "Submenu"}}
     *             \{{#dropdownItem}}My Subitem\{{/dropdownItem}}
     *         \{{/dropdownSubmenu}}
     *     \{{/dropdownMenu}}
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>,
     *           <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.dropdownMenu
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown menu element
     */
    var dropdownMenu = function (options) {
        logger.assert(!!options.fn, 'Templates: Dropdown menu helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-dropdown-menu-');

        var oldButtonGroupContext;

        if (this.__buttonGroupContext__) {
            oldButtonGroupContext = this.__buttonGroupContext__;
            this.__buttonGroupContext__.hasDropdownMenu = true;
            this.__buttonGroupContext__.menuId = options.hash.id;

            // Need to mask the menu Ids of submenus from the root menu item
            this.__buttonGroupContext__ = {};
        }

        if (this.__tabContext__) {
            this.__tabContext__.menuId = options.hash.id;
        }

        var innerText = options.fn(this);

        if (oldButtonGroupContext) {
            this.__buttonGroupContext__ = oldButtonGroupContext;
        }

        return renderer.renderElementWithContent('ul', {
            'class' : utils.mergeClasses(['dropdown-menu'], options.hash['class']),
            role : 'menu',
            'aria-hidden' : 'true',
        }, utils.getElementAttributes(options.hash), innerText);
    };

    dropdownMenu.handlebarsType = 'helper';
    dropdownMenu.handlebarsName = 'dropdownMenu';

    return {
        dropdownMenu : dropdownMenu,
        dropdownSubmenu : dropdownSubmenu,
        dropdownItem : dropdownItem,
        divider : divider
    };
});


define('primutils/templates/helpers/emptyState',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/button',
    'primutils/utils/language'
], function (logger, _, utils, renderer, icon, button, lang) {
    

    var STATE_ICONS = {
        'error' : 'indicator-error',
        'warning' : 'warning',
        'success' : 'checkmark-checked',
        'info' : '', // needs work
        'in-progress' : 'throbber'
    };

    var makeIcon = function (type) {
        if (type !== 'info') {
            var hash = {};
            if (type === 'in-progress') {
                // TODO get this into Blue Sky or get an actual icon
                hash = {
                    'class' : 'square',
                    'style' : 'color: #fff; background-color: #009C92; border-radius: 50%;'
                };
            }

            return icon.call({}, STATE_ICONS[type], '', {
                hash: hash
            });
        } else {
            // TODO an actual icon eventually
            return renderer.renderElementWithContent('i', {
                'class' : 'pgbu-icon'
            }, null, 'i');
        }
    };

    /**
     * Allow the empty state to be used statically (instead of relatively positioned) in places like modals.
     *
     * @property static
     * @type {Boolean}
     * @default false
     */

    /**
     * Sets the text that appears on the button at the bottom of the empty state.
     *
     * If this property is omitted then no button is generated.
     *
     * @property btnText
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * Add additional classes to the generated button.
     *
     * This property is ignored if the `btnText` property is missing.
     *
     * @property btnClass
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * The ID to give the empty state button.
     *
     * This property is ignored if the `btnText` property is missing.
     *
     * @property btnId
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * Creates an empty static notification.
     *
     * Available as: `\{{#emptyState}}\{{/emptyState}}`.
     *
     * Empty state is a block helper and requires inner content.
     *
     * In order for an empty state to work properly, it must be placed inside of an element with the ".empty-state-parent" class or declare `static=true`.
     *
     * Example:
     *
     *     <div class="empty-state-parent">
     *         ... other content ...
     *
     *         \{{#emptyState 'info' 'Information' btnText="Call to Action..."}}
     *             ... Empty State Message ...
     *         \{{/emptyState}}
     *     </div>
     *
     *     \{{#emptyState 'in-progress' 'In progress' static=true}}
     *         ... Empty State Message ...
     *     \{{/emptyState}}
     *
     * @class templates.helpers.emptyState
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the empty state (info, warning, error, success, in-progress)
     * @param header {String} The header shown at the top of the empty state
     * @param [options] {Object} Handlebars options
     * @returns {string} Empty state element
     */
    var emptyState = function (type, header, options) {
        logger.assert(arguments.length === 3, 'Templates: Empty state helper requires a "type" and "header" argument');

        logger.assert(!!options.fn, 'Templates: Empty state helper must be used as a block helper.');
        logger.assert(_.contains(_.keys(STATE_ICONS), type), 'Templates: Empty state type must be: error, warning, success, info, in-progress');

        var isStatic = (utils.hasClass(options.hash['class'], 'empty-state-static')) || (!!options.hash.static);

        var headerElement = renderer.renderElementWithContent('div', {
            'class' : 'empty-state-header'
        }, null, header);

        var iconElement = makeIcon(type);

        var messageElement = renderer.renderElementWithContent('div', {
            'class' : 'empty-state-message'
        }, null, options.fn(this));

        var buttonElement = '';
        if (!!options.hash.btnText) {
            buttonElement = button.call({}, 'callout', '', {
                hash: {
                    'class' : utils.mergeClasses(['btn-huge'], options.hash.btnClass),
                    'id' : options.hash.btnId
                },
                fn: function (ctx) {
                    return options.hash.btnText;
                }
            });
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['empty-state', type, (isStatic ? 'empty-state-static' : '')], options.hash['class']),
            // TODO add to Blue Sky
            'style' : isStatic ? 'background-color: transparent; position: static; padding-top: 0;' : undefined
        }, utils.getElementAttributes(options.hash), iconElement + headerElement + messageElement + buttonElement);
    };

    emptyState.handlebarsType = 'helper';
    emptyState.handlebarsName = 'emptyState';

    return {
        'emptyState' : emptyState
    };
});


define('primutils/templates/helpers/fieldset',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, utils, renderer) {
    

    var FIELDSET_ATTRIBUTES = ['form', 'name'],
        FIELDSET_BOOL_ATTRIBUTES = ['disabled'];

    /**
     * Specifies the message shown to the user directing them that all the inputs in the section are required.
     * This property is **not** controlled by any contained <a href="templates.helpers.controlGroup.html">\{{#controlGroup}}s</a> or <a href="templates.helpers.input.html">\{{input}}s</a>.
     *
     * @property requiredMessage
     * @type {String|null}
     * @default null
     */

    /**
     * Creates a fieldset element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms">Forms on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#fieldset}}\{{/fieldset}}`.
     *
     * Fieldset is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#fieldset 'My Form Section'}}
     *         ... Form Content ...
     *     \{{/fieldset}}
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/fieldset#HTML_Attributes">W3C Fieldset Attributes</a>.
     *
     * @class templates.helpers.fieldset
     * @module templates.helpers
     * @constructor
     * @param name {String} The label shown at the top of the fieldset
     * @param [options] {Object} Handlebars options
     * @returns {string} Fieldset element
     */
    var fieldset = function (name, options) {
        logger.assert(arguments.length === 2, 'Templates: Fieldset legend name value is required');
        logger.assert(!!options.fn, 'Templates: Fieldset helper must be used as a block helper.');

        var classList = utils.classList(options.hash['class']);

        var legend = renderer.renderElementWithContent('legend', {
            'class' : utils.hasClass(options.hash['class'], 'small') ? 'small' : undefined
        }, null, name);

        var inner = renderer.renderElementWithContent('div', {
            'class' : 'fieldset-content fit-parent'
        }, null, options.fn(this));

        if (options.hash.requiredMessage) {
            classList.push('required');
            legend += renderer.renderElementWithContent('span', {
                'class' : 'required-note'
            }, null, options.hash.requiredMessage);
        }

        return renderer.renderElementWithContent('fieldset', {
            'class' : utils.mergeClasses(['bordered'], _.without(classList, 'small').join(' '))
        }, utils.getElementAttributes(options.hash, FIELDSET_ATTRIBUTES, FIELDSET_BOOL_ATTRIBUTES), legend + inner);
    };

    fieldset.handlebarsType = 'helper';
    fieldset.handlebarsName = 'fieldset';

    return fieldset;
});


define('primutils/templates/helpers/input',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function(logger, _, utils, renderer) {
    

    //Note: 'title' is added here because it's ignored from the global list
    //      Since I don't want to do anything fancy with it now, we'll list it here

    var INPUT_ATTRIBUTES = ['accept', 'alt', 'autocomplete', 'formaction', 'formenctype', 'formmethod', 'formtarget',
            'height', 'list', 'max', 'maxlength', 'min', 'name', 'pattern', 'placeholder', 'size', 'src', 'step', 'title', 'width'],
        INPUT_BOOL_ATTRIBUTES = ['checked', 'disabled', 'formnovalidate', 'multiple', 'readonly', 'required'];

    /**
     * Sets the ID of the internal `<label>` generated for the checkboxes and radio buttons.
     * This option can only be used when the type of input is 'checkbox' or 'radio' or with the \{{checkbox}} or \{{radio}} helper.
     *
     * When set, an additional `<span>` is generated with the ID you provided.
     * This is generally the approach taken when changing the label text programatically.
     *
     *     \{{checkbox "Default Label" id="my-checkbox" labelId="my-checkbox-label}}
     *
     *     <label class="checkbox" id="pgbu-label-X" for="my-checkbox">
     *         <input type="checkbox" id="my-checkbox" aria-labelledby="pgbu-label-X" />
     *         <span id="my-checkbox-label">Default Label</span>
     *     </label>
     *
     * If you need to actually programatically adjust the actual `<label>`, you do a select on the parent of the generated input, e.g.
     *
     *     $('#my-checkbox').parent()
     *
     * @property labelId
     * @type {String|null}
     * @default null
     */

    /**
     * Places a small description below the input.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms&subtopic=Hints">Hint text on the Blue Sky UX Guide</a>.
     *
     * @property hint
     * @type {String|null}
     * @default null
     */

    /**
     * Specifies the ID attribute of the generated hint text.
     * Used in cases where the page author needs to access the hint text once its created.
     *
     * If not specified, it will receive an auto-incrementing ID so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property hintId
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Hides the hint text by default using the Bootstrap class `hidden`.
     * Useful for error messages.
     *
     * Change the visibility by using:
     *
     *     $hint.toggleClass('hidden', !shouldShow);
     *
     * See also <a href="#property_hintNotDisplayed">hintNotDisplayed</a>.
     * It is *not* recommended to use both hintHidden and hintNotDisplayed.
     *
     * @property hintHidden
     * @type {Boolean}
     * @default false
     */

    /**
     * Hides the hint text by default using `style="display:none;"`.
     * Useful for error messages.
     *
     * Change the visibility by using:
     *
     *     $hint.toggle(shouldShow);
     *
     * See also <a href="#property_hintNotDisplayed">hintNotDisplayed</a>.
     * It is *not* recommended to use both hintHidden and hintNotDisplayed.
     *
     * @property hintNotDisplayed
     * @type {Boolean}
     * @default false
     */

    /**
     * Places the icon inside a radio button or checkbox input to the left of the input instead of the right.
     *
     * This only applies to checkbox and radio button inputs. This overrides the value from the control group if one was passed in.
     *
     * @property leftIcon
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates an input element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=InputTypes">Inputs on the Blue Sky UX Guide</a>.
     * **Note:** This helper does not automatically create <a href="http://10.140.183.183:8080/view/Inputs%20(OUI)/job/oui/javadoc/">OUI elements</a>.
     * You will still need to do that through javascript.
     *
     * Available as: `\{{input}}`.
     *
     * Input can be used as either a regular helper or a block helper.
     *
     * Examples:
     *
     *     \{{input 'text' 'Default Value'}}
     *
     *     \{{input 'text' value="Default Value"}}
     *
     *     \{{#input 'text'}}Default Value\{{/input}}
     *
     * You can use any of the <a href="http://www.w3.org/wiki/HTML/Elements/input">HTML 5 input types</a>.
     * You can also specify `'textarea'` to create a text area.
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/input#HTML_Attributes">W3C Input Attributes</a>.
     * The helper is not smart enough to determine which attributes are valid for which types.
     * Use other attributes at your own peril.
     *
     * Checkboxes and radio buttons, in addition to specifying them with the type parameter, can be created with:
     *
     *     \{{checkbox 'Checkbox Label' 'checkbox_value'}}
     *     \{{checkbox 'Checkbox Label'}}checkbox_value\{{/checkbox}}
     *     \{{radio 'Radio Label'}}
     *
     * If created using the `\{{input 'checkbox'}}` syntax, you must specify the input's value explicitly
     *
     *     \{{#input 'radio' 'radio_value'}}Radio Label\{{/input}}
     *
     * **Important**:
     * If you attempt to hide a checkbox or a radio button using either the `hidden="hidden"` attribute or with
     * `class="hidden"`, neither or these will be applied to the `input`. Instead, a `class="hidden"` will be applied
     * to the surrounding label. This means that when you later want to unhide the checkbox or radio button, you
     * must remove the `hidden` class from that label instead.
     *
     * Generally, this can be accomplished with:
     *
     *     $('#myRadioButton').closest('.radio').toggleClass('hidden');
     *     $('#myCheckbox').closest('.checkbox').toggleClass('hidden');
     *
     * Textareas can also be created using:
     *
     *     \{{#textarea}}Value\{{/textarea}}
     *
     * See also: <a href="templates.helpers.controlGroup.html">\{{#controlGroup}}</a>
     *
     * @class templates.helpers.input
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the input
     * @param [value] {String} The default value of the input (Default: empty)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var input = function (type, value, options) {
        logger.assert(arguments.length >= 2, 'Templates: Input helper requires a "type" argument');

        if (arguments.length === 2) {
            options = value;
            value = null;
        }

        if (_.contains(['radio', 'checkbox'], type) && !options.hash._receivedLabel) {
            return checkboxRadioBase.call(this, type, options.fn(this), value, options);
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-input-');

        // Prefer inner text over specified value
        value = (options.fn && (!options.hash._receivedLabel || value == null)) ? options.fn(this) : value;

        // See controlGroup.js helper for details on the controlGroupContext
        if (this.__controlGroupContext__) {
            options.hash['aria-labelledby'] = this.__controlGroupContext__.labelId;
            options.hash.name = options.hash.name || this.__controlGroupContext__.name;
            if (options.hash.id) {
                this.__controlGroupContext__.inputs.push(options.hash.id);
            }
            if (options.hash.required) {
                this.__controlGroupContext__.inputRequired = this.__controlGroupContext__.inputRequired || !!options.hash.required;
            }
            if (type === 'radio') {
                this.__controlGroupContext__.isRadioGroup = true;
            }
        }

        var input,
            hint = '',
            hintId = options.hash.hintId || _.uniqueId('pgbu-hint-text-');

        if (options.hash.hint != null) {
            hint = renderer.renderElementWithContent('span', {
                'class' : 'help-block' + (options.hash.hintHidden ? ' hidden' : ''),
                'style' : options.hash.hintNotDisplayed ? 'display: none;' : undefined,
                id : hintId
            }, null, options.hash.hint);
        }

        if (type !== 'textarea') {
            input = renderer.renderSelfClosingElement('input', {
                type: type,
                value: value,
                'class': utils.mergeClasses([], options.hash['class']), // Unnecessary initially, but if we ever have default classes it's ready
                'aria-required': (!!options.hash.required).toString(),
                'aria-describedby': (options.hash.hint != null) ? hintId : undefined,
                'role' : (type === 'radio') ? 'radio' : undefined
            }, utils.getElementAttributes(options.hash, INPUT_ATTRIBUTES, INPUT_BOOL_ATTRIBUTES));
        } else {
            input = renderer.renderElementWithContent('textarea', {
                'class': utils.mergeClasses([], options.hash['class']), // Unnecessary initially, but if we ever have default classes it's ready
                'aria-required': (!!options.hash.required).toString(),
                'aria-describedby': (options.hash.hint != null) ? hintId : undefined
            }, utils.getElementAttributes(options.hash, _.union(INPUT_ATTRIBUTES, ['rows', 'cols']), INPUT_BOOL_ATTRIBUTES), value ? value : '');
        }

        return input + hint;
    };

    input.handlebarsType = 'helper';
    input.handlebarsName = 'input';

    var checkboxRadioBase = function (type, label, value, options) {
        logger.assert(arguments.length >= 3, 'Templates: Checkbox/Radio helper requires a "label" argument');

        if (arguments.length === 3) {
            options = value;
            value = null;
        }

        var hasLabelId = !!options.hash.labelId;
        options.hash.outerLabelId = _.uniqueId('pgbu-label-');
        options.hash.labelId = options.hash.labelId || _.uniqueId('pgbu-label-');

        if (hasLabelId) {
            label = renderer.renderElementWithContent('span', {
                id : options.hash.labelId
            }, null, label);
        }

        var oldControlGroupContext = this.__controlGroupContext__;
        this.__controlGroupContext__ = { inputs : [],
            labelId : options.hash.outerLabelId,
            inputRequired : false,
            name : oldControlGroupContext && oldControlGroupContext.name
        };

        options.hash = _.omit(options.hash, 'hint');
        options.hash._receivedLabel = true;

        var inputText = input.call(this, type, value, options),
            inputIsHidden = utils.hasClass(options.hash.class, 'hidden') || !!options.hash.hidden,
            leftIcons = (!!options.hash.leftIcon) || (oldControlGroupContext && oldControlGroupContext.leftIcons);

        // Remove the hidden-y things
        options.hash.class = utils.mergeClasses([_.without(utils.classList(options.hash.class), 'hidden')]);
        options.hash.hidden = false;

        var result = renderer.renderElementWithContent('label', {
            'class' : type + (options.hash.disabled ? ' disabled ' : '') + (inputIsHidden ? ' hidden ' : '') + (leftIcons ? ' left-icon ' : ''),
            id : options.hash.outerLabelId,
            for : options.hash.id
        }, null, inputText + label);

        if (oldControlGroupContext) {
            oldControlGroupContext.isRadioGroup = this.__controlGroupContext__.isRadioGroup;
        }
        this.__controlGroupContext__ = oldControlGroupContext;

        return result;
    };

    if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }

    var checkbox = _.partial(checkboxRadioBase, 'checkbox');

    checkbox.handlebarsType = 'helper';
    checkbox.handlebarsName = 'checkbox';

    var radio = _.partial(checkboxRadioBase, 'radio');

    radio.handlebarsType = 'helper';
    radio.handlebarsName = 'radio';

    var textarea = _.partial(input, 'textarea');

    textarea.handlebarsType = 'helper';
    textarea.handlebarsName = 'textarea';

    return {
        input : input,
        checkbox : checkbox,
        radio : radio,
        textarea : textarea
    };
});


define('primutils/templates/helpers/join',['primutils.lodash'], function (_) {
    

    /**
     * Concatenates variables and string expressions together when needed in hashes.
     *
     * Example:
     *
     *     \{{#button 'toolbar' class=(_ myVariableClass " myOtherClass")}}Button\{{/button}}
     *
     * No spaces are inserted into the concatenation.
     * If you need to separate strings, insert an empty string between the two variables
     *
     *     \{{#button 'toolbar' class=(_ myVariableClass " " myOtherVariable)}}Button\{{/button}}
     *
     * @class templates.helpers._
     * @module templates.helpers
     * @constructor
     * @param ... {Array} Strings to join
     * @param [options] {Object} Handlebars options
     * @returns {string} Joined string
     */
    var join = function () {
        return _.initial(_.toArray(arguments)).join('');
    };

    join.handlebarsType = 'helper';
    join.handlebarsName = '_';

    return join;
});


define('primutils/templates/helpers/modal',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/button',
    'primutils/utils/language'
], function (logger, _, utils, renderer, link, icon, button, lang) {
    

    var modalCloseButton = function () {
        return link.call(this, '#', lang.getString('dialogs.close'), {
            hash : {
                'class' : 'modal-close',
                'data-dismiss' : 'modal'
            },
            fn : function () {
                return icon.call(this, 'dialog-close', lang.getString('dialogs.close'), { hash : {} });
            }
        });
    };

    /**
     * Hides the close button on the modal.
     *
     * Make sure that users have other ways of dismissing the modal: buttons in the footer or setting keyboard=true on the modal itself.
     *
     * @property hideClose
     * @for templates.helpers.modalControlBar
     * @type {Boolean}
     * @default false
     */

    /**
     * Sets the control bar in the modal header.
     *
     * Available as: `\{{#modalControlBar}}\{{/modalControlBar}}`.
     *
     * Modal control bar is a block helper.
     *
     * Modal header can only be used inside of an <a href="templates.helpers.modalHeader.html">modalHeader</a>.
     *
     * Unless you specify `hideClose` at either the \{{#modalHeader}} or \{{#modalControlBar}} level, the close button will be added for you automatically.
     *
     * @class templates.helpers.modalControlBar
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Modal control bar element
     */
    var modalControlBar = function (options) {
        logger.assert(this.__modalContext__, 'Templates: Modal control bar helper must be called from within a {{#modal}}');
        logger.assert(this.__modalContext__.__headerContext__, 'Templates: Modal control bar helper must be called from within a {{#modalHeader}}');
        logger.assert(!!options.fn, 'Templates: Modal control bar helper must be used as a block helper.');
        this.__modalContext__.__headerContext__.hasControlBar = true;
        var hideClose = this.__modalContext__.__headerContext__.hideClose = this.__modalContext__.__headerContext__.hideClose || !!options.hash.hideClose;

        var innerContent = options.fn(this);
        if (!hideClose) {
            innerContent += modalCloseButton.call(this);
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['control-bar'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), innerContent);
    };

    modalControlBar.handlebarsType = 'helper';
    modalControlBar.handlebarsName = 'modalControlBar';

    /**
     * Hides the close button on the modal.
     *
     * Make sure that users have other ways of dismissing the modal: buttons in the footer or setting keyboard=true on the modal itself.
     *
     * @property hideClose
     * @for templates.helpers.modalHeader
     * @type {Boolean}
     * @default false
     */

    /**
     * Hides the title on the modal.
     *
     * @property hideTitle
     * @for templates.helpers.modalHeader
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a modal header.
     *
     * Available as: `\{{#modalHeader}}\{{/modalHeader}}`.
     *
     * Modal header can be used as either a regular helper or a block helper.
     *
     * Modal header can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
     *
     * As a non-block helper, the modal header will insert the title of the modal unless `hideTitle` is set.
     *
     * As a block helper, any content passed inside the block will be rendered *after* the title.
     *
     * @class templates.helpers.modalHeader
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Modal header element
     */
    var modalHeader = function (options) {
        logger.assert(this.__modalContext__, 'Templates: Modal header must be called from within a {{#modal}}');
        this.__modalContext__.hasHeader = true;

        this.__modalContext__.__headerContext__ = {
            hideClose : !!options.hash.hideClose,
            hideTitle : !!options.hash.hideTitle,
            hasControlBar : false
        };

        var innerContent = '',
            fnContent = '';

        if (!this.__modalContext__.__headerContext__.hideTitle) {
            innerContent += renderer.renderElementWithContent('h3', {
                id : this.__modalContext__.labeledBy,
                'class' : 'modal-title'
            }, null, this.__modalContext__.title);
        }

        if (options.fn) {
            fnContent = options.fn(this);
        }

        if (!this.__modalContext__.__headerContext__.hasControlBar && !this.__modalContext__.__headerContext__.hideClose) {
            innerContent += modalCloseButton.call(this);
        }

        innerContent += fnContent;

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['modal-header'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), innerContent);
    };

    modalHeader.handlebarsType = 'helper';
    modalHeader.handlebarsName = 'modalHeader';

    /**
     * Creates a modal body.
     *
     * Available as: `\{{#modalBody}}\{{/modalBody}}`.
     *
     * Modal body is a block helper and requires inner content.
     *
     * Modal body can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
     *
     * @class templates.helpers.modalBody
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Modal body element
     */
    var modalBody = function (options) {
        logger.assert(this.__modalContext__, 'Templates: Modal header must be called from within a {{#modal}}');
        logger.assert(!!options.fn, 'Templates: Modal body helper must be used as a block helper.');

        this.__modalContext__.hasBody = true;

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['modal-body'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), options.fn(this));
    };

    modalBody.handlebarsType = 'helper';
    modalBody.handlebarsName = 'modalBody';

    /**
     * Specifies whether the auto generated buttons should come before or after the inner block content.
     *
     * @property buttonsLast
     * @for templates.helpers.modalFooter
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a modal footer.
     *
     * Available as: `\{{#modalFooter}}\{{/modalFooter}}`.
     *
     * Modal footer can be used as either a regular helper or a block helper.
     *
     * Modal footer can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
     *
     * You can either specify your own content, placing it inside of the block helper, or
     * you can use the default buttons parameter to create dummy buttons in the footer that will close the modal.
     *
     * Valid button names: 'ok', 'cancel'
     *
     * @class templates.helpers.modalFooter
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [defaultButtons] {String} A comma delimited string of buttons to create in the footer
     * @param [options] {Object} Handlebars options
     * @returns {string} Modal footer element
     */
    var modalFooter = function (defaultButtons, options) {
        logger.assert(this.__modalContext__, 'Templates: Modal header must be called from within a {{#modal}}');
        this.__modalContext__.hasFooter = true;

        var that = this,
            buttonText = '',
            innerText = '';

        if (arguments.length === 1) {
            options = defaultButtons;
            defaultButtons = '';
        }

        _.each(defaultButtons.split(','), function (buttonName) {
            buttonName = buttonName.trim();
            switch(buttonName) {
                case 'ok':
                    buttonText += button.call(that, 'primary', lang.getString('dialogs.ok'), {
                        hash : {
                            'data-dismiss' : 'modal'
                        },
                        fn : function () { return lang.getString('dialogs.ok'); }
                    });
                    break;
                case 'cancel':
                    buttonText += button.call(that, 'secondary', lang.getString('dialogs.cancel'), {
                        hash : {
                            'data-dismiss' : 'modal'
                        },
                        fn : function () { return lang.getString('dialogs.cancel'); }
                    });
                    break;
            }
        });

        if (options.fn) {
            if (options.hash.buttonsLast) {
                innerText += options.fn(this) + buttonText;
            } else {
                innerText += buttonText + options.fn(this);
            }
        } else {
            innerText = buttonText;
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['modal-footer'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), innerText);
    };

    modalFooter.handlebarsType = 'helper';
    modalFooter.handlebarsName = 'modalFooter';

    /**
     * Let's assistive applications know that this dialog does not have any user editable content and is solely a displaying an alert.
     * Most of the shared dialogs (with the exception of ConfirmDialog) would be alert=true.
     *
     * @property alert
     * @for templates.helpers.modal
     * @type {Boolean}
     * @default false
     */

    /**
     * Controls how the backdrop behaves with the modal.
     * See <a href="http://getbootstrap.com/2.3.2/javascript.html#modals">Bootstrap modals</a>.
     *
     * Valid values: true, false, 'static'
     *
     * @property backdrop
     * @for templates.helpers.modal
     * @type {String|Boolean}
     * @default 'static'
     */

    /**
     * Controls whether escape can be used to close the modal
     * See <a href="http://getbootstrap.com/2.3.2/javascript.html#modals">Bootstrap modals</a>.
     *
     * Valid values: true, false
     *
     * @property keyboard
     * @for templates.helpers.modal
     * @type {Boolean}
     * @default true
     */

    /**
     * Creates a modal element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs">Dialogs on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#modal}}\{{/modal}}`.
     *
     * Modal is a block helper and requires inner content.
     *
     * Modals should have several inner elements:
     *  - <a href="templates.helpers.modalHeader.html">modalHeader</a>
     *  - <a href="templates.helpers.modalBody.html">modalBody</a>
     *  - <a href="templates.helpers.modalFooter.html">modalFooter</a>
     *
     * Example:
     *
     *     \{{#modal 'My Sample Modal' alert=true}}
     *         \{{#modalHeader}}
     *             Additional Header Content
     *         \{{/modalHeader}}
     *         \{{#modalBody}}
     *             ... content ...
     *         \{{/modalBody}}
     *         \{{#modalFooter}}
     *             Footer content
     *         \{{/modalFooter}}
     *     \{{/modal}}
     *
     * If a modalHeader or modalFooter section is skipped inside the modal block, the default header and footer will be inserted automatically.
     * If no modalHeader, modalBody, or modalFooter section is found, the inner content of the modal will automatically be placed inside a modalBody section, and
     * the default header and footers appended automatically.
     * The default header includes a title and a close button.
     * The default footer includes an "Ok" button.
     *
     * @class templates.helpers.modal
     * @module templates.helpers
     * @constructor
     * @param title {String} The title of the modal
     * @param [options] {Object} Handlebars options
     * @returns {string} Modal element
     */
    var modal = function (title, options) {
        if (arguments.length === 1) {
            options = title;
            title = '';
        }

        logger.assert(!!options.fn, 'Templates: Modal helper must be used as a block helper.');

        if (!options.hash.id) {
            options.hash.id = _.uniqueId('pgbu-modal-');
        }

        var oldModalContext = this.__modalContext__,
            modalContext = {
                id : options.hash.id,
                labeledBy : options.hash.id + '-label',
                type : options.hash.alert ? 'alertdialog' : 'dialog',
                title : title
            };

        this.__modalContext__ = modalContext;

        var innerText = options.fn ? options.fn(this) : '',
            bodyConverted = false;

        // If there's nothing from the context, assume that the whole thing is for the body
        if (!modalContext.hasHeader && !modalContext.hasFooter && !modalContext.hasBody) {
            bodyConverted = true;
            innerText = modalBody.call(this, {
                hash : {},
                fn : function () {
                    return innerText;
                }
            });
        }

        if (!modalContext.hasHeader) {
            logger.info('Modal Template: Template did not specify a modal-header, adding default header');
            innerText = modalHeader.call(this, { hash : {} }) + innerText;
        }

        if (!modalContext.hasFooter) {
            logger.info('Modal Template: Template did not specify a modal-footer, adding default footer');
            innerText = innerText + modalFooter.call(this, bodyConverted ? 'ok' : '', { hash : {} });
        }

        this.__modalContext__ = oldModalContext;

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['modal', 'pgbu-modal', 'hide', 'fade'], options.hash['class']),
            tabindex : '-1',
            role : modalContext.type,
            'aria-hidden' : 'true',
            'aria-labelledby' : modalContext.labeledBy,
            'data-backdrop' : _.has(options.hash, 'backdrop') ? options.hash.backdrop.toString() : 'static',
            'data-keyboard' : _.has(options.hash, 'keyboard') ? options.hash.keyboard.toString() : 'true'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    modal.handlebarsType = 'helper';
    modal.handlebarsName = 'modal';

    return {
        modalHeader : modalHeader,
        modalControlBar : modalControlBar,
        modalBody : modalBody,
        modalFooter : modalFooter,
        modal : modal
    };

});


define('primutils/templates/helpers/panel',[
    'primutils.logger',
    'primutils.lodash',
    'primutils.handlebars',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/caret',
    'primutils/utils/language'
], function (logger, _, Handlebars, utils, renderer, link, icon, caret, lang) {
    

    var PANEL_TYPE_MAP = {
            'detail' : 'nav nav-tabs detail-panel-tabs',
            'page' : 'nav nav-tabs with-notch',
            'content' : 'nav nav-tabs content-tabs',
            'scenario' : 'nav nav-tabs scenario-tabs',
            'wizard' : 'steps'
        },
        PANEL_TYPE_NO_NAV_MAP = {
            'detail' : 'detail-panel-tabs',
            'page' : 'with-notch',
            'content' : 'content-tabs',
            'scenario' : 'scenario-tabs',
            'wizard' : 'steps'
        },
        PANEL_PANELS_MAP = {
            'detail' : 'detail-panel',
            'page' : 'page-panel',
            'content' : 'content-panel',
            'scenario' : 'scenario-panel',
            'wizard' : 'wizard-panel'
        };

    /**
     * Creates a panel.
     *
     * Available as: `\{{#panel}}\{{/panel}}`.
     *
     * Panel is a block helper and requires inner content.
     *
     * Panel can only be used inside of an <a href="templates.helpers.panels.html">panels</a>.
     *
     * If the ID parameter is omitted the panel element will not function with the <a href="http://getbootstrap.com/2.3.2/javascript.html#tabs">Bootstrap tab widget</a>.
     *
     * @class templates.helpers.panel
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [id] {String} The HTML id of the panel
     * @param [options] {Object} Handlebars options
     * @returns {string} The panel element
     */
    var panel = function (id, options) {
        logger.assert(this.__panelContext__, 'Templates: Tab helper must be called from within a {{#panels}}');

        var active = '';

        if (arguments.length === 1) {
            options = id;
            id = null;
            logger.warn('Panel was not given an ID. It may never be visible.');
        }

        logger.assert(!!options.fn, 'Templates: Panel helper must be used as a block helper.');

        if (id) {
            this.__panelContext__.panelsCreated.push(id);

            if (this.__panelContext__.activeTab === id) {
                active = 'active';
            }
        }

        if (!this.__panelContext__.noSizePanels) {
            if (!options.hash.style) {
                options.hash.style = 'height: 100%;';
            } else {
                options.hash.style = options.hash.style + '; height: 100%;';
            }
        }

        return renderer.renderElementWithContent('div', {
            'id' : id,
            'class' : utils.mergeClasses(['tab-pane', active], options.hash['class']),
            'role' : 'tabpanel'
        }, utils.getElementAttributes(options.hash), options.fn(this));
    };

    panel.handlebarsType = 'helper';
    panel.handlebarsName = 'panel';

    /**
     * Creates a collection of panels.
     *
     * Available as: `\{{#panels}}\{{/panels}}`.
     *
     * Panels is a block helper and requires inner content.
     *
     * Panels can only be used inside of an <a href="templates.helpers.panelContainer.html">panelContainer</a>.
     *
     * Panels should contain several <a href="templates.helpers.panel.html">panel</a> elements.
     *
     * **Note:** The panels section should be created *after* the tabList.
     *
     * @class templates.helpers.panels
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Collection of panels element
     */
    var panels = function (options) {
        logger.assert(this.__panelContext__, 'Templates: Tab helper must be called from within a {{#panelContainer}}');
        logger.assert(!!options.fn, 'Templates: Panels helper must be used as a block helper.');

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['tab-content', PANEL_PANELS_MAP[this.__panelContext__.type]], options.hash['class'])
        }, utils.getElementAttributes(options.hash), options.fn(this));
    };

    panels.handlebarsType = 'helper';
    panels.handlebarsName = 'panels';

    /**
     * Creates a menu button for a tab.
     *
     * Available as: `\{{tabMenuButton}}`.
     *
     * Tab menu button is a helper that does not have inner content.
     *
     * Tab menu button can only be used inside of an <a href="templates.helpers.tab.html">tab</a>.
     *
     * Menus are only supported in 'content' and 'scenario' tabs.
     *
     * @class templates.helpers.tabMenuButton
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab menu button element
     */
    var tabMenuButton = function (options) {
        logger.assert(this.__panelContext__, 'Templates: Tab Menu Button helper must be called from within a {{#tab}}');
        logger.assert(this.__tabContext__, 'Templates: Tab Menu Button helper must be called from within a {{#tab}}');
        logger.assert(_.contains(['content', 'scenario'], this.__panelContext__.type), 'Templates: Menus are only supported in content and scenario tabs');

        this.__tabContext__.hasMenu = true;

        var linkHash = _.extend({}, options.hash, {
            'data-toggle' : 'dropdown',
            'aria-haspopup' : 'true',
            'aria-owns' : '{{menuId}}'
        });

        return link.call(this, '#', null, {
            hash : linkHash,
            fn : function () { return caret.call(this); }
        });
    };

    tabMenuButton.handlebarsType = 'helper';
    tabMenuButton.handlebarsName = 'tabMenuButton';

    /**
     * Creates a notification bubble for a tab.
     *
     * Available as: `\{{#notification}}\{{/notification}}`.
     *
     * Notification is a block helper and requires inner content.
     *
     * Notification should only be used inside of an <a href="templates.helpers.tab.html">tab</a>.
     *
     * Notifications are only supported in 'content' and 'scenario' tabs.
     *
     * @class templates.helpers.notification
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Notification bubble element
     */
    var notification = function (options) {
        logger.assert(this.__panelContext__ && _.contains(['content', 'scenario'], this.__panelContext__.type), 'Templates: Notifications are only supported in content and scenario tabs');
        logger.assert(!!options.fn, 'Templates: Notification helper must be used as a block helper.');

        return renderer.renderElementWithContent('span', {
            'class' : utils.mergeClasses(['notifications'], options.hash['class'])
        }, utils.getElementAttributes(options.hash), options.fn(this));
    };

    notification.handlebarsType = 'helper';
    notification.handlebarsName = 'notification';

    /**
     * When used as an option on the block helper, it places the contents of the block inside the `<a>` instead of the `<li>`.
     *
     * @property isAnchorContent
     * @for templates.helpers.tab
     * @type {Boolean}
     * @default false
     */

    /**
     * Allows you to specify a different href than tab target.
     *
     * (Advanced feature)
     *
     * @property href
     * @for templates.helpers.tab
     * @type {Boolean|null}
     * @default null
     */

    /**
     * Creates a tab.
     *
     * Available as: `\{{#tab}}\{{/tab}}`.
     *
     * Tab can be used as either a regular helper or a block helper.
     *
     * Tab can only be used inside of an <a href="templates.helpers.tabList.html">tabList</a>.
     *
     * If the controls parameter is omitted the panel element will not function with the <a href="http://getbootstrap.com/2.3.2/javascript.html#tabs">Bootstrap tab widget</a>.
     *
     * You can specify `class="active"` to make sure that this tab and its corresponding panel are default visible when executing the template.
     *
     * Tabs can contain the following elements:
     *  - <a href="templates.helpers.notification.html">notification</a>
     *  - <a href="templates.helpers.tabMenuButton.html">tabMenuButton</a>
     *  - <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>
     *
     * All attributes will be applied to the underlying anchor (`<a>`) tag
     * with the exception of four clases: 'header', 'active', 'disabled', 'dropdown'.
     * If specified, these classes will appear on the list item (`<li>`).
     *
     * When used as a block helper, you can specify content that will appear after the tab link.
     * This is most often used for either notifications or dropdown menus.
     *
     * @class templates.helpers.tab
     * @module templates.helpers
     * @protected
     * @constructor
     * @param name {String} The name of the tab
     * @param [controls] {String} The HTML id of the panel
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab element
     */
    var tab = function (name, controls, options) {
        logger.assert(this.__panelContext__, 'Templates: Tab helper must be called from within a {{#tabList}}');
        logger.assert(arguments.length >= 2, 'Templates: Tab name is required');

        if (arguments.length === 2) {
            options = controls;
            controls = null;
        }

        var classList = utils.classList(options.hash['class']),
            linkName,
            isActive = false,
            isAnchorContent = options.hash.isAnchorContent,
            innerText;

        if (options.hash.href) {
            linkName = options.hash.href;
        } else if (controls) {
            linkName = '#' + controls;
        } else {
            linkName = '#';
        }

        var linkHash = _.extend({}, _.omit(options.hash, 'href'), {
            '_internalRole' : 'tab',
            'aria-controls' : controls
        });
        linkHash['class'] = _.without(classList, 'header', 'active', 'disabled').join(' ');

        if (controls) {
            linkHash['data-toggle'] = 'tab';
            linkHash['data-target'] = '#' + controls;

            this.__panelContext__.tabsCreated.push(controls);

            if (_.contains(classList, 'active')) {
                if (this.__panelContext__.activeTab) {
                    logger.error('More than one tab was selected as the active tab, current active: ' + this.__panelContext__.activeTab + ', new tab: ' + controls);
                    classList = _.without(classList, 'active');
                } else {
                    this.__panelContext__.activeTab = controls;
                    isActive = true;
                }
            }
        }

        linkHash['aria-selected'] = isActive;

        if (_.contains(classList, 'header') || this.__panelContext__.type === 'wizard') {
            innerText = name;
        } else {
            innerText = link.call(this, linkName, name, {
                hash: linkHash,
                fn: function () {
                    return (isAnchorContent && options.fn) ? options.fn(this) : name;
                }
            });
        }

        if (options.fn) {
            this.__tabContext__ = {};

            if (!isAnchorContent) {
                innerText += options.fn(this);
            }

            if (this.__tabContext__.hasMenu) {
                classList.push('dropdown');
                innerText = innerText.replace(/\{\{menuId\}\}/g, this.__tabContext__.menuId);
            }

            delete this.__tabContext__;
        }

        return renderer.renderElementWithContent('li', {
            'class' : _.intersection(classList, ['header', 'active', 'disabled', 'dropdown']).join(' '),
            'style' : !!options.hash.hidden ? 'display: none;' : undefined,
            'data-wizard-step' : this.__panelContext__.type === 'wizard' ? controls : undefined
        }, null, innerText);
    };

    tab.handlebarsType = 'helper';
    tab.handlebarsName = 'tab';

    var scrollers = function (scrollClass) {
        var scroller = function (clz, title) {
            return link.call(this, '#', title, {
                hash : { 'class' : clz, tabindex : '-1' },
                fn : function () {
                    return renderer.renderElementWithContent('span', {
                        'class' : 'spoken-text'
                    }, null, title);
                }
            });
        };

        return renderer.renderElementWithContent('span', {
            'class' : 'scrollers'
        }, null, scroller(scrollClass, lang.getString(scrollClass === 'left' ? 'panel.scrollLeft' : 'panel.scrollRight')));
    };

    var collapseHandle = function () {
        var collapseStr = lang.getString('panel.collapse');
        return renderer.renderElementWithContent('span', {
            'class' : 'collapse-handle'
        }, null, link.call(this, '#', collapseStr, {
            hash : {},
            fn : function () {
                return renderer.renderElementWithContent('i', { 'class' : 'collapse-icon' }, null,
                    renderer.renderElementWithContent('span', { 'class' : 'spoken-text' }, null, collapseStr));
            }
        }));
    };

    var tabsMenu = function () {
        var linkText = link.call(this, '#', lang.getString('panel.tabsMenu'), {
            hash : { 'data-toggle' : 'dropdown' },
            fn : function () {
                return icon.call(this, 'navmenu-expanded', lang.getString('panel.tabsMenu'), {hash:{}});
            }
        });

        return renderer.renderElementWithContent('span', {
            'class' : 'tabs-menu'
        }, null, linkText);
    };

    var tabListSizing = function (options) {
        var mainTabs = tabListNoSizing.call(this, options),
            collapseHandleText = '', beforeTabs, afterTabs;

        if (this.__panelContext__.type === 'detail') {
            if (!options.hash.noCollapse || !this.__panelContext__.noCollapse) {
                collapseHandleText = collapseHandle();
            }
        }

        beforeTabs = renderer.renderElementWithContent('div', {
            'class' : 'before-tabs'
        }, null, tabsMenu() + scrollers('left'));

        afterTabs = renderer.renderElementWithContent('div', {
            'class' : 'after-tabs'
        }, null, scrollers('right') + collapseHandleText);

        return renderer.renderElementWithContent('div', {
            'class' : 'nav-tabs-scroll-container has-before-tabs has-after-tabs ' + PANEL_TYPE_NO_NAV_MAP[this.__panelContext__.type]
        }, null, beforeTabs + mainTabs + afterTabs);
    };

    var tabListNoSizing = function (options) {
        var innerText = options.fn(this);

        return renderer.renderElementWithContent(this.__panelContext__.type === 'wizard' ? 'ol' : 'ul', {
            'class' : utils.mergeClasses([PANEL_TYPE_MAP[this.__panelContext__.type], (this.__panelContext__.numbered ? 'numbered' : '')], options.hash['class']),
            'role' : 'tablist'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    /**
     * Creates a tab list.
     *
     * Available as: `\{{#tabList}}\{{/tabList}}`.
     *
     * Tab list is a block helper and requires inner content.
     *
     * Tabs should list several <a href="templates.helpers.tab.html">tabs</a>.
     *
     * Tab list can only be used inside of an <a href="templates.helpers.panelContainer.html">panelContainer</a>.
     *
     *
     * @class templates.helpers.tabList
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab list element
     */
    var tabList = function (options) {
        logger.assert(this.__panelContext__, 'Templates: Tab List helper must be called from within a {{#panelContainer}}');
        logger.assert(!!options.fn, 'Templates: Tab list helper must be used as a block helper.');

        if (options.hash.noSizing || _.contains(['wizard', 'page'], this.__panelContext__.type)) {
            return tabListNoSizing.call(this, options);
        } else {
            return tabListSizing.call(this, options);
        }
    };

    tabList.handlebarsType = 'helper';
    tabList.handlebarsName = 'tabList';

    /**
     * Prevents a detail panel from automatically creating a collapse button.
     *
     * Only available when type='detail'
     *
     * @property noCollapse
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Skips creating the outer <div class="panel-container">.
     *
     * While you can set this to true, you are still required to use the \{{#panelContainer}} helper to set up the tabs and panels.
     * It creates a context that is shared between the tabs and panels.
     *
     * @property noWrap
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Do not automatically set panels to 100% of the height of the panels contains (&lt;div class="tab-container"&gt;).
     *
     * @property noSizePanels
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a panel container element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation">Panels on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#panelContainer}}\{{/panelContainer}}`.
     *
     * Panel container is a block helper and requires inner content.
     *
     * Panels should have the following inner elements (in this order):
     *  - <a href="templates.helpers.tabList.html">tabList</a>
     *  - <a href="templates.helpers.panels.html">panels</a>
     *
     * Example:
     *
     *     \{{#panelContainer 'detail'}}
     *         \{{#tabList}}
     *             \{{tab 'Foo' 'foo' class="active"}}
     *             \{{tab 'Bar' 'bar'}}
     *         \{{/tabList}}
     *         \{{#panels}}
     *             \{{#panel 'foo'}}
     *                 ...
     *             \{{/panel}}
     *             \{{#panel 'bar'}}
     *                 ...
     *             \{{/panel}}
     *         \{{/panels}}
     *     \{{/panelContainer}}
     *
     * @class templates.helpers.panelContainer
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of panel to create
     * @param [options] {Object} Handlebars options
     * @returns {string} Panel container element
     */
    var panelContainer = function (type, options) {
        logger.assert(arguments.length === 2, 'Templates: Panel helper requires type argument');
        logger.assert(_.has(PANEL_TYPE_MAP, type), 'Templates: Panel type must be one of: ' + JSON.stringify(_.keys(PANEL_TYPE_MAP)));
        logger.assert(!!options.fn, 'Templates: Panel container helper must be used as a block helper.');

        var panelsContext = {
            type : type,
            noCollapse : options.hash.noCollapse,
            noSizePanels : options.hash.noSizePanels,
            numbered : options.hash.numbered,
            tabsCreated : [],
            activeTab : null,
            panelsCreated : []
        };

        var oldPanelContext = this.__panelContext__;
        this.__panelContext__ = panelsContext;

        var innerText = options.fn(this);

        this.__panelContext__ = oldPanelContext;

        var missingTabs = _.difference(panelsContext.panelsCreated, panelsContext.tabsCreated),
            missingPanels = _.difference(panelsContext.tabsCreated, panelsContext.panelsCreated);

        if (missingTabs.length > 0) {
            logger.warn('Panels created without corresponding tab: ' + missingTabs.join(', '));
        }
        if (missingPanels.length > 0) {
            logger.warn('Tabs created without corresponding panel: ' + missingPanels.join(', '));
        }

        if (options.hash.noWrap) {
            return innerText;
        } else {
            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses([type === 'wizard' ? 'wizard-context' : 'panel-container'], options.hash['class'])
            }, utils.getElementAttributes(options.hash), innerText);
        }
    };

    panelContainer.handlebarsType = 'helper';
    panelContainer.handlebarsName = 'panelContainer';

    return {
        panelContainer : panelContainer,
        tabList : tabList,
        tab : tab,
        notification : notification,
        tabMenuButton : tabMenuButton,
        panels : panels,
        panel : panel
    };

});


define('primutils/templates/helpers/search',[
    'primutils.logger',
    'primutils.lodash',
    'primutils.jquery',
    'primutils/utils/language',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/input'
], function (logger, _, $, lang, utils, renderer, link, icon, input) {
    

    /**
     * Makes the search box take up the full width of its container
     *
     * @property fluid
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a search element.
     * See <a href="http://localhost/style-guide/?section=web&topic=Navigation&subtopic=LeftHandInPageNavigation">Search on the Blue Sky UX Guide</a>.
     *
     * Input can be used as either a regular helper or a block helper.
     *
     * Examples:
     *
     *     \{{search 'value'}}
     *
     *     \{{search}}Value\{{/value}}
     *
     * See also: <a href="templates.helpers.input.html">\{{#input}}</a>
     *
     * @class templates.helpers.search
     * @module templates.helpers
     * @constructor
     * @param [value] {String} The default value of the search
     * @param [options] {Object} Handlebars options
     * @returns {string} Search element
     */
    var search = function (value, options) {
        if (arguments.length === 1) {
            options = value;
            value = null;
        }

        // Prefer value from block helper body if available
        value = options.fn ? options.fn(this) : value;

        var innerText = '',
            inputHash = _.omit(options.hash, 'class', 'fluid', 'style'),
            outerHash = _.pick(options.hash, 'class', 'style');

        inputHash['class'] = 'filterbox search-query';

        innerText += input.input.call(this, 'search', value, {
            hash : inputHash
        });

        innerText += renderer.renderElementWithContent('span', {
            'class' : 'search-icon'
        }, null, icon.call(this, 'magnifying-glass', lang.getString('search.search'), {hash:{ 'size' : 'pgbu-icon-small' }}));

        innerText += link.call(this, '#', lang.getString('search.clear'), {
            hash : { 'class' : 'search-clear-icon' },
            fn : function () {
                return icon.call(this, 'find-clear', lang.getString('search.clear'), {hash:{}});
            }
        });

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['filter-div', options.hash.fluid ? 'filter-div-fluid' : ''], options.hash['class'])
        }, utils.getElementAttributes(outerHash), innerText);
    };

    search.handlebarsType = 'helper';
    search.handlebarsName = 'search';

    return search;
});


define('primutils/templates/helpers/slider',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/button'
], function (logger, _, utils, renderer, button) {
    

    /**
     * Creates a toggle switch element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ToggleControl">Slider on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{slider}}`.
     *
     * Slider is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{slider 'On' 'Off'}}
     *
     * See also: <a href="templates.helpers.toggle.html">\{{#toggle}}</a>
     *
     * @class templates.helpers.slider
     * @module templates.helpers
     * @constructor
     * @param onText {String} The screen reader spoken text when the button is on
     * @param offText {String} The screen reader spoken text when the button is off
     * @param [options] {Object} Handlebars options
     * @returns {string} Slider element
     */
    var slider = function (onText, offText, options) {
        logger.assert(arguments.length === 3, 'Templates: Slider helper requires on/off text arguments');

        return button.call(this, 'slider', '', {
            hash : _.extend(options.hash, { 'data-toggle' : 'button' }),
            fn : function () { return ''; }
        });
    };

    slider.handlebarsType = 'helper';
    slider.handlebarsName = 'slider';

    return slider;
});


define('primutils/templates/helpers/spinner',[
    'primutils.logger',
    'primutils.lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, utils, renderer) {
    

    /**
     * Concatenates variables and string expressions together when needed in hashes.
     *
     * Example:
     *
     *     \{{spinner}}
     *
     * @class templates.helpers.spinner
     * @module templates.helpers
     * @constructor
     * @param [size] {String|Number} Size of the spinner in font-compatible units (px, em, etc.)
     * @param [options] {Object} Handlebars options
     * @returns {string} Joined string
     */
    var spinner = function (size, options) {
        if (arguments.length === 1) {
            options = size;
            size = '60px'; // default size
        }

        if (_.isNumber(size)) {
            size = size.toString() + 'px';
        }

        // If we are making this an icon, do not set the size
        if (options.hash['class'] && options.hash['class'].indexOf('pgbu-icon-') > -1) {
            size = null;
        }

        var dots = '';
        _.times(3, function (n) {
            dots += renderer.renderElement('div', { 'class' : 'dot dot' + (n+1) }, {});
        });
        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['pulse-spinner'], options.hash['class']),
            'style' : size != null ? (options.hash.style + ';font-size: ' + size) : undefined
        }, utils.getElementAttributes(options.hash), dots);
    };

    spinner.handlebarsType = 'helper';
    spinner.handlebarsName = 'spinner';

    return spinner;
});


define('primutils/templates/helpers/steps',[
    'primutils.logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function(logger, utils, renderer) {
    

    /**
     * Shows the step numbers of each step.
     * @property numbered
     * @for templates.helpers.stepList
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a stepList element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs&subtopic=StepIndicator">Step Lists on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#stepList}}\{{/stepList}}`.
     *
     * Step list is a block helper and requires inner content.
     * Only <a href="templates.helpers.stepItem.html">\{{#stepItem}}s</a> should be placed inside step lists.
     *
     * Example:
     *
     *     \{{#stepList numbered=true}}
     *         \{{#stepItem}}First Step\{{/stepItem}}
     *         \{{#stepItem}}Second Step\{{/stepItem}}
     *     \{{/stepList}}
     *
     * @class templates.helpers.stepList
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Step list element
     */
    var stepList = function (options) {
        logger.assert(!!options.fn, 'Templates: Step list helper must be used as a block helper.');

        return renderer.renderElementWithContent('ol', {
                class : utils.mergeClasses(['steps', options.hash.numbered ? 'numbered' : ''], options.hash['class'])
            },
            utils.getElementAttributes(options.hash), options.fn(this));
    };

    stepList.handlebarsType = 'helper';
    stepList.handlebarsName = 'stepList';

    /**
     * Creates a stepItem element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs&subtopic=StepIndicator">Step Lists on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#stepItem}}\{{/stepItem}}`.
     *
     * Step item is a block helper and requires inner content.
     *
     * Step items should only be used inside of a <a href="templates.helpers.stepList.html">stepList</a>.
     *
     * @class templates.helpers.stepItem
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Step item element
     */
    var stepItem = function (options) {
        logger.assert(!!options.fn, 'Templates: Step item helper must be used as a block helper.');

        return renderer.renderElementWithContent('li', null, utils.getElementAttributes(options.hash, ['class']), options.fn(this));
    };

    stepItem.handlebarsType = 'helper';
    stepItem.handlebarsName = 'stepItem';

    return {
        stepList : stepList,
        stepItem : stepItem
    };
});


define('primutils/markup/primutils/templates/helpers/termDef.hbs', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/icon'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/templates/helpers/termDef.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"orcl-popover ";
  stack1 = lambda((depth0 != null ? depth0.side : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n    <div class=\"orcl-popover-outer\">\n        ";
  stack1 = helpers.icon.call(depth0, (depth0 != null ? depth0.icon : depth0), (depth0 != null ? depth0.title : depth0), {"name":"icon","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    </div>\n    <div class=\"orcl-popover-inner\">\n        <div class=\"orcl-popover-content\">\n            <div class=\"header\">";
  stack1 = lambda((depth0 != null ? depth0.header : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n            <div>";
  stack1 = lambda((depth0 != null ? depth0.content : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n        </div>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

/**
 * @module templates.helpers
 */
define('primutils/templates/helpers/termDef',['primutils/templates/utils/renderer', 'primutils/markup/primutils/templates/helpers/termDef.hbs', 'primutils.logger'], function (renderer, markup, logger) {
    

    /**
     * Creates a terminology definition.
     *
     * Available as: `\{{#termDef}}\{{/termDef}}`.
     *
     * Terminology definition is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#termDef}}\{{/termDef}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.termDef
     * @constructor
     * @param header {String} Header inside terminology definition
     * @param title {String} Hover text for the terminology definition
     * @param [options] {Object} Handlebars options
     * @returns {string} Caret element
     */
    var termDef = function (header, title, options) {
        if (!require.defined('oui')) {
            logger.warn('Terminology definition created without OUI being loaded. It may not function correctly');
        }
        return markup({
            icon : options.hash.icon || 'popover-help',
            side : options.hash.side || '',
            title : title,
            header : header,
            content : options.fn(this)
        });
    };

    termDef.handlebarsType = 'helper';
    termDef.handlebarsName = 'termDef';

    return termDef;
});


define('primutils/templates/helpers/toggle',[
    'primutils.logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, utils, renderer) {
    

    var TOGGLE_TYPE_MAP = {
        'page' : 'btn-toggle-control',
        'data' : 'btn-toggle-switch'
    };

    /**
     * Creates a toggle button group.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ToggleControl">Toggles on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#toggle}}\{{/toggle}}`.
     *
     * Toggle button group is a block helper and requires inner content.
     * Two button elements should be specified.
     * Their 'type' parameter should be omitted.
     * If you need a title on your buttons, you should pass '' (empty string) as the button type.
     *
     * Example:
     *
     *     \{{#toggle 'page'}}
     *         \{{#button}}EPS View\{{/button}}
     *         \{{#button}}Project View\{{/button}}
     *     \{{/toggle}}
     *
     * There are two types of toggles: 'page' and 'data'.
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>, <a href="templates.helpers.slider.html">\{{slider}}</a>
     *
     * @class templates.helpers.toggle
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of toggle to create
     * @param [options] {Object} Handlebars options
     * @returns {string} Toggle element
     */
    var toggle = function (type, options) {
        logger.assert(arguments.length === 2, 'Templates: Toggle helper requires a "type" argument');
        logger.assert(!!options.fn, 'Templates: Toggle helper must be used as a block helper.');

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-group', TOGGLE_TYPE_MAP[type]], options.hash['class']),
            'data-toggle' : 'buttons-radio'
        }, utils.getElementAttributes(options.hash), options.fn(this));
    };

    toggle.handlebarsType = 'helper';
    toggle.handlebarsName = 'toggle';

    return toggle;
});


define('primutils/templates/helpers/toolbar',[
    'primutils.logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, utils, renderer) {
    

    /**
     * Creates a button toolbar.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=GridActionButtons">Toolbars on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#toolbar}}\{{/toolbar}}`.
     *
     * Toolbar is a block helper and requires inner content, generally <a href="templates.helpers.button.html">\{{#button}}s</a> and <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}s</a>.
     *
     * Example:
     *
     *     \{{#toolbar}}
     *         \{{#button 'toolbar'}}Add\{{/button}}
     *         \{{#buttonGroup}}
     *             \{{#button 'toolbar'}}Edit\{{/button}}
     *             \{{#button 'toolbar'}}Super Edit\{{/button}}
     *         \{{/buttonGroup}}
     *     \{{/toolbar}}
     *
     * **Note:** Buttons inside of a toolbar do not automatically get the 'toolbar' button type.
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>, <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>
     *
     * @class templates.helpers.toolbar
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Toolbar element
     */
    var toolbar = function (options) {
        logger.assert(!!options.fn, 'Templates: Toolbar helper must be used as a block helper.');

        this.__toolbarContext__ = {}; // Placeholder for inline inputs
        var innerText = options.fn(this);
        delete this.__toolbarContext__;

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-toolbar'], options.hash['class']),
            'role' : 'toolbar'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    toolbar.handlebarsType = 'helper';
    toolbar.handlebarsName = 'toolbar';

    return toolbar;
});

define('primutils/templates/helpers/wizard',['primutils.lodash', 'primutils/templates/helpers/button'], function (_, button) {
    

    var wizardBaseButton = function (type) {
        var args = _.rest(_.toArray(arguments)),
            options = _.last(args);

        options.hash['data-wizard'] = type;

        return button.apply(this, args);
    };

    if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }

    var wizardNextButton = _.partial(wizardBaseButton, 'next');
    wizardNextButton.handlebarsType = 'helper';
    wizardNextButton.handlebarsName = 'wizardNextButton';

    var wizardPrevButton = _.partial(wizardBaseButton, 'prev');
    wizardPrevButton.handlebarsType = 'helper';
    wizardPrevButton.handlebarsName = 'wizardPrevButton';

    var wizardCancelButton = _.partial(wizardBaseButton, 'cancel');
    wizardCancelButton.handlebarsType = 'helper';
    wizardCancelButton.handlebarsName = 'wizardCancelButton';

    var wizardSaveButton = _.partial(wizardBaseButton, 'save');
    wizardSaveButton.handlebarsType = 'helper';
    wizardSaveButton.handlebarsName = 'wizardSaveButton';

    var wizardResetButton = _.partial(wizardBaseButton, 'reset');
    wizardResetButton.handlebarsType = 'helper';
    wizardResetButton.handlebarsName = 'wizardResetButton';

    return {
        wizardNextButton : wizardNextButton,
        wizardPrevButton : wizardPrevButton,
        wizardCancelButton : wizardCancelButton,
        wizardSaveButton : wizardSaveButton,
        wizardResetButton : wizardResetButton
    };
});


define('primutils/utils/validator',['primutils.jquery', 'primutils.lodash', 'primutils.logger', 'primutils.moment'], function ($, _, logger, moment) {
    

    var __emptyValidator__ = {
            name : '',
            required : [],
            banned : [],
            customValidations : []
        },
        __defaultOptions__ = {
            name : 'Unnamed Validator',
            ignoredSubkeys: [],
            validators : {}
        };

    function optionsKeys(options, ignoredSubkeys, prefix) {
        var result = [], ignoresSubs;

        prefix = prefix || '';

        _.each(_.keys(options), function (key) {
            var obj = options[key];
            ignoresSubs = _.any(ignoredSubkeys, function (ignoreKey) {
                return ignoreKey === prefix + key;
            });

            result.push(prefix + key);
            if (moment.isMoment(obj)) {
                // Skip over date (Moment, really) subobjects
                return;
            }
            if (moment.isDuration(obj)) {
                // Durations, too
                return;
            }
            if (obj instanceof $) {
                return;
            }

            if (!ignoresSubs && _.isObject(obj) && !_.isArray(obj)) {
                result = _.union(result, optionsKeys(options[key], ignoredSubkeys, (prefix + key)+'.'));
            }
        });

        return result;
    }

    function deepValue(rootOptions, key) {
        var path = key.split('.'), pathLength = path.length,
            obj, i;
        obj = rootOptions;

        for(i = 0; i < pathLength && !_.isUndefined(obj) && !_.isNull(obj); i++) {
            obj = obj[path[i]];
        }

        return obj;
    }

    function fnToString(fn) {
        var str = fn.toString();

        if (str.split('\n').length > 1) {
            return str.split('\n')[0] + '...}';
        }
        else {
            return str;
        }
    }

    function valueString(obj) {
        if (_.isFunction(obj) && obj.toString().split('\n').length > 1) {
            return fnToString(obj);
        }
        else {
            return ('' + obj);
        }
    }

    function checkType(type, value) {
        if (_.isArray(type)) {
            return _.any(type, function (possibleType) {
                return checkType(possibleType, value);
            });
        }
        else if (_.isFunction(type)) {
            return type(value);
        }
        else if (type === 'Jquery' || type === 'jQuery') {
            return (value instanceof $) && value.jquery;
        }
        else if (type === 'Date') {
            return moment.isMoment(value);
        }
        else if (type === 'Duration') {
            return moment.isDuration(value);
        }
        else if (_.isString(type) && _.has(_, 'is' + type)) {
            return _['is' + type](value);
        }
        else {
            // Some options can be {Number | 'none'}, this way we can give a list of types like ['Number', 'none']
            return value === type;
        }
    }

    function printError(value, type, keyName) {
        var message = 'Illegal value for \'' + keyName + '\' option. The option\'s value was \'' +
            valueString(value) + '\', which is of type \'' + (typeof value) + '\' but should match one of the following rules/types: ';

        if (_.isArray(type)) {
            message += '[';
            _.each(type, function (t) {
                message += '\'' + t + '\'';
                if (_.last(type) !== t) {
                    message += ', ';
                }
            });
            message += ']';
        }
        else {
            message += '\'' + type + '\'';
        }

        return message;
    }

    /**
     * A helper for running validation on options provided to your component.
     *
     * For simple cases, consider using the static method <a href="#method_checkType">checkType</a>.
     *
     *     validator.checkType('a', ['b', 'c', 'Number'], 'Key name');
     *         // Would logger.warn a message, and return false,
     *         //  because the option given by the user did not match:
     *         //  the string 'b',the string 'c', or pass the _.isNumber function.
     *
     * For a simple for checking everything in a given object (the following example works as of primutils 2.7.3):
     *
     *     var myValidator = validator({
     *         validOptions: {
     *             optionA: ['none', 'Array'],
     *             optionB: ['jQuery']
     *         }
     *     });
     *     myValidator({optionC: 'value', optionB: [], optionA: 'none'});
     *         // Would logger.warn that 'optionC' is unrecognized
     *         // Would logger.warn that 'optionB' doesn't pass any of the validOptions rules
     *         // Would accept optionA without any complaints, as it matches one of the rules (an exact match for the string 'none')
     *
     * For an example of a more complex case:
     *
     *     var myValidator = validator({
     *          // included in error messages when unrecognized/banned/missing options are given by the user
     *         name: 'Your component name',
     *          // all the options that could be recognized by your component; warnings will be fired for any options given that aren't indicated here
     *         validOptions: {
     *              // your component has a option called 'minValue' that must pass the check _.isNumber
     *             minValue: 'Number',
     *              // your component has an option called 'stringValue' that must pass the check _.isString
     *             stringValue: 'String',
     *              // your component has an option called 'anything' that can be of any type
     *             anything: undefined,
     *              // your component has an option called 'someObject' that must be an Object
     *             someObject: 'Object',
     *              // your component has an option called 'someObject', with a subkey of 'subKey', which must be a MomentJS date object
     *             'someObject.subKey': 'Date',
     *              // your component has an option called 'positiveNumeric' that must pass the given function
     *             positiveNumeric: function (n) { return _.isNumber(n) && n > 0; }
     *         },
     *          // any subkeys of the 'anything' option will not be type checked for correctness
     *         ignoredSubkeys: ['anything'],
     *         validators: {
     *              // Rules specific to when you call `myValidator.validate('initialize', {...options to check...});`
     *             initialize: {
     *                  // your component requires that the provided options hash specify the 'minValue' option when you call `myValidator.validate('initialize', userOptions);`
     *                 required: ['minValue'],
     *                  // your component will not allow that the provided options hash specify the 'anything' option when you call `myValidator.validate('initialize', userOptions);`
     *                 banned: ['anything'],
     *             },
     *             configure
     *         }
     *     });
     *     myValidator.validate('initialize', userOptions); // Checks all provided userOptions keys against the spec given on creation
     *
     * @class utils.validator
     * @module utils
     * @param options {Object} see class description example
     * @constructor
     */
    function validatorFactory (options) {
        var _options = {},
            _hasValidated = false;

        _options = _.defaults({}, options || {}, _options, __defaultOptions__);

                validatorFactory.checkType(_options.name, ['String'], 'name');
        validatorFactory.checkType(_options.validators, [undefined, 'Object'], 'validators');
        validatorFactory.checkType(_options.validOptions, ['Object'], 'validOptions');
        validatorFactory.checkType(_options.ignoredSubkeys, [undefined, 'Array'], 'ignoredSubkeys');
        validatorFactory.checkType(_options.required, [undefined, 'Array'], 'required');
        validatorFactory.checkType(_options.banned, [undefined, 'Array'], 'banned');
        

        /**
         * Runs the validation rules provided at validator creation time against the provided options hash.
         * See the more complex example in the class description for more details.
         *
         * @method validate
         * @param [validatorName] {String} the validator required/banned ruleset to run
         * @param options {Object} the object to run validation rules against
         * @returns true if any test failed, false otherwise
         */
        function that (validatorName, userOptions) {
            if (arguments.length === 1) {
                userOptions = validatorName;
                validatorName = __emptyValidator__.name;
            }

            var clonedValidator = _.clone(__emptyValidator__),
                validator = _.has(_options.validators, validatorName) ?
                    _.defaults({}, _options.validators[validatorName], __emptyValidator__) :
                    clonedValidator,
                keys = optionsKeys(userOptions, _options.ignoredSubkeys),
                unrecognized = _.difference(keys, _.keys(_options.validOptions)),
                banned = _.intersection(keys, _.union(validator.banned, _options.banned)),
                missing = _.difference(_.union(validator.required, _options.required), keys),
                message;

            if (validator === clonedValidator && arguments.length === 2) {
                logger.error('No validator for ' + validatorName + ' found');
            }

            _.each(unrecognized, function (key) {
                message = 'Unrecognized ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.warn(message);
            });

            _.each(banned, function (key) {
                message = 'Illegal ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.fatal(message);
                throw new Error(message);
            });

            _.each(missing, function (key) {
                message = 'Missing required ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.fatal(message);
                throw new Error(message);
            });

            _.each(keys, function (key) {
                var value, type;
                if (_.has(_options.validOptions, key)) {
                    value = deepValue(userOptions, key);
                    type = _options.validOptions[key];

                    if (type === undefined || checkType(type, value)) {
                        return;
                    }

                    message = printError(deepValue(userOptions, key), type, key);
                    logger.warn(message);
                }
            });

            _.each(validator.customValidations, function (v) {
                var result, level;
                if (_.isFunction(v)) {
                    result = {
                        options: userOptions,
                        level : false,
                        message : 'Custom validation step failed'
                    };

                    v(result);

                    if (result.level !== false) {
                        message = result.message;
                        level = logger[result.level] ? result.level : 'warn';
                        logger[level](result.message);
                        if (level === 'fatal') {
                            throw new Error(result.message);
                        }
                    }
                } else {
                    logger.warn('Validator was not a function');
                }
            });

            return !!message;
        }
        that.validate = that;

        // First call will use 'initialize' validator options, second will use 'configure'
        that.autoValidate = function (_options) {
            if (!_hasValidated) {
                _hasValidated = true;
                return that.validate('initialize', _options);
            } else {
                return that.validate('configure', _options);
            }
        };

        /**
         * Creates a new validator, based on the current validator, with the provided new options overriding and/or augmenting existing options.
         *
         * @Example
         *
         *     var myValidator = validator({
         *         name: 'myValidator',
         *         validOptions: { jq: 'jQuery' }
         *     });
         *     var myExtendedValidator = myValidator.withOptions({
         *         validOptions: { dat: 'Date' }
         *     });
         *
         *     // Here, `myExtendedValidator` would recognize both the fields `jq` and `dat` to be jQuery selected objects and Moment objects, respectively.
         *
         * @method withOptions
         * @param newOptions {Object} the new options
         * @returns a new validator with new options overriding old ones
         * @since 2.7.2
         */
        that.withOptions = function withOptions(newOptions) {
            return validatorFactory($.extend(true, {}, _options, newOptions));
        };

        /**
         * Gets the valid options recognized by this validator instance.
         *
         * This returns the map of recognized keys to their valid types, so if all you want are the keys, `_.keys(myValidator.getValidOptions())` should suffice.
         *
         * @method getValidOptions
         * @returns the map of recognized options to their acceptable keys
         * @since 2.7.2
         */
        that.getValidOptions = function getValidOptions() {
            return _.clone(_options.validOptions);
        };

        return that;
    }

    /**
     * See simple example in class description.
     *
     * Simple static method to check a single value against a list of types.
     * Will logger.warn if the test fails.
     *
     * @method checkType
     * @static
     * @returns true if the value passed the test, false otherwise
     */
    validatorFactory.checkType = function (value, type, keyName) {
        if (!checkType(type, value)) {
            keyName = keyName || '';
            logger.warn(printError(value, type, keyName));
            return false;
        }
        return true;
    };

    return validatorFactory;

});


define('primutils/markup/primutils/ui/changeBar/changeBar.hbs:.save-controls', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/changeBar.hbs:.save-controls'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"save-cancel-controls\" aria-hidden=\"true\" data-error-count=\"0\" style=\"display: none;\">\n    ";
  stack1 = lambda((depth0 != null ? depth0.content : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/changeBar/changeBar.hbs:.errors-list', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/button','primutils/templates/helpers/buttonGroup','primutils/templates/helpers/dropdown','primutils/templates/helpers/join'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/changeBar.hbs:.errors-list'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"errors-list\" id=\"change-bar-error-list-container-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n    ";
  stack1 = helpers.button.call(depth0, "error", {"name":"button","hash":{
    'class': ("error-button"),
    'accesskey': ((depth0 != null ? depth0.accesskey : depth0)),
    'aria-relevant': ("text"),
    'aria-live': ("polite"),
    'aria-owns': (helpers._.call(depth0, "change-bar-errors-list-", (depth0 != null ? depth0.changeBarId : depth0), {"name":"_","hash":{},"data":data})),
    'aria-haspopup': ("true"),
    'data-toggle': ("dropdown")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    <div class=\"dropdown-menu pull-center\">\n        <ul id=\"change-bar-errors-list-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" data-target=\"#change-bar-error-list-container-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\" class=\"notification-list error-list\" role=\"menu\" aria-region=\"log\" aria-live=\"polite\" aria-relevant=\"additions text\"></ul>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/ui/changeBar/helpers',['primutils.lodash', 'primutils.logger', 'primutils/utils/language',
        'primutils/utils/validator',
        'primutils/templates/helpers/button',
        'primutils/templates/helpers/icon',
        'primutils/templates/utils/utils',
        'primutils/templates/utils/renderer',
        'primutils/markup/primutils/ui/changeBar/changeBar.hbs:.save-controls',
        'primutils/markup/primutils/ui/changeBar/changeBar.hbs:.errors-list'
       ], function (_, logger, lang, validator, buttonBase, icon, utils, renderer, hbrSaveControls, hbrErrorsList) {
    

    var saveControls = function (options) {
        logger.assert(!!options.fn, 'change-bar-controls must be used as a block helper');
        return hbrSaveControls({
            content: options.fn.call(this, this)
        });
    };
    saveControls.handlebarsType = 'helper';
    saveControls.handlebarsName = 'change-bar-controls';

    var errorList = function (type, options) {
        if (arguments.length === 1) {
            type = 'default';
        }
                else {
            logger.assert(arguments.length === 2, 'change-bar-errors only has one required parameter (the type, either \'default\' or \'simple\', passing in more is an error');
            validator.checkType(type, ['default', 'simple']);
        }
        
        return hbrErrorsList({
            accesskey: type === 'default' ? 'e' : undefined,
            changeBarId : (this && this.changeBarId) || _.uniqueId('')
        });
    };
    errorList.handlebarsType = 'helper';
    errorList.handlebarsName = 'change-bar-errors';

    var getCancelText = function () { return lang.getString('changeBar.cancel'); };
    var cancelButton = function (type, title, options) {
        if (arguments.length === 1) {
            return buttonBase.call(this, 'secondary', {
                hash: {
                    'disabled': true,
                    accesskey: 'r',
                    'class': 'cancel-button',
                },
                fn: getCancelText
            });
        }
        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        var mergedOptions = _.defaults({
            hash: {
                'disabled': true,
                'class': utils.mergeClasses(['cancel-button'], options.hash['class'])
            }
        }, options, {
            fn: getCancelText
        });
        return buttonBase.call(this, type, title, mergedOptions);
    };
    cancelButton.handlebarsType = 'helper';
    cancelButton.handlebarsName = 'change-bar-cancel';

    var getSaveText = function () {
        var unsaved = renderer.renderElementWithContent('div', {
            'class' : 'unsaved save-button-state'
        }, {}, lang.getString('changeBar.save'));

        var saving = renderer.renderElementWithContent('div', {
            'class' : 'saving save-button-state hidden'
        }, {}, icon.call(this, 'throbber', '', { hash : {} }));

        var saved = renderer.renderElementWithContent('div', {
            'class' : 'saved save-button-state hidden'
        }, {}, icon.call(this, 'checkmark-checked', '', { hash : {} }));

        return unsaved + saving + saved;
    };

    var saveButton = function (type, title, options) {
        if (arguments.length === 1) {
            return buttonBase.call(this, 'primary', {
                hash: {
                    'disabled': true,
                    accesskey: 's',
                    'class': 'save-button',
                },
                fn: getSaveText
            });
        }
        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        var mergedOptions = _.defaults({
            hash: {
                'disabled': true,
                'class': utils.mergeClasses(['save-button'], options.hash['class'])
            }
        }, options, {
            fn: getSaveText
        });
        return buttonBase.call(this, type, title, mergedOptions);
    };
    saveButton.handlebarsType = 'helper';
    saveButton.handlebarsName = 'change-bar-save';

    var statusText = function () {
        return renderer.renderElement('span', {
            'class' : 'status-message'
        });
    };

    statusText.handlebarsType = 'helper';
    statusText.handlebarsName = 'change-bar-status';

    return {
        'change-bar-controls': saveControls,
        'change-bar-errors': errorList,
        'change-bar-cancel': cancelButton,
        'change-bar-save': saveButton,
        'change-bar-status' : statusText
    };
});



define('primutils/templates/templateUtilities',[
    'primutils.markup',
    'primutils.lodash',
    'primutils.handlebars',
    'primutils.logger',
    'primutils/templates/helpers/localizer', // This is here because it is used by this module, it is listed twice on purpose

    // All exported helpers and partials must be listed here
    'primutils/templates/helpers/accordion',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/buttonGroup',
    'primutils/templates/helpers/caret',
    'primutils/templates/helpers/controlGroup',
    'primutils/templates/helpers/dock',
    'primutils/templates/helpers/dropdown',
    'primutils/templates/helpers/emptyState',
    'primutils/templates/helpers/fieldset',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/input',
    'primutils/templates/helpers/join',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/localizer',
    'primutils/templates/helpers/modal',
    'primutils/templates/helpers/panel',
    'primutils/templates/helpers/search',
    'primutils/templates/helpers/slider',
    'primutils/templates/helpers/spinner',
    'primutils/templates/helpers/steps',
    'primutils/templates/helpers/termDef',
    'primutils/templates/helpers/toggle',
    'primutils/templates/helpers/toolbar',
    'primutils/templates/helpers/wizard',
    'primutils/ui/changeBar/helpers',

], function (hbr, _, Handlebars, logger, localizerHelper) {
    

    var importModules = _.rest(_.toArray(arguments), 5),
        registered = false;

    /**
     * Utility functions for exporting Handlebars templates from primutils
     *
     * @class templates.templateUtilities
     * @module templates
     */

    return {
        /**
         * By default primutils will not register its Handlebars helpers and partials globally.
         * Calling this function assigns all of the known primutils helpers and partials globally to Handlebars.
         *
         * This function can only be called once.
         * Additional calls to this function will not do anything.
         *
         * @method registerImports
         */
        registerImports : function () {
            if (registered) {
                return;
            }

            var imports = hbr.buildImported(importModules, 'primutils');

            _.each(imports.helpers, function (fn, name) {
                Handlebars.registerHelper(name, fn);
            });

            _.each(imports.partials, function (fn, name) {
                Handlebars.registerPartial(name, fn);
            });

            registered = true;
        },

        /**
         * Handlebars provides no built-in checking to make sure that globally registered helpers and partials do not conflict with previously registered global handlers.
         * This method can be called to throw a warning in the console when a duplicate handler is attempted to register.
         *
         * To completely prevent duplicate handlers from overwrite each other, pass `true`.
         * Otherwise the new handler will overwrite the old one.
         *
         * Note: This does not prevent individual templates from "shadowing" global templates by using the "deps" option.
         * See the <a href="#">markup plugin</a> for more information.
         *
         * @method protectHandlebarsImports
         * @param [noOverwrite] {Boolean} Prevents a second handler with the same name from being registered at all.
         */
        protectHandlebarsImports : function (noOverwrite) {
            Handlebars.registerHelper = _.wrap(Handlebars.registerHelper, function (fn, name) {
                var names = _.isObject(name) ? _.keys(name) : [name],
                    existingHelpers = _.intersection(names, _.keys(Handlebars.helpers)),
                    hasDupes = existingHelpers.length > 0;

                if (hasDupes) {
                    logger[noOverwrite ? 'error' : 'warn']('The following helpers are already registered globally' +
                        (noOverwrite ? ' and will not be overwritten' : '') +
                        ': ' + existingHelpers.join(', '));
                }

                if (!noOverwrite || !hasDupes) {
                    return fn.apply(this, _.rest(_.toArray(arguments)));
                } else {
                    throw new Error('Cannot override previously registered helpers: ' + existingHelpers.join(', '));
                }
            });

            Handlebars.registerPartial = _.wrap(Handlebars.registerPartial, function (fn, name) {
                var names = _.isObject(name) ? _.keys(name) : [name],
                    existingPartials = _.intersection(names, _.keys(Handlebars.partials)),
                    hasDupes = existingPartials.length > 0;

                if (hasDupes) {
                    logger[noOverwrite ? 'error' : 'warn']('The following partials are already registered globally' +
                        (noOverwrite ? ' and will not be overwritten' : '') +
                        ': ' + existingPartials.join(', '));
                }

                if (!noOverwrite || !hasDupes) {
                    return fn.apply(this, _.rest(_.toArray(arguments)));
                } else {
                    throw new Error('Cannot override previously registered helpers: ' + existingPartials.join(', '));
                }
            });
        },

        /**
         * Sets the localizer to use in the localizer helper.
         *
         * @method setLocalizerForLocalizerHelper
         * @param localizer {Localizer} The localizer to use in the localizer helper
         */
        setLocalizerForLocalizerHelper : function (localizer) {
            // Type checking is done inside the helper
            localizerHelper.useLocalizer(localizer);
        }
    };
});


define('primutils/formatters/wrappers',[], function () {
    

    return {
        timezoneSensitiveDateTimeUnserializeValidation: function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            }
            else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            }
            else if (str && str.indexOf('.') === -1) {
                throw new Error('TimezoneSensitive dates must include timezone marker');
            }
            else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        globalDateUnserializeValidation: function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            }
            else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            }
            else if (str && str.indexOf('.') !== -1) {
                throw new Error('Global dates must not include timezone marker');
            }
            else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        globalDateTimeUnserializeValidation: function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            }
            else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            }
            else if (str && str.indexOf('.') !== -1) {
                throw new Error('Global dates must not include timezone marker');
            }
            else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        integerValidation: function (origFn, options, str) {
            var origResult = origFn(options, str);
            var decimalSeparator = options.decimalSeparator === '\\.' ? '.' : options.decimalSeparator;
            if (origResult === true && str.indexOf(decimalSeparator) !== -1) {
                return false;
            }
            return origResult;
        }
    };
});



define('primutils/formatters/formats',{
    timezoneSensitiveDateTimeSerialization:   'YYYY-MM-DD[T]HH:mm:ss.ZZ',
    timezoneSensitiveDateTimeUnserialization: 'YYYY-MM-DD[T]HH:mm:ss.ZZ',

    globalDateSerialization:       'YYYY-MM-DD[T00:00:00]',
    globalDateUnserialization:     'YYYY-MM-DD[T00:00:00]',
    globalDateTimeSerialization:   'YYYY-MM-DD[T]HH:mm:ss',
    globalDateTimeUnserialization: 'YYYY-MM-DD[T]HH:mm:ss'
});


define('primutils/ui/widgets/dock',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils/utils/keyCode',
    'primutils/utils/language'
], function ($, _, keyCode, lang) {
    

    var ORIENTATIONS = {
        'horizontal' : 'dock-bottom',
        'vertical'   : 'dock-right'
    };

    function getOrientation ($dock) {
        if ($dock.hasClass('dock-left') || $dock.hasClass('dock-right')) {
            return 'vertical';
        } else if ($dock.hasClass('dock-top') || $dock.hasClass('dock-bottom')) {
            return 'horizontal';
        }
    }

    // TODO Implement remainder of ARIA splitter recommendations
    // End - max size
    // Home - min size
    // Enter - Restore previous position

    function Dock (el) {
        var that = this;
        this.$el = $(el);
        this.throttledResize = _.throttle(function () {
            that.$el.trigger('dock-resized');
        }, 250, { trailing : true });
    }

    Dock.prototype = {
        toggle : function (state) {
            state = (state == null) ? this.$el.hasClass('collapsed') : state;
            return this[state ? 'expand' : 'collapse']();
        },

        expand : function () {
            if (!this.$el.hasClass('collapsed')) {
                return;
            }

            var e = $.Event('dock-expanding');
            this.$el.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            this.$el.removeClass('collapsed').attr('aria-expanded', true)
                .find('.dock-header > .collapse-handle').attr('title', lang.getString('dock.collapse')).find('> .spoken-text').text(lang.getString('dock.collapse'));
            this.$el.trigger($.Event('dock-expanded'));
        },

        collapse : function () {
            if (this.$el.hasClass('collapsed')) {
                return;
            }

            var e = $.Event('dock-collapsing');
            this.$el.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            this.$el.addClass('collapsed').attr('aria-expanded', false)
                .find('.dock-header > .collapse-handle').attr('title', lang.getString('dock.expand')).find('> .spoken-text').text(lang.getString('dock.expand'));
            this.$el.trigger($.Event('dock-collapsed'));
        },

        orient : function (orientation) {
            var e = $.Event('dock-before-orientation');
            this.$el.trigger(e, { orientation : orientation });

            if (e.isDefaultPrevented()) {
                return;
            }

            if (orientation === 'vertical') {
                this._storedHeight = this.$el.css('height');
            } else if (orientation === 'horizontal') {
                this._storedWidth = this.$el.css('width');
            }

            this.$el.css({
                height : (orientation === 'vertical')  ? '' : this._storedHeight,
                width : (orientation === 'horizontal') ? '' : this._storedWidth
            });

            this.$el.removeClass('dock-left dock-right dock-top dock-bottom').addClass(ORIENTATIONS[orientation])
                .find('.resize-handle').attr('aria-orientation', orientation);
            this.$el.find('.dock-controls button').removeClass('active').filter('[data-orientation="' + orientation + '"]').addClass('active');

            this.$el.trigger($.Event('dock-orientation-changed'), { orientation : orientation });
        }
    };

    $.fn.extend({
        prim_dock : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_dock');

                if (!data) {
                    data = new Dock(this);
                    $this.data('prim_dock', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_dock.Constructor = Dock;

    $(document)
        .on('click.dock.data-api', '.dock .collapse-handle', function (e) {
            var $dock = $(e.target).closest('.dock');
            e.preventDefault();
            $dock.prim_dock('toggle');
        })
        .on('click.dock.data-api', '.dock .dock-controls button', function (e) {
            var $btn = $(e.currentTarget),
                orient = $btn.data('orientation'),
                $dock = $(e.target).closest('.dock');
            $dock.prim_dock('orient', orient);
        })
        .on('keydown', '.dock .resize-handle', function (e) {
            var $dock = $(e.target).closest('.dock'),
                dock,
                orientation = getOrientation($dock),
                multiplier = ($dock.hasClass('dock-left') || $dock.hasClass('dock-top')) ? -1 : 1,
                diff = 10 * multiplier;

            dock = $dock.prim_dock().data('prim_dock'); // Make sure that dock has been initialized internally

            if ($dock.hasClass('collapsed')) { return; }

            if (orientation === 'horizontal') {
                if (e.keyCode === keyCode.UP) {
                    $dock.css('height', (parseInt($dock.css('height'), 10) + diff) + 'px');
                    e.preventDefault();
                } else if (e.keyCode === keyCode.DOWN) {
                    $dock.css('height', (parseInt($dock.css('height'), 10) - diff) + 'px');
                    e.preventDefault();
                }
            } else if (orientation === 'vertical') {
                if (e.keyCode === keyCode.LEFT) {
                    $dock.css('width', (parseInt($dock.css('width'), 10) + diff) + 'px');
                    e.preventDefault();
                } else if (e.keyCode === keyCode.RIGHT) {
                    $dock.css('width', (parseInt($dock.css('width'), 10) - diff) + 'px');
                    e.preventDefault();
                }
            }

            if (e.isDefaultPrevented()) {
                dock.throttledResize();
            }
        })
        .on('mousedown.dock.data-api', '.dock .resize-handle', function (e) {
            var $dock = $(e.target).closest('.dock');

            if ($dock.hasClass('collapsed')) { return; }

            $('body').css({
                '-webkit-user-select' : 'none',
                '-moz-user-select' : 'none',
                '-ms-user-select' : 'none',
                '-o-user-select' : 'none',
                'user-select' : 'none'
            });

            var lastX = e.pageX,
                lastY = e.pageY,
                orientation = getOrientation($dock),
                namespace = _.uniqueId('.pgbu-dock-resize-'),
                multiplier = ($dock.hasClass('dock-left') || $dock.hasClass('dock-top')) ? -1 : 1;

            var moveHandler = function (event) {
                if (orientation === 'horizontal') {
                    $dock.css('height', (parseInt($dock.css('height'), 10) - (event.pageY - lastY) * multiplier) + 'px');
                } else if (orientation === 'vertical') {
                    $dock.css('width', (parseInt($dock.css('width'), 10) - (event.pageX - lastX) * multiplier) + 'px');
                }

                lastX = event.pageX;
                lastY = event.pageY;
            };

            var removeHandler = function () {
                $('body').css({
                    '-webkit-user-select': '',
                    '-moz-user-select': '',
                    '-ms-user-select': '',
                    '-o-user-select': '',
                    'user-select': ''
                });
                $(document).off( 'mousemove' + namespace + ' mouseleave' + namespace + ' mouseup' + namespace);
                $dock.trigger('dock-resized');
            };

            $(document)
                .on('mousemove' + namespace, moveHandler)
                .on('mouseleave' + namespace + ' mouseup' + namespace, removeHandler);
        });
});

define('primutils/ui/widgets/modal',[
    'primutils.jquery',
    'primutils.lodash',
], function ($, _) {
    

    function startDraggable() {
        var $body = $('body');

        function attachMoveListeners($modal, originalX, originalY) {
            var originalOffset = $modal.offset();

            $modal.find('.modal-header').css('cursor', 'move');

            $body.css({
                    '-webkit-user-select' : 'none',
                    '-moz-user-select' : 'none',
                    '-ms-user-select' : 'none',
                    '-o-user-select' : 'none',
                    'user-select' : 'none'
                })
                .on('mousemove.modal.data-api', _.throttle(function (ev) {
                    $modal.offset({
                        left: originalOffset.left + (ev.pageX - originalX),
                        top: originalOffset.top + (ev.pageY - originalY)
                    });
                }, 50))
                .on('mouseup.modal.data-api', function (ev) {
                    $body.off('mousemove.modal.data-api mouseup.modal.data-api').css({
                        '-webkit-user-select': '',
                        '-moz-user-select': '',
                        '-ms-user-select': '',
                        '-o-user-select': '',
                        'user-select': ''
                    });

                    $modal.find('.modal-header').css('cursor', '');
                });
        }

        $body
            .on('mousedown.modal.data-api', '.modal .modal-header [data-dismiss=modal], .modal .modal-header .control-bar', function (ev) {
                ev.stopPropagation();
            })
            .on('mousedown.modal.data-api', '.modal .modal-header', function (ev) {
                var $modal = $(this).closest('.modal');
                if ($modal.attr('data-draggable') !== 'false') {
                    attachMoveListeners($modal, ev.pageX, ev.pageY);
                }
            });
    }
    startDraggable();
});

define('primutils/ui/widgets/search',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils/utils/keyCode'
], function ($, _, keyCode) {
    

    // TODO Implement remainder of ARIA splitter recommendations
    // End - max size
    // Home - min size
    // Enter - Restore previous position

    function Search (el) {
        this.$el = $(el);
    }

    Search.prototype = {
        clear : function (state) {
            this.$el.find('input').val('');
            this.changed(false);
            this.$el.trigger('search-cleared');
        },

        changed : function (hasText) {
            this.$el.find('.search-clear-icon').toggle(hasText);
        }
    };

    $.fn.extend({
        prim_search : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_search');

                if (!data) {
                    data = new Search(this);
                    $this.data('prim_search', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_search.Constructor = Search;

    $(document)
        .on('keyup.search.data-api change.search.data-api', '.filter-div > input', function (e) {
            var $this = $(this);
            $this.parent().prim_search('changed', $this.val().length > 0);
        })
        .on('click.search.data-api', '.filter-div > .search-clear-icon', function (e) {
            e.preventDefault();
            $(this).parent().prim_search('clear').find('input').focus();
        });
});


define('primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs:#item', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/tabMenuItemTemplate.hbs:#item'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "checked";
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<li role=\"menuitem\">\n    <a href=\"#\" data-tab-id=\"";
  stack1 = lambda((depth0 != null ? depth0.target : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" class=\"";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.visible : depth0), {"name":"if","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n        <i class=\"check\"></i>\n        <span class=\"tab-title\">";
  stack1 = lambda((depth0 != null ? depth0.title : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>\n    </a>\n</li>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs:#menu', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/dropdown'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/tabMenuItemTemplate.hbs:#menu'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.dropdownMenu.call(depth0, {"name":"dropdownMenu","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/widgets/tab',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils.logger',
    'primutils/utils/language',
    'primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs:#item',
    'primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs:#menu',

    'primutils.bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _, logger, lang, tabMenuItemTemplate, tabMenuTemplate) {
    

    $(document.head).append('<style type="text/css">.nav-tabs-scroll-container .nav-tabs>li.scroll-tab-hidden, .nav-tabs-scroll-container .nav-tabs>li.scroll-tab-scrolled, .nav-tabs-scroll-container .nav-tabs>li.scroll-tab-user-hidden { display: none !important; visibility: hidden !important; }</style>');

    var SQUEEZABLE_TABS = '.nav-tabs.scenario-tabs, .nav-tabs.content-tabs';

    function Tabbar (el) {
        this.$element = $(el);
        this._fixResizeContainer();
    }

    Tabbar.prototype = {
        get $tabs () {
            return this.$element.children(':not([data-noresize])');
        },
        get $nonTabs () {
            return this.$element.children('[data-noresize]');
        },
        get $menuTabs () {
            return this.$tabs.add(this.$nonTabs);
        },
        get $scrollEligibleTabs () {
            return this.$menuTabs.filter(':visible,.scroll-tab-hidden');
        },

        _resize : function (force) {
            //this._squeeze(force);
            this._scollResize(force);
        },

        _squeeze : function (force) {
            if (!this.$element.is(SQUEEZABLE_TABS)) {
                return;
            }

            var tabbarWidth = this.$element.width(),
                nonResizableWidth = _.reduce(this.$nonTabs, function (elem) {
                    return $(elem).outerWidth(true);
                }, 0),
                $allTabs = this.$tabs.removeClass('scroll-tab-hidden').filter(':visible'),
                $activeTab = $allTabs.filter('.active');

            // Prevent the scroll bar from appearing while resizing
            this.$element.css({ 'max-height' : $activeTab.outerHeight() - 1, 'overflow' : 'hidden' });
            $allTabs.css({'max-width' : 'none', 'float' : 'none', 'display' : 'inline-block', 'white-space':'nowrap'});
            // TODO put notch on <li> instead of <a> to do proper ellipses

            var activeTabWidth = $activeTab.outerWidth(true),
                grabbedMinWidth = parseInt($activeTab.css('min-width'), 10),
                minWidth = grabbedMinWidth > 0 ? grabbedMinWidth : $(window).width(),
                $otherTabs = $allTabs.not($activeTab),
                otherTabsCount = $otherTabs.length,
                usefulWidth = tabbarWidth - activeTabWidth - nonResizableWidth,
                baseTabSize = Math.max(Math.floor(usefulWidth / otherTabsCount), minWidth),
                reclaimableSize = 0,
                reclaimableTabs = 0;

            $otherTabs.each(function () {
                var space = Math.max(0, baseTabSize - $(this).outerWidth(true));
                reclaimableSize += space;
                if (space > 0) {
                    reclaimableTabs++;
                }
            });

            if (reclaimableTabs > 0) {
                baseTabSize += Math.floor(reclaimableSize / otherTabsCount);
            }

            $otherTabs.css('max-width', baseTabSize).find('> a');

            // Restore previous values for tabs
            $allTabs.css({'float' : '','display' : ''});
            this.$element.css({ 'max-height' : '', 'overflow' : '' });
        },

        _fixResizeContainer : function () {
            this.$resizeContainer = this.$element.closest('.nav-tabs-scroll-container');
        },

        _unscroll : function () {
            this.$tabs.removeClass('scroll-tab-hidden');
        },

        _scollResize : function (force) {
            var $scrollers = this.$resizeContainer.find('.scrollers'),
                tabbarWidth = this.$element.width() - this.$element.find('.collapse-handle').outerWidth(true),
                $allTabs = this.$scrollEligibleTabs,
                leftOffset = parseInt(this.$element.css('padding-left'), 10) || 0,
                right = null;

            // Prevent the scroll bar from appearing while resizing
            this.$resizeContainer.addClass('has-before-tabs has-after-tabs');
            this.$element.css({ 'max-height' : $allTabs.filter(':visible').find('a').outerHeight(), 'overflow' : 'hidden' });
            $allTabs.removeClass('scroll-tab-hidden');
            $scrollers.children().removeClass('available');

            $allTabs.each(function () {
                var $tab = $(this);
                if (!right) {
                    right = $tab.position().left - leftOffset;
                }
                right += $tab.outerWidth(true);
                if (right > tabbarWidth) {
                    $tab.addClass('scroll-tab-hidden');
                }
            });

            var $firstVisibleTab = this.$tabs.filter(':visible').first(),
                $lastVisibleTab = this.$tabs.filter(':visible').last(),
                hasLeftScroll = $firstVisibleTab.prevAll('.scroll-tab-hidden,.scroll-tab-scrolled').length > 0,
                hasRightScroll = $lastVisibleTab.nextAll('.scroll-tab-hidden,.scroll-tab-scrolled').length > 0,
                hasHiddenTabs = this.$tabs.hasClass('scroll-tab-user-hidden'),
                hasBeforeTabs = hasLeftScroll || hasRightScroll || hasHiddenTabs,
                isDetailPanel = this.$element.hasClass('detail-panel-tabs');

            $scrollers.find('.left').toggleClass('available', hasLeftScroll);
            $scrollers.find('.right').toggleClass('available', hasRightScroll);
            this.$resizeContainer
                .toggleClass('has-before-tabs', hasBeforeTabs)
                .toggleClass('has-after-tabs', hasBeforeTabs || isDetailPanel);
            this.$resizeContainer.find('.before-tabs').toggleClass('hidden', !hasBeforeTabs);
            this.$resizeContainer.find('.after-tabs').toggleClass('hidden', !(hasBeforeTabs || isDetailPanel));

            // Restore previous values for tabs
            this.$element.css({ 'max-height' : '', 'overflow' : '' });
        },

        scrollRight : function () {
            var $firstVisibleTab = this.$tabs.filter(':visible').first(),
                moreScrollHiddenTabs = this.$tabs.filter(':visible').last().nextAll('.scroll-tab-hidden').length > 0;
            if (moreScrollHiddenTabs) {
                $firstVisibleTab.addClass('scroll-tab-scrolled');
            }
            this._resize();
        },

        scrollLeft : function () {
            var $firstVisibleTab = this.$tabs.filter(':visible').first(),
                $previousTab = $firstVisibleTab.prevAll('.scroll-tab-scrolled').first();
            $previousTab.removeClass('scroll-tab-scrolled');
            this._resize();
        },

        resize : function (force) {
            this._fixResizeContainer();
            if (this.$resizeContainer.length) {
                this._resize(force);
            } else {
                this._unscroll();
            }
        },

        buildMenu : function () {
            var $parent = this.$resizeContainer.find('.tabs-menu'),
                $menu, $visibleTabs;

            if ($parent.hasClass('open')) {
                $menu = $(tabMenuTemplate({}));
                    $visibleTabs = this.$menuTabs.filter(':visible,.scroll-tab-user-hidden,.scroll-tab-scrolled,.scroll-tab-hidden');
                $parent.find('.dropdown-menu').remove();

                $visibleTabs.each(function () {
                    var $tab = $(this).find('> [data-target]');
                    $menu.append(tabMenuItemTemplate({
                        title: $tab.text(),
                        target: $tab.attr('data-target'),
                        visible: !$(this).hasClass('scroll-tab-user-hidden')
                    }));
                });

                $menu.appendTo($parent);
            }
        },

        toggleTab : function (tabTarget) {
            var $tab = this.$menuTabs.find('[data-target="' + tabTarget + '"]').parent(),
                $firstVisibleTab = this.$tabs.filter(':visible').first(),
                $dropdownMenu = this.$resizeContainer.find('.tabs-menu');

            if (!$tab.hasClass('active')) {
                if ($firstVisibleTab.prevAll().has($tab)) {
                    $tab.addClass('.scroll-tab-scrolled');
                }

                $tab.toggleClass('scroll-tab-user-hidden');
                $dropdownMenu.find('[data-tab-id="' + tabTarget + '"]').toggleClass('checked', !$tab.hasClass('scroll-tab-user-hidden'));
                this._resize();
            }
        },

        selectTab : function (tabTarget) {
            var $tab = this.$menuTabs.find('[data-target="' + tabTarget + '"]'),
                $li = $tab.parent(),
                $dropdownMenu = this.$resizeContainer.find('.tabs-menu');

            if (!$li.hasClass('active') && (this.$scrollEligibleTabs.is($li) || $li.hasClass('scroll-tab-user-hidden'))) {
                $tab.click();
                if ($li.hasClass('scroll-tab-user-hidden')) {
                    $li.removeClass('scroll-tab-user-hidden');
                    $dropdownMenu.find('[data-tab-id="' + tabTarget + '"]').toggleClass('checked', !$li.hasClass('scroll-tab-user-hidden'));
                    this._resize();
                }

                $tab.focus();
            }
        },

        collapse : function () {
            var e = $.Event('panel-collapsing');
            this.$element.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            var expandStr = lang.getString('panel.expand');
            this.$element.closest('.panel-container').addClass('collapsed')
                .find('.nav-tabs .collapse-handle a').attr('title', expandStr)
                .find('> i > .spoken-text').text(expandStr);
            this.$element.trigger($.Event('panel-collapsed'));
        },

        expand : function () {
            var e = $.Event('panel-expanding');
            this.$element.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            var collapseStr = lang.getString('panel.collapse');
            this.$element.closest('.panel-container').removeClass('collapsed')
                .find('.nav-tabs .collapse-handle a').attr('title', collapseStr)
                .find('> i > .spoken-text').text(collapseStr);
            this.$element.trigger($.Event('panel-expanded'));
        },

        toggle : function (state) {
            var $container = this.$element.closest('.panel-container');

            state = (state == null) ? $container.hasClass('collapsed') : state;

            if (state) {
                this.expand();
            } else {
                this.collapse();
            }
        }
    };

    $.fn.extend({
        prim_tabbar : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_tabbar');

                if (!data) {
                    data = new Tabbar(this);
                    $this.data('prim_tabbar', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_tabbar.Constructor = Tabbar;

    function resizeAllTabbarsImpl () {
        $('.nav-tabs').each(function () {
            $(this).prim_tabbar('resize');
        });
    }
    var resizeAllTabbars = _.throttle(resizeAllTabbarsImpl, 100);

    $(window).on('resize.tab.data-api', resizeAllTabbars);
    $(resizeAllTabbars);

    if (window.MutationObserver) {
        (function () {
            var observer = new MutationObserver(resizeAllTabbars);
            var config = { childList : true, subtree : true };
            observer.observe(document.body, config);
        })();
    }

    $(document)
        .on('show.tab.data-api', '[data-toggle="tab"]', function (e) {
            var $next = $(this).parents('li'),
                $prev = $(e.relatedTarget).parents('li'),
                slideLeft = $next.index() < $prev.index();
            $next.add($prev).toggleClass('slide-left', slideLeft);
        })
        .on('click.tab.data-api', '[data-toggle="tab"]', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs').prim_tabbar('resize');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .scrollers > .left', function (e) {
            e.preventDefault();
            if ($(this).hasClass('available')) {
                $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('scrollLeft');
            }
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .scrollers > .right', function (e) {
            e.preventDefault();
            if ($(this).hasClass('available')) {
                $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('scrollRight');
            }
        })
        .on('click.tab.data-api', '.panel-container .collapse-handle a', function (e) {
            e.preventDefault();
            $(this).closest('.panel-container, .nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('toggle');
        })
        .on('mousedown.tab.data-api', '.nav-tabs-scroll-container .tabs-menu > a', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('resize');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu > a', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('buildMenu');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu [data-tab-id] .check', function (e) {
            e.preventDefault();
            e.stopPropagation();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('toggleTab', $(this).closest('[data-tab-id]').attr('data-tab-id'));
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu [data-tab-id]', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('selectTab', $(this).attr('data-tab-id'));
        })
        .on('mousewheel.tab.data-api DOMMouseScroll.tab.data-api', '.nav-tabs-scroll-container .nav-tabs', function (e) {
            e.preventDefault();
            if (e.originalEvent.wheelDelta > 0) {
                $(this).prim_tabbar('scrollLeft');
            } else {
                $(this).prim_tabbar('scrollRight');
            }
        })
        .on('dock-expanded.tab.data-api dock-orientation-changed.tab.data-api dock-resized.tab.data-api', '.dock', function (e) {
            $(this).find('.nav-tabs').prim_tabbar('resize');
        })
        .on('shown.tab.data-api', '.modal', function () {
            $(this).find('.nav-tabs').prim_tabbar('resize');
        })
        .on('shown.tab.data-api', '[data-toggle="tab"]', function (e) {
            var $this = $(this),
                pane = $this.attr('data-target');

            if (!pane) {
                pane = $this.attr('href');
                pane = pane && pane.replace(/.*(?=#[^\s]*$)/, ''); //strip for ie7
            }

            $(pane).find('.nav-tabs').prim_tabbar('resize');
        })
        .on('wizard-step-changed.tab.data-api', '.wizard-context', function () {
            $(this).find('.nav-tabs').prim_tabbar('resize');
        });

    function startDraggable() {
        var sortableChildren = ':not([data-fixed],.scrollers,.collapse-handle,.scroll-tab-hidden,.scroll-tab-scrolled)',
            tabAnimateDuration = 125, // ms; these two values look chrome-like to me
            tabAnimationType = 'linear';

        var $draggableTabs = false,
            $tabbar = false,
            $tab = false,
            blankImage = false;

        function createBlankImage() {
            // http://stackoverflow.com/questions/9126105/blank-image-encoded-as-data-uri
            var blankDataURIImage = 'data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==';
            blankImage = document.createElement('img');
            blankImage.src = blankDataURIImage;
        }

        function setBlankImage(rawEvent) {
            if (!blankImage) {
                createBlankImage();
            }
            if (rawEvent.dataTransfer.setDragImage) {
                rawEvent.dataTransfer.setDragImage(blankImage, 0, 0);
            }
        }

        function positionTabs($newBar) {
            var lastLeft;

            $draggableTabs = $newBar.children(sortableChildren);
            lastLeft = $draggableTabs.first().offsetParent().is($newBar) ? $draggableTabs.first().position().left : $draggableTabs.first().position().left - $newBar.position().left;

            $tabbar = $newBar
                .data('_dragOriginalPosition', $newBar.get(0).style.position)
                .css({'position': 'relative', 'min-height' : $draggableTabs.outerHeight() - 1});

            $draggableTabs
                .each(function (index) {
                    $(this)
                        .data('index', index)
                        .data('_dragOriginalPosition', this.style.position)
                        .data('_dragOriginalLeft', this.style.left)
                        .data('_dragLeftPosition', lastLeft)
                        .css({
                            position : 'absolute',
                            left : lastLeft
                        });

                    lastLeft += $(this).outerWidth(true);
                });
        }

        function doTrySwap() {
            if (!$tab) {
                return;
            }

            var swapAllowance = 15;
            var tabIndex = $tab.data('index');
            var tabWidth = $tab.outerWidth(true);

            var tabLeft = $tab.offset().left;
            var tabKnownLeft = $tab.data('_dragLeftPosition');
            var tabRight = tabLeft + tabWidth;
            var $previousTab = $tab.prev(sortableChildren);
            var previousCenter = $previousTab.length ? $previousTab.offset().left + ($previousTab.outerWidth(true) / 2) : 0;
            var $nextTab = $tab.next(sortableChildren);
            var nextCenter =     $nextTab.length ? $nextTab.offset().left + ($nextTab.outerWidth(true) / 2) : 0;

            if ((tabIndex >= 1) && (previousCenter - tabLeft > swapAllowance)) {
                $previousTab.data('index', tabIndex);
                $tab.data('index', tabIndex - 1);

                $tab.data('_dragLeftPosition', $previousTab.data('_dragLeftPosition'));
                $previousTab.data('_dragLeftPosition', tabKnownLeft);

                $previousTab.animate({
                    left: tabKnownLeft
                }, tabAnimateDuration, tabAnimationType);
                $tab.insertBefore($previousTab);
            }
            else if ((tabIndex < $draggableTabs.length - 1) && (tabRight - nextCenter > swapAllowance)) {
                $nextTab.data('index', tabIndex);
                $tab.data('index', tabIndex + 1);

                $tab.data('_dragLeftPosition', $nextTab.data('_dragLeftPosition'));
                $nextTab.data('_dragLeftPosition', tabKnownLeft);

                $nextTab.animate({
                    left: tabKnownLeft
                }, tabAnimateDuration, tabAnimationType);
                $tab.insertAfter($nextTab);
            }
        }
        var trySwap = _.throttle(doTrySwap, 50);

        function resetTabs() {
            $draggableTabs
                .each(function () {
                    var $this = $(this);
                    $this.css({
                        position: $this.data('_dragOriginalPosition') || '',
                        left: $this.data('_dataOriginalLeft') || ''
                    });
                    delete $this.data()._dragOriginalPosition;
                    delete $this.data()._dragOriginalLeft;
                });

            $tabbar.css({'position': $tabbar.data('_dragOriginalPosition'), 'min-height': ''});
            delete $tabbar.data()._dragOriginalPosition;
        }

        var $body = $('body');

        $body
            .on('dragstart.tab.data-api', '[data-sortable] [data-toggle="tab"]', function (ev) {
                var $selectedTab = $(this).parent();

                var beforeReorderEvent = $.Event('tabs-before-reordering');
                $selectedTab.parent().trigger(beforeReorderEvent);

                if (beforeReorderEvent.isDefaultPrevented()) {
                    return;
                }

                $tab = $selectedTab;
                $tab.addClass('dragging');
                positionTabs($tab.parent());
                setBlankImage(ev.originalEvent);

                $tabbar.trigger('tabs-reordering');
            })
            .on('dragend.tab.data-api', '[data-sortable] [data-toggle="tab"]', function (ev) {
                resetTabs();
                $tab.removeClass('dragging');
                $tab = false;
                $draggableTabs = false;

                $tabbar.trigger('tabs-reordered');

                $tabbar = false;
            });

        $body
            .on('dragover.tab.data-api', function (ev) {
                if ($tab) {
                    var tabWidth = $tab.outerWidth(true),
                        position = ev.originalEvent.clientX - (tabWidth / 2),
                        tabbarPos = $tabbar.position().left;
                    $tab.offset({
                        left: Math.max(tabbarPos, Math.min($tabbar.outerWidth(true) + tabbarPos - tabWidth, position))
                    });
                    trySwap();
                    ev.preventDefault();
                }
            })
            .on('dragenter.tab.data-api', function (ev) {
                if ($tab) {
                    ev.preventDefault();
                }
            })
            .on('drop.tab.data-api', function (ev) {
                if ($tab) {
                    ev.preventDefault();
                }
            });
    }
    startDraggable();

    // --------------------
    // Tab enhancements

    var Tab = $.fn.tab.Constructor,
        TabProto = Tab.prototype;

    // Need to be able to pass arguments to Bootstrap tab function
    $.fn.tab = function ( option ) {
        var applyArgs = Array.prototype.slice.call(arguments, 1);
        return this.each(function () {
            var $this = $(this),
                data = $this.data('tab');

            if (!data) {
                $this.data('tab', (data = new Tab(this)));
            }

            if (typeof option === 'string') {
                data[option].apply(data, applyArgs);
            }
        });
    };

    TabProto.toggle = function (state) {
        state = (state == null) ? !this.element.parent().is(':visible') : state;

        if (state) {
            this.available();
        } else {
            this.unavailable();
        }
    };

    TabProto.available = function () {
        this.element.parent().show();
        this.element.closest('.nav-tabs').prim_tabbar('resize');
    };

    TabProto.unavailable = function () {
        this.element.parent().hide();
        this.element.closest('.nav-tabs').prim_tabbar('resize');
    };
});

define('primutils/ui/widgets/wizard',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils.logger',
    'primutils/templates/helpers/panel',

    'primutils.bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _, logger, panelHelpers) {
    

    function Wizard (el) {
        this.$element = $(el);
    }

    function getFakePanelContext () {
        return {
            __panelContext__ : {
                type: 'wizard',
                tabsCreated: [],
                panelsCreated: []
            }
        };
    }

    function noContent () {
        return '';
    }

    Wizard.prototype = {
        goToStep : function (stepId, force) {
            if (!(force || this.canChangeStep(stepId))) {
                return;
            }

            this.$element.find('.steps > li').removeClass('active')
                .filter('[data-wizard-step="' + stepId + '"]').addClass('active');

            this.$element.find('.wizard-panel > .tab-pane').removeClass('active')
                .filter('#' + stepId).addClass('active');

            this._updateButtons();

            this.$element.trigger('wizard-step-changed', { step: stepId });
        },

        _updateButtons : function () {
            var nextStep = this.getNext(),
                prevStep = this.getPrev(),
                canNext = false, canPrev = false;

            if (nextStep) {
                canNext = this.canChangeStep(nextStep);
            }

            if (prevStep) {
                canPrev = this.canChangeStep(prevStep);
            }

            this.$element.find('[data-wizard=prev]').toggle(!!prevStep && canPrev);
            this.$element.find('[data-wizard=next]').toggle(!!nextStep && canNext);
            //this.$element.find('[data-wizard=cancel]').toggle(!prevStep);
            this.$element.find('[data-wizard=save]').toggle(!nextStep);
        },

        canChangeStep : function (newStep) {
            var current = this.getCurrent(),
                eventInfo = {
                    currentStep : current,
                    newStep : newStep
                },
                e = $.Event('wizard-can-change-step', eventInfo);
            this.$element.trigger(e, eventInfo);
            return e.result !== false;
        },

        get $steps () {
            return this.$element.find('.steps');
        },

        get $panels () {
            return this.$element.find('.wizard-panel');
        },

        get $currentLi () {
            return this.$steps.find(' > .active');
        },

        getCurrent : function () {
            return this.$currentLi.attr('data-wizard-step');
        },

        getNext : function () {
            return this.$currentLi.nextAll(':visible').first().attr('data-wizard-step');
        },

        next : function () {
            var nextStep = this.getNext();
            if (nextStep) {
                this.goToStep(nextStep);
            } else {
                this._updateButtons();
            }
        },

        getPrev : function () {
            return this.$currentLi.prevAll(':visible').first().attr('data-wizard-step');
        },

        prev : function () {
            var prevStep = this.getPrev();
            if (prevStep) {
                this.goToStep(prevStep);
            } else {
                this._updateButtons();
            }
        },

        toggleStep : function (step, visible) {
            this.$steps.find('[data-wizard-step="' + step + '"]').toggle(visible);
        },

        addStep : function (stepId, stepName, after) {
            var context = getFakePanelContext(),
                $newStep = panelHelpers.tab.call(context, stepName, stepId, {hash:{}}),
                $newPanel = panelHelpers.panel.call(context, stepId, {hash:{},fn:noContent});

            if (after) {
                this.$steps.find('[data-wizard-step="' + after + '"]').after($newStep);
            } else {
                this.$steps.append($newStep);
            }

            this.$panels.append($newPanel);
        },

        deleteStep : function (stepId) {
            var $step = this.$steps.find('[data-wizard-step="' + stepId + '"]');
            if ($step.hasClass('active')) {
                if (this.getPrev()) {
                    this.goToStep(this.getPrev(), true);
                } else {
                    this.goToStep(this.getNext(), true);
                }
            }
            $step.remove();
            this.$panels.find('#'+stepId).remove();
        },

        cancel : function () {
            this.$element.trigger('wizard-cancel');
        },

        save : function () {
            this.$element.trigger('wizard-save');
        },

        reset : function () {
            this.$element.trigger('wizard-reset');
        },

        resize : function () {

        }
    };

    $.fn.extend({
        prim_wizard : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_wizard');

                if (!data) {
                    data = new Wizard(this);
                    $this.data('prim_wizard', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_wizard.Constructor = Wizard;

    $(document).on('click.wizard.data-api', '.wizard-context [data-wizard]', function (e) {
        var $this = $(this),
            type = $this.attr('data-wizard');
        $this.closest('.wizard-context').prim_wizard(type);
    });
});

define('primutils/ui/widgets/widgets',[
    'primutils/ui/widgets/dock',
    'primutils/ui/widgets/modal',
    'primutils/ui/widgets/search',
    'primutils/ui/widgets/tab',
    'primutils/ui/widgets/wizard',
], function () {
    
});


/**
 * Primary export of primutils.  TODO document me.
 *
 * @class primutils
 * @module primutils
 * @private
 */
define('primutils',[
    'primutils/i18n/Localizer',
    'primutils/formatters/numberFormatter',
    'primutils/formatters/dateFormatter',
    'primutils/formatters/durationFormatter',
    'primutils/validators/validatorFactory',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils',
    'primutils/templates/templateUtilities',
    'primutils/formatters/wrappers',
    'primutils/formatters/formats',

    'primutils/ui/widgets/widgets'

    // Items that should be available in the compiled file, but are not available as a property of primutils
    // should be listed in Gruntfile.js
],
    function (Localizer, numberFormatter, dateFormatter, durationFormatter, validatorFactory, language, accessibilityUtils, templateUtilities, formatterWrappers, formatterFormats) {
        

        return {

            version : '2.10.0',

            lang : function (stringMap) {
                language.setStringMap(stringMap);
            },

            numberFormatter : numberFormatter,
            dateFormatter : dateFormatter,
            durationFormatter: durationFormatter,
            validatorFactory : validatorFactory,

            formats: formatterFormats,
            wrappers: formatterWrappers,

            accessibilityUtils : accessibilityUtils,

            registerGlobalHandlebarsImports : templateUtilities.registerImports.bind(templateUtilities),

            protectGlobalHandlebarsImports : templateUtilities.protectHandlebarsImports.bind(templateUtilities),

            setLocalizerForLocalizerHelper : templateUtilities.setLocalizerForLocalizerHelper.bind(templateUtilities)
        };
    }
);

define('primutils/a11y/bootstrap/button',['primutils.jquery', 'primutils.bootstrap'], function ($) {
    

    var ButtonPrototype = $.fn.button.Constructor.prototype,
        origButtonToggle = ButtonPrototype.toggle;

    ButtonPrototype.toggle = function () {
        if (!this.$element.prop('disabled')) {
            origButtonToggle.apply(this, arguments);
            var $parent = this.$element.closest('[data-toggle="buttons-radio"]');
            if ($parent) {
                $parent.find('[aria-pressed="true"]').attr('aria-pressed', false);
            }
            this.$element.attr('aria-pressed', this.$element.hasClass('active'));
        }
    };
});


define('primutils/a11y/bootstrap/collapse',['primutils.jquery',
    'primutils/utils/keyCode',

    'primutils.bootstrap'
], function ($, keyCode) {
    

    /**
     * Enhancements to Bootstrap's collapse plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.collapse
     * @module a11y.bootstrap
     * @private
     */

    var CollapsePrototype = $.fn.collapse.Constructor.prototype,
        origCollapseTransition = CollapsePrototype.transition;

    CollapsePrototype.transition = function (method) {
        origCollapseTransition.apply(this, arguments);
        var $accordionGroup = this.$element.parents('.accordion-group').first();

        $accordionGroup[method]('open');

        //add aria-expanded true if you are adding open class.
        $accordionGroup.find('.accordion-toggle')
            .attr('aria-expanded', method === 'addClass')
            .attr('aria-selected', method === 'addClass');

        $accordionGroup.find('.accordion-inner')
            .attr('tabindex', method === 'addClass' ? '' : '-1')
            .attr('aria-hidden', method !== 'addClass');
    };

    $(document).on('focusin.collapse.data-api, click.collapse.data-api', '.accordion', function (e) {
        var $accordion = $(e.currentTarget),
            focusedOnce = $accordion.data('focusedOnce'),
            $group = $accordion.find('.accordion-body.in').parent(),
            $selected = $(document.activeElement);

        if (!$group.length) {
            $group = $accordion.find('.accordion-group').first();
        }

        $accordion.data('focusedOnce', true).find('.accordion-toggle')
            .attr('tabindex', '-1') // prevent tabbing, use arrow keys instead
            .attr('aria-selected', false);

        $group.find('.accordion-toggle')
            .attr('tabindex', '0')
            .attr('aria-selected', true);

        if (!focusedOnce && !$selected.closest('.accordion-group').is($group)) {
            $group.find('.accordion-toggle').focus();
        }
    });

    $(document).on('keydown.collapse.data-api', '.accordion-group', function (e) {
        var $target = $(e.target),
            $group = $(e.currentTarget),
            $newGroup;

        if ($target.hasClass('accordion-toggle')) {
            if (e.keyCode === keyCode.UP || e.keyCode === keyCode.LEFT) {
                $newGroup = $group.prevAll('.accordion-group').first();
            } else if (e.keyCode === keyCode.DOWN || e.keyCode === keyCode.RIGHT) {
                $newGroup = $group.nextAll('.accordion-group').first();
            } else if (e.keyCode === keyCode.HOME) {
                $newGroup = $group.prevAll('.accordion-group').last();
            } else if (e.keyCode === keyCode.END) {
                $newGroup = $group.nextAll('.accordion-group').last();
            } else if (e.keyCode === keyCode.SPACE) {
                $target.click();
                e.preventDefault();
            }

            if ($newGroup) {
                e.preventDefault();
                $newGroup.find('.accordion-toggle').focus();
            }
        } else if (e.keyCode === keyCode.UP && e.ctrlKey) {
            $group.find('.accordion-toggle').focus();
            e.stopPropagation();
        }
    });

    $(document).on('show.collapse.data-api hidden.collapse.data-api', '.accordion-body', function (e) {
        $(this).toggleClass('hidden', e.type !== 'show');
    });
});


define('primutils/a11y/bootstrap/dropdown',['primutils.jquery', 'primutils.lodash',
    'primutils/a11y/collapsibleDropdown',
    'primutils/utils/keyCode',
    'primutils.bootstrap'
], function ($, _, collapsibleDropdown, keyCode) {
    

    /**
     * Enhancements to Bootstrap's dropdown plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.dropdown
     * @module a11y.bootstrap
     * @private
     */

    // Copied from Bootstrap (but made to fit our style)
    function getParent($this) {
        var selector = $this.attr('data-target'),
            $parent;

        if (!selector) {
            selector = $this.attr('href');
            selector = selector && /#/.test(selector) && selector.replace(/.*(?=#[^\s]*$)/, ''); //strip for ie7
        }

        $parent = selector && $(selector);

        if (!$parent || !$parent.length) {
            $parent = $this.parent();
        }

        return $parent;
    }

    function shouldDropUp($popup, $parent) {
        var popupHeight = $popup.outerHeight(true),
            parentHeight = $parent.outerHeight(),
            offset = $parent.offset();

        return offset.top + popupHeight + parentHeight > window.pageYOffset + $(window).height();
    }

    // Needs to be attached _after_ bootstrap
    $(document)
        .on('click.dropdown.data-api', '[data-toggle=dropdown]', function () {
            var $this = $(this),
                $parent = getParent($this),
                $popup = $parent.children('.dropdown-menu'),
                $body = $('body'),
                $menu = $parent.children('.dropdown-menu'),
                isActive = $parent.hasClass('open');

            $this.attr('aria-expanded', isActive);
            $popup.attr('aria-hidden', !isActive);

            if (isActive) {
                $body.css('overflow', 'hidden'); // Prevent scrollbars while calculating need to dropup
                $parent.toggleClass('dropup', shouldDropUp($popup, $parent));
                $body.css('overflow', '');
            }

            if ($menu.hasClass('pull-center')) {
                var currentMargin = parseFloat($menu.css('margin-right')) || 0;
                $menu.css('margin-right', isActive ? (($menu.outerWidth(true) - currentMargin) / -2) : '');
            }

            if (!$popup.data('collapsibleDropdown')) {
                $popup.data('collapsibleDropdown', collapsibleDropdown($popup));
            }

            return false;
        })
        .on('keydown.dropdown.data-api', '[data-toggle=dropdown]', function (e) {
            if (e.keyCode === keyCode.DOWN) {
                var $parent = getParent($(this)),
                    $items = $('[role=menu] li:not(.divider):visible > a', $parent);
                $items.first().focus();
            }
        });

    $(document).on('keydown.dropdown.data-api', '[role=menu]', function(event) {
        var target = $(event.target),
            $parent = getParent($(event.currentTarget)),
            $items = $('[role=menu] li:not(.divider):visible > a', $parent);


        if($items.length === 1) {
            if(event.keyCode === keyCode.TAB) { //Both TAB and Shift-TAB should remove open on single child menu
                target.closest('.open').removeClass('open');
            }
        }else if(target.parent().is(':first-child')) {
            if (event.keyCode === keyCode.TAB && event.shiftKey) {
                target.closest('.open').removeClass('open');
            } else if (event.keyCode === keyCode.UP) {
                $items.last().focus();
            }
        }else if(target.parent().is(':last-child')) {
            if (event.keyCode === keyCode.TAB && !event.shiftKey) {
                target.closest('.open').removeClass('open');
            } else if (event.keyCode === keyCode.DOWN) {
                $items.first().focus();
            }
        }
    });
});

define('primutils/utils/emulateTransitionEnd',['primutils.jquery', 'primutils.bootstrap'], function ($) { //Require that Bootstrap be loaded first for $.support.transition.end
    

    // Based on: http://api.jquery.com/jQuery.cssHooks/
    function venderPrefix( prop, style ) {
        var vendorProp,
            // Capitalize first character of the prop to test vendor prefix
            capProp = prop.charAt( 0 ).toUpperCase() + prop.slice( 1 ),
            prefixes = [ 'Moz', 'Webkit', 'O', 'ms' ];

        if ( prop in style ) { // Browser supports standard CSS property name
            return prop;
        } else {
            // Otherwise test support for vendor-prefixed property names
            for ( var i = 0; i < prefixes.length; i++ ) {
                vendorProp = prefixes[ i ] + capProp;
                if ( vendorProp in style ) {
                    return vendorProp;
                }
            }
        }
    }

    function parseToMS(value) {
        value = value.trim();
        if (value.slice(-2) === 'ms') {
            return parseFloat(value);
        } else if (value.slice(-1) === 's') {
            return 1000 * parseFloat(value);
        }
    }

    function transitionDurationMS($elem, defaultValue) {
        var venderTransitionDuration = venderPrefix('TransitionDuration', $elem.get(0).style),
            value = $elem.css(venderTransitionDuration);
        if (value.indexOf(',') > -1) {
            return defaultValue;
        } else {
            return parseToMS(value) || defaultValue;
        }
    }

    // Based on: http://blog.alexmaccaw.com/css-transitions
    $.fn.emulateTransitionEnd = function(guessDuration) {
        if ($.support.transition) {
            var called = false, $el = this;
            $(this).one($.support.transition.end, function() { called = true; });
            // Add 25ms to guessed duration to give slow animations time to finish
            setTimeout(function() {
                if (!called) {
                    $($el).trigger($.support.transition.end);
                }
            }, transitionDurationMS($(this), guessDuration ? (guessDuration + 25) : 10000));
        }
    };
});


define('primutils/a11y/bootstrap/modal',['primutils.jquery', 'primutils.lodash', 'primutils/a11y/accessibilityUtils',
    'primutils.bootstrap',
    'primutils/utils/emulateTransitionEnd' // For $.fn.emulateTransitionEnd
], function ($, _, utils) {
    

    /**
     * Enhancements to Bootstrap's modal plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.modal
     * @module a11y.bootstrap
     * @private
     */

    var ModalPrototype = $.fn.modal.Constructor.prototype,
        origModalBackdrop = ModalPrototype.backdrop,
        origModalEscape = ModalPrototype.escape,
        origModalShow = ModalPrototype.show;

    ModalPrototype.backdrop = function () {
        origModalBackdrop.apply(this, arguments);
        if (this.$backdrop) {
            // If there was no animation, then Bootstrap will have already removed $backdrop
            // and set it to null, so we don't need to do this
            this.$backdrop.emulateTransitionEnd(200);
        }
    };

    ModalPrototype.show = function () {
        this._originalFocusedElement = document.activeElement;

        origModalShow.apply(this, arguments);

        //Happens after origModalShow is called so this.$backdrop is set up for moving
        var shownModalsWithoutSelf = $('.modal.in').not(this.$element);
        if(shownModalsWithoutSelf.length > 0) {
            //Shift some z-indices...
            var max = _.chain(shownModalsWithoutSelf)
                .map(function(modal) {
                    return parseInt($(modal).css('z-index'), 10);
                })
                .max()
                .value();

            //Modals and backdrops originally start at 1050 and 1040 respectively
            this.$backdrop.css('z-index', max + 1);
            this.$element.css('z-index', max + 10);
        }

        // Move the modal backdrop to right after the modal, but only if it's already in the DOM
        // This is necessary due to potentially having modals in new stacking contexts, (e.g. position:absolute)
        // as well as when you have modals inside of modals. Additionally, if z-indexing was touched above
        // backdrop needs to be in same parent as child modal
        if (this.$element.parent().length) {
            this.$element.after(this.$backdrop);
        }
    };

    ModalPrototype.escape = function () {
        var that = this;

        if (this.isShown) {
            this.$element.on('keyup.defaultAction.modal', function (e) {
                var $defaultEnter = that.$element.find('[data-default="enter"]'),
                    $defaultEsc = that.$element.find('[data-default="esc"]');
                if (e.which === 27 && $defaultEsc.length) {
                    $defaultEsc.click();
                } else if (e.which === 13 && $defaultEnter.length) {
                    $defaultEnter.click();
                }
            });
        } else if (!this.isShown) {
            this.$element.off('keyup.defaultAction.modal');

            this.$element.one('hidden', function () {
                if (that._originalFocusedElement && that._originalFocusedElement.parentNode) {
                    that._originalFocusedElement.focus();
                    delete that._originalFocusedElement;
                }
            });
        }

        origModalEscape.apply(this, arguments);
    };

    $(document).on('keydown.modal', '.modal', function(event) {
        if(event.keyCode === 9) { //Tab
            var focusable = utils.findFocusable($(this)),
                length = focusable.length,
                forward = !event.shiftKey;


            event.preventDefault();
            event.stopPropagation();

            if($.contains(this, document.activeElement)) {
                //Go on to the next one;
                var index = focusable.index(document.activeElement);
                if(forward) {
                    index = (index + 1) % length;
                } else {
                    index = ((index - 1) % length + length) % length;
                }

                focusable.get(index).focus();
            } else if (forward && focusable.length > 0){
                focusable.get(0).focus();
            } else if (!forward && focusable.length > 0) {
                focusable.get(focusable.length - 1).focus();
            }

            // Added because bootstrap modal was not highlighting textfield input when tabbed to.
            // The textfield was the focus however its input was not selected.
            if (document.activeElement.type === 'text' && !document.activeElement.readOnly) {
                document.activeElement.select();
            }
        }
    });
});


define('primutils/a11y/bootstrap/tab',['primutils.jquery',
    'primutils/utils/keyCode',

    'primutils.bootstrap'
], function ($, keyCode) {
    

    var tabSelector = '[data-toggle="tab"]',
        tabishSelector = tabSelector + ',.scrollers > a,.collapse-handle > a',
        tabishAndMenusSelector = tabishSelector + ',[data-toggle=dropdown]';

    /**
     * Enhancements to Bootstrap's tab plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.tab
     * @module a11y.bootstrap
     * @private
     */

    $(document).on('focusin.tab.data-api, click.collapse.data-api', tabishSelector, function (e) {
        var $tab = $(e.currentTarget),
            $tabbar = $tab.closest('.nav-tabs'),
            focusedOnce = $tabbar.data('focusedOnce'),
            $selected = $(document.activeElement);

        if ($tab.attr('data-toggle') !== 'tab') {
            $tab = $tabbar.find(tabSelector).first();
        }

        $tabbar.data('focusedOnce', true).find(tabishAndMenusSelector)
            .attr('tabindex', '-1') // prevent tabbing, use arrow keys instead
            .attr('aria-selected', false);

        $tab
            .attr('tabindex', '0')
            .attr('aria-selected', $tab.attr('data-toggle') === 'tab');

        if (!focusedOnce && !$selected.is($tab)) {
            $tab.focus();
        }
    });

    $(document).on('keydown.tab.data-api', tabishSelector, function (e) {
        var $tab = $(e.currentTarget),
            $tabbar = $tab.closest('.nav-tabs'),
            $li = $tab.closest('li'),
            isScroller = $li.hasClass('scrollers'),
            fwd,
            $newTab,
            $menu;

        if (e.keyCode === keyCode.UP || e.keyCode === keyCode.LEFT) {
            $newTab = (!isScroller || $tab.prevAll().length === 0) ? $li.prevAll(':visible,.scroll-tab-scrolled').first() : $li;
            fwd = false;
            if ($newTab.length && !$newTab.is(':visible')) {
                $tabbar.prim_tabbar('scrollLeft');
                if (!$newTab.is(':visible')) {
                    $newTab = $newTab.prevAll(':visible').first();
                }
            }
        } else if (e.keyCode === keyCode.DOWN || e.keyCode === keyCode.RIGHT) {
            $menu = $li.find('[data-toggle=dropdown]');
            if (e.keyCode === keyCode.DOWN && $menu.length > 0) {
                $menu.focus().click();
                e.preventDefault();
            } else {
                $newTab = (!isScroller || $tab.next().length === 0) ? $li.nextAll(':visible,.scroll-tab-hidden').first() : $li;
                fwd = true;
                if ($newTab.length && !$newTab.is(':visible')) {
                    $tabbar.prim_tabbar('scrollRight');
                    if (!$newTab.is(':visible')) {
                        $newTab = $newTab.nextAll(':visible').first();
                    }
                }
            }
        } else if (e.keyCode === keyCode.HOME) {
            $newTab = $li.prevAll(':visible').find(tabSelector).parent().first();
            fwd = false;
        } else if (e.keyCode === keyCode.END) {
            $newTab = $li.nextAll(':visible').find(tabSelector).parent().last();
            fwd = true;
        } else if (e.keyCode === keyCode.SPACE) {
            $tab.click();
            e.preventDefault();
        } else if (e.keyCode === keyCode.F10 && e.shiftKey) {
            $menu = $li.find('[data-toggle=dropdown]');
            if ($menu.length > 0) {
                $menu.focus().click();
                e.preventDefault();
            }
        }

        if ($newTab) {
            e.preventDefault();
            $newTab.filter(':visible').find('> a:not([data-toggle=dropdown])')[fwd ? 'last' : 'first']().focus();
        }
    });

    $(document).on('keydown.tab.data-api', '.nav-tabs [data-toggle=dropdown]', function (e) {
        var $dropdown = $(this),
            $li = $dropdown.closest('li');
        if (e.keyCode === keyCode.LEFT || e.keyCode === keyCode.RIGHT || (e.keyCode === keyCode.F10 && e.shiftKey)) {
            $li.removeClass('open');
            $li.find('[data-toggle=tab]').focus();
        }
    });

    $(document).on('keydown.tab.data-api', '.tab-pane', function (e) {
        var id = $(this).attr('id');
        if (e.keyCode === keyCode.UP && e.ctrlKey) {
            e.preventDefault();
            $('a[href=#' + id + ']').focus();
            e.stopPropagation();
        }
    });
});


define('primutils/bootstrapAccessibilityFixes',[
    'primutils/a11y/bootstrap/button',
    'primutils/a11y/bootstrap/collapse',
    'primutils/a11y/bootstrap/dropdown',
    'primutils/a11y/bootstrap/modal',
    'primutils/a11y/bootstrap/tab'
], function () {
    

    /**
     * Enhancements to Bootstrap that make it more accessible and usable.
     *
     * This module loads all of the <a href="../modules/a11y.bootstrap.html">accessibility fixes</a> for Bootstrap.
     *
     * @class a11y.bootstrap.accessibilityFixes
     * @module a11y.bootstrap
     */
});

define('primutils/store/OrderedMap',['primutils.lodash'], function (_) {
    

    /**
     * Notes:
     * collection = ordered list of items
     * indexes = lookup map between item key and collection index
     * TODO - could probably use some underscore functions here for managing the index
     */

    return function (options) {

        var that = this,
            defaults = _.defaults(options, {
                key : 'key'
            });
        that.keys = {};
        that.collection = [];

        var populateIndexes = function () {
            that.keys = {};
            for (var index = 0; index < that.collection.length; index++) {
                var item = that.collection[index];
                that.keys[item[defaults.key]] = index;
            }
        };

        var isKeyValid = function (key) {
            if (key == null) {
                throw new Error('key undefined', 'OrderedMap.js');
            }
            return true;
        };

        return {

            /**
             * Return all items
             * @return array
             */
            items : function () {
                return that.collection;
            },

            /**
             * Return item referenced by key
             * @param key - string
             * @return object
             */
            item : function (key) {
                if (isKeyValid(key)) {
                    var item = that.collection[that.keys[key]];
                    return item == null ? null : item;
                }
                return null;
            },

            /**
             * Add array of items
             * @param items - array
             */
            addAll : function (items) {
                for (var index = 0; index < items.length; index++) {
                    this.add(items[index]);
                }
            },

            /**
             * Add item
             * @param item - object
             */
            add : function (item) {
                that.collection.push(item);
                that.keys[item[defaults.key]] = that.collection.length - 1;
            },

            /**
             * Insert item into a specific position
             * Note - SLOW due to repopulate of indexes
             * @param item - object
             * @param position - integer
             */
            insertAt : function (item, position) {
                if (position < 0 || position > that.collection.length - 1) {
                    throw new RangeError('position out of index bounds', 'orderedmap.js');
                }

                that.collection.splice(position, 0, item);
                populateIndexes();
            },

            /**
             * Remove item
             * Note - SLOW due to repopulate of indexes
             * @param key - string
             */
            remove : function (key) {
                if (isKeyValid(key)) {
                    var position = this.position(key);
                    if (position > -1) {
                        delete that.keys[key];
                        that.collection.splice(position, 1);
                        populateIndexes();
                    }
                }
            },

            /**
             * Return the position of the item
             * @param key - integer
             * @return integer
             */
            position : function (key) {
                if (isKeyValid(key)) {
                    var thisKey = that.keys[key];
                    return thisKey == null ? -1 : thisKey;
                }
                return -1;
            },

            /**
             * Return whether the ordered map is empty
             * @return boolean
             */
            isEmpty : function () {
                return this.count() === 0;
            },

            /**
             * Return the size of the ordered map
             * @return integer
             */
            count : function () {
                return that.collection.length;
            },

            /**
             * Remove all items from ordered map
             */
            clear : function () {
                that.keys = {};
                that.collection = [];
            },

            _repopulateIndexes : function () {
                populateIndexes();
            }
        };
    };
});

define('primutils/store/Collection',['primutils.lodash', 'primutils/store/OrderedMap'], function (_, OrderedMap) {
    

    var Collection = function (name, options) {
        this.name = name;
        this.map = {};
        this.options = _.defaults(options, {
            getKey : function (type) {
                return 'id';
            }
        });
    };

    var addNewMap = function (type) {
        this.map[type] = new OrderedMap({
            key : this.options.getKey.call(this, type)
        });
    };

    Collection.prototype.constructor = Collection;

    Collection.prototype.add = function (type, item) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }
        this.map[type].add(item);
        return item;
    };

    Collection.prototype.insertAt = function (type, item, position) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }
        this.map[type].insertAt(item, position);
        return item;
    };

    Collection.prototype.position = function (type, key) {
        if (this.hasType(type)) {
            return this.map[type].position(key);
        }
        return -1;
    };

    Collection.prototype.remove = function (type, key) {
        if (this.hasType(type)) {
            this.map[type].remove(key);
        }
    };

    Collection.prototype.item = function (type, key) {
        if (this.hasType(type)) {
            return this.map[type].item(key);
        }

        return null;
    };

    Collection.prototype.items = function (type) {

        if (this.hasType(type)) {
            return this.map[type].items();
        }
        return [];
    };

    Collection.prototype.hasType = function (type) {
        if (type == null) {
            throw new Error('storeCollection.js - "type" parameter not defined');
        }

        return _.has(this.map, type);
    };

    Collection.prototype.clear = function (type, keys, purge) {
        if (this.hasType(type)) {
            if (_.isArray(keys)) {
                var that = this;
                var itemKeys = _.pluck(this.items(type), that.options.getKey(type));
                _.each(itemKeys, function (itemKey) {
                    if (_.contains(keys, itemKey)) {
                        that.remove(type, itemKey, '', purge);
                    }
                });
            } else {
                delete this.map[type];

            }
        }
    };

    Collection.prototype.clearAll = function () {
        this.map = {};
    };

    Collection.prototype.count = function (type) {
        if (this.hasType(type)) {
            return this.map[type].count();
        }
        return 0;
    };

    Collection.prototype.isEmpty = function (type) {
        if (this.hasType(type)) {
            return this.map[type].isEmpty();
        }
        return true;
    };

    Collection.prototype.isEmptyAll = function () {
        var that = this;
        var empty = true;
        _.each(this.map, function (value, type) {
            if (that.count(type) > 0) {
                empty = false;
                return true;
            }
        });
        return empty;
    };

    Collection.prototype.types = function () {
        return _.keys(this.map);
    };

    Collection.prototype._repopulateIndexes = function (type) {
        if (this.hasType(type)) {
            this.map[type]._repopulateIndexes();
        }
    };

    return Collection;
});

define('primutils/utils/nestedPropertyAccessors',[], function () {
    

    var DEFAULT_PATH_DELIMITER = '.';

    var api = {

        getValue : function (item, field, delimiter) {
            delimiter = delimiter || DEFAULT_PATH_DELIMITER;

            if (item == null) {
                return item;
            }

            // get last part of path
            var depthIndex = field.lastIndexOf(delimiter);
            if (depthIndex === -1) {
                // return the value
                return item[field];
            }

            // move down a level in the object
            var firstLevelIndex = field.indexOf(delimiter);
            var depth = field.slice(0, firstLevelIndex);
            return api.getValue(item[depth], field.slice(firstLevelIndex + 1));
        },

        setValue : function (item, field, value, delimiter) {
            delimiter = delimiter || DEFAULT_PATH_DELIMITER;

            // get last part of path
            var depthIndex = field.lastIndexOf(delimiter);
            if (depthIndex === -1) {
                // set value
                item[field] = value;
            } else {
                // move down a level in the object
                var firstLevelIndex = field.indexOf(delimiter);
                var depth = field.slice(0, firstLevelIndex);
                if (item[depth] == null) {
                    // no object at this depth
                    throw new Error('nesterPropertyAccessors - no object at this depth: ' + field);
                }
                return api.setValue(item[depth], field.slice(firstLevelIndex + 1), value);
            }
        }
    };

    return api;

});

define('primutils/store/ModifyCollection',['primutils.jquery', 'primutils.lodash', 'primutils/store/Collection', 'primutils/utils/nestedPropertyAccessors'], function ($, _, Collection, propHelpers) {
    

    var PATH_DELIMITER = '.';

    var BREAK = {};

    function pathItemEach(item, path, callback, context) {
        var piece = item,
            parts = path.split(PATH_DELIMITER),
            fullPath = '';
        for (var i = 0; i < parts.length; i++) {
            fullPath += parts[i];
            if (!_.isUndefined(piece)) {
                piece = piece[parts[i]];
            }
            if (callback.call(context, parts[i], fullPath, piece) === BREAK) {
                return;
            }
            fullPath += PATH_DELIMITER;
        }
    }

    function reversePathEach(path, callback, context) {
        var pathPart = path,
            parent, key, delimLoc;
        while ((delimLoc = pathPart.lastIndexOf(PATH_DELIMITER)) > 0) {
            parent = pathPart.substr(0, delimLoc);
            key = pathPart.substr(delimLoc + 1);
            if (callback.call(context, key, pathPart) === BREAK) {
                return;
            }
            pathPart = parent;
        }
        // Root path
        callback.call(context, pathPart, pathPart);
    }

    function groupAllByPath(fields) {
        return _.reduce(fields, function (res, field) {
            reversePathEach(field, function (key, path) {
                res[path] = res[path] || [];
                res[path].push(field);
            });
            return res;
        }, {});
    }

    function ModifyStoreCollection (name, options) {
        Collection.call(this, name, options);
    }

    ModifyStoreCollection.prototype = _.extend(Object.create(Collection.prototype), {
        hasOriginalValue: function (type, id, field) {
            var item = this.item(type, id),
                found = false;

            if (item != null) {
                if (_.has(item.oldValues, field)) {
                    return true;
                }

                reversePathEach(field, function (key, path) {
                    if (_.has(item.oldValues, path)) {
                        found = true;
                        return BREAK;
                    }
                });
            }

            return found;
        },

        getOriginalValue: function (type, id, field) {
            var item = this.item(type, id),
                valueAtDepth, foundAtPath;

            if (item != null) {
                if (_.has(item.oldValues, field)) {
                    return item.oldValues[field];
                }

                reversePathEach(field, function (key, path) {
                    if (_.has(item.oldValues, path)) {
                        valueAtDepth = item.oldValues[path];
                        foundAtPath = path;
                        return BREAK;
                    }
                });

                if (foundAtPath != null && valueAtDepth != null) {
                    if (field.indexOf(foundAtPath + PATH_DELIMITER) === 0) {
                        var value = propHelpers.getValue(valueAtDepth, field.substr((foundAtPath.length + PATH_DELIMITER.length)), PATH_DELIMITER);
                        item.oldValues[field] = value;
                    }
                    return false;
                }
            }

            return undefined;
        },

        _allModifiedFields : function (type, id) {
            var item = this.item(type, id);
            if (item != null && item.oldValues != null) {
                return _.keys(item.oldValues);
            }
            return [];
        },

        modifiedFields : function (type, id) {
            var item = this.item(type, id);
            if (item != null && item.trueModifiedFields != null) {
                return _.keys(item.trueModifiedFields);
            }
            return [];
        },

        isSameValue: function (type, id, field, other) {
            return _.isEqual(this.getOriginalValue(type, id, field), other);
        },

        _forceMarkModified : function (type, id, field) {
            var item = this.item(type, id);

            if (item == null) {
                throw new Error('ModifyStore() - _forceMarkModified expects the item to be modified already');
            }

            item.trueModifiedFields[field] = true;
        },

        add: function (type, id, field, originalItem) {
            var item = this.item(type, id),
                key = this.options.getKey.call(this, type),
                oldValues;

            if (item == null) {
                item = {
                    oldValues: {},
                    trueModifiedFields: {}
                };
                item[key] = id;
                Collection.prototype.add.call(this, type, item);
            }

            oldValues = item.oldValues;

            item.trueModifiedFields[field] = true;

            if (!_.has(oldValues, field)) {
                if (this.hasOriginalValue(type, id, field)) {
                    oldValues[field] = this.getOriginalValue(type, id, field);
                } else {
                    pathItemEach($.extend(true, {}, originalItem), field, function (key, path, value) {
                        if (!_.has(oldValues, path)) {
                            oldValues[path] = value;
                        }
                    });
                }
            }
        },

        remove: function (type, id, field, purge) {
            var item = this.item(type, id);

            if (item != null) {
                var oldValues = item.oldValues,
                    oldValueKeys = _.keys(oldValues);

                if (field != null) {
                    delete oldValues[field];
                    delete item.trueModifiedFields[field];

                    _.each(oldValueKeys, function (path) {
                        if (path.indexOf(field + PATH_DELIMITER) === 0) {
                            delete oldValues[path];
                            delete item.trueModifiedFields[path];
                        }
                    });

                    var trueModifiedGroups = groupAllByPath(_.keys(item.trueModifiedFields)),
                        extraOldValues = _.difference(_.keys(oldValues), _.keys(trueModifiedGroups));

                    _.each(extraOldValues, function (path) {
                        delete oldValues[path];
                    });
                }

                // remove the item if no modified fields exist
                // If purge is true, get rid of the whole record - this is
                //     because the remove method for this collection is by
                //     field and not by id.
                if (_.isEmpty(oldValues) || purge === true) {
                    Collection.prototype.remove.call(this, type, id);
                }
            }

            // remove the type if type is empty
            if (this.isEmpty(type)) {
                this.clear(type);
            }
        }
    });

    return ModifyStoreCollection;
});


/**
 * A simple eventing system which supports namespacing.
 *
 *     var s = new Subscribable();
 *     s.on('test', function () { console.log('test'); }
 *     s.trigger('test');
 *     s.off('test');
 *
 * Other classes can inherit from Subscribable like so. It will create two variables on your object: `_subscribableListeners` and `_subscribableOptions`.
 *
 *     function MySubscribable () {
 *         // ...
 *         Subscribable.call(this, options); // options can be null/undefined
 *         // ...
 *     }
 *
 *     MySubscribable.prototype = Object.create(Subscribable.prototype);
 *     MySubscribable.prototype.otherMethod = function () { ... };
 *
 * @class utils.Subscribable
 * @module utils
 * @constructor
 */
define('primutils/utils/Subscribable',['primutils.lodash', 'primutils/utils/validator', 'primutils.logger'], function(_, validator, logger) {
    

    var SUBSCRIBABLE_LISTENERS_KEY = '_subscribableListeners',
        NO_NAMEPSACE = 'noNamespace';

    var removeAllListeners = function () {
        var listenerMap = this[SUBSCRIBABLE_LISTENERS_KEY];
        _.each(listenerMap, function (e, key) {
            delete listenerMap[key];
        });
    };

    function Subscribable (options) {
        logger.assert(this && this instanceof Subscribable, 'Subscribable must be created with \'new\'');

        /**
         * The map of event names to their proper functions.
         *
         * @example
         *      var _subscribableListeners = {
         *          select : {
         *              noNamespace : [ 'callBack1', 'callBack2' ],
         *              view : 'callBack',
         *              customNameSpace : 'callBack'
         *          }
         *      };
         *
         * @property _subscribableListeners
         * @type {Object}
         * @private
         * @default {}
         */
        Object.defineProperty(this, SUBSCRIBABLE_LISTENERS_KEY, {
            value : {},
            writable : false,
            enumerable : false
        });

        options = options || {};
        validator.checkType(options, ['Object']);
        this._validator(options);
        Object.defineProperty(this, '_subscribableOptions', {
            value: _.extend({}, options),
            writable: false,
            enumerable: false,
        });
    }

    var checkEventValidity = function (eventName) {
        var split = eventName.split('.'), ns = split[0], ev = split[1];

        if (ns === eventName) {
            ev = ns;
        }

        if (!_.contains(this._subscribableOptions.allowedEvents, ev)) {
            logger.error('The event spec: ' + eventName + ' is not allowed for this Subscribable.  Valid events are: [' + this._subscribableOptions.allowedEvents.join(',') + ']');
            return false;
        }

        return true;
    };

    Subscribable.prototype = {
        constructor : Subscribable,

        /**
         * Calls each function that matches the argument.
         * Argument can either be namespace.event, or just event
         *
         * If argument is just event name than all functions will be called otherwise
         * only listeners in the namespace will be called.
         *
         * Chainable since 2.7.3.
         *
         * @method trigger
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @chainable
         */
        trigger : function (eventName) {
            var split = eventName.split('.'),
                args = _.rest(_.toArray(arguments), 1);

            if (this._subscribableOptions.allowedEvents && !checkEventValidity.call(this, eventName)) {
                return;
            }

            if(split[0] === eventName) {
                // No name space given - fire all the 'eventName'
                // events even the namespaced ones
                _.each(this[SUBSCRIBABLE_LISTENERS_KEY][eventName], function(e, key) {
                    if (key === NO_NAMEPSACE) {
                        _.each(e, function(noNameEvent) {
                            noNameEvent.fn.apply(noNameEvent.context, args);
                        });
                    } else {
                        e.fn.apply(e.context, args);
                    }
                });
            } else {
                if (this[SUBSCRIBABLE_LISTENERS_KEY] && this[SUBSCRIBABLE_LISTENERS_KEY][split[1]]) {
                    var ev = this[SUBSCRIBABLE_LISTENERS_KEY][split[1]][split[0]];
                    if (!_.isUndefined(ev)) {
                        ev.fn.apply(ev.context, args);
                    }
                }
            }

            if(!_.isUndefined(this[SUBSCRIBABLE_LISTENERS_KEY]['*'])) {
                args = _.union(split, args);
                _.each(this[SUBSCRIBABLE_LISTENERS_KEY]['*'], function (e, key) {
                    if (key === NO_NAMEPSACE) {
                        _.each(e, function (noNameEvent) {
                            noNameEvent.fn.apply(noNameEvent.context, args);
                        });
                    } else {
                        e.fn.apply(e.context, args);
                    }
                });
            }

            return this;
        },

        /**
         * Registers a listener for the event name possibly under the given namespace.
         *
         * If no name space if given then it goes under the array 'noNamespace'.
         *
         * '*' as the event name is a special property that will be
         * fired for any event called.
         *
         * @method on
         * @chainable
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @param {Function} callBack The function to call when triggered.
         * @param {Object} [context] The context to use when calling the callBack.
         */
        on : function(eventName, callBack, context) {
            var that = this;

            function addEvent(ev, cb, ctx) {
                var events = ev.split(' ');

                if (that._subscribableOptions.allowedEvents && !checkEventValidity.call(that, eventName)) {
                    return;
                }

                _.each(events, function(event) {
                    var split = event.split('.'), nameSpace = split[0], e = split[1];
                    if (event === split[0]) {
                        //No name space was given
                        e = split[0];

                        if (_.isUndefined(that[SUBSCRIBABLE_LISTENERS_KEY][e])) {
                            that[SUBSCRIBABLE_LISTENERS_KEY][e] = {};
                        }
                        if (_.isUndefined(that[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE])) {
                            that[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE] = [];
                        }

                        that[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE].push({ fn: cb, context: ctx });
                    } else {
                        if (_.isUndefined(that[SUBSCRIBABLE_LISTENERS_KEY][e])) {
                            that[SUBSCRIBABLE_LISTENERS_KEY][e] = {};
                        }

                                                if (_.isObject(that[SUBSCRIBABLE_LISTENERS_KEY][e][nameSpace])) {
                            logger.info('A listener already exists for ' + event + ', and is being replaced.');
                        }
                        
                        that[SUBSCRIBABLE_LISTENERS_KEY][e][nameSpace] = { fn: cb, context: ctx };
                    }
                });
            }

            if(_.isObject(eventName)) {
                //Given a map
                //Ignore callback and context
                _.each(eventName, function (value, key) {
                    if (_.isFunction(value)) {
                        addEvent(key, value, undefined);
                    } else {
                        addEvent(key, value.fn, value.context);
                    }
                });
            } else if (_.isString(eventName)) {
                addEvent(eventName, callBack, context);
            }

            return this;
        },

        /**
         * Removes the function(s) for the given `namespace.event`.
         *
         * If no namespace is given - that is, the passed eventName does not include a '.' - AND no function is provided - all listeners for the given eventname, under any namespace, are removed.
         *
         * (Since 2.7.3) If a second parameter, the callback function, is provided, AND the eventName does not include a namespace, then the given function will be removed from the list of listeners not gathered under specific namespaces.
         *
         * @method off
         * @chainable
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @param {Function} [fn] a specific function to remove from the list of non-namespaced listeners
         */
        off : function(eventName, fn) {
            var events = eventName.split(' '),
                hasFnToRemove = arguments.length === 2;

            function filterByFn(listeners) {
                return _.filter(listeners, function (listener) {
                    return listener && listener.fn !== fn;
                });
            }

            var removeEvent = function (eventName) {
                if (this._subscribableOptions.allowedEvents && !checkEventValidity.call(this, eventName)) {
                    return;
                }

                var split = eventName.split('.'), nameSpace = split[0], e = split[1];
                if(eventName === split[0]){
                    //No namespace given
                    e = split[0];

                    if (!hasFnToRemove) {
                        //Remove all 'e' - including namespaced ones
                        delete this[SUBSCRIBABLE_LISTENERS_KEY][e];
                    }
                    else {
                        if (this[SUBSCRIBABLE_LISTENERS_KEY][e] && this[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE]) {
                            this[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE] = filterByFn(this[SUBSCRIBABLE_LISTENERS_KEY][e][NO_NAMEPSACE]);
                        }
                    }
                } else {
                    if (this && this[SUBSCRIBABLE_LISTENERS_KEY] && this[SUBSCRIBABLE_LISTENERS_KEY][e]) {
                        if (!_.isUndefined(this[SUBSCRIBABLE_LISTENERS_KEY][e][nameSpace])) {
                            delete this[SUBSCRIBABLE_LISTENERS_KEY][e][nameSpace];
                        }
                    }
                }
            }.bind(this);

            events.forEach(removeEvent);

            return this;
        },

        /**
         * Removes all events listeners tracked by this Subscribable.
         *
         * @method clear
         * @chainable
         * @since 2.0.8
         */
        clear: function () {
            if (!arguments.length) {
                removeAllListeners.call(this);
            }

            return this;
        },

        /**
         * The validator for this Subscribable.
         *
         * @property _validator
         * @protected
         * @since 2.7.2
         */
        _validator: validator({
            name: 'Subscribable',
            validOptions: {
                /**
                 * An array of event names that are valid for `on`/`off`/`trigger` method calls.
                 * If this option is present, any attempts to call any methods on this Subscribable for events not listed in the array will trigger logger.error calls and be ignored.
                 *
                 * @property allowedEvents
                 * @public
                 * @type {Array|undefined}
                 * @default undefined
                 */
                allowedEvents: ['Array', undefined]
            }
        })
    };

    /**
     * Alias for `.on`.
     *
     * @method subscribe
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.subscribe = Subscribable.prototype.on;

    /**
     * Alias for `.off`.
     *
     * @method unsubscribe
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.unsubscribe = Subscribable.prototype.off;

    /**
     * Alias for `.trigger`.
     *
     * @method publish
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.publish = Subscribable.prototype.trigger;

    return Subscribable;

});

/**
 * Store allows you to work with different models(objects) for your page.  Some of the key benefits of using the store are:
 * <ul>
 *     <li>Ordered keyed list.  Allows for item retrieval via key.</li>
 *     <li>Hold many different models.  EG collection of projects, collection of categories, object of meta data, etc.</li>
 *     <li>Original values are held after modification.  This gives an easy mechanism to revert changes.</li>
 *     <li>Supplies a pre-determined batch save object of any added, deleted and modified objects</li>
 *     <li>Triggers post events on the main operations</li>
 * </ul>
 *
 * *Information about Keys*
 *
 * The store only works with keys that are strings.
 * The reason behind this is that for many operations the key is used as the key in a JavaScript object hash.
 * This key could be strictly compared against your object's key, and if you are not using a string key this comparison will fail.
 *
 * *Updating Fields*
 *
 * Updating fields is done through the `.update` method.
 * You can update entire objects or navigate directly through to the path you want to update.
 * Paths are always separated by periods (.).
 * Avoid using complex keys (those needing bracket notation `foo['bar.bar']`) which contain periods.
 *
 *     .update('type', 'id', 'path.to.update', _value_);
 *
 * You should never update your object's key property directly.
 * The store uses several internal collections to maintain modification (also added/deleted) state of your object.
 * If you change the key of your object you will lose all of the modification information.
 * You should only update your object's key through either of the `.commit` or `.mergeAndCommit` methods.
 *
 * You should also avoid updating specific keys in arrays.
 * Always refresh the entire array instead of adding a new index.
 *
 * Example Object:
 *
 *     {
 *         'foo' : [ 1, 2, 3, 4 ]
 *     }
 *
 * Bad:
 *
 *     .update('type', 'id', 'foo.5', 5);
 *
 * Not Great (but would work):
 *
 *     .update('type', 'id', 'foo.1', 100);
 *
 * Best:
 *
 *     .update('type', id', 'foo', [ 100, 2, 3, 4, 5 ]);
 *
 * *What is Committing?*
 *
 * Committing your store indicates that your model is no longer dirty and that you no longer want to be able to revert
 * any of the changes that have already been made.
 * It is also a way of indicating that your model is synchronized with an outside source like a web server.
 *
 * There are two types of commit: plain committing and merge-and-committing.
 * Plain committing means that the previous history about your object is lost and the next time you reset your store
 * modifications will be restored to this point.
 * Merging and committing means that you would also like to take in updates from an outside source as part of your commit.
 * See information about the types of merging that the store supports in the `mergeMode` option.
 *
 * Both types of merging take a list of "new keys".
 * The new keys are a way of safely updating keys on new store objects to a new value from an outside source like a web server.
 * Generally when you add a new record, you give it a negative id to indicate it is new (see `_isItemNew`).
 * Passing new keys allows you to convert those new records to regular modifications on subsequent saves.
 *
 *     //Example
 *     {
 *         // Old Key    New Key
 *         "-1"        : "23",
 *         "-2"        : "24"
 *     }
 *
 * Any objects that do not have a corresponding new key will be left alone, but their modification state will be dropped.
 *
 * *About Resetting*
 *
 * Resetting allows your to return some or all of your objects back to their state before modification (i.e. since the last
 * call to `.commit` or `.add`.
 * Resetting new items will remove them from the store, and resetting deleted items will add them back to their last modified value.
 * There is essentially only one "checkpoint" which is created when you call `.commit`.
 * You can only reset back to the last commit state, so if you plan on resetting to some specific state make sure you call commit first.
 *
 * *Updating Store Options*
 *
 * The store does not provide a way to update its options because of the multitude of side-effects that most of the options have.
 * For example, changing the `key` option could affect unique key constraints and force reindexing on the store.
 * How would changing the options for flex and code values work? Do the old codes get reserialized, dropped, left alone?
 * Do the updated ones get auto-deserialized?
 *
 * The only place this is not enforced is for dates (see the `setDateFields` option).
 *
 * Example:
 *
 *      var META_ID = '1';
 *
 *      // returned server json
 *      var jsonResponse = {
 *          meta : {
 *              columns : {}
 *          },
 *          data : [{projectA}, {projectB}]
 *      };
 *
 *      var store = new Store({key : 'id'});
 *      store.add('projects', jsonResponse.data);
 *      store.add('meta', jsonResponse.meta, function(type, item){
 *          // add an id to the meta for any retrieval and updating
 *          item.id = META_ID;
 *          return item;
 *      });
 *
 *      store.on('update', function(args){
 *          if (this.isDirtyAll()){
 *              // store is dirty
 *          } else {
 *              // store is clean
 *          }
 *      });
 *
 *      // get objects
 *      var project = store.item('project', 'projectA');
 *      var columns = store.item('meta', META_ID).columns;
 *
 *      // modify a value
 *      store.update('project', 'projectA', 'code', 'prj-002');
 *
 *      // get save object
 *      // {
 *      //      added : [],
 *      //      deleted : [],
 *      //      modified : [{projectA}]
 *      // }
 *      var saveObject = store.saveAll();
 *
 *      // commit changes
 *      store.commitAll({
 *          storeTypeFoo: { // new IDs for the items in storeType 'storeTypeFoo'
 *              "-5": "100",
 *              "-6": "101"
 *          }
 *      });
 *
 *      // cancel changes
 *      store.resetAll();
 *
 *
 * @class store.Store
 * @module store
 * @param options Options for the store.
 * @constructor
 */

define('primutils/store/Store',['primutils.jquery', 'primutils.lodash', 'primutils.logger', 'primutils/store/Collection', 'primutils/store/ModifyCollection', 'primutils/utils/Subscribable',
    'primutils.moment', 'primutils/utils/nestedPropertyAccessors'],
    function ($, _, logger, Collection, ModifyCollection, Subscribable, moment, propHelpers) {
        

        var MERGE_MODES = {
            NONE : 'none',
            REPLACE : 'replace',
            EXTEND_EXISTING : 'extendExisting',
            EXTEND_EXISTING_SHALLOW : 'extendExistingShallow',
            EXTEND_NEW : 'extendNew',
            EXTEND_NEW_SHALLOW : 'extendNewShallow'
        };
        MERGE_MODES.default = MERGE_MODES.EXTEND_EXISTING;

        var DEFAULT_OPTIONS = {
            /**
             * Field used to index the object
             *
             * @property key
             * @type String
             * @optional
             * @default 'id'
             */
            key : 'id',
            /**
             * Types that cannot be changed through the store
             *
             * @property readOnly
             * @type Array(String)
             * @optional
             */
            readOnly : [],
            /**
             * Allows the creation of dependent sub-objects that are updated when their parent object is updated.
             *
             * The relationship must be defined through a `parentId` field.
             *
             * (This feature is experimental.)
             * @property associatedTypes
             * @type Object
             * @optional
             */
            associatedTypes : {},
            /**
             * Field used in specific types to index the object
             *
             * @property typeKeys
             * @type Map(String,String)
             * @Optional
             */
            typeKeys : {},
            /**
             * Override options per type
             *
             * Only some options are supported: `key`, `dateFields`, `recursiveSerialization`, `flex*Key` (e.g. `flexOtherKey`),
             * `flex*Search` (e.g. `flexTestSearch`), `codeKey`, `codePrefix`, `codeSearch`
             *
             * @property typeOptions
             * @type Map(String, Object)
             * @Optional
             */
            typeOptions : {},
            /**
             * Date formatter to be used to unserialize and serialize dates
             *
             * Must provide two methods:
             *  * `unserialize` - convert string from server into MomentJS
             *  * `serialize` - convert MomentJS into string for server
             *
             * @property dateFormatter
             * @type Object
             */
            dateFormatter : {
                unserialize : function (value) {
                    return value;
                },
                serialize : function (value) {
                    return value;
                }
            },

            /**
             * Fields which are automatically deserialized into moment objects on objects
             *
             * @property dateFields
             * @type Array
             */
            dateFields : [],
            /**
             * Recursively do serialization and deserialization
             *
             * Recursive serialization may be useful if you have nested ROs (Rest Objects) coming back from the server.
             *
             * For example, a Scope Item Assignment has its own information as well as the Scope Item itself.
             *
             *     {
             *         "plannedUnits" : 30,
             *
             *         // Scope Item Assignment Flex Fields, Codes, Dates
             *
             *         "scopeItem" : {
             *             "name" : "Scope Item 1",
             *             "availableUnits" : 40,
             *
             *             // Scope Item Flex Fields, Codes, Dates
             *         }
             *     }
             *
             * By default, the store will not attempt to deserialize the scope item's flex fields (code values, date fields) for you.
             * In the past, you had to use the `modifyItemFn` to do subitem deserialization.
             *
             *     store.addAll('scopeItemAssignments', [...], function (item) {
             *         store._deserialize('scopeItemAssignments', item.scopeItem);
             *     }, dateFields);
             *
             * Setting this property to true means you no longer have to do this.
             *
             * @property recursiveSerialization
             * @type Boolean
             */
            recursiveSerialization : false,
            /**
             * How the merge methods work:
             *
             * * Store.mergeModes.NONE : Only add the item to the store if a value with that key (or corresponding "old key") does not already exist. This is the legacy behavior.
             * * Store.mergeModes.REPLACE : Replace the object currently in the store with that from the server.
             * * Store.mergeModes.EXTEND_EXISTING : (Default) Do a deep extends with the current store object as the target. Values from the merge object will take precedence over those currently in the store.
             * * Store.mergeModes.EXTEND_EXISTING_SHALLOW : Do an extends with the current store object as the target.
             * * Store.mergeModes.EXTEND_NEW : Do a deep extends with the merge object as the target. Values from the store object will take precedence over those from the merge object.
             * * Store.mergeModes.EXTEND_NEW_SHALLOW : Do an extends with the merge object as the target.
             *
             * @property mergeMode
             * @type {String}
             * @default EXTEND_EXISTING
             */
            mergeMode : MERGE_MODES.default,

            /**
             * Key under which other flex fields live
             *
             * @property flexOtherKey
             * @type String
             * @default 'flex'
             */
            flexOtherKey : 'flex',
            /**
             * Key under which text flex fields live
             *
             * @property flexTextKey
             * @type String
             * @default 'textFlex'
             */
            flexTextKey : 'textFlex',
            /**
             * Key under which cost flex fields live
             *
             * @property flexCostKey
             * @type String
             * @default 'costFlex'
             */
            flexCostKey : 'costFlex',

            /**
             * Regular expression for determining whether a field is an other flex field
             *
             * @property flexOtherSearch
             * @type RegExp
             * @default /_FO-/
             */
            flexOtherSearch : /_FO-/,
            /**
             * Regular expression for determining whether a field is a text flex field
             *
             * @property flexTextSearch
             * @type RegExp
             * @default /_FT-/
             */
            flexTextSearch : /_FT-/,
            /**
             * Regular expression for determining whether a field is a cost flex field
             *
             * @property flexCostSearch
             * @type RegExp
             * @default /_FC-/
             */
            flexCostSearch : /_FC-/,

            /**
             * Key under which object code values live
             *
             * @property codeKey
             * @type String
             * @default 'projectCategories'
             */
            codeKey : 'projectCategories',
            /**
             * Prefix applied when deserializing object code values
             *
             * @property codePrefix
             * @type String
             * @default 'CO_CODE_TYPE-'
             */
            codePrefix : 'CO_CODE_TYPE-',
            /**
             * Regular expression for determining whether a field is a object code value
             *
             * Generally you want this to match your code prefix
             *
             * @property codeSearch
             * @type RegExp
             * @default /CO_CODE_TYPE-/
             */
            codeSearch : /CO_CODE_TYPE-/
        };

        var PATH_DELIMITER = '.';
        var STUB = 'DO-NOT-DELETE';
        var UNSERIALIZED_MARKER = '$unserialized$';
        var ASSOCIATED_DELIM = '__associated_';

        function serializeAssociatedField(fieldName) {
            return ASSOCIATED_DELIM + fieldName;
        }

        var getKey = function (type) {
            return getTypedOption.call(this, type, 'key');
        };

        var getTypedOption = function (type, option) {
            var value = this.options.typeOptions[type] && this.options.typeOptions[type][option];
            if (value == null && option === 'key') {
                value = this.options.typeKeys[type];
            }
            return value == null ? this.options[option] : value;
        };

        var _unserialize = function (type, item, dateFields, recursive) {
            var that = this,
                flexCostKey = getTypedOption.call(this, type, 'flexCostKey'),
                flexTextKey = getTypedOption.call(this, type, 'flexTextKey'),
                flexKey = getTypedOption.call(this, type, 'flexOtherKey'),
                projectCategoryKey = getTypedOption.call(this, type, 'codeKey'),
                projectCategoryPrefix = getTypedOption.call(this, type, 'codePrefix'),
                recursiveFields;

            function fieldsOfFlexType(item, flexType) {
                return _.has(item, flexType) ? _.keys(item[flexType]) : [];
            }

            function hasUnserialzedFlex(item, flexType) {
                return _.has(item, flexType) && item[flexType][UNSERIALIZED_MARKER] === true;
            }

            // Unserialize flex fields
            _.each([flexKey, flexCostKey, flexTextKey], function (flexType) {
                if (!hasUnserialzedFlex(item, flexType)) {
                    _.each(fieldsOfFlexType(item, flexType), function (fieldName) {
                        item[fieldName] = item[flexType][fieldName];
                        item[flexType][fieldName] = STUB;
                    });
                }

                if (_.has(item, flexType)) {
                    item[flexType][UNSERIALIZED_MARKER] = true;
                }
            });

            // Unserialize project category fields
            if (_.has(item, projectCategoryKey) && item[projectCategoryKey][UNSERIALIZED_MARKER] !== true) {
                _.each(item[projectCategoryKey], function (projectCategory, typeId) {
                    item[projectCategoryPrefix + typeId] = projectCategory;
                    item[projectCategoryKey][typeId] = STUB;
                });
                item[projectCategoryKey][UNSERIALIZED_MARKER] = true;
            }

            // Unserialize date fields
            _.each(dateFields, function (dateField) {
                if (_.has(item, dateField) && !moment.isMoment(item[dateField])) {
                    item[dateField] = that.options.dateFormatter.unserialize(item[dateField]);
                }
            });

            // Do recursion
            if (recursive) {
                recursiveFields = _.difference(_.keys(item), dateFields || [], [flexCostKey, flexTextKey, flexKey, projectCategoryKey]);
                _.each(recursiveFields, function (fieldName) {
                    if (_.isObject(item[fieldName])) {
                        if (_.isArray(item[fieldName])) {
                            item[fieldName] = _.map(item[fieldName], function (value) {
                                return _unserialize.call(that, type, value, dateFields, recursive);
                            });
                        } else {
                            item[fieldName] = _unserialize.call(that, type, item[fieldName], dateFields, recursive);
                        }
                    }
                });
            }

            return item;
        };

        var _serialize = function (type, item, recursive) {
            var that = this,
                flexKey = getTypedOption.call(this, type, 'flexOtherKey'),
                flexCostKey = getTypedOption.call(this, type, 'flexCostKey'),
                flexTextKey = getTypedOption.call(this, type, 'flexTextKey'),
                projectCategoryKey = getTypedOption.call(this, type, 'codeKey'),
                flexFields = _.chain(item).keys().groupBy(function (fieldName) {
                        if (isFlexField.call(that, type, fieldName)) {
                            return flexKey;
                        }
                        else if (isFlexCostField.call(that, type, fieldName)) {
                            return flexCostKey;
                        }
                        else if (isFlexTextField.call(that, type, fieldName)) {
                            return flexTextKey;
                        }
                        return '__private_not_flex__';
                    }).omit('__private_not_flex__').value(),
                projectCategoryFields = _.filter(_.keys(item), function (fieldName) {
                    return isProjectCategoryField.call(that, type, fieldName);
                }),
                recursiveFields = recursive ? _.difference(_.keys(item), [flexCostKey, flexTextKey, flexKey, projectCategoryKey, flexFields, projectCategoryFields]) : [];

            // Serialize date fields and do recursion
            if (recursive) {
                _.each(recursiveFields, function (fieldName) {
                    if (moment.isMoment(item[fieldName])) {
                        item[fieldName] = that.options.dateFormatter.serialize(item[fieldName]);
                    } else if (_.isObject(item[fieldName])) {
                        if (_.isArray(item[fieldName])) {
                            item[fieldName] = _.map(item[fieldName], function (value) {
                                return _serialize.call(that, type, value, recursive);
                            });
                        } else {
                            item[fieldName] = _serialize.call(that, type, item[fieldName], recursive);
                        }
                    }
                });
            } else { // Or just date fields
                _.each(item, function (value, fieldName) {
                    if (moment.isMoment(value)) {
                        item[fieldName] = that.options.dateFormatter.serialize(value);
                    }
                });
            }

            // Serialize flex fields
            _.each(flexFields, function (fieldNames, flexType) {
                _.each(fieldNames, function (fieldName) {
                    item[flexType] = item[flexType] || {};
                    item[flexType][fieldName] = item[fieldName];
                    delete item[fieldName];
                });
            });

            if (item[flexKey]) {
                delete item[flexKey][UNSERIALIZED_MARKER];
            }
            if (item[flexCostKey]) {
                delete item[flexCostKey][UNSERIALIZED_MARKER];
            }
            if (item[flexTextKey]) {
                delete item[flexTextKey][UNSERIALIZED_MARKER];
            }

            // Serialize project categories
            if (projectCategoryFields.length > 0) {
                item[projectCategoryKey] = item[projectCategoryKey] || {};

                _.each(projectCategoryFields, function (fieldName) {
                    var projectCategory = item[fieldName];

                    if (projectCategory != null && _.has(projectCategory, 'typeId')) {
                        item[projectCategoryKey][projectCategory.typeId] = projectCategory;
                    }

                    delete item[fieldName];
                });
            }

            if (item[projectCategoryKey]) {
                // Filter out all of the projectCategories that were deleted from the object
                item[projectCategoryKey] = _.reduce(item[projectCategoryKey], function (result, catVal, typeId) {
                    if (catVal !== STUB) {
                        result[typeId] = catVal;
                    }
                    return result;
                }, {});

                delete item[projectCategoryKey][UNSERIALIZED_MARKER];
            }

            return item;
        };

        var updateAssociatedObjects = function (type, item) {
            var that = this,
                key = getKey.call(this, type);

            // update parent object if an associated object
            if (_.has(this.options.associatedTypes, type)) {
                var parentAssociatedType = that.options.associatedTypes[type];

                var parentItems = that.items(parentAssociatedType.type, function (parentItem) {
                    return that._itemKey(parentAssociatedType.type, parentItem) === item[parentAssociatedType.parentId];
                });

                if (parentItems.length > 1) {
                    throw new Error('store.js - can\'t have more than one parent');
                } else if (parentItems.length) {
                    var parentId = that._itemKey(parentAssociatedType.type, _.first(parentItems));
                    var childItems = $.extend(true, [], that.items(type,
                        function (childItem) {
                            return parentId === childItem[parentAssociatedType.parentId];
                        }));
                    _.each(childItems, function (item) {
                        delete item[key];
                        delete item[parentAssociatedType.parentId];
                    });
                    that.update(parentAssociatedType.type, parentId,
                        serializeAssociatedField(type), childItems);
                }

            }
        };

        var isFlexField = function (type, fieldName) {
            return fieldName.search(getTypedOption.call(this, type, 'flexOtherSearch')) > -1;
        };

        var isFlexCostField = function (type, fieldName) {
            return fieldName.search(getTypedOption.call(this, type, 'flexCostSearch')) > -1;
        };

        var isFlexTextField = function (type, fieldName) {
            return fieldName.search(getTypedOption.call(this, type, 'flexTextSearch')) > -1;
        };

        var isProjectCategoryField = function (type, fieldName) {
            return fieldName.search(getTypedOption.call(this, type, 'codeSearch')) > -1;
        };

        var getDateFields = function (type) {
            if (!_.has(this._dateFieldsCache, type)) {
                var global = this.options.dateFields || [],
                    typed = getTypedOption.call(this, type, 'dateFields') || [];
                this._dateFieldsCache[type] = _.union(global, typed);
            }
            return this._dateFieldsCache[type];
        };

        var Store = function (options) {
            var that = this,
                storeCollectionOptions = {
                    getKey : function (type) {
                        return getKey.call(that, type);
                    }
                };

            this.options = $.extend(true, {}, DEFAULT_OPTIONS, options);
            this._dateFieldsCache = {};

            // remap associations for other perspective - needed for commit
            this.options.associatedChildTypes = {};

            _.each(this.options.associatedTypes, function (parentAssociation, type) {
                if (!_.has(that.options.associatedChildTypes, parentAssociation.type)) {
                    that.options.associatedChildTypes[parentAssociation.type] = [];
                }
                that.options.associatedChildTypes[parentAssociation.type].push(type);
            });

            // event handling
            this._subscribable = new Subscribable();
            this.storeCollection = new Collection('store', storeCollectionOptions);
            this.modifyStoreCollection = new ModifyCollection('modify', storeCollectionOptions);
            this.deleteStoreCollection = new Collection('delete', storeCollectionOptions);
            this.newStoreCollection = new Collection('new', storeCollectionOptions);
        };

        Store.prototype = {
            constructor: Store,

            utilities: {

                /**
                 * Return a value at any item depth
                 *
                 * If the path extends beyond the actual object no exception is thrown and `null` is returned.
                 *
                 * Example:
                 *
                 *      var item = {
                 *          bar : 'bar',
                 *          foo : {
                 *              foo : 'foo'
                 *          }
                 *      };
                 *      console.log(store.utilities._getValue(item, 'bar'));              // 'bar'
                 *      console.log(store.utilities._getValue(item, 'foo'));              // Object {foo: "foo"}
                 *      console.log(store.utilities._getValue(item, 'foo.foo'));          // 'foo'
                 *      console.log(store.utilities._getValue(item, 'foo.foo.foo'));      // null (no exception)
                 *
                 * @method utilities._getValue
                 * @param {Object} item The item to search
                 * @param {String} field The field name to retrieve
                 * @return {*} The value of the object at that field
                 * @private
                 */
                _getValue: function (item, field) {
                    return propHelpers.getValue(item, field, PATH_DELIMITER);
                },

                /**
                 * Set a value at any item depth
                 *
                 * If the path extends beyond the actual object an exception thrown
                 *
                 * Example:
                 *
                 *      var item = {
                 *          bar : 'bar',
                 *          foo : {
                 *              foo : 'foo'
                 *          }
                 *      };
                 *      store.utilities._setValue(item, 'bar', 'foo');
                 *      console.log(store.utilities._getValue(item, 'bar'));             // 'foo'
                 *      store.utilities._setValue(item, 'foo.foo', 'bar');
                 *      console.log(store.utilities._getValue(item, 'foo.foo'));         // 'bar'
                 *      store.utilities._setValue(item, 'foo.bar.buzz', 'fizzbuzz'));    // Exception
                 *      store.utilities._setValue(item, 'foo.bar', 'fizzbuzz'));         // No Exception
                 *      console.log(store.utilities._getvalue(item, 'foo.bar'));         // 'fizzbuzz'
                 *
                 * @method utilities._setValue
                 * @param {Object} item The item to search
                 * @param {String} field The field name to set
                 * @param {*} value The new value of the field
                 * @private
                 */
                _setValue: function (item, field, value) {
                    return propHelpers.setValue(item, field, value, PATH_DELIMITER);
                },

                /**
                 * Ajax get call
                 *
                 * Use with $.when()
                 *
                 * @method utilities.GET
                 * @param {String} url The URL to get
                 * @return {jQuery Promise} A jQuery promise
                 */
                GET: function (url) {
                    return $.ajax({
                        url: url,
                        dataType: 'json',
                        type: 'GET'
                    });
                },

                /**
                 * Ajax put call
                 *
                 * Use with $.when()
                 *
                 * @method utilities.PUT
                 * @param {String} url The URL to PUT to
                 * @param {Object} data The data to send
                 * @param {Boolean} async Whether to send the data asynchronously
                 * @return {jQuery Promise} A jQuery Promise
                 */
                PUT: function (url, data, async) {
                    async = (async == null) ? true : async;
                    return $.ajax({
                        url: url,
                        type: 'PUT',
                        data: (_.isObject(data) ? JSON.stringify(data) : data),
                        contentType: 'application/json',
                        async: async
                    });
                }
            },

            /**
             * Unserializes fields in a object that can be used client-side
             *
             * Does all unserializations: date, flex objects, code values
             *
             * @param type {String} Store type being unserialized
             * @param item {Object} Object being unserialized
             * @param dateFields {Array} Array of data fields to convert to moment objects
             * @param modifyItemFn {Function} (type, object) -> object: Additional modifications to perform after unserialization
             * @private
             */
            _unserialize : function (type, item, dateFields, modifyItemFn) {
                var recursive = getTypedOption.call(this, type, 'recursiveSerialization');

                if (_.isFunction(modifyItemFn)) {
                    var newItem = modifyItemFn(type, item);
                    if (newItem != null) {
                        item = newItem;
                    }
                }

                _unserialize.call(this, type, item, dateFields, recursive);
            },

            /**
             * Serializes fields in a object
             *
             * Does all serializations: date, flex objects, code values
             *
             * @param type {String} Store type being serialized
             * @param item {Object} Object being serialized
             * @param serializeFn {Function} (type, object) -> object: Additional modifications to perform after serialization
             * @private
             */
            _serialize : function (type, item, serializeFn) {
                var recursive = getTypedOption.call(this, type, 'recursiveSerialization');
                _serialize.call(this, type, item, recursive);
                if (_.isFunction(serializeFn)) {
                    serializeFn(type, item);
                }
            },

            /**
             * Determines if the item belongs in the new item collection
             *
             * The default behavior is to treat all items with keys that parse to negative numbers as new
             *
             * -1 : New
             * 1  : Not new
             *
             * @method _isItemNew
             * @param {String} type
             * @param {Object} item
             * @return Boolean
             * @private
             */
            _isItemNew: function (type, item) {
                if (item != null) {
                    return parseInt(this._itemKey(type, item), 10) < 0;
                }

                throw new Error('store._isItemNew() - item undefined');
            },

            /**
             * Return item key property
             *
             * @method _itemKey
             * @param {String} type
             * @param {Object} item
             * @return String
             * @private
             */
            _itemKey: function (type, item) {
                if (item != null) {
                    var key = getKey.call(this, type);
                    return item[key];
                }

                throw new Error('store._itemKey() - item undefined');
            },

            /**
             * Return whether the store contains any objects of the given type
             *
             * @method isEmpty
             * @param {String} type
             * @return Boolean
             */
            isEmpty: function (type) {
                return this.storeCollection.isEmpty(type);
            },

            /**
             * Return whether the store is completely empty
             *
             * @method isEmptyAll
             * @return Boolean
             */
            isEmptyAll: function () {
                return this.storeCollection.isEmptyAll();
            },

            /**
             * Return the amount of items for a given type
             *
             * @method count
             * @param {String} type
             * @return Integer
             */
            count: function (type) {
                return this.storeCollection.count(type);
            },

            /**
             * Add an array of items to the store to a given type
             *
             * <b>Note</b> - If the item key property meets the `_isItemNew` criteria it is treated as new
             *
             * Triggers the `add` event for each item added to the store.
             * Triggers the 'add-all' event.
             *
             * Example:
             *
             *      store.addAll('type', [{}, {}], function(type, item){
             *          item.foo = 'bar';   // add another property
             *          return item;
             *      });
             *
             * @method addAll
             * @param {String} type
             * @param {Array} items
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             */
            addAll: function (type, items, modifyItemFn, dateFields) {
                var that = this;
                _.each(items, function (item) {
                    that.add(type, item, modifyItemFn, dateFields);
                });

                /**
                 * Triggered by the store when `store.addAll` is called.
                 *
                 * Args:
                 * * type: the type of which all items were added
                 *
                 * @event add-all
                 * @since 2.1.0
                 */
                this._subscribable.trigger('add-all', {
                    type: type
                });
            },

            /**
             * Add item to store for a given type
             *
             * <b>Note</b> - If the item key property meets the `_isItemNew` criteria it is treated as new
             *
             * Triggers Event - 'add'
             *
             * @method add
             * @param {String} type
             * @param {Object} item
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             */
            add: function (type, item, modifyItemFn, dateFields) {
                if (item == null) {
                    return;
                }

                var allDateFields = getDateFields.call(this, type);
                if (dateFields != null) {
                    allDateFields = _.union(allDateFields, dateFields);
                }

                this._unserialize(type, item, allDateFields, modifyItemFn);

                this.storeCollection.add(type, item);

                if (this._isItemNew(type, item)) {
                    if (!_.contains(this.options.readOnly, type)) {
                        this.newStoreCollection.add(type, item);
                        updateAssociatedObjects.call(this, type, item);
                    }
                }

                /**
                 * Triggered when an item is added to the store
                 *
                 * Args:
                 *  * type: The type of the added item
                 *  * item: The item added
                 *  * position: The position of the item in the store
                 *
                 * @event add
                 * @param args {Object}
                 *
                 */
                this._subscribable.trigger('add', {
                    type: type,
                    item: item,
                    position: this.count(type) - 1
                });
            },

            /**
             * Add an array of items to the store for a given type.
             * If the item already exists in the store it is ignored, otherwise it is added
             *
             * Since 2.7.0 this no longer triggers the 'add' event.
             *
             * @method merge
             * @deprecated prefer mergeAndCommit
             * @param {String} type
             * @param {Array} items
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             */
            merge: function (type, items, modifyItemFn, dateFields) {
                logger.warn('The merge method is deprecated and only uses the legacy "None" merge mode. Prefer mergeAndCommit.');
                this.mergeAndCommit(type, {}, items, modifyItemFn, dateFields, MERGE_MODES.NONE, true);
            },

            /**
             * Merges and commits several types at once.
             *
             * See additional information in the `mergeAndCommit` method.
             *
             * This method also triggers the 'commit-all' method.
             *
             * @method mergeAndCommitAll
             * @param newKeyMap {Object}
             * @param newItemsMap {Object}
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             * @param mergeMode {String}
             */
            mergeAndCommitAll: function (newKeyMap, newItemsMap, modifyItemFn, dateFields, mergeMode) {
                var that = this,
                    types = _.union(_.keys(newKeyMap), _.keys(newItemsMap));
                _.each(types, function (type) {
                    that.mergeAndCommit(type, newKeyMap[type] || {}, newItemsMap[type] || [], modifyItemFn, dateFields, mergeMode);
                });

                // Trigger commit all since we're not actually really calling the commitAll function
                that._subscribable.trigger('commit-all');
            },

            /**
             * Combines new data with existing data according to the merge mode and then performs a commit.
             *
             * See information about the merge modes under the `mergeMode` option.
             *
             * This method triggers the 'commit' method.
             *
             * @method mergeAndCommit
             * @param {String} type
             * @param {Object} newKeys
             * @param {Array} items
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             * @param {String} [mergeMode]
             */
            mergeAndCommit: function (type, newKeys, items, modifyItemFn, dateFields, mergeMode, skipCommit) {
                var that = this,
                    oldKeys = _.invert(newKeys),
                    key = getKey.call(this, type),
                    shallowExtend;
                mergeMode = mergeMode || getTypedOption.call(this, type, 'mergeMode');
                shallowExtend = mergeMode === MERGE_MODES.EXTEND_EXISTING_SHALLOW || mergeMode === MERGE_MODES.EXTEND_NEW_SHALLOW;

                if (mergeMode === MERGE_MODES.NONE) {
                    _.each(items, function (item) {
                        var id = that._itemKey(type, item),
                            hasOldId = _.has(oldKeys, id),
                            oldId = hasOldId ? oldKeys[id] : null;
                        if (that.storeCollection.item(type, id) == null &&
                                that.deleteStoreCollection.item(type, id) == null &&
                                that.modifyStoreCollection.item(type, id) == null &&
                                (!hasOldId || (that.storeCollection.item(type, oldId) == null &&
                                that.deleteStoreCollection.item(type, oldId) == null &&
                                that.modifyStoreCollection.item(type, oldId) == null))) {
                            that.add(type, item, modifyItemFn, dateFields);
                        }
                    });
                } else if (mergeMode === MERGE_MODES.REPLACE) {
                    _.each(items, function (item) {
                        that._unserialize(type, item, dateFields, modifyItemFn);
                        // Use the old key to swap keys so the commit is always last
                        item[key] = _.has(oldKeys, item[key]) ? oldKeys[item[key]] : item[key];
                        that.storeCollection.remove(type, item[key]);
                        that.storeCollection.add(type, item);
                    });
                } else if (mergeMode === MERGE_MODES.EXTEND_EXISTING || mergeMode === MERGE_MODES.EXTEND_EXISTING_SHALLOW) {
                    _.each(items, function (item) {
                        var oldItem = that.storeCollection.item(type, oldKeys[item[key]]) || {};
                        that._unserialize(type, item, dateFields, modifyItemFn);
                        // Use the old key to swap keys so the commit is always last
                        var newItem = $.extend(!shallowExtend, {}, oldItem, item);
                        newItem[key] = _.has(oldKeys, item[key]) ? oldKeys[item[key]] : item[key];
                        that.storeCollection.remove(type, newItem[key]);
                        that.storeCollection.add(type, newItem);
                    });
                } else if (mergeMode === MERGE_MODES.EXTEND_NEW || mergeMode === MERGE_MODES.EXTEND_NEW_SHALLOW) {
                    _.each(items, function (item) {
                        var oldItem = that.storeCollection.item(type, oldKeys[item[key]]) || {};
                        that._unserialize(type, item, dateFields, modifyItemFn);
                        // Use the old key to swap keys so the commit is always last
                        var newItem = $.extend(!shallowExtend, {}, item, oldItem);
                        newItem[key] = _.has(oldKeys, item[key]) ? oldKeys[item[key]] : item[key];
                        that.storeCollection.remove(type, newItem[key]);
                        that.storeCollection.add(type, newItem);
                    });
                }

                if (!skipCommit) {
                    that.commit(type, newKeys);
                }
            },

            /**
             * Add item to store at a particular position to a given type
             *
             * Triggers Event - 'insert'
             *
             * @method insertAt
             * @param {String} type
             * @param {Object} item
             * @param {Integer} position
             * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
             * @param {Array} [dateFields] Additional date fields to deserialize
             */
            insertAt: function (type, item, position, modifyItemFn, dateFields) {
                if (item == null) {
                    return;
                }

                var allDateFields = getDateFields.call(this, type);
                if (dateFields != null) {
                    allDateFields = _.union(allDateFields, dateFields);
                }

                this._unserialize(type, item, allDateFields, modifyItemFn);

                this.storeCollection.insertAt(type, item, position);

                if (this._isItemNew(type, item)) {
                    if (!_.contains(this.options.readOnly, type)) {
                        this.newStoreCollection.add(type, item);
                        updateAssociatedObjects.call(this, type, item);
                    }
                }

                /**
                 * Triggered when an item is inserted into the store
                 *
                 * Args:
                 *  * type: The type of the inserted item
                 *  * item: The item added
                 *  * position: The position of the item in the store
                 *
                 * @event insert
                 * @param args {Object}
                 */
                this._subscribable.trigger('insert', {
                    type: type,
                    item: item,
                    position: position
                });
            },

            /**
             * Return position of item in store for a given type
             *
             * @method position
             * @param {String} type
             * @param {String} key
             * @return Integer
             */
            position: function (type, key) {
                return this.storeCollection.position(type, key);
            },

            /**
             * Return an array of items from the store for a given type
             *
             * To include items that have been deleted set the <i>includeDeleted</i> param to true
             *
             * To filter the return list pass a function in for <i>filterFn</i> param
             *
             * Example:
             *
             *      var items = store.items('type', function(type, item) {
             *          // evaluate to true to include item
             *          return item.foo === 'foo';
             *      });
             *
             * @method items
             * @param {String} type
             * @param {Function} [filterFn] (item) -> boolean: Filter items that match a condition
             * @param {Boolean} [includeDeleted] Also return items that are marked for deletion
             * @return Array
             */
            items: function (type, filterFn, includeDeleted) {

                var items = [];
                if (includeDeleted) {
                    items = _.union(this.storeCollection.items(type), this.deleteStoreCollection.items(type));
                } else {
                    items = this.storeCollection.items(type);
                }

                if (_.isFunction(filterFn)) {
                    var tmpItems = [];
                    _.each(items, function (item) {
                        if (filterFn(item)) {
                            tmpItems.push(item);
                        }
                    });
                    items = tmpItems;
                }

                return items;
            },

            /**
             * Return item from store for a given type with that key
             *
             * To include items that have been deleted set the <i>includeDeleted</i> param to true
             *
             * If the item is not found, this method returns `null`.
             *
             * @method item
             * @param {String} type
             * @param {String} key
             * @param {Boolean} [includeDeleted] Also try looking for the item in the deleted items collection
             * @return object
             */
            item: function (type, key, includeDeleted) {
                var item = this.storeCollection.item(type, key);
                if (item == null && includeDeleted) {
                    item = this.deleteStoreCollection.item(type, key);
                }
                return item;
            },

            /**
             * Update item in store for a given type with that key.
             *
             * The store holds the original value so value can be reverted.
             *
             * Values can be set at any level for the item.
             *
             * If the value is set back to a value which passes the `_.isEqual` test, the field will be reset and will
             * no longer appear in the modified fields list.
             *
             * Example:
             *
             *      var item = {
             *          key : '1',
             *          bar : 'bar',
             *          foo : {
             *              foo : 'foo'
             *          }
             *      };
             *      store.add('type', item);
             *      console.log(store.isModifiedField('type', '1', 'foo.foo'));             // false
             *      store.update('type', '1', 'foo.foo', 'bar');
             *      console.log(store.isModifiedField('type', '1', 'foo.foo'));             // true
             *      console.log(store.isModifiedField('type', '1', 'bar'));                 // false
             *      store.update('type', '1', 'bar', 'foo');
             *      console.log(store.isModifiedField('type', '1', 'bar'));                 // true
             *      store.update('type', '1', 'bar', 'bar');
             *      console.log(store.isModifiedField('type', '1', 'bar'));                 // false
             *
             * <b>Note</b> For the store to work correctly all values must be updated through the store using this function
             *
             * Triggers Event - 'update'
             *
             * @method update
             * @param {String} type
             * @param {String} key
             * @param {String} field A period-delimited path to the field to update
             * @param value
             */
            update: function (type, key, field, value) {

                if (_.contains(this.options.readOnly, type)) {
                    throw new Error('store.update() - You can not update a read only type - ' + type);
                }

                var item = this.storeCollection.item(type, key);
                if (item != null) {

                    // is the new value different from present value
                    var currentValue = this.utilities._getValue(item, field);
                    if (_.isEqual(currentValue, value)) {
                        return;
                    }

                    // tell the modify store value has changed
                    if (!this._isItemNew(type, item)) {
                        if (this.modifyStoreCollection.hasOriginalValue(type, key, field)) {
                            // is the updated value the same as the original
                            if (this.modifyStoreCollection.isSameValue(type, key, field, value)) {
                                // remove value from modified collection
                                this.modifyStoreCollection.remove(type, key, field);
                            } else {
                                this.modifyStoreCollection._forceMarkModified(type, key, field);
                            }
                        } else {
                            this.modifyStoreCollection.add(type, key, field, item);
                        }
                    }

                    // change the value
                    if (!this.hasField(type, key, field)) {
                        var parts = field.split(PATH_DELIMITER);
                        var itemPart = item;
                        for (var i = 0; i < parts.length; i++) {
                            var token = parts[i];
                            if (!_.has(itemPart, token)) {
                                // create
                                if (i === parts.length - 1) {
                                    itemPart[token] = undefined;
                                } else {
                                    itemPart[token] = {};
                                }
                            }
                            itemPart = itemPart[token];
                        }
                    }

                    this.utilities._setValue(item, field, value);

                    updateAssociatedObjects.call(this, type, item);

                    /**
                     * Triggered when an item is updated in the store
                     *
                     * Args:
                     *  * type: The type of the updated item
                     *  * key: The key of the item updated
                     *  * path: The path to the modified field
                     *
                     * @event update
                     * @param args {Object}
                     */
                    this._subscribable.trigger('update', {
                        type: type,
                        key: key,
                        path: field
                    });
                }
            },

            /**
             * Remove item from store for a given type with that key
             *
             * If you attempt to remove an item from the store that does not exist, no error is thrown.
             *
             * Triggers Event - 'remove'
             *
             * @method remove
             * @param {String} type
             * @param {String} key
             */
            remove: function (type, key, silent) {

                var item = this.item(type, key);
                if (item != null) {

                    // if the item is not new then add to deleted store
                    if (!this._isItemNew(type, item)) {
                        if (!_.contains(this.options.readOnly, type)) {
                            this.reset(type, key, true);
                            this.deleteStoreCollection.add(type, item);
                        }
                    }

                    // remove item from new store
                    this.newStoreCollection.remove(type, key);

                    // remove item from modify store
                    this.modifyStoreCollection.remove(type, key, null, true);

                    this.storeCollection.remove(type, key);

                    updateAssociatedObjects.call(this, type, item);

                    if (!silent) {
                        /**
                         * Triggered when an item is removed from the store
                         *
                         * Args:
                         *  * type: The type of the removed item
                         *  * key: The key of the item removed
                         *
                         * @event remove
                         * @param args {Object}
                         */
                        this._subscribable.trigger('remove', {
                            type: type,
                            key: key
                        });
                    }
                }
            },

            /**
             * Does the field exist for a given type with that key.
             *
             * If a field has been set to undefined explicitly this method will still return false.
             *
             * Example:
             *
             *      var item = {
             *          key : '1',
             *          foo : {
             *              foo : 'foo',
             *              bar : undefined
             *          }
             *      };
             *      store.add('type', item);
             *      console.log(store.hasField('type', '1', 'key'));                    // true
             *      console.log(store.hasField('type', '1', 'bar'));                    // false
             *      console.log(store.hasField('type', '1', 'foo.foo'));                // true
             *      console.log(store.hasField('type', '1', 'foo.bar'));                // false
             *
             * @method hasField
             * @param {String} type
             * @param {String} key
             * @param {String} field
             * @returns Boolean
             */
            hasField: function (type, key, field) {
                var item = this.item(type, key);
                if (item != null) {
                    var value = this.utilities._getValue(item, field);
                    return !_.isUndefined(value);
                }
                return false;
            },

            /**
             * Determines whether or not a particular type or type-key pair has been updated in the store.
             * (Adds and removals do not count as updates for this method; for that, see `isDirty` and `isDirtyAll`)
             *
             * If no arguments are provided this behaves exactly like isModifiedAll.
             * If one argument is provided, it checks to see if a particular type has any changes.
             * With both arguments it checks to see if a particular key in a type has been updated.
             *
             * @method isModified
             * @param {String} [type]
             * @param {String} [key]
             * @return Boolean
             */
            isModified: function (type, key) {
                if (!arguments.length) {
                    return this.isModifiedAll();
                }
                else if (arguments.length === 1) {
                    return !(this.modifyStoreCollection.isEmpty(type));
                }
                else {
                    return !_.isEmpty(this.modifyStoreCollection.item(type, key));
                }
            },

            /**
             * Has the store had any items updated for all types.
             * (Adds and removals do not count as updates for this method; for that, see `isDirty` and `isDirtyAll`)
             *
             * @method isModifiedAll
             * @return Boolean
             */
            isModifiedAll: function () {
                return !this.modifyStoreCollection.isEmptyAll();
            },

            /**
             * Determines whether or not a particular type or type-key pair has been added in the store.
             *
             * If no arguments are provided this behaves exactly like isNewAll.
             * If one argument is provided, it checks to see if a particular type has any changes.
             * With both arguments it checks to see if a particular key in a type has been added.
             *
             * @method isNew
             * @param {String} [type]
             * @param {String} [key]
             * @return Boolean
             */
            isNew: function (type, key) {
                if (!arguments.length) {
                    return this.isNewAll();
                }
                else if (arguments.length === 1) {
                    return !(this.newStoreCollection.isEmpty(type));
                }
                else {
                    return !_.isEmpty(this.newStoreCollection.item(type, key));
                }
            },

            /**
             * Has the store had any items added for all types.
             *
             * @method isNewAll
             * @return Boolean
             */
            isNewAll: function () {
                return !this.newStoreCollection.isEmptyAll();
            },

            /**
             * Determines whether or not a particular type or type-key pair has been deleted in the store.
             *
             * If no arguments are provided this behaves exactly like isDeletedAll.
             * If one argument is provided, it checks to see if a particular type has any changes.
             * With both arguments it checks to see if a particular key in a type has been deleted.
             *
             * @method isDeleted
             * @param {String} [type]
             * @param {String} [key]
             * @return Boolean
             */
            isDeleted: function (type, key) {
                if (!arguments.length) {
                    return this.isModifiedAll();
                }
                else if (arguments.length === 1) {
                    return !(this.deleteStoreCollection.isEmpty(type));
                }
                else {
                    return !_.isEmpty(this.deleteStoreCollection.item(type, key));
                }
            },

            /**
             * Has the store had any items deleted for all types.
             *
             * @method isDeletedAll
             * @return Boolean
             */
            isDeletedAll: function () {
                return !this.deleteStoreCollection.isEmptyAll();
            },

            /**
             * Determines whether or not a particular type or type-key pair has been modified in the store; added, updated, and removed items all count as dirty.
             *
             * If no arguments are provided this behaves exactly like isDirtyAll.
             * If one argument is provided, it checks to see if a particular type has any changes.
             * With both arguments it checks to see if a particular key in a type has been modified.
             *
             * @method isDirty
             * @param {String} [type]
             * @param {String} [key]
             * @return Boolean
             */
            isDirty: function (type, key) {
                if (arguments.length === 0) {
                    return this.isDirtyAll();
                } else if (arguments.length === 1) {
                    return !(this.newStoreCollection.isEmpty(type) && this.modifyStoreCollection.isEmpty(type) && this.deleteStoreCollection.isEmpty(type));
                } else {
                    return this.isModified(type, key) || this.isNew(type, key) || this.isDeleted(type, key);
                }
            },

            /**
             * Has the store had any items added, deleted or modified for all types
             *
             * @method isDirtyAll
             * @return Boolean
             */
            isDirtyAll: function () {
                return !(this.newStoreCollection.isEmptyAll() && this.modifyStoreCollection.isEmptyAll() && this.deleteStoreCollection.isEmptyAll());
            },

            /**
             * Has the field been modified for a given type with that key
             *
             * Fields which have been updated back to their original value (matching the `_.isEqual` test) will no longer
             * appear in this list.
             *
             * @method isModifiedField
             * @param {String} type
             * @param {String} key
             * @param {String} field
             * @return Boolean
             */
            isModifiedField: function (type, key, field) {
                return this.modifyStoreCollection.hasOriginalValue(type, key, field);
            },

            /**
             * Return a list of all modified fields for a given type with that key
             *
             * Example:
             *
             *      var item = {
             *          key : '1',
             *          bar : 'bar',
             *          foo : {
             *              foo : 'foo'
             *          }
             *      };
             *      store.add('type', item);
             *      store.update('type', '1', 'foo.foo', 'bar');
             *      store.update('type', '1', 'bar', 'foo');
             *      console.log(store.modifiedFields('type', '1'));                 // ['foo.foo', 'bar']
             *
             * @method modifiedFields
             * @param {String} type
             * @param {String} key
             * @return Array
             */
            modifiedFields: function (type, key) {
                return this.modifyStoreCollection.modifiedFields(type, key);
            },

            /**
             * Returns the original value for a particular path on an object before any updates.
             *
             * @param type {String}
             * @param key {String}
             * @param field {String}
             * @returns {*}
             */
            getOriginalValue: function (type, key, field) {
                if (this.modifyStoreCollection.hasOriginalValue(type, key, field)) {
                    return this.modifyStoreCollection.getOriginalValue(type, key, field);
                } else {
                    return this.utilities._getValue(this.storeCollection.item(type, key), field);
                }
            },

            /**
             * Reset item for a given type with that key or the entire type
             *
             * Reverts any modified values for item
             *
             * Example:
             *
             *      var item = {
             *          key : '1',
             *          bar : 'bar'
             *      };
             *      store.add('type', item);
             *      store.update('type', '1', 'bar', 'foo');
             *      console.log(store.item('type', '1').bar);                 // 'foo'
             *      store.reset('type', '1');
             *      console.log(store.item('type', '1').bar);                 // 'bar'
             *
             * Example whole type:
             *
             *      var items = [
             *          { key : '1', bar : 'bar' }, { key : '2', bar : 'another' }
             *      ];
             *      store.addAll('type', items);
             *      store.update('type', '1', 'bar', 'foo');
             *      store.update('type', '2', 'bar', 'more');
             *      console.log(store.item('type', '1').bar);                 // 'foo'
             *      console.log(store.item('type', '2').bar);                 // 'more'
             *      store.reset('type');
             *      console.log(store.item('type', '1').bar);                 // 'bar'
             *      console.log(store.item('type', '2').bar);                 // 'another'
             *
             * Triggers Event - `reset`
             *
             * @method reset
             * @param {String} type
             * @param {String} [key]
             */
            reset: function (type, key, silent) {
                var that = this;
                if (key != null) {
                    // Reset one particular item
                    var modifyItem = that.modifyStoreCollection.item(type, key),
                        storeItem = that.storeCollection.item(type, key);

                    if (modifyItem != null) {
                        _.each(that.modifyStoreCollection._allModifiedFields(type, key), function (field) {
                            if (field.indexOf(PATH_DELIMITER) < 0) {
                                that.utilities._setValue(storeItem, field, modifyItem.oldValues[field]);
                                that.modifyStoreCollection.remove(type, key, field);
                            }
                        });

                        /**
                         * Triggered when an item is reset back to its original value
                         *
                         * Args:
                         *  * type: The type of the reset item
                         *  * key: The key of the reset item
                         *
                         * @event reset
                         * @param args {Object}
                         */
                        if (!silent) {
                            that._subscribable.trigger('reset', {
                                type: type,
                                key: key
                            });
                        }
                    } else {
                        // The item was added or deleted
                        var deleteItem = that.deleteStoreCollection.item(type, key),
                            addItem = that.newStoreCollection.item(type, key);

                        if (deleteItem != null) {
                            // New items do not end up in the deleteStoreCollection so we don't need to check for new here.
                            that.storeCollection.add(type, deleteItem);
                            that.deleteStoreCollection.remove(type, key);
                        } else if (addItem != null) {
                            that.remove(type, key, true);
                        }
                    }
                } else {
                    // Reset whole type
                    // TODO Why do some of these use a while loop and others _.each?

                    if (this.deleteStoreCollection.map[type]) {
                        _.each(this.deleteStoreCollection.map[type].items(), function (item) {
                            // New items do not end up in the deleteStoreCollection so we don't need to check for new here.
                            that.storeCollection.add(type, item);
                        });
                    }

                    if (this.modifyStoreCollection.map[type]) {
                        _.each(this.modifyStoreCollection.map[type].items(), function (modifyItem) {
                            var key =  getKey.call(that, type),
                                storeItem = that.storeCollection.item(type, modifyItem[key]);

                            _.each(that.modifyStoreCollection._allModifiedFields(type, modifyItem[key]), function (field) {
                                if (field.indexOf(PATH_DELIMITER) < 0) {
                                    that.utilities._setValue(storeItem, field, modifyItem.oldValues[field]);
                                    that.modifyStoreCollection.remove(type, key, field);
                                }
                            });
                        });
                    }

                    if (this.newStoreCollection.map[type]) {
                        var orderedMap = this.newStoreCollection.map[type],
                            typeKey = getKey.call(that, type);
                        while (orderedMap.items().length) {
                            var newItem = orderedMap.items()[0];
                            // Want a silent here to prevent triggering 'remove' event
                            that.remove(type, newItem[typeKey], true);
                        }
                    }

                    this.modifyStoreCollection.clear(type);
                    this.deleteStoreCollection.clear(type);
                    this.newStoreCollection.clear(type);

                    if (!silent) {
                        that._subscribable.trigger('reset', {
                            type: type
                        });
                    }
                }
            },

            /**
             * Reset all items in store.
             *
             * Removes all new items.
             * Reverts all deleted items.
             * Reverts any modified values for all items.
             *
             * Example:
             *
             *      var newItem = {
             *          key : '-1'
             *      };
             *      var deletedItem = {
             *          key : '2'
             *      };
             *      var modifyItem = {
             *          key : '3'
             *      };
             *      store.add('type', newItem);
             *      store.add('type', deletedItem);
             *      store.add('type', modifyItem);
             *      store.remove('type, '2');
             *      store.update('type, '3', 'bar', 'bar');
             *      console.log(store.isDirtyAll());                                // true
             *      store.resetAll();
             *      console.log(store.isDirtyAll());                                // false
             *
             * Triggers Event - 'reset-all'
             *
             * @method resetAll
             */
            resetAll: function () {
                var that = this,
                    types = _.union(_.keys(this.storeCollection.map), _.keys(this.modifyStoreCollection.map), _.keys(this.newStoreCollection.map), _.keys(this.deleteStoreCollection.map));

                _.each(types, function (type) {
                    that.reset(type, undefined, true);
                });

                /**
                 * Triggered when the entire store is reset with `store.resetAll()`.
                 * Doesn't provide any parameters.
                 *
                 * @event reset-all
                 */
                this._subscribable.trigger('reset-all');
            },

            /**
             * Clear items from store for a given type.
             *
             * Pass in an array of keys to only clear items that meet those keys.
             *
             * Trigger Event - 'clear'
             *
             * @method clear
             * @param {String} type
             * @param {Array} [keys]
             */
            clear: function (type, keys) {
                this.storeCollection.clear(type, keys);
                this.modifyStoreCollection.clear(type, keys, true);
                this.deleteStoreCollection.clear(type, keys);
                this.newStoreCollection.clear(type, keys);

                /**
                 * Triggered when an individual store type is cleared.
                 *
                 * Args:
                 *  * type: The type of the cleared store
                 *
                 * @event clear
                 * @param {String} type
                 */
                this._subscribable.trigger('clear', {
                    type: type,
                    keys: keys
                });
            },

            /**
             * Clear all items from store.
             *
             * Trigger Event - `clear-all`
             *
             * @method clearAll
             */
            clearAll: function () {
                this.storeCollection.clearAll();
                this.modifyStoreCollection.clearAll();
                this.deleteStoreCollection.clearAll();
                this.newStoreCollection.clearAll();

                /**
                 * Triggered when the entire store is cleared with `store.clearAll()`.
                 * Doesn't provide any parameters.
                 *
                 * @event clear-all
                 */
                this._subscribable.trigger('clear-all');
            },

            /**
             * Return a save object for all item types.
             *
             * Example:
             *
             *      {
             *          foo : {
             *              added: [],      // array of objects
             *              modified: [],   // array of objects
             *              deleted: []     // array of keys
             *          },
             *          bar : {
             *              added: [],      // array of objects
             *              modified: [],   // array of objects
             *              deleted: []     // array of keys
             *          }
             *      }
             *
             * @param {Function} [serializeFn]
             * @method saveAll
             * @return Object
             */
            saveAll: function (serializeFn) {
                var that = this;
                var saveData = {};

                // remove any associated type
                var associatedTypes = _.keys(that.options.associatedTypes);
                var types = _.filter(that.storeCollection.types(), function (type) {
                    return !_.contains(associatedTypes, type);
                });
                _.each(types, function (type) {
                    var saveObj = that.save(type, serializeFn);
                    if (saveObj != null && (saveObj[type].added.length || saveObj[type].deleted.length || saveObj[type].modified.length)) {
                        saveData[type] = saveObj[type];
                    }
                });

                return saveData;
            },

            /**
             * Return save object for a given type.
             *
             * The type still appears as a top level key on the returned object.
             * See example.
             *
             * Example:
             *
             *      {
             *          foo : {
             *              added: [],      // array of objects
             *              modified: [],   // array of objects
             *              deleted: []     // array of keys
             *          }
             *      }
             *
             * @method save
             * @param {String} type
             * @param {Function} [serializeFn] (type, item) -> item: Additional serialization to perform on the item
             * @return Object
             */
            save: function (type, serializeFn) {
                var that = this,
                    key = getKey.call(this, type);
                if (this.storeCollection.hasType(type)) {

                    var saveData = {};
                    saveData[type] = {
                        added: [],
                        deleted: [],
                        modified: []
                    };

                    if (!_.contains(that.options.readOnly, type)) {

                        var deserializeAssociatedType = function (item) {
                            _.each(_.keys(that.options.associatedTypes), function (assocType) {
                                if (_.has(item, ASSOCIATED_DELIM + assocType)) {
                                    item[assocType] = item[ASSOCIATED_DELIM + assocType];
                                    delete item[ASSOCIATED_DELIM + assocType];
                                    if (item[assocType] == null) {
                                        delete item[assocType];
                                    }
                                }
                            });
                            return item;
                        };

                        // new items - whole object
                        _.each(this.newStoreCollection.items(type), function (item) {
                            var cloneItem = deserializeAssociatedType($.extend(true, {}, item));
                            that._serialize(type, cloneItem, serializeFn);
                            saveData[type].added.push(cloneItem);
                        });

                        // removed items - just the id
                        _.each(this.deleteStoreCollection.items(type), function (item) {
                            saveData[type].deleted.push(item[key]);
                        });

                        // modified items
                        _.each(this.modifyStoreCollection.items(type), function (item, id) {
                            var cloneItem = deserializeAssociatedType($.extend(true, {}, that.storeCollection.item(type, item[key])));
                            that._serialize(type, cloneItem, serializeFn);
                            saveData[type].modified.push(cloneItem);
                        });

                    }
                    return saveData;
                }

                return null;
            },

            /**
             * Commit all changes in store.
             *
             * New items are persisted with a real key.
             *
             * To persist new items pass in the following format:
             *
             *      // old and new key map for each type
             *      var newKeys = {
             *          foo : {
             *              '-1' : '17261',
             *              '-2' : '17262'
             *          },
             *          bar : {
             *              '-5' : '2323'
             *          }
             *      };
             *      store.commitAll(newKeys);
             *
             * Deleted items are permanently removed.
             *
             * Modified values on an item are permanent.
             *
             * Triggers the event `commit-all`.
             * Triggers the event `commit` for each committed storeType.
             *
             * @method commitAll
             * @param {Object} newKeys
             */
            commitAll: function (newKeys) {
                var that = this;
                _.each(newKeys, function (data, type) {
                    that.commit(type, data);
                });
                if (!newKeys) {
                    logger.warn('Store.commitAll() should be called with an object describing the storeTypes and keys to be committed.  If not provided, the `commit` event will not get called for each store type that gets cleared.');
                }
                that.modifyStoreCollection.clearAll();
                that.deleteStoreCollection.clearAll();
                that.newStoreCollection.clearAll();

                /**
                 * Triggered when `store.commitAll` is called.
                 * Doesn't provide any parameters.
                 *
                 * @event commit-all
                 * @since 2.0.9
                 */
                that._subscribable.trigger('commit-all');
            },

            /**
             * Commit items to store for a given type.
             *
             * New items are persisted with a real key.
             *
             * To persist new items pass in the following format:
             *
             *      // old and new key map
             *      var newKeys = {
             *          '-1' : '17261',
             *          '-2' : '17262'
             *      };
             *      store.commit('type', newKeys);
             *
             * Deleted items are permanently removed
             *
             * Modified values on an item are permanent
             *
             * Trigger Event - 'commit'
             *
             * @method commit
             * @param {String} type
             * @param {Object} newKeys
             */
            commit: function (type, newKeys) {
                var that = this,
                    key = getKey.call(this, type);
                // update the new item ids
                _.each(newKeys, function (newId, oldId) {
                    var item = that.storeCollection.item(type, oldId);

                    if (item != null) {
                        // find associated objects by oldId so we can update
                        if (_.has(that.options.associatedChildTypes, type)) {
                            _.each(that.options.associatedChildTypes[type], function (childType) {
                                var childItems = that.items(childType, function (item) {
                                    return item[that.options.associatedTypes[childType].parentId] === oldId;
                                });
                                _.each(childItems, function (item) {
                                    item[that.options.associatedTypes[childType].parentId] = newId;
                                });
                            });
                        }

                        item[key] = newId;
                        that.storeCollection._repopulateIndexes(type);
                    }
                });

                if (_.has(that.options.associatedChildTypes, type)) {
                    // flip any new ids for any associated children
                    _.each(that.options.associatedChildTypes[type], function (childType) {
                        var childKey = getKey.call(that, childType);
                        _.each(that.items(childType), function (item) {
                            if (that._isItemNew(childType, item)) {
                                item[childKey] = _.uniqueId(childType);
                            }
                        });
                        that.storeCollection._repopulateIndexes(childType);
                    });
                }

                // clear new
                that.newStoreCollection.clear(type);

                // clear modified
                that.modifyStoreCollection.clear(type);

                // clear deleted
                that.deleteStoreCollection.clear(type);

                that._subscribable.trigger('commit', { type: type });

            },

            /**
             * Attach an event handler
             *
             * Example:
             *
             *      store.on('add', function(args) {
             *          // fired when an item is added to store
             *      });
             *
             * @method on
             * @param {String} eventName
             * @param {Function} callbackFn
             * @param {Object} [context]
             */
            on: function (eventName, callbackFn, context) {
                this._subscribable.on(eventName, callbackFn, context);
            },

            /**
             * Remove event handler
             *
             * Example:
             *
             *      store.remove('add');
             *
             * @method off
             * @param {String} eventName
             */
            off: function (eventName) {
                this._subscribable.off(eventName);
            },

            /**
             * Returns whether or not the store has a collection for the given type.
             * This is different that checking `store.count() > 0` in that even if the store doesn't have a collection for the given type, `.count()` would return 0.
             *
             * @method hasType
             * @param {String} type
             * @returns true if the store has a collection for the given type, false otherwise
             */
            hasType: function (type) {
                return this.storeCollection.hasType(type);
            },

            /**
             * Sets the known date fields that will automatically be deserialized when an item is added to the store.
             * You can set the global date fields or the date fields per type.
             *
             * Global : `store.setDateFields([...]);`
             * Per Type : `store.setDateFields('type', [...]);`
             *
             * NOTE: You should only call this method when the store (or the specific type) is empty!
             * Calling setDateFields will not modify the existing items in the store. E.g if you've added and new date field
             * and removed another, the old field will not be serialized back to a string and the new field will not be turned
             * into a moment object/
             *
             * NOTE: We generally do not all for updating options on the store because there are often weird side effects.
             * We make an exception for date fields because of how the store generally interacts with the meta data from the server.
             * Generally the store is created before your initial rest call, so we don't known the date fields before that server call.
             * This method allows them to be set after that meta call.
             *
             * @method setDateFields
             * @param [type] {String}
             * @param allFields {Array}
             */
            setDateFields: function (type, allFields) {
                if (arguments.length < 2) {
                    allFields = type;
                    type = null;
                }

                if (type == null) {
                    this.options.dateFields = allFields || [];
                    this._dateFieldsCache = {};
                } else {
                    this.options.typeOptions[type] = this.options.typeOptions[type] || {};
                    this.options.typeOptions[type].dateFields = allFields || [];
                    delete this._dateFieldsCache[type];
                }
            }
        };

        Store.mergeModes = MERGE_MODES;

        return Store;

    });


define('primutils/markup/primutils/ui/changeBar/changeBar.hbs:#modal-backdrop', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/changeBar.hbs:#modal-backdrop'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  return "<div class=\"modal-backdrop fade\"></div>";
  },"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/utils/grid',['primutils.jquery', 'primutils.lodash'], function ($, _) {
    

    var GRID_PSEUDO_SELECTOR = 'pgbu-grid';

    var that = {
        _pseudoSelectorDefined: function (pseudo) {
            return _.has($.expr[':'], pseudo);
        },
        gridApiLoaded: function () {
            return that._pseudoSelectorDefined(GRID_PSEUDO_SELECTOR);
        },
        selectGrids: function () {
            if (!that.gridApiLoaded()) {
                return $([]);
            }
            return $(':' + GRID_PSEUDO_SELECTOR);
        }
    };
    return that;

});

define('primutils/ui/changeBar/ChangeBarView',['primutils.jquery', 'primutils.lodash',
    'primutils/utils/language', 'primutils.logger',
    'primutils/markup/primutils/ui/changeBar/changeBar.hbs:#modal-backdrop',
    'primutils/utils/keyCode', 'primutils/utils/grid',
    'primutils/utils/emulateTransitionEnd'
], function ($, _, lang, logger, modalBackdrop, keyCode, gridUtils) {
    

    // why?  https://jira.oraclecorp.com/jira/browse/PRIMCR-404
    //  debouncing save after click aloows grid validation, etc, to take place
    //  before trying to process the save
    var SAVE_DEBOUNCE_TIME = 25,
        BACKDROP_FADE_TIME = 100, // See modals.less (.1s in Blue Sky 3)
        SAVED_STATE_VISIBLE_TIMEOUT = 2000;

    var SAVE_STATES = {
        UNSAVED : 'unsaved',
        SAVING : 'saving',
        SAVED  : 'saved'
    };

    var getValue = function (opt) {
        return _.isFunction(opt) ? opt.apply(this, _.rest(_.toArray(arguments))) : opt;
    };

    var removeListeners = function () {
        this.$errorList.off('click', 'li').off('scroll', 'li').off('keydown', 'li');
        this.buttons.$saveButton.off('click');
        this.buttons.$cancelButton.off('click');
        this.buttons.$errorButton.off('click');
    };

    var initListeners = function () {
        var that = this,
            controller = this.controller,
            refocus = null,
            debouncedSave;

        this.$errorList.on('click', 'li', function (e) {
            var error = $(this).data('error');
            controller._selectError(error);
        }).on('scroll', function () {
            scrollErrorsList.call(that);
            if (refocus === 'first') {
                that.$errorList.find('li:first-child a').focus();
            } else if (refocus === 'last') {
                that.$errorList.find('li').last().find('a').focus();
            }
            refocus = null;
        }).on('keydown', 'li', function (e) {
            var $this = $(this),
                preventDefault = true;

            if (e.keyCode === keyCode.UP) {
                if ($this.is(':first-child')) {
                    refocus = 'first';
                    that.$errorList.scrollTop(that.$errorList.scrollTop() - that.virtualHeights.errorItem);
                } else {
                    $this.prev().children('a').focus();
                }
            } else if (e.keyCode === keyCode.DOWN) {
                if ($this.next().is(':last-child')) { // .virtual-scroller is last child
                    refocus = 'last';
                    that.$errorList.scrollTop(that.$errorList.scrollTop() + that.virtualHeights.errorItem);
                } else {
                    $this.next().children('a').focus();
                }
            } else if (e.keyCode === keyCode.PAGEUP) {
                refocus = 'first';
                that.$errorList.scrollTop(that.$errorList.scrollTop() - that.virtualHeights.errorList);
            } else if (e.keyCode === keyCode.PAGEDOWN) {
                refocus = 'last';
                that.$errorList.scrollTop(that.$errorList.scrollTop() + that.virtualHeights.errorList);
            } else if (e.keyCode === keyCode.HOME) {
                refocus = 'first';
                that.$errorList.scrollTop(0);
            } else if (e.keyCode === keyCode.END) {
                refocus = 'last';
                that.$errorList.scrollTop(that.$errorList.get(0).scrollHeight);
            } else if (e.keyCode === keyCode.TAB && !e.shiftKey && $this.next().is(':last-child') && !$this.is(':first-child')) { // .virtual-scroller is last child
                preventDefault = false;
                that.$errorList.closest('.open').removeClass('open');
            } else {
                return;
            }

            if (preventDefault) {
                e.preventDefault();
                e.stopPropagation();
            }
        });

        this.buttons.$errorButton.on('click', function (e) {
            if (!that._repopulatedOnce) {
                repopulateErrorsList.call(that);
            } else {
                setupVirtualization.call(that);
            }
            _.defer(function () {
                that.$errorList.scrollTop(that._resetScrollPosition ? 0 : that._lastErrorScrollPosition);
                that._resetScrollPosition = false;
            });
        });

        function baseHandleSave() {
            if (!getSaveEnabled.call(that)) {
                return;
            }
            controller._save();
        }
        debouncedSave = _.debounce(baseHandleSave, SAVE_DEBOUNCE_TIME);
        this.buttons.$saveButton.on('click', function (e) {
            // over-opinionated?  seems reasonable but...
            if (!that.buttons.$saveButton.is(':focus')) {
                that.buttons.$saveButton.focus();
            }

            // want to do this before the debounce, so grid/page have a chance to work
            if (controller._options.closeGridEditorsOnSave && gridUtils.gridApiLoaded()) {
                var $grids = gridUtils.selectGrids();
                if ($grids && $grids.length) {
                    $grids.grid('closeEditor', true);
                }
            }

            debouncedSave();
        });

        this.buttons.$cancelButton.on('click', function (e) {
            controller._cancel();
        });
    };

    var initElement = function () {
                logger.assert(this.controller._options.changeBarMarkup || this.controller._options.fromExisting, 'ChangeBar must be instantiated with either DOM provided or with the fromExisting option');
                var $el = false;
        if (this.controller._options.fromExisting) {
            var $fromExisting = $(getValue.call(this, this.controller._options.fromExisting));
            if ($fromExisting.is('.save-cancel-controls')) {
                $el = $fromExisting.filter('.save-cancel-controls');
                logger.assert($el.length === 1, 'when using ChangeBar fromExisting, change-bar-controls to attach to must be unambiguous, but more than one was found');
            }
            else if ($fromExisting.has('.save-cancel-controls')) {
                $el = $fromExisting.find('.save-cancel-controls');
                logger.assert($el.length === 1, 'when using ChangeBar fromExisting, change-bar-controls to attach to must be unambiguous, but more than one was found');
            }
            else {
                var message = 'The `fromExisting` option was specified, but the provided selector doesn\'t appear to contain a save cancel bar created with the change-bar-controls helper.';
                logger.fatal(message);
                throw new Error(message);
            }
        }
        else {
            $el = $(this.controller._options.changeBarMarkup);
        }
        this.$el = $el;
        this.$el.toggleClass('fade', this.controller._options.centerChangeBar);

        this._savingButtonClass = _.compact(_.without((this.$el.find('.save-button').attr('class') || '').split(' '), 'btn', 'save-button')).join(' ');

        this.$modalBackdrop = $(modalBackdrop());

        this.clearStatusMessage();
    };

    var setupVirtualization = function () {
        var $dropdownContainer, isOpen, $errorItem;

        if (!this.virtualHeights) {
            this.virtualHeights = {
                canVirtualize: this.$errorList.css('max-height') !== 'none',
                hasValidErrorItemHeight : false,
                errorList: parseInt(this.$errorList.css('max-height'), 10)
            };
            this.virtualHeights.errorItem = this.virtualHeights.errorList;
        }

        if (this.$el.css('display') !== 'none' && !this.virtualHeights.hasValidErrorItemHeight && this.virtualHeights.canVirtualize) {
            $dropdownContainer = this.buttons.$errorButton.is('[data-toggle=dropdown]') ? this.buttons.$errorButton.parent() : $();
            isOpen = $dropdownContainer.hasClass('open');
            $errorItem = makeErrorItem.call(this, { message: 'sizing' }).appendTo(this.$errorList);

            $dropdownContainer.addClass('open');

            this.virtualHeights.errorItem = parseInt($errorItem.outerHeight(true), 10);
            this.virtualHeights.hasValidErrorItemHeight = true;

            $errorItem.remove();
            $dropdownContainer.toggleClass('open', isOpen);
        }
    };

    var repopulateErrorsList = function () {
        var that = this,
            errors = this.controller.getVisibleErrors();

        this._repopulatedOnce = true;

        setupVirtualization.call(this);

        if (this.virtualHeights.canVirtualize) {
            that._resetScrollPosition = true;
            scrollErrorsList.call(that);
        } else {
            this.$errorList.empty();
            _.each(errors, function (error) {
                that.$errorList.append(makeErrorItem.call(that, error));
            });
        }
    };

    // Makes save and cancel buttons the same width so switching the save state doesn't affect the save button's size
    // Also do this to the cancel button to make the change bar appear "even"
    var fixSaveButtonWidth = function () {
        if (this._fixedSaveButtonWidth) { return; }

        var barVisible = this.isVisible();

        toggle.call(this, true, true);

        this.buttons.$saveButton.children('.save-button-state').removeClass('hidden');

        var $buttons = this.buttons.$saveButton.add(this.buttons.$cancelButton);

        var maxWidth = _.max(_.map($buttons, function (btn) {
            return $(btn).outerWidth(true);
        }));
        $buttons.css('width', maxWidth);

        toggle.call(this, barVisible, true);

        this._fixedSaveButtonWidth = true;
    };

    var scrollErrorsList = function () {
        if (!this.virtualHeights || !this.virtualHeights.canVirtualize) {
            return;
        }

        if (this.$errorList.find('.virtual-scroller').length === 0) {
            this.$errorList.append('<div class="virtual-scroller"></div>');
        }

        var scrollTop = this.$errorList.scrollTop(),
            errors = this.controller.getVisibleErrors(),
            maxVisibleErrors = Math.ceil(this.virtualHeights.errorList / this.virtualHeights.errorItem),
            countOffset = Math.max(Math.min(Math.floor(scrollTop / this.virtualHeights.errorItem), errors.length - maxVisibleErrors), 0),
            maxCount = Math.min(maxVisibleErrors, errors.length),
            $virtualScroller = this.$errorList.find('.virtual-scroller');

        $virtualScroller.css('height', ((errors.length - countOffset - maxVisibleErrors) * this.virtualHeights.errorItem) + 'px');
        this.$errorList.children(':not(.virtual-scroller)').remove();
        for(var i = countOffset; i <= (countOffset + maxCount - 1); i++) {
            if (errors[i] != null) {
                makeErrorItem.call(this, errors[i]).css('margin-top', i === countOffset ? scrollTop : 0).insertBefore($virtualScroller);
            }
        }

        this._lastErrorScrollPosition = scrollTop;
    };

    function ChangeBarView (controller) {
        logger.assert(this && this instanceof ChangeBarView, 'ChangeBarVew must be created with new');
        this.controller = controller;

        this._savingState = SAVE_STATES.UNSAVED;

        initElement.call(this);
        toggle.call(this, false);

        this.errorItemMarkup = this.controller._options.errorItemMarkup;

        this.buttons = {
            $saveButton : this.$el.find('.save-button'),
            $cancelButton : this.$el.find('.cancel-button'),
            $errorButton : this.$el.find('.error-button')
        };

        this.buttons.saveButtonStates = {
            unsaved : this.buttons.$saveButton.find('.unsaved'),
            saving : this.buttons.$saveButton.find('.saving'),
            saved : this.buttons.$saveButton.find('.saved')
        };

        this.$errorList = this.$el.find('.error-list');
        this.$errorListParent = this.$el.find('.errors-list');

        initListeners.call(this);
    }

    var centerChangeBar = function () {
        if (!this.controller._options.centerChangeBar) { return; }
        this.$el.css('margin-left', this.$el.outerWidth(false) / -2);
    };

    var toggle = function (visible, immediate) {
        var that = this;

        function doToggle() {
            var isFade = that.$el.hasClass('fade');

            if (visible) {
                that.$el.css('display', '');
                /* jshint -W030 */ //Suppress the fact that we aren't assigning this to anything
                that.$el[0].offsetWidth; // force reflow
                /* jshint +W030 */
            } else if ($.support.transition && isFade && !immediate) {
                that.$el.one($.support.transition.end, function () {
                    that.$el.css('display', that.isVisible() ? '' : 'none');
                }).emulateTransitionEnd(200);
            }

            that.$el
                .toggleClass('in', visible)
                .attr('aria-hidden', !visible)
                .data('visible', visible);

            if (!visible && (!$.support.transition || !isFade || immediate)) {
                that.$el.css('display', 'none');
            }
        }

        if (this._hideTimeout) {
            clearTimeout(this._hideTimeout);
            this._hideTimeout = null;
        }

        if (visible || immediate) {
            doToggle();
        } else {
            this._hideTimeout = setTimeout(doToggle, SAVED_STATE_VISIBLE_TIMEOUT);
        }
    };

    var makeErrorItem = function (error) {
        var $item = $(this.errorItemMarkup({message: error.message}));
        $item.data('error', error);
        return $item;
    };

    var getSaveEnabled = function () {
        var barVisible = this.$el.data('visible'),
            validSavingState = this._savingState === SAVE_STATES.UNSAVED,
            saveEnabled = this.controller._checkAllowSave() && barVisible && validSavingState;

        return saveEnabled;
    };

    var updateSavingThreeState = function () {
        this.buttons.$saveButton
            .toggleClass('success', this._savingState === SAVE_STATES.SAVED)
            .toggleClass('btn-blue-dark', this._savingState === SAVE_STATES.SAVING)
            .toggleClass(this._savingButtonClass, this._savingState !== SAVE_STATES.SAVING);
        this.buttons.$saveButton.children('.save-button-state').addClass('hidden');
        this.buttons.saveButtonStates[this._savingState].removeClass('hidden');
    };

    var toggleModal = function (visible, immediate) {
        var that = this;

        // Make is a boolean
        immediate = !!immediate;

        if (visible) {
            this.$modalBackdrop.appendTo(document.body);
            /* jshint -W030 */ //Suppress the fact that we aren't assigning this to anything
            this.$modalBackdrop[0].offsetWidth; // force reflow
            /* jshint +W030 */
            that.$modalBackdrop.addClass('in');
        } else if (this.$modalBackdrop.parents('html').length > 0) {
            if ($.support.transition && !immediate) {
                this.$modalBackdrop.one($.support.transition.end, function () {
                    that.$modalBackdrop.remove();
                }).removeClass('in').emulateTransitionEnd(BACKDROP_FADE_TIME);
            } else {
                this.$modalBackdrop.removeClass('in').remove();
            }
        }
    };

    var setButtonStates = function () {
        var errors = this.controller.getVisibleErrors(),
            barVisible = this.$el.data('visible'),
            saveEnabled = getSaveEnabled.call(this),
            cancelEnabled = barVisible && this._savingState === SAVE_STATES.UNSAVED,
            errorEnabled = (errors.length !== 0) && this._savingState === SAVE_STATES.UNSAVED,
            errorVisible = (errors.length !== 0);

        this.buttons.$saveButton.prop('disabled', !saveEnabled);
        this.buttons.$cancelButton.prop('disabled', !cancelEnabled);
        this.buttons.$errorButton.prop('disabled', !errorEnabled).attr('aria-hidden', !errorEnabled);
        // Swap the button group visibility instead of the button itself otherwise the margin is still visible
        this.$errorListParent.css('display', errorVisible ? '' : 'none');
        this.$el.toggleClass('has-errors', errorVisible);

        updateSavingThreeState.call(this);
        centerChangeBar.call(this);
    };

    ChangeBarView.prototype = {
        show : function () {
            logger.assert(this._savingState === SAVE_STATES.UNSAVED, 'ChangeBar.show should only be called when change bar is unsaved');

            toggle.call(this, true);
            toggleModal.call(this, false);
            setButtonStates.call(this);
        },

        updateButtonStates : function () {
            setButtonStates.call(this);
        },

        hide : function () {
            logger.assert(this._savingState !== SAVE_STATES.SAVING, 'ChangeBar.hide should not be called while change bar is saving');

            this._savingState = SAVE_STATES.UNSAVED;
            toggle.call(this, false, true);
            toggleModal.call(this, false, true);
            setButtonStates.call(this);
        },

        saving : function () {
            logger.assert(this._savingState === SAVE_STATES.UNSAVED && this.isVisible(), 'ChangeBar.save should only be called when change bar is unsaved and visible');

            this._savingState = SAVE_STATES.SAVING;
            toggleModal.call(this, true);
            setButtonStates.call(this);
        },

        saveSucceeded : function () {
            logger.assert(this._savingState === SAVE_STATES.SAVING, 'ChangeBar.saveSucceeded should only be called when change bar is saving');

            this._savingState = SAVE_STATES.SAVED;
            toggle.call(this, false, false);
            toggleModal.call(this, false, false);
            setButtonStates.call(this);

            // Switch back to unsaved state after running all of the redraw
            this._savingState = SAVE_STATES.UNSAVED;
        },

        saveFailed : function () {
            logger.assert(this._savingState === SAVE_STATES.SAVING, 'ChangeBar.saveFailed should only be called when change bar is saving');

            this._savingState = SAVE_STATES.UNSAVED;
            toggleModal.call(this, false, false);
            setButtonStates.call(this);
        },

        getSavingState : function () {
            return this._savingState;
        },

        redraw : function () {
            var errors = this.controller.getVisibleErrors();

            if (this.$el.closest(document.documentElement).length > 0) {
                repopulateErrorsList.call(this);
                fixSaveButtonWidth.call(this);
            }

            this.buttons.$errorButton
                .text(lang.formatString(this.controller._options.locale.errors, errors.length))
                .attr('data-error-count', errors.length);
            this.$el.attr('data-error-count', errors.length);

            setButtonStates.call(this);
        },

        isVisible : function () {
            return !!this.$el.data('visible');
        },

        removeListeners: function () {
            removeListeners.call(this);
        },

        setStatusMessage: function (message) {
            this.$el.find('.status-message').text(message);
            centerChangeBar.call(this);
        },

        clearStatusMessage: function () {
            this.setStatusMessage(this.controller._options.locale.defaultStatusMessage);
        }
    };

    return ChangeBarView;
});


define('primutils/markup/primutils/ui/changeBar/changeBar.hbs:#change-bar', ['primutils.handlebars','primutils.markup','primutils/ui/changeBar/helpers'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/changeBar.hbs:#change-bar'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers['change-bar-status'].call(depth0, {"name":"change-bar-status","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-errors'].call(depth0, {"name":"change-bar-errors","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-cancel'].call(depth0, {"name":"change-bar-cancel","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-save'].call(depth0, {"name":"change-bar-save","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers['change-bar-controls'].call(depth0, {"name":"change-bar-controls","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/changeBar/changeBar.hbs:#error-item', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/icon','primutils/templates/helpers/dropdown'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/changeBar.hbs:#error-item'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<span class=\"message\" title=\"";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.dropdownItem.call(depth0, {"name":"dropdownItem","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * A standardized, flexible ChangeBar, suitable for use anywhere save and cancel are used, optimized for use at the page level in Prime.
 *
 * ChangeBar attempts to namespace functions into logical groupings.
 * For example, the `errors` functions (`errors.all`, `errors.removeMatching`, etc) all deal with errors.
 * The `events` functions deal with events, while `visibility` deal with visibility.
 *
 * There are three main topics covered by this class description:
 *
 * * [Creating a ChangeBar](#creation)
 * * [Dealing with errors](#errors)
 * * [Customizing the ChangeBar DOM](#custom_dom)
 *
 * <a id="creation"></a>
 * #### Creating a ChangeBar
 *
 * ChangeBar, by default, does not attach to the DOM.
 * This makes it very flexible, and indeed, can mean that more than one ChangeBar can exist peacefully on a page.
 *
 * We have three options for attaching to the DOM.
 * The first, setting the `attachTo` option and calling `attach`, can be used like the following:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         // ....
 *         var myChangeBar = new ChangeBar({
 *             attachTo: '.content .banner' // this is the default
 *         });
 *         myChangeBar.attach();
 *     });
 *
 * Or, making that shorter with chainable functions, assuming we like all the default options:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar().attach();
 *     });
 *
 * Alternatively, since ChangeBar by default creates and maintains it's own element, it can be attached more manually:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar();
 *         var $element = $(...);
 *         $element.append(myChangeBar.$element);
 *     });
 *
 * Finally, if we have already created some ChangeBar on some existing element [as described below](#custom_dom), we could avoid calling `attach()` or `$element` entirely:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar({
 *             // since version 2.0.6
 *             fromExisting: $('.my-changebar-parent'),
 *         });
 *     });
 *
 * In this final form, we create our ChangeBar by using the `fromExisting` option and passing either an element which `.is` a `{{#change-bar-controls}}` or can `.find` a `{{#change-bar-controls}}`.
 *
 * <a id="errors"></a>
 * #### Dealing with errors
 *
 * ChangeBar attempts to deal with errors in a very non-opinionated, flexible manner.
 * For every error, the only required field is `message` - any other fields to the object can be added according to your needs.
 * (Though note how ChangeBar uses <a href="#hidden_errors">hidden errors</a>).
 * Internally, all functions use <a href="http://devdocs.io/underscore/index#where">_.where</a> when trying to match particular errors.
 *
 * So, how do we use this?
 * Let's say that we have multiple detail panels and navigation tabs, and errors can be associated with each.
 * We might have error objects like the following:
 *
 *     myChangeBar.errors.add([
 *         {detailPane: 'pane-events', navTab: 'project-1', message: '...'},
 *         {detailPane: 'pane-discussions', navTab: 'project-1', message: '...'},
 *         {detailPane: 'pane-discussions', navTab: 'project-2', message: '...'},
 *         {detailPane: 'pane-events', navTab: 'project-3', message: '...'}
 *     ]);
 *
 * We could clear out all errors associated with project-1 with the following:
 *
 *     myChangeBar.errors.removeMatching({navTab: 'project-1'});
 *
 * We could find all the errors associated with project-1 by the following:
 *
 *     myChangeBar.errors.findMatching({navTab: 'project-1'});
 *
 * In addition, `errors.all` returns all errors, `errors.visible` returns all visible errors, and `errors.removeAll` can be useful for dealing with the `cancel` event.
 *
 * <a id="selecting_errors"></a>
 * ##### Selecting Errors (AKA focusing errors)
 *
 * When the user clicks on an error in the list of errors, pages should bring browser focus to the error in question.
 * The <a data-anchor="#event_select-error" data-tabid="#events" href="#event_select-error">select-error</a> event exists to support this functionality.
 *
 * <a id="hidden_errors"></a>
 * ##### Hidden Errors
 *
 * If errors include the property `{hidden: true}`, those errors are considered hidden.
 * Hidden errors do not increment the ChangeBar's error count and are not visible in the dropdown when the user clicks to see the errors.
 * However, if there are no visible errors, but there are hidden errors, and the user clicks save, the hidden errors will become visible (that is, ChangeBar will loop through the list of errors, update the hidden errors to have `{hidden: false}`, and repopulate the error button/list).
 *
 * <a id="custom_dom"></a>
 * #### Customizing the ChangeBar DOM (since 2.0.6)
 *
 * ChangeBar uses several helpers in order to support generating custom DOM.
 * The default is a Handlebars template with the following content:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors}}
 *         \{{change-bar-cancel}}
 *         \{{change-bar-save}}
 *     \{{/change-bar-controls}}
 *
 * The above is a typical page-level ChangeBar, with access keys and the normal text for all the three buttons.
 * If you were using ChangeBar in a modal, something like the following would be more appropriate:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors "simple"}}
 *         \{{change-bar-cancel "secondary"}}
 *         \{{change-bar-save "primary"}}
 *     \{{/change-bar-controls}}
 *
 * Since no accesskeys would be created, but the controls are otherwise suitable for ChangeBar to make use of.
 *
 * If custom text for the save/cancel buttons is necessary, something like the following would work:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors "simple"}}
 *         \{{#change-bar-cancel "secondary"}}My cancel text\{{/change-bar-cancel}}
 *         \{{#change-bar-save "primary"}}My save text\{{/change-bar-save}}
 *     \{{/change-bar-controls}}
 *
 * When generating custom DOM as any of the above examples, see the `fromExisting` and `changeBarMarkup` options on how to use them.
 * Basically, `changeBarMarkup` allows you to specify the entire string that ChangeBar will wrap (and subsequent `.attach()` calls would append to DOM).
 * `fromExisting` allows you to add the DOM whenever works best for you, and ChangeBar would then grab that DOM on creation.
 *
 * <a id="helper_details"></a>
 * ##### Some helper details
 *
 * * `change-bar-controls` requires block content
 * * `change-bar-errors:` only one parameter.  "default", assumed if no parameters provided, gives the error button an accesskey.  No accesskey is provided if "simple" is passed.
 * * `change-bar-cancel` if no parameters are provided, suitable for the page level in Prime (i.e. has an accesskey).
 *     If parameters are provided, the `button` helper is used, provided with default text/attributes as required for this ChangeBar, but otherwise options are passed right along to the `button` helper.
 * * `change-bar-save` if no parameters are provided, suitable for the page level in Prime (i.e. has an accesskey).
 *     If parameters are provided, the `button` helper is used, provided with default text/attributes as required for this ChangeBar, but otherwise options are passed right along to the `button` helper.
 *
 * @class ui.ChangeBar
 * @module ui
 * @param [options] {Object} a map of options; see 'Properties' tab below for available options
 * @param [actions] {Object} handlers for events. See the events tab for <a data-anchor="#event_save" data-tabid="#events" href="#event_save">save</a>, <a data-anchor="#event_cancel" data-tabid="#events" href="#event_cancel">cancel</a>, <a data-anchor="#event_select-error" data-tabid="#events" href="#event_select-error">selectError</a>, <a data-anchor="#event_pre-destroy" data-tabid="#events" href="#event_pre-destroy">preDestroy</a>, and <a data-anchor="#event_unhide-errors" data-tabid="#events" href="#event_unhide-errors">unhideErrors</a>.
 *     Passing any of these values as part of this actions hash is equivalent to attaching listeners to ChangeBar events later, e.g. `myChangeBar.on('select-error', actions.selectError)`.
 * @constructor
 */

define('primutils/ui/changeBar/ChangeBar',['primutils.jquery', 'primutils.lodash', 'primutils/utils/Subscribable', 'primutils.logger',
    'primutils/utils/language',
    'primutils/utils/validator',
    'primutils/ui/changeBar/ChangeBarView',
    'primutils/markup/primutils/ui/changeBar/changeBar.hbs:#change-bar',
    'primutils/markup/primutils/ui/changeBar/changeBar.hbs:#error-item'
], function ($, _, Subscribable, logger, lang, validator, ChangeBarView, changeBarTemplate, errorItemTemplate) {
    

    var ALLOWED_EVENTS = ['save', 'cancel', 'select-error', 'unhide-errors', 'pre-destroy'];

    function failOnceDestroyed() {
        throw new Error('ChangeBar has been destroyed, calling methods is no longer legal');
    }

    function isJquerySelector($node) {
        return !!$node.jquery;
    }

    var __validator__ = validator({
        name: 'ChangeBar',
        validOptions: {
            id: 'String',
            /**
             * Specifies the element to attach to, in the event of using `myChangeBar.attach()`.
             * Can be a String, in which case changeBar will select for that selector using jQuery to attach to, a jQuery selector already, or a Function that should return either of the same two types.
             *
             * @property attachTo
             * @type {String|$jquerySelector|Function}
             */
            attachTo: ['String', 'Function', isJquerySelector],
            /**
             * Will be used to sort any errors before rendering them to the DOM.
             * Will be called every time errors are repopulated; ChangeBar does not attempt to detect that the errors have not been updated in order to avoid doing work.
             * The function should function as per the native <a href="http://devdocs.io/javascript/global_objects/array/sort">Array.sort</a>.
             *
             * @property sortErrors
             * @type {undefined|Function}
             */
            sortErrors: [undefined, 'Function'],
            /**
             * Whether or not save should be allowed.
             * Regardless of the result of this option, save will not be allowed if the ChangeBar is not visible.
             * An example choice for this function might be something like:
             *
             *     allowSave: function (cb) {
             *         return cb.util.defaultFns.allowSave() || myPage.allowSaveWithErrors();
             *     }
             *
             * See also the method `util.defaultFns.allowSave`
             *
             * @property allowSave
             * @type {undefined|Function}
             * @since 2.0.3
             * @default a function that returns true if there are no visible errors in the changebar and false if there are visible errors
             */
            allowSave: [undefined, 'Function'],
            /**
             * Called every time errors are populated in this changebar, to determine whether or not the changebar should be shown.
             * If this function returns true, `changebar.visibility.show()` will be called, otherwise it will not.
             *
             * An example choice for this function might be something like:
             *
             *     checkAutoShow: function (cb) {
             *         return myPage.isDirty();
             *     }
             *
             * See also the method `util.defaultFns.checkAutoShow`
             *
             * @property checkAutoShow
             * @type {undefined|Function}
             * @default a function that returns true if there are errors in the changebar, false otherwise
             * @since 2.0.3
             */
            checkAutoShow: [undefined, 'Function'],
            /**
             * Used to configure the markup created by this ChangeBar.
             * Not used if the option `fromExisting` is specified.
             * See the class description for more details on DOM customization.
             *
             * @property changeBarMarkup
             * @type {undefined|String}
             * @default {String} the default DOM suitable for the page level in Prime
             */
            changeBarMarkup: [undefined, 'String'],
            /**
             * Used to attach this ChangeBar to existing DOM.
             * If this is used, overrides the `changeBarMarkup` option.
             * Must be attached to existing DOM that uses the DOM as created by the ChangeBar helpers `change-bar-controls`, `change-bar-errors`, `change-bar-cancel`, and `change-bar-save`.
             * See the class description for more details on DOM customization.
             *
             * @since 2.0.6
             * @property fromExisting
             * @type {String|$jquerySelector|Function}
             * @default undefined
             */
            fromExisting: ['String', 'Function', isJquerySelector],
            /**
             * Whether the change bar should be centered over its DOM anchor point.
             *
             * This option is `true` by default because most change bars go in a page banner where Blue Sky dictates that change bars should be centered.
             * If you are using a change bar for a different reason, you will probably want to set this option to `false`.
             *
             * @since 2.7.11
             * @property centerChangeBar
             * @type {Boolean}
             * @default true
             */
            centerChangeBar: 'Boolean',
            /**
             * Whether or not to automatically close and commit any editors on grids on the page.
             * This typically only matters if the save button accesskey is used, and it's recommended to leave this as true.
             *
             * @since 2.6.10
             * @property closeGridEditorsOnSave
             * @type {Boolean}
             * @default true
             */
            closeGridEditorsOnSave: 'Boolean',
        },
        validators: {
            initialize: {}
        },
        ignoredSubkeys: ['attachTo', 'fromExisting']
    });

    var __actionValidator = validator({
        name: 'ChangeBar Actions',
        validOptions: {
            save: 'Function',
            cancel: 'Function',
            selectError: 'Function',
            unhideErrors: 'Function',
        },
        validators: {
            initialize: {}
        }
    });

    var repopulateErrors = function () {
        if (_.isFunction(this._options.sortErrors)) {
            this._errors.sort(this._options.sortErrors);
        }

        this._view.redraw();
        if (this._options.checkAutoShow(this)) {
            this.show();
        }
    };

    var addErrorWithoutRepopulate = function (error) {
        validator.checkType(error.hidden, [undefined, 'Boolean']);
        if (!validator.checkType(error.message, 'String', 'changeBarError.message')) {
            error.message = lang.getString('changeBar.error');
        }

        this._errors.push(error);
    };

    var getValue = function (opt) {
        return _.isFunction(opt) ? opt.apply(this, _.rest(_.toArray(arguments))) : opt;
    };

    /**
     * The default function behavior for the `checkAutoShow` option.
     * Returns true if there are errors managed by this changebar, false otherwise.
     *
     * @method util.defaultFns.checkAutoShowss
     * @since 2.0.3
     */
    var checkAutoShow = function (cb) {
        return cb.errors.all().length > 0;
    };

    /**
     * The default function behavior for the `allowSave` option.
     * Returns true if there are no visible errors managed by this changebar, false otherwise.
     *
     * @method util.defaultFns.allowSave
     * @since 2.0.3
     */
    var allowSave = function (cb) {
        return cb.errors.visible().length === 0;
    };

    function getDefaults () {
        var id = _.uniqueId('change-bar-');
        return {
            id: id,
            attachTo: '.content .banner',
            centerChangeBar: true, // Gonna default this to true because most changebars are in the banner
            fromExisting: undefined,
            changeBarMarkup : changeBarTemplate({
                changeBarId : id
            }),
            errorItemMarkup : errorItemTemplate,
            locale : {
                errors : lang.getString('changeBar.errorsWithCounter'),
                defaultStatusMessage : lang.getString('changeBar.defaultStatusMessage')
            },
            allowSave: allowSave,
            checkAutoShow: checkAutoShow,
            closeGridEditorsOnSave: true,
        };
    }

    function ChangeBar (options, actions) {
        logger.assert(this && this instanceof ChangeBar, 'ChangeBar must be created with new');
        Subscribable.call(this, {
            allowedEvents: ALLOWED_EVENTS
        });

        options = options || {};
        validator.checkType(options, ['Object']);
        __validator__.validate('initialize', options);
        Object.defineProperty(this, '_options', {
            value: _.defaults({}, options, getDefaults()),
            writable: false,
            enumerable: false,
        });

        actions = actions || {};
        validator.checkType(actions, ['Object']);
        __actionValidator.validate('initialize', actions);
        if (_.isFunction(actions.save)) {
            this.on('save', actions.save);
        }
        if (_.isFunction(actions.cancel)) {
            this.on('cancel', actions.cancel);
        }
        if (_.isFunction(actions.selectError)) {
            this.on('select-error', actions.selectError);
        }
        if (_.isFunction(actions.unhideErrors)) {
            this.on('unhide-errors', actions.unhideErrors);
        }
        if (_.isFunction(actions.preDestroy)) {
            this.on('pre-destroy', actions.preDestroy);
        }

        Object.defineProperty(this, '_view', {
            value: new ChangeBarView(this),
            writable: false,
            enumerable: false
        });

        this._errors = [];

        this.visibility = this.vis = {};
        this.visibility.toggle = this.toggleChanged.bind(this);
        this.visibility.show = this.show.bind(this);
        this.visibility.hide = this.hide.bind(this);
        this.visibility.saving = this.saving.bind(this);
        this.visibility.saveSucceeded = this.saveSucceeded.bind(this);
        this.visibility.saveFailed = this.saveFailed.bind(this);
        this.visibility.up = this.isVisible.bind(this);

        this.errors = {};
        this.errors.all = this.getErrors.bind(this);
        this.errors.visible = this.getVisibleErrors.bind(this);
        this.errors.add = this.addError.bind(this);
        this.errors.findMatching = this.findMatchingErrors.bind(this);
        this.errors.removeMatching = this.removeMatchingErrors.bind(this);
        this.errors.removeAll = this.removeAllErrors.bind(this);

        this.util = {};
        this.util.attach = this.attach.bind(this);
        this.util.detach = this.detach.bind(this);
        this.util.destroy = this.destroy.bind(this);
        this.util.destroyWithoutDOM = this.destroyWithoutDOM.bind(this);
        this.util.defaultFns = {};
        this.util.defaultFns.checkAutoShow = checkAutoShow.bind(this);
        this.util.defaultFns.allowSave = allowSave.bind(this);
        this.util.updateButtons = this.updateButtons.bind(this);
        this.util.getSavingState = this.getSavingState.bind(this);

        this.events = {};
        /**
         * Attaches event handlers for this change bar.
         * For details on event syntax, see <a href="utils.Subscribable.html">Subscribable</a>.
         *
         * @method events.on
         */
        this.events.on = this.on.bind(this);
        /**
         * Detaches event handlers for this change bar.
         * For details on event syntax, see <a href="utils.Subscribable.html">Subscribable</a>.
         *
         * @method events.off
         */
        this.events.off = this.off.bind(this);
        this.events.trigger = this.trigger.bind(this);
    }

    var cbp = ChangeBar.prototype = Object.create(Subscribable.prototype);

    Object.defineProperty(cbp, '$element', {
        get : function () {
            return this._view.$el;
        }
    });

    // maybe consider documenting this as `this.util.attach`.  Keep backwards compatability either way
    /**
     * Simple chainable utility for attaching this ChangeBar to the page.
     * Respects the option `attachTo` for what element to append to.
     *
     * @method attach
     * @chainable
     */
    cbp.attach = function () {
        logger.info(this._options.id + ' is attaching to: ' + this._options.attachTo);
        $(getValue.call(this, this._options.attachTo)).append(this.$element);
        repopulateErrors.call(this);
        return this;
    };

    /**
     * Simple chainable utility for detaching this ChangeBar from the DOM.
     * Note: uses $.fn.detach internally, so does not disconnect event listeners, etc.
     * For that purpose, see <a href="#method_destroy">destroy</a>.
     *
     * @method detach
     * @chainable
     */
    cbp.detach = function () {
        logger.info(this._options.id + ' is detaching from DOM');
        this.$element.detach();
        return this;
    };

    /**
     * Simple chainable utility for destroying this ChangeBar.
     * Note: uses $.fn.destroy internally, so does disconnect event listeners, etc.
     * If this is not desired, consider <a href="#method_detach">detach</a>.
     *
     * @method destroy
     */
    cbp.destroy = function () {
        logger.info(this._options.id + ' is getting destroyed');
        /**
         * Triggered immediately before this ChangeBar is destroyed.
         *
         * @event pre-destroy
         * @since 2.0.8
         * @param cb {ChangeBar} the ChangeBar that's getting destroyed (this)
         */
        this.trigger('pre-destroy', this);
        this.detach();
        this.$element.remove();

        this.events = failOnceDestroyed;
        this.on = failOnceDestroyed;
        Subscribable.prototype.clear.call(this);
        this.off = failOnceDestroyed;
        this.trigger = failOnceDestroyed;
    };

    cbp.destroyWithoutDOM = function () {
        logger.info(this._options.id + ' is getting destroyed (but leaving DOM in place)');
        this.trigger('pre-destroy', this);
        this._errors = [];
        repopulateErrors.call(this);
        this._checkAllowSave = function () {
            return false;
        };
        this.hide();
        this._view.removeListeners();
        this._view.$el = $([]);

        this.events = failOnceDestroyed;
        this.on = failOnceDestroyed;
        Subscribable.prototype.clear.call(this);
        this.off = failOnceDestroyed;
        this.trigger = failOnceDestroyed;
    };

    /**
     * Returns whether or not this ChangeBar is visible.
     * Please note: if the ChangeBar is `detach`'d from the DOM, it will still return true, while this would be false after `destroy`.
     *
     * @method visibility.up
     * @returns {Boolean} true if this ChangeBar is visible, false otherwise
     * @since 2.0.2
     */
    cbp.isVisible = function () {
        return !!this._view.isVisible();
    };

    /**
     * Toggles the visibility of the changeBar.
     * If any errors are present, the changeBar will not hide.
     *
     * @method visibility.toggle
     * @chainable
     */
    cbp.toggleChanged = function (changed) {
        if (changed) {
            this.show();
        } else {
            this.hide();
        }
        return this;
    };

    /**
     * Update the state of the buttons (save, cancel, errors) managed by this changebar.
     *
     * @method util.updateButtons
     * @chainable
     * @since 2.3.9
     */
    cbp.updateButtons = function () {
        this._view.updateButtonStates();
        return this;
    };

    /**
     * Shows the changeBar.
     *
     * @method visibility.show
     * @chainable
     */
    cbp.show = function () {
        this._view.show();
        return this;
    };

    /**
     * Sets the change bar status to saving.
     *
     * *This method is NOT chainable.*
     *
     * @method visibility.saving
     * @return {{succeeded:Function,failed:Function}}
     */
    cbp.saving = function () {
        this._view.saving();

        return {
            succeeded : this.saveSucceeded.bind(this),
            failed : this.saveFailed.bind(this)
        };
    };

    /**
     * Sets the change bar status to saved and hides the change bar after flashing the success button.
     *
     * @method visibility.saveSucceeded
     * @chainable
     */
    cbp.saveSucceeded = function () {
        this._view.saveSucceeded();
        return this;
    };

    /**
     * Sets the change bar status to unsaved and keeps the change bar visible.
     *
     * @method visibility.saveFailed
     * @chainable
     */
    cbp.saveFailed = function () {
        this._view.saveFailed();
        return this;
    };

    cbp.getSavingState = function () {
        return this._view.getSavingState();
    };

    /**
     * Hides the changeBar.
     *
     * @method visibility.hide
     * @chainable
     */
    cbp.hide = function (force) {
        if (this._errors.length === 0 || force === true) {
            this._view.hide();
        }
        return this;
    };

    /**
     * Gets all errors, including invisible ones.
     *
     * @method errors.all
     * @returns {Array} all errors, hidden or not, possibly empty
     */
    cbp.getErrors = function () {
        return this._errors;
    };

    /**
     * Gets all visible errors.
     *
     * @method errors.visible
     * @returns {Array} all visible errors, possibly empty
     */
    cbp.getVisibleErrors = function () {
        return _.difference(this._errors, this.findMatchingErrors({ hidden: true }));
    };

    /**
     * Registers an error, or array of errors, with this changeBar.
     * When registering errors, the `message` property is required.
     * If `hidden` is set to true, the error is considered an invisible one and will only show up if the user attempts to save before dealing with all invisible errors.
     *
     * Note: ChangeBar does *not* attempt to deduplicate errors.
     * If you add the same error object twice, that's your prerogative.
     * Consider `errors.removeMatching` for deduplicating errors, and TODO Brian Sacksteder's blog post for dealing with grid errors.
     *
     * @method errors.add
     * @chainable
     * @param errorItem {Array|Object} the error Object, or Array of error Objects, to be added to this changeBar
     */
    cbp.addError = function (errorItem) {
        logger.assert(errorItem && (_.isObject(errorItem) || _.isArray(errorItem)), 'changeBar.errors.add only takes objects or arrays');

        if (_.isArray(errorItem)) {
            this._addBulkErrors(errorItem);
        }
        else {
            addErrorWithoutRepopulate.call(this, errorItem);
            repopulateErrors.call(this);
        }

        return this;
    };

    cbp._addBulkErrors = function (errors) {
        _.each(errors, function (err) {
            logger.assert(err && _.isObject(err), 'When adding multiple errors, the array must be an array of objects');
            addErrorWithoutRepopulate.call(this, err);
        }.bind(this));

        repopulateErrors.call(this);

        return this;
    };

    /**
     * Returns all errors matching the given example.
     * As <a href="http://devdocs.io/underscore/index#where">_.where</a> used on the list of errors.
     *
     * @method errors.findMatching
     * @param example {Object} an Object to be used as a filter for the errors with this change bar.
     *      For example, if the object `{controller:'foo', tab:'bar'}` is used as the example, any error with both of those properties present and matching will be included in the return of this function.
     * @returns {Array} all matching errors
     */
    cbp.findMatchingErrors = function (example) {
        return _.where(this._errors, example);
    };

    /**
     * Removes all errors matching the given example.
     * Internally uses <a href="http://devdocs.io/underscore/index#where">_.where</a> to find the matching errors to be removed.
     *
     * @method errors.removeMatching
     * @param example {Object} an Object to be used as a filter for the errors to be removed.
     *      For example, if the object `{controller:'foo', tab:'bar'}` is used as the example, any error with both of those properties present and matching will be removed.
     * @chainable
     */
    cbp.removeMatchingErrors = function (example) {
        var matchingErrors = this.findMatchingErrors(example);
        this._errors = _.difference(this._errors, matchingErrors);
        repopulateErrors.call(this);
        return this;
    };

    /**
     * Removes all the errors associated with this change bar.
     *
     * @method errors.removeAll
     * @chainable
     */
    cbp.removeAllErrors = function () {
        this._errors.splice(0, this._errors.length);
        repopulateErrors.call(this);
        return this;
    };

    /**
     * Triggered when an error is selected (aka clicked) by the user.
     *
     * @event select-error
     * @param {Object} error the error that was selected (clicked) by the user.
     */
    cbp._selectError = function (error) {
        this.trigger('select-error', error);
        return this;
    };

    /**
     * Triggered when the user clicks save.
     * Will only trigger from ChangeBar itself if no errors are present.
     *
     * @event save
     */
    cbp._save = function () {
        if (this.getVisibleErrors().length) {
            // not sure how this would happen, but doesn't hurt to be paranoid
            $.noop();
        }
        else if (this.getErrors().length) {
            var unhiddenErrors = [];
            this._errors = _.map(this._errors, function (err) {
                if (err.hidden === true) {
                    unhiddenErrors.push(err);
                    err.hidden = false;
                }
                return err;
            });
            repopulateErrors.call(this);

            /**
             * Triggered when the user clicks save, when no visible errors were present but hidden errors were, and those errors have been made visible.
             *
             * @event unhide-errors
             * @param {Array} errors the errors that were now made visible
             */
            this.trigger('unhide-errors', unhiddenErrors);
        }
        else {
            this.trigger('save');
        }
        return this;
    };

    /**
     * Triggered when the user clicks cancel.
     *
     * @event cancel
     */
    cbp._cancel = function () {
        this.trigger('cancel');
        return this;
    };

    // This function is exposed on ChangeBar only for ChangeBarView to use it: buyer beware, do not use this private API if you're looking at this comment
    cbp._checkAllowSave = function () {
        return this._options.allowSave.call(undefined, this);
    };

    return ChangeBar;
});


define('primutils/ui/ChangeBar',['primutils/ui/changeBar/ChangeBar'], function (ChangeBar) {
    

    return ChangeBar;
});



define('primutils/markup/primutils/ui/dialogs/alert.hbs', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/alert.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <span class=\"modal-icon\">";
  stack1 = lambda((depth0 != null ? depth0.icon : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</span>\r\n        <p>";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</p>\r\n        <div class=\"additional-details\">\r\n            <a href=\"#\" class=\"additional-details-label\">";
  stack1 = lambda((depth0 != null ? depth0.additionalDetailsLabel : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</a>\r\n            <div class=\"additional-details-body\">\r\n                ";
  stack1 = lambda((depth0 != null ? depth0.additionalDetails : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n            </div>\r\n        </div>\r\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.ok : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("pgbu-modal-with-icon message-modal"),
    'id': ("alert-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/dialogs/doNotShow.hbs', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/doNotShow.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"pull-left\">\n    <div class=\"controls\">\n        <label class=\"checkbox suppress-modal-checkbox\" for=\"doNotShowAgain\">\n            <input type=\"checkbox\" id=\"doNotShowAgain\">\n            ";
  stack1 = lambda((depth0 != null ? depth0.doNotShowAgain : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n        </label>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/dialogs/doNotShowHelper',[], function () {
    

    var DO_NOT_SHOW_KEY = 'ORA_CPP_DIALOG_DO_NOT_SHOW_DATA';

    function getCurrentDoNotShowData () {
        var currentDoNotShow = {};
        if (window.localStorage && window.localStorage[DO_NOT_SHOW_KEY]) {
            currentDoNotShow = JSON.parse(window.localStorage[DO_NOT_SHOW_KEY]);
        }
        return currentDoNotShow;
    }

    return {
        toggleDoNotShow : function (key, state) {
            var currentDoNotShow = getCurrentDoNotShowData();
            currentDoNotShow[key] = state;
            if (window.localStorage) {
                window.localStorage[DO_NOT_SHOW_KEY] = JSON.stringify(currentDoNotShow);
            }
        },

        isDoNotShow : function (key) {
            return !!getCurrentDoNotShowData()[key];
        }
    };
});


/**
 * Displays a message to the user that they must acknowledge by pressing "ok".
 * Additional information can be provided which is hidden by default.
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "ok" or closes the modal, see the `ok` method or `okCallback` property.
 *
 * @class ui.dialogs.AlertDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/AlertDialog',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils/markup/primutils/ui/dialogs/alert.hbs',
    'primutils/markup/primutils/ui/dialogs/doNotShow.hbs',
    'primutils/utils/language',
    'primutils/ui/dialogs/doNotShowHelper',
    'primutils/a11y/accessibilityUtils',
    'primutils.logger'
], function ($, _, alertTemplate, doNotShowTemplate, lang, doNotShowHelper, accessibilityUtils, logger) {
    

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again", the <a href="#property_okCallback">okCallback</a> will automatically
         * be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false,
        /**
         * Whether to automatically call `show` when the spinner is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay : true
        /**
         * A method which will be called when the user clicks the Ok button. This method will *not* be called if the
         * user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes no arguments.
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default "Alert" - localized
             */
            title : lang.getString('dialogTitles.alert'),
            /**
             * The message to display inside the dialog.
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * A string representing an icon to display. You must specify the full markup.
             * @property icon {String}
             * @optional
             * @default The warning icon from the Blue Sky Glyphs font set
             */
            icon : '<i class="pgbu-icon pgbu-icon-warning" title="'+ lang.getString('dialogTitles.alert') +'"></i>',
            /**
             * The label to display on the collapsed additional information section
             * @property additionalDetailsLabel {String}
             * @optional
             * @default "Additional Details" - localized
             */
            additionalDetailsLabel : lang.getString('dialogs.additionalInfo'),
            /**
             * Additional information about the alert that the user can expand if they care about it.
             * @property additionalDetails {String}
             * @optional
             * @default ""
             */
            additionalDetails : '',
            /**
             * The text to display on the default Ok button
             * @property ok {String}
             * @optional
             * @default "Ok" - localized
             */
            ok : lang.getString('dialogs.ok')
        };
    }

    function bindOptionsCallbacks(options, dfd) {
        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }
    }

    function AlertDialog (options) {
        var that = this,
            dfd = $.Deferred();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        if (options.iconClass) {
            logger.warn('The option "iconClass" is deprecated, use "icon" instead.');
            options.icon = '<i class="'+ options.iconClass +'"></i>';
        }

        if (!_.isString(options.doNotShowAgain)) {
            options.doNotShowAgain = false;
        }

        /**
         * Adds a method which will be called when the user clicks the Ok button. This method will *not* be called if
         * the user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes no arguments.
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {AlertDialog} The alert dialog instance
         */
        this.ok = dfd.promise().done;

        if (options.doNotShowAgain !== false && doNotShowHelper.isDoNotShow(options.doNotShowAgain)) {
            dfd.resolve();
            bindOptionsCallbacks(options, dfd);
            return; // Don't bother constructing the rest of the dialog if we're not showing it
        }

        this.$modal = $(alertTemplate({
            close : lang.getString('dialogs.close'),
            title: options.title,
            message: options.message,
            icon: options.icon,
            additionalDetailsLabel: options.additionalDetailsLabel,
            additionalDetails: options.additionalDetails,
            ok: options.ok
        }));

        if (options.additionalDetails) {
            var $additionalDetails = this.$modal.find('.additional-details');
            $additionalDetails.show().on('click', '.additional-details-label', function (e) {
                e.preventDefault();
                $additionalDetails.toggleClass('open');
            });
        }

        if (options.doNotShowAgain !== false) {
            this.$modal.find('.modal-footer').prepend(doNotShowTemplate({
                doNotShowAgain : lang.getString('dialogs.dontShowAgain')
            }));
        }

        this.$modal.find('[data-dismiss=modal]').on('click', function(event) {
            // Do not toggle the do not show
            dfd.resolve();
            event.stopPropagation();
        });

        this.$modal.find('.proceedModal').on('click', function (event) {
            if (options.doNotShowAgain !== false) {
                doNotShowHelper.toggleDoNotShow(options.doNotShowAgain, that.$modal.find('#doNotShowAgain').prop('checked'));
            }

            dfd.resolve();

            event.stopPropagation();
        });

        this.$modal.on('shown', function () {
            $('.proceedModal').focus();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        bindOptionsCallbacks(options, dfd);

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : options.autoDisplay && (dfd.state() === 'pending'),
            backdrop : 'static'
        });
    }

    AlertDialog.prototype = {
        constructor : AlertDialog,

        /**
         * Shows the modal
         * @method show
         */
        show : function () {
            this.$modal.modal('show');
        },

        /**
         * Opens the additional details panel
         * @method openAdditionalDetails
         */
        openAdditionalDetails : function () {
            this.$modal.find('.additional-details').addClass('open');
        }
    };

    return AlertDialog;
});


define('primutils/markup/primutils/ui/dialogs/cancel.hbs', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/cancel.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <p>";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</p>\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <div class=\"pull-left doNotShowAgainContainer\">\n            <div class=\"controls\">\n                <label class=\"checkbox suppress-modal-checkbox\" for=\"doNotShowAgain\">\n                    <input type=\"checkbox\" id=\"doNotShowAgain\"> ";
  stack1 = lambda((depth0 != null ? depth0.doNotShowAgain : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n                </label>\n            </div>\n        </div>\n        <div class=\"pull-right\">\n            ";
  stack1 = helpers.button.call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(7, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n        </div>\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.no : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"7":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.yes : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("cancel-modal"),
    'id': ("cancel-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Displays a message to the user asking them if they would like to cancel their changes.
 * The user can either cancel (by clicking "yes") or not cancel their changes (by clicking "no").
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "yes", see the `yes` method or `yesCallback` property.
 * To add a handler for when the user clicks "no" or closes the modal, see the `no` method or `noCallback` property.
 *
 * @class ui.dialogs.CancelDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/CancelDialog',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils/markup/primutils/ui/dialogs/cancel.hbs',
    'primutils/utils/language',
    'primutils/ui/dialogs/doNotShowHelper',
    'primutils/a11y/accessibilityUtils'
], function ($, _, cancelTemplate, lang, doNotShowHelper, accessibilityUtils) {
    

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again", the <a href="#property_okCallback">okCallback</a> will automatically
         * be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false
        /**
         * A method which will be called when the user clicks the Yes button.
         *
         * The callback takes no arguments.
         *
         * @property yesCallback {Function}
         * @optional
         */
        //yesCallback,
        /**
         * A method which will be called when the user clicks the No button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property noCallback {Function}
         * @optional
         */
        //noCallback
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default "Cancel Changes" - localized
             */
            title : lang.getString('dialogTitles.cancel'),
            /**
             * The message to display inside the dialog.
             * @property message {String}
             * @optional
             * @default "Are you sure you want to cancel your changes?" - localized
             */
            message : lang.getString('dialogs.cancelMessage'),
            /**
             * The text to display on the default Yes button
             * @property yes {String}
             * @optional
             * @default "Yes" - localized
             */
            yes : lang.getString('dialogs.yes'),
            /**
             * The text to display on the No button
             * @property no {String}
             * @optional
             * @default "No" - localized
             */
            no : lang.getString('dialogs.no')
        };
    }

    function bindOptionsCallbacks(options, dfd) {
        if (_.isFunction(options.yesCallback)) {
            dfd.done(options.yesCallback);
        }

        if (_.isFunction(options.noCallback)) {
            dfd.fail(options.noCallback);
        }
    }

    function CancelDialog (options) {
        var that = this,
            dfd = $.Deferred(),
            promise = dfd.promise();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        if (!_.isString(options.doNotShowAgain)) {
            options.doNotShowAgain = false;
        }

        /**
         * Adds a method which will be called when the user clicks the Yes button.
         *
         * The callback takes no arguments.
         *
         * @method yes
         * @param fn {Function} The callback function
         * @chainable
         * @returns {CancelDialog} The cancel dialog instance
         */
        this.yes = promise.done;
        /**
         * Adds a method which will be called when the user clicks the No button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method no
         * @param fn {Function} The callback function
         * @chainable
         * @returns {CancelDialog} The cancel dialog instance
         */
        this.no = promise.fail;

        if (options.doNotShowAgain !== false && doNotShowHelper.isDoNotShow(options.doNotShowAgain)) {
            dfd.resolve();
            bindOptionsCallbacks(options, dfd);
            return; // Don't bother constructing the rest of the dialog if we're not showing it
        }

        this.$modal = $(cancelTemplate({
            close : lang.getString('dialogs.close'),
            title : options.title,
            message : options.message,
            yes : options.yes,
            no : options.no,
            doNotShowAgain : lang.getString('dialogs.dontShowAgain')
        }));

        if (options.doNotShowAgain === false) {
            this.$modal.find('.doNotShowAgainContainer').hide();
        }

        this.$modal.find('.proceedModal').on('click', function (event) {
            if (options.doNotShowAgain !== false) {
                doNotShowHelper.toggleDoNotShow(options.doNotShowAgain, that.$modal.find('#doNotShowAgain').prop('checked'));
            }

            dfd.resolve();

            event.stopPropagation();
        });

        this.$modal.find('.cancelModal, [data-dismiss=modal]').on('click', function (event) {
            // Do not toggle the do not show
            dfd.reject();
            event.stopPropagation();
        });

        this.$modal.on('shown', function () {
            $('.proceedModal').focus();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        bindOptionsCallbacks(options, dfd);

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : (dfd.state() === 'pending'),
            backdrop : 'static'
        });
    }

    CancelDialog.prototype = {
        constructor : CancelDialog
    };

    return CancelDialog;
});


define('primutils/markup/primutils/ui/dialogs/confirm.hbs', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/confirm.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{
    'id': ("confirm-modal-message")
  },"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <p class=\"break-long-words\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</p>\r\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(7, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.cancel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"7":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.accept : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("confirm-modal"),
    'id': ("confirm-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Displays a custom message to the user to confirm whether or not they really want to perform the action which triggers this ConfirmDialog.
 * Contrary to most modals, ConfirmDialog does not provide a default `message`, so you will always want to configure that option.
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "yes", see the `ok` method or `okCallback` property.
 * To add a handler for when the user clicks "no" or closes the modal, see the `cancel` method or `cancelCallback` property.
 *
 * @class ui.dialogs.ConfirmDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/ConfirmDialog',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils.logger',
    'primutils/markup/primutils/ui/dialogs/confirm.hbs',
    'primutils/markup/primutils/ui/dialogs/doNotShow.hbs',
    'primutils/utils/language',
    'primutils/ui/dialogs/doNotShowHelper',
    'primutils/a11y/accessibilityUtils'
], function ($, _, logger, confirmTemplate, doNotShowTemplate, lang, doNotShowHelper, accessibilityUtils) {
    

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again" and clicks Ok, the <a href="#property_okCallback">okCallback</a> will
         * automatically be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false
        /**
         * A method which will be called when the user clicks the Ok button.
         *
         * The callback takes no arguments.
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback,
        /**
         * A method which will be called when the user clicks the Cancel button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property cancelCallback {Function}
         * @optional
         */
        //cancelCallback
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default "Confirm" - localized
             */
            title : lang.getString('dialogTitles.confirm'),
            /**
             * The message to display inside the dialog.
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * The text to display on the default Yes button
             * @property acceptLabel {String}
             * @optional
             * @default "Yes" - localized
             */
            acceptLabel : lang.getString('dialogs.yes'),
            /**
             * The text to display on the No button
             * @property cancelLabel {String}
             * @optional
             * @default "No" - localized
             */
            cancelLabel : lang.getString('dialogs.no')
        };
    }

    function bindOptionsCallbacks(options, dfd) {
        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }

        if (_.isFunction(options.cancelCallback)) {
            dfd.fail(options.cancelCallback);
        }
    }

    function ConfirmDialog (options) {
        var that  = this,
            dfd = $.Deferred(),
            promise = dfd.promise();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());
                if (!options.message) {
            logger.warn('message property is falsey - no default message is provided, but you probably want to provide one');
        }
        
        if (!_.isString(options.doNotShowAgain)) {
            options.doNotShowAgain = false;
        }

        /**
         * Adds a method which will be called when the user clicks the Ok button.
         *
         * The callback takes no arguments.
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {ConfirmDialog} The confirm dialog instance
         */
        this.ok = promise.done;
        /**
         * Adds a method which will be called when the user clicks the Cancel button. This method will also be called if
         * the user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method cancel
         * @param fn {Function} The callback function
         * @chainable
         * @returns {ConfirmDialog} The confirm dialog instance
         */
        this.cancel = promise.fail;

        if (options.doNotShowAgain !== false && doNotShowHelper.isDoNotShow(options.doNotShowAgain)) {
            dfd.resolve();
            bindOptionsCallbacks(options, dfd);
            return; // Don't bother constructing the rest of the dialog if we're not showing it
        }

        this.$modal = $(confirmTemplate({
            close : lang.getString('dialogs.close'),
            title: options.title,
            message: options.message,
            accept: options.acceptLabel,
            cancel: options.cancelLabel
        }));

        if (options.doNotShowAgain !== false) {
            this.$modal.find('.modal-footer').prepend(doNotShowTemplate({
                doNotShowAgain : lang.getString('dialogs.dontShowAgain')
            }));
        }

        this.$modal.find('.proceedModal').on('click', function (event) {
            if (options.doNotShowAgain !== false) {
                doNotShowHelper.toggleDoNotShow(options.doNotShowAgain, that.$modal.find('#doNotShowAgain').prop('checked'));
            }

            dfd.resolve();

            event.stopPropagation();
        });

        this.$modal.find('.cancelModal, [data-dismiss=modal]').on('click', function(event) {
            // Do not toggle the do not show
            dfd.reject();
            event.stopPropagation();
        });

        this.$modal.on('shown', function () {
            $('.proceedModal').focus();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        bindOptionsCallbacks(options, dfd);

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : (dfd.state() === 'pending'),
            backdrop : 'static'
        });
    }

    ConfirmDialog.prototype = {
        constructor : ConfirmDialog
    };

    return ConfirmDialog;
});


define('primutils/markup/primutils/ui/dialogs/loadingDialog.hbs', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/loadingDialog.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"page-spinner page-loading-spinner page-spinner-static\" aria-labelledby=\"loading-modal-label\">\n    <div class=\"spinner-label\" id=\"loading-modal-label\">";
  stack1 = lambda((depth0 != null ? depth0.loading : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n    <div class=\"spinner-large loading-spinner-large\">\n        <div class=\"dot dot1\"></div>\n        <div class=\"dot dot2\"></div>\n        <div class=\"dot dot3\"></div>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Creates a loading spinner over a particular component to signify that the content fetched may take a while to load.
 *
 * Unlike the <a href="ui.dialogs.SavingDialog.html">SavingDialog</a>, the LoadingDialog does not create a mask over
 * the page content. The user is free to navigate away from loading content at any time.
 *
 * Example:
 *
 *     require(['jquery', 'primutils/ui/dialogs/LoadingDialog'], function ($, LoadingDialog) {
 *
 *         // creating a new spinner showing the defaults
 *         var spinner = new LoadingDialog({
 *             autoDisplay : true,
 *             target : $('.my-container')
 *         });
 *
 *         $.get('mystuff')
 *             .done(function (result) {
 *                 // do something with result
 *             })
 *             .always(function () {
 *                 // always close the spinner, even if AJAX call fails
 *                 spinner.complete();
 *             });
 *     });
 *
 * @class ui.dialogs.LoadingDialog
 * @module ui.dialogs
 */
define('primutils/ui/dialogs/LoadingDialog',['primutils.jquery', 'primutils.lodash', 'primutils.logger', 'primutils/markup/primutils/ui/dialogs/loadingDialog.hbs', 'primutils/utils/language'],
        function ($, _, logger, loadingSpinnerTemplate, lang) {
    

    var defaultTarget = $('body');

    var DEFAULT_OPTIONS = {
        /**
         * Whether to automatically call `show` when the spinner is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay : true
        /**
         * The jQuery element to display the spinner over
         *
         * @property target {jQuery Object}
         * @optional
         * @default $('body')
         */
        //target
        /**
         * The text to place above the spinner
         *
         * @propery loading {String}
         * @optional
         * @default "Loading" - localized
         */
        //loading
    };

    function LoadingDialog (options) {

        this.options = _.defaults({}, options, DEFAULT_OPTIONS, {
            // Need to specify this here, so it pulls the latest from "global"
            target : defaultTarget
        });

        // So that stop can be passed into promise.always / promise.then.
        this.complete = _.bind(this.complete, this);
        this.done = function () {
            logger.warn('LoadingDialog.done() is deprecated.  Use .complete() instead.  .done() may be removed without further warning.  Full docs at: http://10.140.183.183:8080/job/primutils/javadoc/classes/ui.dialogs.LoadingDialog.html');
            this.complete();
        }.bind(this);

        if (this.options.autoDisplay) {
            this.show();
        }
    }

    LoadingDialog.prototype = {
        constructor : LoadingDialog,

        /**
         * Shows the spinners over the current target(s)
         *
         * @method show
         */
        show: function () {
            var that = this,
                $spinner = $();
            this.options.target.each(function () {
                var $this = $(this);

                // Need to set position
                // The $target of the spinner needs to be positioned something non-"static" in order for the spinner to
                // be positioned properly.
                // TODO be smarter about this. Store original position in data so that the proper value can be restored
                // if it was set in inline CSS
                if ($this.css('position') === 'static') {
                    $this.css('position', 'relative');
                    $this.data('setPosition', true);
                }

                var $result = $(loadingSpinnerTemplate({
                        loading: that.options.loading || lang.getString('dialogTitles.loading')
                    }))
                    .css('display', 'none')
                    .appendTo($this)
                    .fadeIn();

                $spinner = $spinner.add($result);
            });

            this.$spinner = $spinner;
        },

        /**
         * Tells the spinner that it is complete and should hide and remove itself.
         *
         * @method complete
         */
        complete: function () {
            var that = this;
            this.$spinner.fadeOut(function () {
                $(this).remove();
                that.options.target.each(function () {
                    var $this = $(this);
                    // unset position if it was set
                    if ($this.data('setPosition')) {
                        $this.css('position', '').removeData('setPosition');
                    }
                });
            });
            this.$spinner = $();
        }
    };

    /**
     * Sets the default target for all created loading dialogs.
     *
     * Useful when your application has a main content area that isn't the document body
     *
     * @method setDefaultTarget
     * @static
     * @param $target {jQuery Object} The default target of the loading dialog
     */
    LoadingDialog.setDefaultTarget = function ($target) {
        defaultTarget = $target;
    };

    return LoadingDialog;

});


define('primutils/markup/primutils/ui/dialogs/promptMarkup.hbs', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/controlGroup','primutils/templates/helpers/input','primutils/templates/helpers/join'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/promptMarkup.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <div class=\"short-message\">";
  stack1 = lambda((depth0 != null ? depth0.directions : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\r\n";
  stack1 = helpers.controlGroup.call(depth0, (depth0 != null ? depth0.message : depth0), {"name":"controlGroup","hash":{},"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, buffer = "            ";
  stack1 = helpers.input.call(depth0, "text", {"name":"input","hash":{
    'hintHidden': (true),
    'hintId': ("prompt-help"),
    'hint': ((depth0 != null ? depth0.helpText : depth0)),
    'maxlength': ((depth0 != null ? depth0.maxLength : depth0)),
    'class': ("prompt-text input-block-level"),
    'id': ("prompt-input")
  },"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(6, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(8, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"6":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.cancel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"8":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.accept : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("prompt-modal"),
    'id': ("prompt-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Prompts the user for a value with a textual input.
 * You can supply a message to the user and provide a default value.
 * The user can then either choose to submit the value or cancel entering a value.
 *
 * To add a handler for when the user has entered text and clicks "ok", see the `ok` method or `okCallback` property.
 * To add a handler for when the user clicks "cancel" or closes the modal, see the `cancel` method or `cancelCallback` property.
 *
 * @class ui.dialogs.PromptDialog
 * @module ui.dialogs
 * @param options {Object} See properties
 * @constructor
 */
define('primutils/ui/dialogs/PromptDialog',[
    'primutils.jquery',
    'primutils.lodash',
    'primutils/markup/primutils/ui/dialogs/promptMarkup.hbs',
    'primutils/utils/keyCode',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils'
], function ($, _, confirmTemplate, keyCode, lang, accessibilityUtils) {
    

    var DEFAULT_OPTIONS = {
        /**
         * Whether or not the user must entire a value in the dialog before proceeding.
         *
         * @property required {Boolean}
         * @optional
         * @default false
         */
        required : false
        /**
         * A callback to check whether the value entered in the dialog is unique. A unique check is only run when the
         * user attempts to submit the value.
         *
         * The callback takes one parameter, the value entered in the box.
         *
         *     checkUnique : function (value) { ... }
         *
         * The return value can either be an object the returns the following properties or a `$.when` compatible
         * promise that resolves to the same. `message` is optional and will default to the value of the `helpText` option.
         *
         *     return {
         *         result : true // or false; Whether this item already exists, i.e. returns true when the item is *NOT* unique
         *         message : ''  // or missing; A message to show the user when the violated the constraint.
         *     };
         *
         * You can also return a simple `true` or `false` which matches `result` in the object above.
         *
         * @property checkUnique {Function}
         * @optional
         */
        //checkUnique,
        /**
         * A method which will be called when the user clicks the Ok button. This method will *not* be called if the
         * user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes one arguments, the value entered in the box.
         *
         *     okCallback : function (value) { ... }
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback,
        /**
         * A method which will be called when the user clicks the Cancel button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property cancelCallback {Function}
         * @optional
         */
        //cancelCallback
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default "Prompt" - localized
             */
            title : lang.getString('dialogTitles.prompt'),
            /**
             * The initial value of the text box.
             * @property defaultText {String}
             * @optional
             * @default ""
             */
            defaultText : '',
            /**
             * The label to display above the text box.
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * A short paragraph explaining the purpose of the prompt. Displays above the text box and label.
             * @property directions {String}
             * @optional
             * @default ""
             */
            directions : '',
            /**
             * The default error message to show when `checkUnique` or `required` constraints are not met.
             * @property helpText {String}
             * @optional
             * @default ""
             */
            //helpText,
            /**
             * The text to display on the default Ok button
             * @property acceptLabel {String}
             * @optional
             * @default "Ok" - localized
             */
            acceptLabel : lang.getString('dialogs.ok'),
            /**
             * The text to display on the Cancel button
             * @property cancelLabel {String}
             * @optional
             * @default "Cancel" - localized
             */
            cancelLabel : lang.getString('dialogs.cancel'),
            /**
             * The max length of the value in the text box.
             *
             * @property maxLength {Number|String}
             * @optional
             */
            maxLength : ''
        };
    }

    function PromptDialog (options) {
        var that = this,
            dfd = $.Deferred(),
            promise = dfd.promise();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        /**
         * Adds a method which will be called when the user clicks the Ok button.
         *
         * The callback takes one arguments, the value entered in the box.
         *
         *     function (value) { ... }
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {PromptDialog} The prompt dialog instance
         */
        this.ok = promise.done;
        /**
         * Adds a method which will be called when the user clicks the Cancel button. This method will also be called if
         * the user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method cancel
         * @param fn {Function} The callback function
         * @chainable
         * @returns {PromptDialog} The prompt dialog instance
         */
        this.cancel = promise.fail;

        this.$modal = $(confirmTemplate({
            close : lang.getString('dialogs.close'),
            title : options.title,
            helpText : options.helpText,
            message : options.message,
            directions : options.directions,
            accept : options.acceptLabel,
            cancel : options.cancelLabel,
            maxLength : options.maxLength
        }));

        if (options.required) {
            this.$modal.find('.control-group').addClass('required');
            this.$modal.find('.prompt-text').attr('required', 'required').attr('aria-required', 'true').attr('aria-describedby', 'prompt-help');
        }

        function getPromptValue() {
            return that.$modal.find('.prompt-text').val();
        }

        function setError(msg) {
            var message = msg || options.helpText;
            that.$modal.find('.control-group').addClass('error').find('.help-block').text(message).css('display', '');
            that.$modal.find('.proceedModal').prop('disabled', true);
        }

        function clearError() {
            that.$modal.find('.control-group').removeClass('error').find('.help-block').css('display', 'none');
            that.$modal.find('.proceedModal').prop('disabled', false);
        }

        var validate = function () {
            if (_.isFunction(options.checkUnique)) {
                $.when(options.checkUnique(getPromptValue())).done(function (response) {
                    var result, message;

                    if (_.isObject(response)) {
                        result = response.result;
                        message = response.message || options.helpText;
                    } else {
                        result = response;
                        message = options.helpText;
                    }

                    if (!result) {
                        dfd.resolve(getPromptValue());
                    }
                    else {
                        setError(message);
                    }
                }).fail(function () {
                    setError(lang.getString('dialogs.serverValidationError'));
                    // Re-enable the proceed button, as we cannot say the value is invalid so let them try again
                    that.$modal.find('.proceedModal').prop('disabled', false);
                });
            }
            else if (options.required) {
                // Accounts for if user hits enter with a required field left blank
                if (getPromptValue() !== '') {
                    dfd.resolve(getPromptValue());
                }
            }
            else {
                dfd.resolve(getPromptValue());
            }
        };

        that.$modal.find('.prompt-text')
            .val(options.defaultText)
            .on('keydown', function (ev) {
                /* This function was put in strictly for some odd IE10 behavior. Modals
                 * lose focus when users hit 'Enter', and the keydown event was firing on
                 * the 'Close Tab' buttons. This prevents 'Enter' keydown from triggering
                 * the wrong events in IE, and doesn't affect anything else.
                 */
                if (ev.keyCode === keyCode.ENTER) {
                    return false;
                }
            })
            .on('keydown', function () {
                clearError();
            })
            .on('blur', function () {
                if (getPromptValue().length === 0 && options.required) {
                    setError(lang.getFormattedString('dialogs.required', options.message));
                }
            })
            .on('keyup', function (e) {
                var valid = !options.required || $(this).val().length > 0;
                that.$modal.find('.proceedModal').prop('disabled', !valid);
                if (e.keyCode === keyCode.ESC) {
                    dfd.reject();
                }
            });

        this.$modal.find('.proceedModal')
            .prop('disabled', options.defaultText.length === 0 && options.required)
            .on('click', function (event) {
                //disable button when clicked so users cannot hit button twice.
                that.$modal.find('.proceedModal').prop('disabled', true);
                validate();
            });

        this.$modal.find('.cancelModal, [data-dismiss=modal]').on('click', function (event) {
            dfd.reject();
        });

        this.$modal.on('shown', function () {
            that.$modal.find('.prompt-text').focus().get(0).select();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }

        if (_.isFunction(options.cancelCallback)) {
            dfd.fail(options.cancelCallback);
        }

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : true,
            backdrop : 'static'
        });
    }

    PromptDialog.prototype = {
        constructor : PromptDialog
    };

    return PromptDialog;
});


define('primutils/markup/primutils/ui/dialogs/savingMarkup.hbs:#saving-progress', ['primutils.handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/join'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/savingMarkup.hbs:#saving-progress'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{
    'hideClose': (true),
    'hideTitle': (true)
  },"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <h3 id=\"";
  stack1 = lambda(((stack1 = (depth0 != null ? depth0.__modalContext__ : depth0)) != null ? stack1.labeledBy : stack1), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</h3>\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <div class=\"progress progress-striped progress-success active\">\n            <div class=\"bar\" style=\"width: 0;\"></div>\n        </div>\n        <div style=\"display: none !important;\" class=\"saving-aria-message\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" role=\"alert\">";
  stack1 = lambda((depth0 != null ? depth0.savingInProgress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'aria-describedby': (helpers._.call(depth0, (depth0 != null ? depth0.spinnerId : depth0), "-aria-message", {"name":"_","hash":{},"data":data})),
    'style': ("top: 30% !important;"),
    'id': ("saving-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/dialogs/savingMarkup.hbs:#saving-indeterminate', ['primutils.handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/savingMarkup.hbs:#saving-indeterminate'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"page-spinner page-saving-spinner page-spinner-static hide fade saving-dialog\" aria-labelledby=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-saving-modal-label\" aria-describedby=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" style=\"outline: none;\" tabindex=\"-1\" role=\"dialog\">\n    <div class=\"spinner-label\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-saving-modal-label\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n    <div class=\"spinner-large saving-spinner-large\">\n        <div class=\"dot dot1\"></div>\n        <div class=\"dot dot2\"></div>\n        <div class=\"dot dot3\"></div>\n    </div>\n    <div style=\"display: none !important;\" class=\"saving-aria-message\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" role=\"alert\">";
  stack1 = lambda((depth0 != null ? depth0.savingInProgress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Creates a saving dialog in either a progress bar or spinner format.
 *
 * Unlike the <a href="ui.dialogs.LoadingDialog.html">LoadingDialog</a>, the SavingDialog creates a mask over the
 * page content which prevents the user from interacting with the application while the save operation is running.
 * It is imperative that you call `complete` in order to return control of the application back to the user.
 *
 * Normally, `type: 'spinner'` (the default) should be used.
 * If the progress of your operation is known, `type: 'progress'` may be preferred.
 *
 * You can change the name of the spinner using the `message` option.
 *
 *     require(['jquery', 'primutils/ui/dialogs/SavingDialog'], function ($, SavingDialog) {
 *
 *         // creating a new dialog showing the defaults
 *         var dialog = new SavingDialog({
 *             type : 'spinner',
 *             autoDisplay : true
 *         });
 *
 *         $.get('mystuff')
 *             .done(function (result) {
 *                 // do something with result
 *             })
 *             .always(function () {
 *                 // always close the spinner, even if AJAX call fails
 *                 dialog.complete();
 *             });
 *     });
 *
 * @class ui.dialogs.SavingDialog
 * @module ui.dialogs
 */
define('primutils/ui/dialogs/SavingDialog',['primutils.jquery', 'primutils.lodash', 'primutils.logger',
        'primutils/markup/primutils/ui/dialogs/savingMarkup.hbs:#saving-progress',
        'primutils/markup/primutils/ui/dialogs/savingMarkup.hbs:#saving-indeterminate',
        'primutils/utils/language',
        'primutils/a11y/accessibilityUtils',

        'primutils/utils/emulateTransitionEnd' // $.fn.emulateTransitionEnd
], function ($, _, logger, progressTemplate, indeterminateTemplate, lang, accessibilityUtils) {
    

    var DEFAULT_OPTIONS = {
        /**
         * Is this a progress bar or a spinner.
         *
         * Valid options are 'spinner' or 'progress'
         *
         * @property type {String}
         * @optional
         * @default 'spinner'
         */
        type: 'spinner',
        /**
         * Whether to automatically call `show` when the dialog is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay: true
    };

    function getDefaultStrings () {
        return {
            /**
             * The default message to display, can be overridden. It's probably best to keep this somewhat short.
             *
             * @property message {String}
             * @optional
             * @default "Saving" - localized
             */
            message: lang.getString('dialogTitles.saving')
        };
    }

    function SavingDialog (options) {
        logger.assert(this && this instanceof SavingDialog, 'SavingDialog must be instantiated with new.');

        this.options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        // So that stop can be passed into promise.always / promise.then.
        this.complete = _.bind(this.complete, this);
        this.done = function () {
            logger.warn('SavingDialog.done() is deprecated.  Use .complete() instead.  .done() may be removed without further warning.  Full docs at: http://10.140.183.183:8080/job/primutils/javadoc/classes/ui.dialogs.SavingDialog.html');
            this.complete();
        }.bind(this);

        this._percent = 0.0;
        this._shown = false;

        this.savingProgress = progressTemplate({
            spinnerId : _.uniqueId('saving-spinner-'),
            message: this.options.message,
            savingInProgress: lang.getString('dialogs.savingInProgress')
        });
        this.savingIndeterminate = indeterminateTemplate({
            spinnerId : _.uniqueId('saving-spinner-'),
            message: this.options.message,
            savingInProgress: lang.getString('dialogs.savingInProgress')
        });


        if (this.options.autoDisplay) {
            this.show();
        }
    }

    SavingDialog.prototype = {
        constructor : SavingDialog,

        /**
         * Shows the dialog
         *
         * @method show
         */
        show : function () {
            // Do not show twice
            if (this._shown) { return; }

            if (this.options.type === 'progress') {
                this._$dialog = $(this.savingProgress);
            } else {
                this._$dialog = $(this.savingIndeterminate);
            }

            // Remove old modal focus traps
            $(document).off('focusin.modal');

            accessibilityUtils.modalAppend(this._$dialog, {allowClosingParent: true});

            this._$dialog.modal({
                show: true,
                backdrop: 'static',
                keyboard: false
            });

            this._shown = true;
        },

        /**
         * Tells the spinner that it is complete and should hide and remove itself.
         *
         * @method complete
         */
        complete : function () {
            var that = this;

            if (this.options.type === 'progress') {
                this.percentage(1.0);
            }

            this._$dialog.one('hidden.hide', function () {
                that._$dialog.remove();
                that.announceComplete();
            }).modal('hide');

            this._shown = false;
        },

        /**
         * Changes the percentage of the progress bar.
         * The newPercent value should always be greater than the previous percentage otherwise display glitches may
         * result.
         *
         * Only available if type is 'progress'.
         *
         * @method percentage
         * @param newPercent {Number} If specified, set the percent to this value (0.0 - 1.0)
         */
        percentage : function (newPercent) {
            if (this.options.type !== 'progress') {
                this._percent = 0.0;
            }

            if (!arguments.length) {
                return this._percent;
            }

            if (this.options.type === 'progress') {
                // Clamp between 0.0 and 1.0
                this._percent = Math.min(Math.max(0, newPercent), 1);

                this._$dialog.find('.progress .bar').css('width', (this._percent * 100) + '%');

                if (this._percent === 1.0) {
                    this.complete();
                }
            }
        },

        announceComplete : function () {
            var $saveComplete = $('<div style="position: fixed; top: -10000px; left: -10000px;" class="saving-complete-alert" role="alert" aria-live="assertive" aria-relevant="all" aria-atomic="true">' + lang.getString('dialogs.savingCompleted') + '</div>').appendTo(document.body);
            setTimeout(function () {
                $saveComplete.remove();
            }, 0);
        }
    };

    return SavingDialog;

});


/**
 * This class is still available as the `primutils/utils/promisifyError` module, but the documentation is now located at [promises](utils.promises.html#method_promisifyError).
 *
 * @private
 * @module utils
 * @class utils.promisifyError
 */

define('primutils/utils/promisifyError',['primutils.jquery', 'primutils.lodash', 'primutils.logger'], function ($, _, logger) {
    

    return function (fn) {
        return function () {
            var prom = $.Deferred();
            try {
                var result = fn.apply(this, _.toArray(arguments));
                if (result && result.then && _.isFunction(result.then)) { // we'll assume it's a promise
                    return result;
                }
                return prom.resolve(result).promise();
            } catch (err) {
                logger.error('ERROR occurred, wrapping as a rejected promise: ' + err);
                return prom.reject(err).promise();
            }
        };
    };

});



/**
 * This helper is documented at [promises](utils.promises.html#method_stripResolveData).
 *
 * @private
 * @module utils
 * @class utils.stripPromiseResolveData
 */

define('primutils/utils/stripPromiseResolveData',['primutils.lodash', 'primutils.jquery'], function (_, $) {
    

    return function (promise) {
        var dfd = $.Deferred();

        promise
            .done(function () {
                dfd.resolve();
            })
            .fail(function () {
                dfd.reject.apply(dfd, arguments);
            });

        return dfd.promise();
    };
});



/**
 * A collection of small utilities for use with asynchronous control flow in JS, specifcally methods around dealing with promises.
 *
 * All methods in this module are static.  See methods list, below, for details.
 *
 * @module utils
 * @class utils.promises
 */

define('primutils/utils/promises',['primutils.jquery',
        'primutils.lodash',
        'primutils/utils/promisifyError',
        'primutils/utils/stripPromiseResolveData'], function ($, _, promisifyError, stripPromiseResolveData) {
    

    return {

        /**
         * `promisifyError` is a utility designed to improve error handling that may occur in promise chains.
         *
         * To describe it's use, say we have the following:
         *
         *     // define statement, etc.
         *     $.get(...)
         *         .then(appendElementsToPage)
         *         .fail(showAlertDialogOnFailure);
         *
         * This code could work: GET some information from the server, and use it to insert something into the DOM.
         * If the AJAX request fails, the alert dialog would get shown.
         * But if the AJAX request goes through and `appendElementsToPage` throws an uncaught error, showing the alert dialog wouldn't happen (e.g. if the function was `function appendElementsToPage() { throw new Error(); }`)
         * This utility function allows us to address the error.
         * To update the above example, let's try the following:
         *
         *     define([..., 'primutils/utils/promises', function (..., promises) {
         *         $.get(...)
         *            .then(promises.promisifyError(appendElementsToPage))
         *            .fail(showAlertDialogOnFailure);
         *     });
         *
         * With this snippet, both an error in the AJAX request and the `appendElementsToPage` function will cause `showAlertDialogOnFailure` to get called.
         * If both the AJAX request and `appendElementsToPage` operate successfully, the behavior is the same as the first snippet.
         *
         * *Note*: this function will always return a promise, even if the result was passing.
         *
         * *Note to Prime developers*: if you're using [AJAXErrorHandler](http://pgbuci.us.oracle.com:8080/job/ui-platform-cont/Javascript_YUIDocs/classes/shared.dialogs.AJAXErrorHandler.html) as your failure handler, AJAXErrorHandler will propagate the underlying JS error for better debugging.
         *
         * *alias*: `primutils/utils/promisifyError` module.
         *
         * @method promisifyError
         * @param fn {Function} the function to be called, which is given the promise's resolved data (as if a `done` call)
         * @return a promise, which will be rejected() if the given fn throws some error
         * @static
         * @since 2.9.0 (`primutils/utils/promisifyError` module was available earlier)
         */
        promisifyError: promisifyError,

        /**
         * `stripResolveData` is a utility designed to encourage clean contracts among promise consumers.
         *
         * To describe it's use, say we have the following:
         *
         *     // define statement, etc.
         *     function loadData() {
         *         return $.get(...)
         *             .done(function (data) {
         *                 store.addAll(..., data.values);
         *             });
         *     }
         *     // ...
         *     loadData().done(function (data) { // well, I have data, so what they heck... });
         *
         * This function works, but tends to promote a bad practice - since loadData returns the original promise (`done` does not change the promise chaining context), consumers of the loadData function may attempt to use the same data.values member, even though it has already been loaded into the store.
         *
         * `stripResolveData` can be used as such to strip any success data given by the original promise resolution (e.g. the data, for AJAX requests), but keep any failure data:
         *
         *     define([..., 'primutils/utils/promises'], function (..., promises) {
         *         function loadData() {
         *             return promises.stripResolveData($.get(...)
         *                 .done(function (data) {
         *                     store.addAll(..., data.values);
         *                 }));
         *         }
         *         loadData().done(function () { // arguments is an empty list here, can't use it });
         *     });
         *
         * *alias*: `primutils/utils/stripPromiseResolveData` module.
         *
         * @method stripResolveData
         * @param promise {Promise} the promise to mirror in terms of success/failure results, but whose resolved data should not be given to consumers
         * @return a promise, which will be resolved() if the original promise is resolved, and will maintain failure data if the original promise is rejected
         * @static
         * @since 2.9.0
         */
        stripResolveData: stripPromiseResolveData,
    };
});


define('primutils/utils/urls',['primutils.lodash'], function (_) {
    

    /**
     * A collection of utilities for dealing with URLs.
     *
     * @class utils.urls
     * @module utils
     */

    var that = {
        /**
         * Gets a param from the current `window.location`.
         * Does not handle arrays of data in the URL such as `aParam[]=val1&aParam[]=val2`.
         *
         * For example, if the URL is:
         *
         *     http://HOST:PORT/primavera/p/error?tab=5&loggerLevel=DEBUG&encoded=my%20user#other
         *
         * Then:
         *
         *     getSingleParam('tab', 'default');    // '5'
         *     getSingleParam('nonParam', '567');   // '567'
         *     getSingleParam('loggerLevel', null); // 'DEBUG'
         *     getSingleParam('encoded', null);     // 'my user' (as of 2.7.7, previous versions did not decode parts)
         *
         * @method getSingleParam
         * @param paramName {String} the name of the parameter to get
         * @param defaultValue the default return value if the param is not present in the URL - this will not get decoded
         * @returns the parameter value in the URL if found, defaultValue otherwise
         */
        getSingleParam: function (paramName, defaultValue) {
            return that.getSingleParamFromURL(window.location.search, paramName, defaultValue);
        },

        getSingleParamFromURL: function (url, paramName, defaultValue) {
            var urlParams = url.replace(/^.*\?/, '').split('&');

            var result = defaultValue;
            for (var params_index = 0, len = urlParams.length; params_index < len; params_index++) {
                var param = urlParams[params_index];
                var pair = param.split('=');
                var key = pair[0] || '';
                var value = pair[1] || '';

                if (key === paramName) {
                    result = window.decodeURIComponent(value);
                }
            }

            return result;
        }
    };

    return that;

});



define('primutils/utils/History',['primutils.jquery', 'primutils.lodash', 'primutils.logger', 'primutils/utils/urls', 'primutils/utils/Subscribable'], function ($, _, logger, urls, Subscribable) {
    

    /**
     * History is a thin wrapper around the HTML5 history API.
     *
     * It's purpose is to track one param in the URL (as configured by the `param` option).
     *
     * It provides a helper method for updating this param in the URL (`pushParam`).
     *
     * It provides an event (`param-updated`) anytime this param is updated through any method other than the `pushParam` method.
     *
     * @class utils.History
     * @module utils
     */

    function History(options) {
        logger.assert(this && this instanceof History, 'History must be created with \'new\'');

        /**
         * The param to be tracked.
         * For example, if the param given is `user`, the URLs `...&user=5` `...&user=6` represent two possible states for the `param-updated` event to switch between.
         *
         * @property param
         * @type {String}
         * @required
         */
        logger.assert(options && _.isString(options.param), 'History must be created with a param to track as part of provided options');
        this._options = _.defaults({}, options, {
            eventNamespace: _.uniqueId('history_'),
        }, this._options);

        /**
         * Triggered every time the param is updated, except when `pushParam` is called.
         *
         * @event param-updated
         * @param param {String} the param that changed (same as options.param)
         * @param newValue {String} the new value of the tracked param
         */
        Subscribable.call(this, {
            allowedEvents: ['param-updated']
        });
        this._init();
    }
    History.prototype = Object.create(Subscribable.prototype);

    var hp = History.prototype;

    hp._options = {};

    hp._init = function () {
        this._supportsHistory = window.history && window.history.pushState;
        this._paramMatcher = new RegExp(this._options.param + '=[^&#]+');
        this._attach();
    };

    hp._attach = function () {
        var that = this;

        if (!this._supportsHistory) {
            return;
        }

        var ns = this._options.eventNamespace;
        $(window).on('popstate.' + ns, function () {
            /**
             * A default value provided to the `param-updated` event in the case that the tracked URL param is no longer present.
             *
             * For example, on the social user profile page, `options.defaultValue` is the user's ID, meaning that the user's own profile is loaded if the URL no longer contains some user identifier.
             *
             * @property defaultValue
             * @type {String}
             * @optional
             */
            that.trigger('param-updated', that._options.param, urls.getSingleParam(that._options.param, that._options.defaultValue));
        });

        $('body').on('click.' + ns, this._linkFilter(), function (ev) {
            // let users modify keypress if they'd like
            if (ev.ctrlKey || ev.shiftKey) {
                return;
            }

            ev.preventDefault();

            var newParam = $(this).attr('data-historyreplace-value');

            if (newParam && !that._skipValue(newParam)) {
                that.trigger('param-updated', that._options.param, newParam, that._getOldValue());
                that.pushParam(newParam);
            }
        });
    };

    hp._getOldValue = function () {
        return urls.getSingleParam(this._options.param, undefined);
    };

    hp._skipValue = function (newValue) {
        var oldValue = this._getOldValue();
        return oldValue !== undefined && oldValue === newValue;
    };

    hp._linkFilter = function () {
        return '[data-historyreplace-name="' + this._options.param + '"]';
    };

    /**
     * Cleans up event listeners, etc.
     *
     * @method destroy
     */
    hp.destroy = function () {
        var ns = this._options.eventNamespace;
        $(window).off('popstate.' + ns);
        $('body').off('click.' + ns, this._linkFilter());
    };

    /**
     * Pushes a new param state to the URL, if possible (i.e. if the browser supports the HTML5 history API), otherwise does nothing.
     *
     * @method pushParam
     * @param paramValue {String} the new value of the param for the URL
     */
    hp.pushParam = function (paramValue) {
        if (!this._supportsHistory || this._skipValue(paramValue)) {
            return;
        }

        var loc = window.location;
        var paramName = this._options.param;
        var newLocation;
        if (loc.search.indexOf(paramName) !== -1) {
            newLocation = loc.href.replace(this._paramMatcher, paramName + '=' + window.encodeURIComponent(paramValue));
        }
        else {
            var baseLocation = loc.origin + loc.pathname + loc.search;
            var newParamPrefix = loc.search ? '&' : '?';
            newLocation = baseLocation + newParamPrefix + paramName + '=' + window.encodeURIComponent(paramValue) + loc.hash;
        }

        history.pushState(null, null, newLocation);
    };

    return History;
});



define('primutils/formatters/formatterFactory',['primutils.lodash',
        'primutils/formatters/wrappers',
        'primutils/formatters/formats',
        'primutils/formatters/numberFormatter',
        'primutils/formatters/dateFormatter',
        'primutils/formatters/durationFormatter'], function (_, builtinWrappers, builtinFormats, numFmt, dateFmt, durFmt) {
    

    var DEFAULT_OPTIONS = {
        fourDigitYear: true,
        monthName: true,
        leadingZeros: true,

        dateSeparator: '/',
        dateFormat: '0',
            // day-month-year: 1
            // month-day-year: 0
            // year-month-day: 2

        showDecimals: true,
        numberDecimals: 2,
        currencyShowSymbol: true,
        currencySymbol: '$',
        decimalSeparator: '.',
        digitSeparator: ',',

        positiveFormat: '#1.1',

        // TODO: primutils shouldn't add a '-' character in this case
        negativeFormat: '(#1.1)',

        // duration formatting:
        hourChar: 'h',
        dayChar: 'd',
        weekChar: 'w',
        monthChar: 'm',
        hoursPerDay: '8',
        hoursPerWeek: '40',
        hoursPerMonth: '172'
    };

    function makeZeros(count) {
        var result = '';
        for (var i = 0; i < count; i++) {
            result += '0';
        }
        return result;
    }

    function convertServerFormats(options) {
        options.positiveFormat = options.positiveFormat
            .replace('#', '$')
            .replace('1.1', '0,0');
        options.negativeFormat = options.negativeFormat
            .replace('#', '$')
            .replace('1.1', '0,0');
    }

    function formatterFactory(givenOptions) {
        var options = _.defaults({}, givenOptions || {}, DEFAULT_OPTIONS);
        var formatters = {};
        // # -> $, 1.1 -> 0,0
        convertServerFormats(options);

        var decimalFormat = '';
        if (options.showDecimals) {
            decimalFormat = '.' + makeZeros(options.numberDecimals);
        }

        var positiveCurrencyFormat = options.positiveFormat.replace('0,0', '0,0' + decimalFormat);
        var negativeCurrencyFormat = options.negativeFormat.replace('0,0', '0,0' + decimalFormat);
        var positiveNumberFormat = '0,0' + decimalFormat;
        var positiveIntegerFormat = '0,0';
        var bareIntegerFormat = '0';

        var currencySymbol = options.currencyShowSymbol ? options.currencySymbol : '';

        var decimalSeparator = options.decimalSeparator;
        var digitSeparator = options.digitSeparator;

        var baseNumberFormatter = numFmt({
            alwaysShowNegativeSymbol: false,
            decimalSeparator: decimalSeparator,
            thousandsSeparator: digitSeparator
        });

        var year = options.fourDigitYear ? 'YYYY' : 'YY';
        var month = options.monthName ? 'MMM' : options.leadingZeros ? 'MM' : 'M';
        var day = options.leadingZeros ? 'DD' : 'D';
        var dateOrder =
            options.dateFormat === '0' ? [month, day, year] :
            options.dateFormat === '1' ? [day, month, year] :
                                                 [year, month, day];

        var baseDateFormat = dateOrder.join(options.dateSeparator);

        // TODO baseTimeFormat should respect actual user preferences
        var baseTimeFormat = 'HH:mm';
        var timezoneSensitiveDateTimeFormat = baseDateFormat + ' ' + baseTimeFormat;
        var timezoneSensitiveTimeFormat = baseTimeFormat;
        var globalDateFormat = baseDateFormat;
        var globalDateTimeFormat = baseDateFormat + ' ' + baseTimeFormat;
        var globalTimeFormat = baseTimeFormat;

        var hoursPerDay = parseFloat(options.hoursPerDay);
        var hoursPerWeek = parseFloat(options.hoursPerWeek);
        var hoursPerMonth = parseFloat(options.hoursPerMonth);
        var durationFormatterOptions = {
            workableDecimalPlaces: 6,
            decimalsToShow : options.showDecimals ? options.numberDecimals : 0,

            decimalSeparator: decimalSeparator,
            thousandsSeparator: digitSeparator,

            showIn: options.hourChar,
            units: {
            }
        };
        durationFormatterOptions.units[options.hourChar] = {
            scale: 1,
            sub: null
        };
        durationFormatterOptions.units[options.dayChar] = {
            scale: hoursPerDay,
            sub: options.hourChar
        };
        durationFormatterOptions.units[options.weekChar] = {
            scale: hoursPerWeek,
            sub: options.dayChar
        };
        durationFormatterOptions.units[options.monthChar] = {
            scale: hoursPerMonth,
            sub: options.weekChar
        };

        var baseDateFmt = dateFmt.withOptions({
            forgiveLeadingZeros: true
        });

        formatters = {
            options: {
                currency: {
                    symbol: options.currencySymbol
                },
                numeric: {
                    decimalSeparator: decimalSeparator,
                    thousandsSeparator: digitSeparator,
                    showDecimals: options.showDecimals,
                    decimalPlaces: options.numberDecimals
                },
                datetime: {
                    year: year,
                    month: month,
                    day: day,
                    dateSeparator: options.dateSeparator,
                },
                duration: {
                    hoursPerDay: hoursPerDay,
                    hoursPerWeek: hoursPerWeek,
                    hoursPerMonth: hoursPerMonth
                },
            },

            timezoneSensitiveDateTimeFormatter: baseDateFmt.withOptions({
                format: timezoneSensitiveDateTimeFormat,
                unserializeFormat: builtinFormats.timezoneSensitiveDateTimeUnserialization,
                serializeFormat: builtinFormats.timezoneSensitiveDateTimeSerialization,

                // TODO not in production, for speed improvement?
                unserializeWrapper: builtinWrappers.timezoneSensitiveDateTimeUnserializeValidation
            }),

            timezoneSensitiveTimeFormatter: baseDateFmt.withOptions({
                format: timezoneSensitiveTimeFormat,
                unserializeFormat: builtinFormats.timezoneSensitiveDateTimeUnserialization,
                serializeFormat: builtinFormats.timezoneSensitiveDateTimeSerialization,

                // TODO not in production, for speed improvement?
                unserializeWrapper: builtinWrappers.timezoneSensitiveDateTimeUnserializeValidation
            }),

            globalDateFormatter: baseDateFmt.withOptions({
                format: globalDateFormat,
                unserializeFormat: builtinFormats.globalDateUnserialization,
                serializeFormat: builtinFormats.globalDateSerialization,

                // TODO not in production, for speed improvement?
                unserializeWrapper: builtinWrappers.globalDateUnserializeValidation
            }),

            globalDateTimeFormatter: baseDateFmt.withOptions({
                format: globalDateTimeFormat,
                unserializeFormat: builtinFormats.globalDateTimeUnserialization,
                serializeFormat: builtinFormats.globalDateTimeSerialization,

                // TODO not in production, for speed improvement?
                unserializeWrapper: builtinWrappers.globalDateTimeUnseriailzeValidation
            }),

            globalTimeFormatter: baseDateFmt.withOptions({
                format: globalTimeFormat,
                unserializeFormat: builtinFormats.globalDateTimeUnserialization,
                serializeFormat: builtinFormats.globalDateTimeSerialization,

                // TODO not in production, for speed improvement?
                unserializeWrapper: builtinWrappers.globalDateTimeUnseriailzeValidation
            }),

            genericNumberFormatter: baseNumberFormatter.withOptions({
                workableDecimalPlaces: 6,

                format: positiveNumberFormat
            }),

            genericIntegerFormatter: baseNumberFormatter.withOptions({
                validateWrapper: builtinWrappers.integerValidation,
                workableDecimalPlaces: 0,

                format: positiveIntegerFormat
            }),

            bareIntegerFormatter: baseNumberFormatter.withOptions({
                validateWrapper: builtinWrappers.integerValidation,
                workableDecimalPlaces: 0,

                format: bareIntegerFormat
            }),

            percentFormatter: baseNumberFormatter.withOptions({
                workableDecimalPlaces: 4, // TODO what do we use for percents?

                formattedMultiplier: 100,

                // Doesn't do negatives in P6, so probably doesn't need any special handling
                //  (though maybe validate() should return false for negatives?)
                format: '0,0.0$',
                symbol: '%',

                decimalsOptional: true
            }),

            costFormatter: baseNumberFormatter.withOptions({
                workableDecimalPlaces: 5,

                format: positiveCurrencyFormat,
                negativeFormat: negativeCurrencyFormat,
                symbol: currencySymbol
            }),

            durationFormatter: durFmt.withOptions(durationFormatterOptions),
        };

        // See http://10.140.183.183/redmine/issues/3604
        // Base the long term duration formatter off of the base duration formatter to auto-absorb changes to workable
        // decimal places, and separators
        formatters.longTermDurationFormatter = formatters.durationFormatter.withOptions({
            showIn : options.monthChar, // Show in months
            showAbbr : true,
            showSubUnits : false
        });

        // Alias
        formatters.cppDurationFormatter = formatters.longTermDurationFormatter;

        return formatters;
    }

    return formatterFactory;

});

